/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.io.File;
import java.io.IOException;

/**
 * This class implements utility methods for creating files (with engine results).
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class FileCreator {

  // The default directory name.
  private static String defaultDirectory;

  private FileCreator() {
    /* Empty. */
  }

  /**
   * Returns the name of the default directory.
   *
   * @return the default directory name.
   */
  public static String getDefaultDirectory() {
    return defaultDirectory;
  }

  /**
   * Sets the name of the default directory.
   *
   * @param dirName the name to be set.
   */
  public static void setDefaultDirectory(final String dirName) {
    defaultDirectory = dirName;
  }

  /**
   * Creates a new file with the given name.
   * <p>
   * If a file with the specified name exists, no new file will be created.
   * </p>
   * @param fileName the file name.
   * @return the reference to the created file.
   */
  public static File newFile(final String fileName) {
    final String path = new File(fileName).isAbsolute() || defaultDirectory == null ? fileName :
        String.format("%s%s%s", defaultDirectory, File.separator, fileName);

    final File file = new File(path);

    try {
      if (!file.exists() && !file.createNewFile()) {
        Log.error("%s'%s'.", LogMessage.CANT_CREATE_FILE, file.toPath().toString());
      }
    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }

    return file;
  }
}
