/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import java.io.PrintWriter;

/**
 * GraphML utils.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public final class GraphMlUtils {

  /* GraphML figures. */
  public static final String FORM_TRIANGLE = "triangle";
  public static final String FORM_ELLIPSE = "ellipse";
  public static final String FORM_DIAMOND = "diamond";
  public static final String FORM_PARALLELOGRAM = "parallelogram";
  public static final String FORM_RECTANGLE = "rectangle";
  public static final String FORM_OCTAGON = "octagon";
  public static final String FORM_TRAPEZOID = "trapezoid";
  public static final String FORM_TRAPEZOID2 = "trapezoid2";

  /* GraphML colors. */
  private static final String COLOR_WHITE = "#ffffff";
  public static final String COLOR_YELLOW = "#ffcc00";
  public static final String COLOR_PURPLE = "#ccccff";
  public static final String COLOR_GREEN = "#008000";
  public static final String COLOR_RED = "#ff0000";
  public static final String COLOR_BLUE = "#0000ff";
  public static final String COLOR_ORANGE = "#ff8400";
  public static final String COLOR_GRAY = "#c0c0c0";
  public static final String COLOR_LIGHT_GRAY = "#cccccc";

  /* GraphML alignments. */
  public static final String ALIGN_LEFT = "left";
  public static final String ALIGN_RIGHT = "right";
  public static final String ALIGN_CENTER = "center";

  /* GraphML font styles. */
  public static final String FONT_BOLD = "bold";
  public static final String FONT_ITALIC = "italic";
  public static final String FONT_BOLDITALIC = "bolditalic";
  private static final String FONT_PLAIN = "plain";

  /* GraphML edge style types */
  public static final String STYLE_LINE = "line";
  public static final String STYLE_DASH = "dashed";


  // Cropping text in the node, if label size of text > node size
  public static final Boolean CROPPING_TEXT = true;

  private GraphMlUtils() {
    /* Empty. */
  }

  /**
   * Print the GraphML title tag to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphMlBegin(final PrintWriter printWriter) {
    // Write xml data
    printWriter.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    // Open GraphML tag
    printWriter.println("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\""
                        + " xmlns:y=\"http://www.yworks.com/xml/graphml\">");
    //
    printWriter.println("<key for=\"node\" id=\"d0\" yfiles.type=\"nodegraphics\"/>");
    printWriter.println("<key for=\"edge\" id=\"d1\" yfiles.type=\"edgegraphics\"/>");
  }

  /**
   * Print the GraphML close tag to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphMlEnd(final PrintWriter printWriter) {
    // Close GraphML tag
    printWriter.format("</graphml>%n");
  }

  /**
   * Print the graph open tag.
   *
   * @param printWriter the GraphML file writer.
   * @param modelId the model identifier.
   */
  public static void onGraphBegin(final PrintWriter printWriter, final String modelId) {
    printWriter.format("<graph id=\"model%s\" edgedefault=\"undirected\">%n", modelId);
  }

  /**
   * Print the graph open tag.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphBegin(final PrintWriter printWriter) {
    onGraphBegin(printWriter, "");
  }

  /**
   * Print the graph open tag.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphEnd(final PrintWriter printWriter) {
    printWriter.format("</graph>%n");
  }

  /**
   * Close group node.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void closeGroup(final PrintWriter printWriter) {
    printWriter.print("</node>");
  }

  /**
   * Print group node with this label.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeName the group node identifier.
   * @param nodeLabel the Label for group node.
   */
  public static void printGroup(
      final PrintWriter printWriter,
      final String nodeName,
      final String nodeLabel) {
    printGroup(printWriter, nodeName, nodeLabel, FORM_RECTANGLE, COLOR_WHITE);
  }

  /**
   * Prints group node with the specified identifier, label and color.
   *
   * @param writer The GraphML file writer.
   * @param nodeId The group node unique identifier.
   * @param nodeLabel The group node label.
   * @param color The group node color.
   */
  public static void printGroup(
      final PrintWriter writer,
      final String nodeId,
      final String nodeLabel,
      final String color) {
    printGroup(writer, nodeId, nodeLabel, FORM_RECTANGLE, color);
  }

  /**
   * Print group node with this label, form, color to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeName the group node identifier.
   * @param nodeLabel the Label for group node.
   * @param form the group node form.
   * @param color the group node color.
   */
  private static void printGroup(
      final PrintWriter printWriter,
      final String nodeName,
      final String nodeLabel,
      final String form,
      final String color) {
    printWriter.format("<node id=\"%s\" yfiles.foldertype=\"group\">%n", nodeName);
    printWriter.format("   <data key=\"d0\">%n");
    printWriter.format("      <y:GroupNode>%n");
    printWriter.format("           <y:Fill color=\"%s\"/>%n", color);
    printWriter.format("      <y:NodeLabel alignment=\"center\" autoSizePolicy=\"node_size\">"
                      + " <![CDATA[%s]]> </y:NodeLabel>%n", nodeLabel);
    printWriter.format("           <y:Shape type=\"%s\"/>%n", form);
    printWriter.format("       </y:GroupNode>%n");
    printWriter.format("   </data>%n");
  }

  /**
   * Print node.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeName the node identifier.
   */
  public static void printNode(final PrintWriter printWriter, final String nodeName) {
    printWriter.format("<node id=\"%s\"/>%n", nodeName);
  }

  /**
   * Print node with this label, form, color to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeName the node identifier.
   * @param nodeLabel the Label for node.
   * @param form the node form.
   * @param color the node color.
   */
  public static void printNode(
      final PrintWriter printWriter,
      final String nodeName,
      final String nodeLabel,
      final String form,
      final String color) {
    printNode(printWriter, nodeName, nodeLabel, form, color, false);
  }

  /**
   * Print node with this cropping label, form, color to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeName the node identifier.
   * @param nodeLabel the Label for node.
   * @param form the node form.
   * @param color the node color.
   * @param croppingText flag of cropping text.
   */
  public static void printNode(
      final PrintWriter printWriter,
      final String nodeName,
      final String nodeLabel,
      final String form,
      final String color,
      final boolean croppingText) {
    String croppingLabel = "";
    if (croppingText) {
      croppingLabel = " autoSizePolicy=\"node_size\" configuration=\"CroppingLabel\"";
    }
    printWriter.format("<node id=\"%s\">%n", nodeName);
    printWriter.format("   <data key=\"d0\">%n");
    printWriter.format("      <y:ShapeNode>%n");
    printWriter.format("           <y:Fill color=\"%s\"/>%n", color);
    printWriter.format("      <y:NodeLabel%s> <![CDATA[%s]]> </y:NodeLabel>%n", croppingLabel,
        nodeLabel);
    printWriter.format("           <y:Shape type=\"%s\"/>%n", form);
    printWriter.format("       </y:ShapeNode>%n");
    printWriter.format("   </data>%n");
    printWriter.format("</node>%n");
  }

  /**
   * Print edge to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   * @param nodeSourceName the source node identifier.
   * @param nodeTargetName the target node identifier.
   */
  public static void printEdge(
      final PrintWriter printWriter,
      final String nodeSourceName,
      final String nodeTargetName) {
    printWriter.format("<edge source=\"%s\" target=\"%s\"/>%n", nodeSourceName, nodeTargetName);
  }

  /**
   * Print edge with label to GraphML file.
   *
   * @param printWriter the GraphML file writer.
   * @param srcId the source node identifier.
   * @param edgeStyle the style of edge
   * @param dstId the target node identifier.
   */
  public static void printEdge(
      final PrintWriter printWriter,
      final String srcId,
      final String edgeStyle,
      final String dstId) {
    printEdge(
        printWriter,
        srcId,
        edgeStyle,
        "",
        GraphMlUtils.ALIGN_CENTER,
        GraphMlUtils.FONT_PLAIN,
        dstId);
  }

  /**
   * Print edge with the specified parameters to GraphML file.
   * @param printWriter GraphML file writer
   * @param srcId source node identifier
   * @param edgeStyle edge style
   * @param edgeLabel edge label
   * @param dstId destination node identifier
   */
  public static void printEdge(
      final PrintWriter printWriter,
      final String srcId,
      final String edgeStyle,
      final String edgeLabel,
      final String dstId) {
    printEdge(
        printWriter,
        srcId,
        edgeStyle,
        edgeLabel,
        GraphMlUtils.ALIGN_CENTER,
        GraphMlUtils.FONT_PLAIN, dstId);
  }

  /**
   * Prints edge with the specified parameters to GraphML file.
   * <p>Prints edge from the source node with the specified {@code sourceId} identifier
   * to the destination node with the specified {@code dstId}. The edge will be labeled with
   * the specified {@code label} that is aligned in the specified way and has the specified
   * font style (bold, italic, plain, etc.). The edge will be printed as the line of the specified
   * style to the file that the specified writer is set to.</p>
   * @param printWriter the GraphML file writer
   * @param srcId source node identifier
   * @param styleType edge style
   * @param label edge label
   * @param labelAlignment edge label alignment
   * @param labelFontStyle edge label font style
   * @param dstId destination node identifier
   */
  public static void printEdge(
      final PrintWriter printWriter,
      final String srcId,
      final String styleType,
      final String label,
      final String labelAlignment,
      final String labelFontStyle,
      final String dstId) {
    printWriter.format("<edge source=\"%s\" target=\"%s\">%n", srcId, dstId);
    printWriter.format("    <data key=\"d1\">%n");
    printWriter.format("        <y:PolyLineEdge>%n");
    printWriter.format("        <y:LineStyle type=\"%s\" />", styleType);
    printWriter.format(
        "        <y:EdgeLabel alignment=\"%s\" fontStyle=\"%s\"> <![CDATA[%s]]> </y:EdgeLabel>%n",
        labelAlignment,
        labelFontStyle,
        label);
    printWriter.format("        <y:Arrows source=\"none\" target=\"standard\"/>%n");
    printWriter.format("        </y:PolyLineEdge>%n");
    printWriter.format("    </data>%n");
    printWriter.format("</edge>%n");
  }

  /**
   * Close graphml file.
   *
   * @param printWriter the GraphML file writer.
   */
  public static void closeGraphMlFile(final PrintWriter printWriter) {
    printWriter.close();
  }
}
