/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Backend;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgBfsWalker;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.util.LogMessage;

/**
 * Base class for control flow graph model visitor-based back-ends.
 *
 * <p>Base class for {@link Backend} back-ends that are
 * based on {@link CfgVisitor} control flow graph model visitor.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgVisitorBackend extends Backend {

  private final CfgVisitor visitor;
  private final WalkerMode mode;

  /**
   * Creates object with the specified model visitor.
   *
   * @param name The backend name.
   * @param visitor Model visitor.
   * @param mode Mode of walking.
   *
   * @throws IllegalArgumentException when mode argument is {@code null}.
   */
  public CfgVisitorBackend(
      final String name,
      final CfgVisitor visitor,
      final WalkerMode mode) {
    super(name);
    InvariantChecks.checkNotNull(mode);

    this.visitor = visitor == null ? new CfgEmptyVisitor() : visitor;
    this.mode = mode;
  }

  /**
   * Elaborates the specified entity.
   *
   * @param entity Entity to be elaborated.
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified entity is neither
   *         of {@link ru.ispras.retrascope.model.cfg.CfgModel} model
   *         nor of {@link GaddModel} model kinds.
   */
  @Override
  public void start(final Entity entity) {
    InvariantChecks.checkNotNull(entity);

    Model model;

    if (entity.getEntityType() == CfgModel.TYPE) {

      model = (CfgModel) entity;

    } else if (entity.getEntityType() == GaddModel.TYPE) {

      model = (GaddModel) entity;

    } else {
      throw RetrascopeException.exception(LogMessage.WRONG_TYPE, entity.getEntityType());
    }

    final ModelWalker walker =
        getMode().equals(WalkerMode.BFS)
            ? new CfgBfsWalker(getVisitor()) : new CfgDfsWalker(getVisitor(), getMode());
    walker.start(model.getEntityType() == GaddModel.TYPE ? ((GaddModel)model).getModel() : model);
  }

  /**
   * Returns the backend's internal visitor.
   * @return The model visitor.
   */
  protected CfgVisitor getVisitor() {
    return this.visitor;
  }

  private WalkerMode getMode() {
    return this.mode;
  }
}
