/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.generator.test.random;

import org.apache.commons.cli.ParseException;
import ru.ispras.castle.ast.IntegerRange;
import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.randomizer.Randomizer;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.basis.TestGeneratorEngine;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.io.File;
import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Random test generator for Control Flow Graph model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRandomTestGenerator extends TestGeneratorEngine {

  /**
   * Engine identifier.
   */
  public static final String ENGINE_ID = "cfg-random-test-generator";

  /**
   * Generated test sequence length.
   */
  public static final Parameter TEST_LEN =
      new Parameter("test-len", true, "Length of generated test sequence.");

  /**
   * Random seed value.
   */
  public static final Parameter TEST_SEED = new Parameter("test-seed", true, "Test seed value.");

  /**
   * Name of clock signal.
   */
  public static final Parameter CLK_NAME = new Parameter("clk-name", true, "Name of clock signal.");

  /**
   * Active level of clock signal.
   */
  public static final Parameter CLK_LVL =
      new Parameter("clk-lvl", true, "Clock signal active level");

  /**
   * Name of reset signal.
   */
  public static final Parameter RST_NAME = new Parameter("rst-name", true, "Name of reset signal.");

  /**
   * Active level of reset signal.
   */
  public static final Parameter RST_LVL =
      new Parameter("rst-lvl", true, "Reset signal active level");

  /**
   * Delay in ticks through which the reset signal should be kept active for completed resetting.
   */
  public static final Parameter RST_DELAY =
      new Parameter("rst-delay", true, "Reset signal delay.");

  /* Input signal names and their possible value ranges. */
  public static final Parameter INPUT_VALUES =
      new Parameter("input-values", true, "Input signals and their possible values.");

  class ClockInfo {

    private final String name;
    private final int activeLevel;

    ClockInfo(
        final String name,
        final int level) {

      this.name = name;
      this.activeLevel = level;
    }

    String getName() {
      return this.name;
    }

    int getLevel() {
      return this.activeLevel;
    }
  }

  class ResetInfo extends ClockInfo {

    private final int delay;

    ResetInfo(
        final String name,
        final int level,
        final int delay) {
      super(name, level);

      this.delay = delay;
    }

    int getDelay() {
      return this.delay;
    }
  }

  /**
   * Default constructor.
   */
  public CfgRandomTestGenerator() {
    super(ENGINE_ID, CfgInterface.TYPE);

    addParameter(TEST_LEN);
    addParameter(TEST_SEED);
    addParameter(CLK_NAME);
    addParameter(CLK_LVL);
    addParameter(RST_NAME);
    addParameter(RST_LVL);
    addParameter(RST_DELAY);
    addParameter(INPUT_VALUES);
  }

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    final CfgInterface cfgInterface = (CfgInterface) inputs.get(CfgInterface.TYPE);
    final TestModel testModel = new TestModel();

    try {

      final Map<String, DeclarationContainer> signalInterfaces = cfgInterface.getMapping();

      if (signalInterfaces.isEmpty()) {
        return testModel;
      }

      final Arguments arguments =
          parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());

      Randomizer.get().setSeed(Integer.parseInt(arguments.getValue(TEST_SEED, "0")));

      final Map<String, IntegerRange> inputValues =
          getInputValues(arguments.getValue(INPUT_VALUES, ""), signalInterfaces.values());

      for (final Map.Entry<String, DeclarationContainer> entry : signalInterfaces.entrySet()) {

        final DeclarationContainer moduleIface = entry.getValue();

        /* Process invariants that are of "integer range" kind. */
        final Collection<NodeVariable> inVariables = moduleIface.getInputs();

        for (final NodeVariable inVariable : inVariables) {

          final String name = inVariable.getName();
          final Declaration declaration = moduleIface.getDeclaration(inVariable);
          final Range invariant = declaration.getInvariant();

          if (invariant != null && invariant.isIntegerRange()) {

            final IntegerRange invRange = invariant.toIntegerRange();

            if (inputValues.containsKey(name)) {

              final IntegerRange inputRange = inputValues.get(name);

              if (inputRange.overlaps(invRange)) {

                if (inputRange.contains(invRange) && !inputRange.equals(invRange)) {

                  Log.warn("Wrong range for %s variable: substituted by %s.", name, invRange);
                  inputValues.remove(name);
                  inputValues.put(name, invRange);

                } else if (!invRange.contains(inputRange)) {

                  final IntegerRange newRange = inputRange.intersect(invRange);
                  inputValues.put(name, newRange);
                }
              } else {
                Log.error(
                    "Wrong range for %s variable: must be overlapped with %s.",
                    name,
                    invRange);
              }
            } else {
              inputValues.put(name, invRange);
            }
          }
        }

        final Test test = new Test();
        testModel.addTest(test);

        final Sequence testSequence = new Sequence();
        test.addSequence(testSequence);

        final ClockInfo clockInfo =
            new ClockInfo(
                arguments.getValue(CLK_NAME, "clk"),
                Integer.parseInt(arguments.getValue(CLK_LVL, "1")));

        final ResetInfo resetInfo =
            new ResetInfo(
                arguments.getValue(RST_NAME, "rst"),
                Integer.parseInt(arguments.getValue(RST_LVL, "1")),
                Integer.parseInt(arguments.getValue(RST_DELAY, "1")));

        addResetVectors(testSequence, moduleIface, clockInfo, resetInfo);
        final int testLength = Integer.parseInt(arguments.getValue(TEST_LEN));

        for (int i = 0; i < testLength; i++) {
          addRandomTestVector(testSequence, moduleIface, clockInfo, inputValues);
        }
      }

    } catch (final ParseException e) {
      throw new RetrascopeException(e);
    }
    return testModel;
  }

  private Map<String, IntegerRange> getInputValues(
      final String value,
      final Collection<DeclarationContainer> interfaces) {

    final Map<String, IntegerRange> inputValues = new LinkedHashMap<>();

    if (value.length() > 0) {

      final String[] pairs = value.split(File.pathSeparator);

      for (final String inputValuePair : pairs) {

        final String[] keyValue = inputValuePair.split("=");
        final String signalName = keyValue[0];
        final String values = keyValue[1];
        final DataType signalType = getDataType(signalName, interfaces);

        if (signalType == null) {
          continue;
        }

        final IntegerRange range = getRange(values, signalType);
        inputValues.put(signalName, range);
      }
    }
    return inputValues;
  }

  /* String 'values' has the following form: "[x,y,radix]" */
  private IntegerRange getRange(
      final String values,
      final DataType signalType) {

    final int firstBracketIndex = values.indexOf('[');
    final int lastBracketIndex = values.indexOf(']');

    final String noBracket = values.substring(firstBracketIndex + 1, lastBracketIndex);
    final String[] ranges = noBracket.split(",");
    final int radix = Integer.parseInt(ranges[2]);

    final Data min = signalType.valueOf(ranges[0], radix);
    final Data max = signalType.valueOf(ranges[1], radix);

    return new IntegerRange(min.getInteger(), max.getInteger());
  }

  private DataType getDataType(
      final String signalName,
      final Collection<DeclarationContainer> interfaces) {

    for (final DeclarationContainer signalInterface : interfaces) {

      if (signalInterface.containsInput(signalName)
          || signalInterface.containsOutput(signalName)) {

        return signalInterface.getVariable(signalName).getDataType();
      }
    }
    return null;
  }

  private void addResetVectors(
      final Sequence sequence,
      final DeclarationContainer signalInterface,
      final ClockInfo clockInfo,
      final ResetInfo resetInfo) {

    if (!containsSignal(signalInterface, resetInfo.getName())) {
      return;
    }

    final String clkName = clockInfo.getName();
    final NodeVariable clkVariable = getVariable(signalInterface, clkName);
    final boolean clkExists = clkVariable != null;
    final Event enableClockEvent = getClkEvent(clkVariable, clockInfo.getLevel());

    /* Activate a reset for one tick. */
    final Transaction enReset = new Transaction();
    final String rstName = resetInfo.getName();
    final NodeValue rstActiveValue = NodeValue.newInteger(resetInfo.getLevel());
    enReset.addVariable(rstName, rstActiveValue);
    addZeroOther(enReset, signalInterface, clkName, rstName);
    final TestVector enResetVector = new TestVector(enReset);

    if (clkExists) {
      enResetVector.addEvent(enableClockEvent);
    }

    sequence.addVector(enResetVector);

    /* Make a flip-flop till delay. */
    for (int i = 0; i < resetInfo.getDelay() - 1; i++) {

      final Transaction waitTransaction = new Transaction();
      addZeroOther(waitTransaction, signalInterface, clkName, rstName);
      final TestVector enableClkVector = new TestVector(waitTransaction);

      if (clkExists) {
        enableClkVector.addEvent(enableClockEvent);
      }

      sequence.addVector(enableClkVector);
    }

    /* Deactivate reset. */
    final Transaction disableReset = new Transaction();
    disableReset.addVariable(
        rstName,
        NodeValue.newInteger(resetInfo.getLevel() == 0 ? 1 : 0));
    addZeroOther(disableReset, signalInterface, clkName, rstName);
    final TestVector disable = new TestVector(disableReset);

    if (enableClockEvent != null) {
      disable.addEvent(enableClockEvent);
    }

    sequence.addVector(disable);
  }

  private void addZeroOther(
      final Transaction transaction,
      final DeclarationContainer signalInterface,
      final String ... otherNames) {

    final Collection<NodeVariable> inputs = signalInterface.getInputs();

    for (final NodeVariable input : inputs) {

      if (!hasNameAmong(input, otherNames)) {

        transaction.addVariable(input.getName(), getZeroValue(input.getDataType()));
      }
    }
  }

  private boolean hasNameAmong(
      final NodeVariable variable,
      final String ... names) {

    for (final String name : names) {

      if (variable.getName().equals(name)) {

        return true;
      }
    }
    return false;
  }

  private Event getClkEvent(
      final NodeVariable clkVariable,
      final int activeLevel) {

    if (clkVariable == null) {
      return null;
    }

    final EventType type = activeLevel == 0 ? EventType.NEGATIVE_EDGE : EventType.POSITIVE_EDGE;

    return new Event(clkVariable, type);
  }

  private NodeVariable getVariable(
      final DeclarationContainer signalInterface,
      final String name) {

    for (final NodeVariable inputSignal : signalInterface.getInputs()) {

      if (inputSignal.getName().equals(name)) {
        return inputSignal;
      }
    }
    return null;
  }

  private void addRandomTestVector(
      final Sequence sequence,
      final DeclarationContainer moduleInterface,
      final ClockInfo clockInfo,
      final Map<String, IntegerRange> permissible) {

    final Transaction transaction = new Transaction();

    for (final NodeVariable input : moduleInterface.getInputs()) {

      transaction.addVariable(input.getName(), getRandomValue(input, permissible, moduleInterface));
    }

    final TestVector vector = new TestVector(transaction);
    final NodeVariable clkVariable = getVariable(moduleInterface, clockInfo.getName());
    final boolean clkExists = clkVariable != null;

    if (clkExists) {
      vector.addEvent(getClkEvent(clkVariable, clockInfo.getLevel()));
    }
    sequence.addVector(vector);
  }

  private NodeValue getRandomValue(
      final NodeVariable variable,
      final Map<String, IntegerRange> permissible,
      final DeclarationContainer moduleInterface) {

    final String varName = variable.getName();
    final boolean hasPerm = permissible.containsKey(varName);

    BigInteger min = null;
    BigInteger max = null;

    if (hasPerm) {

      min = permissible.get(varName).getMin();
      max = permissible.get(varName).getMax();
    }

    NodeValue rndValue = getRandom(variable, hasPerm, min, max);
    final Node invariant = moduleInterface.getInvariant(varName);

    if (invariant == null) {

      return rndValue;

    } else {

      Node equation = Nodes.eq(variable, rndValue);
      boolean compatible = SolverUtils.areCompatible(equation, invariant);

      while (!compatible) {

        rndValue = getRandom(variable, hasPerm, min, max);
        equation = Nodes.eq(variable, rndValue);
        compatible = SolverUtils.areCompatible(equation, invariant);
      }
      return rndValue;
    }
  }

  private NodeValue getRandom(
      final NodeVariable variable,
      final boolean hasPerm,
      final BigInteger min,
      final BigInteger max) {

    final DataType dataType = variable.getDataType();

    switch (dataType.getTypeId()) {

      case BIT_VECTOR:
        return getRndBitVector(hasPerm, min, max, dataType);

      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(Randomizer.get().nextBoolean());

      case LOGIC_INTEGER:
        return hasPerm
            ? NodeValue.newInteger(Randomizer.get().nextBigIntegerRange(min, max))
            : NodeValue.newInteger(Randomizer.get().next());

      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, dataType);
    }
  }

  private NodeValue getRndBitVector(
      final boolean hasPermValues,
      final BigInteger min,
      final BigInteger max,
      final DataType dataType) {

    final BitVector bitVector = BitVector.newEmpty(dataType.getSize());
    Randomizer.get().fill(bitVector);

    return NodeValue.newBitVector(
        hasPermValues
            ? BitVector.valueOf(
                Randomizer.get().nextBigIntegerRange(min, max),
                dataType.getSize())
            : bitVector);
  }

  private NodeValue getZeroValue(final DataType dataType) {

    switch (dataType.getTypeId()) {

      case BIT_VECTOR:
        return NodeValue.newBitVector(0, dataType.getSize());

      case LOGIC_BOOLEAN:
        return (NodeValue) Nodes.FALSE;

      case LOGIC_INTEGER:
        return NodeValue.newInteger(0);

      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, dataType);
    }
  }

  private boolean containsSignal(
      final DeclarationContainer signalInterface,
      final String signalName) {

    for (final NodeVariable inputSignal : signalInterface.getInputs()) {

      if (inputSignal.getName().equals(signalName)) {

        return true;
      }
    }
    return false;
  }
}
