/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import org.apache.commons.cli.ParseException;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.cfg.CfgVisitorBackend;
import ru.ispras.retrascope.engine.cfg.transformer.gadd.backend.GaddClockDetectBackend;
import ru.ispras.retrascope.engine.cfg.transformer.gadd.backend.GaddPathCounterBackend;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * Transformer from {@link CfgModel} to {@link GaddModel} model.
 *
 * <p>The transformer of control flow graph (CFG)
 * to guarded actions decision diagram (GADD).</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgGaddTransformer extends Engine {

  /**
   *  The engine identifier.
   */
  public static final String ENGINE_ID = "cfg-gadd-transformer";

  /**
   * The 'clock variable name' command line parameter.
   */
  public static final Parameter CLV_NAME =
      new Parameter("clk-name", true, "Name of clock variable.");

  private GaddClockDetectBackend clockDetector;

  /**
   * Default constructor.
   */
  public CfgGaddTransformer() {
    super(CfgGaddTransformer.ENGINE_ID);

    addInputType(CfgModel.TYPE);
    setOutputType(GaddModel.TYPE);
    addParameter(CLV_NAME);

    this.clockDetector = new GaddClockDetectBackend();
    addBackend(
        new CfgVisitorBackend(
            GaddClockDetectBackend.NAME, this.clockDetector, WalkerMode.DFS_NO_RPT));
    addBackend(new GaddPathCounterBackend());
  }

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    /* Walk the model with the specified visitor. */
    final CfgGaddTransformerVisitor visitor = new CfgGaddTransformerVisitor();
    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS_NO_RPT);

    walker.start((CfgModel) inputs.get(CfgModel.TYPE));

    /* Build an atomic actions model. */
    final CfgGaddTransformerBuilder builder = new CfgGaddTransformerBuilder(visitor.getModuleMap());
    final GaddModel model = builder.build();

    /* Elaborate command line parameters. */
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      if (engineArgs.contains(CLV_NAME)) {

        final String clockName = engineArgs.getValue(CLV_NAME);

        NodeVariable clockVar = null;
        String moduleName = null;

        for (final Module module : model.getModel().getModules()) {

          if (module.declares(clockName)) {
            clockVar = module.getVariable(clockName);
            moduleName = module.getName();
            break;
          }
        }

        if (clockVar == null) {
          Log.warn("Can't find clock variable with name: '%s'; an heuristic applied.", clockName);
          this.clockDetector.setModel(model);
        } else {
          model.addClock(moduleName, clockVar);
          deactivateBackends(GaddClockDetectBackend.NAME);
        }
      } else {
        this.clockDetector.setModel(model);
      }
    } catch (final ParseException e) {
      throw new RetrascopeException(e);
    }

    return model;
  }
}
