/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.model.gadd.Phase;
import ru.ispras.retrascope.util.Log;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Builder for guarded actions diagram.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgGaddTransformerBuilder {

  /* The mapping of new module to old processes. */
  private final Map<CfgModule, List<CfgProcess>> moduleMap;

  private final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processPhaseMap;

  /**
   * Constructs object with the specified mapping.
   *
   * @param map Mapping of modules to their processes.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  CfgGaddTransformerBuilder(final Map<CfgModule, List<CfgProcess>> map) {
    InvariantChecks.checkNotNull(map);

    this.moduleMap = map;
    this.processPhaseMap = new LinkedHashMap<>();
  }

  private Map<CfgModule, List<CfgProcess>> getModuleMap() {
    return Collections.unmodifiableMap(this.moduleMap);
  }

  /**
   * Creates {@link GaddModel} model.
   *
   * <p>Builds guarded actions decision diagram
   * from {@link CfgModel} control flow graph model representation.</p>
   *
   * @return Guarded actions decision diagram.
   */
  public GaddModel build() {

    final CfgModel cfgModel = new CfgModel();

    for (final CfgModule module : getModuleMap().keySet()) {

      final CfgModule gaddModule = buildModule(module);

      if (module.getParent() == null) {

        cfgModel.addModule(gaddModule);
      }
    }

    return new GaddModel(cfgModel, this.processPhaseMap);
  }

  private CfgModule buildModule(final CfgModule module) {

    /* Apply transformations of the current module's process. */
    final List<CfgProcess> processes = getModuleMap().get(module);

    for (final CfgProcess process : processes) {

      final int processNum = processes.indexOf(process);
      applyToProcess(module, process, processNum);
    }

    return module;
  }

  private void applyToProcess(final CfgModule module, final CfgProcess cfgProcess, final int num) {

    /* Create new process container. */
    final CfgProcess process = new CfgProcess(cfgProcess);
    module.add(process);

    final CfgProcessPhasePlacer phasePlacer = new CfgProcessPhasePlacer(cfgProcess);

    /* Phase variable. */

    final int maxPhase = phasePlacer.getMaxPhase();
    final int phaseVarSize = (int) (Math.log(maxPhase) / Math.log(2)) + 1;

    final String prefix = module.getParent() == null ? "" : module.getName() + "_";
    final String processId = String.format("%sP%d_PHASE", prefix, num);
    final NodeVariable phaseVar = NodeVariable.newBitVector(processId, phaseVarSize);
    process.addMetaInfo(MetaInfoType.PHASE_VAR_NAME, phaseVar.getName());

    final VariableDescriptor phaseDescriptor = new VariableDescriptor(VariableType.REG);
    phaseVar.setUserData(phaseDescriptor);

    final NodeValue phaseInit = NodeValue.newBitVector(0, phaseVarSize);
    final NodeValue phaseMax = NodeValue.newBitVector(maxPhase, phaseVarSize);
    final Range invariant = new Range(phaseInit, phaseMax);
    module.declareVariable(phaseVar, phaseInit, invariant);

    /* Associate phases with old process statements. */

    final Graph<Statement, PhaseEdge> phaseGraph = phasePlacer.getPhaseGraph();

    final Graph<Phase, PhaseSeqEdge> phaseSeqGraph = makePhaseSeqGraph(phaseGraph);
    this.processPhaseMap.put(process, phaseSeqGraph);

    final CfgSwitchStatement phaseVarSwitch = new CfgSwitchStatement(phaseVar);
    process.add(phaseVarSwitch);

    final Set<NodeValue> terminalPhases = new LinkedHashSet<>();

    final Collection<Statement> vertices = phaseGraph.getVertices();

    for (final Statement vertex : vertices) {

      final Collection<PhaseEdge> outEdges = phaseGraph.getOutEdges(vertex);

      /* Calculate possible values of phase variable to reach this switch. */
      final Set<NodeValue> phaseValues = new LinkedHashSet<>();
      final Collection<PhaseEdge> inEdges = phaseGraph.getInEdges(vertex);

      for (final PhaseEdge inEdge : inEdges) {

        final Phase inEdgePhase = inEdge.getPhase();
        phaseValues.add(NodeValue.newBitVector(inEdgePhase.getValue(), phaseVarSize));
      }

      final boolean isProcess = vertex.equals(cfgProcess);

      if (isProcess || vertex instanceof CfgBlockStatement) {

        if (outEdges.size() != 1) {
          Log.error("Wrong number of out edges for '%s': %d\n", vertex.toString(), outEdges.size());
          Log.error("The edge destinations are:");

          for (final PhaseEdge edge : outEdges) {

            Log.error("%s\n", phaseGraph.getDest(edge));
          }
        }

        final PhaseEdge onlyEdge = outEdges.iterator().next();

        final Phase phase = onlyEdge.getPhase();
        final List<Statement> blocks =
            isProcess ? onlyEdge.getPathBlocks() : Collections.singletonList(vertex);

        final NodeValue newPhaseVal = NodeValue.newBitVector(phase.getValue(), phaseVarSize);
        final EventList processEvents = cfgProcess.getEventList();
        final CfgCaseStatement phaseCase = isProcess
            ? new CfgCaseStatement(processEvents, phaseInit)
            : new CfgCaseStatement(phaseValues);

        for (final Statement block : blocks) {
          phaseCase.add(block.deepCopy());
        }

        phaseCase.add(new CfgBlockStatement(phaseVar, newPhaseVal));
        phaseVarSwitch.add(phaseCase);

        final Statement onlyEdgeDst = phaseGraph.getDest(onlyEdge);
        final boolean processIsDst = onlyEdgeDst.equals(cfgProcess);

        if (processIsDst) {
          terminalPhases.add(newPhaseVal);
        }
      } else {

        final CfgSwitchStatement switchVertex = (CfgSwitchStatement) vertex;
        final CfgCaseStatement phaseCase = new CfgCaseStatement(phaseValues);
        phaseVarSwitch.add(phaseCase);

        final CfgSwitchStatement newSwitch = new CfgSwitchStatement(switchVertex.getCondition());
        newSwitch.addMetaInfo(switchVertex.getAllMetaInfo());
        phaseCase.add(newSwitch);

        for (final PhaseEdge outEdge : outEdges) {
          final CfgCaseData data = outEdge.getCaseData();
          final Phase edgePhase = outEdge.getPhase();
          final List<Statement> blocks = outEdge.getPathBlocks();
          final CfgCaseStatement newCase =
              (data == null || data.isValueFree())
                  ? new CfgCaseStatement(true)
                      : data.hasEvents()
                          ? new CfgCaseStatement(data.getEventList(), data.getValues())
                              : new CfgCaseStatement(data.getValues());

          for (final Statement block : blocks) {
            newCase.add(block.deepCopy());
          }

          final NodeValue phase = NodeValue.newBitVector(edgePhase.getValue(), phaseVarSize);
          newCase.add(new CfgBlockStatement(phaseVar, phase));
          newSwitch.add(newCase);

          final Statement dst = phaseGraph.getDest(outEdge);
          final boolean processIsDst = dst.equals(cfgProcess);
          if (processIsDst) {
            terminalPhases.add(phase);
          }
        }
      }
    }

    final CfgCaseStatement terminalCase = new CfgCaseStatement(terminalPhases);
    final CfgBlockStatement resetPhaseStmt = new CfgBlockStatement(phaseVar, phaseInit);
    terminalCase.add(resetPhaseStmt);
    phaseVarSwitch.add(terminalCase);

    CfgSsaTransformer.toSsa(module, process);
  }

  private Graph<Phase, PhaseSeqEdge> makePhaseSeqGraph(final Graph<Statement, PhaseEdge> graph) {

    final Graph<Phase, PhaseSeqEdge> phaseSeqGraph = new DirectedSparseMultigraph<>();

    final Collection<PhaseEdge> edges = graph.getEdges();

    for (final PhaseEdge edge : edges) {

      final Phase phase = makePhaseVertex(phaseSeqGraph, edge);

      final Statement src = graph.getSource(edge);
      final Collection<PhaseEdge> inEdges = graph.getInEdges(src);

      for (final PhaseEdge inEdge : inEdges) {

        final Phase inEdgePhase = makePhaseVertex(phaseSeqGraph, inEdge);
        phaseSeqGraph.addEdge(new PhaseSeqEdge(), inEdgePhase, phase, EdgeType.DIRECTED);
      }

      final Statement dst = graph.getDest(edge);
      final Collection<PhaseEdge> outEdges = graph.getOutEdges(dst);

      for (final PhaseEdge outEdge : outEdges) {

        final Phase outEdgePhase = makePhaseVertex(phaseSeqGraph, outEdge);
        phaseSeqGraph.addEdge(new PhaseSeqEdge(), phase, outEdgePhase, EdgeType.DIRECTED);
      }
    }

    return phaseSeqGraph;
  }

  private Phase makePhaseVertex(
      final Graph<Phase, PhaseSeqEdge> phaseSeqGraph,
      final PhaseEdge edge) {

    final Phase phase = edge.getPhase();

    if (!phaseSeqGraph.containsVertex(phase)) {
      phaseSeqGraph.addVertex(phase);
    }
    return phase;
  }


}
