/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The SSA transformer component that collects basic block statements in accordance with
 * their blocking\concurrent nature.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSsaBlockVisitor extends CfgEmptyVisitor {

  private final List<Assignment> blocking;
  private final List<Assignment> concurrent;

  CfgSsaBlockVisitor() {
    this.blocking = new LinkedList<>();
    this.concurrent = new LinkedList<>();
  }

  @Override
  public void onBlockEnd(final CfgBlockStatement block) {

    final Statement predecessor = StatementUtils.getPredecessor(block);
    if (predecessor.isType(CfgStatementType.BASIC_BLOCK)) {
      return;
    }

    this.blocking.clear();
    this.concurrent.clear();

    classify(block);

    /* Mapping of statements to be removed to their parents. */
    final Map<Statement, Statement> parentsOfRemoved = new LinkedHashMap<>();

    final Collection<Statement> blockSuccessors = StatementUtils.getSuccessors(block);

    if (!blockSuccessors.isEmpty()) {

      Statement current = blockSuccessors.iterator().next();

      while (current.isType(CfgStatementType.BASIC_BLOCK)) {

        final CfgBlockStatement currentBlock = (CfgBlockStatement) current;

        classify(currentBlock);

        parentsOfRemoved.put(current, current.getParent());

        final Collection<Statement> currentSuccessors = StatementUtils.getSuccessors(current);

        if (currentSuccessors.isEmpty()) {
          break;

        } else {
          current = currentSuccessors.iterator().next();
        }
      }
    }
    removeDescentsFromProcess(parentsOfRemoved);

    block.removeAssignments();

    if (this.blocking.isEmpty()) {
      block.addAssignments(this.concurrent);
      block.setConcurrent();
    } else {
      block.addAssignments(this.blocking);
      block.setBlocking();

      if (!this.concurrent.isEmpty()) {
        final CfgBlockStatement concurrent = new CfgBlockStatement(this.concurrent, true);
        block.getParent().add(concurrent);
      }
    }
  }

  private void classify(final CfgBlockStatement block) {
    if (block.isConcurrent()) {
      this.concurrent.addAll(block.getAssignments());
    } else {
      this.blocking.addAll(block.getAssignments());
    }
  }

  private void removeDescentsFromProcess(final Map<Statement, Statement> toBeRemoved) {

    for (final Map.Entry<Statement, Statement> entry : toBeRemoved.entrySet()) {
      entry.getValue().remove(entry.getKey());
    }
  }
}
