/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd.backend;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.util.Log;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Component that detects clock-like variables in the {@link GaddModel} model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class GaddClockDetectBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "clock_detector";

  /* The current model. */
  private GaddModel model;

  /* The clock-like variables for current process. */
  private Set<NodeVariable> clocks;

  /* The mapping from processes to their related clock-like variables. */
  private Map<ModuleProcess, Collection<NodeVariable>> processClockMap;

  /**
   * Sets the specified model as the target one.
   * @param model The target model.
   */
  public void setModel(final GaddModel model) {
    this.model = model;
  }

  @Override
  public void onModelBegin(final Model model) {

    this.processClockMap = new LinkedHashMap<>();
  }

  @Override
  public void onModuleEnd(final Module module) {

    /* Associate the most frequent signal as clock-like variable for the module. */
    final List<NodeVariable> clockList = new LinkedList<>();
    final Set<NodeVariable> clockSet = new LinkedHashSet<>();

    for (final Collection<NodeVariable> processClocks : this.processClockMap.values()) {
      clockList.addAll(processClocks);
      clockSet.addAll(processClocks);
    }

    final List<NodeVariable> clocks = new LinkedList<>(clockSet);

    final Comparator<NodeVariable> frequencyComparator = (first, second) -> {

      final int firstFreq = Collections.frequency(clockList, first);
      final int secondFreq = Collections.frequency(clockList, second);

      /* The most frequent variable should go earlier that the least one. */
      return Integer.compare(secondFreq, firstFreq);
    };

    clocks.sort(frequencyComparator);

    /* Print names for clock-like variables. */
    final List<String> names = new LinkedList<>();
    for (final NodeVariable var : clocks) {
      names.add(var.getName());
    }

    Log.info(
        "Clock-like variables (CLV): %s.",
        clocks.isEmpty() ? "<none>" : String.join(", ", names));

    if (clocks.isEmpty()) {
      return;
    }

    final NodeVariable clock = clocks.iterator().next();

    if (clocks.size() > 1) {

      Log.warn("More than one CLV detected; chose the most frequent one: '%s'", clock);
    }
    this.model.addClock(module.getName(), clock);

    this.processClockMap.clear();
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    this.clocks = new LinkedHashSet<>();

    for (final Event event : process.getEventList()) {

      final NodeVariable variable = event.getVariable();

      if (VariableType.isOneBitSignal(variable)) {
        this.clocks.add(variable);
      }
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    this.processClockMap.put(process, new LinkedHashSet<>(this.clocks));

    this.clocks.clear();
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {

    for (final Assignment statement : block.getAssignments()) {

      for (final NodeVariable stmtUse : statement.getUses()) {
        this.clocks.remove(stmtUse);
      }

      for (final NodeVariable stmtDefine : statement.getDefines()) {
        this.clocks.remove(stmtDefine);
      }
    }
  }
}
