/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.property;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmDfsWalker;
import ru.ispras.retrascope.model.efsm.EfsmEmptyVisitor;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.property.ModuleProperty;
import ru.ispras.retrascope.result.property.ProcessProperty;
import ru.ispras.retrascope.result.property.PropertyEntity;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Visitor for EfsmTransitionPropertyExtractor.
 *
 * @see EfsmTransitionPropertyExtractor
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class EfsmTransitionPropertyVisitor extends EfsmEmptyVisitor {

  private PropertyEntity entity;
  private ModuleProperty container;
  private ProcessProperty group;
  private Efsm efsm;

  /**
   * Returns object's output.
   *
   * @return output object that is built by visiting\walking
   */
  public PropertyEntity getOutput() {
    return this.entity;
  }

  private Efsm getEfsm() {
    return this.efsm;
  }

  @Override
  public void onModelBegin(final Model efsmModel) {
    this.entity = new PropertyEntity();
  }

  @Override
  public void onModuleBegin(final Module efsmModule) {

    if (this.container == null) {

      this.container = new ModuleProperty(efsmModule, efsmModule.getName(), null);

      /* Set clock-like variable for the top-level container. */
      final Collection<NodeVariable> clocks = getClocks((EfsmModule) efsmModule);

      if (!clocks.isEmpty()) {

        this.container.setClock(clocks.iterator().next());
      }

      this.entity.addContainer(this.container);

    } else {

      final ModuleProperty parent = this.container;
      this.container = new ModuleProperty(efsmModule, efsmModule.getName(), parent);
      parent.addModuleProperty(this.container);
    }
  }

  private Collection<NodeVariable> getClocks(final EfsmModule efsmModule) {

    class ClockCollector extends EfsmEmptyVisitor {
      private Set<NodeVariable> clocks;

      private Set<NodeVariable> resets;

      private ClockCollector() {
        this.clocks = new LinkedHashSet<>();
        this.resets = new LinkedHashSet<>();
      }

      private Collection<NodeVariable> getClocks() {

        return CollectionUtils.complementSets(this.clocks, this.resets);
      }

      @Override
      public void onProcessBegin(final ModuleProcess process) {

        final Efsm efsm = (Efsm) process;

        this.clocks.addAll(efsm.getClocks());

        final Collection<GuardedAction> resetActions = efsm.getResetGuardedActions();

        if (!resetActions.isEmpty()) {

          for (final GuardedAction resetAction : resetActions) {
            this.resets.addAll(ExprUtils.getVariables(resetAction.getGuardNode()));
          }
        }
      }
    }

    final ClockCollector collector = new ClockCollector();
    final ModelWalker walker = new EfsmDfsWalker(collector);
    walker.visitNode(efsmModule);

    return collector.getClocks();
  }

  @Override
  public void onModuleEnd(final Module efsmModule) {
    this.container = this.container.getParent();
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    this.efsm = (Efsm)process;

    final String efsmName = getEfsm().getName();
    final EfsmState initState = getEfsm().getInitialState();
    final Node initStateNode = initState.getNode();
    final List<GuardedAction> resets = getEfsm().getResetGuardedActions();

    this.group = resets.isEmpty()
        ? new ProcessProperty(efsmName, this.container, initStateNode)
        : new ProcessProperty(efsmName, this.container, initStateNode, resets);

    this.container.addProcessProperty(this.group);

    /* Set reset-like variable. */
    final Collection<GuardedAction> rstGuardedActions = getEfsm().getResetGuardedActions();

    if (!rstGuardedActions.isEmpty()) {

      final GuardedAction rstGuardedAction = rstGuardedActions.iterator().next();

      if (rstGuardedAction != null) {

        final NodeVariable resetSignal = getResetSignal(getEfsm(), rstGuardedAction);

        if (resetSignal != null) {
          setResetWhereDeclared(this.group, resetSignal);
        }
      }
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    if (getEfsm().singleTestable()) {

      final int efsmTransNum = getEfsm().getTransitionNum();
      final int assertNum = this.group.getPropertyNum();

      if (efsmTransNum != assertNum) {
        Log.warn(
            "EFSM transition num '%d' doesn't equal to property num '%d', probable dead code.",
            efsmTransNum,
            assertNum);
      }
    }
  }

  /* It is supposed that the guard depends on exactly one variable. */
  private NodeVariable getResetSignal(final Efsm efsm, final GuardedAction guardedAction) {

    final Node guardCondition = guardedAction.getGuardNode();
    final Collection<NodeVariable> uses = ExprUtils.getVariables(guardCondition);

    if (uses.size() == 1) {

      return uses.iterator().next();

    } else if (uses.size() > 1) {

      final String phaseVarName = efsm.getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);

      if (uses.size() == 2 && getNames(uses).contains(phaseVarName)) {

        for (final NodeVariable variable : uses) {

          final String varName = variable.getName();

          if (!varName.equals(phaseVarName)) {
            return variable;
          }
        }
      }
      throw new RetrascopeException(
          String.format(
              "Resetting guarded action cannot depend on more than one variable: %s.",
              Retrascope.getConfiguration().getStylePrinter().toString(guardCondition)));
    }
    return null;
  }

  private Collection<String> getNames(final Collection<NodeVariable> variables) {

    final Collection<String> names = new LinkedHashSet<>();

    for (final NodeVariable variable : variables) {
      names.add(variable.getName());
    }

    return names;
  }

  private void setResetWhereDeclared(
      final ProcessProperty group,
      final NodeVariable resetSignal) {

    ModuleProperty parentContainer = group.getParent();

    while (parentContainer != null) {

      if (parentContainer.containsVariable(resetSignal)) {

        parentContainer.setResetVariable(resetSignal);
        break;
      }

      parentContainer = parentContainer.getParent();
    }
  }

  @Override
  public void onTransitionBegin(final EfsmTransition transition) {

    final EfsmState srcState = transition.getSrc();
    final Node srcNode = srcState.getNode();

    final Integer srcPhase = srcState.getPhase();
    final Node srcPhaseNode = NodeValue.newInteger(srcPhase);
    final String phaseVarName = getEfsm().getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);
    final NodeVariable phaseVariable = getEfsm().getParentContainer().getVariable(phaseVarName);

    final Node phaseNode = TypeConversion.coerce(srcPhaseNode, phaseVariable.getDataType());
    final Node phaseEq = Nodes.eq(phaseVariable, phaseNode);

    final Node transitionGuard = transition.getClockGuardNode();
    final Node[] predicates = new Node[] {srcNode, phaseEq, transitionGuard};
    final Node conjunction = ExprUtils.getConjunction(predicates);
    final Node assertion = TransformerUtils.standardize(conjunction);

    /* Filter FALSE properties. */
    if (assertion.equals(Nodes.FALSE)) {

      Log.warn("FALSE condition for: %s.", transition.toString());
    } else {

      this.group.addProperty(assertion);
    }
  }
}
