/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.basis.TestGeneratorEngine;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * The test generator engine for EFSM-based hardware models.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmModel
 */
public final class EfsmTestGenerator extends TestGeneratorEngine {

  /**
   * The engine ID of all {@code EfsmTestGenerator} instances.
   */
  public static final String ENGINE_ID = "efsm-test-generator";

  /**
   * Engine's command line parameter encoding loop iteration limit.
   */
  public static final Parameter LOOP_LIMIT =
      new Parameter("loop-limit", true, "Loop iteration limit");

  private static final int DEFAULT_LIMIT = 1;

  private static final String LOG_ENTRY_HEADER = "EFSM.TestGenerator: ";

  /**
   * Constructs a generator.
   */
  public EfsmTestGenerator() {
    super(EfsmTestGenerator.ENGINE_ID, EfsmModel.TYPE);
    super.addParameter(EfsmTestGenerator.LOOP_LIMIT);
  }

  /**
   * Generates and returns tests for the specified EFSM modules.
   *
   * @param inputs a map containing the EFSM modules and command line arguments. It must contain a
   *        mapping from the {@link EfsmModel} entity type to an {@link EfsmModel} object.
   *        Optionally, it may include a mapping from the {@link Arguments} entity type to an
   *        {@link Arguments} object.
   *
   * @throws IllegalArgumentException if {@code inputs} is {@code null}
   * @throws IllegalArgumentException if {@code inputs} contains no mapping from the
   *         {@link EfsmModel} entity type to an {@link EfsmModel} object
   *
   * @return the generated tests
   */
  @Override
  public TestModel start(final Map<EntityType, Entity> inputs) {

    final EfsmModel efsmModel = Engine.getInputEntity(inputs, EfsmModel.class, EfsmModel.TYPE);

    int loopLimit = -1;

    final Arguments args = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    if (args != null) {
      try {

        final Arguments parsedArgs = super.parseCommandLine(args.getCommandLine());

        loopLimit = Integer.parseInt(parsedArgs.getValue(EfsmTestGenerator.LOOP_LIMIT, "-1"));

      } catch (final ParseException | NumberFormatException e) {
        throw new IllegalArgumentException("The exception has occurred while parsing command line "
            + "arguments", e);
      }
    }

    if (loopLimit == -1) {

      loopLimit = EfsmTestGenerator.DEFAULT_LIMIT;

      Log.info("%sloop limit isn't specified. Setting it to the default value: %s",
          EfsmTestGenerator.LOG_ENTRY_HEADER, loopLimit);
    } else {
      Log.debug("%sthe specified loop limit: %s", EfsmTestGenerator.LOG_ENTRY_HEADER, loopLimit);
    }

    final TestModel testModel = new TestModel();

    for (final Module module : efsmModel.getModules()) {

      final String logPrefix =
          String.format("%smodule %s: ", EfsmTestGenerator.LOG_ENTRY_HEADER, module.getName());

      Log.info("%sstarting test generation", logPrefix);

      final Generator generator = new Generator((EfsmModule)module, loopLimit);
      final Test test = generator.generateTest();
      testModel.addTest(test);

      Log.info("%s test generation has been finished successfully", logPrefix);
      final String covStr = generator.getCoverage().toString();
      Log.info("%stest coverage: %s", logPrefix, covStr);
      Log.info("%samount of sequences: %s", logPrefix, test.getSequences().size());

      int sequenceLength = 0;

      for (final Sequence sequence : test.getSequences()) {
        sequenceLength = sequenceLength + sequence.getLength();
      }

      Log.info("%samount of vectors: %s", logPrefix, sequenceLength);
    }

    return testModel;
  }
}
