/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.engine.basis.TestGenerator;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.engine.efsm.simulator.ModuleSimulator;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestCoverage;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The test generator for EFSM-based hardware models.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class Generator extends TestGenerator {

  private final String logEntryHeader;

  private final ModuleSimulator simulator;

  // TODO: this field is assigned but never accessed
  private final int loopLimit;

  private final EfsmTransitionCoverage totalCoverage;

  private final Map<Efsm, Set<List<EfsmTransition>>> traversedPaths;

  // TODO: this field is assigned but never accessed
  private final boolean oneSeqMode;

  /**
   * Constructs a generator for the specified EFSM module.
   *
   * @param module the EFSM module to be tested
   * @param loopLimit a loop iteration limit
   *
   * @throws IllegalArgumentException either if {@code module} is {@code null} or if
   *     {@code loopLimit} is lesser than zero
   */
  public Generator(final EfsmModule module, final int loopLimit) {

    this.simulator = new ModuleSimulator(module);

    if (loopLimit < 0) {
      throw new IllegalArgumentException("loopLimit is lesser than zero");
    }

    this.logEntryHeader = "EFSM.TestGenerator." + module.getName() + ": ";

    this.loopLimit = loopLimit;

    this.totalCoverage = new EfsmTransitionCoverage(this.simulator.getCoverage());

    this.traversedPaths = new HashMap<>();

    boolean oneSeqMode = false;

    for (final Efsm efsm : module.getNonInitialEfsms()) {

      if (efsm.singleTestable()) {
        this.traversedPaths.put(efsm, new HashSet<List<EfsmTransition>>());

        if (efsm.getResetGuardedActions().isEmpty()) {
          oneSeqMode = true;
        }
      } else {
        Log.warn("The efsm can't be tested as a single one: %s", efsm.getName());
      }
    }

    this.oneSeqMode = oneSeqMode;
  }

  //TODO process initial EFSMs
  @Override
  public Test generateTest() {

    Test test = new Test();
    final List<Efsm> efsms =
        new ArrayList<>(this.simulator.getSimulatedModel().getNonInitialEfsms());

    if (!efsms.isEmpty()) {
      //sequences.putAll(this.generateDirectly());

      Log.debug("%sgeneration has been finished successfully. Total coverage: %s",
          this.logEntryHeader, this.totalCoverage);
      Log.debug(
          "%stotal amount of sequences in the generated test before minimisation: %s",
          this.logEntryHeader,
          test.getSequenceNum());
      Log.debug(
          "%stotal amount of vectors in the generated test before minimisation: %s",
          this.logEntryHeader,
          + test.getVectorNum());

      final Map<Sequence, EfsmTransitionCoverage> sequences = this.generateRandomly();
      test = TestMinimiser.minimise(sequences);

      Log.debug(
          "%stotal amount of sequences in the generated test after minimisation: %s",
          this.logEntryHeader,
          test.getSequenceNum());
      Log.debug(
          "%stotal amount of vectors in the generated test after minimisation: %s",
          this.logEntryHeader,
          test.getVectorNum());

      for (final EfsmTransition transition : this.totalCoverage.getUncovered()) {

        Log.debug("%suncovered transition: %s", this.logEntryHeader, transition);
      }
    }

    return test;
  }

  private Map<Sequence, EfsmTransitionCoverage> generateRandomly() {

    final Map<Sequence, EfsmTransitionCoverage> sequences = new HashMap<>();
    final Set<Efsm> efsms = this.simulator.getSimulatedModel().getNonInitialEfsms();
    final Map<Efsm, Iterator<TestVector>> vectorIterators = new HashMap<>(efsms.size());

    final int noProgressLimit = this.getNoProgressLimit();
    final List<RandomGenerator> generators = new LinkedList<>();

    for (final Efsm efsm : efsms) {

      if (efsm.singleTestable()) {
        generators.add(new RandomGenerator(this.simulator.getSubSimulator(efsm)));
      }
    }

    int seqNoProgress = 0;

    final Comparator<RandomGenerator> comparator = new Comparator<RandomGenerator>() {

      @Override
      public int compare(final RandomGenerator gen1, final RandomGenerator gen2) {
        BigDecimal coverage1 =
            simulator.getSubSimulator(gen1.getEfsm()).getCoverage().getPercentage();
        BigDecimal coverage2 =
            simulator.getSubSimulator(gen2.getEfsm()).getCoverage().getPercentage();

        return coverage2.compareTo(coverage1);
      }
    };

    while (!this.totalCoverage.isFull() && (seqNoProgress <= noProgressLimit)) {

      Log.debug("%sstarting a new sequence", this.logEntryHeader);

      this.simulator.reset();

      final TestVector resetVector = this.resetModuleSimulator();

      if (resetVector == null) {
        break;
      }

      this.simulator.simulateVector(resetVector);

      for (final RandomGenerator generator : generators) {

        vectorIterators.put(generator.getEfsm(), generator.getNextSequenceIterator());
      }

      final Sequence sequence = new Sequence(resetVector);
      final Map<Efsm, List<EfsmTransition>> currentPaths = new HashMap<>(efsms.size());

      for (final Efsm efsm : efsms) {
        currentPaths.put(efsm, new ArrayList<EfsmTransition>());
      }

      int vectorNoProgress = 0;

      final List<TestVector> bufferedVectors = new ArrayList<>();
      EfsmTransitionCoverage seqCov = this.simulator.getCoverage();

      while (vectorNoProgress <= noProgressLimit) {

        Collections.sort(generators, comparator);
        TestVector currentVector = new TestVector();
        boolean success = false;

        for (final RandomGenerator generator : generators) {

          final Iterator<TestVector> sequenceIterator = vectorIterators.get(generator.getEfsm());
          generator.fixInputs(currentVector);

          final TestVector generatedVector = sequenceIterator.next();

          if (generatedVector != null) {
            success = true;
            currentVector = generatedVector;
          }
        }

        if (!success) {
          break;
        }

        TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(),
            currentVector);

        this.simulator.simulateVector(currentVector);

        for (final Efsm efsm : efsms) {
          if (efsm.singleTestable()) {
            final EfsmSimulator subSimulator = this.simulator.getSubSimulator(efsm);
            final EfsmTransition lastTraversed = subSimulator.getLastTraversed();
            final List<EfsmTransition> currentPath = currentPaths.get(efsm);
            currentPath.add(lastTraversed);
          }
        }

        final BigDecimal oldCovPercentage = seqCov.getPercentage();
        final EfsmTransitionCoverage bufSeqCov = this.simulator.getCoverage();
        final BigDecimal newCovPercentage = bufSeqCov.getPercentage();

        if (newCovPercentage.compareTo(oldCovPercentage) > 0) {
          Log.debug("%s: the new vector has increased current sequence coverage from %s to %s",
              this.logEntryHeader,
              oldCovPercentage,
              newCovPercentage);

          if (!bufferedVectors.isEmpty()) {
            sequence.addVectors(bufferedVectors);
            bufferedVectors.clear();
            vectorNoProgress = 0;
          }

          sequence.addVector(currentVector);
          seqCov = bufSeqCov;

          final EfsmTransitionCoverage totalCov = new EfsmTransitionCoverage(this.totalCoverage);
          totalCov.include(seqCov);

          if (totalCov.isFull()) {

            Log.debug("%s: total coverage is full now", this.logEntryHeader);

            this.totalCoverage.include(seqCov);

            break;
          }
        } else {

          vectorNoProgress++;
          bufferedVectors.add(currentVector);

          Log.debug(
              "%s: there is no progress in sequence coverage. Buffering the last vector. "
                  + "No progress counter: %s. No progress limit: %s",
              this.logEntryHeader,
              vectorNoProgress,
              noProgressLimit);
        }
      }

      final BigDecimal oldCovPercentage = this.totalCoverage.getPercentage();
      final EfsmTransitionCoverage bufTotalCov = new EfsmTransitionCoverage(this.totalCoverage);
      bufTotalCov.include(seqCov);
      final BigDecimal newCovPercentage = bufTotalCov.getPercentage();

      if (newCovPercentage.compareTo(oldCovPercentage) > 0) {
        Log.debug(
            "%s: the new sequence has increased total coverage from %s to %s%%",
            this.logEntryHeader,
            oldCovPercentage,
            newCovPercentage);

        this.totalCoverage.include(bufTotalCov);

        for (final Efsm efsm : efsms) {

          if (efsm.singleTestable()) {

            final List<EfsmTransition> path = currentPaths.get(efsm);

            for (int i = 1; i <= bufferedVectors.size(); i++) {
              path.remove(path.size() - 1);
            }

            this.traversedPaths.get(efsm).add(path);
          }
        }

        sequences.put(sequence, new EfsmTransitionCoverage(seqCov));
      } else {
        seqNoProgress++;

        Log.debug(
            "%s: there is no progress in total coverage. Skipping the last sequence. "
                + "No progress counter: %s. No progress limit: %s",
            this.logEntryHeader,
            seqNoProgress,
            noProgressLimit);
      }
    }

    Log.debug(
        "%s: the random phase of generation has been finished successfully. Total coverage: %s",
        this.logEntryHeader,
        this.totalCoverage);

    return sequences;
  }


//TODO
  /*private Map<Sequence, EfsmTransitionCoverage> generateDirectly() {
    if (!this.totalCoverage.isFull()) {
      Set<Efsm> efsms = this.simulator.getSimulatedModel().getNonInitialEfsms();

      if (efsms.size() == 1) {
        Generator.LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": starting the directed phase of generation");

        for (RandomGenerator randomGenerator: generators) {
          Efsm efsm = randomGenerator.getEfsm();

          if (randomGenerator.getTotalCoverage().isFull()) {
            Generator.LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the " + efsm + " EFSM is fully covered. Skipping it");
          } else {
            Generator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the " + efsm +
                " EFSM has uncovered transitions. Processing it");

            DirectedGenerator directedGenerator =
                new DirectedGenerator(efsm, traversedPaths.get(efsm), loopLimit);

            for (Iterator<TestVector> vectorIterator = directedGenerator.getNextSequenceIterator();
                vectorIterator != null;
                vectorIterator = directedGenerator.getNextSequenceIterator()) {
              Sequence sequence = new Sequence();

              while (vectorIterator.hasNext()) {
                TestVector vector = vectorIterator.next();

                TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(),
                    vector);

                sequence.addVector(vector);
              }

              test.addSequence(sequence);

              EfsmTransitionCoverage sequenceCoverage = directedGenerator.getSequenceCoverage();
              totalCoverage.include(sequenceCoverage);
              sequencesCoverage.put(sequence, sequenceCoverage);

              Generator.LOGGER.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": the directed sequence has been generated successfully");
              Generator.LOGGER.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": current coverage: " + totalCoverage);
            }
          }
        }

        Generator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader +
                ": the directed phase of generation has been finished successfully");
      } else {
        Generator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader +
            ": directed generation for multi-EFSM systems isn't supported yet");

        return Collections.emptyMap();
      }
    }
  }*/

  private TestVector resetModuleSimulator() {

    Log.debug("%s: resetting the module", this.logEntryHeader);
    TestVector vector = new TestVector();

    for (final EfsmSimulator efsmSimulator : this.simulator.getSubSimulators()) {

      final Collection<GuardedAction> rstActions = efsmSimulator.getEfsm().getResetGuardedActions();

      for (final Iterator<GuardedAction> iterator = rstActions.iterator(); iterator.hasNext();) {

        final GuardedAction next = iterator.next();
        final EventList eventList = vector.getEventList();
        final Map<String, NodeValue> assignments = vector.getAssignments();

        final TestVector candidateVector =
            efsmSimulator.enableGuardedAction(next, eventList, assignments);

        if (candidateVector == null) {
          if (!iterator.hasNext()) {
            Log.warn("%s: failed to reset the process: %s",
                this.logEntryHeader,
                efsmSimulator.getName());
          }
        } else {
          vector = candidateVector;

          break;
        }
      }
    }

    Log.debug("%s: the module has been reset", this.logEntryHeader);

    TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(), vector);

    return vector;
  }

  private int getNoProgressLimit() {
    int transitionsAmount = 0;
    int statesAmount = 0;

    for (final Efsm efsm : this.simulator.getSimulatedModel().getNonInitialEfsms()) {

      transitionsAmount = transitionsAmount + efsm.getTransitionNum();
      statesAmount = statesAmount + efsm.getStateNum();
    }

    return transitionsAmount / statesAmount;
  }

  @Override
  public TestCoverage getCoverage() {
    return this.totalCoverage;
  }
}
