/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction;

import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * The class representing a random-based heuristic for choosing a transitions to traverse.
 * <br>
 * Each instance of {@code SemiRandomEfsmTransitionChooser} is bounded to the specified
 * EFSM, its state, and test coverage.
 * <br>
 * This class is very similar to {@link RandomEfsmTransitionChooser} but during choosing it uses
 * coverage information in order to firstly choose uncovered transitions.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomEfsmTransitionChooser
 */
public final class SemiRandomEfsmTransitionChooser extends EfsmTransitionChooser {

  private final List<EfsmTransition> uncovered;
  private final List<EfsmTransition> covered;

  private final Random rndNumGen;

  /**
   * Creates a new SemiRandomEfsmTransitionChooser for the specified EFSM, its state, and coverage.
   * <br>
   * Note that list of choosable transitions is constructed only during a
   * SemiRandomEfsmTransitionChooser instance initialisation, so any future changes in the
   * specified EFSM, state, and coverage instances will not affect a result of choosing by that
   * {@code SemiRandomEfsmTransitionChooser} instance.
   *
   * @param efsm the EFSM
   * @param state the state of the EFSM
   * @param coverage the coverage of the EFSM
   * @param seed a seed for an inner random number generator
   */
  public SemiRandomEfsmTransitionChooser(
      final Efsm efsm,
      final EfsmState state,
      final EfsmTransitionCoverage coverage,
      final long seed) {

    this.rndNumGen = new Random(seed);
    this.uncovered = new ArrayList<>();
    this.covered = new ArrayList<>();

    for (final EfsmTransition transition : efsm.getOutgoing(state)) {
      if (coverage.isCovered(transition)) {
        this.covered.add(transition);
      } else {
        this.uncovered.add(transition);
      }
    }
  }

  /**
   * Chooses and returns a transition for traversing.
   * <br>
   * The transition is chosen at random among previously unchosen transitions outgoing from the
   * EFSM state which this SemiRandomEfsmTransitionChooser is bounded to.
   * <br>
   * Firstly SemiRandomEfsmTransitionChooser chooses priority transitions which were specified
   * during construction of this chooser. If there are no such previously unchosen transitions, it
   * switches subsequently to uncovered transitions and finally to covered transitions.
   *
   * @return the transition proposed for traversing. {@code null} is returned if all choosable
   *         transitions were chosen or there are no choosable transitions at all
   */
  @Override
  public EfsmTransition chooseNext() {

    if (this.uncovered.isEmpty()) {

      if (this.covered.isEmpty()) {

        return null;

      } else {
        final int chosenIndex = this.rndNumGen.nextInt(this.covered.size());

        return this.covered.remove(chosenIndex);
      }

    } else {
      final int chosenIndex = this.rndNumGen.nextInt(this.uncovered.size());

      return this.uncovered.remove(chosenIndex);
    }
  }

  @Override
  public boolean hasNext() {
    return !this.uncovered.isEmpty() || !this.covered.isEmpty();
  }
}
