/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.printer.graphml;

import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.engine.basis.GraphMlUtils;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.efsm.EfsmEmptyVisitor;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.File;
import java.io.PrintWriter;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * EFSM visitor for GraphML printer.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EfsmGraphMlVisitor extends EfsmEmptyVisitor {

  /* Writer to GraphML output file. */
  private PrintWriter printWriter;

  /* Map of identifiers for transitions. */
  private Map<EfsmTransition, Integer> transitionMap;

  /* Map of identifiers for states. */
  private Map<EfsmState, Integer> stateMap;

  /* Model identifier. */
  private int modelId;
  private int groupId;

  /**
   * Default constructor.
   */
  public EfsmGraphMlVisitor() {
    super();
  }

  /**
   * Create GraphML file with the specified name.
   *
   * @param fileName file name
   */
  public void initFile(final String fileName) {

    final File graphMlFile = FileCreator.newFile(fileName);

    modelId = 0;
    printWriter = ReadWriteUtils.newPrintWriter(graphMlFile);
  }

  @Override
  public void onProcessBegin(final ModuleProcess efsm) {
    GraphMlUtils.onGraphBegin(printWriter, Integer.toString(modelId));
    stateMap = new LinkedHashMap<>();
    transitionMap = new LinkedHashMap<>();
    modelId++;
  }

  @Override
  public void onProcessEnd(final ModuleProcess efsm) {
    GraphMlUtils.onGraphEnd(printWriter);
  }

  @Override
  public void onStateBegin(final EfsmState state) {
    final int idState = getStateId(state);

    GraphMlUtils.printNode(printWriter,
        Integer.toHexString(idState) + Integer.toString(modelId), state.getId(),
        GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  private int getStateId(final EfsmState state) {

    final Integer idState = stateMap.get(state);

    if (idState == null) {
      final int mapSize = stateMap.size();
      stateMap.put(state, mapSize);
      return mapSize;
    }

    return idState;
  }

  @Override
  public void onTransitionBegin(final EfsmTransition transition) {

    final Integer idTransition = transitionMap.get(transition);

    if (idTransition == null) {

      final int mapSize = transitionMap.size();
      transitionMap.put(transition, mapSize);
    } else {
      return;
    }

    /* Get child & parent nodes. */
    final int sourceState = getStateId(transition.getSrc());

    final String guardId = "Guard" + Integer.toHexString(System.identityHashCode(transition));
    final String guardNodeId =
        String.format("{%s : %s}", transition.getClocks(), transition.getGuardNode());
    GraphMlUtils.printNode(printWriter, guardId, guardNodeId,
        GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_BLUE, GraphMlUtils.CROPPING_TEXT);

    GraphMlUtils.printEdge(
        printWriter,
        Integer.toHexString(sourceState)  + Integer.toString(modelId), guardId);

    final String actionId = "Action" + Integer.toHexString(System.identityHashCode(transition));
    GraphMlUtils.printNode(printWriter, actionId, transition.getAction().getId(),
        GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_GREEN, GraphMlUtils.CROPPING_TEXT);
    GraphMlUtils.printEdge(printWriter, guardId, actionId);

    final int destinationState = getStateId(transition.getDst());

    GraphMlUtils.printEdge(
        printWriter, actionId,
        Integer.toHexString(destinationState)  + Integer.toString(modelId));
  }

  @Override
  public void onModelBegin(final Model efsmModel) {
    GraphMlUtils.onGraphMlBegin(printWriter);
  }

  @Override
  public void onModelEnd(final Model efsmModel) {
    GraphMlUtils.onGraphMlEnd(printWriter);

    GraphMlUtils.closeGraphMlFile(printWriter);
  }

  @Override
  public void onModuleBegin(final Module efsmContainer) {
    GraphMlUtils.printGroup(printWriter, "Group" + groupId, efsmContainer.getName());
    groupId ++;
  }

  @Override
  public void onModuleEnd(final Module efsmContainer) {
    GraphMlUtils.closeGroup(printWriter);
  }
}
