/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods targeted at EFSM model simulation features.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmSimulatorUtils {

  private EfsmSimulatorUtils() {
    /* Empty. */
  }

  /**
   * Collect EFSM transition coverage information for the specified test model on the specified
   * EFSM model.
   * @param testModel Test to that the coverage is measured.
   * @param efsmModel EFSM model that is to be simulated to retrieve the coverage.
   * @return Mapping from test sequence to related EFSM model transition coverage.
   */
  public static Map<Sequence, EfsmTransitionCoverage> collectCoverage(
      final TestModel testModel,
      final EfsmModel efsmModel) {

    final Map<Sequence, EfsmTransitionCoverage> covMap = new LinkedHashMap<>();

    final List<Test> tests = testModel.getTests();
    final List<Module> topModules = new LinkedList<>(efsmModel.getModules());
    final int testNum = tests.size();
    final int topModuleNum = topModules.size();
    InvariantChecks.checkTrue(testNum == topModuleNum,
        String.format("testNum %d != %d topModuleNum", testNum, topModuleNum));

    for (int i = 0; i < testNum; i++) {

      final Module topModule = topModules.get(i);
      final ModuleSimulator simulator = new ModuleSimulator((EfsmModule) topModule);

      final Test test = tests.get(i);
      final List<Sequence> sequences = test.getSequences();

      for (final Sequence sequence : sequences) {

        simulator.reset();
        simulator.resetCoverage();

        simulator.simulateSequence(sequence);
        final EfsmTransitionCoverage coverage = simulator.getCoverage();

        if (covMap.containsKey(sequence)) {
          Log.error("Need to add more covered transitions to one sequence.");
        } else {
          covMap.put(sequence, coverage);
        }
      }
    }

    return covMap;
  }

  /**
   * Print information about EFSM transition coverage that the specified test reaches.
   * <p>All the information about coverage that different sequences reach is stored
   * into the specified mapping.</p>
   * @param map Mapping from test sequence to the related coverage of EFSM transitions.
   * @param test Test to which the coverage should be printed.
   */
  public static void printEfsmTransitionCoverage(
      final Map<Sequence, EfsmTransitionCoverage> map,
      final Test test) {

    final List<Sequence> sequences = test.getSequences();

    if (sequences.isEmpty()) {
      Log.info("EFSM transition coverage: 0%%.");
      return;
    }
    final Iterator<Sequence> iterator = sequences.iterator();
    final EfsmTransitionCoverage coverage = new EfsmTransitionCoverage(map.get(iterator.next()));

    while (iterator.hasNext()) {
      final EfsmTransitionCoverage sequenceCoverage = map.get(iterator.next());

      final Set<EfsmTransition> coveredBySequence = sequenceCoverage.getCovered();

      for (final EfsmTransition transition : coveredBySequence) {
        if (!coverage.isCovered(transition)) {
          coverage.markTransitionAsCovered(transition);
        }
      }
    }

    final Set<EfsmTransition> uncovered = coverage.getUncovered();
    final Set<EfsmTransition> covered = coverage.getCovered();

    Log.info(
            "EFSM transition coverage: %s%% (%d/%d).%n",
            String.valueOf(coverage.getPercentage()),
            covered.size(),
            covered.size() + uncovered.size());

    for (final EfsmTransition transition : uncovered) {
      Log.warn("EFSM transition remains uncovered: %s", transition.toString());
    }
  }
}
