/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.basis.SimulationSnapshot;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator.EfsmSimulationSnapshot;
import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmModuleState;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

/**
 * A simulator for {@code EfsmModule}.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 *
 * @see EfsmModule
 */
public class ModuleSimulator extends Simulator<EfsmModule> {

  private final Set<EfsmSimulator> efsmSimulators;
  private final Map<String, Map<EfsmSimulator, NodeValue>> varModifications;

  private final String logEntryHeader;

  private final Random rndGen;

  /**
   * Constructs a simulator for the specified EFSM module.
   *
   * @param efsmModule the EFSM module which the simulator is constructed for
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public ModuleSimulator(final EfsmModule efsmModule) {
    super(efsmModule);

    this.efsmSimulators = new HashSet<>();

    final String name = this.simulatedModel.getName();
    this.logEntryHeader = String.format("EFSM.Simulator.%s:", name);

    for (final Efsm efsm : this.simulatedModel.getNonInitialEfsms()) {
      if (efsm.singleTestable()) {
        this.efsmSimulators.add(new EfsmSimulator(this, efsm));
      } else {
        Log.warn("The efsm can't be tested as a single one: %s", efsm.getName());
      }
    }

    final Collection<String> registerNames = this.simulatedModel.getRegisterNames();
    final Collection<String> outputNames = this.simulatedModel.getOutputNames();

    this.varModifications = new HashMap<>(registerNames.size() + outputNames.size());
    resetModifications();

    this.reset();

    this.rndGen = new Random(0L);
  }

  private void resetModifications() {
    final Collection<String> registerNames = this.simulatedModel.getRegisterNames();
    final Collection<String> outputNames = this.simulatedModel.getOutputNames();

    for (final String variable : registerNames) {
      this.varModifications.put(variable, new HashMap<>());
    }

    for (final String variable: outputNames) {
      this.varModifications.put(variable, new HashMap<>());
    }
  }

  @Override
  public EfsmModule getSimulatedModel() {
    return this.simulatedModel;
  }

  @Override
  public TestVector simulateVector(final TestVector inputVector) {
    super.readVector(inputVector);

    Log.debug("%s processing vector: ", this.logEntryHeader, inputVector);

    final Transaction outTransaction = new Transaction();

    for (final EfsmSimulator efsmSimulator : this.efsmSimulators) {
      efsmSimulator.simulateVector(inputVector);
    }

    updateMemory(outTransaction);

    final TestVector outputVector = new TestVector(outTransaction);

    Log.debug("%s vector has been processed successfully. Output vector: %s",
        this.logEntryHeader, outputVector);

    return outputVector;
  }

  private void updateMemory(final Transaction outTransaction) {

    for (final Entry<String, NodeValue> modification: resolveAssignmentConflicts().entrySet()) {

      final String variableName = modification.getKey();
      final NodeValue variableValue = modification.getValue();

      if (this.regMemory.containsVariable(variableName)) {

        this.regMemory.setValue(variableName, variableValue);

      } else {

        if (this.outputsMemory.containsVariable(variableName)) {

          this.outputsMemory.setValue(variableName, variableValue);
          outTransaction.addVariable(variableName, variableValue);
        } else {
          throw new RetrascopeException("The variable " + variableName
              + " isn't found neither among inner variables nor among outputs");
        }
      }
    }
  }

  @Override
  public Sequence simulateSequence(final Sequence inputSequence) {
    InvariantChecks.checkNotNull(inputSequence);

    Log.debug("%s processing sequence: %s", this.logEntryHeader, inputSequence);

    final Sequence outputSequence = new Sequence();

    inputSequence.initIterator();
    while (inputSequence.hasNext()) {
      outputSequence.addVector(this.simulateVector(inputSequence.next()));
    }

    Log.debug("%s sequence has been processed successfully. Output sequence: %s",
        this.logEntryHeader, outputSequence);

    return outputSequence;
  }

  @Override
  public void reset() {
    Log.debug("%s resetting", this.logEntryHeader);

    super.reset();

    for (final EfsmSimulator efsmSimulator : this.efsmSimulators) {
      efsmSimulator.reset();
    }

    resetModifications();

    Log.debug("%s reset has been finished successfully", this.logEntryHeader);
  }

  @Override
  public EfsmTransitionCoverage getCoverage() {
    final EfsmTransitionCoverage coverage = new EfsmTransitionCoverage();

    for (final EfsmSimulator efsmSimulator: this.efsmSimulators) {
      coverage.include(efsmSimulator.getCoverage());
    }

    return coverage;
  }

  /**
   * Returns a coverage of the specified EFSM.
   *
   * @param efsm the target EFSM
   *
   * @return the coverage of the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public EfsmTransitionCoverage getCoverage(final Efsm efsm) {

    for (final EfsmSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getEfsm().equals(efsm)) {
        return efsmSimulator.getCoverage();
      }
    }

    return null;
  }

  /**
   * Returns a simulator for the specified EFSM. The returned simulator represents a current state
   * of EFSM simulation.
   *
   * @param efsm the target EFSM
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public EfsmSimulator getSubSimulator(final Efsm efsm) {

    for (final EfsmSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getEfsm().equals(efsm)) {
        return efsmSimulator;
      }
    }

    return null;
  }

  /**
   * Returns all process simulators of this simulator. The returned simulators represent a current
   * state of simulation of the processes.
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public Set<EfsmSimulator> getSubSimulators() {
    return Collections.unmodifiableSet(this.efsmSimulators);
  }

  /**
   * Submits a variable modification request to this simulator. If there is no inner/output variable
   * with the specified name in a simulated module, this method has no any effects.
   * <br>
   * The request is valid only for the current simulation cycle. At the end of the cycle the
   * simulator performs all requested modifications.
   *
   * @param simulator a process simulator which caused variable modification
   * @param variableName the name of a variable to be modified
   * @param value a new requested value of the variable
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  void modifyVariable(
      final EfsmSimulator simulator,
      final String variableName,
      final NodeValue value) {
    final Map<EfsmSimulator, NodeValue> modifications =
        this.varModifications.get(variableName);

    if (modifications == null) {
      Log.warn("%s a request to modify unknown variable %s has been received",
          this.logEntryHeader, variableName);
    } else {
      modifications.put(simulator, value);
    }
  }

  private Map<String, NodeValue> resolveAssignmentConflicts() {

    final Map<String, NodeValue> actualMods = new HashMap<>(this.varModifications.size());

    for (final Entry<String, Map<EfsmSimulator, NodeValue>> modifications :
        this.varModifications.entrySet()) {
      final String variableName = modifications.getKey();
      final Map<EfsmSimulator, NodeValue> assignedValues = modifications.getValue();

      switch (assignedValues.size()) {
        case 0: {
          break;
        }

        case 1: {
          actualMods.put(variableName, assignedValues.values().iterator().next());

          break;
        }

        default: {
          Log.warn("%s multiple simultaneous assignments to '%s': %s",
              this.logEntryHeader,
              variableName,
              assignedValues);

          final int chosenAssignment = this.rndGen.nextInt(assignedValues.size());
          final NodeValue chosenValue =
              new ArrayList<>(assignedValues.values()).get(chosenAssignment);

          Log.debug("%s randomly chosen assigned value: %s", this.logEntryHeader, chosenValue);

          actualMods.put(variableName, chosenValue);
          break;
        }
      }
    }

    return actualMods;
  }

  @Override
  public EfsmModuleState getCurrentState() {
    final Collection<State> subStates = new ArrayList<>();

    for (final EfsmSimulator subSimulator: this.efsmSimulators) {
      subStates.add(subSimulator.getCurrentState());
    }

    return new EfsmModuleState(subStates);
  }

  @Override
  public void resetCoverage() {
    for (final EfsmSimulator subSimulator: this.efsmSimulators) {
      subSimulator.resetCoverage();
    }
  }

  public class ModuleSimulationSnapshot extends SimulationSnapshot {
    private final Map<EfsmSimulator, EfsmSimulationSnapshot> snapshots;

    ModuleSimulationSnapshot() {
      super(ModuleSimulator.this);

      this.snapshots = new HashMap<>(ModuleSimulator.this.efsmSimulators.size());

      for (final EfsmSimulator subSimulator: ModuleSimulator.this.efsmSimulators) {
        this.snapshots.put(subSimulator, subSimulator.makeSimulationSnapshot());
      }
    }
  }

  @Override
  public ModuleSimulationSnapshot makeSimulationSnapshot() {
    return new ModuleSimulationSnapshot();
  }

  /**
   * Loads the specified simulation snapshot discarding current simulation state and history.
   *
   * @param snapshot the snapshot
   */
  public void loadSnapshot(final SimulationSnapshot snapshot) {

    Log.debug("%s: loading a simulation snapshot", this.logEntryHeader);

    super.loadSnapshot(snapshot);

    if (snapshot instanceof ModuleSimulationSnapshot) {

      final ModuleSimulationSnapshot moduleSnapShot = (ModuleSimulationSnapshot) snapshot;

      for (final EfsmSimulator subSimulator : ModuleSimulator.this.efsmSimulators) {
        subSimulator.loadSnapshot(moduleSnapShot.snapshots.get(subSimulator));
      }
    }
  }
}
