/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.extractor.property;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.result.usedef.UseDefContainer;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class GaddUseDefVisitor extends CfgEmptyVisitor {

  private Map<NodeVariable, UseDefContainer> useDefMap;
  private Stack<Node> nodeStack = new Stack<>();
  private Stack<Node> conditionStack = new Stack<>();
  private ModuleProcess currentProcess;

  @Override
  public void onModuleBegin(final Module module) {
    useDefMap = new HashMap<>();

    final Set<NodeVariable> variables = new HashSet<>();
    final Collection<NodeVariable> defines = new ArrayList<>(module.getDefines());
    final Collection<NodeVariable> uses = new ArrayList<>(module.getUses());
    removeVersionsAndPhaseVars(module, defines);
    removeVersionsAndPhaseVars(module, uses);
    variables.addAll(defines);
    variables.addAll(uses);
    addKeys(variables);
  }

  public Map<NodeVariable, UseDefContainer> getUseDefMap() {
    return useDefMap;
  }

  private void removeVersionsAndPhaseVars(final Module module,
      final Collection<NodeVariable> variables) {
    final List<NodeVariable> varsToRemove = new ArrayList<>();
    final List<NodeVariable> phaseVars = getPhaseVars(module);
    for (NodeVariable variable : variables) {
      if (GaddHlddUtils.getOriginalVariable(variable) != null || phaseVars.contains(variable)) {
        varsToRemove.add(variable);
      }
    }
    variables.removeAll(varsToRemove);
  }

  private List<NodeVariable> getPhaseVars(final Module module) {
    final List<NodeVariable> phaseVars = new ArrayList<>();
    for (ModuleProcess process : module.getProcesses()) {
      final NodeVariable phaseVar = GaddHlddUtils.getPhaseVariable(module, process);
      if (phaseVar != null) {
        phaseVars.add(phaseVar);
      }
    }
    return phaseVars;
  }

  /*
   * Fills the use/def map with key variables.
   */
  private void addKeys(final Collection<NodeVariable> keys) {
    for (NodeVariable variable : keys) {
      useDefMap.put(variable, new UseDefContainer());
    }
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    nodeStack = new Stack<>();
    conditionStack = new Stack<>();
    currentProcess = process;
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {
    for (Assignment assignment : block.getAssignments()) {
      addNodes(assignment);
    }
  }

  /*
   * Extracts def/use conditions from assignment.
   */
  private void addNodes(final Assignment assignment) {
    final Collection<NodeVariable> defKeys = assignment.getDefines();
    final Collection<NodeVariable> useKeys = assignment.getUses();
    final Collection<NodeVariable> keys = new HashSet<>();
    keys.addAll(defKeys);
    keys.addAll(useKeys);
    for (NodeVariable key : keys) {
      final NodeVariable originalKey = GaddHlddUtils.getOriginalVariable(key);
      final NodeVariable finalKey;
      if (originalKey != null) {
        finalKey = originalKey;
      } else {
        finalKey = key;
      }
      final UseDefContainer useDefContainer = useDefMap.get(finalKey);
      if (useDefContainer != null) {

        final Node extraCondition = getIteCondition(assignment.getValue(), finalKey);

        if (extraCondition != null) {
          nodeStack.push(extraCondition);
        }

        if (defKeys.contains(finalKey)) {
          final Node finalNode =
              getCondition(useDefContainer.getDefineCondition(currentProcess.getId()));
          useDefContainer.putDefineCondition(currentProcess, finalNode);
        }
        if (useKeys.contains(finalKey)) {
          final Node finalNode =
              getCondition(useDefContainer.getUseCondition(currentProcess.getId()));
          useDefContainer.putUseCondition(currentProcess, finalNode);
        }

        if (extraCondition != null) {
          nodeStack.pop();
        }
      }
    }
  }

  /*
   * Extracts condition from if-then-else expression
   */
  private Node getIteCondition(final Node node, final NodeVariable key) {
    if (node instanceof NodeOperation
        && ((NodeOperation) node).getOperationId() == StandardOperation.ITE) {

      final NodeOperation ite = (NodeOperation) node;
      final Node condition = ite.getOperand(0);
      final Node firstExpr = ite.getOperand(1);
      final Node secondExpr = ite.getOperand(2);

      final Collection<NodeVariable> firstExprVars = ExprUtils.getVariables(firstExpr);
      final Collection<NodeVariable> secondExprVars = ExprUtils.getVariables(secondExpr);

      if (firstExprVars.contains(key)) {
        final Node nextCondition = getIteCondition(firstExpr, key);
        return nextCondition == null ? condition : Nodes.and(condition, nextCondition);
      }
      if (secondExprVars.contains(key)) {
        final Node nextCondition = getIteCondition(secondExpr, key);
        return nextCondition == null
            ? Nodes.not(condition)
            : Nodes.and(Nodes.not(condition), nextCondition);
      }
    }

    return null;
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {
    if (caseNode.isDefault()) {
      nodeStack.push(getDefaultCondition(caseNode));
    } else {
      final CfgCaseData caseData = caseNode.getCfgCaseData();
      final Set<NodeValue> values = caseData.getValues();
      nodeStack.push(getCaseCondition(values));
    }
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    nodeStack.pop();
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    conditionStack.push(switchNode.getCondition());
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {
    conditionStack.pop();
  }

  @Override
  public void onModuleEnd(final Module module) {
    printAssertions();
  }

  /*
   * Generates def/use condition.
   */
  private Node getCondition(final Node storedNode) {
    if (storedNode != null) {
      return Nodes.or(storedNode, getPathCondition());
    } else {
      return getPathCondition();
    }
  }

  /*
   * Generates condition using current stack state.
   */
  private Node getPathCondition() {
    if (nodeStack.empty()) {
      return Nodes.TRUE;
    } else if (nodeStack.size() == 1) {
      return nodeStack.peek();
    } else {
      return Nodes.and(nodeStack);
    }
  }

  /*
   * Generates condition for 'default' branch.
   */
  private Node getDefaultCondition(final CfgCaseStatement caseNode) {
    final CfgSwitchStatement parent = (CfgSwitchStatement) caseNode.getParent();
    final Collection<NodeValue> values = new ArrayList<>();
    for (int i = 0; i < parent.getStatementNum(); i++) {
      final CfgCaseStatement child = (CfgCaseStatement) parent.getStatement(i);
      if (child != caseNode) {
        values.addAll(child.getValues());
      }
    }
    return Nodes.not(getCaseCondition(values));
  }

  /*
   * Generates condition using case values.
   */
  private Node getCaseCondition(final Collection<NodeValue> values) {
    final Collection<Node> nodes = new ArrayList<>();
    for (NodeValue value : values) {
      nodes.add(Nodes.eq(conditionStack.peek(), value));
    }
    if (nodes.size() > 1) {
      return Nodes.or(nodes);
    } else {
      return nodes.iterator().next();
    }
  }

  private void printAssertions() {
    for (Map.Entry<NodeVariable, UseDefContainer> useDefMapEntry : useDefMap.entrySet()) {
      for (Node node : getReadWriteConditions(useDefMapEntry.getValue())) {
        Log.info("Read-write condition for %s:%n%s", useDefMapEntry.getKey(), node.toString());
      }

      for (Node node : getWriteWriteConditions(useDefMapEntry.getValue())) {
        Log.info("Write-write condition for %s:%n%s", useDefMapEntry.getKey(), node.toString());
      }
    }
  }

  private Collection<Node> getReadWriteConditions(final UseDefContainer container) {
    final Collection<Node> result = new ArrayList<>();
    final List<String> defKeys = new ArrayList<>(container.getDefProcesses());
    final List<String> useKeys = new ArrayList<>(container.getUseProcesses());
    for (String defKey : defKeys) {
      for (String useKey : useKeys) {
        if (defKey.compareTo(useKey) != 0) {
          result.add(
              Nodes.and(container.getDefineCondition(defKey), container.getUseCondition(useKey)));
        }
      }
    }
    return result;
  }

  private Collection<Node> getWriteWriteConditions(final UseDefContainer container) {
    final Collection<Node> result = new ArrayList<>();
    final List<Node> defines = new ArrayList<>(container.getDefines().values());
    if (defines.size() > 1) {
      for (int i = 0; i < defines.size(); i++) {
        for (int j = i + 1; j < defines.size(); j++) {
          result.add(Nodes.and(defines.get(i), defines.get(j)));
        }
      }
    }
    return result;
  }

}
