/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.formula.GaddHlddFormulaVisitor;
import ru.ispras.retrascope.model.basis.walker.ModelVisitor;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.model.hldd.HlddBuilder;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddWalker;

import java.util.Map;

/**
 * This engine transforms the GADD model into the HLDD model.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class GaddHlddTransformer extends Engine {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "gadd-hldd-transformer";

  /**
   * Delete unused variables parameter.
   */
  public static final Parameter DELETE_UNUSED =
      new Parameter("delete-unused", false, "Delete the variables, that are never used");

  /**
   * Use formula based visitor parameter.
   */
  public static final Parameter FORMULA_BASED =
      new Parameter("formula-based", false, "Use formula-based visitor");

  public static final Parameter COLLECT_USEDEF =
      new Parameter("collect-usedef", false, "Collect variable defs/uses");

  /**
   * Constructs a new engine.
   */
  public GaddHlddTransformer() {
    super(ENGINE_ID);
    addInputType(GaddModel.TYPE);
    setOutputType(HlddModel.TYPE);
    addParameter(DELETE_UNUSED);
    addParameter(FORMULA_BASED);
    addParameter(COLLECT_USEDEF);
  }

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {
    final Arguments arguments =
        Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);
    boolean deleteUnused = false;
    boolean formulaBased = false;
    boolean collectUsedef = false;

    try {
      if (arguments != null) {
        final Arguments parsedArgs = parseCommandLine(arguments.getCommandLine());
        deleteUnused = parsedArgs.contains(DELETE_UNUSED);
        formulaBased = parsedArgs.contains(FORMULA_BASED);
        collectUsedef = parsedArgs.contains(COLLECT_USEDEF);
      }
    } catch (ParseException e) {
      e.printStackTrace();
    }

    final GaddModel cgaaModel = (GaddModel) inputs.get(GaddModel.TYPE);
    final CfgModel cfgModel = cgaaModel.getModel();
    final ModelVisitor visitor =
        formulaBased ? new GaddHlddFormulaVisitor() : new GaddHlddVisitor();

    if (formulaBased) {
      ((GaddHlddFormulaVisitor) visitor).setCollectUseDef(collectUsedef);
    }

    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS);
    walker.start(cfgModel);
    final HlddModel model = ((HlddBuilder) visitor).getModel();

    final HlddUnusedVariableBackend unused = new HlddUnusedVariableBackend(deleteUnused);
    final HlddWalker unusedWalker = new HlddWalker(unused);
    unusedWalker.start(model);

    final HlddVariableAnalyzerBackend analyzer = new HlddVariableAnalyzerBackend();
    final HlddWalker analyzerWalker = new HlddWalker(analyzer);
    analyzerWalker.start(model);

    return model;
  }

}
