/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class represents a visitor which is used to search for unused variables in the design.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddUnusedVariableBackend extends HlddEmptyVisitor {

  private Set<NodeVariable> usedVars = new HashSet<>();
  private boolean deleteUnused;

  HlddUnusedVariableBackend(final boolean delUnused) {
    deleteUnused = delUnused;
  }

  @Override
  public void onAssignment(final HlddTerminalNode node) {
    getUsedVars(node.getAssignedValue());
  }

  @Override
  public void onConditionBegin(final HlddConditionNode node) {
    getUsedVars(node.getExpression());
  }

  @Override
  public void onEdgeBegin(final HlddEdge edge) {
    getUsedVars(edge.getConditionEquation());
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    final NodeVariable variable = hldd.getVariable();
    if (variable != null) {
      usedVars.add(variable);
    }
  }

  @Override
  public void onModuleBegin(final Module module) {
    for (Map.Entry<NodeVariable, Declaration> entry : module.getDeclarationMap().entrySet()) {
      getUsedVars(entry.getValue().getInitialValue());
      getUsedVars(DeclarationUtils.getInvariant(entry.getValue(), entry.getKey()));
    }
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    getUsedVars(((HlddProcess) process).getInvariant());
  }

  private void getUsedVars(final Node expression) {
    if (expression != null) {
      usedVars.addAll(ExprUtils.getVariables(expression));
    }
  }

  @Override
  public void onModuleEnd(final Module module) {
    final Set<NodeVariable> unusedVars = new HashSet<>(module.getDeclarationMap().keySet());
    unusedVars.removeAll(module.getOutputs());
    unusedVars.removeAll(usedVars);
    removeArrayItems(unusedVars);
    if (!unusedVars.isEmpty()) {
      Log.warn("The following variables are never used: %s", unusedVars.toString());
      if (deleteUnused) {
        removeHldds(unusedVars, module);
        Log.warn("Unused variables have been removed from the model");
      }
    }
  }

  private void removeArrayItems(final Set<NodeVariable> unusedVars) {
    final Collection<NodeVariable> itemsToRemove = new ArrayList<>();
    for (NodeVariable var : unusedVars) {
      if (var.getUserData() instanceof NodeOperation) {
        itemsToRemove.add(var);
      }
    }
    unusedVars.removeAll(itemsToRemove);
  }

  private void removeHldds(final Set<NodeVariable> unusedVars, final Module module) {
    for (NodeVariable variable : unusedVars) {
      module.removeDeclaration(variable);
      for (ModuleProcess process : module.getProcesses()) {
        final HlddProcess hlddProcess = (HlddProcess) process;
        final Collection<Hldd> hlddsToDelete = new ArrayList<>();
        for (Hldd hldd : hlddProcess.getHldds()) {
          if (hldd.getVariable().equals(variable)) {
            hlddsToDelete.add(hldd);
          }
        }
        ((HlddProcess) process).removeHldds(hlddsToDelete);
      }
    }
  }

}
