/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents a visitor which is used to remove duplicate variable assignments
 * in different processes.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddVariableAnalyzerBackend extends HlddEmptyVisitor {

  private Map<NodeVariable, List<ModuleProcess>> assignments = new HashMap<>();
  private Module parentModule;
  private ModuleProcess parentProcess;

  private static final String DUPLICATE_VAR_PROCESS = "duplicate_vars_assignments";
  private static final String INITIAL_FLAG = "initial_flag";
  private static final String ASSIGNMENT = "assignment";
  private static final String MASK = "mask";

  HlddVariableAnalyzerBackend() {
  }

  /*
   * Transforms multiple assignments of a variable.
   */
  private void transformMultipleAssignments() {
    final ArrayList<Hldd> newHldds = new ArrayList<>();
    for (Map.Entry<NodeVariable, List<ModuleProcess>> entry : assignments.entrySet()) {
      // if more than one process assigns a variable
      if (entry.getValue().size() > 1) {
        final NodeVariable oldVar = entry.getKey();
        final Map<NodeVariable, NodeVariable> newVariables =
            constructNewVariables(oldVar, entry.getValue());
        final Hldd newHldd = constructAssignmentHldd(oldVar, newVariables);
        newHldds.add(newHldd);
      }
    }
    addNewProcess(parentModule, newHldds);
  }

  /*
   * Constructs a new variable for each process where the original variable is assigned
   * and substitutes it.
   */
  private Map<NodeVariable, NodeVariable> constructNewVariables(final NodeVariable oldVar,
      final List<ModuleProcess> processes) {
    final Map<NodeVariable, NodeVariable> newVariables = new HashMap<>();
    int idx = 0;
    for (ModuleProcess process : processes) {
      final NodeVariable newVar = getNewVariable(oldVar, ASSIGNMENT, idx);
      newVar.setUserData(oldVar);
      substituteOldVariable(oldVar, newVar, (HlddProcess) process);
      final NodeVariable maskVar = getNewVariable(oldVar, MASK, idx++);
      if (!parentModule.declares(maskVar)) {
        parentModule.declareVariable(maskVar);
      }
      addMaskAssignment(process, oldVar, newVar, maskVar);
      newVariables.put(newVar, maskVar);
    }
    return newVariables;
  }

  /*
   * Constructs a new variable.
   */
  private NodeVariable getNewVariable(final NodeVariable oldVar, final String infix,
      final int index) {
    final String newVarName = String.format("%s_%s_%s", oldVar.getName(), infix, index);
    return new NodeVariable(newVarName, oldVar.getDataType());
  }

  /*
   * Substitutes the original variable in it's HLDD.
   */
  private void substituteOldVariable(final NodeVariable oldVar, final NodeVariable newVar,
      final HlddProcess process) {
    final Hldd oldHldd = process.getHldd(oldVar);
    oldHldd.setVariable(newVar);
    if (parentModule.declares(oldVar)) {
      final Declaration oldDeclaration = parentModule.getDeclaration(oldVar);
      parentModule
          .declareVariable(newVar, oldDeclaration.getInitialValue(), oldDeclaration.getInvariant());
    } else {
      parentModule.declareVariable(newVar);
    }
  }

  private void addMaskAssignment(final ModuleProcess process, final NodeVariable oldVar,
      final NodeVariable newVar, final NodeVariable maskVar) {
    final Hldd maskHldd = new Hldd(maskVar);
    final Node nextNewVar = getNext(newVar);
    Node extend = null;
    if (process.isInitial()) {
      final NodeVariable initialFlag = getInitialFlag(process);
      DataType varType = newVar.getDataType();
      if (varType.isTypeId(DataTypeId.MAP)) {
        varType = (DataType) varType.getParameters()[1];
      }
      final int bitSize = varType.getSize();
      extend = Nodes.bvzeroext(bitSize - 1, initialFlag);
    }
    final Node arg = (oldVar.getUserData() instanceof Node) ? (Node) oldVar.getUserData() : oldVar;
    final Node bvxor = Nodes.bvxor(arg, nextNewVar);
    final Node value = extend == null ? bvxor : Nodes.mul(bvxor, extend);
    final HlddTerminalNode node = new HlddTerminalNode(maskVar, value);
    node.setLevel(0);
    maskHldd.addNode(node);
    maskHldd.setRootNode(node);
    ((HlddProcess) process).add(maskHldd);
  }

  /*
   * Returns the initial flag variable (it is true only at the initial moment of time).
   */
  private NodeVariable getInitialFlag(final ModuleProcess process) {
    final String flagName = String.format("%s_%s", INITIAL_FLAG, process.getName());
    final NodeVariable flag = new NodeVariable(flagName, DataType.bitVector(1));
    if (!parentModule.declares(flag)) {
      parentModule.declareVariable(flag, NodeValue.newBitVector(1, 1));
      ((HlddProcess) process).add(getInitalFlagHldd(flag));
    }
    return flag;
  }

  /*
   * Constructs the initial flag variable assignment HLDD.
   */
  private Hldd getInitalFlagHldd(final NodeVariable flag) {
    final Hldd hldd = new Hldd(flag);
    final HlddTerminalNode assignment =
        new HlddTerminalNode(flag, NodeValue.newBitVector(0, 1));
    assignment.setLevel(0);
    hldd.addNode(assignment);
    hldd.setRootNode(assignment);
    return hldd;
  }

  /*
   * Constructs the original variable assignment HLDD using substitution variables.
   */
  private Hldd constructAssignmentHldd(final NodeVariable oldVar,
      final Map<NodeVariable, NodeVariable> newVariables) {
    final Hldd newHldd = new Hldd();
    newHldd.setVariable(oldVar);
    final List<Node> nodes = new ArrayList<>();
    for (Map.Entry<NodeVariable, NodeVariable> variableEntry : newVariables.entrySet()) {
      nodes.add(getMaskedNext(variableEntry.getKey(), variableEntry.getValue()));
    }
    final Node oldValue = oldVar.getUserData() instanceof NodeOperation
        ? (NodeOperation) oldVar.getUserData()
        : oldVar;
    final Node maskedOldVal =
        Nodes.bvand(oldValue, Nodes.bvnot(Nodes.bvor(getNext(newVariables.values()))));
    nodes.add(maskedOldVal);
    final Node newValNode = Nodes.bvor(nodes);
    final HlddTerminalNode rootNode = new HlddTerminalNode(oldVar, newValNode);
    rootNode.setLevel(0);
    newHldd.addNode(rootNode);
    newHldd.setRootNode(rootNode);
    return newHldd;
  }

  private Node getMaskedNext(final NodeVariable newVar, final NodeVariable mask) {
    return Nodes.bvand(getNext(mask), getNext(newVar));
  }

  private Node getNext(final NodeVariable var) {
    // FIXME: implement normal NEXT operation in printer.
    return new NodeOperation(/*SmvReservedWords.Operation.NEXT*/StandardOperation.FUNCTION, var);
  }

  private Collection<Node> getNext(final Collection<NodeVariable> variables) {
    final Collection<Node> result = new ArrayList<>();
    for (NodeVariable variable : variables) {
      result.add(getNext(variable));
    }
    return result;
  }

  @Override
  public void onModuleBegin(final Module module) {
    parentModule = module;
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    parentProcess = process;
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    final NodeVariable variable = hldd.getVariable();
    if (parentModule.declares(variable) || variable.getUserData() instanceof Node) {
      addToMap(assignments, variable, parentProcess);
    }
  }

  @Override
  public void onModuleEnd(final Module module) {
    transformMultipleAssignments();
  }

  /*
   * Adds a new process to the module with a specified name and a list of HLDDs.
   */
  private void addNewProcess(final Module parentModule, final List<Hldd> hldds) {
    if (!hldds.isEmpty()) {
      final HlddProcess newProcess = new HlddProcess(DUPLICATE_VAR_PROCESS);
      parentModule.add(newProcess);
      hldds.forEach(newProcess::add);
    }
  }

  private <T> void addToMap(final Map<NodeVariable, List<T>> map, final NodeVariable key,
      final T value) {
    final List<T> list = map.containsKey(key) ? map.get(key) : new ArrayList<>();
    list.add(value);
    map.put(key, list);
  }

}
