/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.formula;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.version.VersionUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddBuilder;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This visitor constructs a formula-based HLDD from a GADD.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class GaddHlddFormulaVisitor extends CfgEmptyVisitor implements HlddBuilder {

  private static final String GUARD = "guard";

  // The phase variable of the current process.
  private NodeVariable currentPhaseVar;
  // The current case condition.
  private Node currentCase;
  // The HlddModule stack.
  private final LinkedList<Module> moduleStack;
  // The current HlddProcess.
  private HlddProcess hlddProcess;
  // The resulting model.
  private final HlddModel hlddModel;
  // The list of phase values leading to the current GADD branch.
  private List<Integer> prevPhases;
  // The least phase in a switch.
  private int basicPhase;
  // The variables defined in a process.
  private Set<NodeVariable> definedVariables;
  // The version-value map of a process.
  private Map<Node, Node> versionValuesMap;
  // The phase-block map of a process.
  private Map<Integer, BlockBuilder> blocksMap;
  // The map of substituted phases.
  private Map<Integer, Integer> phaseMap;
  // The map of variable assignments in a switch.
  private Map<NodeVariable, Map<BlockBuilder, Node>> switchAssignments;
  // The use/def condition collection flag.
  private boolean collectUseDef;
  // The condition collector class.
  private UseDefCollector collector;

  /**
   * Constructs a visitor.
   */
  public GaddHlddFormulaVisitor() {
    super();
    hlddModel = new HlddModel();
    moduleStack = new LinkedList<>();
    collectUseDef = false;
    collector = new UseDefCollector();
    basicPhase = -1;
  }

  /**
   * Enable or disable use/def condition collection.
   *
   * @param collectUseDef - the specified value.
   */
  public void setCollectUseDef(final boolean collectUseDef) {
    this.collectUseDef = collectUseDef;
  }

  @Override
  public HlddModel getModel() {
    return hlddModel;
  }

  @Override
  public void onModelEnd(final Model model) {
    if (collectUseDef) {
      hlddModel.setUseDefMap(collector.getUseDefMap());
    }
  }

  @Override
  public void onModuleBegin(final Module module) {
    final HlddModule hlddModule = new HlddModule(module);
    moduleStack.push(hlddModule);
  }

  @Override
  public void onModuleEnd(final Module module) {
    final Module hlddModule = moduleStack.pop();
    if (moduleStack.isEmpty()) {
      hlddModel.addModule(hlddModule);
    } else {
      getCurrentModule().add(hlddModule);
    }
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    hlddProcess = new HlddProcess(process);
    final Module currentModule = getCurrentModule();
    currentModule.add(hlddProcess);
    currentPhaseVar = GaddHlddUtils.getPhaseVariable(currentModule, process);
    if (currentPhaseVar != null) {
      currentModule.removeDeclaration(currentPhaseVar);
    }
    definedVariables = new HashSet<>();
    versionValuesMap = new HashMap<>();
    blocksMap = new LinkedHashMap<>();
    phaseMap = new HashMap<>();
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    addHldds();

    for (BlockBuilder block : sortBlocks()) {
      refineUses(block);
    }

    hlddProcess.setInvariant(getInvariant());

    if (collectUseDef) {
      collector.updateContainer(process);
    }
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    if (!isPhase(switchNode)) {
      final List<Integer> nextPhases = getChildPhases(switchNode);
      basicPhase = nextPhases.get(0);
      for (int i = 1; i < nextPhases.size(); i++) {
        phaseMap.put(nextPhases.get(i), basicPhase);
      }
      switchAssignments = new HashMap<>();
    }
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {
    if (!isPhase(switchNode)) {
      mergeBlocks();
      basicPhase = -1;
      switchAssignments = null;
    }
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {
    if (isPhase(caseNode)) {
      prevPhases = getPrevPhases(caseNode);
      currentCase = Nodes.TRUE;
    } else {
      if (caseNode.isDefault()) {
        currentCase = getDefaultCaseCondition(caseNode);
      } else {
        currentCase = getCaseCondition(caseNode);
      }
      if (collectUseDef) {
        collector.collectUsesGuard(hlddProcess, currentCase, getPrevGuardsDisj(prevPhases));
      }
    }
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    if (isPhase(caseNode)) {
      prevPhases = null;
    }
    currentCase = null;
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {
    final Map<Node, Node> mapping = new HashMap<>();
    int phase = 0;

    for (Map.Entry<NodeVariable, Node> entry : getAssignmentMap(block.getAssignments())
        .entrySet()) {

      if (entry.getKey().equals(currentPhaseVar)) {
        phase = getInt((NodeValue) entry.getValue());
      } else {
        mapping.put(entry.getKey(), entry.getValue());
        definedVariables.add(entry.getKey());
      }
    }

    final BlockBuilder blockNode = new BlockBuilder(phase, prevPhases);
    final Node guardExpr = labelVariables(currentCase, prevPhases);
    blockNode.addUses(ExprUtils.getVariables(guardExpr));
    final NodeVariable guardExt = getGuardVariable(phase);
    final Node guardExtExpr = getGuardExtExpression(guardExpr, phase, prevPhases);
    versionValuesMap.put(guardExt, guardExtExpr);

    if (basicPhase == -1) {
      setBlockDefsUses(blockNode, mapping);
    } else {
      collectAssignments(blockNode, mapping);
    }

    if (collectUseDef) {
      collector.collectUseDefs(hlddProcess, mapping, guardExt);
    }

    blocksMap.put(phase, blockNode);
  }

  /*
   * Get the assignment map from a collection of assignments.
   */
  private Map<NodeVariable, Node> getAssignmentMap(final Collection<Assignment> assignments) {
    InvariantChecks.checkNotNull(assignments);
    final Map<NodeVariable, Node> mapping = new LinkedHashMap<>();
    for (Assignment assignment : assignments) {
      final Node value = assignment.getValue();
      final NodeVariable variable = assignment.getVariable();
      mapping.put(variable, value);
    }
    return mapping;
  }

  /*
   * Sort blocks by phase value.
   */
  private List<BlockBuilder> sortBlocks() {
    final List<BlockBuilder> sortedBlocks = new LinkedList<>(blocksMap.values());
    sortedBlocks.sort(Comparator.comparingInt(BlockBuilder::getPhase));
    return sortedBlocks;
  }

  /*
   * Construct the current process invariant.
   */
  private Node getInvariant() {
    final Collection<Node> assignments = new LinkedList<>();
    for (Map.Entry<Node, Node> entry : versionValuesMap.entrySet()) {
      assignments.add(Nodes.eq(entry.getKey(), entry.getValue()));
    }
    return Nodes.and(assignments);
  }

  /*
   * Collect ancestor blocks' phases.
   */
  private List<Integer> getChildPhases(final CfgSwitchStatement switchStatement) {
    final List<Integer> phases = new LinkedList<>();
    for (int i = 0; i < switchStatement.getStatementNum(); i++) {
      final CfgCaseStatement caseStatement = (CfgCaseStatement) switchStatement.getStatement(i);
      final CfgBlockStatement blockStatement = (CfgBlockStatement) caseStatement.getStatement(0);
      final int phase = getPhase(blockStatement.getAssignments());
      if (phase != -1) {
        phases.add(phase);
      }
    }
    phases.sort(Integer::compareTo);
    return phases;
  }

  /*
   * Search for a phase value in a collection of assignments.
   */
  private int getPhase(final Collection<Assignment> assignments) {
    for (Assignment assignment : assignments) {
      if (assignment.getVariable().equals(currentPhaseVar)) {
        return getInt((NodeValue) assignment.getValue());
      }
    }
    return -1;
  }

  /*
   * Collect assignments of a block.
   */
  private void collectAssignments(final BlockBuilder blockNode, final Map<Node, Node> mapping) {

    for (Map.Entry<Node, Node> entry : mapping.entrySet()) {
      if (entry.getKey() instanceof NodeVariable) {
        final NodeVariable variable = (NodeVariable) entry.getKey();
        final Node value = entry.getValue();
        final Map<BlockBuilder, Node> assignments;

        if (switchAssignments.containsKey(variable)) {
          assignments = switchAssignments.get(variable);
        } else {
          assignments = new HashMap<>();
        }
        assignments.put(blockNode, value);

        switchAssignments.put(variable, assignments);
      }
    }
  }

  /*
   * Merge assignments of switch blocks into one block.
   */
  private void mergeBlocks() {
    for (Map.Entry<NodeVariable, Map<BlockBuilder, Node>> entry : switchAssignments.entrySet()) {
      final NodeVariable variable = entry.getKey();
      final Map<BlockBuilder, Node> assignments = entry.getValue();

      if (!assignments.isEmpty()) {
        final HashMap<Node, Node> varAssignment = new HashMap<>();
        if (variable.isType(DataTypeId.MAP)) {
          final Integer[] ranges = getRanges(variable);
          if (ranges != null) {
            // Split at-once array assignment into array elements assignments.
            for (int i = ranges[0]; i <= ranges[1]; i++) {
              final Node index = NodeValue.newInteger(i);
              varAssignment.put(Nodes.select(variable, index),
                  getIte(variable, index, assignments.entrySet().iterator(), new HashMap<>()));
            }
          }
        } else {
          varAssignment.put(variable, getIte(variable, assignments.entrySet().iterator()));
        }

        setBlockDefsUses(blocksMap.get(basicPhase), varAssignment);
      }
    }
  }

  /*
   * If not defined, define the variables used in the given block.
   */
  private void refineUses(final BlockBuilder block) {
    final Collection<NodeVariable> variables = block.getAllUses();
    for (NodeVariable variable : variables) {
      if (!versionValuesMap.containsKey(variable) && variable
          .getUserData() instanceof PhasedDescriptor) {
        final NodeVariable original = ((PhasedDescriptor) variable.getUserData()).getOriginal();
        final List<Integer> defPhases = whereDef(original, block.getPrevPhases(), new HashSet<>());
        final Node versionValue = getPreviousVar(defPhases, original);
        versionValuesMap.put(variable, versionValue);
      }
    }
  }

  /*
   * Get definition variable using definition phases list.
   */
  private NodeVariable getPreviousVar(final List<Integer> defPhases, final NodeVariable original) {
    if (!defPhases.isEmpty()) {
      Collections.sort(defPhases);
      final NodeVariable choice = getVarVersion(original, defPhases);
      if (!versionValuesMap.containsKey(choice)) {
        versionValuesMap.put(getIte(original, choice, defPhases, defPhases.size() - 1), Nodes.TRUE);
      }
      return choice;
    } else {
      return original;
    }
  }

  /*
   * Search for variable definition phases.
   */
  private List<Integer> whereDef(final NodeVariable variable, final List<Integer> phases,
      final Set<Integer> visited) {
    final List<Integer> defList = new LinkedList<>();
    for (Integer prevPhase : phases) {
      final int phase = phaseMap.getOrDefault(prevPhase, prevPhase);
      if (!visited.contains(phase) && (phase != 0)) {
        visited.add(phase);
        defList.addAll(whereDef(variable, phase, visited));
      }
    }
    return defList;
  }

  /*
   * Search for variable definition phases.
   */
  private List<Integer> whereDef(final NodeVariable variable, final int currentPhase,
      final Set<Integer> visited) {
    final List<Integer> defList = new LinkedList<>();
    final BlockBuilder block = blocksMap.get(currentPhase);
    if (block.defines(variable)) {
      defList.add(currentPhase);
    } else {
      defList.addAll(whereDef(variable, block.getPrevPhases(), visited));
    }
    return defList;
  }

  /*
   * Construct a guarding expression of a given phase.
   */
  private Node getGuardExpression(final int phase) {
    final Set<Integer> subPhases =
        phaseMap.entrySet().stream().filter(entry -> entry.getValue() == phase)
            .collect(HashSet::new, (set, entry) -> set.add(entry.getKey()),
                AbstractCollection::addAll);

    subPhases.add(phase);
    final Set<Node> guards = new HashSet<>();
    for (Integer subPhase : subPhases) {
      guards.add(getGuardVariable(subPhase));
    }
    return guards.size() > 1 ? Nodes.or(guards) : guards.iterator().next();
  }

  /*
   * Construct the following ite expression for phases [n,...,m]:
   *
   * guard_n_ext ? (choice = var_n) : (guard_n-1_ext ? (choice = var_n-1) : (
   * ... guard_m_ext ? (choice = var_m) : (choice = original_var))
   */
  private Node getIte(final NodeVariable original, final NodeVariable choice,
      final List<Integer> phases, final int position) {
    if (position >= 0) {
      final int phase = phases.get(position);
      final Node guardExpr = getGuardExpression(phase);
      final NodeVariable version = getVarVersion(original, phase);
      final Node expr = Nodes.eq(choice, version);
      return Nodes.ite(guardExpr, expr, getIte(original, choice, phases, position - 1));
    } else {
      return Nodes.eq(choice, original);
    }
  }

  /*
   * Construct the following ite expression for blocks labeled by phases [n,...,m]
   * & original variable values:
   *
   * guard_n_ext ? value_n : (guard_n-1_ext ? value_n-1 : (
   * ... guard_m_ext ? value_m : original_var)
   */
  private Node getIte(final Node original, final Iterator<Map.Entry<BlockBuilder, Node>> iterator) {
    if (iterator.hasNext()) {
      final Map.Entry<BlockBuilder, Node> entry = iterator.next();
      final int phase = entry.getKey().getPhase();
      final NodeVariable extGuard = getGuardVariable(phase);
      return Nodes.ite(extGuard, entry.getValue(), getIte(original, iterator));
    } else {
      return original;
    }
  }

  /*
   * Construct the ite expression for a rhs of an array element assignment in a merged block.
   */
  private Node getIte(final NodeVariable original, final Node index,
      final Iterator<Map.Entry<BlockBuilder, Node>> iterator,
      final Map<BlockBuilder, Node> guardSrcMap) {

    if (iterator.hasNext()) {
      final Map.Entry<BlockBuilder, Node> entry = iterator.next();
      final int phase = entry.getKey().getPhase();
      final NodeVariable extGuard = getGuardVariable(phase);
      final Node condition;
      final Node value;

      // assigned value is a STORE
      if (GaddHlddUtils.isStore(entry.getValue())) {
        final NodeOperation store = (NodeOperation) entry.getValue();
        final Node source = store.getOperand(0);
        final Node srcIndex = store.getOperand(1);
        final int intVal = getInt((NodeValue) index);
        final Node castIndex = NodeUtils.getValue(intVal, srcIndex.getDataType());
        value = store.getOperand(2);
        condition = Nodes.and(extGuard, Nodes.eq(srcIndex, castIndex));
        guardSrcMap.put(entry.getKey(), Nodes.select(source, index));

        // assigned value is an array
      } else if (entry.getValue() instanceof NodeVariable && entry.getValue()
          .isType(DataTypeId.MAP)) {
        condition = extGuard;
        value = Nodes.select(entry.getValue(), index);

      } else {
        condition = extGuard;
        value = entry.getValue();
      }

      return Nodes.ite(condition, value, getIte(original, index, iterator, guardSrcMap));

    } else {
      // default value
      return getIte(Nodes.select(original, index), guardSrcMap.entrySet().iterator());
    }
  }

  /*
   * Mark variables of a node by the given phases, declare obtained versions.
   */
  private Node labelVariables(final Node node, final List<Integer> phases) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule rule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {
        return node.getKind().equals(Node.Kind.VARIABLE);
      }

      @Override
      public Node apply(final Node node) {
        if (node instanceof NodeVariable) {
          return getVarVersion((NodeVariable) node, phases);
        }
        return node;
      }
    };
    transformer.addRule(Node.Kind.VARIABLE, rule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  /*
   * Construct final HLDDs.
   */
  private void addHldds() {
    for (NodeVariable variable : definedVariables) {
      if (!isSsaVersion(variable) && !variable.equals(currentPhaseVar)) {
        final List<Integer> finalPhases = blocksMap.get(0).getPrevPhases();
        final NodeVariable finalVersion = getVarVersion(variable, finalPhases);
        blocksMap.get(0).addUse(finalVersion);
        if (variable.isType(DataTypeId.MAP)) {
          for (int i = 0; i < GaddHlddUtils.getArraySize(variable); i++) {
            final Node index = NodeValue.newInteger(i);
            final NodeVariable element = VersionUtils.getArrayElement(variable, index);
            final Node select = Nodes.select(finalVersion, index);
            addHldd(element, select);
          }
        } else {
          addHldd(variable, finalVersion);
        }
      }
    }
  }

  /*
   * Add a HLDD to the current process.
   */
  private void addHldd(final NodeVariable variable, final Node value) {
    final Hldd hldd = new Hldd(variable);
    final HlddTerminalNode node = new HlddTerminalNode(variable, value);
    hldd.addNode(node);
    hldd.setRootNode(node);
    hlddProcess.add(hldd);
  }

  /*
   * Get the current module.
   */
  private Module getCurrentModule() {
    final Module module = moduleStack.peek();
    if (module == null) {
      throw new RetrascopeException("Module stack is empty!");
    } else {
      return module;
    }
  }

  /*
   * Label variables by phases, add assignments to the version-value map,
   * set a block's defs & uses.
   */
  private void setBlockDefsUses(final BlockBuilder block, final Map<Node, Node> mapping) {
    for (Map.Entry<Node, Node> entry : mapping.entrySet()) {
      final Node variable = entry.getKey();
      final Node assignment = entry.getValue();
      final Node transformedAssignment = labelVariables(assignment, prevPhases);
      final Node nextPhaseVersion = getNextPhaseVersion(variable, block.getPhase());
      final NodeVariable definedVariable = getDefinedVariable(variable);

      if (nextPhaseVersion != null && definedVariable != null) {
        block.addDefine(definedVariable);
        block.addUses(ExprUtils.getVariables(transformedAssignment));
        if (nextPhaseVersion instanceof NodeVariable && nextPhaseVersion.isType(DataTypeId.MAP)
            && GaddHlddUtils.isStore(transformedAssignment)) {
          addArray((NodeVariable) nextPhaseVersion, (NodeOperation) transformedAssignment);
        } else {
          versionValuesMap.put(nextPhaseVersion, transformedAssignment);
        }
      }
    }
  }

  /*
   * Label the specified node by the specified phase.
   */
  private Node getNextPhaseVersion(final Node node, final int phase) {
    if (node instanceof NodeVariable) {
      return getVarVersion((NodeVariable) node, phase);
    } else if (GaddHlddUtils.isSelect(node)) {
      final List<Integer> phases = new ArrayList<>();
      phases.add(phase);
      return labelVariables(node, phases);
    }
    return null;
  }

  /*
   * If the specified node is a variable, return it; if the specified node is a select operation,
   * return the array variable; return null otherwise.
   */
  private NodeVariable getDefinedVariable(final Node node) {
    if (node instanceof NodeVariable) {
      return (NodeVariable) node;
    } else if (GaddHlddUtils.isSelect(node)) {
      return ExprUtils.getVariables(node).iterator().next();
    }
    return null;
  }

  /*
   * Add an array assignment to the version-value map.
   */
  private void addArray(final NodeVariable dst, final NodeOperation store) {
    final Node source = store.getOperand(0);
    final Node index = store.getOperand(1);
    final Node value = store.getOperand(2);
    final Integer[] ranges = getRanges(dst);
    if (ranges != null) {
      for (int i = ranges[0]; i <= ranges[1]; i++) {
        addElement(dst, source, value, index, i);
      }
    }
  }

  /*
   * Extract variable ranges, if possible.
   */
  private Integer[] getRanges(final NodeVariable variable) {
    final VariableDescriptor descriptor = getDescriptor(variable);
    if (descriptor != null) {
      final List<Range> ranges = descriptor.getIndexRanges();
      if (ranges != null && !ranges.isEmpty()) {
        final Range range = ranges.get(0);
        final int minRange = getInt((NodeValue) range.getLow());
        final int maxRange = getInt((NodeValue) range.getHigh());
        return new Integer[] {minRange, maxRange};
      }
    }
    return null;
  }

  /*
   * Put dst[dstIndex] = dstIndex == srcIndex ? value : source[dstIndex] to version-value map.
   */
  private void addElement(final Node dst, final Node source, final Node value, final Node srcIndex,
      final int dstIndex) {
    final NodeValue currentIndex = getValue(srcIndex.getDataType(), dstIndex);
    final Node selectSrc = Nodes.select(source, currentIndex);
    final Node selectDst = Nodes.select(dst, currentIndex);
    final Node expr = getArrayElemValue(value, selectSrc, srcIndex, currentIndex);
    versionValuesMap.put(selectDst, expr);
  }

  /*
   * Construct a rhs value for array element assignment.
   */
  private Node getArrayElemValue(final Node value, final Node defaultValue, final Node srcIndex,
      final Node dstIndex) {
    if (srcIndex instanceof NodeValue) {
      if (getInt((NodeValue) dstIndex) == getInt((NodeValue) srcIndex)) {
        return value;
      } else {
        return defaultValue;
      }
    } else {
      final Node cond = Nodes.eq(srcIndex, dstIndex);
      return Nodes.ite(cond, value, defaultValue);
    }
  }

  /*
   * Get a variable descriptor.
   */
  private VariableDescriptor getDescriptor(final NodeVariable variable) {
    final Object userData = variable.getUserData();
    if (userData instanceof PhasedDescriptor) {
      return getDescriptor(((PhasedDescriptor) userData).getOriginal());
    }
    if (userData instanceof VariableDescriptor) {
      return (VariableDescriptor) userData;
    } else {
      return null;
    }
  }

  /*
   * Get a correct NodeValue from an int value.
   */
  private NodeValue getValue(final DataType targetType, final int intValue) {
    if (targetType.isTypeId(DataTypeId.BIT_VECTOR)) {
      final int size = targetType.getSize();
      return NodeValue.newBitVector(intValue, size);
    } else if (targetType.isTypeId(DataTypeId.LOGIC_INTEGER)) {
      return NodeValue.newInteger(intValue);
    } else {
      throw new IllegalArgumentException(
          String.format("Can't transform Integer to %s", targetType.toString()));
    }
  }

  /*
   * Get an int value from a NodeValue.
   */
  private int getInt(final NodeValue value) {
    if (value.isType(DataTypeId.LOGIC_INTEGER)) {
      return value.getInteger().intValue();
    } else if (value.isType(DataTypeId.BIT_VECTOR)) {
      return value.getBitVector().bigIntegerValue(false).intValue();
    } else {
      return ((NodeValue) TransformerUtils.toInt(value)).getInteger().intValue();
    }
  }


  /*
   * Get a list of phase values that mark a case statement.
   */
  private List<Integer> getPrevPhases(final CfgCaseStatement caseNode) {
    final List<Integer> result = new LinkedList<>();
    if (caseNode.isDefault()) {
      throw new RetrascopeException("Phase switch has a default case!");
    } else {
      for (NodeValue value : caseNode.getValues()) {
        result.add(getInt(value));
      }
    }
    Collections.sort(result);
    return result;
  }

  /*
   * Get a version of the specified variable and the specified phase.
   */
  private NodeVariable getVarVersion(final NodeVariable variable, final int phase) {
    final LinkedList<Integer> list = new LinkedList<>();
    list.add(phase);
    return getVarVersion(variable, list);
  }

  /*
   * Get a version of the specified variable and the specified phases list.
   */
  private NodeVariable getVarVersion(final NodeVariable variable, final List<Integer> phases) {

    if (prevPhaseIs0(phases) || isInput(variable) || isSsaVersion(variable) || isPhasedVersion(
        variable)) {
      // Return original variable if the phase value is 0, it is input or ssa or phased version.
      return variable;
    } else {
      final PhasedDescriptor descriptor = new PhasedDescriptor(variable, phases);
      return getVariable(getPhasedName(variable.getName(), phases), variable.getDataType(),
          descriptor);
    }
  }

  /*
   * A list of phases contains only phase 0.
   */
  private boolean prevPhaseIs0(final List<Integer> phases) {
    return phases.size() == 1 && phases.get(0) == 0;
  }

  /*
   * A variable is an input.
   */
  private boolean isInput(final NodeVariable variable) {
    return getCurrentModule().declaresInput(variable.getName());
  }

  /*
   * A variable is a SSA version.
   */
  private boolean isSsaVersion(final NodeVariable variable) {
    return GaddHlddUtils.getOriginalVariable(variable) != null;
  }

  /*
   * A variable is labeled by phase value.
   */
  private boolean isPhasedVersion(final NodeVariable variable) {
    return variable.getUserData() instanceof PhasedDescriptor;
  }

  /*
   * Get a variable by name and declare it, if not yet declared.
   */
  private NodeVariable getVariable(final String varName, final DataType dataType,
      final Object userData) {
    final Module currentModule = getCurrentModule();
    if (currentModule.declares(varName)) {
      return currentModule.getVariable(varName);
    } else {
      final NodeVariable variable = new NodeVariable(varName, dataType);
      variable.setUserData(userData);
      currentModule.declareVariable(variable);
      return variable;
    }
  }

  /*
   * Get a guard or an extended guard variable.
   */
  private NodeVariable getGuardVariable(final int phase) {
    final PhasedDescriptor descriptor = new PhasedDescriptor();
    final String guardName = String.format("%s_%s_%s", GUARD, currentPhaseVar.getName(), phase);
    return getVariable(guardName, DataType.BOOLEAN, descriptor);
  }

  /*
   * Construct a name for a variable version.
   */
  private String getPhasedName(final String varName, final List<Integer> phases) {
    final StringBuilder nameBuilder = new StringBuilder();
    nameBuilder.append(String.format("%s_%s", varName, currentPhaseVar.getName()));
    for (int phase : phases) {
      nameBuilder.append(String.format("_%s", phase));
    }
    return nameBuilder.toString();
  }

  /*
   * Constructs the expression for guard_ext variable (guard_n_ext = expr & prev_gurads_ext).
   */
  private Node getGuardExtExpression(final Node guardExpr, final int phase,
      final List<Integer> previousPhases) {
    if (phase == 1) {
      // Return guard_1 to avoid cyclic dependencies, because phase 1 is
      // the beginning of the process.
      return guardExpr;
    } else {
      return Nodes.and(guardExpr, getPrevGuardsDisj(previousPhases));
    }
  }

  /*
   * Get a disjunction of extended guards for the given phases.
   */
  private Node getPrevGuardsDisj(final List<Integer> previousPhases) {
    final List<NodeVariable> prevGuards = new LinkedList<>();
    for (Integer prevPhase : previousPhases) {
      prevGuards.add(getGuardVariable(prevPhase));
    }
    return prevGuards.size() == 1 ? prevGuards.get(0) : Nodes.or(prevGuards);
  }

  /*
   * Get a condition for a default case.
   */
  private Node getDefaultCaseCondition(final CfgCaseStatement defCase) {
    final CfgSwitchStatement parent = (CfgSwitchStatement) defCase.getParent();
    final List<Node> cases = new LinkedList<>();
    for (int i = 0; i < parent.getStatementNum(); i++) {
      final CfgCaseStatement caseStatement = (CfgCaseStatement) parent.getStatement(i);
      if (!caseStatement.isDefault()) {
        cases.add(getCaseCondition(caseStatement));
      }
    }
    return Nodes.not(Nodes.or(cases));
  }

  /*
   * Get a case condition based on the current guard and the case values.
   */
  private Node getCaseCondition(final CfgCaseStatement caseNode) {
    final Set<NodeValue> values = caseNode.getValues();
    final Node condition = ((CfgSwitchStatement) caseNode.getParent()).getCondition();
    if (values.size() > 0) {
      final List<Node> conditions = new LinkedList<>();
      for (NodeValue value : values) {
        conditions.add(Nodes.eq(condition, value));
      }
      return conditions.size() > 1 ? Nodes.or(conditions) : conditions.get(0);
    } else {
      throw new RetrascopeException(
          String.format("Non-default case of %s has no values!", condition));
    }
  }

  /*
   * Check whether the specified case is marked by a phase variable.
   */
  private boolean isPhase(final CfgCaseStatement caseNode) {
    return isPhase((CfgSwitchStatement) caseNode.getParent());
  }

  private boolean isPhase(final CfgSwitchStatement switchNode) {
    return switchNode.getCondition().equals(currentPhaseVar);
  }

}
