/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.formula;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.result.usedef.UseDefContainer;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class is used to collect use/def information.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class UseDefCollector {

  Map<NodeVariable, UseDefContainer> getUseDefMap() {
    return useDefMap;
  }

  private Map<NodeVariable, UseDefContainer> useDefMap;

  private Map<NodeVariable, Set<Node>> currentUses;
  private Map<NodeVariable, Set<Node>> currentDefs;

  UseDefCollector() {
    useDefMap = new HashMap<>();
    clearMaps();
  }

  /**
   * Collects use and define conditions in a basic block.
   *
   * @param process - the specified process.
   * @param mapping - the specified assignment mapping of a basic block.
   * @param guardExt - the specified block guard.
   */
  void collectUseDefs(final ModuleProcess process, final Map<Node, Node> mapping,
      final Node guardExt) {
    collectDefsBlock(process, mapping, guardExt);
    collectUsesBlock(process, mapping, guardExt);
  }

  /*
   * Collects defines in block assignments.
   */
  private void collectDefsBlock(final ModuleProcess process, final Map<Node, Node> mapping,
      final Node guardExt) {
    for (Node node : mapping.keySet()) {
      if (node instanceof NodeVariable) {
        addCondition(process, (NodeVariable) node, guardExt, currentDefs);
      }
    }
  }

  /*
   * Collects uses in block assignments.
   */
  private void collectUsesBlock(final ModuleProcess process, final Map<Node, Node> mapping,
      final Node guardExt) {
    final Set<NodeVariable> usedVariables = new HashSet<>();
    for (Node node : mapping.values()) {
      for (NodeVariable variable : ExprUtils.getVariables(node)) {
        if (!usedVariables.contains(variable)) {
          usedVariables.add(getOriginal(variable));
        }
      }
    }
    usedVariables.forEach(variable -> addCondition(process, variable, guardExt, currentUses));
  }

  /*
   * Returns the original variable, if possible.
   */
  private NodeVariable getOriginal(final NodeVariable variable) {
    if (variable.getUserData() instanceof PhasedDescriptor) {
      return  ((PhasedDescriptor) variable.getUserData()).getOriginal();
    } else if (variable.getUserData() instanceof VariableDescriptor && !notSsaVersion(
        variable)) {
      return  ((VariableDescriptor) variable.getUserData()).getOriginalVariable();
    }
    return variable;
  }

  /*
   * A variable is not an input.
   */
  private boolean notInput(final NodeVariable variable, final ModuleProcess process) {
    return !((Module) process.getParent()).declaresInput(variable.getName());
  }

  /*
   * A variable is not an SSA version.
   */
  private boolean notSsaVersion(final NodeVariable variable) {
    return GaddHlddUtils.getOriginalVariable(variable) == null;
  }

  /**
   * Collects use conditions in a guard condition.
   *
   * @param process - the specified process.
   * @param guardCondition - the specified guard condition.
   * @param prevCondition - the specified previous guards condition.
   */
  void collectUsesGuard(final ModuleProcess process, final Node guardCondition,
      final Node prevCondition) {
    final Collection<NodeVariable> variables = ExprUtils.getVariables(guardCondition);
    for (NodeVariable variable : variables) {
      addCondition(process, variable, prevCondition, currentUses);
    }
  }

  /*
   * Adds a condition to the specified def/use map.
   */
  private void addCondition(final ModuleProcess process, final NodeVariable variable,
      final Node condition, final Map<NodeVariable, Set<Node>> conditionMap) {
    if (notInput(variable, process) && notSsaVersion(variable)) {
      final Set<Node> conditions =
          conditionMap.containsKey(variable) ? conditionMap.get(variable) : new HashSet<>();
      conditions.add(condition);
      conditionMap.put(variable, conditions);
    }
  }

  /*
   * Adds the collected use & define conditions to the use/def map.
   */
  void updateContainer(final ModuleProcess process) {
    updateDefs(process);
    updateUses(process);
    clearMaps();
  }

  /*
   * Adds the collected define conditions to the use/def map.
   */
  private void updateDefs(final ModuleProcess process) {
    for (Map.Entry<NodeVariable, Set<Node>> useEntry : currentUses.entrySet()) {
      final Node useCondition = Nodes.or(useEntry.getValue());
      final UseDefContainer container = getContainer(useEntry.getKey());
      container.putUseCondition(process, useCondition);
      useDefMap.put(useEntry.getKey(), container);
    }
  }

  /*
   * Adds the collected use conditions to the use/def map.
   */
  private void updateUses(final ModuleProcess process) {
    for (Map.Entry<NodeVariable, Set<Node>> defEntry : currentDefs.entrySet()) {
      final Node defCondition = Nodes.or(defEntry.getValue());
      final UseDefContainer container = getContainer(defEntry.getKey());
      container.putDefineCondition(process, defCondition);
      useDefMap.put(defEntry.getKey(), container);
    }
  }

  /*
   * Inits the current use & def maps.
   */
  private void clearMaps() {
    currentDefs = new HashMap<>();
    currentUses = new HashMap<>();
  }

  /*
   * Returns a use/def container for a variable.
   */
  private UseDefContainer getContainer(final NodeVariable variable) {
    if (useDefMap.containsKey(variable)) {
      return useDefMap.get(variable);
    } else {
      return new UseDefContainer();
    }
  }
}
