/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.version;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;

/**
 * This class represents a version of a variable: it contains the original variable, the version
 * variable, the version number and the current value.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class Version {

  private NodeVariable version;
  private NodeVariable original;
  private Node[] values;
  private int versionNumber;

  /**
   * Constructs a new Version of the specified variable.
   *
   * @param versionVar  - the specified version variable.
   * @param originalVar - the specified original variable.
   * @param num         - the specified version number.
   * @param size        - the specified quantity of values (e.g. array size)
   */
  public Version(final NodeVariable versionVar, final NodeVariable originalVar, final int num,
      final int size) {
    if (versionVar.equals(originalVar) && num != 0) {
      throw new IllegalStateException("Version number of the original variable can only be 0!");
    }
    version = versionVar;
    original = originalVar;
    values = new Node[size];
    versionNumber = num;
  }

  /**
   * Indicates that this Version represents the original variable.
   *
   * @return {@code true} if this Version represents the original variable, {@code false} if
   *     this Version represents some version of the original variable.
   */
  public boolean isOriginal() {
    return version.equals(original);
  }

  /**
   * Returns the corresponding version variable.
   *
   * @return the corresponding {@link NodeVariable}.
   */
  public NodeVariable getVariable() {
    return version;
  }

  /**
   * Returns the corresponding original variable.
   *
   * @return the corresponding {@link NodeVariable}.
   */
  public NodeVariable getOriginal() {
    return original;
  }

  /**
   * Return the current values of the Version.
   *
   * @return the corresponding array of current values.
   */
  public Node[] getValues() {
    return values;
  }

  /**
   * Returns the version number.
   *
   * @return the version number. {@code 0} indicates the original variable.
   */
  public int getVersionNumber() {
    return versionNumber;
  }

}
