/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.version;

import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents a registry that contains variables and their versions.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class VersionRegistry {

  private final Map<NodeVariable, Versions> versions = new HashMap<>();

  /**
   * Constructs a registry.
   */
  public VersionRegistry() {
  }

  /**
   * Adds the specified versions container to the registry.
   *
   * @param vars - the specified {@link Versions} object.
   */
  public void addVersions(final Versions vars) {
    versions.put(vars.getOriginalVariable(), vars);
  }

  /**
   * Returns the current values of the specifed variable.
   *
   * @param variable - the specifed variable.
   * @return {@code null} if the specified variable does not belong to this registry, a list of
   *     current values otherwise.
   */
  public List<Node> getValues(final NodeVariable variable) {
    if (versions.containsKey(variable)) {
      return versions.get(variable).getValues(variable);
    } else {
      final NodeVariable original = GaddHlddUtils.getOriginalVariable(variable);
      if (versions.containsKey(original)) {
        return versions.get(original).getValues(variable);
      }
    }
    return null;
  }

  /**
   * Indicates whether the registry contains the specified variable.
   *
   * @param variable - the specified variable.
   * @return {@code true} if the registry contains the specified variable, {@code false} otherwise.
   */
  public boolean contains(final NodeVariable variable) {
    return getValues(variable) != null;
  }

  /**
   * Returns the versions container of the specified variable.
   *
   * @param originalVariable - the specified variable.
   * @return the {@link Versions} container if the registry contains the specified variable,
   *     {@code null} otherwise.
   */
  public Versions getVersions(final NodeVariable originalVariable) {
    return versions.get(originalVariable);
  }

  /**
   * Inits the contents of the registry.
   */
  public void initAllValues() {
    versions.values().forEach(Versions::initValues);
  }

  /**
   * Updates the value of the specified variable.
   *
   * @param variable - the specified variable.
   * @param index    - the specified index of the value.
   * @param value    - the specified value.
   */
  public void updateValue(final NodeVariable variable, final int index, final Node value) {
    final Versions vars = getVersions(GaddHlddUtils.getOriginalVariable(variable));
    if (vars != null) {
      vars.updateValue(variable, index, value);
    }
  }

  /**
   * Indicates whether the specified variable is the original variable.
   *
   * @param variable - the specified variable.
   * @return {@code true} if the registry contains the specified original variable, {@code false}
   *     otherwise.
   */
  private boolean isOriginal(final Node variable) {
    return versions.keySet().contains(variable);
  }

  /**
   * Returns all versions of all variables.
   *
   * @return the collection of version variables.
   */
  public Collection<NodeVariable> getAllVersions() {
    final Collection<NodeVariable> result = new ArrayList<>();
    for (Versions vars : versions.values()) {
      result.addAll(vars.getVersions());
    }
    return result;
  }


  /**
   * Searches for and assigns the values of the specified right-hand side variable to the specified
   * original variable. Then the values of the subsequent variables are searched. The search is
   * performed until the final values do not contain any version variables. The variable-value
   * map is filled during the search.
   *
   * @param original - the left-hand side of the assignment.
   * @param rhs      - the right-hand side of the assignment.
   * @return The variable-value map of the assigned variables.
   */
  public Map<NodeVariable, Node> collectAssignments(final NodeVariable original,
      final NodeVariable rhs) {
    if (original.getUserData() instanceof VariableDescriptor && rhs
        .getUserData() instanceof VariableDescriptor) {
      final VariableDescriptor lhsDescriptor = (VariableDescriptor) original.getUserData();
      final VariableDescriptor rhsDescriptor = (VariableDescriptor) rhs.getUserData();
      if (!lhsDescriptor.getIndexRanges().equals(rhsDescriptor.getIndexRanges())) {
        throw new RetrascopeException("Array types are not equal!");
      }
    }

    final List<Node> originalValues = getValues(original);
    final List<Node> rhsValues = getValues(rhs);

    final Versions versions = getVersions(original);

    final Map<NodeVariable, Node> mapping = new HashMap<>();
    for (int i = 0; i < versions.getSize(); i++) {
      final Node finalValue = getFinalValue(rhs, rhsValues.get(i));
      originalValues.set(i, finalValue);
      mapping.put(VersionUtils.getArrayElement(original, NodeValue.newInteger(i)), finalValue);
    }

    return collectSubsequent(rhs, mapping);
  }

  /*
   * Searches for the subsequent assignments.
   */
  private Map<NodeVariable, Node> collectSubsequent(final NodeVariable rhs,
      final Map<NodeVariable, Node> values) {
    final Map<NodeVariable, Node> arrayItems = new HashMap<>();
    final ExprTreeWalker walker = new ExprTreeWalker(new ExprTreeVisitorDefault() {

      @Override
      public void onOperationBegin(final NodeOperation node) {
        if (GaddHlddUtils.isSelect(node)) {
          final Node src = node.getOperand(0);
          final Node index = node.getOperand(1);
          if (src instanceof NodeVariable && index instanceof NodeValue && !isOriginal(src)
              && getValues((NodeVariable) src) != null && !src.equals(rhs)) {
            final int intIndex = VersionUtils.tryGetIntIndex(index);
            final Node finalNode = getFinalValue(rhs, getValues((NodeVariable) src).get(intIndex));
            arrayItems.put(VersionUtils.getArrayElement((NodeVariable) src, index), finalNode);
          }
        }
      }
    });
    walker.visit(values.values());

    if (arrayItems.isEmpty()) {
      return values;
    } else {
      values.putAll(collectSubsequent(rhs, arrayItems));
      return values;
    }
  }

  /*
   * Searches for the final value on the specified place.
   */
  private Node getFinalValue(final NodeVariable dst, final Node value) {
    if (GaddHlddUtils.isSelect(value)) {
      final Node src = ((NodeOperation) value).getOperand(0);
      final Node prevIndex = Reducer.reduce(((NodeOperation) value).getOperand(1));
      final List<Node> prevValues = getValues((NodeVariable) src);
      if (prevIndex instanceof NodeValue) {
        final int prevIntIndex = VersionUtils.tryGetIntIndex(prevIndex);
        final Node prevValue = prevValues != null ? prevValues.get(prevIntIndex) : null;
        if (!isOriginal(src) && !value.equals(prevValue) && prevValue != null && !src.equals(dst)) {
          return getFinalValue(dst, prevValue);
        }
      }
    }
    return value;
  }

}
