/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.property;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.HlddClockedVisitorBase;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.ModelCheckingMethod;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvReservedWords;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.property.ModuleProperty;
import ru.ispras.retrascope.result.property.ProcessProperty;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * This class represents a visitor for property-smv printer.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddPropertySmvVisitor extends HlddClockedVisitorBase {

  private boolean printCtl;

  private NodeVariable currentPhaseVar;

  protected Map<NodeVariable, Node> inputs;

  private Collection<GuardedAction> resets;

  // Model properties.
  protected Map<String, ModuleProperty> properties;
  // Current property container (module).
  private ModuleProperty curPropertyContainer;
  // Current property group (process).
  private ProcessProperty curAssertGroup;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddPropertySmvVisitor(final String fileName) {
    super(fileName);
    printCtl = false;
    inputs = new HashMap<>();
    resets = new ArrayList<>();
    properties = new HashMap<>();
  }

  /**
   * Sets the type of specifications to be checked.
   *
   * @param method The specification method.
   */
  public void setSpecType(final String method) {
    switch (ModelCheckingMethod.getEnumValue(method)) {
      case CTL_BDD:
        printCtl = true;
        break;
      case LTL_BMC:
      case LTL_MSAT:
        printCtl = false;
        break;
      case PSL_BDD:
      case PSL_BMC:
        printCtl = false;
        Log.warn("Option %s cannot be applied to %s. Using ltl-bmc instead.", method,
            HlddPropertySmvPrinter.ENGINE_ID);
        break;
      default:
        throw new RetrascopeException("Unknown model checking method: " + method);
    }
  }

  /**
   * Adds model properties to the visitor.
   *
   * @param assertions - the specified properties.
   */
  public void addAssertions(final Map<String, ModuleProperty> assertions) {
    this.properties = assertions;
  }

  @Override
  public void onModuleBegin(final Module hlddModule) {
    setCurPropertyContainer(hlddModule.getName());
    super.onModuleBegin(hlddModule);
    if (curPropertyContainer != null && curPropertyContainer.getClock() != null) {
      for (ProcessProperty group : curPropertyContainer.getProcessMap().values()) {
        if (group.getRstGuardedAction() != null && group.getRstGuardedAction().size() > 0) {
          resets.addAll(group.getRstGuardedAction());
        }
      }
    }
    for (final NodeVariable input : hlddModule.getInputs()) {
      inputs.put(input, hlddModule.getDeclaration(input).getInitialValue());
    }
  }

  @Override
  protected NodeVariable getClock() {
    if (curPropertyContainer != null) {
      return curPropertyContainer.getClock();
    }
    return null;
  }

  @Override
  public void onModuleEnd(final Module hlddModule) {
    resets = new ArrayList<>();
  }


  @Override
  public void onProcessBegin(final ModuleProcess process) {
    setCurAssertGroup(process.getName());
    super.onProcessBegin(process);

    if (process.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
      final String phaseName = process.getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);
      currentPhaseVar = currentModule.getVariable(phaseName);
    } else {
      currentPhaseVar = null;
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {
    if (curAssertGroup != null) {
      printAssertions();
    }
    super.onProcessEnd(process);
  }

  /*
   * Searches for ModuleProperty using the specified key name. If found, assigns the value to
   * the corresponding variable, assigns null otherwise.
   */
  private void setCurPropertyContainer(final String name) {
    curPropertyContainer = properties.getOrDefault(name, null);
  }

  /*
   * Sets the current property group.
   */
  private void setCurAssertGroup(final String name) {
    if (curPropertyContainer != null) {
      final Map<String, ProcessProperty> groups = curPropertyContainer.getProcessMap();
      curAssertGroup = groups.getOrDefault(name, null);
    }
  }

  /*
   * Prints the current property group.
   */
  private void printAssertions() {
    final String assertionDecl =
        printCtl ? SmvReservedWords.CTLSPEC.toString() : SmvReservedWords.LTLSPEC.toString();
    final String futureOperator = printCtl ? "EF" : "F";
    for (final Node node : curAssertGroup.getProperties()) {
      append(assertionDecl);

      // Reachability condition.
      final String reachCond = String.format("(%s %s)", futureOperator,
          getExprDescription(node));

      if (resets.size() == 0) {
        appendLine(String.format("  ! %s;", reachCond));
      } else {
        // Initial reset condition.
        final Node resetNode = getResetCondition();
        final String rstCond =
            String.format("(%s)", getExprDescription(resetNode));

        if (printCtl) {
          // Input initial values condition.
          final String inputsCond =
              getExprDescription(getInitialInputsCondition());
          appendLine(String.format(" ! E [(%s) U (%s & %s)];", inputsCond, rstCond, reachCond));
        } else {
          appendLine(String.format(" ! ( X %s & %s);", rstCond, reachCond));

        }
      }
    }
    appendLine();
  }

  /*
   * Generates the reset condition based on current reset guarded actions.
   */
  private Node getResetCondition() {
    final List<Node> guards = new ArrayList<>();
    for (GuardedAction guardedAction : resets) {
      final Node guard = guardedAction.getGuardNode();
      final Set<NodeVariable> uses = guardedAction.getGuardUses();
      if (currentPhaseVar == null || uses.size() > 1) {
        guards.add(guard);
      }
    }
    final Node resetGuardNode = ExprUtils.getConjunction(guards.toArray(new Node[0]));
    Node initStateNode = curAssertGroup.getInitValueState();
    final Node resetNode;
    if (initStateNode.getKind() == Node.Kind.VALUE && initStateNode.isType(DataType.BOOLEAN)) {
      resetNode = resetGuardNode;
    } else {
      resetNode = ExprUtils.getConjunction(initStateNode, resetGuardNode);
    }
    return resetNode;
  }

  /*
   * Generates the initial values condition.
   */
  private Node getInitialInputsCondition() {
    final Collection<Node> equations = new ArrayList<>();
    for (final Entry<NodeVariable, Node> var : inputs.entrySet()) {
      if (var.getValue() != null) {
        equations.add(Nodes.eq(var.getKey(), var.getValue()));
      }
    }
    return ExprUtils.getConjunction(equations.toArray(new Node[0]));
  }

}
