/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.usedef;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddNonTerminalNode;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;
import ru.ispras.retrascope.result.usedef.UseDefContainer;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/**
 * This visitor adds various checker processes to the HLDD model.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddUseDefCheckerVisitor extends HlddEmptyVisitor {

  private Map<NodeVariable, UseDefContainer> useDefMap;
  private HlddModule currentModule;
  private CheckerVariables checkerVariables;


  private enum Type {
    USE("use"), DEF("def");

    public String name;

    Type(final String name) {
      this.name = name;
    }
  }

  public void setUseDefMap(final Map<NodeVariable, UseDefContainer> useDefMap) {
    this.useDefMap = useDefMap;
  }

  /**
   * Constructs a new visitor.
   */
  public HlddUseDefCheckerVisitor() {
    checkerVariables = new CheckerVariables();
    checkerVariables.counters = new HashMap<>();
    checkerVariables.useFlags = new HashMap<>();
    checkerVariables.defFlags = new HashMap<>();
  }

  public CheckerVariables getCheckerVariables() {
    return checkerVariables;
  }

  @Override
  public void onModuleBegin(final Module module) {
    currentModule = (HlddModule) module;
    checkerVariables.counters.put(module, new HashSet<>());
    checkerVariables.useFlags.put(module, new HashMap<>());
    checkerVariables.defFlags.put(module, new HashMap<>());
  }

  @Override
  public void onModuleEnd(final Module module) {
    generateCheckers();
  }

  private void generateCheckers() {
    for (Map.Entry<NodeVariable, UseDefContainer> useDefMapEntry : useDefMap.entrySet()) {
      addRwCheckerProcess(useDefMapEntry.getKey(), useDefMapEntry.getValue());
      addWwCheckerProcess(useDefMapEntry.getKey(), useDefMapEntry.getValue());
    }
  }

  /*
   * Creates the write-write checker process for the specified variable based on a define event
   * counter.
   */
  private void addWwCheckerProcess(final NodeVariable variable,
      final UseDefContainer useDefContainer) {
    if (useDefContainer.getDefines().size() > 1) {
      final String processName = String.format("%s_ww_checker_process", variable.getName());
      final ModuleProcess process = new HlddProcess(processName);
      currentModule.add(process);
      addCounterHldd(process, declareCounter(variable), useDefContainer.getDefines().values());
    }
  }

  private NodeVariable declareCounter(final NodeVariable variable) {
    final String counterName = String.format("%s_def_counter", variable.getName());
    final NodeVariable counter = new NodeVariable(counterName, DataType.bitVector(2));
    currentModule.declareVariable(counter, NodeValue.newBitVector(0, 2));
    checkerVariables.counters.get(currentModule).add(counter);
    return counter;
  }

  /*
   * Creates the CFG for the specified counter variable.
   */
  private void addCounterHldd(final ModuleProcess process, final NodeVariable counter,
      final Collection<Node> defConditions) {
    final Node counterPlusOne = Nodes.add(counter, NodeValue.newBitVector(1, 2));

    final Hldd counterHldd = new Hldd(counter);
    final HlddConditionNode initialSwitch = new HlddConditionNode(null, getZeroedPhases());
    final HlddEdge trueEdge = getTrueCase(initialSwitch);
    final HlddEdge falseEdge = getFalseCase(initialSwitch);
    final HlddTerminalNode zeroCnt =
        new HlddTerminalNode(trueEdge, counter, NodeValue.newBitVector(0, 2));

    initialSwitch.addChild(trueEdge);
    initialSwitch.addChild(falseEdge);
    trueEdge.addChild(zeroCnt);

    ((HlddProcess) process).add(counterHldd);
    counterHldd.addNode(initialSwitch).addNode(trueEdge).addNode(falseEdge).addNode(zeroCnt);

    counterHldd.setRootNode(initialSwitch);

    HlddNode prevNode = null;
    for (Node defCondition : defConditions) {
      final HlddNode parent = prevNode == null ? falseEdge : prevNode;
      final HlddConditionNode defSwitch = new HlddConditionNode(parent, defCondition);
      final HlddEdge defTrue = getTrueCase(defSwitch);
      final HlddEdge defFalse = getFalseCase(defSwitch);
      final HlddTerminalNode incCnt = new HlddTerminalNode(defTrue, counter, counterPlusOne);

      ((HlddNonTerminalNode) parent).addChild(defSwitch);
      defSwitch.addChild(defTrue);
      defSwitch.addChild(defFalse);
      defTrue.addChild(incCnt);

      counterHldd.addNode(defSwitch).addNode(defTrue).addNode(defFalse).addNode(incCnt);

      prevNode = defFalse;
    }
    if (prevNode != null) {
      ((HlddNonTerminalNode) prevNode).addChild(new HlddTerminalNode(prevNode, counter, counter));
    } //else {
    //falseEdge.addChild(new HlddTerminalNode(falseEdge, counter, counter));
    //}
  }

  /*
   * Creates the read-write checker process for the specified variable based on the use/define event
   * flags.
   */
  private void addRwCheckerProcess(final NodeVariable variable,
      final UseDefContainer useDefContainer) {
    if (useDefContainer.getDefines().size() > 0 && useDefContainer.getUses().size() > 0
        && !HlddUseDefUtils.useDefSameProcess(useDefContainer)) {
      addFlags(Type.DEF, variable, useDefContainer.getDefines());
      addFlags(Type.USE, variable, useDefContainer.getUses());
    }
  }

  private void addFlags(final Type type, final NodeVariable variable, final Map<String, Node> map) {
    final Map<NodeVariable, Map<String, NodeVariable>> flags =
        declareFlags(variable, map.keySet(), type);
    if (flags.size() > 0) {
      addFlagAssignmentProcess(variable, flags, type, map);
    }
  }

  /*
   * Creates the flags of the specified type and adds them to the current module.
   */
  private Map<NodeVariable, Map<String, NodeVariable>> declareFlags(final NodeVariable variable,
      final Collection<String> names, final Type flagType) {
    final Map<NodeVariable, Map<String, NodeVariable>> varProcessFlagMap = new HashMap<>();
    final Map<String, NodeVariable> flags = new HashMap<>();
    for (String processName : names) {
      final String flagName =
          String.format("%s_%s_in_%s", variable.getName(), flagType.name, processName);
      final NodeVariable flag = new NodeVariable(flagName, DataType.BOOLEAN);
      currentModule.declareVariable(flag, Nodes.FALSE);
      flags.put(processName, flag);
    }
    varProcessFlagMap.put(variable, flags);
    switch (flagType) {
      case DEF:
        checkerVariables.defFlags.get(currentModule).putAll(varProcessFlagMap);
        break;
      case USE:
        checkerVariables.useFlags.get(currentModule).putAll(varProcessFlagMap);
        break;
      default:
        break;
    }
    return varProcessFlagMap;
  }

  /*
   * Creates the assignment processes for the specified variable use/define flags.
   */
  private void addFlagAssignmentProcess(final NodeVariable variable,
      final Map<NodeVariable, Map<String, NodeVariable>> flags, final Type flagType,
      final Map<String, Node> conditions) {
    for (Map<String, NodeVariable> variableMap : flags.values()) {
      for (Map.Entry<String, NodeVariable> entry : variableMap.entrySet()) {
        final String originalName = entry.getKey();
        final String processName =
            String.format("%s_%s_%s", variable.getName(), flagType.name, originalName);
        final ModuleProcess process = new HlddProcess(processName);
        currentModule.add(process);
        addFlagHldd(process, GaddHlddUtils
                .getPhaseVariable(currentModule, getProcessByName(currentModule, originalName)),
            entry.getValue(), conditions.get(originalName));
      }
    }
  }

  private ModuleProcess getProcessByName(final Module module, final String name) {
    for (ModuleProcess process : module.getProcesses()) {
      if (process.getName().compareTo(name) == 0) {
        return process;
      }
    }
    return null;
  }

  /*
   * Creates the assignment CFG for the specified flag variable.
   */
  private void addFlagHldd(final ModuleProcess process, final NodeVariable phaseVariable,
      final NodeVariable flag, final Node useDefCondition) {
    final Node assignmentCondition = Nodes.and(useDefCondition, Nodes.noteq(flag, Nodes.TRUE));
    final Node ite = Nodes.ite(assignmentCondition, Nodes.TRUE, flag /*Nodes.FALSE*/);

    final Hldd flagHldd = new Hldd(flag);
    ((HlddProcess) process).add(flagHldd);

    final HlddTerminalNode assignTrue = new HlddTerminalNode(flag, ite);

    if (phaseVariable != null) {
      final HlddConditionNode initialSwitch =
          new HlddConditionNode(null, Nodes.eq(phaseVariable, getZero(phaseVariable)));
      final HlddEdge trueEdge = getTrueCase(initialSwitch);
      final HlddEdge falseEdge = getFalseCase(initialSwitch);
      final HlddTerminalNode assignFalse = new HlddTerminalNode(trueEdge, flag, Nodes.FALSE);
      assignTrue.setParent(falseEdge);

      initialSwitch.addChild(trueEdge);
      initialSwitch.addChild(falseEdge);
      trueEdge.addChild(assignFalse);
      falseEdge.addChild(assignTrue);

      flagHldd.addNode(initialSwitch).addNode(trueEdge).addNode(falseEdge).addNode(assignTrue)
          .addNode(assignFalse);
      flagHldd.setRootNode(initialSwitch);

    } else {
      flagHldd.addNode(assignTrue);
      flagHldd.setRootNode(assignTrue);
    }
  }

  /*
   * Returns the 'all-zero' condition for the current module's phase variables.
   */
  private Node getZeroedPhases() {
    Node result = null;
    for (ModuleProcess process : currentModule.getProcesses()) {
      final NodeVariable variable = GaddHlddUtils.getPhaseVariable(currentModule, process);
      if (variable != null) {
        final Node zero = getZero(variable);
        final Node eq = Nodes.eq(variable, zero);
        if (result != null) {
          result = Nodes.and(result, eq);
        } else {
          result = eq;
        }
      }
    }
    return result;
  }

  private HlddEdge getTrueCase(final HlddNode parent) {
    final HlddEdge edge = new HlddEdge(parent);
    edge.addExpression(Nodes.TRUE);
    return edge;
  }

  private HlddEdge getFalseCase(final HlddNode parent) {
    final HlddEdge edge = new HlddEdge(parent);
    edge.addExpression(Nodes.FALSE);
    return edge;
  }

  private Node getZero(final NodeVariable variable) {
    return NodeValue.newBitVector(0, variable.getDataType().getSize());
  }


}
