/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class contains static methods for proper MAP data type printing.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class MapUtils {

  // Represents MAP data type node-name pairs.
  private static Map<Node, String> arrays;
  // Counts MAP data type values in the program.
  private static int mapCounter;

  private MapUtils() {
    /* Empty. */
  }

  static {
    resetMapCounter();
    resetMaps();
  }

  /**
   * Resets MAP data type container.
   */
  public static void resetMaps() {
    arrays = new HashMap<>();
  }

  /**
   * Returns the MAP data type values with their SMV variable name.
   *
   * @return The map which keys are MAP data type nodes and values are corresponding variable names.
   */
  public static Map<Node, String> getMaps() {
    return arrays;
  }

  /**
   * Resets data maps counter.
   */
  private static void resetMapCounter() {
    mapCounter = 0;
  }

  /**
   * Searches for DataTypeId.MAP values in the specified node and replaces them by generated
   * array variables.
   *
   * @param node - the specified {@link Node}.
   * @return The node with substitution performed.
   */
  public static Node substituteMaps(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule mapRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.SELECT)
            && ((NodeOperation) arg0).getOperand(0).isType(DataTypeId.MAP);
      }

      @Override
      public Node apply(final Node arg0) {

        final Node map = ((NodeOperation) arg0).getOperand(0);
        Node index = ((NodeOperation) arg0).getOperand(1);
        if (map.getKind() == Kind.VALUE) {
          final DataMap dataMap = (DataMap) ((NodeValue) map).getValue();
          final NodeValue size = new NodeValue(Data.newInteger(dataMap.size()));
          index = Nodes.mod(index, size);

          if (!arrays.containsKey(map)) {
            final String mapVar = "map" + mapCounter;
            arrays.put(map, mapVar);
            mapCounter++;
          }
        }

        return Nodes.select(map, index);
      }

    };

    transformer.addRule(StandardOperation.SELECT, mapRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  /**
   * Returns the description of data arrays in SMV format using the specified printer.
   *
   * @param printer The SMV expression printer.
   *
   * @return String representation of data arrays.
   */
  public static String printMaps(final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();

    if (!arrays.isEmpty()) {
      result.append(SmvReservedWords.DEFINE.toString() + "\n");
    }

    for (Entry<Node, String> entry : arrays.entrySet()) {
      final DataMap dataMap = (DataMap) ((NodeValue) entry.getKey()).getValue();
      final String mapName = entry.getValue();

      result.append(String.format("  %s := [", mapName));

      int ind = 0;
      for (Entry<Data, Data> mapEntry : dataMap.entrySet()) {
        if (ind != 0) {
          result.append(", ");
        }
        result.append(printer.getConstant(mapEntry.getValue()));
        ind++;
      }
      result.append("];\n");
    }
    return result.toString();
  }

}
