/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.MapBasedPrinter;
import ru.ispras.fortress.expression.printer.OperationDescription.Type;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.formula.PhasedDescriptor;
import ru.ispras.retrascope.model.basis.VariableDescriptor;

import java.math.BigInteger;
import java.util.Collection;
import java.util.List;
import java.util.Stack;

/**
 * This class represents the printer into *.smv format.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvExprPrinter extends MapBasedPrinter {


  /**
   * The size of integer type.
   */
  private static final int INTEGER_SIZE = 32;
  /**
   * Substituted variable suffix.
   */
  public static final String SUB_ENDING = "_SUBSTITUTED";

  // Integer printing mode flag.
  private boolean intAsBv;
  // Hexadecimal bit vector printing flag.
  private boolean bvHex;
  // Binary bit vector printing flag.
  private boolean bvBin;
  // Decimal bit vector printing flag.
  private boolean bvDec;
  // Operation stack.
  private final Stack<Enum<?>> operationIdStack;
  // Current operand number.
  private int currentOperandIndex;
  // BVCONCAT operation flag.
  private boolean insideConcat;


  /**
   * Replaces lowercase {@code true} and {@code false} values with their SMV equivalents.
   *
   * @param string - the specified string.
   * @return A string with substitution performed.
   */
  public static String replaceLogicValues(final String string) {
    return string.replaceAll("true", SmvReservedWords.TRUE.toString())
        .replaceAll("false", SmvReservedWords.FALSE.toString());
  }

  /**
   * Types of NuXMV constraints.
   */
  public enum ConstraintType {
    FAIRNESS, INVAR, CTLSPEC, LTLSPEC, PSLSPEC
  }

  /**
   * Default constructor. Sets operation mappings.
   */
  public SmvExprPrinter() {
    // Supported operations.
    addMapping(StandardOperation.MINUS, "-", Type.PREFIX);
    addMapping(StandardOperation.PLUS, "+", Type.PREFIX);
    addMapping(StandardOperation.NOT, "!", Type.PREFIX, false);
    addMapping(StandardOperation.BVNOT, "!", Type.PREFIX, false);
    addMapping(StandardOperation.BVANDR, "&", Type.PREFIX);
    addMapping(StandardOperation.BVORR, "|", Type.PREFIX);
    addMapping(StandardOperation.BVXORR, "xor", Type.PREFIX);
    addMapping(StandardOperation.BVXNORR, "xnor", Type.PREFIX);
    addMapping(StandardOperation.BVUREM, "mod", Type.INFIX);
    addMapping(StandardOperation.MUL, "*", Type.INFIX);
    addMapping(StandardOperation.BVMUL, "*", Type.INFIX);
    addMapping(StandardOperation.DIV, "/", Type.INFIX);
    addMapping(StandardOperation.BVSDIV, "/", Type.INFIX);
    addMapping(StandardOperation.REM, "mod", Type.INFIX);
    addMapping(StandardOperation.MOD, "mod", Type.INFIX);
    addMapping(StandardOperation.BVSMOD, "mod", Type.INFIX);
    addMapping(StandardOperation.ADD, "+", Type.INFIX);
    addMapping(StandardOperation.BVADD, "+", Type.INFIX);
    addMapping(StandardOperation.SUB, "-", Type.INFIX);
    addMapping(StandardOperation.BVSUB, "-", Type.INFIX);
    addMapping(StandardOperation.BVNEG, "-", Type.PREFIX);
    addMapping(StandardOperation.BVLSHR, ">>", Type.INFIX);
    addMapping(StandardOperation.BVLSHL, "<<", Type.INFIX);
    addMapping(StandardOperation.BVASHR, ">>", Type.INFIX);
    addMapping(StandardOperation.BVASHL, "<<", Type.INFIX);
    addMapping(StandardOperation.LESS, "<", Type.INFIX);
    addMapping(StandardOperation.BVULT, "<", Type.INFIX);
    addMapping(StandardOperation.BVSLT, "<", Type.INFIX);
    addMapping(StandardOperation.LESSEQ, "<=", Type.INFIX);
    addMapping(StandardOperation.BVULE, "<=", Type.INFIX);
    addMapping(StandardOperation.BVSLE, "<=", Type.INFIX);
    addMapping(StandardOperation.GREATER, ">", Type.INFIX);
    addMapping(StandardOperation.BVUGT, ">", Type.INFIX);
    addMapping(StandardOperation.BVSGT, ">", Type.INFIX);
    addMapping(StandardOperation.GREATEREQ, ">=", Type.INFIX);
    addMapping(StandardOperation.BVUGE, ">=", Type.INFIX);
    addMapping(StandardOperation.BVSGE, ">=", Type.INFIX);
    addMapping(StandardOperation.EQ, "=", Type.INFIX);
    addMapping(StandardOperation.NOTEQ, "!=", Type.INFIX);
    addMapping(StandardOperation.EQCASE, "=", Type.INFIX);
    addMapping(StandardOperation.NOTEQCASE, "!=", Type.INFIX);
    addMapping(StandardOperation.BVAND, "&", Type.INFIX);
    addMapping(StandardOperation.BVXOR, "xor", Type.INFIX);
    addMapping(StandardOperation.BVXNOR, "xnor", Type.INFIX);
    addMapping(StandardOperation.BVOR, "|", Type.INFIX);
    addMapping(StandardOperation.AND, "&", Type.INFIX);
    addMapping(StandardOperation.OR, "|", Type.INFIX);
    addMapping(StandardOperation.IMPL, "->", Type.INFIX);
    addMapping(StandardOperation.ITE, new String[] {"?", ":"});
    addMapping(StandardOperation.BVCONCAT, "(", " :: ", ")");
    addMapping(StandardOperation.BVREPEAT, "(", "(", "))");
    addMapping(StandardOperation.BVEXTRACT, "", new String[] {"[", ":"}, "]", new int[] {2, 0, 1});
    addMapping(StandardOperation.SELECT, "", "[", "]");
    addMapping(StandardOperation.BV2BOOL, SmvReservedWords.TOBOOL.toString() + "(", "", ")");
    addMapping(StandardOperation.BOOL2BV, SmvReservedWords.TOBIT.toString() + "(", "", ")");
    addMapping(StandardOperation.BV2INT, SmvReservedWords.TOINT.toString() + "(", "", ")");
    addMapping(StandardOperation.INT2BV, SmvReservedWords.INT2BV.toString() + "(",
        new String[] {","}, ")", new int[] {1, 0});
    addMapping(StandardOperation.STORE, new String[] {"[", "] :="});
    addMapping(StandardOperation.BVZEROEXT, "extend(", new String[] {","}, ")", new int[] {1, 0});
    addMapping(StandardOperation.BVSIGNEXT, "extend(", new String[] {","}, ")", new int[] {1, 0});

    // SMV next operation hack.
    addMapping(StandardOperation.FUNCTION, "next(", "", ")");

    // Unsupported operations.
    addMapping(StandardOperation.BVROL, "BVROL(", ", ", ")");
    addMapping(StandardOperation.BVROR, "BVROR(", ", ", ")");
    addMapping(StandardOperation.BVUREM, "BVUREM(", ", ", ")");
    addMapping(StandardOperation.BVSREM, "BVSREM(", ", ", ")");

    addMapping(StandardOperation.ABS, "ABS(", ", ", ")");
    addMapping(StandardOperation.MIN, "MIN(", ", ", ")");
    addMapping(StandardOperation.MAX, "MAX(", ", ", ")");

    addMapping(StandardOperation.BVNANDR, "~&", Type.PREFIX);
    addMapping(StandardOperation.BVNORR, "~|", Type.PREFIX);
    addMapping(StandardOperation.POWER, "**", Type.INFIX);
    addMapping(StandardOperation.BVNAND, "~&", Type.INFIX);
    addMapping(StandardOperation.BVNOR, "~|", Type.INFIX);


    setVisitor(new Visitor());
    printIntsAsBv();
    setDec();

    operationIdStack = new Stack<>();
    insideConcat = false;
  }

  /**
   * Enables integer printing as 32-bit vector.
   */
  private void printIntsAsBv() {
    intAsBv = true;
  }

  /**
   * Disables integer printing as 32-bit vector.
   */
  public void printIntsNormal() {
    intAsBv = false;
  }

  /**
   * Sets binary bit vector printing mode.
   */
  public void setBin() {
    bvBin = true;
    bvDec = false;
    bvHex = false;
  }

  /**
   * Sets decimal bit vector printing mode.
   */
  private void setDec() {
    bvBin = false;
    bvDec = true;
    bvHex = false;
  }

  /**
   * Sets hexadecimal bit vector printing mode.
   */
  public void setHex() {
    bvBin = false;
    bvDec = false;
    bvHex = true;
  }

  /*
   * Returns the description of a constraint in SMV format using the given ConstraintType and node.
   */
  private String getConstraintDescription(final ConstraintType type, final Node node) {
    return type.name() + "\n  " + this.toString(node);
  }

  /**
   * Returns the description of a constraint in SMV format using the given {@link ConstraintType}
   * and the collection of nodes.
   *
   * @param type  - the specified constraint type.
   * @param nodes - the specified node collection.
   * @return The description string.
   */
  public String getConstraintDescription(final ConstraintType type, final Collection<Node> nodes) {
    final StringBuilder builder = new StringBuilder();
    for (Node node : nodes) {
      builder.append(getConstraintDescription(type, node));
    }
    return builder.toString();
  }

  /* This visitor is used to print expressions. */
  private final class Visitor extends ExprTreeVisitor {

    @Override
    public void onValue(final NodeValue value) {
      if (value.isType(DataTypeId.MAP)) {
        appendText(MapUtils.getMaps().get(value));
      } else {
        NodeValue coercedValue = null;
        final boolean intAsBvOld = intAsBv;
        final Enum<?> currentOperation =
            operationIdStack.isEmpty() ? null : operationIdStack.peek();

        if (!operationIdStack.isEmpty() && currentOperation == StandardOperation.BVEXTRACT) {
          intAsBv = false;
          if (currentOperandIndex != 0) {
            coercedValue = (NodeValue) TypeConversion.coerce(value, DataType.INTEGER);
          }
        }

        if (!operationIdStack.isEmpty() && (currentOperation == StandardOperation.BVZEROEXT
            || currentOperation == StandardOperation.BVSIGNEXT) && currentOperandIndex != 0) {
          intAsBv = false;
        }

        if (!operationIdStack.isEmpty() && (currentOperation == StandardOperation.BVASHL) && value
            .isType(DataType.INTEGER) && intAsBv) {
          coercedValue = (NodeValue) TypeConversion.coerce(value, DataType.bitVector(INTEGER_SIZE));
        }

        if (null != coercedValue) {
          appendText(getConstant(coercedValue.getData()));
        } else {
          appendText(getConstant(value.getData()));
        }
        intAsBv = intAsBvOld;
      }
    }

    @Override
    public void onOperationBegin(final NodeOperation operation) {
      if (insideConcat && operationIdStack.peek() == StandardOperation.BVCONCAT && operation
          .isType(DataType.BOOLEAN)) {
        appendText(SmvReservedWords.TOBIT.toString() + "(");
      }

      operationIdStack.push(operation.getOperationId());
      if (ExprUtils.isOperation(operation, StandardOperation.BVCONCAT)) {
        insideConcat = true;
      }

      super.onOperationBegin(operation);
    }

    @Override
    public void onOperationEnd(final NodeOperation operation) {
      super.onOperationEnd(operation);
      operationIdStack.pop();
      if (insideConcat && !operationIdStack.isEmpty()
          && operationIdStack.peek() == StandardOperation.BVCONCAT && operation
          .isType(DataType.BOOLEAN)) {
        appendText(")");
      }
      if (ExprUtils.isOperation(operation, StandardOperation.BVCONCAT) && (
          operationIdStack.isEmpty() || operationIdStack.peek() != StandardOperation.BVCONCAT)) {
        insideConcat = false;
      }
    }

    @Override
    public void onOperandBegin(final NodeOperation operation, final Node operand, final int index) {
      currentOperandIndex = index;
      if (ExprUtils.isOperation(operation, StandardOperation.ITE)) {
        if (index == 0 && operand.getDataTypeId() != DataTypeId.LOGIC_BOOLEAN) {
          appendText(SmvReservedWords.TOBOOL.toString() + "(");
        }
      }
      super.onOperandBegin(operation, operand, index);
    }

    @Override
    public void onOperandEnd(final NodeOperation operation, final Node operand, final int index) {
      super.onOperandEnd(operation, operand, index);
      if (ExprUtils.isOperation(operation, StandardOperation.ITE)) {
        if (index == 0 && operand.getDataTypeId() != DataTypeId.LOGIC_BOOLEAN) {
          appendText(")");
        }
      }
    }

    @Override
    public void onVariable(final NodeVariable variable) {
      String varName = SmvPrinterUtils.getName(variable);
      final Enum<?> currentOperation = operationIdStack.isEmpty() ? null : operationIdStack.peek();
      if (!operationIdStack.isEmpty() && (currentOperation == StandardOperation.BVASHL) && variable
          .isType(DataType.INTEGER) && intAsBv) {
        varName = String.format("%s(%s)", SmvReservedWords.UNSIGNED.toString(), varName);
      }
      if (insideConcat && currentOperation != null && currentOperation == StandardOperation.BVCONCAT
          && variable.isType(DataType.BOOLEAN)) {
        varName = String.format("%s(%s)", SmvReservedWords.TOBIT.toString(), varName);
      }
      appendText(varName);
    }

  }

  /**
   * Returns {@link Data} description in SMV format.
   *
   * @param node - the specified data node.
   * @return Data string description.
   */
  public String getConstant(final Data node) {
    String result = "";
    String signed = "";
    final DataType dataType = node.getType();

    if (dataType.getTypeId() == DataTypeId.LOGIC_BOOLEAN) {
      result =
          node.getBoolean() ? SmvReservedWords.TRUE.toString() : SmvReservedWords.FALSE.toString();
    } else {
      BigInteger value = BigInteger.ZERO;
      int size = 0;
      if (dataType.getTypeId() == DataTypeId.BIT_VECTOR) {
        value = node.getBitVector().bigIntegerValue(false);
        size = node.getBitVector().getBitSize();
      }
      if (dataType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
        value = node.getInteger();
        size = INTEGER_SIZE;
        signed = "s";
      }

      final String sign = (value.signum() == -1) ? "-" : "";

      if (intAsBv || dataType.getTypeId() != DataTypeId.LOGIC_INTEGER) {
        if (bvBin) {
          result = sign + "0" + signed + "b" + size + "_" + value.abs().toString(2);
        }
        if (bvDec) {
          result = sign + "0" + signed + "d" + size + "_" + value.abs().toString(10);
        }
        if (bvHex) {
          result = sign + "0" + signed + "h" + size + "_" + value.abs().toString(16);
        }
      } else {
        result = value.toString();
      }
    }
    return result;
  }

  /**
   * Returns a data type in SMV format.
   *
   * @param dataType - the specified {@link DataType}.
   * @return The data type string.
   */
  public String getType(final DataType dataType) {
    return getType(dataType, null);
  }

  /**
   * Returns variable's data type in SMV format.
   *
   * @param variable - the specified {@link NodeVariable}.
   * @return The data type string.
   */
  public String getType(final NodeVariable variable) {
    final Object userData = variable.getUserData();
    if (userData instanceof PhasedDescriptor) {
      final NodeVariable original = ((PhasedDescriptor) userData).getOriginal();
      if (original != null) {
        return getType(original);
      }
    }
    if (userData instanceof VariableDescriptor) {
      final VariableDescriptor descriptor = (VariableDescriptor) userData;
      final List<Range> ranges = descriptor.getIndexRanges();
      return getType(variable.getDataType(), ranges);
    } else {
      return getType(variable.getDataType());
    }
  }

  /**
   * Returns a data type in SMV format.
   *
   * @param dataType - the specified {@link DataType}.
   * @param ranges   - the specified map size (if the dataType corresponds to map).
   * @return The data type string.
   */
  public String getType(final DataType dataType, final List<Range> ranges) {
    final StringBuilder result = new StringBuilder();
    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        result.append(String.format("%s[%s]", SmvReservedWords.BIT_VECTOR.toString(),
            dataType.getSize()));
        break;
      case LOGIC_BOOLEAN:
        result.append(SmvReservedWords.BOOLEAN.toString());
        break;
      case LOGIC_INTEGER:
        if (intAsBv) {
          result.append(String.format("%s %s[%s]", SmvReservedWords.SIGNED.toString(),
              SmvReservedWords.BIT_VECTOR.toString(), INTEGER_SIZE));
        } else {
          result.append(SmvReservedWords.INTEGER.toString());
        }
        break;
      case MAP:
        final Object[] types = dataType.getParameters();
        final DataType keyType = (DataType) types[0];
        final DataType valueType = (DataType) types[1];

        if (null == ranges || ranges.isEmpty()) {
          result.append(String.format("%s 0..", SmvReservedWords.ARRAY.toString()));
          if (keyType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
            result.append(Integer.MAX_VALUE);
          }
          if (keyType.getTypeId() == DataTypeId.BIT_VECTOR) {
            final BigInteger upperRange =
                BigInteger.valueOf(2).pow(keyType.getSize()).subtract(BigInteger.ONE);
            result.append(upperRange.toString());
          }
          result.append(" of ");
        } else {
          for (Range range : ranges) {
            if (range.isIntegerRange()) {
              final BigInteger minRange = ((NodeValue) range.getLow()).getInteger();
              final BigInteger maxRange = ((NodeValue) range.getHigh()).getInteger();
              result.append(String
                  .format("%s %s..%s of ", SmvReservedWords.ARRAY.toString(), minRange.toString(),
                      maxRange.toString()));
            } else {
              throw new IllegalArgumentException(
                  String.format("%s is not an integer range!", range.toString()));
            }
          }
        }

        result.append(String.format("%s", getType(valueType)));
        break;
      default:
        result.append(SmvReservedWords.BOOLEAN.toString());
        break;
    }
    return result.toString();
  }

}
