/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class contains static methods for *.smv file printing.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvPrinterUtils {

  /* Statement end with double new line. */
  private static final String ENDLINE = ";\n\n";
  private static final String BASE = "  ";

  private SmvPrinterUtils() {
    /* Empty. */
  }

  /**
   * Generates a module declaration in SMV format.
   *
   * @param name   - the specified module name.
   * @param inputs - the specified collection of parameters.
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name,
      final Collection<NodeVariable> inputs) {

    final StringBuilder result = new StringBuilder();
    result.append(String.format("%s %s", SmvReservedWords.MODULE.toString(), name));
    if (!inputs.isEmpty()) {
      result.append(String.format("(%s)", getModuleParametersDescription(inputs)));
    }
    result.append("\n");
    return result.toString();
  }

  /**
   * Generates a module declaration in SMV format with empty parameter list.
   *
   * @param name - the specified module name.
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name) {
    return getModuleDeclaration(name, new ArrayList<>());
  }

  /**
   * Generates a module instantiation in SMV format.
   *
   * @param instanceName - the specified module instance name.
   * @param moduleName   - the specified module name.
   * @param parameters   - the specified collection of parameters.
   * @return The string containing the module instantiation.
   */
  public static String getModuleInstantiation(final String instanceName, final String moduleName,
      final Collection<NodeVariable> parameters) {
    return String.format("%s%n%s : %s(%s);%n", SmvReservedWords.VARIABLE.toString(), instanceName,
        moduleName, getModuleParametersDescription(parameters));
  }

  /**
   * Generates process instantiations in SMV format.
   *
   * @param processes      - the specified process collection.
   * @param inputVariables - the specified input variable collection.
   * @return The string containing the process instantiations.
   */
  public static String getProcessInstantiation(final Collection<ModuleProcess> processes,
      final Collection<NodeVariable> inputVariables) {
    final StringBuilder builder = new StringBuilder();
    for (final ModuleProcess process : processes) {
      final String processName = getName(process.getName());
      final String instanceName = "  m_" + processName;
      builder.append(
          SmvPrinterUtils.getModuleInstantiation(instanceName, processName, inputVariables));
    }
    return builder.toString();
  }

  /*
   * Generates a list of module parameters from a collection of variables.
   */
  private static String getModuleParametersDescription(final Collection<NodeVariable> parameters) {
    final StringBuilder result = new StringBuilder();
    final ArrayList<NodeVariable> variables = new ArrayList<>(parameters);
    for (int j = 0; j < variables.size(); j++) {
      if (!(variables.get(j).getUserData() instanceof NodeOperation)) {
        if (j != 0) {
          result.append(", ");
        }
        result.append(getName(variables.get(j)));
      }
    }
    return result.toString();
  }

  /**
   * Checks whether the variable name is acceptable by SMV syntax.
   *
   * @param variable - the specified variable.
   * @return The modified variable name if it is forbidden by SMV, lower case variable name
   *     otherwise.
   */
  public static String getName(final NodeVariable variable) {
    return getName(variable.getName());
  }

  /**
   * Checks whether the specified name is acceptable by SMV syntax.
   *
   * @param name - the specified name.
   * @return The modified name if it is forbidden by SMV, lower case variable name
   *     otherwise.
   */
  public static String getName(final String name) {
    final String formattedName = getFormattedName(name);
    if (Arrays.asList(SmvReservedWords.RESERVED_WORDS).contains(formattedName)) {
      return formattedName + SmvExprPrinter.SUB_ENDING;
    }
    return formattedName;
  }

  /**
   * Generates a VAR declaration section in SMV format from a collection of variables.
   *
   * @param variables - the variable-declaration mapping
   * @param printer   - the specified {@link SmvExprPrinter}.
   * @return The string containing the VAR section.
   */
  public static String getVarDeclareSection(final Map<NodeVariable, Declaration> variables,
      final SmvExprPrinter printer) {
    return getVarsDescription(variables, SmvReservedWords.VARIABLE.toString(), printer);
  }

  /*
   * Generates a variable declaration in SMV format.
   */
  private static String getVariableDeclaration(final NodeVariable variable,
      final Declaration declaration, final SmvExprPrinter printer, final String keyword) {
    final StringBuilder result = new StringBuilder();
    final String varName = getName(variable);
    result.append(String.format("%s%n", keyword));
    result.append(String.format("  %s : %s%s", varName, printer.getType(variable), ENDLINE));
    result.append(getInvar(variable, declaration, printer));
    result.append(getInit(variable, declaration, varName, printer));
    return result.toString();
  }

  /*
   * Generates variable's invariant in SMV format.
   */
  private static String getInvar(final NodeVariable variable, final Declaration declaration,
      final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
    if (declaration != null && declaration.containsInvariant()) {
      result.append(String.format("%s%n", SmvReservedWords.INVAR.toString()));
      final Node invariant =
          makeSubstitutions(DeclarationUtils.getInvariant(declaration, variable), true);
      final String exprDescription = SmvExprPrinter.replaceLogicValues(printer.toString(invariant));
      result.append(String.format("  %s%s", exprDescription, ENDLINE));
    }
    return result.toString();
  }

  /*
   * Generates variable's initialization in SMV format.
   */
  private static String getInit(final NodeVariable variable, final Declaration declaration,
      final String varName, final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
    if (declaration != null && declaration.containsInitialValue()
        && GaddHlddUtils.getOriginalVariable(variable) == null) {
      final Node initialValue = makeSubstitutions(declaration.getInitialValue(), true);
      if (variable.isType(DataTypeId.MAP)) {
        if (initialValue.isType(DataTypeId.MAP) && initialValue.getKind() == Kind.VALUE) {
          final DataMap map = ((NodeValue) initialValue).getMap();
          result.append(SmvReservedWords.ASSIGNMENT.toString()).append("\n");
          for (Entry<Data, Data> dataEntry : map.entrySet()) {
            final Node index = new NodeValue(dataEntry.getKey());
            final Node value = new NodeValue(dataEntry.getValue());
            final Node element = new NodeOperation(StandardOperation.SELECT, variable, index);
            result.append(String.format("  %s(%s) := %s%s", SmvReservedWords.INIT.toString(),
                printer.toString(element), printer.toString(value), ENDLINE));
          }
        }
      } else {
        result.append(SmvReservedWords.ASSIGNMENT.toString()).append("\n");
        result.append(String.format("  %s(%s) := %s%s", SmvReservedWords.INIT.toString(), varName,
            printer.toString(initialValue), ENDLINE));
      }
    }
    return result.toString();
  }

  /*
   * Generates a declaration section in SMV format from variable collection.
   */
  private static String getVarsDescription(final Map<NodeVariable, Declaration> variables,
      final String keyword, final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
    result.append("\n");
    final List<Entry<NodeVariable, Declaration>> entries = new ArrayList<>(variables.entrySet());
    entries.sort((o1, o2) -> GaddHlddUtils.compareVariablesByName(o1.getKey(), o2.getKey()));
    for (Entry<NodeVariable, Declaration> entry : entries) {
      if (!(entry.getKey().getUserData() instanceof NodeOperation)) {
        result.append(getVariableDeclaration(entry.getKey(), entry.getValue(), printer, keyword));
      }
    }
    return result.toString();
  }

  /**
   * Returns the indentation space according to the specified indentation level.
   *
   * @param level - the specified indentation level.
   * @return The whitespace string.
   */
  public static String getWhitespace(final int level) {
    StringBuilder builder = new StringBuilder();
    for (int i = 0; i < level; i++) {
      builder.append(BASE);
    }
    return builder.toString();
  }

  /*
   * Searches for power operations in the node and replaces them by arithmetic left shift operation.
   */
  private static Node substitutePower(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule powerRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.POWER);
      }

      @Override
      public Node apply(final Node arg0) {
        return Nodes.bvashl(NodeValue.newInteger(1), ((NodeOperation) arg0).getOperand(1));
      }
    };

    transformer.addRule(StandardOperation.POWER, powerRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  /*
   * Searches for BVREPEAT operations in a node and replaces them by concatenation.
   */
  private static Node substituteBvrepeat(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule bvrepeatRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.BVREPEAT);
      }

      @Override
      public Node apply(final Node arg0) {
        final NodeOperation operation = (NodeOperation) arg0;
        final Node pattern = operation.getOperand(1);
        Node result = pattern;
        if (operation.getOperand(0).getKind() == Kind.VALUE) {
          final NodeValue repeats = (NodeValue) TransformerUtils.toInt(operation.getOperand(0));
          assert repeats != null;
          for (int i = 0; i < repeats.getInteger().intValue() - 1; i++) {
            result = Nodes.bvconcat(pattern, result);
          }
        }
        return result;
      }
    };

    transformer.addRule(StandardOperation.BVREPEAT, bvrepeatRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  /*
   * Searches for BVNOR operations in the node and replaces them by BVOR operation negation.
   */
  private static Node substituteNor(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule norRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.BVNORR, StandardOperation.BVNOR);
      }

      @Override
      public Node apply(final Node arg0) {
        final NodeOperation nor = (NodeOperation) arg0;
        final List<Node> operands = nor.getOperands();
        final StandardOperation op = ExprUtils.isOperation(arg0, StandardOperation.BVNORR)
            ? StandardOperation.BVORR
            : ExprUtils.isOperation(arg0, StandardOperation.BVNOR) ? StandardOperation.BVOR : null;
        final NodeOperation or = new NodeOperation(op, operands);
        return Nodes.bvnot(or);
      }
    };

    transformer.addRule(StandardOperation.BVNORR, norRule);
    transformer.addRule(StandardOperation.BVNOR, norRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  /*
   * Searches for BVORR operations in the node and replaces them by zero comparison operation.
   */
  private static Node substituteBvorr(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule bvorrRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVORR);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation bvorr = (NodeOperation) node;
        final Node operand = bvorr.getOperand(0);
        final Node zero = new NodeValue(Data.newBitVector(0, operand.getDataType().getSize()));
        final Node greater = Nodes.greater(operand, zero);
        return Nodes.bool2bv(greater);
      }
    };

    transformer.addRule(StandardOperation.BVORR, bvorrRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  private static Node substituteBvxorr(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule bvxorrRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVXORR, StandardOperation.BVXNORR);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;
        final Node operand = operation.getOperand(0);
        final Enum<?> operationId = operation.getOperationId();
        final int size = operand.getDataType().getSize();
        final ArrayList<Node> bits = new ArrayList<>();
        for (int i = 0; i < size; i++) {
          bits.add(Nodes.bvzeroext(31, Nodes.bvextract(i, operand)));
        }
        final NodeOperation sumBits = Nodes.add(bits);
        final NodeOperation mod = Nodes.mod(sumBits, NodeValue.newBitVector(2, 32));
        final NodeOperation odd = Nodes.eq(mod, NodeValue.newBitVector(1, 32));
        final NodeOperation even = Nodes.eq(mod, NodeValue.newBitVector(0, 32));
        if (operationId == StandardOperation.BVXORR) {
          return Nodes.bool2bv(odd);
        } else if (operationId == StandardOperation.BVXNORR) {
          return Nodes.bool2bv(even);
        }
        return null;
      }
    };

    transformer.addRule(StandardOperation.BVXORR, bvxorrRule);
    transformer.addRule(StandardOperation.BVXNORR, bvxorrRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  private static Node substituteBvandr(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();
    final TransformerRule rule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVANDR, StandardOperation.BVNANDR);
      }

      @Override
      public Node apply(final Node node) {
        final Node operand = ((NodeOperation) node).getOperand(0);
        final Enum<?> operationId = ((NodeOperation) node).getOperationId();
        final int size = operand.getDataType().getSize();
        final ArrayList<Node> bits = new ArrayList<>();
        for (int i = 0; i < size; i++) {
          bits.add(Nodes.bvextract(i, operand));
        }
        final Node mul = Nodes.mul(bits);
        if (operationId == StandardOperation.BVANDR) {
          return mul;
        } else if (operationId == StandardOperation.BVNANDR) {
          return Nodes.not(mul);
        }
        return null;
        //return Nodes.eq(mul, NodeValue.newBitVector(1, 1));
      }
    };
    transformer.addRule(StandardOperation.BVANDR, rule);
    transformer.addRule(StandardOperation.BVNANDR, rule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }


  /**
   * Performs the following substitutions on the specified Node:
   * <ul>
   * <li>Substitutes a power operation by an arithmetic left shift operation.
   * <li>If a map value is found, a corresponding variable is created in the declarations section
   * and the value is replaced by this variable.
   * <li>If a BVEXTRACT operation extracts only one bit it is replaced by a SELECT operation.
   * <li>Substitutes BVREPEAT operation by concatenation.
   * <li>Substitutes BVNOR (BVNORR) operation by BVOR (BVORR) negation.
   * <li>Substitutes BVORR operation by comparison to zero.
   * <li>Substitutes BVXORR (BVXNORR) by sum bits count.
   * <li>Substitutes BVANDR (BVNANDR) by bits multiplication.
   * </ul>
   *
   * @param node - the specified node.
   * @param standardize - denotes whether the node should be standardized.
   * @return The node with the substitutions made.
   */
  public static Node makeSubstitutions(final Node node, final boolean standardize) {
    Node subNode = standardize ? TransformerUtils.standardize(node) : node;
    subNode = substitutePower(subNode);
    subNode = MapUtils.substituteMaps(subNode);
    subNode = substituteBvrepeat(subNode);
    subNode = substituteNor(subNode);
    subNode = substituteBvorr(subNode);
    subNode = substituteBvxorr(subNode);
    subNode = substituteBvandr(subNode);
    return Reducer.reduce(subNode);
  }

  /**
   * Returns the type conversion string from the source type into the target type if possible.
   *
   * @param sourceType - the specified source type.
   * @param targetType - the specified target type.
   * @return Type conversion expression in SMV format if conversion is possible, empty string
   *     otherwise.
   */
  public static String getTypeConversionPrefix(final DataTypeId sourceType,
      final DataTypeId targetType) {
    String result = "";
    if (targetType == DataTypeId.BIT_VECTOR && sourceType == DataTypeId.LOGIC_BOOLEAN) {
      result = SmvReservedWords.TOBIT.toString();
    }
    if (targetType == DataTypeId.LOGIC_BOOLEAN && sourceType == DataTypeId.BIT_VECTOR) {
      result = SmvReservedWords.TOBOOL.toString();
    }
    if (targetType == DataTypeId.LOGIC_INTEGER && (sourceType == DataTypeId.LOGIC_BOOLEAN
        || sourceType == DataTypeId.BIT_VECTOR)) {
      result = SmvReservedWords.TOINT.toString();
    }
    return result;
  }

  /**
   * Generates clock assignment.
   *
   * @param clock     - the specified clock variable.
   * @param phaseVars - the collection of phase variables.
   * @param printer   - the SMV expression printer.
   * @return Clock assignment in SMV format.
   */
  public static String getClockAssignment(final NodeVariable clock,
      final Collection<NodeVariable> phaseVars, final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
    final String clockName = getName(clock);


    if (phaseVars.isEmpty()) {
      result.append(SmvReservedWords.ASSIGNMENT.toString()).append("\n").append(String
          .format("  %s(%s) := !%s;%n", SmvReservedWords.NEXT.toString(), clockName, clockName));

    } else {
      result.append(SmvReservedWords.ASSIGNMENT.toString()).append("\n")
          .append(String.format("  %s(%s) :=%n", SmvReservedWords.NEXT.toString(), clockName))
          .append(String.format("    %s%n", SmvReservedWords.CASE_BEGIN.toString()));

      if (phaseVars.size() == 1) {
        final NodeVariable phaseVar = phaseVars.iterator().next();
        final String phaseName = phaseVar.getName();
        final int size = phaseVar.getDataType().getSize();
        result.append(String.format("      (%s = 0d%d_0) : !%s;%n", phaseName, size, clockName));
      } else {
        final ArrayList<Node> conditions = new ArrayList<>();
        for (NodeVariable phaseVar : phaseVars) {
          final int size = phaseVar.getDataType().getSize();
          conditions.add(Nodes.eq(phaseVar, new NodeValue(Data.newBitVector(0, size))));
        }
        final Node conjunction = ExprUtils.getConjunction(conditions.toArray(new Node[0]));
        result.append(String.format("      %s : !%s;%n", printer.toString(conjunction), clockName));
      }

      result
          .append(String.format("      %s : %s;%n", SmvReservedWords.TRUE.toString(), clockName))
          .append(String.format("    %s;%n", SmvReservedWords.CASE_END.toString()));
    }

    return result.toString();
  }

  /**
   * Converts a string of process identifiers into a single identifier.
   *
   * @param name The string of process identifiers.
   * @return The formatted name.
   */
  private static String getFormattedName(final String name) {
    String result = name.replaceAll("[\\s.,(){}:=\\-\\[\\]@]", "_");
    result = result.replaceAll("_+", "_");
    return result;
  }

  /**
   * Returns a list of registers of the specified module.
   *
   * @param module - the specified {@link Module}.
   * @return The list of register variables.
   */
  public static List<NodeVariable> getRegs(final Module module) {
    final List<NodeVariable> result = module.getVariables();
    result.removeAll(module.getInputs());
    result.removeAll(module.getOutputs());
    return result;
  }

  /**
   * Generates input variable constraints that prohibit value changes while clock = 1.
   *
   * @param inputs - the collection of input variables.
   * @param clock  - the clock variable.
   * @return The string containing a set of constraints.
   */
  public static String getInputConstraints(final Collection<NodeVariable> inputs,
      final NodeVariable clock) {
    final StringBuilder result = new StringBuilder();
    final String clockName = clock.getName();
    for (NodeVariable variable : inputs) {
      final String inputName = getName(variable);
      result.append(SmvReservedWords.TRANSITION.toString()).append("\n").append(String
          .format("  (%s = 0d1_1 & %s(%s) != 0d1_0) -> (%s(%s) = %s)%n", clockName,
              SmvReservedWords.NEXT.toString(), clockName, SmvReservedWords.NEXT.toString(),
              inputName, inputName)).append(SmvReservedWords.TRANSITION.toString()).append("\n")
          .append(String.format("  (%s = 0d1_0 & %s(%s) != 0d1_1) -> (%s(%s) = %s)%n", clockName,
              SmvReservedWords.NEXT.toString(), clockName, SmvReservedWords.NEXT.toString(),
              inputName, inputName)).append("\n");
    }
    return result.toString();
  }

}
