/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.launcher;

import org.apache.commons.cli.ParseException;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.input.ReversedLinesFileReader;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;
import org.stringtemplate.v4.misc.ErrorBuffer;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.ModelCheckingMethod;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.SmvPrinterEngine;
import ru.ispras.retrascope.engine.smv.test.parser.parsers.SmvSpecificationParser;
import ru.ispras.retrascope.engine.smv.test.parser.parsers.SmvStatisticsParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.ReadWriteUtils;
import ru.ispras.retrascope.util.SmvChecker;
import ru.ispras.retrascope.util.SmvChecker.Result;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

/**
 * This engine checks the generated *.smv files with the model checker.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvModelCheckerLauncher extends Engine {

  /**
   * The engine ID.
   */
  public static final String ENGINE_ID = "smv-modelchecker-launcher";

  /**
   * The engine output entity type.
   */
  public static final EntityType OUT_TYPE = EntityType.get("smv-launch");

  /**
   * Default bounded model checking bound.
   */
  public static final int DEFAULT_BOUND = 1;

  // Checker log file name.
  private static final String CHECKER_LOG_FILE = "smv_traces.log";

  // Command file name.
  private static final String CHECKER_COMMANDS = "smv_commands.txt";

  // Checker commands templates.
  private static final String COMMAND_FILE_TEMPLATE = "stg/smv/SmvCommandFile.stg";
  private static final String TEMPLATE_NAME = "smv_command_file";

  private static final String BOUND_PLACEHOLDER = "%BOUND%";
  private static final String COMMENT = "***";

  private static final String SOURCE_OPTION = "-source";
  private static final String VERBOSE_OPTION = "-v";

  private static final String ZERO = "0";

  private static final int READ_LIMIT = 50;

  /**
   * Model checking method parameter.
   */
  public static final Parameter CHECK_METHOD =
      new Parameter("check-method", ModelCheckingMethod.getStringValues(), "Model checking method");

  public static final Parameter COMMAND_FILE =
      new Parameter("command-file", true, "Command file for nuXmv");

  public static final Parameter USE_NUSMV =
      new Parameter("use-nusmv", false, "Use the NuSMV model checker");

  public static final Parameter LOG_LEVEL =
      new Parameter("smv-verbosity", true, "Verbosity level of the model checker");

  private int counterexamples;

  // BMC bound value.
  private static int BOUND = DEFAULT_BOUND;

  /**
   * Constructs a new engine object.
   */
  public SmvModelCheckerLauncher() {
    super(ENGINE_ID);
    addParameter(CHECK_METHOD);
    addParameter(COMMAND_FILE);
    addParameter(USE_NUSMV);
    addParameter(LOG_LEVEL);
    addInputType(SmvPrinterEngine.OUT_TYPE);
    setOutputType(OUT_TYPE);
    counterexamples = 0;
  }

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    final FileNames smvFiles = (FileNames) inputs.get(SmvPrinterEngine.OUT_TYPE);
    final Arguments arguments = parseArguments(inputs);
    final ModelCheckingMethod checkMethod = getCheckingMethod(arguments);
    final String commandFile = getCommandFilePath(checkMethod, arguments);
    final SmvChecker checker = getChecker(arguments);
    final String logLevel = arguments == null ? ZERO : arguments.getValue(LOG_LEVEL, ZERO);

    return runChecker(smvFiles, commandFile, logLevel, checker);
  }

  /*
   * Performs model checking.
   */
  private FileNames runChecker(final FileNames smvFiles, final String commandFile,
      final String logLevel, final SmvChecker checker) {
    final FileNames result = new FileNames(OUT_TYPE.getId());

    for (String file : smvFiles.getArray()) {
      checker.runChecker(file, getCheckerArgs(commandFile, logLevel));
      analyzeResult(result, checker, file);
    }

    return result;
  }

  private void analyzeResult(final FileNames result, final SmvChecker checker, final String file) {
    Log.info("Checker result: " + checker.getResult().name());
    if (checker.getResult() == Result.OK && checker.getLogFile() != null) {
      printRunInfo(file);
      result.add(CHECKER_LOG_FILE);
    }
    if (checker.getResult() == Result.ERROR) {
      throw new RetrascopeException(getErrorMessage(file));
    }
  }

  /*
   * Generates the model checker's command line.
   */
  private String[] getCheckerArgs(final String commandFile, final String logLevel) {
    final List<String> args = new LinkedList<>();
    args.add(SOURCE_OPTION);
    args.add(commandFile);
    if (!logLevel.equals(ZERO)) {
      args.add(VERBOSE_OPTION);
      args.add(logLevel);
    }
    return args.toArray(new String[0]);
  }

  /*
   * Returns the checker instance.
   */
  private SmvChecker getChecker(final Arguments arguments) {
    final SmvChecker checker = SmvChecker.getChecker();
    // Set nuXmv by default
    checker.setNuXmv();
    if (arguments != null && arguments.contains(USE_NUSMV)) {
      checker.setNuSmv();
    }
    return checker;
  }

  /*
   * Returns the checking method.
   */
  private ModelCheckingMethod getCheckingMethod(final Arguments arguments) {
    if (arguments != null && arguments.contains(CHECK_METHOD)) {
      return ModelCheckingMethod.getEnumValue(arguments.getValue(CHECK_METHOD));
    } else {
      return ModelCheckingMethod.LTL_BMC;
    }
  }

  /*
   * Parses the command line.
   */
  private Arguments parseArguments(final Map<EntityType, Entity> inputs) {
    try {
      final Arguments arguments =
          Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);
      if (arguments != null) {
        return parseCommandLine(arguments.getCommandLine());
      }
    } catch (ParseException e) {
      e.printStackTrace();
    }
    return null;
  }

  /*
   * Generates error message.
   */
  private String getErrorMessage(final String file) {
    final String logFile = FilenameUtils.getBaseName(file) + SmvChecker.SMVLOG;
    final StringBuilder errorMsg = new StringBuilder(
        String.format("Model checker exited with error! See '%s' for details:", logFile));
    try {
      if (ReadWriteUtils.exists(logFile)) {
        errorMsg.append("\n");
        final ReversedLinesFileReader fileReader = new ReversedLinesFileReader(new File(logFile));

        final Stack<String> readLines = new Stack<>();
        String line;
        int counter = 0;

        while ((line = fileReader.readLine()) != null && counter < READ_LIMIT) {
          if (!line.startsWith(COMMENT) && !line.equals("")) {
            readLines.push(line);
          }
          counter++;
        }
        fileReader.close();

        errorMsg.append(String.format("Last %s lines of the log file:\n\n", READ_LIMIT));

        while (!readLines.isEmpty()) {
          errorMsg.append(readLines.pop()).append("\n");
        }

      }
    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
    return errorMsg.toString();
  }

  private String getCommandFilePath(final ModelCheckingMethod checkMethod, final Arguments args) {
    return getCommandFile(checkMethod, args).getAbsolutePath();
  }

  /*
   * Returns the checker's command file.
   */
  private File getCommandFile(final ModelCheckingMethod checkMethod, final Arguments args) {
    if (args.contains(COMMAND_FILE) && args.getValue(COMMAND_FILE) != null) {
      final String commandFileName = args.getValue(COMMAND_FILE);
      final File commandFile = new File(commandFileName);
      if (commandFile.exists()) {
        return commandFile;
      } else {
        throw new RetrascopeException(String.format("%s does not exist!", commandFileName));
      }
    } else {
      final File commandFile = new File(CHECKER_COMMANDS);
      final STGroup group = new STGroupFile(COMMAND_FILE_TEMPLATE);
      final ST template = group.getInstanceOf(TEMPLATE_NAME);
      template.add("time", new Date().toString());
      template.add("bound", BOUND);
      template.add("log_file", CHECKER_LOG_FILE);
      switch (checkMethod) {
        case CTL_BDD:
          template.add("bdd", true);
          template.add("type", "ctl");
          break;
        case LTL_BMC:
          template.add("bmc", true);
          template.add("type", "ltl");
          break;
        case LTL_MSAT:
          template.add("msat", true);
          template.add("type", "ltl");
          break;
        case PSL_BMC:
          template.add("bmc", true);
          template.add("type", "psl");
          break;
        case PSL_BDD:
          template.add("bdd", true);
          template.add("type", "psl");
          break;
        default:
          throw new RetrascopeException("Unknown model checking method: " + checkMethod);
      }
      try {
        template.write(commandFile, new ErrorBuffer());
      } catch (IOException e) {
        e.printStackTrace();
      }
      return commandFile;
    }
  }

  /*
   * Generates a command file for BMC.
   */
  private File generateCommandFile(final InputStream templatePath) {
    final File generatedFile = new File(CHECKER_COMMANDS);
    final BufferedReader reader = ReadWriteUtils.newBufferedReader(templatePath);
    final PrintWriter writer = ReadWriteUtils.newPrintWriter(generatedFile);
    try {
      while (reader.ready()) {
        final String line = reader.readLine();
        if (line != null) {
          writer.println(line.replace(BOUND_PLACEHOLDER, String.valueOf(BOUND)));
        }
      }
      reader.close();
      writer.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return generatedFile;
  }

  /**
   * Returns the number of counterexamples found after model checking.
   *
   * @return The number of counterexamples.
   */
  public int getCounterexampleCount() {
    return counterexamples;
  }

  /*
   * Prints model checking information.
   */
  private void printRunInfo(final String fileName) {
    final SmvSpecificationParser specParser = new SmvSpecificationParser();
    final SmvStatisticsParser statParser = new SmvStatisticsParser();
    statParser.parseStats(fileName + "log");
    specParser.parseSpecifications(fileName);
    Log.info("Specifications total: " + specParser.getOverallSpecs());
    Log.info("Counterexamples generated: " + statParser.getCounterexamples());
    Log.info("Specifications proven TRUE: " + statParser.getTrueSpecs());
    counterexamples = statParser.getCounterexamples();
  }

  /**
   * Sets the specified value as the bound for bounded model checking.
   *
   * @param bound - the specified bound value.
   */
  public static void setBmcBound(final int bound) {
    BOUND = bound;
  }

  /**
   * Returns the current value of the bound for bounded model checking.
   *
   * @return The current bound value.
   */
  public static int getBmcBound() {
    return BOUND;
  }
}
