/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.test.parser.parsers;

import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.regex.Pattern;

/**
 * This class is the base class for SMV file parsers.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
abstract class LogParser {

  /* Patterns. */
  private static final String COMMENT_PATTERN = "^\\*{3}";
  private static final String COUNTEREXAMPLE_PATTERN = "Trace Type: Counterexample";
  private static final String TRUE_SPEC_PATTERN = "(specification).+(is true)";
  private static final String SPEC_PATTERN = "[CL]TLSPEC";
  private static final String STATE_PATTERN = "(State: )\\d+\\.\\d+";
  private static final String VALUE_PATTERN = "(^\\s+[\\w_]+(\\.[\\w_]+)*) = (-)*([\\w_]+)";

  final Pattern specification;
  final Pattern trueSpec;
  final Pattern comment;
  final Pattern counterexample;
  final Pattern state;
  final Pattern value;

  /**
   * Constructs a new parser.
   */
  LogParser() {
    specification = Pattern.compile(SPEC_PATTERN);
    trueSpec = Pattern.compile(TRUE_SPEC_PATTERN);
    comment = Pattern.compile(COMMENT_PATTERN);
    counterexample = Pattern.compile(COUNTEREXAMPLE_PATTERN, Pattern.CASE_INSENSITIVE);
    state = Pattern.compile(STATE_PATTERN);
    value = Pattern.compile(VALUE_PATTERN, Pattern.CASE_INSENSITIVE);
  }

  /**
   * Parses the specified file line.
   *
   * @param line - the specified line.
   */
  protected abstract void parseLine(final String line);

  /* Parses the specified file. */
  void parse(final String fileName) {

    final BufferedReader fileReader = ReadWriteUtils.newBufferedReader(fileName);

    try {
      while (fileReader.ready()) {
        final String line = fileReader.readLine();
        if (line != null) {
          parseLine(line);
        }
      }
      fileReader.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

}
