/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench;

import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import ru.ispras.castle.codegen.FileGenerator;
import ru.ispras.castle.codegen.FileGeneratorStringTemplate;
import ru.ispras.castle.codegen.StringTemplateBuilder;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.testbench.Testbench;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Testbench printer abstract engine.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class TestbenchPrinterEngine extends Engine {

  /**
   * Path to directory for testbenches to be generated.
   */
  private static final Parameter TBENCH_DIR =
      new Parameter(
          "testbench-dir",
          true,
          "A directory where testbenches are saved");

  /**
   * When specified, allows the engine to overwrite testbench files.
   */
  public static final Parameter OVERWRITE =
      new Parameter(
          "overwrite-existing",
          false,
          "Overwrite target files if they already exist");

  private static final String CLOCK_PERIOD = "10";
  private static final String CLOCK_UNIT = "ns";

  private static final String DEFAULT_DIRECTORY = "testbenches/";

  private static final String FILE_EXISTS = "The specified file already exists: ";

  /**
   * Creates a new {@code TestVhdlTestbenchPrinter} instance.
   *
   * @param engineId The engine identifier.
   * @param outType The engine output type.
   */
  protected TestbenchPrinterEngine(final String engineId, final EntityType outType) {
    super(engineId, TestModel.TYPE, outType);

    addInputType(CfgInterface.TYPE);
    addParameter(TBENCH_DIR);
    addParameter(OVERWRITE);
  }

  /**
   * Returns an extension of the testbench files.
   *
   * @return Testbench files extension.
   */
  protected abstract String getTestbenchExt();

  /**
   * Returns the name for ST template directory.
   * @return The name for ST template directory.
   */
  protected abstract String getTemplateDirName();

  /**
   * Returns the name for ST template.
   * @return The name for ST template.
   */
  protected abstract String getTemplateName();

  /**
   * Returns the TST parser template name.
   * @return The TST parser template name.
   */
  protected abstract String getTstParserName();

  /**
   * Returns ST top entity (group) name.
   * @return the StringTemplate top entity (group) name.
   */
  protected abstract String getStEntityName();

  /**
   * Returns string description of input port.
   *
   * @return Input port string description.
   */
  protected abstract String getInputPortKind();

  /**
   * Returns string description of output port.
   *
   * @return Output port string description.
   */
  protected abstract String getOutputPortKind();

  /**
   * Returns the string prefix for bit vector variable declaration.
   * @return The string prefix for bit vector variable declaration.
   */
  protected abstract String getBitVectorPrefix();

  /**
   * Returns the string delimiter for bit vector variable declaration.
   * @return The string delimiter for bit vector variable declaration.
   */
  protected abstract String getBitVectorDelim();

  /**
   * Returns the string suffix for bit vector variable declaration.
   * @return The string suffix for bit vector variable declaration.
   */
  protected abstract String getBitVectorSuffix();

  /**
   * Returns input signal declaration format.
   *
   * @return Input signal declaration format.
   */
  protected abstract String getInputDeclarationFmt();

  /**
   * Returns output signal declaration format.
   *
   * @return Output signal declaration format.
   */
  protected abstract String getOutputDeclarationFmt();

  /**
   * Returns mapping from data types to their string representations.
   *
   * @return Map from data types to their string representations.
   */
  protected abstract Map<DataType, String> getDataTypeMap();

  /**
   * Returns mapping from data type string description to "string-to" conversion functions names.
   *
   * @return Map from data types to function names extracting them from strings.
   */
  protected abstract Map<String, String> getConvertFunctionMap();

  /**
   * Builds the collection of testbenches from the specified inputs.
   *
   * @param inputs The inputs that contain testbenches.
   *
   * @return The collection of testbenches.
   */
  protected List<Set<Testbench>> makeTestbenches(final Map<EntityType, Entity> inputs) {

    final boolean overwrite = isOverwrite(inputs);

    final TestModel testModel = getInputEntity(inputs, TestModel.class, TestModel.TYPE);
    final CfgInterface cfgInterface = getInputEntity(inputs, CfgInterface.class, CfgInterface.TYPE);
    final List<Test> tests = testModel.getTests();
    final Map<String, DeclarationContainer> interfaceMapping = cfgInterface.getMapping();

    InvariantChecks.checkTrue(tests.size() == interfaceMapping.size());

    final Iterator<Test> testIterator = tests.iterator();
    final Iterator<Map.Entry<String, DeclarationContainer>> entryIterator =
        interfaceMapping.entrySet().iterator();
    final String directory = getTestbenchDirectory(inputs);
    final List<Set<Testbench>> testbenches = new ArrayList<>(tests.size());

    while (testIterator.hasNext() && entryIterator.hasNext()) {
      final Map.Entry<String, DeclarationContainer> interfaceEntry = entryIterator.next();
      final Test test = testIterator.next();

      final Testbench testbench = printTestbench(test, interfaceEntry, directory, overwrite);
      testbenches.add(new LinkedHashSet<>(Collections.singletonList(testbench)));
    }

    return testbenches;
  }

  private boolean isOverwrite(final Map<EntityType, Entity> inputs) {

    final Arguments args = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    boolean overwrite = false;

    if (args != null) {
      try {
        final Arguments parsedArgs = parseCommandLine(args.getCommandLine());
        overwrite = parsedArgs.contains(OVERWRITE);
      } catch (final ParseException e) {
        throwParseException(args, e);
      }
    }
    return overwrite;
  }

  private String getTestbenchDirectory(final Map<EntityType, Entity> inputs) {

    final Arguments args = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    String directory = DEFAULT_DIRECTORY;

    if (args != null) {
      try {
        final Arguments parsedArgs = parseCommandLine(args.getCommandLine());
        directory = parsedArgs.getValue(TBENCH_DIR, directory);
      } catch (final ParseException e) {
        throwParseException(args, e);
      }
    }
    return directory;
  }

  private Testbench printTestbench(
      final Test test,
      final Map.Entry<String, DeclarationContainer> entry,
      final String testbenchDir,
      final boolean overwrite) {

    final String moduleName = normaliseModuleName(entry.getKey());
    final Path dirPath = Paths.get(processFileName(testbenchDir, moduleName));

    try {
      Files.createDirectories(dirPath);
    } catch (final IOException exception) {
      throwCreateException(dirPath.toString(), exception);
    }

    final DeclarationContainer signals = entry.getValue();
    final Collection<String> inputSignalNames = NodeUtils.getVariableNames(signals.getInputs());

    final Set<Path> additionalFiles = new LinkedHashSet<>();

    int valueLength = 0;
    int lineLength = 0;
    int sequenceNumber = 0;

    for (final Sequence sequence : test.getSequences()) {
      final Path tstFile = dirPath.resolve(getTestFileName(moduleName, sequenceNumber));
      try {
        DeciderLikeTestPatternPrinter.print(sequence, tstFile, inputSignalNames, overwrite);
      } catch (final IOException e) {

        final String msg = LogMessage.makeMessage(LogMessage.CANT_CREATE_FILE, tstFile);
        throw new RetrascopeException(msg, e);
      }
      sequenceNumber++;

      additionalFiles.add(tstFile);

      try (final BufferedReader testReader = ReadWriteUtils.newBufferedReader(tstFile)) {

        int maxLineLength = 0;
        int maxValueLength = 0;

        for (String line = testReader.readLine(); line != null; line = testReader.readLine()) {
          if (line.length() > maxLineLength) {
            maxLineLength = line.length();
          }

          final String[] values = line.split(" ");

          for (final String value : values) {
            if (value.length() > maxValueLength) {
              maxValueLength = value.length();
            }
          }

          if (maxLineLength > lineLength) {
            lineLength = maxLineLength;
          }
          if (maxValueLength > valueLength) {
            valueLength = maxValueLength;
          }
        }
      } catch (final IOException exception) {
        throw new RetrascopeException(exception);
      }
    }

    final Path testPatternFile = dirPath.resolve(getTestFileName(moduleName, 0));

    try {

      if (getTstParserName() != null) {
        final Path tstParserFile = processTstParser(dirPath, moduleName, valueLength, 0, overwrite);
        additionalFiles.add(tstParserFile);
      }

      final Path testbenchFile =
          processTestbenchFile(dirPath, moduleName, testPatternFile,
              lineLength, signals, inputSignalNames, 0, overwrite);

      return new Testbench(testbenchFile, additionalFiles);
    } catch (final IOException e) {
      throw new RetrascopeException("The exception has occurred while printing a file", e);
    }
  }

  private static void throwParseException(
      final Arguments arguments,
      final ParseException exception) {
    throwException(arguments.toString(), exception, LogMessage.CANT_PARSE);
  }

  private static void throwCreateException(
      final String file,
      final IOException exception) {
    throwException(file, exception, LogMessage.CANT_CREATE_FILE);
  }

  private static void throwException(
      final String description,
      final Exception exception,
      final LogMessage type) {

    final String msg = LogMessage.makeMessage(type, description);
    throw new RetrascopeException(msg, exception);
  }

  private Path processTstParser(
      final Path directory,
      final String moduleName,
      final int maxValueLength,
      final int testNumber,
      final boolean overwrite) throws IOException {

    final String fileName = getTstParserFileName(moduleName, testNumber);
    final Path tstParserFile = directory.resolve(fileName);

    final String[] tstParserTemplate = new String[] {getTemplateDirName() + getTstParserName()};

    final FileGenerator tstParserGenerator =
        new FileGeneratorStringTemplate(
            tstParserFile.toString(),
            tstParserTemplate,
            new StringTemplateBuilder() {
            @Override
            public ST build(final STGroup stGroup) {
              final ST st = stGroup.getInstanceOf("vhdl_tst_parser");
              st.add("time", new Date().toString());
              st.add("max_value_length", maxValueLength);

              return st;
            }
        });

    tstParserGenerator.generate();

    if (!overwrite && Files.exists(tstParserFile)) {
      throw new FileAlreadyExistsException(FILE_EXISTS + tstParserFile);
    }

    return tstParserFile;
  }

  private Path processTestbenchFile(
      final Path dir,
      final String moduleName,
      final Path testPatternFile,
      final int maxLineLength,
      final DeclarationContainer externalSignals,
      final Collection<String> inputSignalNames,
      final int testNumber,
      final boolean overwrite) throws IOException {
    InvariantChecks.checkNotNull(testPatternFile);
    InvariantChecks.checkNotNull(externalSignals);

    final String fileName = getTestbenchFileName(moduleName, testNumber);
    final Path testbenchFile = dir.resolve(fileName);
    final Map<String, HdlSignal> signalMap = mapExternalSignals(externalSignals);
    final Collection<HdlSignal> values = signalMap.values();

    final String[] testbenchTemplate = new String[] {getTemplateDirName() + getTemplateName()};

    final FileGenerator testbenchGenerator =
        new FileGeneratorStringTemplate(
            testbenchFile.toString(),
            testbenchTemplate,
            new StringTemplateBuilder() {
            @Override
            public ST build(final STGroup stGroup) {
              final ST st = stGroup.getInstanceOf(getStEntityName());
              st.add("time", new Date().toString());
              st.add("module_name", moduleName);
              st.add("test_file_name", testPatternFile.getFileName().toString());
              st.add("clock_unit", CLOCK_UNIT);
              st.add("clock_period", CLOCK_PERIOD);
              st.add("max_line_length", maxLineLength);
              st.add("vector_format", getInputFormat(externalSignals.getInputs()));
              st.add("vector_signals", StringUtils.join(inputSignalNames, ", "));

              for (final NodeVariable input : externalSignals.getInputs()) {

                st.add("port_dec_names", input.getName());
                st.add("port_directions", getInputPortKind());
                st.add("port_types", mapDataType(input));
              }

              for (final NodeVariable output : externalSignals.getOutputs()) {

                st.add("port_dec_names", output.getName());
                st.add("port_directions", getOutputPortKind());
                st.add("port_types", mapDataType(output));
              }

              for (final HdlSignal signal : values) {

                final String format =
                    signal.isInput ? getInputDeclarationFmt() : getOutputDeclarationFmt();

                st.add("signal_dec_fmt", format);
                st.add("signal_dec_names", signal.name);
                st.add("signal_types", signal.type);
              }

              for (final Map.Entry<String, HdlSignal> entry : signalMap.entrySet()) {

                st.add("port_map_names", entry.getKey());
                st.add("signal_names", entry.getValue().name);
              }

              for (final HdlSignal signal : values) {

                st.add("buf_names", signal.name);
                st.add("buf_types", signal.type);
              }

              for (final String signalName : inputSignalNames) {

                final HdlSignal inSignal = signalMap.get(signalName);
                st.add("input_names", signalName);

                final String signalCoreType = inSignal.type.split("[ (]")[0];
                final DataType signalDataType = inSignal.dataType;

                if (getConvertFunctionMap().isEmpty()) {
                  continue;
                }

                if (getConvertFunctionMap().containsKey(signalCoreType)) {
                  st.add("converters", getConvertFunctionMap().get(signalCoreType));
                } else if (getDataTypeMap().containsKey(signalDataType)) {
                  st.add("converters", getDataTypeMap().get(signalDataType));
                } else {
                  throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, inSignal.type);
                }
              }

              return st;
            }
          });

    testbenchGenerator.generate();

    if (!overwrite && Files.exists(testbenchFile)) {
      throw new FileAlreadyExistsException(FILE_EXISTS + testbenchFile);
    }

    return testbenchFile;
  }

  private CharSequence getInputFormat(final Collection<NodeVariable> inputs) {

    final Iterator<NodeVariable> iterator = inputs.iterator();
    final StringBuilder builder = new StringBuilder();

    while (iterator.hasNext()) {
      builder.append(getSpecifier(iterator.next().getDataType()));

      if (iterator.hasNext()) {
        builder.append(' ');
      }
    }

    return builder.toString();
  }

  private String getSpecifier(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return "%b";
      case LOGIC_INTEGER:
        return "%d";
      case BIT_VECTOR:
        return "%" + dataType.getSize() + "b";
      case LOGIC_REAL:
        return "%f";
      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private String getTestFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "test", testNumber, "tst");
  }

  private String getTestbenchFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "testbench", testNumber, getTestbenchExt());
  }

  private String getTstParserFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "tst_parser", testNumber, getTestbenchExt());
  }

  private static String getTestbenchFile(
      final String prefix,
      final String meaning,
      final int fileNum,
      final String ext) {
    return String.format("%s_%s%s.%s", prefix, meaning, (fileNum > 0) ? "_" + fileNum : "", ext);
  }

  private String mapDataType(final NodeVariable variable) {

    final VariableDescriptor descriptor = (VariableDescriptor) variable.getUserData();

    if (descriptor != null) {
      final String vhdlDataType = descriptor.getStringMetaInfo(MetaInfoType.HDL_TYPE_STR.name());

      if (vhdlDataType != null) {
        return vhdlDataType;
      }
    }

    final DataType dataType = variable.getDataType();

    if (getDataTypeMap().containsKey(dataType)) {
      return getDataTypeMap().get(dataType);
    } else {
      if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
        return String.format("%s%s%s%s%s",
            getBitVectorPrefix(),
            String.valueOf(dataType.getSize() - 1),
            getBitVectorDelim(),
            "0",
            getBitVectorSuffix());
      } else {
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, dataType);
      }
    }
  }

  private Map<String, HdlSignal> mapExternalSignals(final DeclarationContainer externalSignals) {

    final Collection<NodeVariable> inputs = externalSignals.getInputs();
    final Collection<NodeVariable> outputs = externalSignals.getOutputs();

    final Map<String, HdlSignal> mapping = new LinkedHashMap<>(inputs.size() + outputs.size());

    for (final NodeVariable input : inputs) {
      final HdlSignal inSignal =
          new HdlSignal(input.getName(), mapDataType(input), input.getDataType(), true);
      mapping.put(input.getName(), inSignal);
    }

    for (final NodeVariable output : outputs) {
      final HdlSignal outSignal =
          new HdlSignal(output.getName(), mapDataType(output), output.getDataType(), false);
      mapping.put(output.getName(), outSignal);
    }

    return mapping;
  }

  private static class HdlSignal {

    final String name;
    final String type;
    final DataType dataType;
    final boolean isInput;

    HdlSignal(
        final String name,
        final String type,
        final DataType dataType,
        final boolean isInput) {

      this.name = name;
      this.type = type;
      this.dataType = dataType;
      this.isInput = isInput;
    }
  }

  private static String normaliseModuleName(final String moduleName) {
    InvariantChecks.checkNotNull(moduleName);

    final String[] splitName = moduleName.split("\\.");

    if (splitName.length == 1) {
      return moduleName;
    } else {
      return splitName[splitName.length - 1].replaceAll("\\(.*\\)", "");
    }
  }

  private static String processFileName(final String name, final String moduleName) {
    return name.concat(moduleName);
  }
}
