/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Statement that contains variable declarations.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class DecStatement extends Statement {

  private final DeclarationContainer container;

  /**
   * Default constructor.
   */
  DecStatement() {
    super();

    this.container = new DeclarationContainer();
  }

  /**
   * Constructs object as a copy of the specified statement.
   * @param decStatement The declaration statement that is a source for this object construction.
   */
  DecStatement(final DecStatement decStatement) {
    super(decStatement);

    this.container = new DeclarationContainer(decStatement.getDeclarationMap());
  }

  private DeclarationContainer getContainer() {
    return this.container;
  }

  /**
   * Returns variables declarations.
   *
   * @return declarations of node variables
   */
  public Collection<Declaration> getDeclarations() {
    return getContainer().getDeclarations();
  }

  /**
   * Returns the declaration of the specified variable.
   *
   * @param variable variable
   *
   * @return declaration of the specified variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Declaration getDeclaration(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return getContainer().getDeclaration(variable);
  }

  /**
   * Mapping of internal variables names to variable declarations.
   *
   * @return mapping of variable names to declarations
   */
  public Map<NodeVariable, Declaration> getDeclarationMap() {
    return getContainer().getDeclarationMap();
  }

  /**
   * Returns the number of declared variables.
   * @return The number of declared variables.
   */
  private int getVariableNum() {
    return getContainer().getVariableNum();
  }

  /**
   * Returns declared variable with specified name.
   *
   * @param name name of variable to be returned
   *
   * @return declared variable with the specified name,
   *         or {@code null} if there is no such variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public NodeVariable getVariable(final String name) {
    InvariantChecks.checkNotNull(name);

    return getContainer().getVariable(name);
  }

  /**
   * Returns declared variables.
   *
   * @return variables collection
   */
  public List<NodeVariable> getVariables() {
    return getContainer().getVariables();
  }

  /**
   * Returns a collection of this object variables names.
   *
   * @return names of variables
   */
  public Collection<String> getVariableNames() {

    return getContainer().getVariableNames();
  }

  /**
   * Returns a collection of names of the declared input signal variables.
   *
   * @return input variables names
   */
  public Collection<String> getInputNames() {

    return getContainer().getInputNames();
  }

  /**
   * Returns a collection of input signal variables that are declared into this object.
   *
   * @return Input signal variables.
   */
  public Collection<NodeVariable> getInputs() {

    return getContainer().getInputs();
  }

  /**
   * Returns a collection of names of the declared output signal variables.
   *
   * @return output variables names
   */
  public Collection<String> getOutputNames() {

    return getContainer().getOutputNames();
  }

  /**
   * Returns a collection of output signal variables that are declared into this object.
   *
   * @return Output signal variables.
   */
  public Collection<NodeVariable> getOutputs() {

    return getContainer().getOutputs();
  }

  /**
   * Returns a collection of names of the declared register variables.
   *
   * @return register variables names
   */
  public Collection<String> getRegisterNames() {

    return getContainer().getRegisterNames();
  }

  /**
   * Returns data type of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return data type if exists the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public DataType getDataType(final String name) {

    return getContainer().getDataType(name);
  }

  /**
   * Returns initial value of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return initial value if it exist for the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public Node getInitialValue(final String name) {

    return getContainer().getInitialValue(name);
  }

  /**
   * Returns invariant of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return invariant if it exist for the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public Node getInvariant(final String name) {

    return getContainer().getInvariant(name);
  }

  /**
   * Checks does this object contain the specified variable.
   *
   * @param variable variable to be checked
   *
   * @return {@code true} if module already has such variable declared, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean declares(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return getContainer().containsVariable(variable);
  }

  /**
   * Checks whether object contains a variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if object contains a variable with the specified name,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean declares(final String name) {
    InvariantChecks.checkNotNull(name);

    return getContainer().containsVariable(name);
  }

  /**
   * Checks whether this object contain input signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an input signal variable with the specified name,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean declaresInput(final String name) {
    InvariantChecks.checkNotNull(name);

    return getContainer().containsInput(name);
  }

  /**
   * Checks whether this object contain output signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an output signal variable with the specified name,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean declaresOutput(final String name) {
    InvariantChecks.checkNotNull(name);

    return getContainer().containsOutput(name);
  }

  /**
   * Checks whether this object contain register variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains a register variable with the specified name,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean declaresRegister(final String name) {
    InvariantChecks.checkNotNull(name);

    return getContainer().containsRegister(name);
  }

  /**
   * Declares the specified collection of variables.
   *
   * @param declarations variables declarations
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void declareVariables(final Map<NodeVariable, Declaration> declarations) {
    InvariantChecks.checkNotNull(declarations);

    getContainer().declareVariables(declarations);
  }

  /**
   * Declares non-existing variable.
   *
   * <p>Object MUST NOT contain the specified variable.
   * Check this fact by corresponding check-method first.</p>
   *
   * @param variable variable to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when this object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable) {

    declareVariable(variable, null, null);
  }

  /**
   * Declares the specified variable with the specified initial value.
   * @param variable The variable to be declared.
   * @param initValue The initial value of the specified variable.
   * @throws IllegalArgumentException when initial value argument is {@code null}.
   */
  public void declareVariable(final NodeVariable variable, final Node initValue) {

    declareVariable(variable, initValue, null);
  }

  /**
   * Declares the specified variable with the specified invariant.
   *
   * <p>mObject MUST NOT contain the specified variable.
   * Check this fact by corresponding method first.</p>
   *
   * @param variable variable to be added
   * @param invariant variable invariant (can be {@code null})
   *
   * @throws IllegalArgumentException when variable argument is {@code null}.
   * @throws IllegalStateException when this module object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable, final Range invariant) {

    if (invariant == null) {
      declareVariable(variable);
    } else {
      declareVariable(variable, null, invariant);
    }
  }

  /**
   * Declares the specified variable with the specified initial value and the specified invariant.
   *
   * <p>Invariant keeps a constraint on variable value.</p>
   *
   * @param variable variable
   * @param initialValue initial value of variable
   * @param invariant variable invariant
   * @throws IllegalArgumentException when variable argument is {@code null}.
   */
  public void declareVariable(
      final NodeVariable variable,
      final Node initialValue,
      final Range invariant) {

    getContainer().declareVariable(variable, initialValue, invariant);
  }

  /**
   * Declares the specified variable with the related data.
   * @param variable The variable to be declared.
   * @param declaration The variable data.
   */
  public void declareVariable(final NodeVariable variable, final Declaration declaration) {

    declareVariable(variable, declaration.getInitialValue(), declaration.getInvariant());
  }

  /**
   * Declares new variable of the specified data type.
   *
   * <p>NOTE: the {@link VariableDescriptor} variable data is not set here
   * because of Verilog parser features.</p>
   *
   * @param dataType new variable data type
   *
   * @return new variable that is already declared
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public NodeVariable declareNewVariable(final DataType dataType) {
    InvariantChecks.checkNotNull(dataType);

    /* Create unique variable name. */
    final int varNum = getVariableNum();
    final String varName = DeclarationUtils.NEW_VAR_PREFIX + Integer.toString(varNum);

    final NodeVariable nodeVariable = new NodeVariable(varName, dataType);

    declareVariable(nodeVariable);
    return nodeVariable;
  }

  /**
   * Removes declaration of variable with the specified name.
   *
   * @param name variable name
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void removeDeclaration(final String name) {
    InvariantChecks.checkNotNull(name);

    getContainer().removeDeclaration(name);
  }

  /**
   * Removes declaration of the specified variable.
   *
   * @param variable The variable that is declared here.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void removeDeclaration(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    getContainer().removeDeclaration(variable);
  }
}
