/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.expression.Node;

/**
 * Variable declaration.
 *
 * <p>The variable declaration includes an invariant
 * (a condition, to that at every moment variable value must meet)
 * and an initial value.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Declaration {

  /* Variable initial value. */
  private Node initialValue;

  /* A constraint on variable value (invariant). */
  private final Range invariant;

  /**
   * Constructs an object with the specified initial value and it's invariant.
   *
   * @param initialValue Initial value of the specified variable.
   * @param invariant Invariant og the specified variable.
   *
   * @throws java.lang.NullPointerException when variable argument is {@code null}.
   */
  public Declaration(final Node initialValue, final Range invariant) {

    this.initialValue = initialValue;
    this.invariant = invariant;
  }

  /**
   * Creates a copy of the specified declaration object.
   * @param declaration The object to be copied.
   */
  public Declaration(final Declaration declaration) {

    this.initialValue = declaration.initialValue;
    this.invariant = declaration.invariant;
  }

  /**
   * Checks whether object contains invariant.
   *
   * @return {@code true} if there is not-null invariant in this object,
   *         {@code false} otherwise.
   */
  public boolean containsInvariant() {
    return this.invariant != null;
  }

  /**
   * Checks whether object contains it's initial value.
   *
   * @return {@code true} if there is not-null initial value in this object,
   *         {@code false} otherwise.
   */
  public boolean containsInitialValue() {
    return this.initialValue != null;
  }

  /**
   * Returns variable invariant.
   *
   * @return Variable invariant.
   */
  public Range getInvariant() {
    return this.invariant;
  }

  /**
   * Returns initial value of declared variable.
   *
   * @return Variable initial value.
   */
  public Node getInitialValue() {
    return this.initialValue;
  }

  /**
   * Sets new initial value to the declared variable.
   * @param value Variable's initial value.
   */
  public void setInitialValue(final Node value) {
    this.initialValue = value;
  }
}
