/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * The module's concurrent component.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class ModuleProcess extends Statement {

  /* The optional identifier. */
  private final String name;

  /* The collection of events this object is sensitive to. */
  private final EventList sensitivityList = new EventList();

  private final boolean initial;

  /* Process-level properties. */
  private List<Node> properties;

  /**
   * Default constructor.
   */
  protected ModuleProcess() {
    this(null, null, false);
  }

  /**
   * Constructs the process object with the specified sensitivity list of events.
   * @param events The process list of events.
   */
  protected ModuleProcess(final EventList events) {
    this(events, false);
  }

  /**
   * Constructs the process object with the specified value of initial flag.
   *
   *  @param initial The initial flag.
   */
  protected ModuleProcess(final boolean initial) {
    this(null, null, initial);
  }

  /**
   * Constructs the process object with the specified parameters.
   *
   * @param events The sensitivity list events.
   * @param initial The initial flag.
   */
  protected ModuleProcess(final EventList events, final boolean initial) {
    this(null, events, initial);
  }

  /**
   * Constructs the process object with the specified parameters.
   * @param name The process name.
   * @param events The process sensitivity list.
   * @param initial The initial flag.
   */
  protected ModuleProcess(final String name, final EventList events, final boolean initial) {
    super();

    this.name = name;
    setEventList(events);
    this.initial = initial;
    this.properties = new LinkedList<>();
  }

  /**
   * Constructs process object with the specified name.
   * @param name The name of the constructed object.
   */
  protected ModuleProcess(final String name) {
    this(name, null, false);
  }

  /**
   * Constructs object as a copy of the specified process.
   * @param process The object to be an original for being copied.
   */
  protected ModuleProcess(final ModuleProcess process) {
    super(process);

    this.name = process.getName();
    setEventList(process.getEventList());
    this.initial = process.isInitial();
    this.properties = new LinkedList<>(process.getProperties());
  }

  /**
   * Returns the process name.
   * @return The process name.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Checks whether process is of initial kind.
   *
   * @return {@code true} if process is initial, {@code false} otherwise.
   */
  public boolean isInitial() {
    return this.initial;
  }

  /**
   * Returns {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @return sensitivity list
   */
  public EventList getEventList() {
    return this.sensitivityList.deepCopy();
  }

  /**
   * Sets new {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @param eventList The event list.
   *
   * @throws RetrascopeException when object is initial and argument is non-empty.
   */
  public void setEventList(final EventList eventList) {
    if (eventList == null) {
      return;
    }

    if (isInitial() && !eventList.isEmpty()) {
      throw new RetrascopeException(
          "Cannot set to initial process a non-empty event collection.");
    }

    this.sensitivityList.clear();
    this.sensitivityList.addEvents(eventList);
  }

  /**
   * Add the specified event which this process is sensible to.
   *
   * @param event The event to be added.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when this process is initial.
   */
  public void addEvent(final Event event) {
    InvariantChecks.checkNotNull(event);

    if (isInitial()) {
      throw new RetrascopeException("Cannot add event to initial process");
    }
    this.sensitivityList.add(event);
  }

  /**
   * Removes all the events that are related on the variables from the specified collection.
   * @param variables The variables the events to be removed are related to.
   */
  protected void clearEvents(final Collection<NodeVariable> variables) {

    final Collection<Event> toBeRemoved = new LinkedList<>();

    for (final Event event : getEventList()) {

      final NodeVariable eventVar = event.getVariable();

      if (variables.contains(eventVar)) {
        toBeRemoved.add(event);
      }
    }

    for (final Event rmEvent : toBeRemoved) {
      removeEvent(rmEvent);
    }
  }

  private void removeEvent(final Event event) {
    this.sensitivityList.remove(event);
  }

  /**
   * Checks whether this object is sensitive to event for the specified variable.
   * @param variable The variable that may be used in the object events.
   * @return {@code true} when this object is sensitive to an event for the specified variable,
   *         {@code false} otherwise.
   */
  public boolean hasEvent(final NodeVariable variable) {

    return getEventList().contains(variable);
  }

  /**
   * Returns process-level properties' list.
   * @return properties.
   */
  public List<Node> getProperties() {

    return this.properties;
  }

  /**
   * Adds the specified process-level property.
   * @param property The property to be added.
   */
  public void addProperty(final Node property) {

    this.properties.add(property);
  }

  @Override
  public boolean isSequential() {
    return true;
  }

  @Override
  public StatementType getType() {
    return StatementType.PROCESS;
  }

  @Override
  public void setParent(final Statement parent) {

    if (!parent.isType(StatementType.MODULE)) {
      throw RetrascopeException.exception(LogMessage.WRONG_TYPE, parent);
    }
    super.setParent(parent);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return getName();
  }
}
