/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;

/**
 * Utility methods for {@link Statement} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class StatementUtils {

  private StatementUtils() {
    /* Empty. */
  }

  /**
   * Checks whether the specified nodes are of the specified type.
   *
   * @param type node type
   * @param nodes nodes to be checked
   *
   * @return {@code true} if all the specified nodes are of the specified type,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when either the specified type is {@code null}
   *         or the specified array of nodes is empty.
   */
  public static boolean isType(final Enum<?> type, final Statement... nodes) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotEmpty(nodes);

    boolean result = true;

    for (final Statement node : nodes) {

      if (!node.isType(type)) {
        result = false;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether the specified nodes are of the specified type.
   *
   * @param type node type
   * @param nodes nodes to be checked
   *
   * @return {@code true} if all the specified nodes are of the specified type,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when either the specified type is {@code null}
   *         or the specified array of nodes is empty.
   */
  public static boolean isType(final CfgStatementType type, final Statement... nodes) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotEmpty(nodes);

    boolean result = true;

    for (final Statement node : nodes) {

      if (!node.isType(type)) {
        result = false;
        break;
      }
    }

    return result;
  }

  /**
   * Returns the collection of top level statements of the specified one.
   *
   * @param stmt The statement which top level sub-statements should be returned.
   *
   * @return The collection of top level statements of the specified one.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Collection<Statement> getTopStatements(final Statement stmt) {
    InvariantChecks.checkNotNull(stmt);

    final Collection<Statement> topStatements = new LinkedList<>();

    for (int i = 0; i < stmt.getStatementNum(); i++) {
      topStatements.add(stmt.getStatement(i));
    }

    return topStatements;
  }

  /**
   * Returns the top level parent for the specified statement.
   * @param stmt The statement which top level parent should be returned.
   * @return The top level parent for the specified one, or the statement itself
   *         if it has no parents.
   */
  public static Statement getTopParent(final Statement stmt) {

    Statement parent = stmt.getParent();
    if (parent == null) {
      return stmt;
    }
    while (parent.getParent() != null) {
      parent = parent.getParent();
    }
    return parent;
  }

  /**
   * Returns the statement that comes before the specified statement.
   * @param stmt The statement which predecessor is returned.
   * @return The statement that comes before the specified statement.
   */
  public static Statement getPredecessor(final Statement stmt) {

    final Statement parent = stmt.getParent();
    int index = parent.indexOf(stmt);

    if (index > 0) {
      return parent.getStatement(index - 1);
    } else {
      return parent;
    }
  }

  /**
   * Returns statements that go after the specified statement.
   *
   * @param stmt statement which successors are returned
   * @return statements that go after the specified
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Collection<Statement> getSuccessors(final Statement stmt) {
    InvariantChecks.checkNotNull(stmt);

    final Collection<Statement> successors = new LinkedHashSet<>();
    final Statement statementParent = stmt.getParent();

    if (stmt.isSequential()) {

      if (stmt.containStatements()) {

        successors.add(stmt.getStatement(0));

      } else {

        Statement current = stmt;
        Statement parent = statementParent;

        while (parent != null && !parent.isType(StatementType.MODULE)) {

        /* If parent node is SWITCH, ignore it's other children since they're not successors. */
          if (parent.isSequential()) {

            final int currentIndex = parent.indexOf(current);

            if (currentIndex < parent.getStatementNum() - 1) {
              successors.add(parent.getStatement(currentIndex + 1));
              break;
            }
          }
          current = parent;
          parent = current.getParent();
        }
      }
    } else {

      for (int i = 0; i < stmt.getStatementNum(); i++) {

        successors.add(stmt.getStatement(i));
      }
    }

    return successors;
  }

  /**
   * Checks whether the specified statement has successors or not.
   *
   * @param stmt statement that is checked
   *
   * @return {@code true} when the specified statement has no successors, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean isLast(final Statement stmt) {
    InvariantChecks.checkNotNull(stmt);

    final Statement statementParent = stmt.getParent();

    if (stmt.isSequential()) {

      if (stmt.containStatements()) {

        return false;

      } else {

        Statement current = stmt;
        Statement parent = statementParent;

        while (parent != null && !parent.isType(StatementType.MODULE)) {

        /* If parent node is SWITCH, ignore it's other children since they're not successors. */
          if (parent.isSequential()) {

            final int currentIndex = parent.indexOf(current);

            if (currentIndex < parent.getStatementNum() - 1) {
              return false;
            }
          }
          current = parent;
          parent = current.getParent();
        }
        return true;
      }
    } else {
      return !stmt.containStatements();
    }
  }

  /**
   * Replaces the first statement by the second statement.
   * <p>Finds parent statement for the specified first statement and substitutes it
   * by the second statement.</p>
   * @param orig statement to be replaced
   * @param substitute statement that substitutes the first argument
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}
   * @throws IllegalArgumentException when first argument does not have parent statement.
   */
  public static void replace(final Statement orig, final Statement substitute) {
    InvariantChecks.checkNotNull(orig);
    InvariantChecks.checkNotNull(substitute);

    final Statement parent = orig.getParent();

    if (parent == null) {
      throw new IllegalArgumentException("Node to be replaced does not belong to any statement.");
    }

    final int oneIndex = parent.indexOf(orig);
    parent.remove(orig);
    parent.add(substitute, oneIndex);
  }

  /**
   * Returns node that contains a declaration of variable with the specified name.
   *
   * @param node root node which and which child nodes are checked
   * @param name variable name
   *
   * @return node that contains the declaration of variable with the specified name
   *         or {@code null} when the specified node and all the it's children
   *         do not contain such declaration
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public static DecStatement getVariableContainer(final Statement node, final String name) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(name);

    if (node instanceof DecStatement
        && ((DecStatement)node).declares(name)) {

      return (DecStatement) node;

    } else if (node.getStatementNum() > 0) {

      for (int i = 0; i < node.getStatementNum(); i++) {

        final DecStatement result = getVariableContainer(node.getStatement(i), name);

        if (result != null) {
          return result;
        }
      }
      return null;
    }
    return null;
  }

  /**
   * Returns node that contains a declaration of the specified variable.
   *
   * @param stmt The statement that (and which parents) should be checked for declaration.
   * @param var The variable which declaration should be returned.
   * @return Either a declaration statement for the specified variable (the specified one,
   *         or it's parent at some level) or {@code null} if there is no such at any level.
   */
  public static DecStatement getVariableContainer(final Statement stmt, final NodeVariable var) {

    final String name = var.getName();

    return getVariableContainer(stmt, name);
  }

  /**
   * Returns variable declaration container that is closest to the specified statement.
   *
   * @param stmt statement that is included in some degree to the resulting container
   * @return declaration container that is closest to the specified statement in terms of inclusion
   */
  public static DecStatement nearStatement(final Statement stmt) {
    InvariantChecks.checkNotNull(stmt);

    if (stmt instanceof DecStatement) {

      return (DecStatement) stmt;

    } else if (stmt.getParent() != null) {

      return nearStatement(stmt.getParent());
    }
    return null;
  }

  /**
   * Returns closest statement of the specified type to the specified statement.
   * <p>Closest statement is in terms of statement inclusion. For example, we have a
   * {@link CfgModule} object {@code x} that includes {@link CfgProcess} object {@code y} that
   * includes some statement. For that statement the closest statement of the
   * {@code CfgStatementType.MODULE} type is an {@code x} object.</p>
   *
   * @param stmt statement that is included in some degree to the resulting one
   * @param type type of the returned object
   *
   * @return the statement that is closest to the specified statement in terms
   *         of inclusion and has the specified type or {@code null} if there is no such.
   *
   * @throws RetrascopeException when there is no variable declaration container
   *         of the specified type among parents of the specified statement.
   */
  public static Statement nearStatement(final Statement stmt, final Enum<?> type) {
    InvariantChecks.checkNotNull(stmt);

    if (stmt.isType(type)) {

      return stmt;

    } else if (stmt.getParent() != null) {

      return nearStatement(stmt.getParent(), type);
    }
    throw new RetrascopeException(
        String.format(
            "%s container of %s type for %s.",
            LogMessage.CANT_FIND,
            type,
            stmt.getDescription()));
  }
}
