/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.LogMessage;

/**
 * Variable types.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public enum VariableType {
  IN,    // input signal
  OUT,   // output signal
  INOUT, // input-output signal
  REG, varType;     // internal variable (register)

  /**
   * Checks whether identifier is of signal type.
   *
   * @param typeId Type identifier to be checked.
   *
   * @return {@code true} if identifier is of input/output/input-output signal type,
   *         {@code false} otherwise.
   */
  public static boolean isSignal(final Enum<?> typeId) {

    return typeId.getClass().equals(VariableType.class)
      && (typeId == IN || typeId == OUT || typeId == INOUT);
  }

  /**
   * Checks whether the specified identifier is of input or input-output signal type.
   *
   * @param typeId Type identifier to be checked.
   *
   * @return {@code true} if identifier is of input/input-output signal type,
   *         {@code false} otherwise.
   */
  public static boolean isInputSignal(final Enum<?> typeId) {

    return isSignal(typeId) && typeId != OUT;
  }

  /**
   * Check whether the specified identifier is of input-output or output type.
   *
   * @param typeId Type identifier to be checked.
   *
   * @return {@code true} if identifier is of input-output/output signal type,
   *         {@code false} otherwise.
   */
  public static boolean isOutputSignal(final Enum<?> typeId) {

    return isSignal(typeId) && typeId != IN;
  }

  /**
   * Check whether the specified identifier is of register type.
   *
   * @param typeId Type identifier to be checked.
   *
   * @return {@code true} if identifier is of register type, {@code false} otherwise.
   */
  public static boolean isRegister(final Enum<?> typeId) {

    return typeId == REG;
  }

  /**
   * Checks whether the specified variable is a one-bit length signal.
   *
   * @param variable variable
   *
   * @return {@code true} if variable is one bit long signal, {@code false} otherwise
   *
   * @throws RetrascopeException in the following situations:
   *         (1) when any of arguments is {@code null};
   *         (2) when argument's user data is not of {@link VariableDescriptor} type
   */
  public static boolean isOneBitSignal(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final Object data = variable.getUserData();

    if (data instanceof VariableDescriptor) {

      final VariableDescriptor descriptor = (VariableDescriptor) data;
      final boolean isSignal = VariableType.isSignal(descriptor.getVariableType());
      final DataType dataType = variable.getDataType();
      final boolean booleanType = dataType.equals(DataType.BOOLEAN);
      final boolean oneBitVector = dataType.equals(DataType.bitVector(1));

      return isSignal && (booleanType || oneBitVector);

    } else {
      throw RetrascopeException.exception(LogMessage.WRONG_TYPE, data.getClass());
    }
  }
}
