/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Deque;

/**
 * The factory for new variable version creation.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class VariableVersionFactory {

  /**
   * Creates new version of the specified variable.
   * <p>Already existing versions of the variable are stored in the specified deque.</p>
   * @param var The variable that version is to be created.
   * @param deque The existing versions storage.
   * @return The new version of the specified variable.
   */
  public static NodeVariable getVersion(final NodeVariable var, final Deque<NodeVariable> deque) {
    InvariantChecks.checkNotNull(deque, "No versions for variable: " + var.getName());

    final int versionNum = deque.size();
    final String versionName = getVersionName(var.getName(), versionNum);
    final NodeVariable version = new NodeVariable(versionName, var.getDataType());
    final Object varData = var.getUserData();

    final VariableDescriptor versionData = new VariableDescriptor((VariableDescriptor)varData);
    versionData.setVersionNumber(versionNum);
    versionData.setOriginalVariable(var);
    version.setUserData(versionData);

    return version;
  }

  private static String getVersionName(final String name, final int postfix) {

    return String.format("%s_VERSION_%s", name, postfix);
  }
}
