/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis.memory;

import ru.ispras.fortress.expression.NodeValue;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * The hash map-based store of variables.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class HashMapMemory implements Memory {

  private final Map<String, NodeValue> store;

  /**
   * Constructs an empty (values of variables are equal to {@code null}) hash map-based store of
   * variables.
   *
   * @param variableNames a collection of names of variables to be stored
   */
  public HashMapMemory(final Collection<String> variableNames) {

    this.store = new HashMap<>(variableNames.size());

    for (final String variableName : variableNames) {
      this.store.put(variableName, null);
    }
  }

  /**
   * Constructs hash map-based store of variables as a copy of the specified {@link HashMapMemory}
   * instance.
   *
   * @param memory the memory to be copied
   */
  public HashMapMemory(final HashMapMemory memory) {
    this.store = new HashMap<>(memory.getValues());
  }

  @Override
  public void setValue(final String name, final NodeValue value) {

    this.store.put(name, value);
  }

  @Override
  public void setValues(final Map<String, NodeValue> values) {
    this.store.putAll(values);
  }

  @Override
  public boolean containsVariable(final String name) {
    return this.store.containsKey(name);
  }

  @Override
  public NodeValue getValue(final String name) {
    return this.store.get(name);
  }

  @Override
  public Map<String, NodeValue> getValues() {
    return this.store;
  }

  @Override
  public int hashCode() {
    return 31 + this.store.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }

    if (obj == null) {
      return false;
    }

    if (!(obj instanceof HashMapMemory)) {
      return false;
    }

    final HashMapMemory other = (HashMapMemory) obj;

    return this.store.equals(other.store);
  }

  @Override
  public String toString() {

    final StringBuilder stringBuilder = new StringBuilder("{");

    final int initLength = stringBuilder.length();

    for (final Map.Entry<String, NodeValue> entry : this.store.entrySet()) {
      if (stringBuilder.length() > initLength) {
        stringBuilder.append("; ");
      }

      stringBuilder.append(entry.getKey());
      stringBuilder.append(" = ");
      stringBuilder.append(entry.getValue());
    }

    stringBuilder.append('}');

    return stringBuilder.toString();
  }

  @Override
  public void reset() {
    for (final Entry<String, NodeValue> entry : this.store.entrySet()) {
      entry.setValue(null);
    }
  }
}
