/*
 * Copyright 2013-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.Locale;
import java.util.Map;

/**
 * Process of control flow graph model.
 *
 * <p>The object contains the embedded control flow graph nodes
 * and {@link EventList} sensitivity list. Also it can be either
 * of initial or of non-initial type.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgProcess extends ModuleProcess {

  /**
   * Default constructor.
   *
   * <p>Constructs non-initial process.</p>
   */
  public CfgProcess() {
    this(false);
  }

  /**
   * Constructs new process with specified parameters.
   * @param name Process name.
   * @param list Process sensitivity list.
   * @param initial Process initial flag.
   *
   * @throws IllegalArgumentException when event list argument is {@code null}.
   */
  public CfgProcess(final String name, final EventList list, final boolean initial) {
    super(name, list, initial);
  }

  /**
   * Constructs new process with the specified value of initial flag.
   *
   * @param initial initial flag
   */
  public CfgProcess(final boolean initial) {
    super(initial);
  }

  /**
   * Constructs new process with the specified sensitivity list of events.
   *
   * @param events The process sensitivity list of events.
   */
  public CfgProcess(final EventList events) {
    super(events);
  }

  /**
   * Constructs the object that copies the data of the specified one.
   * <p>The internal statements are omitted in the copy.</p>
   * @param process The object the data is copied.
   */
  public CfgProcess(final ModuleProcess process) {
    this(process.getName(), process.getEventList(), process.isInitial());
  }

  private CfgProcess(final CfgProcess process) {
    super(process);
  }

  @Override
  public String getName() {
    return super.getName() == null ? defaultName() : super.getName();
  }

  @Override
  public String getId() {
    return String.format("%s (%s)", getName(), getSuffix());
  }

  private String defaultName() {

    final String typeName = getType().name();
    final boolean emptyList = getEventList().isEmpty();
    final String list = emptyList ? "" : String.format(" (%s)", getEventList());

    return String.format("%s %s%s", getPrefix(), typeName.toLowerCase(Locale.US), list);
  }

  @Override
  public void apply(final Map<String, Node> bindings) {
    getEventList().applyToEventList(bindings);
    super.apply(bindings);
  }

  @Override
  public Statement deepCopy() {
    return new CfgProcess(this);
  }

  private String getPrefix() {
    final Statement parent = getParent();

    return parent == null ? "" : ((CfgModule)parent).getName();
  }

  private String getSuffix() {
    return Integer.toHexString(hashCode());
  }
}
