/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

/**
 * Wait statement of the control flow graph model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgWaitStatement extends CfgConditionStatement {

  private final EventList list;

  /**
   * Constructs the statement with the specified events and condition fields.
   * @param list The events container.
   * @param condition The statement condition.
   */
  public CfgWaitStatement(
      final EventList list,
      final Node condition) {
    super(condition);
    checkNode(condition);
    this.list = list;
  }

  /**
   * Constructs the statement with the specified events.
   * @param list The events container.
   */
  public CfgWaitStatement(final EventList list) {
    super();
    this.list = list;
  }

  /**
   * Constructs the statement with the specified condition field.
   * @param node The statement condition,
   */
  public CfgWaitStatement(final Node node) {
    super(node);
    checkNode(node);
    this.list = null;
  }

  private CfgWaitStatement(final CfgWaitStatement statement) {
    super(statement.getCondition());
    this.list = statement.getEventList();
  }

  /**
   * Returns event list.
   *
   * @return event list
   */
  public EventList getEventList() {
    return this.list;
  }

  /**
   * Checks whether this object contains an event.
   *
   * @return {@code true} when this object containt an event,
   *         {@code false} otherwise
   */
  private boolean hasEvents() {
    return this.list != null && !this.list.isEmpty();
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.WAIT;
  }

  @Override
  public Statement deepCopy() {
    return new CfgWaitStatement(this);
  }

  @Override
  public int getStatementNum() {
    return 0;
  }

  @Override
  public Statement getStatement(final int num) {
    return null;
  }

  @Override
  public void add(final Statement statement) {
    unsupported();
  }

  @Override
  public void remove(final Statement statement) {
    unsupported();
  }

  @Override
  public void addAll(final Collection<Statement> list) {
    unsupported();
  }

  private void unsupported() {
    throw new UnsupportedOperationException(
      LogMessage.makeMessage(LogMessage.NO_CHILD, getClass().getName()));
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {

    final String events = hasEvents() ? getEventList().toString() : "";
    final String suffix = hasCondition() ? " until " + printer.toString(getCondition()) : "";

    return String.format("%s %s%s", getType().name(), events, suffix);
  }

  private static void checkNode(final Node node) {

    if (node.isType(DataType.UNKNOWN)) {

      throw new RetrascopeException(
          String.format(
              "%s%s; only %s is allowed.",
              LogMessage.WRONG_TYPE,
              node.getDataType(),
              DataType.UNKNOWN));
    }
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return Collections.emptySet();
  }

  @Override
  public void apply(final Map<String, Node> bindings) {
    getEventList().applyToEventList(bindings);
    super.apply(bindings);
  }
}
