/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * The EFSM representation of an HDL module. It consists of EFSMs which share all variables declared
 * in {@code EfsmModule} and nested modules having their own variable space.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmModule extends Module {

  private final Map<ModuleProcess, Boolean> efsmToInitialMap;

  /**
   * Constructs the object with the specified name.
   *
   * @param name The object name.
   *
   * @throws IllegalArgumentException if {@code name} is {@code null}.
   */
  public EfsmModule(final String name) {
    super(name);

    InvariantChecks.checkNotNull(name);

    this.efsmToInitialMap = new LinkedHashMap<>();
  }

  /**
   * Constructs the object that copies data from thw specified one.
   * <p>The internal statements are omitted in the copy.</p>
   * @param module The object which data is to be copied.
   */
  public EfsmModule(final Module module) {
    this(module.getName());

    declareVariables(module.getDeclarationMap());
  }

  private EfsmModule(final EfsmModule efsmModule) {
    super(efsmModule);

    this.efsmToInitialMap = new LinkedHashMap<>(efsmModule.getEfsmMap());
  }

  private Map<ModuleProcess, Boolean> getEfsmMap() {

    return Collections.unmodifiableMap(this.efsmToInitialMap);
  }

  /**
   * Adds the specified non-initial EFSM into this module.
   *
   * @param efsm the efsm to be added
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  public void addNonInitialEfsm(final Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.efsmToInitialMap.put(efsm, Boolean.FALSE);

    if (!contains(efsm)) {
      add(efsm);
    }
  }

  /**
   * Adds the specified EFSM into this module as so-called "initial EFSM" which must be
   * simulated in the very beginning of this module simulation. If there are more than one initial
   * EFSMs in this module, they are executed in parallel.
   *
   * @param efsm the efsm to be added
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  public void addInitialEfsm(final Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.efsmToInitialMap.put(efsm, Boolean.TRUE);

    if (!contains(efsm)) {
      add(efsm);
    }
  }

  /**
   * Returns top-level non-initial EFSMs of this EFSM module.
   *
   * @return The collection of non-initial EFSMs.
   */
  public Set<Efsm> getNonInitialEfsms() {

    return getEfsms(false);
  }

/*  *//**
   * Returns top-level initial EFSMs of this EFSM module.
   *
   * @return the EFSMs
   *//*
  public Set<Efsm> getInitialEfsms() {
    return getEfsms(true);
  }*/

  private Set<Efsm> getEfsms(final boolean initial) {
    final Set<Efsm> efsms = new LinkedHashSet<>();

    for (final ModuleProcess efsmProcess : getProcesses()) {
      if (this.efsmToInitialMap.containsKey(efsmProcess)) {
        final Boolean initialValue = this.efsmToInitialMap.get(efsmProcess);
        if (initial == initialValue) {
          efsms.add((Efsm)efsmProcess);
        }
      }
    }

    return efsms;
  }

  @Override
  public Statement deepCopy() {
    return new EfsmModule(this);
  }

  @Override
  public int hashCode() {

    return 31 * super.hashCode() + this.efsmToInitialMap.hashCode();
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof EfsmModule)) {
      return false;
    }

    final EfsmModule otherModule = (EfsmModule) object;

    return this.efsmToInitialMap.equals(otherModule.efsmToInitialMap);
  }

  @Override
  public String toString() {
    return getName();
  }
}
