/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for extended finite state machine model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmUtils {

  private EfsmUtils() {
    /* Empty. */
  }

  /**
   * Returns mapping between modified variables and affecting variables.
   * <p>
   * Returns mappings between modified variables and variables which affects corresponding
   * modifications.
   * </p>
   * <br>
   * <p>
   * For example, if this action consists of two subsequent assignments {@code x = y} and
   * {@code z = x + i}, the resulting map contains mappings from {@code x} to {@code y} and from
   * {@code z} to {@code y} and {@code i}.
   * </p>
   *
   * @param action The object to be analyzed.
   *
   * @return a map containing modified variables and variables which affect corresponding
   *         modifications
   */
  public static Map<NodeVariable, Set<NodeVariable>> getResultingDependencies(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Set<NodeVariable>> substitutions = new HashMap<>(assignments.size());

    for (final Assignment assignment : assignments) {

      final Collection<NodeVariable> usedVariables = ExprUtils.getVariables(assignment.getValue());

      Set<NodeVariable> affecting;

      if (usedVariables.isEmpty()) {

        affecting = Collections.emptySet();

      } else {

        affecting = new HashSet<>(usedVariables.size());

        for (final NodeVariable variable : usedVariables) {

          if (substitutions.containsKey(variable)) {

            affecting.addAll(substitutions.get(variable));

          } else {
            affecting.add(variable);
          }
        }
      }
      substitutions.put(assignment.getVariable(), affecting);
    }
    return substitutions;
  }

  /**
   * Returns mapping between variables and corresponding substitutions.
   * <p>
   * Returns mappings between variables and corresponding substitutions representing total variable
   * modifications during execution of this action. That is, execution of this action is equal to
   * parallel evaluation of the substitutions and subsequent assignment of derived values to the
   * corresponding variables.
   * </p>
   *
   * @param action The object to be analyzed.
   *
   * @return a map containing modified variables and the substitutions representing these
   *         modifications
   */
  public static Map<NodeVariable, Node> getResultingSubstitutions(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Node> substitutions = new LinkedHashMap<>();

    for (final Assignment assignment : assignments) {
      Node expression = assignment.getValue();

      for (final NodeVariable variable : ExprUtils.getVariables(expression)) {

        if (substitutions.containsKey(variable)) {
          expression =
              Transformer.substitute(expression, variable.getName(), substitutions.get(variable));
        }
      }

      final RangedVariable rangedVariable = assignment.getTarget();
      final NodeVariable nodeVariable = rangedVariable.getVariable();

      substitutions.remove(nodeVariable);
      substitutions.put(nodeVariable,
          RangedVariableUtils.transformNodeWithRanged(rangedVariable, expression));
    }

    return substitutions;
  }

  /**
   * Constructs a synchronous composition from the specified list of EFSMs.
   *
   * @param module - the specified EFSM container
   * @param efsms - the specified list of EFSMs
   * @return the synchronous composition of the EFSMs
   * @throws IllegalArgumentException if the specified list of EFSMs is {@code null}
   *         or contains less than two models
   */
  public static Efsm product(final EfsmModule module, final Efsm ... efsms) {

    if ((efsms == null) || (efsms.length < 2)) {
      throw new IllegalArgumentException("The number of EFSMs should be greater than 1!");
    }

    // This data structure represents a state with its outgoing transitions.

    class StateWithTransitions {

      private final EfsmState state;
      private final List<EfsmTransition> transitions;

      private StateWithTransitions(final EfsmState state, final List<EfsmTransition> transitions) {
        this.state = state;
        this.transitions = transitions;
      }
    }

    // Constructing a new EFSM name.

    final StringBuilder builder = new StringBuilder(efsms[0].getName());

    for (int i = 1; i < efsms.length; i++) {
      builder.append(" || ").append(efsms[i].getName());
    }
    final Efsm product = new Efsm(module, builder.toString());

    // Each EFSM is represented as a list of states with transitions
    final List<List<StateWithTransitions>> oldStates = new ArrayList<>(efsms.length);
    final List<EfsmState> initialStates = new ArrayList<>();

    for (final Efsm efsm : efsms) {

      // Collecting initial states.
      initialStates.add(efsm.getInitialState());

      final List<StateWithTransitions> swt = new ArrayList<>(efsm.getStateNum());

      for (final EfsmState state : efsm.getStates()) {

        final Collection<EfsmTransition> trans = efsm.getOutgoing(state);

        if (trans.isEmpty()) {

          // If a state has no outgoing transitions, a new transition is constructed
          // (guard = true, action = null). This is needed by the construction algorithm.

          trans.add(
              new EfsmTransition(
                  state,
                  state,
                  new GuardedAction(Nodes.TRUE, new Action(new Assignment(), true))));
        } else {

          final List<Node> guards = new ArrayList<>();

          // If the guards of the outgoing transitions are not complete (a situation is
          // possible when no transition is enabled), a negation of disjunction of guards
          // is constructed.

          for (final EfsmTransition item : trans) {
            guards.add(item.getGuardNode());
          }

          if (!areComplete(trans)) {
            trans.add(
                new EfsmTransition(
                    state,
                    state,
                    new GuardedAction(guards, new Action(new Assignment(), true))));
          }
        }
        swt.add(new StateWithTransitions(state, new ArrayList<>(trans)));
      }
      oldStates.add(swt);
    }

    // Constructing a new initial state.

    Node initialStateNode = null;

    for (final EfsmState initState : initialStates) {

      if (initialStateNode == null) {

        initialStateNode = initState.getNode();
      } else {
        initialStateNode = Nodes.and(initialStateNode, initState.getNode());
      }
    }

    final EfsmState newInitialState = new EfsmState(initialStateNode);

    // Combining old EFSM states to construct new EFSM states.

    final List<List<StateWithTransitions>> stateCombinations = combineLists(oldStates);

    final Map<EfsmState, List<List<EfsmTransition>>> transMap = new LinkedHashMap<>();

    // Combining transitions of new state combinations.

    for (final List<StateWithTransitions> combinedStateList : stateCombinations) {

      Node stateNode = null;
      final List<List<EfsmTransition>> newStateItems = new ArrayList<>();
      final List<List<EfsmTransition>> combined;

      for (final StateWithTransitions item : combinedStateList) {

        if (stateNode == null) {
          stateNode = item.state.getNode();
        } else {
          stateNode = Nodes.and(stateNode, item.state.getNode());
        }
        newStateItems.add(item.transitions);
      }

      combined = combineLists(newStateItems);

      // Creating a new state.

      final EfsmState newState = product.addState(stateNode);
      transMap.put(newState, combined);

      if (newState.equals(newInitialState)) {
        product.setInitialState(newState);
      }

    }

    // Constructing new transitions.

    for (final Map.Entry<EfsmState, List<List<EfsmTransition>>> entry :
        transMap.entrySet()) {

      for (final List<EfsmTransition> item : entry.getValue()) {

        final EfsmState destState = getNewDestinationState(item);

        if (transMap.keySet().contains(destState)) {

          final GuardedAction gAction = constructGuardedAction(item);

          if (gAction != null) {
            product.addTransition(entry.getKey(), destState, gAction);
          }
        }
      }
    }

    return product;
  }

  /**
   * Checks whether the guards of the specified transitions are complete.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return {@code true} if the guards are complete, {@code false} otherwise
   */
  private static boolean areComplete(final Collection<EfsmTransition> transitions) {

    Node constraint = null;
    final List<Node> guards = new ArrayList<>();

    for (final EfsmTransition item : transitions) {

      guards.add(item.getGuardNode());

      if (constraint == null) {
        constraint = item.getGuardNode();
      } else {
        constraint = Nodes.and(constraint, item.getGuardNode());
      }
    }

    return SolverUtils.areComplete(guards.toArray(new Node[guards.size()]));
  }

  /**
   * Performs a Cartesian product of several lists (sets) of elements.
   *
   * @param lists {@link List} containing the lists to product
   * @return the new {@link List} containing a Cartesian product of the input lists
   * @throws IllegalArgumentException when argument is {@code null}
   */
  private static <T> List<List<T>> combineLists(final List<List<T>> lists) {
    InvariantChecks.checkNotNull(lists);

    int combinations = 1;

    for (final List<T> item : lists) {
      combinations = combinations * item.size();
    }

    final List<List<T>> product = new ArrayList<>();

    for (int i = 0; i < combinations; i++) {

      final List<T> combination = new ArrayList<>();
      int index = 1;

      for (final List<T> item : lists) {
        combination.add(item.get((i / index) % item.size()));
        index = index * item.size();
      }
      product.add(combination);
    }
    return product;
  }

  /**
   * Constructs a new EfsmState according to the destinations of transitions, contained in the
   * specified collection.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return a new {@link EfsmState}
   */
  private static EfsmState getNewDestinationState(final Collection<EfsmTransition> transitions) {

    final List<EfsmState> result = new ArrayList<>();
    Node constraint = null;

    for (final EfsmTransition item : transitions) {
      result.add(item.getDst());
    }

    for (final EfsmState item : result) {
      if (constraint == null) {
        constraint = item.getNode();
      } else {
        constraint = Nodes.and(constraint, item.getNode());
      }
    }

    return new EfsmState(constraint);

  }

  /**
   * Constructs a single {@link GuardedAction} from the guarded actions of the specified
   * transitions.
   *
   * @param transitions a collection of the transitions
   * @return the new guarded action
   */
  private static GuardedAction constructGuardedAction(
      final Collection<EfsmTransition> transitions) {

    final List<Node> predicates = new ArrayList<>();
    final List<Assignment> assignments = new ArrayList<>();
    final EventList events = new EventList();

    for (final EfsmTransition item : transitions) {

      predicates.add(item.getGuardNode());
      events.addEvents(item.getClocks());

      if (item.getAction() != null) {

        final Action action = item.getAction();

        for (final Assignment atomic : action.getAssignments()) {

          if (atomic.getValue() != null) {
            assignments.add(atomic);
          }
        }
      }
    }

    if (!SolverUtils.areCompatible(predicates.toArray(new Node[predicates.size()]))) {
      return null;
    }

    final GuardedAction gAction = new GuardedAction(predicates, new Action(assignments, true));
    gAction.addClocks(events);

    return gAction;
  }

  /**
   * Checks whether the specified EFSM (or it's ancestor) declares the specified variable.
   *
   * @param efsm EFSM to be checked if it (or it's ancestor) declares variable
   * @param variable variable to be checked for declaration
   * @return {@code true} if the specified EFSM (or some it's ancestor) declares the specified
   *         variable, {@code false} otherwise
   */
  public static boolean hasDeclaringAncestor(final Efsm efsm, final NodeVariable variable) {

    EfsmModule module = efsm.getParentContainer();

    while (module != null) {

      if (module.declares(variable)) {
        return true;
      }
      module = (EfsmModule) module.getParent();
    }
    return false;
  }
}
