/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.gadd;

import edu.uci.ics.jung.graph.Graph;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.cfg.transformer.gadd.PhaseSeqEdge;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgModel;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Guarded Actions Decision Diagram (i.e. GADD) model.
 *
 * <p>The model is control flow graph {@link CfgModel} model
 * that is stored in the form of the guarded actions decision diagram.
 * It means that internal model is an acyclic graph, each path of here contains a
 * continuous sub-sequence of switch/case nodes (may be empty) and after that -
 * a continuous sub-sequence of concurrent (non-blocking) basic blocks.
 * The basic block sequence is called "action", the switch/case sequence is called "guard".</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class GaddModel extends Model {

  /**
   * The entity type.
   */
  public static final EntityType TYPE = EntityType.get("gadd");

  /* The internal model. */
  private final CfgModel model;

  /* Map from model processes to their phase dependency graphs. */
  private final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processPhaseMap;

  /* Mapping from module names to their clock-like variables. */
  private final Map<String, NodeVariable> moduleClockMap;

  /**
   * Constructs an object with the specified parameters.
   *
   * @param model The internal model.
   * @param map The mapping from processes to related phase graphs.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public GaddModel(final CfgModel model, final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> map) {
    super(GaddModel.TYPE);
    InvariantChecks.checkNotNull(model);

    this.model = model;
    this.processPhaseMap = map;
    this.moduleClockMap = new LinkedHashMap<>();
  }

  /**
   * Adds the specified entity.
   *
   * @param entity the entity to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when argument is not of {@code GaddModel} type.
   */
  @Override
  public void add(final Entity entity) {
    InvariantChecks.checkNotNull(entity);
    this.model.add(((GaddModel) entity).getModel());
  }

  /**
   * Returns the internal {@link CfgModel} model.
   *
   * @return model
   */
  public CfgModel getModel() {
    return this.model;
  }

  /**
   * Returns the mapping from model processes to their phase sequence graphs.
   * @return The mapping from model processes to their phase sequence graphs.
   */
  public Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> getProcessPhaseMap() {
    return Collections.unmodifiableMap(this.processPhaseMap);
  }

  /**
   * Associate clock-like variable with the name-specified module.
   * @param name The name of the module.
   * @param clock The clock-like variable.
   */
  public void addClock(final String name, final NodeVariable clock) {

    this.moduleClockMap.put(name, clock);
  }

  /**
   * Checks whether the specified process has an associated clock-like variable.
   * @param name The name of the module.
   * @return {@code true} when the model contains a clock-like variable that is associated
   *         with the process, {@code false} otherwise.
   */
  public boolean hasClock(final String name) {

    return this.moduleClockMap.get(name) != null;
  }

  /**
   * Returns the clock-like variable that is associated with module that has the specified name.
   * @param name The name of the module.
   * @return The clock-like variable for the specified process.
   */
  public NodeVariable getClock(final String name) {

    return this.moduleClockMap.get(name);
  }
}
