/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

/**
 * This class represents a node of a high-level decision diagram (HLDD).
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public abstract class HlddNode {

  // Parent node.
  private HlddNode parent;
  // Type of the node.
  private NodeType type;
  // Node level in the HLDD.
  private int level;

  /**
   * Type of a node.
   */
  protected enum NodeType {
    CONDITION, EDGE, ASSIGNMENT
  }

  /**
   * Constructs a new node.
   *
   * @param parent - the specified parent of a node;
   * @param type - the specified type of a node.
   */
  HlddNode(final HlddNode parent, final NodeType type) {
    setParent(parent);
    setType(type);
  }

  /**
   * Constructs a new node.
   *
   * @param type - the specified type of a node.
   */
  HlddNode(final NodeType type) {
    setType(type);
  }


  HlddNode(final HlddNode node) {
    setType(node.getType());
    setLevel(node.getLevel());
  }

  /**
   * Indicates whether the node is terminal.
   *
   * @return {@code true} if the node is terminal, {@code false} otherwise.
   */
  public boolean isTerminal() {
    return type == NodeType.ASSIGNMENT;
  }

  /**
   * Indicates whether the node is a condition.
   *
   * @return {@code true} if the node is a condition, {@code false} otherwise.
   */
  public boolean isCondition() {
    return type == NodeType.CONDITION;
  }

  /**
   * Indicates whether the node is an edge.
   *
   * @return {@code true} if the node is an edge, {@code false} otherwise.
   */
  public boolean isEdge() {
    return type == NodeType.EDGE;
  }

  /**
   * Returns the level of this node.
   *
   * @return The node level.
   */
  public int getLevel() {
    return level;
  }

  /**
   * Sets the level of this node.
   *
   * @param level - the specified level in the diagram.
   */
  public void setLevel(final int level) {
    this.level = level;
  }

  /**
   * Returns the type of the node.
   *
   * @return {@link NodeType}.
   */
  public NodeType getType() {
    return type;
  }

  /**
   * Sets the type of the node.
   *
   * @param type - the specified {@link NodeType}.
   */
  private void setType(final NodeType type) {
    this.type = type;
  }

  /**
   * Returns the parent of the node.
   *
   * @return The parent node.
   */
  public HlddNode getParent() {
    return parent;
  }

  /**
   * Sets the specified {@link HlddNode} as the parent node.
   *
   * @param parent - the specified parent node.
   */
  public void setParent(final HlddNode parent) {
    this.parent = parent;
  }

  /**
   * Indicates whether the node is non-terminal.
   *
   * @return {@code true} if the node is non-terminal, {@code false} otherwise.
   */
  public abstract boolean isNonTerminal();

  /**
   * Makes a deep copy of the node.
   *
   * @return The deep copy of the node.
   */
  public abstract HlddNode deepCopy();

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((parent == null) ? 0 : parent.hashCode());
    result = prime * result + ((type == null) ? 0 : type.hashCode());
    result = prime * result + level;
    return result;
  }


  @Override
  public boolean equals(final Object object) {
    if (this == object) {
      return true;
    }
    if (object == null) {
      return false;
    }
    if (object.getClass() != this.getClass()) {
      return false;
    } else {
      final HlddNode node = (HlddNode) object;
      if (parent != null && node.getParent() != null) {
        if (!node.getParent().equals(parent)) {
          return false;
        }
      } else {
        if (parent != null || node.getParent() != null) {
          return false;
        }
      }
      if (node.getType() != type) {
        return false;
      }
      return node.getLevel() == level;
    }
  }

  @Override
  public String toString() {
    return "HlddNode [type=" + type + ", level=" + level + "]";
  }


}
