/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.walker.DfsWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;

/**
 * Hldd model walker.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddWalker extends DfsWalker {

  /**
   * Constructs a new walker object.
   *
   * @param visitor - the specified {@link HlddVisitor}.
   */
  public HlddWalker(final HlddVisitor visitor) {
    super(visitor, WalkerMode.DFS);
  }

  @Override
  protected void visitProcess(final ModuleProcess process) {
    getVisitor().onProcessBegin(process);
    for (Hldd hldd : ((HlddProcess) process).getHldds()) {
      visitHldd(hldd);
    }
    getVisitor().onProcessEnd(process);
  }

  /*
   * Visits a high-level decision diagram.
   */
  private void visitHldd(final Hldd hldd) {
    InvariantChecks.checkNotNull(hldd);
    ((HlddVisitor) getVisitor()).onHlddBegin(hldd);
    visitRoot(hldd.getRootNode());
    ((HlddVisitor) getVisitor()).onHlddEnd(hldd);
  }

  /*
   * Visits a root node.
   */
  private void visitRoot(final HlddNode node) {
    InvariantChecks.checkNotNull(node);
    ((HlddVisitor) getVisitor()).onRootBegin(node);
    visitNode(node);
    ((HlddVisitor) getVisitor()).onRootEnd(node);
  }

  /*
   * Visits a collection of nodes.
   */
  private void visit(final Collection<? extends HlddNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    HlddNode visitLast = null;
    for (HlddNode node : nodes) {
      if (node.isEdge() && ((HlddEdge)node).isDefault()) {
        visitLast = node;
      } else {
        visitNode(node);
      }
    }
    if (visitLast != null) {
      visitNode(visitLast);
    }
  }

  /*
   * Visits a node.
   */
  private void visitNode(final HlddNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getType()) {

      case CONDITION:
        final HlddConditionNode condition = (HlddConditionNode) node;
        ((HlddVisitor) getVisitor()).onConditionBegin(condition);
        visit(condition.getChildren());
        ((HlddVisitor) getVisitor()).onConditionEnd(condition);
        break;

      case EDGE:
        final HlddEdge edge = (HlddEdge) node;
        ((HlddVisitor) getVisitor()).onEdgeBegin(edge);
        if (edge.getChild() != null) {
          visitNode(edge.getChild());
        }
        ((HlddVisitor) getVisitor()).onEdgeEnd(edge);
        break;

      case ASSIGNMENT:
        final HlddTerminalNode assignment = (HlddTerminalNode) node;
        ((HlddVisitor) getVisitor()).onAssignment(assignment);
        break;
      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, node);
    }
  }
}
