/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.util.Log;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * HDL parser backend that removes initial processes from model but keeps their data.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgInitialProcessBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "init_process";

  @Override
  public void onModuleBegin(final Module module) {

    final Map<NodeVariable, NodeValue> initValues = new LinkedHashMap<>();
    final Collection<ModuleProcess> initialProcesses = new LinkedList<>();

    final Map<NodeVariable, Map<NodeValue, NodeValue>> arrayData = new LinkedHashMap<>();
    final Map<RangedVariable, Node> varValueMap = new LinkedHashMap<>();

    for (final ModuleProcess process : module.getProcesses()) {

      if (!process.isInitial()) {
        continue;
      }

      final int stmtNum = process.getStatementNum();
      for (int i = 0; i < stmtNum; i++) {

        final Statement stmt = process.getStatement(i);

        if (stmt instanceof CfgBlockStatement) {

          for (final Assignment assignment : ((CfgBlockStatement) stmt).getAssignments()) {

            final RangedVariable target = assignment.getTarget();
            final Node value = assignment.getValue();

            if (!(value instanceof NodeValue)) {
              varValueMap.put(target, value);
              continue;
            }

            final NodeValue nodeValue = (NodeValue) value;

            getValue(initValues, arrayData, target, nodeValue);
          }
        } else {
          Log.warn("Can't parse statement of the following type at 'initial' process: %s",
              stmt.getType());
        }
      }

      /* Calculate initial values for array variables. */
      for (final Map.Entry<NodeVariable, Map<NodeValue, NodeValue>> entry : arrayData.entrySet()) {

        final NodeVariable variable = entry.getKey();
        final Map<NodeValue, NodeValue> data = entry.getValue();

        final DataMap map = new DataMap(getKeyType(data), getValueType(data));

        for (final Map.Entry<NodeValue, NodeValue> dataEntry : data.entrySet()) {

          map.put(dataEntry.getKey().getData(), dataEntry.getValue().getData());
        }

        final NodeValue nodeMap = NodeValue.newMap(map);
        initValues.put(variable, nodeMap);
      }

      /* Try to calculate initial values for non-const-valued assignments. */
      for (final Map.Entry<RangedVariable, Node> entry : varValueMap.entrySet()) {

        final RangedVariable target = entry.getKey();
        final Node value = entry.getValue();
        final Collection<NodeVariable> usedVars = ExprUtils.getVariables(value);

        if (initialized(usedVars, initValues)) {

          final List<NodeBinding.BoundVariable> bindings = getBindings(initValues);

          final Node eval = Transformer.substituteAllBindings(new NodeBinding(value, bindings));

          if (eval instanceof NodeValue) {
            getValue(initValues, arrayData, target, (NodeValue) eval);
          } else  {
            Log.warn("Can't calculate expr: '%s'; not all variables are initialized",
                value.toString());
          }
        } else {
          Log.warn("Can't calculate expr: '%s'; not all variables are initialized",
              value.toString());
        }
      }

      /* Set initial values. */
      for (final Map.Entry<NodeVariable, NodeValue> entry : initValues.entrySet()) {
        module.getDeclaration(entry.getKey()).setInitialValue(entry.getValue());
      }

      initialProcesses.add(process);
    }

    /* Remove initial processes from the model. */
    for (final ModuleProcess initial : initialProcesses) {
      module.remove(initial);
    }
  }

  private List<NodeBinding.BoundVariable> getBindings(final Map<NodeVariable, NodeValue> map) {

    final List<NodeBinding.BoundVariable> list = new LinkedList<>();

    for (final Map.Entry<NodeVariable, NodeValue> entry : map.entrySet()) {

      list.add(NodeBinding.bindVariable(entry.getKey(), entry.getValue()));
    }

    return list;
  }

  private boolean initialized(
      final Collection<NodeVariable> vars,
      final Map<NodeVariable, NodeValue> map) {

    for (final NodeVariable var : vars) {

      if (!map.containsKey(var)) {
        return false;
      }
    }
    return true;
  }

  private void getValue(
      final Map<NodeVariable, NodeValue> initValues,
      final Map<NodeVariable, Map<NodeValue, NodeValue>> arrayData,
      final RangedVariable target,
      final NodeValue nodeValue) {

    if (!target.ranged()) {

      if (target.indexed()) {

        getArrayInitValue(arrayData, target, nodeValue);

      } else {

        initValues.put(target.getVariable(), nodeValue);
      }
    }
  }

  private void getArrayInitValue(
      final Map<NodeVariable, Map<NodeValue, NodeValue>> arrayData,
      final RangedVariable target,
      final NodeValue nodeValue) {
    final List<Node> indices = target.getIndices();

    if (indices.size() > 1) {

      Log.warn("Arrays of '%d' dimensions are unsupported by the '%s' backend.",
          indices.size(), NAME);
      return;
    }

    final Node index = indices.iterator().next();

    if (!(index instanceof NodeValue)) {
      Log.warn("Non-constant indexed assignments to arrays at initial processes"
          + " are unsupported by the backend: '%s'.", NAME);
      return;
    }

    final NodeValue constIndex = (NodeValue)index;
    final NodeVariable variable = target.getVariable();

    if (arrayData.containsKey(variable)) {

      final Map<NodeValue, NodeValue> data = arrayData.get(variable);
      data.put(constIndex, nodeValue);

    } else {

      final Map<NodeValue, NodeValue> data = new LinkedHashMap<>();
      data.put(constIndex, nodeValue);
      arrayData.put(variable, data);
    }
  }

  private DataType getValueType(final Map<NodeValue, NodeValue> map) {

    if (map.isEmpty()) {
      return null;
    }

    final Map.Entry<NodeValue, NodeValue> firstElem = map.entrySet().iterator().next();

    return firstElem.getValue().getDataType();
  }

  private DataType getKeyType(final Map<NodeValue, NodeValue> map) {

    if (map.isEmpty()) {
      return null;
    }

    final Map.Entry<NodeValue, NodeValue> firstElem = map.entrySet().iterator().next();

    return firstElem.getKey().getDataType();
  }
}
