/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.Deque;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Control flow graph model visitor that detects and transforms sequences of switch nodes.
 *
 * <p>The HDL parser backend that optimises sequences of {@link CfgModel} model
 * {@link CfgSwitchStatement} nodes. A sequence may include nested switch statements or sequential
 * switch statements. Collapsed switch nodes must have conditions of the following
 * form: {@code x == const_1 || ... || x == const_n}, and the same {@code x} variable must be used
 * in every statement of the sequence to be transformed. Related {@code CfgCaseStatement} statements
 * should be mutually exclusive.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSwitchSequenceBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "switch_seq";

  private Deque<CfgSwitchStatement> switches;

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    this.switches = new LinkedList<>();
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    while (!this.switches.isEmpty()) {

      final CfgSwitchStatement switchStmt = this.switches.pop();

      processNextSwitches(switchStmt);
    }
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchStmt) {

    processNestedSwitches(switchStmt);

    if (canBeCombinedWithNext(switchStmt)) {

      this.switches.push(switchStmt);
    }
  }

  private void processNextSwitches(final CfgSwitchStatement switchStmt) {

    while (canBeCombinedWithNext(switchStmt)) {

      combineWithNext(switchStmt);
    }
  }

  private void combineWithNext(final CfgSwitchStatement switchStmt) {

    final Statement parent = switchStmt.getParent();
    final int nextIndex = parent.indexOf(switchStmt) + 1;
    final Statement next = parent.getStatement(nextIndex);
    parent.remove(next);

    updateSwitch((CfgSwitchStatement) next);

    final Node switchCond = switchStmt.getCondition();

    if (ExprUtils.isOperation(switchCond, StandardOperation.EQ)) {

      updateSwitch(switchStmt);
    }

    for (int i = 0; i < next.getStatementNum(); i++) {

      final Statement caseStmt = next.getStatement(i);

      if (caseStmt.containStatements()) {
        switchStmt.add(caseStmt.deepCopy());
      }
    }
  }

  private void updateSwitch(final CfgSwitchStatement switchStmt) {

    final Node cond = switchStmt.getCondition();
    final Node commonVarNode = getEqVar(cond);

    assert commonVarNode != null : "Can't get variable from: " + cond;

    if (commonVarNode.equals(cond)) {
      return;
    }

    final NodeValue caseValue = getAssignedValue(cond);
    switchStmt.setCondition(commonVarNode);

    for (int i = 0; i < switchStmt.getStatementNum(); i++) {

      final CfgCaseStatement caseStmt = (CfgCaseStatement) switchStmt.getStatement(i);
      caseStmt.removeValues();

      if (caseStmt.containStatements()) {
        caseStmt.add(caseValue);
      } else {
        caseStmt.setDefault(true);
      }
    }
  }

  private boolean canBeCombinedWithNext(final CfgSwitchStatement switchStmt) {

    final Statement parent = switchStmt.getParent();

    final int switchIndex = parent.indexOf(switchStmt);

    if (switchIndex == parent.getStatementNum() - 1) {

      /* Switch statement is the last. */
      return false;
    }

    final Statement next = parent.getStatement(switchIndex + 1);

    if (!next.isType(CfgStatementType.SWITCH)) {

      /* Next statement is not a switch. */
      return false;
    }

    final Node switchCond = switchStmt.getCondition();
    final CfgSwitchStatement nextSwitch = (CfgSwitchStatement) next;
    final Node nextCond = nextSwitch.getCondition();

    return similar(switchCond, nextCond) && combineSwitches(switchStmt, nextSwitch);
  }

  private boolean combineSwitches(final CfgSwitchStatement one, final CfgSwitchStatement another) {

    for (int i = 0; i < one.getStatementNum(); i++) {

      final CfgCaseStatement oneCase = (CfgCaseStatement) one.getStatement(i);
      final Node oneCaseCond = CfgUtils.getExpression(oneCase, one);

      for (int j = 0; j < another.getStatementNum(); j++) {

        final CfgCaseStatement anotherCase = (CfgCaseStatement) another.getStatement(j);
        final Node anotherCaseCond = CfgUtils.getExpression(anotherCase, another);

        final Node compatBranches = Nodes.and(oneCaseCond, anotherCaseCond);

        if (SolverUtils.isSat(compatBranches)
            && oneCase.containStatements() && anotherCase.containStatements()) {
          return false;
        }
      }
    }
    return true;
  }

  private void processNestedSwitches(final CfgSwitchStatement switchNode) {

    CfgSwitchStatement grandChild;

    while ((grandChild = getCompatibleGrand(switchNode)) != null) {

      collapseSwitchSubTree(switchNode, grandChild);
    }
  }

  private void collapseSwitchSubTree(
      final CfgSwitchStatement parent,
      final CfgSwitchStatement grand) {

    final CfgCaseStatement child = (CfgCaseStatement) grand.getParent();
    parent.remove(child);
    child.remove(grand);

    final Node switchCond = parent.getCondition();
    final Node grandCond = grand.getCondition();
    final NodeVariable commonUsed = getCommonUsedVariable(switchCond, grandCond);

    if (!switchCond.equals(commonUsed)) {

      parent.setCondition(commonUsed);
    }

    final NodeValue switchValue = getAssignedValue(switchCond);
    final NodeValue grandValue = getAssignedValue(grandCond);

    /* If switch condition is not collapsed, modify it and it's non-removed case child. */
    if (switchValue != null) {

      final CfgCaseStatement restChild = (CfgCaseStatement) parent.getStatement(0);

      if (restChild.isDefault()) {
        restChild.setDefault(false);
      }

      restChild.removeValues();
      restChild.add(switchValue);
    }

    final List<Statement> moved = new LinkedList<>();

    /* Change values for grandChild children */
    for (int i = 0; i < grand.getStatementNum(); i++) {

      final CfgCaseStatement toBeMoved = (CfgCaseStatement) grand.getStatement(i);
      final Set<NodeValue> caseValues = new LinkedHashSet<>(toBeMoved.getValues());

      toBeMoved.removeValues();

      if (caseValues.contains(Nodes.TRUE)) {

        toBeMoved.add(grandValue);
      } else {
        toBeMoved.setDefault(true);
      }
      moved.add(toBeMoved);
    }

    /* Re-link grandchild children case statements to root switch statement */
    for (final Statement toBeMoved : moved) {

      grand.remove(toBeMoved);
      parent.add(toBeMoved);
    }
  }

  private static NodeValue getAssignedValue(final Node node) {

    switch (node.getKind()) {

      case OPERATION:

        final NodeOperation operation = (NodeOperation) node;

        for (int i = 0; i < operation.getOperandCount(); i++) {

          final Node operand = operation.getOperand(i);

          if (operand.getKind() == Node.Kind.VALUE) {
            return (NodeValue) operand;
          }
        }
        break;

      case VALUE:
        return (NodeValue) node;

      default:
        return null;
    }
    return null;
  }

  private CfgSwitchStatement getCompatibleGrand(final CfgSwitchStatement switchStmt) {

    for (int i = 0; i < switchStmt.getStatementNum(); i++) {

      final CfgCaseStatement caseChild = (CfgCaseStatement) switchStmt.getStatement(i);

      if (compatGrand(switchStmt, caseChild)) {

        return (CfgSwitchStatement) caseChild.getStatement(0);
      }
    }
    return null;
  }

  private boolean compatGrand(final CfgSwitchStatement parent, final CfgCaseStatement child) {

    final Node parentCond = parent.getCondition();

    if (varOrBool(parentCond) && falseCase(parent, child) && hasOneSwitchChild(child)) {

      final CfgSwitchStatement grandChild = (CfgSwitchStatement) child.getStatement(0);
      final Node grandCond = grandChild.getCondition();

      return similar(parentCond, grandCond) && allReachablePaths(parent, child, grandChild);
    }
    return false;
  }

  private static boolean allReachablePaths(
      final CfgSwitchStatement parent,
      final CfgCaseStatement child,
      final CfgSwitchStatement grandChild) {

    final Node caseExpression = CfgUtils.getExpression(child, parent);

    for (int i = 0; i < grandChild.getStatementNum(); i++) {

      final Statement grandStatement = grandChild.getStatement(i);
      final CfgCaseStatement grandCase = (CfgCaseStatement) grandStatement;

      final Node grandExpression = CfgUtils.getExpression(grandCase, grandChild);

      if (!SolverUtils.isSat(Nodes.and(caseExpression, grandExpression))) {
        return false;
      }
    }

    return true;
  }

  private static boolean varOrBool(final Node node) {

    final Collection<NodeVariable> variables = ExprUtils.getVariables(node);

    return ExprUtils.isKind(Node.Kind.VARIABLE, node)
        || (node.isType(DataType.BOOLEAN) && variables.size() == 1);
  }

  private static boolean hasOneSwitchChild(final CfgCaseStatement stmt) {

    return stmt.getStatementNum() == 1 && stmt.getStatement(0).isType(CfgStatementType.SWITCH);
  }

  private static boolean falseCase(final CfgSwitchStatement parent, final CfgCaseStatement child) {

    final Set<NodeValue> values = child.getValues();

    return child.isDefault() && !hasFalseChildBeside(parent, child)
        || (values.size() == 1 && values.iterator().next().equals(Nodes.FALSE));
  }

  private static boolean hasFalseChildBeside(
      final CfgSwitchStatement parent,
      final CfgCaseStatement child) {

    for (int i = 0; i < parent.getStatementNum(); i++) {

      final CfgCaseStatement caseStmt = (CfgCaseStatement) parent.getStatement(i);

      final Set<NodeValue> values = caseStmt.getValues();

      if (!caseStmt.equals(child) && !values.isEmpty() && values.contains(Nodes.FALSE)) {
        return true;
      }
    }
    return false;
  }

  private Node getEqVar(final Node node) {

    if (ExprUtils.isKind(Node.Kind.OPERATION, node)) {

      final NodeOperation operation = (NodeOperation) node;

      if (!ExprUtils.isOperation(operation, StandardOperation.EQ)) {
        return operation;
      }

      for (int i = 0; i < operation.getOperandCount(); i++) {

        final Node operand = operation.getOperand(i);

        if (!ExprUtils.isKind(Node.Kind.VALUE, operand)) {
          return operand;
        }
      }
    } else if (node.getKind() == Node.Kind.VARIABLE) {
      return node;
    }
    return null;
  }

  private static boolean oneVarNode(final Node node) {

    if (ExprUtils.isOperation(node, StandardOperation.EQ)) {

      int varNum = 0;
      int valNum = 0;
      int opNum = 0;

      for (int i = 0; i < ((NodeOperation) node).getOperandCount(); i++) {

        final Node operand = ((NodeOperation) node).getOperand(i);

        switch (operand.getKind()) {

          case VARIABLE:
            varNum++;
            break;

          case VALUE:
            valNum++;
            break;

          case OPERATION:
            opNum++;
            break;

          default:
            break;
        }
      }

      return (varNum == 1) && (valNum == 1) && (opNum == 0);
    } else {
      return ExprUtils.isKind(Node.Kind.VARIABLE, node);
    }
  }

  private boolean similar(final Node first, final Node second) {

    return oneVarNode(second) && oneVarNode(first) && getCommonUsedVariable(first, second) != null;
  }

  private NodeVariable getCommonUsedVariable(final Node switchCond, final Node grandCond) {

    final Set<NodeVariable> switchVars = new LinkedHashSet<>(ExprUtils.getVariables(switchCond));
    final Set<NodeVariable> grandVars = new LinkedHashSet<>(ExprUtils.getVariables(grandCond));

    final Set<NodeVariable> commonVars = CollectionUtils.intersectSets(switchVars, grandVars);

    return commonVars.size() == 1 ? commonVars.iterator().next() : null;
  }
}
