/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementType;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Control flow graph model visitor that detects and transforms wait nodes.
 *
 * <p>The component transforms input model in such a way that the result does not contain
 * {@link CfgWaitStatement} nodes. It substitutes them
 * by sub-processes which are 'guarded' by predicates on new variables. Situations when
 * the same sequence of graph nodes can be executed both after and not after
 * {@link CfgWaitStatement} node are also elaborated.</p>
 *
 * <p>There is Zamia IG feature which is taken into account in this backend. The IG builder
 * creates {@link CfgWaitStatement} objects every time when it
 * detects some process' sensitivity list. The builder puts this node at the end of process'
 * statements sequence. Such nodes are treated by the backend as finalizing and transformed
 * into {@link EventList} sensitivity list components.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgWaitBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "wait";

  /* The current module. */
  private Module module;

  /* The current process. */
  private ModuleProcess process;

  /* The collection for wait-processes. */
  private Map<Module, List<ModuleProcess>> waitProcessMap;

  /* The collection of finalizing wait nodes. */
  private Set<CfgWaitStatement> finalizing;

  /**
   * Default constructor.
   */
  public CfgWaitBackend() {
    super();
  }

  @Override
  public void onModelBegin(final Model model) {
    this.waitProcessMap = new LinkedHashMap<>();
    this.finalizing = new LinkedHashSet<>();
  }

  @Override
  public void onModelEnd(final Model model) {

    /* Remove finalizing wait nodes from the model. */
    for (final CfgWaitStatement wait : this.finalizing) {

      final Statement waitParent = wait.getParent();
      waitParent.remove(wait);
    }

    /* Add wait sub-processes. */
    for (final Map.Entry<Module, List<ModuleProcess>> entry : this.waitProcessMap.entrySet()) {

      final List<ModuleProcess> waitProcesses = entry.getValue();

      for (final ModuleProcess waitProcess : waitProcesses) {
        entry.getKey().add(waitProcess);
      }
    }

    this.waitProcessMap.clear();
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.module = module;
    this.waitProcessMap.put(module, new LinkedList<>());
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    this.process = process;
  }

  private EventList getSensitivityList(final CfgWaitStatement waitNode) {

    if (waitNode.getEventList().isEmpty()) {

      if (waitNode.hasCondition()) {

        final Set<Event> events = new LinkedHashSet<>();
        final Collection<NodeVariable> variables = ExprUtils.getVariables(waitNode.getCondition());

        for (final NodeVariable variable : variables) {
          events.add(new Event(variable, EventType.ANY_EDGE));
        }
        return new EventList(events);

      } else {
        throw new IllegalArgumentException(
            String.format("Empty node of type: %s.", waitNode.getClass().getName()));
      }
    } else {
      return waitNode.getEventList().deepCopy();
    }
  }

  @Override
  public void onWaitBegin(final CfgWaitStatement wait) {

    if (isFinalizing(wait)) {

      /* The finalizing wait node is detected. */
      this.finalizing.add(wait);

    } else {

      /* The internal wait node is detected. */

      /* Create new process. */
      final CfgProcess waitProcess = new CfgProcess(getSensitivityList(wait));

      /* Create & fill process's control flow graph. */

      /* Start CFG building from creating the switch-case pair
       * that checks the related new variable. */

      final CfgModule decModule =
          (CfgModule) StatementUtils.nearStatement(wait, StatementType.MODULE);

      final NodeVariable newVariable = decModule.declareNewVariable(DataType.BOOLEAN);
      final VariableDescriptor descriptor = new VariableDescriptor(VariableType.REG);
      newVariable.setUserData(descriptor);

      final Node variableCondition = Nodes.eq(newVariable, Nodes.TRUE);
      final Node condition = wait.hasCondition()
          ? Nodes.and(wait.getCondition(), variableCondition)
          : variableCondition;
      final CfgSwitchStatement mainSwitch = new CfgSwitchStatement(condition);
      waitProcess.add(mainSwitch);

      final CfgCaseStatement aCase = new CfgCaseStatement((NodeValue) Nodes.TRUE);
      mainSwitch.add(aCase);

      /* Construct wait-process control flow graph. */

      Statement parent = wait.getParent();
      int fromIndex = parent.indexOf(wait);
      boolean stop = false;

      while (!parent.isType(StatementType.MODULE) && !stop) {

        final List<Statement> toBeRemoved = new LinkedList<>();

        for (int i = fromIndex + 1; i < parent.getStatementNum(); i++) {

          final Statement statement = parent.getStatement(i);

          /* When another wait node is detected, finish graph building. */
          if (statement.isType(CfgStatementType.WAIT)) {
            stop = true;
            break;
          }

          /* Move successor nodes which can be reached only after the specified wait node. */
          if (parent.equals(wait.getParent())) {
            toBeRemoved.add(statement);
          }

          /* Make a copy of the successor because it possibly can be reached other way
           * that does not go through the specified wait node. */
          waitProcess.add(statement.deepCopy());
        }

        for (final Statement removed : toBeRemoved) {
          parent.remove(removed);
        }

        final Statement ancestor = parent.getParent();
        fromIndex = ancestor.indexOf(parent);
        parent = ancestor;
      }

      /* Substitute wait node by block node. */
      final CfgBlockStatement trigger = new CfgBlockStatement(newVariable, Nodes.TRUE);
      StatementUtils.replace(wait, trigger);
      final List<ModuleProcess> moduleProcesses = this.waitProcessMap.get(this.module);
      moduleProcesses.add(waitProcess);
    }
  }

  private boolean isFinalizing(final CfgWaitStatement wait) {
    return !wait.hasCondition()
        && wait.getEventList().equals(this.process.getEventList())
        && StatementUtils.getSuccessors(wait).isEmpty();
  }
}
