/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.instgraph.IGContainer;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSubProgram;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * VHDL parser component that elaborates subprograms (functions) into control flow sub-graphs.
 *
 * @see VhdlParser
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgSubprogramBuilder {

  private final List<Statement> statements;
  private final IgStructureVisitor structureVisitor;

  /**
   * Constructs component with the specified IG structure visitor.
   *
   * @param structureVisitor IG structure visitor.
   */
  IgSubprogramBuilder(final IgStructureVisitor structureVisitor) {

    this.statements = new LinkedList<>();
    this.structureVisitor = structureVisitor;
  }

  /**
   * Builds a sequence of control flow statements from the specified sub-program.
   * <p>All the sub-program's internal variables should have the specified prefix.</p>
   * @param program The sub-program to be processed.
   * @param varPrefix The string prefix for internal variables of the specified sub-program.
   * @return The sequence of control flow statements that are built from the specified sub-program.
   */
  Statement[] build(final IGOperationInvokeSubprogram program, final String varPrefix) {

    if (IgUtils.recursiveProgram(program)) {
      Log.warn("%s recursive subprogram %s", LogMessage.CANT_PARSE, varPrefix);
    } else {

      final IGSubProgram subProgram = program.getSub();
      buildProgram(subProgram, varPrefix);
      getStructureVisitor().clearMappings();
    }

    return this.statements.toArray(new Statement[0]);
  }

  private void buildProgram(final IGSubProgram program, final String varPrefix) {

    final Map<String, Node> bindings = new LinkedHashMap<>();

    /* Declare current copies of the sub-program internal variables. */
    final IGContainer container = program.getContainer();

    final Map<NodeVariable, Declaration> declarations =
        getStructureVisitor().getDeclarations(container);

    for (final Map.Entry<NodeVariable, Declaration> decEntry : declarations.entrySet()) {

      final NodeVariable variable = decEntry.getKey();
      final String varName = variable.getName();
      final Declaration declaration = decEntry.getValue();

      final DecStatement module = getStructureVisitor().getTopLevelModule();

      final String newName = String.format("%s.%s", varPrefix, varName);
      final DataType newType = variable.getDataType();

      final NodeVariable newVar = new NodeVariable(newName, newType);
      bindings.put(varName, newVar);

      module.declareVariable(newVar, declaration);
    }

    /* Sub-program statements processing. */
    final IGSequenceOfStatements programCode = program.getCode();
    getStructureVisitor().visit(programCode);

    for (final Statement statement : getStructureVisitor().getStatements()) {
      statement.apply(bindings);
      this.statements.add(statement);
    }
  }

  private IgStructureVisitor getStructureVisitor() {
    return this.structureVisitor;
  }
}
