/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGOperationRecordField;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGRecordField;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.castle.ast.IntegerRange;
import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.math.BigInteger;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * VHDL parser component that extracts control flow graph model nodes from IG objects.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VariableBuilder {

  private final IgStructureVisitor parentVisitor;

  /**
   * Constructs object with the specified parent visitor of IG structure.
   *
   * @param visitor Parent IG structure visitor for this object.
   */
  VariableBuilder(final IgStructureVisitor visitor) {

    this.parentVisitor = visitor;
  }

  private IgStructureVisitor getParentVisitor() {
    return this.parentVisitor;
  }

  /**
   * Returns event list of the specified IG process object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) process object.</p>
   *
   * @param process IG process to which the event list is extracted.
   * @return Event list of the specified process.
   */
  EventList getProcessEventList(final IGProcess process) {

    final EventList result = new EventList();

    final IGSequenceOfStatements sos = process.getSequenceOfStatements();

    final int maxStmtNum = sos.getNumStatements();

    for (int i = 0; i < maxStmtNum; i++) {

      final IGSequentialStatement sosStmt = sos.getStatement(i);

      if (sosStmt instanceof IGSequentialWait) {
        result.addEvents(getEventList((IGSequentialWait) sosStmt));
      }
    }
    return result;
  }

  /**
   * Returns event list of the specified IG wait object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) wait node object.</p>
   *
   * @param wait IG wait node from which the event list is extracted.
   * @return Event list to be extracted from the specified node.
   */
  public EventList getEventList(final IGSequentialWait wait) {

    final EventList list = new EventList();

    final List<IGOperation> sensList = wait.getSensitivityList();

    if (sensList != null) {

      for (final IGOperation current : sensList) {

        list.addEvents(getEvent(current));
      }

      final IGOperation conditionClause = wait.getConditionClause();

      if (conditionClause != null) {

        list.addEvents(getEvent(conditionClause));
      }
    }
    return list;
  }

  private List<Event> getEvent(final IGOperation igOp) {

    if (igOp.getType().isRecord()) {

      System.out.println("TO DO");

      IGOperationObject igObject = (IGOperationObject) igOp;
      IGObject object = igObject.getObject();
      System.out.println(object.toString());

      return getRecordFields(object);

    } else {
      final RangedVariable rVar = getRangedVariable(igOp);
      EventType type = null;

      if (posEdge(igOp)) {

        type = EventType.POSITIVE_EDGE;

      } else if (negEdge(igOp)) {

        type = EventType.NEGATIVE_EDGE;
      }
      return Collections.singletonList(type == null ? new Event(rVar) : new Event(rVar, type));
    }
  }

  private List<Event> getRecordFields(final IGObject object) {

    final SourceLocation src = object.computeSourceLocation();
    final IGType type = object.getType();
    final List<Event> list = new LinkedList<>();
    try {
      for (int i = 0; i < type.getNumRecordFields(src); i++) {
        final IGRecordField recordField = type.getRecordField(i, src);
        final String recordName = recordField.getId();
        final IGType recordType = recordField.getType();
        final DataType dataType = getDataType(recordType);
        final RangedVariable recordVar = new RangedVariable(new NodeVariable(recordName, dataType));
        final Event recordEvent = new Event(recordVar);
        list.add(recordEvent);
      }
    } catch (final ZamiaException e) {
      e.printStackTrace();
    }

    return list;
  }

  private static boolean posEdge(final IGOperation igOp) {

    return isType(igOp, EventType.POSITIVE_EDGE);
  }

  private static boolean negEdge(final IGOperation igOp) {

    return isType(igOp, EventType.NEGATIVE_EDGE);
  }

  private static boolean isType(final IGOperation igOp, final EventType type) {

    return igOp.toHRString().equalsIgnoreCase(type.name());
  }

  /**
   * Extracts ranged variable from IGOperation object.
   *
   * @param igOp Object that contains data to be extracted.
   *
   * @return Ranged variable that is extracted from the specified object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when argument type is unsupported.
   */
  RangedVariable getRangedVariable(final IGOperation igOp) {
    InvariantChecks.checkNotNull(igOp);
    RangedVariable var;

    if (igOp instanceof IGOperationInvokeSubprogram) {

      var = getSubprogramVariable((IGOperationInvokeSubprogram) igOp);

    } else if (igOp instanceof IGOperationObject) {

      IGObject igObj = ((IGOperationObject)igOp).getObject();

      var = new RangedVariable(getObjectVariable(igObj));

    } else if (igOp instanceof IGOperationIndex) {

      var = getIndexVariable((IGOperationIndex) igOp);

    } else if (igOp instanceof IGOperationRange) {

      var = getRangeVariable((IGOperationRange) igOp);

    } else if (igOp instanceof IGOperationRecordField) {

      final VariableType type = VariableBuilderUtils.getVariableType(igOp);
      var = getRangeVariable((IGOperationRecordField) igOp, type);

    } else {
      throw exception(LogMessage.CANT_PARSE, igOp.toHRString());
    }
    return var;
  }

  private RangedVariable getRangeVariable(
      final IGOperationRecordField igOp,
      final VariableType varType) {

    final String recordName = IgUtils.getRecordName(igOp);

    return new RangedVariable(getRecordField(recordName, igOp.getRecordField(), varType));
  }

  private RangedVariable getRangeVariable(final IGOperationRange opRange) {

    final IGOperation igOperand = opRange.getOperand();
    final Node operand = getNode(igOperand);

    final IGOperation igRange = opRange.getRange();

    checkRangeVariableParams(operand, igRange);

    final String name = ((NodeVariable) operand).getName();
    final DataType dataType = ((NodeVariable) operand).getData().getType();

    Range operandRange = getVariableRangeInvariant((IGRange) igRange);
    final NodeVariable variable = new NodeVariable(name, dataType);
    final VariableType variableType = VariableBuilderUtils.getVariableType(igOperand);
    final VariableDescriptor varData = new VariableDescriptor(variableType);
    varData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, opRange.getType().toHRString());
    variable.setUserData(varData);
    initDescriptor(varData, opRange);

    /* Subtract from range parameters the OFFSET value if it is determined. */
    getOperationVisitor().checkBitVectorMinIndex(igOperand, variable, varData);

    if (varData.hasMetaInfo(IgOperationVisitor.OFFSET)) {

      final Object shift = varData.getMetaInfo(IgOperationVisitor.OFFSET);
      final Node shiftVal = NodeValue.newInteger(Integer.parseInt(shift.toString()));
      final Node young = operandRange.getLow();
      final Node old = operandRange.getHigh();
      final Node lowSubShift = Nodes.sub(young, shiftVal);
      final Node highSubShift = Nodes.sub(old, shiftVal);

      operandRange = new Range(lowSubShift, highSubShift);
    }

    return new RangedVariable(variable, operandRange);
  }

  private static void checkRangeVariableParams(final Node operand, final IGOperation range) {

    if (operand.getKind() != Node.Kind.VARIABLE || !(range instanceof IGRange)) {

      throw exception(LogMessage.UNSUPPORTED_TYPE, range.toHRString());
    }
  }

  private RangedVariable getIndexVariable(final IGOperationIndex opIdx) {

    final IGOperation igOperand = opIdx.getOperand();
    final Node operand = getNode(igOperand);
    final VariableDescriptor operandData = (VariableDescriptor) operand.getUserData();

    operandData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, igOperand.getType().toHRString());
    operand.setUserData(operandData);
    initDescriptor(operandData, opIdx);

    final RangedVariable var = new RangedVariable((NodeVariable) operand);
    final Node index = getNode(opIdx.getIndex());

    if (operand.isType(DataTypeId.MAP)) {

      operandData.setIndexRanges(extractRanges(opIdx.getType()));
      var.addIndex(index);
    } else if (operand.isType(DataTypeId.BIT_VECTOR)) {
      var.setRange(index, index);
    }


    return var;
  }

  /**
   * Extracts variable from the specified IG object.
   *
   * @param igObj IG object that contains data to be extracted.
   *
   * @return Variable that is extracted from the specified object.
   */
  NodeVariable getObjectVariable(final IGObject igObj) {

    IGObject object = igObj;

    if (getParentVisitor().hasMapping(igObj)) {

      object = getParentVisitor().getMapping(igObj);
    }

    final IGType opType = object.getType();

    final VariableType varType = VariableBuilderUtils.getVariableType(object);
    final DataType varValueType = getDataType(opType);

    if (varValueType == null) {

      /* Non-synthesizable variable. */
      return null;
    }

    final String varName = IgUtils.getName(getParentVisitor(), object);
    final NodeVariable variable = new NodeVariable(varName, varValueType);
    final VariableDescriptor descriptor = new VariableDescriptor(varType);
    final String fullType = getFullType(object);

    descriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);

    /* If variable is of array type, set range attribute to the descriptor. */
    if (variable.isType(DataTypeId.MAP)) {

      descriptor.setIndexRanges(extractRanges(opType));
    }

    variable.setUserData(descriptor);
    initDescriptor(descriptor, object);
    getOperationVisitor().checkBitVectorMinIndex(object, variable, descriptor);

    return variable;
  }

  private List<Range> extractRanges(final IGType type) {

    final List<Range> ranges = new LinkedList<>();

    extract(ranges, type);

    return ranges;
  }

  private void extract(final List<Range> ranges, final IGType type) {

    final IGType indexType = type.getIndexType() == null ? type : type.getIndexType();

    if (!(indexType instanceof IGTypeStatic)) {
      Log.warn("Unable to calculate range for: %s", type.toHRString());
      return;
    }

    final IntegerRange intRange = getRange((IGTypeStatic) indexType);
    ranges.add(new Range(intRange));

    final IGType elemType = type.getElementType();

    if (elemType != null && elemType.getCat() == IGType.TypeCat.ARRAY) {
      extract(ranges, type.getElementType());
    }
  }

  private RangedVariable getSubprogramVariable(final IGOperationInvokeSubprogram igSub) {

    RangedVariable var;
    final int opNum = igSub.getNumOperands();

    if (opNum == 1) {
      var = getRangedVariable(igSub.getOperand(0));
    } else {
      throw exception(LogMessage.CANT_PARSE, igSub.toHRString());
    }
    return var;
  }

  /**
   * Returns a full string representation of IG object type.
   *
   * @param igObj IG object that is processed by this method.
   * @return String representation of object type including base type and invariant.
   */
  private static String getFullType(final IGObject igObj) {

    final IGType type = igObj.getType();
    final IGType elemType = type.getElementType();

    /* Bit vectors are treated as arrays in IG representation. */
    final boolean isBitVector = !type.isAtomic()
        && (type.isArray() && elemType.isLogic() || type.isCharEnum());

    /* The type representation. */
    final String typeId = isBitVector
        ? String.format("bit_vector(%s)", type.getIndexType().toHRString())
        : type.getId() == null ? type.getCat().name() : type.getId();

    /* Extract possible range (for integer variables only). */
    final boolean isRangedInteger = type.isInteger() && type.getRange() != null;
    final String range = isRangedInteger ? " range " + type.getRange().toHRString() : "";

    return String.format("%s%s", typeId, range);
  }

  /**
   * Extracts data type of the specified IG type object.
   *
   * @param igType IG object type.
   *
   * @return Data type that is extracted from the specified IG data type.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when argument is not of expected category
   *         (INTEGER, REAL, ENUM, ARRAY) or is not of expected sub-category
   *         (ENUM-bit, ENUM-bool, ENUM-logic, ARRAY-bit, ARRAY-array, ARRAY-integer, ARRAY-logic).
   */
  public DataType getDataType(final IGType igType) {
    InvariantChecks.checkNotNull(igType);

    DataType dataType;
    switch (igType.getCat()) {
      case INTEGER:
        dataType = DataType.INTEGER;
        break;
      case REAL:
        dataType = DataType.REAL;
        break;
      case ENUM:
        dataType = VariableBuilderUtils.getEnumDataType(igType);
        break;
      case ARRAY:
        dataType = getArrayDataType(igType);
        break;
      case PHYSICAL:
        dataType = null;
        break;
      default:
        throw exception(LogMessage.UNSUPPORTED_TYPE, igType.toHRString());
    }
    return dataType;
  }

  private DataType getArrayDataType(final IGType type) {

    DataType dataType;

    final IGType elemType = type.getElementType();
    final IGType indexType = type.getIndexType();

    if (elemType.isBit()) {

      /* Bit vector, calculate it's size. */
      if (indexType instanceof IGTypeStatic) {

        final int size = getTypeSize((IGTypeStatic) indexType);
        dataType = DataType.bitVector(size);

      } else {
        throw exception(LogMessage.UNSUPPORTED_TYPE, type.toHRString());
      }

    } else if (elemType.isArray()) {

      dataType = DataType.map(DataType.INTEGER, getDataType(elemType));

    } else if (elemType.isInteger()) {

      dataType = DataType.map(DataType.INTEGER, DataType.INTEGER);

    } else if (elemType.isLogic()) {

      /* Bit vector, calculate it's size. */
      if (indexType instanceof IGTypeStatic) {

        final int size = getTypeSize((IGTypeStatic) indexType);
        dataType = DataType.bitVector(size);

      } else {

        dataType = getDataType(indexType);
      }

    } else if (elemType.isRecord()) {

      throw exception(LogMessage.CANT_PARSE, type.toHRString());

    } else if (elemType.isCharEnum()) {

      dataType = DataType.STRING;

    } else {

      throw exception(LogMessage.CANT_PARSE, type.toHRString());
    }
    return dataType;
  }

  /**
   * Returns size of the specified type.
   *
   * @param indexType Index type to be processed.
   *
   * @return Size of the specified index type.
   *
   * @throws RetrascopeException when index left/right borders
   *         cannot be transformed into {@link ru.ispras.fortress.expression.NodeValue} objects.
   * @throws RetrascopeException when index left/right borders
   *         are not of integer types.
   */
  int getTypeSize(final IGTypeStatic indexType) {

    final IntegerRange range = getRange(indexType);

    return Math.abs(range.getMax().intValue() - range.getMin().intValue() + 1);
  }

  private IntegerRange getRange(final IGTypeStatic type) {

    final IGStaticValue range = type.getStaticRange();
    final Node left = getNode(range.getLeft());
    final Node right = getNode(range.getRight());

    checkRangeParams(left, right);

    final Object leftVal = ((NodeValue) left).getValue();
    final Object rightVal = ((NodeValue) right).getValue();

    final BigInteger max = ((BigInteger)leftVal).max((BigInteger) rightVal);
    final BigInteger min = ((BigInteger)leftVal).min((BigInteger) rightVal);

    return new IntegerRange(min, max);
  }

  private static void checkRangeParams(final Node left, final Node right) {

    final String format = "%s%s, %s; %s is supported only.";
    final Node.Kind leftKind = left.getKind();
    final Node.Kind rightKind = right.getKind();

    if (leftKind != Node.Kind.VALUE || rightKind != Node.Kind.VALUE) {

      throw new RetrascopeException(
          String.format(format,
              LogMessage.UNSUPPORTED_TYPE, leftKind, rightKind, Node.Kind.VALUE));
    }
    if (!ExprUtils.isType(DataType.INTEGER, left, right)) {

      final DataType leftType = left.getDataType();
      final DataType rightType = right.getDataType();

      throw new RetrascopeException(
          String.format(format,
              LogMessage.UNSUPPORTED_TYPE, leftType, rightType, DataType.INTEGER));
    }
  }



  /**
   * Extracts invariant to the specified variable from the specified IGType object.
   *
   * @param var Variable which invariant is to be extracted.
   * @param type Type which contains an invariant that is to be extracted.
   *
   * @return Invariant for the specified variable that is in the specified type object.
   *
   * @throws RetrascopeException when argument is ranged
   *         and it's range is neither of {@code IGStaticValue}
   *         nor of {@code IGRange} type.
   */
  public Range getInvariant(final NodeVariable var, final IGType type) {

    Range rangeInv;

    if (noTypeInvariant(var, type)) {

      rangeInv = null;

    } else {

      final IGOperation range = type.getRange();

      if (range instanceof IGStaticValue) {

        rangeInv = getStaticRangeInvariant((IGStaticValue) range);

      } else if (range instanceof IGRange) {

        rangeInv = getVariableRangeInvariant((IGRange) range);

      } else {

        throw exception(LogMessage.CANT_PARSE, range.getClass().toString());
      }
    }
    return rangeInv;
  }

  private static boolean noTypeInvariant(final NodeVariable var, final IGType type) {

    /* One-bit-sized vectors or booleans or non-ranged variables
     * do not have any type invariant. */
    final DataType oneBitVec = DataType.bitVector(1);

    return var.isType(oneBitVec) || var.isType(DataType.BOOLEAN) || type.getRange() == null;
  }

  private Range getVariableRangeInvariant(final IGRange range) {

    final Node left = getNode(range.getLeft());
    final Node right = getNode(range.getRight());

    final Node ascending = getNode(range.getAscending());

    final Node trueNode = ascending.isType(DataTypeId.BIT_VECTOR)
        ? NodeValue.newBitVector(BitVector.TRUE) : Nodes.TRUE;

    final Node checkAscend = Nodes.eq(ascending, trueNode);

    /* Determine whether ascending node is true or false. */
    final boolean ascend = SolverUtils.isSat(checkAscend);

    return RangedVariableUtils.getRangeInvariant(left, right, ascend);
  }

  private Range getStaticRangeInvariant(final IGStaticValue range) {

    /* The left/right borders. */
    final Node left = getNode(range.getLeft());
    final Node right = getNode(range.getRight());

    /* The flag that shows whether range is ascending or descending. */
    final boolean ascend = range.getAscending().isTrue();

    return RangedVariableUtils.getRangeInvariant(left, right, ascend);
  }

  /**
   * Fill the specified descriptor with the specified IG operation object.
   *
   * @param desc Descriptor to be filled.
   * @param igOp IG operation object to be analyzed.
   */
  private void initDescriptor(final VariableDescriptor desc, final IGOperation igOp) {

    if (igOp instanceof IGOperationIndex) {
      return;
    }
    if (igOp instanceof IGOperationObject) {

      final IGObject igObj = ((IGOperationObject) igOp).getObject();

      initDescriptor(desc, igObj);

      return;
    }
    if (igOp instanceof IGOperationRange) {
      return;
    }
    throw exception(LogMessage.CANT_PARSE, igOp.toHRString());
  }

  /**
   * Fills the specified descriptor with the data that are extracted from the specified
   * IG object.
   *
   * @param desc Descriptor to be filled.
   * @param object IG object to be analyzed.
   */
  private void initDescriptor(final VariableDescriptor desc, final IGObject object) {

    // BASE attribute
    final IGType igType = object.getType();
    desc.setBase(getDataType(igType));

    try {

      final SourceLocation srcLoc = object.computeSourceLocation();

      final IGTypeStatic staticType =
          igType.isArray() && (igType instanceof IGTypeStatic)
          ? (IGTypeStatic) igType.getIndexType()
          : igType.computeStaticType(getParentVisitor().getRuntimeEnv(), null, null);

      final IGOperation range = staticType.getRange();

      if (range instanceof IGStaticValue) {

        // HIGH attribute
        final IGStaticValue high = staticType.getStaticHigh(srcLoc);
        desc.setHigh(getNode(high));

        // LOW attribute
        final IGStaticValue low = staticType.getStaticLow(srcLoc);
        desc.setLow(getNode(low));

        // ASCENDING attribute
        final IGStaticValue ascending = staticType.getStaticAscending(srcLoc);
        desc.setAscending(getNode(ascending));

        final Node descLow = desc.getLow();
        final Node descHigh = desc.getHigh();

        // LEFT attribute
        //      final IGStaticValue leftValue = IGStaticValue.generateZ(staticType, sourceLocation);
        desc.setLeft(ascend(desc) ? descLow : descHigh);

        // RIGHT attribute
        desc.setRight(ascend(desc) ? descHigh : descLow);

        if (igType.isArray()) {

          // RANGE attribute
          desc.setRange(new Range(descLow, descHigh));

          // REVERSE_RANGE attribute
          //desc.setReverseRange(new Range(descHigh, descLow));

          // LENGTH attribute
          final NodeOperation sub = Nodes.sub(descHigh, descLow);
          desc.setLength(Reducer.reduce(Nodes.add(sub, NodeValue.newInteger(1))));
        }
      } else if (range instanceof IGRange) {

        // LEFT attribute
        desc.setLeft(getNode(range.getRangeLeft(srcLoc)));

        // RIGHT attribute
        desc.setRight(getNode(range.getRangeRight(srcLoc)));

        // ASCENDING attribute
        desc.setAscending(getNode(((IGRange)range).getAscending()));

        // HIGH attribute
        desc.setHigh(ascend(desc) ? desc.getRight() : desc.getLeft());

        // LOW attribute
        desc.setLow(ascend(desc) ? desc.getLeft() : desc.getRight());
      }
    } catch (final ZamiaException e) {
      throw new RetrascopeException(e);
    }
  }

  private static boolean ascend(final VariableDescriptor desc) {

    return desc.getAscending().equals(Nodes.TRUE);
  }

  private IgOperationVisitor getOperationVisitor() {

    return getParentVisitor().getOperationCfgVisitor();
  }

  private Node getNode(final IGOperation value) {

    return getOperationVisitor().getNode(value);
  }

  private static RetrascopeException exception(final Enum<?> desc, final String msg) {

    return RetrascopeException.exception(desc, msg);
  }

  /**
   * Description variable type.
   * @param igField,varType Record field and variable type.
   * @return variable descriptor.
   */
  public NodeVariable getRecordField(
      final String recordName,
      final IGRecordField igField,
      final VariableType varType) {

//    final String prefix = recordName == null ? getRecordName(igField) : recordName;
    final String fieldName = /*prefix*/recordName + "." + igField.getId();
    final IGType recordType = igField.getType();
    final DataType dataType = getDataType(recordType);
    final NodeVariable variable = new NodeVariable(fieldName, dataType);
    final VariableDescriptor descriptor = new VariableDescriptor(varType);
    variable.setUserData(descriptor);
    return  variable;
  }

/*  public String getRecordName(final IGItem field) {

    final SourceLocation fieldLoc = field.computeSourceLocation();

    final Map<SourceLocation, String> locMap = getParentVisitor().getLocationMap();

    if (locMap.containsKey(fieldLoc)) {

      return locMap.get(fieldLoc);

    } else {
      System.out.println("field not found");

      return null;
    }
  }*/
}
