/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl.mapper;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.parser.vhdl.VhdlParser;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between logic and bit-vector operation identifiers.
 *
 * <p>Mapping for transforming operation codes for logic types
 * into operation codes for bit vector.
 * Mapping is used at VHDL expressions elaboration in
 * {@link VhdlParser} component.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class BitVectorOpCodeMapper {

  private static final Map<StandardOperation, StandardOperation> OPERATIONS;

  private BitVectorOpCodeMapper() {
    /* Empty. */
  }

  static {
    OPERATIONS = new LinkedHashMap<>();

    OPERATIONS.put(StandardOperation.EQ, StandardOperation.EQ);
    OPERATIONS.put(StandardOperation.NOTEQ, StandardOperation.NOTEQ);
    OPERATIONS.put(StandardOperation.ADD, StandardOperation.BVADD);
    OPERATIONS.put(StandardOperation.SUB, StandardOperation.BVSUB);
    OPERATIONS.put(StandardOperation.AND, StandardOperation.BVAND);
    OPERATIONS.put(StandardOperation.OR, StandardOperation.BVOR);
    OPERATIONS.put(StandardOperation.XOR, StandardOperation.BVXOR);
    OPERATIONS.put(StandardOperation.NOT, StandardOperation.BVNOT);
    OPERATIONS.put(StandardOperation.MUL, StandardOperation.BVMUL);
    OPERATIONS.put(StandardOperation.MINUS, StandardOperation.BVNEG);
    OPERATIONS.put(StandardOperation.REM, StandardOperation.BVSREM);
    OPERATIONS.put(StandardOperation.MOD, StandardOperation.BVSMOD);
  }

  /**
   * Checks whether operation is supported.
   *
   * @param operationId Identifier for the checked operation.
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasOperation(final StandardOperation operationId) {
    InvariantChecks.checkNotNull(operationId);

    return OPERATIONS.containsKey(operationId);
  }

  /**
   * Gets bit vector analogous operation code of the specified operation code.
   *
   * @param operation Identifier for the operation.
   *
   * @return Corresponding operation code.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static StandardOperation getOperation(final StandardOperation operation) {
    InvariantChecks.checkNotNull(operation);

    return OPERATIONS.get(operation);
  }
}
