/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl.mapper;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG subprogram identifiers and standard operation identifiers that are supported.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class IgSubprogramMapper {

  private static final Map<String, StandardOperation> OPERATIONS;
  private static final Map<String, Macro> MACROS;

  private IgSubprogramMapper() {
    /* Empty. */
  }

  private enum Macro {
    BV_INCREMENT,
    BV2INT
  }

  static {
    OPERATIONS = new LinkedHashMap<>();

    OPERATIONS.put("=", StandardOperation.EQ);
    OPERATIONS.put("<=", StandardOperation.LESSEQ);
    OPERATIONS.put("<", StandardOperation.LESS);
    OPERATIONS.put(">=", StandardOperation.GREATEREQ);
    OPERATIONS.put(">", StandardOperation.GREATER);
    OPERATIONS.put("+", StandardOperation.ADD);
    OPERATIONS.put("-", StandardOperation.SUB);
    OPERATIONS.put("AND", StandardOperation.AND);
    OPERATIONS.put("OR", StandardOperation.OR);
    OPERATIONS.put("NOT", StandardOperation.NOT);
    OPERATIONS.put("/=", StandardOperation.NOTEQ);
    OPERATIONS.put("&", StandardOperation.BVCONCAT);
    OPERATIONS.put("XOR", StandardOperation.BVXOR);
    OPERATIONS.put("*", StandardOperation.MUL);
    OPERATIONS.put("MOD", StandardOperation.MOD);
    OPERATIONS.put("/", StandardOperation.DIV);
    OPERATIONS.put("**", StandardOperation.POWER);
    OPERATIONS.put("NOR", StandardOperation.BVNOR);
    OPERATIONS.put("BV_NEGATE", StandardOperation.BVNEG);
    OPERATIONS.put("BV_ADDER", StandardOperation.BVADD);

    MACROS = new LinkedHashMap<>();
    MACROS.put("BV_INCREMENT", Macro.BV_INCREMENT);
    MACROS.put("CONV_INTEGER", Macro.BV2INT);
  }

  /**
   * Checks whether operation is supported.
   *
   * @param operation Operation identifier.
   *
   * @return {@code true} if operation is supported, {@code true} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasOperation(final String operation) {
    InvariantChecks.checkNotNull(operation);

    return OPERATIONS.containsKey(operation);
  }

  /**
   * Checks whether this macro-definition is supported.
   *
   * @param name The macro-definition name.
   *
   * @return {@code true} if macro-definition is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasMacro(final String name) {
    InvariantChecks.checkNotNull(name);

    return MACROS.containsKey(name);
  }

  /**
   * Gets operation code of the operation with the specified name.
   *
   * @param opName Operation name.
   *
   * @return Operation code.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when {@code IgSubprogramMapper}
   *         does not contain an operation with the specified string name.
   */
  public static StandardOperation getOperationCode(final String opName) {
    InvariantChecks.checkNotNull(opName);

    if (hasOperation(opName)) {
      return OPERATIONS.get(opName);
    }
    throw new UnsupportedOperationException(opName);
  }

  /**
   * Returns the syntax node for the macro-definition that is described by it's name and operands.
   *
   * @param name The macro-definition name.
   * @param operands Operands for the macro-definition.
   * @return The syntax node for the macro-definition that is described by it's name and operands.
   */
  public static Node getMacroOperation(final String name, final Node ... operands) {

    if (hasMacro(name)) {

      final Node operand = operands[0];

      switch (MACROS.get(name)) {
        case BV_INCREMENT:
          return Nodes.bvadd(operand, NodeValue.newBitVector(1, operand.getDataType().getSize()));
        case BV2INT:
          return Nodes.bv2int(operand);
        default:
          throw new UnsupportedOperationException(name);
      }
    }
    throw new UnsupportedOperationException(name);
  }
}
