/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.property;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * Assertion walker.
 * <p>Provides methods for {@link PropertyEntity} entity and it's
 * sub-components walking.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
class PropertyWalker {

  private final PropertyVisitor visitor;

  /**
   * Constructs walker with the specified visitor object.
   *
   * @param visitor Entity visitor to be used through walking.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public PropertyWalker(final PropertyVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;
  }

  /**
   * Starts walking of the specified entity.
   *
   * @param entity Entity to be traversed.
   */
  public void start(final PropertyEntity entity) {
    InvariantChecks.checkNotNull(entity);

    this.visitor.onRootBegin();

    visitEntity(entity);

    this.visitor.onRootEnd();
  }

  private void visitEntity(final PropertyEntity entity) {

    this.visitor.onEntityBegin(entity);

    visitModules(entity.getContainerMap().values());

    this.visitor.onEntityEnd(entity);
  }

  private void visitModules(final Iterable<ModuleProperty> properties) {

    for (final ModuleProperty property : properties) {

      visitModuleProperty(property);
    }
  }

  private void visitModuleProperty(final ModuleProperty property) {

    this.visitor.onContainerBegin(property);

    visitModules(property.getModuleMap().values());

    visitProcesses(property.getProcessMap().values());

    this.visitor.onContainerEnd(property);
  }

  private void visitProcesses(final Iterable<ProcessProperty> properties) {

    for (final ProcessProperty property : properties) {

      this.visitor.onProcessProperty(property);
    }
  }
}
