/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of an input/output vector of variables. A vector is consisted of a set of
 * transactions along with a list of attendant events (for example clock or reset signal change).
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Transaction
 * @see EventList
 */
@XmlType(propOrder = {"events", "transactions"}, name = "vector")
@XmlRootElement
public class TestVector {
  private static final String LEFT_BOUND = "{";
  private static final String EVENT_HEADER = "events: ";
  private static final String EVENT_DELIM = "; ";
  private static final String TRANS_HEADER = "transaction: ";
  private static final String TRANS_DELIM = "; ";
  private static final String RIGHT_BOUND = "}";

  @XmlElement
  private EventList events;
  @XmlElement(name = "transaction")
  private final Set<Transaction> transactions;

  /**
   * Constructs an empty vector.
   */
  public TestVector() {
    this.transactions = new HashSet<>();
    this.events = new EventList();
  }

  /**
   * Constructs a vector containing the specified transaction.
   *
   * @param transaction the transaction to be placed in the vector
   */
  public TestVector(final Transaction transaction) {
    this();

    this.transactions.add(transaction);
  }

  /**
   * Constructs a vector containing the specified transactions and events.
   *
   * @param transactions    a collection of the transactions
   * @param events          the events
   */
  public TestVector(final Collection<Transaction> transactions, final EventList events) {
    this.transactions = new HashSet<>(transactions);
    this.events = events;
  }

  /**
   * Adds the specified transaction to this vector.
   *
   * @param transaction the transaction to be added
   */
  public void addTransaction(final Transaction transaction) {
    this.transactions.add(transaction);
  }

  /**
   * Adds the specified event to the vector.
   *
   * @param event   the event to be added
   */
  public void addEvent(final Event event) {
    this.events.add(event);
  }

  /**
   * Adds the specified events to this vector.
   *
   * @param events   the events to be added
   */
  public void addEvents(final EventList events) {

    for (final Event event : events) {
      this.events.add(event);
    }
  }

  /**
   * Clears all the internal events for this objects.
   */
  public void removeEvents() {
    this.events = new EventList();
  }

  /**
   * Checks whether this object has events.
   *
   * @return {@code true} when this object has events, {@code false} otherwise.
   */
  public boolean hasEvents() {
    return !getEventList().isEmpty();
  }

  /**
   * Returns an {@link EventList} associated with this vector.
   *
   * @return the event list
   */
  public EventList getEventList() {
    return this.events;
  }

  /**
   * Returns all transactions included in this vector.
   *
   * @return a set of the transactions representing this vector
   */
  public Set<Transaction> getTransactions() {
    return new HashSet<>(this.transactions);
  }

  /**
   * Checks if this vector is empty.
   *
   * @return {@code true} if this vector contains neither any non-empty transactions nor any
   *         events. Otherwise returns {@code false}
   */
  public boolean isEmpty() {
    if (this.events.isEmpty()) {
      for (final Transaction transaction : this.transactions) {
        if (!transaction.isEmpty()) {
          return false;
        }
      }

      return true;
    } else {
      return false;
    }
  }

  /**
   * Checks if the specified variable is in this vector.
   *
   * @param name a name of the variable.
   *
   * @return {@code true} if this vector contains the specified variable. Otherwise returns
   * {@code false}
   */
  public boolean containsVariable(final String name) {

    for (final Transaction transaction : this.transactions) {

      if (transaction.containsVariable(name)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Returns assignments.
   *
   * @return assignments
   */
  public Map<String, NodeValue> getAssignments() {

    final Map<String, NodeValue> assignments = new HashMap<>();

    for (final Transaction transaction : this.transactions) {
      assignments.putAll(transaction.getValues());
    }

    return assignments;
  }

  /**
   * Returns a value of the specified variable from this vector.
   *
   * @param variableName    a name of the variable
   *
   * @return the variable's value if the specified variable is in this vector; otherwise returns
   * {@code null}
   */
  public Node getValue(final String variableName) {

    for (final Transaction transaction : this.transactions) {

      if (transaction.containsVariable(variableName)) {
        return transaction.getValue(variableName);
      }
    }

    return null;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.events.hashCode();
    result = prime * result + this.transactions.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this vector for equality. The object is considered as
   * equal to this vector if and only if it is also a vector and it contains the same events
   * and transactions as this vector does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code TestVector} class or its subclass;
   *    <li>its {@code EventList} is equal to this vector's one;
   *    <li>its contains the same amount of transactions as this vector does;
   *    <li>for each its transaction there is a equal one in this vector.
   * </ul>
   *
   * @param obj the object to be compared for equality with this vector
   *
   * @return {@code true} if the specified object is equal to this vector. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof TestVector)) {
      return false;
    }

    final TestVector vectorToCompare = (TestVector) obj;

    return this.events.equals(vectorToCompare.events)
        && this.transactions.equals(vectorToCompare.transactions);
  }

  @Override
  public String toString() {

    final StringBuilder stringBuilder = new StringBuilder(TestVector.LEFT_BOUND);
    stringBuilder.append(
        String.format("%s%s%s", TestVector.EVENT_HEADER, this.events, TestVector.EVENT_DELIM));

    final int initLength = stringBuilder.length();

    for (final Transaction transaction : this.transactions) {

      if (stringBuilder.length() > initLength) {
        stringBuilder.append(TestVector.TRANS_DELIM);
      }

      stringBuilder.append(String.format("%s%s", TestVector.TRANS_HEADER, transaction));
    }

    stringBuilder.append(TestVector.RIGHT_BOUND);

    return stringBuilder.toString();
  }
}
