/*
 * Copyright 2014-2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.testbench;

import ru.ispras.retrascope.basis.FileNames;

import java.nio.file.Path;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * The representation of an abstract testbench.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class Testbench {
  private final Path testbenchFile;
  private final Set<Path> additionalFiles;

  /**
   * Constructs a new testbench with the specified testbench file and additional files (for
   * example, test pattern and library files).
   *
   * @param testbenchFile a path to the testbench file. It is accessible via the
   * {@link Testbench#getTestbenchFile()}, {@link FileNames#getArray()} and
   * {@link FileNames#getList()} methods
   * @param additionalFiles a set containing additional files of the testbench
   *
   * @throws IllegalArgumentException when additional files collection is {@code null}
   */
  public Testbench(final Path testbenchFile, final Set<Path> additionalFiles) {
    this.testbenchFile = testbenchFile;
    this.additionalFiles = new HashSet<>(additionalFiles);
  }

  /**
   * Returns a path to the testbench file.
   *
   * @return a path to the testbench file
   */
  public Path getTestbenchFile() {
    return this.testbenchFile;
  }

  /**
   * Returns additional files (for example, test pattern and library files) related to this
   * testbench.
   *
   * @return a set of the additional files
   */
  public Set<Path> getAdditionalFiles() {
    return Collections.unmodifiableSet(this.additionalFiles);
  }

  @Override
  public int hashCode() {
    return 31 * super.hashCode() + this.additionalFiles.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }

    if (!super.equals(obj)) {
      return false;
    }

    if (!(obj instanceof Testbench)) {
      return false;
    }

    final Testbench other = (Testbench) obj;

    return this.additionalFiles.equals(other.additionalFiles);
  }
}
