/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.AssignmentContainer;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Utility methods for models transformation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ModelTransformUtils {

  private ModelTransformUtils() {
    /* Empty. */
  }

  /**
   * Performs backward substitution on the specified nodes to the specified expression.
   *
   * <p>The substitution process moves in a backward direction (from last node of the list
   * to the first). The result is an expression where all the target variables that are
   * defined at least one {@link AssignmentContainer} node from the specified collection
   * will be substituted by the value expressions that are assigned to them.</p>
   *
   * @param blocks Blocks which should be used as a source of substitution.
   * @param node Expression to be transformed.
   *
   * @return Modified expression.
   */
  public static Node substituteBackward(
      final List<? extends AssignmentContainer> blocks,
      final Node node) {

    Node nodeCopy = node.deepCopy();

    /* Classify assignments on blocking and concurrent. */

    final List<AssignmentContainer> concurrent = new LinkedList<>();
    final List<AssignmentContainer> blocking = new LinkedList<>();

    for (final AssignmentContainer block : blocks) {

      if (block.isConcurrent()) {
        concurrent.add(block);
      } else {
        blocking.add(block);
      }
    }

    /* Apply concurrent assignments simultaneously. */

    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final AssignmentContainer container : concurrent) {

      final List<Assignment> assignments = container.getAssignments();

      for (final Assignment assignment : assignments) {

        final RangedVariable target = assignment.getTarget();
        final Node value = assignment.getValue();
        boundVariables.add(
            NodeBinding.bindVariable(
                assignment.getVariable(),
                (target.ranged() || target.indexed())
                    ? RangedVariableUtils.transformNodeWithRanged(target, value)
                    : value));
      }
    }
    nodeCopy = Transformer.substituteAllBindings(new NodeBinding(nodeCopy, boundVariables));

    /* Apply blocking assignments in backward direction. */
    for (int i = blocking.size() - 1; i >= 0; i--) {

      final AssignmentContainer block = blocking.get(i);
      final List<Assignment> assignments = block.getAssignments();

      /* Elaborate assignments in backward direction. */
      for (int j = assignments.size() - 1; j >= 0; j--) {

        final Assignment assignment = assignments.get(j);
        final RangedVariable target = assignment.getTarget();
        final NodeVariable targetVariable = target.getVariable();
        final Node value = assignment.getValue();

        final Collection<NodeVariable> variables = ExprUtils.getVariables(nodeCopy);

        if (variables.contains(targetVariable)) {

          nodeCopy = substituteNode(nodeCopy, target, value);
        }
      }
    }

    /* Call the reducer to simplify the substitution result. */
    return Reducer.reduce(nodeCopy);
  }

  private static Node substituteNode(
      final Node node,
      final RangedVariable target,
      final Node value) {

    final Node nonRangedValue = RangedVariableUtils.transformNodeWithRanged(target, value);
    final NodeVariable targetVar = target.getVariable();
    return TransformerUtils.substitute(node, targetVar.getName(), nonRangedValue);
  }
}
