/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;

/**
 * Read and write utilities.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ReadWriteUtils {

  private ReadWriteUtils() {
    /* Empty. */
  }

  /**
   * Returns reader for the file with the specified full name.
   * @param fileName The name of the file to be read.
   * @return Reader for the file with the specified name.
   */
  public static BufferedReader newBufferedReader(final String fileName) {

    try {
      return new BufferedReader(
          new InputStreamReader(
              new FileInputStream(fileName), Charset.defaultCharset()));
    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Returns reader for the specified stream object.
   * <p>Default charset is used for reading.</p>
   * @param stream Stream to be read.
   * @return Reader for the specified stream object.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static BufferedReader newBufferedReader(final InputStream stream) {
    InvariantChecks.checkNotNull(stream);

    return new BufferedReader(new InputStreamReader(stream, Charset.defaultCharset()));
  }

  /**
   * Returns reader for the specified path object.
   * <p>Default charset is used for reading.</p>
   * @param path Path from which data to be read.
   * @return Reader for the specified path object.
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when I/O error occurred.
   */
  public static BufferedReader newBufferedReader(final Path path) {
    InvariantChecks.checkNotNull(path);

    try {
      return Files.newBufferedReader(path, Charset.defaultCharset());

    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Returns writer to the specified stream object.
   * <p>Default charset is used for writing.</p>
   * @param stream Stream to write to.
   * @return Writer to the specified stream object.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static BufferedWriter newBufferedWriter(final OutputStream stream) {
    InvariantChecks.checkNotNull(stream);

    return new BufferedWriter(new OutputStreamWriter(stream, Charset.defaultCharset()));
  }

  /**
   * Returns writer to the specified file object.
   * <p>Default charset is used for writing.</p>
   * @param file File to write to.
   * @return Writer to the specified file object.
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when the specified file already contains data
   *         or when I/O error occurred.
   */
  public static BufferedWriter newBufferedWriter(final File file) {
    InvariantChecks.checkNotNull(file);

    try {
      return new BufferedWriter(
          new OutputStreamWriter(
              new FileOutputStream(file),
              Charset.defaultCharset()));

    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Returns writer to the specified path object.
   * @param path Path to write to.
   * @return Writer to the specified path object.
   * @throws IllegalArgumentException is {@code null}.
   * @throws RetrascopeException when the specified file already exists
   *         or when an I/O error occurred.
   */
  public static BufferedWriter newBufferedWriter(final Path path) {
    InvariantChecks.checkNotNull(path);

    try {
      return Files.newBufferedWriter(path, Charset.defaultCharset());
    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Returns writer to the specified file object.
   * <p>Default charset is used for writing.</p>
   * @param file File to write to.
   * @return Writer to the specified file object.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static PrintWriter newPrintWriter(final File file) {
    InvariantChecks.checkNotNull(file);

    return new PrintWriter(newBufferedWriter(file));
  }

  /**
   * Returns the base name of parent directory for the file is described in the specified path.
   * @param filePath The path to file.
   * @return The base name of parent directory for the file is described in the specified path.
   */
  public static String getParentDirName(final String filePath) {

    final String sep = File.separator;
    return filePath.substring(filePath.lastIndexOf(sep, filePath.lastIndexOf(sep) - 1) + 1);
  }

  /**
   * Clears the specified existing file.
   * @param file The existing file to be cleared.
   */
  public static void clear(final File file) {

    if (!file.exists()) {
      return;
    }

    final PrintWriter writer = ReadWriteUtils.newPrintWriter(file);

    writer.write("");
    writer.close();
  }

  /**
   * Checks whether the file with the specified name exists.
   * @param name The file name.
   * @return {@code true} when the specified file exists, {@code false} otherwise.
   */
  public static boolean exists(final String name) {
    final File file = new File(name);
    return file.exists();
  }
}
