/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.util;

import org.apache.commons.lang3.EnumUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;

/**
 * This enum represents SVA, PSL, LTL and CTL operators.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public enum TemporalOperation {
  /*
   * The items below belong to the "SystemVerilog Assertions" group.
   */
  ASSERT(Family.SVA),
  ASSUME(Family.SVA),

  /*
   * The items below belong to the "PSL Assertions" group.
   */
  SERE_CONCAT(Family.PSL),
  SERE_FUSE(Family.PSL),
  SERE_OR(Family.PSL),
  SERE_NON_LENGTH_AND(Family.PSL),
  SERE_LENGTH_AND(Family.PSL),
  SERE_WITHIN(Family.PSL),

  SERE_CONS_REP_N(Family.PSL),
  SERE_CONS_REP_N_M(Family.PSL),
  SERE_CONS_REP_N_INF(Family.PSL),

  SERE_NON_CONS_REP_N(Family.PSL),
  SERE_NON_CONS_REP_N_M(Family.PSL),
  SERE_NON_CONS_REP_N_INF(Family.PSL),

  SERE_GOTO_REP_N(Family.PSL),
  SERE_GOTO_REP_N_M(Family.PSL),
  SERE_GOTO_REP_N_INF(Family.PSL),
  SERE_GOTO_REP_LAST(Family.PSL),

  PSL_ALWAYS(Family.PSL),
  PSL_NEVER(Family.PSL),
  PSL_EVENTUALLY(Family.PSL),
  PSL_NEXT_STRONG(Family.PSL),
  PSL_NEXT(Family.PSL),
  PSL_NEXT_I_STRONG(Family.PSL),
  PSL_NEXT_I(Family.PSL),

  // TODO: add extended operators

  PSL_SUFF_IMPL_INC(Family.PSL),
  PSL_SUFF_IMPL_EXC(Family.PSL),

  PSL_IMPL(Family.PSL),
  PSL_IFF(Family.PSL),
  PSL_AND(Family.PSL),
  PSL_OR(Family.PSL),
  PSL_NOT(Family.PSL),

  /*
   * The items below belong to the "LTL Assertions" group.
   */
  LTL_F(Family.LTL),
  LTL_G(Family.LTL),
  LTL_X(Family.LTL),
  LTL_U(Family.LTL),
  LTL_W(Family.LTL),

  /*
   * The items below belong to the "CTL Assertions" group.
   */

  // TODO: add CTL operators
  CTL_EF(Family.CTL),
  CTL_U(Family.CTL),
  CTL_E(Family.CTL);

  /**
   * Logic family.
   */
  public enum Family {
    SVA, PSL, LTL, CTL
  }

  private Family family;

  TemporalOperation(final Family family) {
    this.family = family;
  }

  /**
   * Checks whether the specified node is a temporal operation.
   *
   * @param node - the specified node.
   * @return {@code true} if the specified node is a temporal NodeOperation, {@code false}
   *     otherwise.
   */
  public static boolean isTemporal(final Node node) {
    return node instanceof NodeOperation && EnumUtils
        .isValidEnum(TemporalOperation.class, ((NodeOperation) node).getOperationId().name());
  }

  /**
   * Returns the logic family of the specified NodeOperation.
   *
   * @param node - the specified operation.
   * @return The logic family of the specified operation if it is temporal, {@code null} otherwise.
   */
  public static Family getFamily(final NodeOperation node) {
    return isTemporal(node) ? ((TemporalOperation) node.getOperationId()).getFamily() : null;
  }

  /**
   * Returns the logic family of the operator.
   *
   * @return The logic family.
   */
  public Family getFamily() {
    return family;
  }

}
