/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.parser.basis.HdlParser;

import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The {@link ToolTest} class is a base class for all JUnit test
 * cases that run the tool with some parameters.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class ToolTest {

  private final Map<String, String> options;
  private List<String> args;

  /**
   * Default constructor.
   */
  public ToolTest() {

    this.options = new LinkedHashMap<>();
    this.args = new LinkedList<>();
  }

  private Map<String, String> getOptions() {

    return this.options;
  }

  private List<String> getArgs() {

    return this.args;
  }

  /**
   * Returns the target module name.
   * @return The target module name.
   */
  protected abstract String getModuleName();

  /**
   * Adds option long name and, probably, it's value to the tool command line.
   * @param optLongName The option long name.
   * @param value The option value.
   */
  protected void addOption(final String optLongName, final String value) {

    this.options.put(optLongName, value);
  }

  /**
   * Adds command line parameter and, probably, it's value to the tool command line.
   * @param param The command line parameter.
   * @param value The option value.
   */
  protected void addOption(final Parameter param, final String value) {

    this.options.put(param.getName(), value);
  }

  /**
   * Adds no-value option with the specified long name to the tool command line.
   *
   * @param optLongName The option long name.
   */
  protected void addOption(final String optLongName) {
    addOption(optLongName, null);
  }

  /**
   * Adds no-value parameter to the tool command line.
   *
   * @param param The parameter.
   */
  protected void addOption(final Parameter param) {
    addOption(param, null);
  }

  /**
   * Runs the tool with the specified arguments.
   * @param args The argument strings.
   */
  public void runTest(final String ... args) {

    final String moduleName = getModuleName();

    if (moduleName != null) {
      addOption(HdlParser.MODULE_NAME, moduleName);
    }
    addArgs(args);
    Retrascope.main(makeArgs());
  }

  private void addArgs(final String ... args) {
    this.args.addAll(Arrays.asList(args));
  }

  private String[] makeArgs() {

    final List<String> args = new LinkedList<>(getArgs());

    for (final Map.Entry<String, String> entry : getOptions().entrySet()) {

      final String key = entry.getKey();
      args.add(key);

      final String value = entry.getValue();

      if (value != null) {
        args.add(value);
      }
    }

    return args.toArray(new String[0]);
  }
}
