/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Set of expression transformation rules.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class TransformerRuleSet {

  /* Operations that produce boolean result. */
  private static final List<StandardOperation> BOOL_GROUP = Arrays.asList(
      StandardOperation.AND,
      StandardOperation.OR,
      StandardOperation.NOT,
      StandardOperation.XOR,
      StandardOperation.IMPL);

  /* Operations that compare same type arguments. */
  private static final List<StandardOperation> COMPARE_GROUP = Arrays.asList(
      StandardOperation.EQ,
      StandardOperation.NOTEQ,
      StandardOperation.GREATER,
      StandardOperation.GREATEREQ,
      StandardOperation.LESS,
      StandardOperation.LESSEQ);

  private static final List<StandardOperation> ARITHM_GROUP = Arrays.asList(
      StandardOperation.ADD,
      StandardOperation.SUB);

  private TransformerRuleSet() {
    /* Empty. */
  }

  static class CompareOpDescriptor {
    NodeValue rhs;
    NodeValue lhs;
    NodeValue module;
    NodeVariable var;

    CompareOpDescriptor() {
      /* Empty. */
    }

    NodeValue getRightHandSide() {
      return (NodeValue) this.rhs.deepCopy();
    }

    void setRightHandSide(final NodeValue value) {
      this.rhs = value;
    }

    NodeValue getLeftHandSide() {
      return (NodeValue) this.lhs.deepCopy();
    }

    void setLeftHandSide(final NodeValue value) {
      this.lhs = value;
    }

    NodeValue getModule() {
      return (NodeValue) this.module.deepCopy();
    }

    void setModule(final NodeValue value) {
      this.module = value;
    }

    NodeVariable getVariable() {
      return (NodeVariable) this.var.deepCopy();
    }

    void setVariable(final  NodeVariable variable) {
      this.var = variable;
    }
  }

  /**
   * Returns set of transformation rules that are applied while standardization process.
   *
   * @return Mapping from operation identifiers to transformation rules.
   */
  static NodeTransformer getRuleSet() {

    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule booleanGroupRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        if (ExprUtils.isKind(Node.Kind.OPERATION, node)) {

          final NodeOperation operation = (NodeOperation) node;

          InvariantChecks.checkTrue(operation.getOperationId() instanceof StandardOperation);

          StandardOperation opId = (StandardOperation) operation.getOperationId();

          return BOOL_GROUP.contains(opId);
        }
        return false;
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation operation = (NodeOperation) node;
        final List<Node> args = new ArrayList<>();

        for (final Node child : operation.getOperands()) {

          final Node converted = TypeConversion.coerce(child, DataType.BOOLEAN);

          // not integral node, maybe throw exception?
          if (converted == null) {
            return node;
          }
          args.add(converted);
        }
        return new NodeOperation(operation.getOperationId(), args);
      }
    };

    final TransformerRule iteRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.ITE);
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation ite = (NodeOperation) node;
        final List<Node> args = new ArrayList<>(ite.getOperands());
        final Node condition = TypeConversion.coerce(args.get(0), DataType.BOOLEAN);

        // not integral node, maybe throw exception?
        if (condition == null) {
          return node;
        }
        args.set(0, condition);

        return Nodes.ite(args.get(0), args.get(1), args.get(2));
      }
    };

    final TransformerRule expandEq = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.EQ)
            && ((NodeOperation) node).getOperandCount() > 2;
      }

      @Override
      public Node apply(final Node node) {

        final List<Node> operands = ((NodeOperation)node).getOperands();
        final Node[] equations = new NodeOperation[operands.size() - 1];

        for (int i = 1; i < operands.size(); i++) {
          equations[i - 1] = Nodes.eq(operands.get(i), operands.get(0));
        }

        return ExprUtils.getConjunction(equations);
      }
    };

    final TransformerRule addRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return ExprUtils.isOperation(node, COMPARE_GROUP)
            && hasAddSubOperand(this.descriptor, (NodeOperation) node, StandardOperation.ADD);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Reducer.reduce(
                Nodes.sub(
                    this.descriptor.getRightHandSide(),
                    this.descriptor.getLeftHandSide())));
      }
    };

    final TransformerRule addByModuleRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return ExprUtils.isOperation(node, COMPARE_GROUP)
            && hasAddByModuleOperand(this.descriptor, (NodeOperation) node, StandardOperation.ADD);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Reducer.reduce(
                Nodes.mod(
                    Nodes.sub(
                        this.descriptor.getRightHandSide(),
                        this.descriptor.getLeftHandSide()),
                this.descriptor.getModule())));
      }
    };

    final TransformerRule subRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return ExprUtils.isOperation(node, COMPARE_GROUP)
            && hasAddSubOperand(this.descriptor, (NodeOperation) node, StandardOperation.SUB);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Reducer.reduce(
                Nodes.add(
                    this.descriptor.getRightHandSide(),
                    this.descriptor.getLeftHandSide())));
      }
    };

    final TransformerRule subByModuleRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return ExprUtils.isOperation(node, COMPARE_GROUP)
            && hasAddByModuleOperand(this.descriptor, (NodeOperation) node, StandardOperation.SUB);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Reducer.reduce(
                Nodes.mod(
                    Nodes.add(
                        this.descriptor.getRightHandSide(),
                        this.descriptor.getLeftHandSide()),
                    this.descriptor.getModule())));
      }
    };

    /* Compress chained BVEXTRACT operations. */
    final TransformerRule chainedBvExtractRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.BVEXTRACT)
            && hasBvExtractOperand((NodeOperation) node);
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation operation = (NodeOperation) node;

        final Node nodeHigh = operation.getOperand(0);
        final Node nodeLow = operation.getOperand(1);
        final NodeOperation operand = (NodeOperation) operation.getOperand(2);

        final Node opHigh = operand.getOperand(0);
        final Node opLow = operand.getOperand(1);

        /* For transformation to be correct the precondition should be met. */
        if (!SolverUtils.isSat(
            Nodes.and(
                Nodes.greatereq(Nodes.sub(opHigh, opLow), nodeHigh),
                Nodes.greatereq(nodeHigh, nodeLow)))) {
          throw new IndexOutOfBoundsException("Wrong indices in: " + node);
        }

        final Node subOperand = operand.getOperand(2);

        return Nodes.bvextract(
            Reducer.reduce(Nodes.add(nodeHigh, opLow)),
            Reducer.reduce(Nodes.add(nodeLow, opLow)),
            subOperand);
      }
    };

    /* Check order of const parameters for BVEXTRACT operation. */
    final TransformerRule constParamOrderBvExtractRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        if (!ExprUtils.isKind(Node.Kind.OPERATION, node)) {
          return false;
        }

        final NodeOperation operation = (NodeOperation) node;

        if (!ExprUtils.isOperation(operation, StandardOperation.BVEXTRACT)) {
          return false;
        }

        final Node high = operation.getOperand(0);
        final Node low = operation.getOperand(1);

        if (!ExprUtils.isKind(Node.Kind.VALUE, high, low)) {
          return false;
        }

        final NodeValue highVal = (NodeValue) high;
        final NodeValue lowVal = (NodeValue) low;

        if (ExprUtils.isType(DataType.INTEGER, highVal, lowVal)) {

          final BigInteger highInt = highVal.getInteger();
          final BigInteger lowInt = lowVal.getInteger();

          return highInt.compareTo(lowInt) < 0;

        } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, highVal, lowVal)) {

          final BigInteger highInt = highVal.getBitVector().bigIntegerValue();
          final BigInteger lowInt = lowVal.getBitVector().bigIntegerValue();

          return highInt.compareTo(lowInt) < 0;
        }

        return false;
      }

      @Override
      public Node apply(final Node node) {

        final Node high = ((NodeOperation)node).getOperand(0);
        final Node low = ((NodeOperation)node).getOperand(1);
        final Node operand = ((NodeOperation)node).getOperand(2);

        return Nodes.bvextract(TransformerUtils.toInt(low), TransformerUtils.toInt(high), operand);
      }
    };

    /* Modify BVEXTRACT operations with non-const parameters. */
    final TransformerRule varParamBvExtractRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        if (!ExprUtils.isKind(Node.Kind.OPERATION, node)) {
          return false;
        }

        final NodeOperation operation = (NodeOperation) node;

        if (!ExprUtils.isOperation(operation, StandardOperation.BVEXTRACT)) {
          return false;
        }

        final Node high = operation.getOperand(0);
        final Node low = operation.getOperand(1);

        return !ExprUtils.isKind(Node.Kind.VALUE, high, low);
      }

      @Override
      public Node apply(final Node node) {

        Node high = ((NodeOperation)node).getOperand(0);
        Node low = ((NodeOperation)node).getOperand(1);

        final DataType highType = high.getDataType();
        final DataType lowType = low.getDataType();

        if (!highType.equals(lowType) && NodeUtils.oneConstAnotherNot(high, low)) {

          /* When both const and var-containing params appear, cast first to last. */
          high = high.getKind() == Node.Kind.VALUE ? TypeConversion.coerce(high, lowType) : high;
          low = low.getKind() == Node.Kind.VALUE ? TypeConversion.coerce(low, highType) : low;
        }

        final Node operand = ((NodeOperation)node).getOperand(2);
        final int size = operand.getDataType().getSize();

        return Nodes.bvand(operand,
            Nodes.ite(
                NodeUtils.greater(high, low),
                mask(high, low, size),
                mask(low, high, size)));
      }
    };

    /* Modify SELECT operations that key type != index type. */
    final TransformerRule selectParamRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.SELECT);
      }

      @Override
      public Node apply(final Node node) {

        final Node array = ((NodeOperation) node).getOperand(0);

        final DataType arrayType = array.getDataType();
        final DataType keyType = (DataType) arrayType.getAttribute(DataTypeId.Attribute.KEY);

        final Node index = ((NodeOperation) node).getOperand(1);

        if (keyType.equals(index.getDataType())) {
          return node;
        } else {
          return Nodes.select(array, TypeConversion.coerce(index, keyType));
        }
      }
    };

    for (final Enum<?> op : BOOL_GROUP) {
      transformer.addRule(op, booleanGroupRule);
    }

    transformer.addRule(StandardOperation.ITE, iteRule);

    for (final Enum<?> op : COMPARE_GROUP) {

      transformer.addRule(op, addRule);
      transformer.addRule(op, subRule);
      transformer.addRule(op, addByModuleRule);
      transformer.addRule(op, subByModuleRule);
    }
    transformer.addRule(StandardOperation.EQ, expandEq);
    transformer.addRule(StandardOperation.BVEXTRACT, chainedBvExtractRule);
    transformer.addRule(StandardOperation.BVEXTRACT, constParamOrderBvExtractRule);
    //transformer.addRule(StandardOperation.BVEXTRACT, varParamBvExtractRule);
    transformer.addRule(StandardOperation.SELECT, selectParamRule);

    return transformer;
  }

  private static Node mask(final Node high, final Node low, final int size) {

    final BitVector ones = BitVector.newEmpty(size);
    ones.setAll();

    final Node init = NodeValue.newBitVector(ones);

    /* It is checked above that both params have same data type. */
    final DataTypeId typeId = high.getDataTypeId();

    Node mask = Nodes.bvlshr(init, low);
    mask = Nodes.bvlshl(mask,
        new NodeOperation(
            sub(typeId),
            new NodeOperation(
                add(typeId),
                low,
                NodeValue.newBitVector(size, low.getDataType().getSize())),
            new NodeOperation(
                add(typeId),
                high,
                one(high.getDataTypeId(), low.getDataType().getSize()))));
    mask = Nodes.bvlshr(mask,
        new NodeOperation(
            sub(typeId),
            NodeValue.newBitVector(size, low.getDataType().getSize()),
            new NodeOperation(
                add(typeId),
                high,
                one(high.getDataTypeId(), low.getDataType().getSize()))));
    return mask;
  }

  private static Enum<?> add(final DataTypeId typeId) {

    return typeId == DataTypeId.BIT_VECTOR ? StandardOperation.BVADD : StandardOperation.ADD;
  }

  private static Enum<?> sub(final DataTypeId typeId) {

    return typeId == DataTypeId.BIT_VECTOR ? StandardOperation.BVSUB : StandardOperation.SUB;
  }

  private static Node one(final DataTypeId typeId, final int size) {

    return typeId == DataTypeId.BIT_VECTOR
        ? NodeValue.newBitVector(1, size)
        : NodeValue.newInteger(1);
  }

  private static boolean hasBvExtractOperand(final NodeOperation operation) {

    final Node operand = operation.getOperand(2);

    return ExprUtils.isOperation(operand, StandardOperation.BVEXTRACT);
  }

  private static boolean hasAddByModuleOperand(
      final CompareOpDescriptor descriptor,
      final NodeOperation node,
      final StandardOperation lowOp) {

    final int opCount = node.getOperandCount();

    if (opCount != 2) {
      return false;
    }

    boolean constOp = false;
    boolean addByModOp = false;

    for (int i = 0; i < opCount; i++) {

      final Node operand = node.getOperand(i);

      if (!addByModOp && ExprUtils.isOperation(operand, StandardOperation.MOD)) {

        final int count = ((NodeOperation)operand).getOperandCount();

        if (count != 2) {
          return false;
        }

        boolean hasConstOperand = false;
        boolean hasVarOperand = false;

        for (int j = 0; j < count; j++) {

          final Node subOperand = ((NodeOperation)operand).getOperand(j);

          if (ExprUtils.isOperation(subOperand, lowOp) && !hasVarOperand) {

            boolean hasVarArg = false;
            boolean hasConstArg = false;

            for (int k = 0; k < ((NodeOperation) subOperand).getOperandCount(); k++) {

              final Node subSubOperand = ((NodeOperation) subOperand).getOperand(k);

              if (!hasVarArg && subSubOperand instanceof NodeVariable) {
                hasVarArg = true;
                descriptor.setVariable((NodeVariable) subSubOperand);
              }

              if (!hasConstArg && subSubOperand instanceof NodeValue) {
                hasConstArg = true;
                descriptor.setLeftHandSide((NodeValue) subSubOperand);
              }
            }
            hasVarOperand = hasVarArg && hasConstArg;
          }

          if (subOperand instanceof NodeValue && !hasConstOperand) {
            hasConstOperand = true;
            descriptor.setModule((NodeValue)subOperand);
          }
        }

        addByModOp = hasConstOperand && hasVarOperand;
      }
      if (operand instanceof NodeValue && !constOp) {
        constOp = true;
        descriptor.setRightHandSide((NodeValue) operand);
      }
    }

    return constOp && addByModOp;
  }

  private static boolean hasAddSubOperand(
      final CompareOpDescriptor descriptor,
      final NodeOperation node,
      final StandardOperation opCode) {

    final int opCount = node.getOperandCount();

    if (opCount != 2) {
      return false;
    }

    boolean constOperand = false;
    boolean addSubOperand = false;

    for (int i = 0; i < opCount; i++) {

      final Node operand = node.getOperand(i);

      if (!addSubOperand && ExprUtils.isOperation(operand, ARITHM_GROUP)) {

        final int subOpCount = ((NodeOperation)operand).getOperandCount();

        if (subOpCount != 2) {
          return false;
        }

        boolean hasVarArg = false;
        boolean hasConstArg = false;

        for (int j = 0; j < subOpCount; j++) {

          final Node subOperand = ((NodeOperation)operand).getOperand(j);

          if (!hasVarArg && subOperand instanceof NodeVariable) {
            hasVarArg = true;
            descriptor.setVariable((NodeVariable) subOperand);
          }
          if (!hasConstArg && subOperand instanceof NodeValue) {
            hasConstArg = true;
            descriptor.setLeftHandSide((NodeValue) subOperand);
          }
        }
        addSubOperand = hasVarArg && hasConstArg && ExprUtils.isOperation(operand, opCode);
      }
      if (operand instanceof NodeValue && !constOperand) {
        constOperand = true;
        descriptor.setRightHandSide((NodeValue) operand);
      }
    }

    return constOperand && addSubOperand;
  }
}
