/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.io.FilenameUtils;

import java.io.File;
import java.net.URL;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * HDL test files for project.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class HdlTestFiles {

  /**
   * The storage for VHDL test suite files.
   */
  public static final Map<String, Map<String, List<String>>> VHDL_MAP = new LinkedHashMap<>();

  /**
   * The storage for Verilog test suite files.
   */
  public static final Map<String, Map<String, List<String>>> VERILOG_MAP = new LinkedHashMap<>();

  /**
   * The stub module name for HDL files that do not have modules inside them.
   */
  public static final String NO_MODULE = "no_module_file";

  private HdlTestFiles() {
    /* Empty. */
  }

  static {

    /* === VHDL test projects === */

    /* bvsample */
    final Map<String, List<String>> bvSample = new LinkedHashMap<>();
    putToMap(bvSample, getFile("/bvsample/bvsample.vhd"));
    VHDL_MAP.put("bvsample", bvSample);

    /* sample */
    final Map<String, List<String>> sample = new LinkedHashMap<>();
    putToMap(sample, getFile("/sample/sample.vhd"));
    VHDL_MAP.put("sample", sample);

    /* example */
    final Map<String, List<String>> example = new LinkedHashMap<>();
    putToMap(example, getFile("/example/stub.vhd"));
    putToMap(example, getFile("/example/driver.vhd"));
    putToMap(example, "example",
        getFile("/example/example.vhd"),
        getFile("/example/stub.vhd"),
        getFile("/example/driver.vhd"));
    VHDL_MAP.put("example", example);

    /* ITC'99  benchmark */
    final Map<String, List<String>> itc99 = new LinkedHashMap<>();

    putToMap(itc99, getFile("/itc99/b01/b01.vhd"));
    putToMap(itc99, getFile("/itc99/b02/b02.vhd"));
    putToMap(itc99, getFile("/itc99/b03/b03.vhd"));
    putToMap(itc99, getFile("/itc99/b04/b04.vhd"));
    putToMap(itc99, getFile("/itc99/b05/b05.vhd"));
    putToMap(itc99, getFile("/itc99/b06/b06.vhd"));
    putToMap(itc99, getFile("/itc99/b07/b07.vhd"));
    putToMap(itc99, getFile("/itc99/b08/b08.vhd"));
    putToMap(itc99, getFile("/itc99/b09/b09.vhd"));
    putToMap(itc99, getFile("/itc99/b10/b10.vhd"));
    putToMap(itc99, getFile("/itc99/b11/b11.vhd"));
    putToMap(itc99, getFile("/itc99/b12/b12.vhd"));
    putToMap(itc99, getFile("/itc99/b13/b13.vhd"));
    putToMap(itc99, getFile("/itc99/b14/b14.vhd"));
    putToMap(itc99, getFile("/itc99/b14_1/b14_1.vhd"));
    putToMap(itc99, getFile("/itc99/b15/b15.vhd"));
    putToMap(itc99, getFile("/itc99/b15_1/b15_1.vhd"));
    putToMap(itc99, getFile("/itc99/b17/b17.vhd"));
    putToMap(itc99, getFile("/itc99/b17_1/b17_1.vhd"));
    putToMap(itc99, getFile("/itc99/b18/b18.vhd"));
    putToMap(itc99, getFile("/itc99/b18_1/b18_1.vhd"));
    putToMap(itc99, getFile("/itc99/b19/b19.vhd"));
    putToMap(itc99, getFile("/itc99/b19_1/b19_1.vhd"));
    putToMap(itc99, getFile("/itc99/b20/b20.vhd"));
    putToMap(itc99, getFile("/itc99/b20_1/b20_1.vhd"));
    putToMap(itc99, getFile("/itc99/b21/b21.vhd"));
    putToMap(itc99, getFile("/itc99/b21_1/b21_1.vhd"));
    putToMap(itc99, getFile("/itc99/b22/b22.vhd"));
    putToMap(itc99, getFile("/itc99/b22_1/b22_1.vhd"));
    putToMap(itc99, "inj_block", getFile("/itc99/b30/b30.vhd"));
    VHDL_MAP.put("itc99", itc99);

    /* MiniMIPS */
    final Map<String, List<String>> minimips = new LinkedHashMap<>();
    // TODO: uncomment designs when related features will be implemented
    //putToMap(minimips, getFile("/minimips/alu.vhd")); // to_integer function
    //putToMap(minimips, getFile("/minimips/banc.vhd"));  // to_integer function
    putToMap(minimips, getFile("/minimips/bus_ctrl.vhd"));
    //putToMap(minimips, getFile("/minimips/minimips.vhd"));
    //putToMap(minimips, getFile("/minimips/pps_di.vhd")); // record type
    putToMap(minimips, getFile("/minimips/pps_ei.vhd"));
    //putToMap(minimips, getFile("/minimips/pps_ex.vhd")); // to_integer function
    putToMap(minimips, getFile("/minimips/pps_mem.vhd"));
    putToMap(minimips, getFile("/minimips/pps_pf.vhd"));
    //putToMap(minimips, getFile("/minimips/ram.vhd")); // to_integer function
    //putToMap(minimips, getFile("/minimips/renvoi.vhd")); // to_integer function
    //putToMap(minimips, getFile("/minimips/rom.vhd")); // file_open function
    //putToMap(minimips, getFile("/minimips/syscop.vhd")); // to_integer function
    VHDL_MAP.put("minimips", minimips);

    /* Plasma */
    final Map<String, List<String>> plasma = new LinkedHashMap<>();
    // TODO: uncomment designs when related features will be implemented
    //putToMap(plasma, getFile("/plasma/alu.vhd")); // bv_adder function
    putToMap(plasma, getFile("/plasma/bus_mux.vhd"));
    putToMap(plasma, getFile("/plasma/control.vhd"));
    putToMap(plasma, getFile("/plasma/mem_ctrl.vhd"));
    putToMap(plasma, getFile("/plasma/mlite2sram.vhd"));
    //putToMap(plasma, getFile("/plasma/mlite2uart.vhd")); // records
    //putToMap(plasma, getFile("/plasma/mlite_cpu.vhd")); // bv_inc6
    //putToMap(plasma, getFile("/plasma/mult.vhd")); // bv_inc6
    putToMap(plasma, getFile("/plasma/pc_next.vhd")); // bv_increment
    putToMap(plasma, getFile("/plasma/pipeline.vhd"));
    //putToMap(plasma, getFile("/plasma/plasma.vhd")); // records
    //putToMap(plasma, getFile("/plasma/ram.vhd")); // Zamia error
    //putToMap(plasma, getFile("/plasma/reg_bank.vhd")); // conv_integer of variable args
    putToMap(plasma, getFile("/plasma/shifter.vhd"));
    putToMap(plasma, getFile("/plasma/sram2mlite.vhd"));
    //putToMap(plasma, getFile("/plasma/uart.vhd")); // records
    VHDL_MAP.put("plasma", plasma);

    /* === Verilog test projects === */

    /* adder */
    final Map<String, List<String>> adder = new LinkedHashMap<>();
    putToMap(adder, getFile("/adder/adder4.v"));
    VERILOG_MAP.put("adder", adder);

    /* array_ssa */
    final Map<String, List<String>> arraySsa = new LinkedHashMap<>();
    putToMap(arraySsa, getFile("/array_ssa_test/array_ssa_test.v"));
    VERILOG_MAP.put("array_ssa_test", arraySsa);

    /* blocks */
    final Map<String, List<String>> blocks = new LinkedHashMap<>();
    putToMap(blocks, getFile("/blocks/blocks.v"));
    VERILOG_MAP.put("blocks", blocks);

    /* bv_non_zero_idx */
    final Map<String, List<String>> bvNonZeroIdx = new LinkedHashMap<>();
    putToMap(bvNonZeroIdx, getFile("/bv_non_zero_idx/bv_non_zero_idx.v"));
    VERILOG_MAP.put("bv_non_zero_idx", bvNonZeroIdx);

    /* empty */
    final Map<String, List<String>> empty = new LinkedHashMap<>();
    putToMap(empty, NO_MODULE, getFile("/empty/empty.v"));
    VERILOG_MAP.put("empty", empty);

    /* fifo */
    final Map<String, List<String>> fifo = new LinkedHashMap<>();
    putToMap(fifo, getFile("/fifo/ram.v"));
    putToMap(fifo, "fifo", getFile("/fifo/fifo.v"), getFile("/fifo/ram.v"));
    VERILOG_MAP.put("fifo", fifo);

    /* fifo_mod */
    final Map<String, List<String>> fifoMod = new LinkedHashMap<>();
    putToMap(fifoMod, getFile("/fifo_mod/ram.v"));
    putToMap(fifoMod, "fifo_mod", getFile("/fifo_mod/fifo.v"), getFile("/fifo_mod/ram.v"));
    VERILOG_MAP.put("fifo_mod", fifoMod);

    /* for_loop */
    final Map<String, List<String>> forLoop = new LinkedHashMap<>();
    putToMap(forLoop, getFile("/for_loop/for_loop.v"));
    VERILOG_MAP.put("for_loop", forLoop);

    /* incompat_cond */
    final Map<String, List<String>> incompat = new LinkedHashMap<>();
    putToMap(incompat, getFile("/incompat_cond/incompat_cond.v"));
    VERILOG_MAP.put("incompat_cond", incompat);

    /* fifo0 */
    final Map<String, List<String>> fifo0 = new LinkedHashMap<>();
    putToMap(fifo0, getFile("/fifo0/mem_2p.v"));
    putToMap(fifo0, "myfifo", getFile("/fifo0/myfifo.v"), getFile("/fifo0/mem_2p.v"));
    putToMap(fifo0, "mul_fifo", getFile("/fifo0/mul_fifo.v"), getFile("/fifo0/myfifo.v"));
    VERILOG_MAP.put("fifo", fifo0);

    /* multi_assignment */
    final Map<String, List<String>> mult = new LinkedHashMap<>();
    putToMap(mult, getFile("/multi_assignment/multiple_assignment.v"));
    VERILOG_MAP.put("multi_assignment", mult);

    /* mux */
    final Map<String, List<String>> mux = new LinkedHashMap<>();
    putToMap(mux, getFile("/mux/mux4_1.v"));
    VERILOG_MAP.put("mux", mux);

    /* ram */
    final Map<String, List<String>> ram = new LinkedHashMap<>();
    putToMap(ram, getFile("/ram/ram.v"));
    VERILOG_MAP.put("ram", ram);

    /* Verilog projects from opencores.org */

    /* MIPS 16 bit */
    final Map<String, List<String>> mips16 = new LinkedHashMap<>();
    putToMap(mips16, getFile("/opencores/mips16/alu.v"));
    putToMap(mips16, getFile("/opencores/mips16/data_mem.v"));
    putToMap(mips16, "EX_stage",
        getFile("/opencores/mips16/EX_stage.v"),
        getFile("/opencores/mips16/alu.v"));
    putToMap(mips16, getFile("/opencores/mips16/hazard_detection_unit.v"));
    putToMap(mips16, getFile("/opencores/mips16/ID_stage.v"));
    putToMap(mips16, "IF_stage",
        getFile("/opencores/mips16/IF_stage.v"), getFile("/opencores/mips16/instruction_mem.v"));
    putToMap(mips16, getFile("/opencores/mips16/instruction_mem.v"));
    putToMap(mips16, getFile("/opencores/mips16/MEM_stage.v"));
    putToMap(mips16, "mips_16_core_top",
        getFile("/opencores/mips16/mips_16_core_top.v"),
        getFile("/opencores/mips16/hazard_detection_unit.v"),
        getFile("/opencores/mips16/MEM_stage.v"),
        getFile("/opencores/mips16/register_file.v"),
        getFile("/opencores/mips16/WB_stage.v"));
    putToMap(mips16, NO_MODULE, getFile("/opencores/mips16/mips_16_defs.v"));
    putToMap(mips16, getFile("/opencores/mips16/register_file.v"));
    putToMap(mips16, getFile("/opencores/mips16/WB_stage.v"));
    VERILOG_MAP.put("mips16", mips16);
  }

  /**
   * Returns a mapping of projects to their files.
   *
   * @return Projects from the test suite.
   */
  public static Map<String, Map<String, List<String>>> getProjects() {

    final Map<String, Map<String, List<String>>> map = new LinkedHashMap<>();
    map.putAll(VHDL_MAP);
    map.putAll(VERILOG_MAP);

    return map;
  }

  private static void putToMap(
      final Map<String, List<String>> map,
      final String topName,
      final File ... files) {

    final List<String> filePaths = new LinkedList<>();
    for (final File file : files) {
      filePaths.add(file.getAbsolutePath());
    }
    map.put(topName, filePaths);
  }

  private static void putToMap(final Map<String, List<String>> map, final File file) {
    putToMap(map, getTopLevelName(file.getAbsolutePath()), file);
  }

  private static String getTopLevelName(final String file) {

    return FilenameUtils.getBaseName(file);
  }

  private static File getFile(final String resourceName) {

    final URL resource = HdlTestFiles.class.getResource(resourceName);

    return new File(resource.getFile());
  }
}
