/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.ArrayUtils;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.parser.basis.HdlParser;
import ru.ispras.retrascope.parser.verilog.VerilogParser;
import ru.ispras.retrascope.util.LogMessage;

import java.io.File;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for testing at module-level.
 * Can be applied for HDL modules written in languages are supported by the project.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class HdlTestUtils {

  private HdlTestUtils() {
    /* Empty. */
  }

  /**
   * Runs the tool on the specified mapping of HDL files with the specified target and options.
   *
   * @param map Mapping of HDL top level module names to related files.
   * @param target Tool target entity.
   * @param options Additional command line options for the tool.
   *
   * @throws RetrascopeException when it encounters at the file with unknown extension.
   */
  public static void runHdl(
      final Map<String, List<String>> map,
      final String target,
      final String ... options) {

    for (final Map.Entry<String, List<String>> entry : map.entrySet()) {

      final Map<String, List<String>> oneElemMap = new LinkedHashMap<>(1);
      oneElemMap.put(entry.getKey(), entry.getValue());

      final String ext = FilenameUtils.getExtension(entry.getValue().get(0));

      if ("vhd".equalsIgnoreCase(ext)) {
        runVhdl(oneElemMap, target, options);
      } else if ("v".equalsIgnoreCase(ext)) {
        runVerilog(oneElemMap, target, options);
      } else {
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, ext);
      }
    }
  }

  /**
   * Runs the tool on all the HDL test projects (excluding files which path suffixes are in the
   * specified set) with the specified target and additional options.
   *
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param options Additional options for tool.
   */
  public static void runHdl(
      final String target,
      final Set<String> ignored,
      final String... options) {

    runVhdl(target, ignored, options);
    runVerilog(target, ignored, options);
  }

  /**
   * Runs tool on HDL modules with the specified target and options.
   *
   * @param target Tool target.
   * @param options Additional options for tool.
   */
  public static void runHdl(final String target, final String ... options) {

    runHdl(target, null, options);
  }

  /**
   * Runs tool on all the VHDL projects from the test suite with the specified options.
   * <p>Runs the tool on all the VHDL test projects (excluding files which path suffixes are in the
   * specified set) with the specified target and additional options.</p>
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param options Additional options for tool.
   */
  public static void runVhdl(
      final String target,
      final Set<String> ignored,
      final String ... options) {

    /* Run Retrascope on each module of each VHDL project. */

    final Collection<Map<String, List<String>>> vhdlFiles = HdlTestFiles.VHDL_MAP.values();

    for (final Map<String, List<String>> projectMap : vhdlFiles) {

      for (final Map.Entry<String, List<String>> entry : projectMap.entrySet()) {

        final String fullFileName = entry.getKey();

        if (ignored == null || !contain(ignored, entry.getKey())) {

          final Map<String, List<String>> onlyFile = new LinkedHashMap<>();
          onlyFile.put(fullFileName, entry.getValue());

          runVhdl(onlyFile, target, options);
        }
      }
    }
  }

  /**
   * Runs tool on VHDL modules with the specified target and options.
   *
   * @param target Tool target.
   * @param options Additional options for tool.
   */
  public static void runVhdl(final String target, final String ... options) {

    runVhdl(target, null, options);
  }

  /**
   * Runs Retrascope on the specified VHDL files with the specified target and additional options.
   *
   * @param map Mapping of VHDL top level module names to related files.
   * @param target Tool target.
   * @param options Additional options for tool.
   */
  public static void runVhdl(
      final Map<String, List<String>> map,
      final String target,
      final String... options) {

    final String[] fileNames = getFileNames(map.values());
    final String topLevelNames = getVhdlModuleNames(map.keySet());

    runRetrascope(fileNames, makeVhdlCmdLine(target, topLevelNames, options));
  }

  /**
   * Runs tool on all the Verilog projects from the test suite with the specified options.
   * <p>Runs the tool on all the Verilog test projects (excluding files which path suffixes
   * are in the specified set) with the specified target and additional options.</p>
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param options Additional options for tool.
   */
  public static void runVerilog(
      final String target,
      final Set<String> ignored,
      final String ... options) {

    /* Run Retrascope on each module of each Verilog project. */

    final Collection<Map<String, List<String>>> verilogFiles = HdlTestFiles.VERILOG_MAP.values();

    for (final Map<String, List<String>> projectMap : verilogFiles) {

      for (final Map.Entry<String, List<String>> entry : projectMap.entrySet()) {

        if (ignored == null || !contain(ignored, entry.getKey())) {

          final Map<String, List<String>> onlyFile = new LinkedHashMap<>();
          onlyFile.put(entry.getKey(), entry.getValue());

          runVerilog(onlyFile, target, options);
        }
      }
    }
  }

  /**
   * Runs tool on Verilog modules with the specified target.
   *
   * @param target Tool target.
   */
  public static void runVerilog(final String target) {

    runVerilog(target, null);
  }

  /**
   * Runs Retrascope on Verilog file with the specified options.
   *
   * @param prjMap Map of HDL top level module names to related files.
   * @param target Value of command line target option.
   * @param options Additional command line options for tool.
   */
  public static void runVerilog(
      final Map<String, List<String>> prjMap,
      final String target,
      final String... options) {

    final String[] fileNames = getFileNames(prjMap.values());
    final String includes = getIncludes(prjMap.values());
    final String moduleNames = getVerilogModuleNames(prjMap.keySet());

    runRetrascope(fileNames, makeVerilogCmdLine(target, includes, moduleNames, options));
  }

  private static String getIncludes(final Collection<List<String>> prjPaths) {

    final Set<String> includePaths = new LinkedHashSet<>();

    for (final List<String> modulePaths : prjPaths) {

      for (final String path : modulePaths) {
        includePaths.add(path.substring(0, path.lastIndexOf(File.separator)));
      }
    }

    return String.join(File.pathSeparator, includePaths);
  }

  /**
   * Checks whether the collection of paths contains the path ending.
   * @param paths The collection of file paths
   * @param pathEnd The path ending
   * @return {@code true} when the specified collection contains a path, that ends with the
   *          specified string, {@code false} otherwise.
   */
  public static boolean contain(final Set<String> paths, final String pathEnd) {

    for (final String substring : paths) {
      if (pathEnd.endsWith(substring)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Create module name parameter for command line.
   *
   * @param names VHDL modules top-level names.
   */
  private static String getVhdlModuleNames(final Collection<String> names) {

    final List<String> topLevelNames = new LinkedList<>();

    for (final String fileName : names) {
      topLevelNames.add(fileName.substring(fileName.lastIndexOf(File.separator) + 1));
    }

    return String.join(File.pathSeparator, topLevelNames);
  }

  private static String getVerilogModuleNames(final Collection<String> names) {

    return String.join(File.pathSeparator, names);
  }

  /**
   * Creates command line for VHDL input with the specified parameters.
   *
   * @param target Value of command line "target" option for VHDL parser.
   * @param moduleNames Value of command line "toplevel" option for VHDL parser.
   * @param options Additional options for VHDL parser.
   *
   * @return Command line for tool.
   */
  private static String[] makeVhdlCmdLine(
      final String target,
      final String moduleNames,
      final String ... options) {

    if (moduleNames.equals(HdlTestFiles.NO_MODULE)) {

      return (String[]) ArrayUtils.addAll(
          new String[] {
              paramName(Retrascope.TARGET_PARAM),
              target},
          options);
    }
    return (String[]) ArrayUtils.addAll(
        new String[] {
            paramName(Retrascope.TARGET_PARAM),
            target,
            paramName(HdlParser.MODULE_NAME),
            moduleNames},
        options);
  }

  private static String[] makeVerilogCmdLine(
      final String target,
      final String includes,
      final String moduleNames,
      final String ... options) {

    if (moduleNames.equals(HdlTestFiles.NO_MODULE)) {

      return (String[]) ArrayUtils.addAll(
          new String[] {
              paramName(Retrascope.TARGET_PARAM),
              target,
              paramName(VerilogParser.INCLUDE_PATH),
              includes},
          options);
    }
    return (String[]) ArrayUtils.addAll(
        new String[] {
            paramName(Retrascope.TARGET_PARAM),
            target,
            paramName(VerilogParser.INCLUDE_PATH),
            includes,
            paramName(HdlParser.MODULE_NAME),
            moduleNames},
      options);
  }

  /**
   * Runs Retrascope.
   *
   * @param fileNames Names of VHDL files for processing.
   * @param options Command line options for the tool.
   */
  private static void runRetrascope(final String[] fileNames, final String ... options) {

    Retrascope.main((String[]) ArrayUtils.addAll(fileNames, options));
  }

  /**
   * Creates an array of file names for the command line.
   *
   * @param files HDL files.
   */
  private static String[] getFileNames(final Collection<List<String>> files) {

    final List<String> names = new LinkedList<>();
    for (final List<String> list : files) {
      names.addAll(list);
    }

    return names.stream().map(String::new).toArray(String[]::new);
  }

  private static String paramName(final Parameter param) {

    return param.getName();
  }
}
