/*
 * Copyright 2016-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.engine.gadd.transformer.efsm.GaddEfsmTransformer;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.ModelCheckingMethod;
import ru.ispras.retrascope.engine.hldd.printer.smv.property.HlddPropertySmvPrinter;
import ru.ispras.retrascope.engine.smv.launcher.SmvModelCheckerLauncher;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents the base class for test cases which use the SMV model checker.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public abstract class SmvUtilTest {

  private static final List<String> EXCLUDED;
  private static final Map<String, ModelCheckingMethod> METHOD = new HashMap<>();
  private static final Map<String, Integer> EXPECTED_COUNTEREXAMPLE_COUNT = new HashMap<>();

  // BMC bound values for various designs.
  private static final Map<String, Integer> BOUND_VALUES = new HashMap<>();

  static {
    BOUND_VALUES.put("b05", 1);
    BOUND_VALUES.put("b11", 1);
    BOUND_VALUES.put("b13", 1);
    BOUND_VALUES.put("b14", 1);
    BOUND_VALUES.put("b14_1", 1);
    BOUND_VALUES.put("b20", 1);
    BOUND_VALUES.put("b20_1", 1);
    BOUND_VALUES.put("b21", 1);
    BOUND_VALUES.put("b21_1", 1);
    BOUND_VALUES.put("b22", 1);
    BOUND_VALUES.put("b22_1", 1);
    BOUND_VALUES.put("mem_ctrl", 1);
    BOUND_VALUES.put("mlite2sram", 1);
    BOUND_VALUES.put("pipeline", 1);
  }

  static {
    EXCLUDED = new ArrayList<>();
    // EXCLUDED.add(Paths.get("b03", "b03.vhd").toString());
    // EXCLUDED.add(Paths.get("b04", "b04.vhd").toString());
    // EXCLUDED.add(Paths.get("b05", "b05.vhd").toString());
    // EXCLUDED.add(Paths.get("b07", "b07.vhd").toString());
    // EXCLUDED.add(Paths.get("b08", "b08.vhd").toString());
    // EXCLUDED.add(Paths.get("b10", "b10.vhd").toString());
    // EXCLUDED.add(Paths.get("b11", "b11.vhd").toString());
    EXCLUDED.add(Paths.get("b12", "b12.vhd").toString()); // SMV: Type System Violation
    // EXCLUDED.add(Paths.get("b13", "b13.vhd").toString());
    // EXCLUDED.add(Paths.get("b14", "b14.vhd").toString());
    // EXCLUDED.add(Paths.get("b14_1", "b14_1.vhd").toString());
    EXCLUDED.add(Paths.get("b15", "b15.vhd").toString()); // Undefined variable in specification.

    /* Undefined variable in specification. */
    EXCLUDED.add(Paths.get("b15_1", "b15_1.vhd").toString());

    EXCLUDED.add(Paths.get("b17", "b17.vhd").toString()); // Array assignments.
    EXCLUDED.add(Paths.get("b17_1", "b17_1.vhd").toString()); // Array assignments.
    EXCLUDED.add(Paths.get("b18", "b18.vhd").toString()); // Array assignments.
    EXCLUDED.add(Paths.get("b18_1", "b18_1.vhd").toString()); // Array assignments.
    EXCLUDED.add(Paths.get("b19", "b19.vhd").toString()); // Array assignments.
    EXCLUDED.add(Paths.get("b19_1", "b19_1.vhd").toString()); // Array assignments.
    // EXCLUDED.add(Paths.get("b20", "b20.vhd").toString());
    // EXCLUDED.add(Paths.get("b20_1", "b20_1.vhd").toString());
    // EXCLUDED.add(Paths.get("b21", "b21.vhd").toString());
    // EXCLUDED.add(Paths.get("b21_1", "b21_1.vhd").toString());
    // EXCLUDED.add(Paths.get("b22", "b22.vhd").toString());
    // EXCLUDED.add(Paths.get("b22_1", "b22_1.vhd").toString());
    EXCLUDED.add(Paths.get("b30", "b30.vhd").toString()); // Multiple variable declarations.
    EXCLUDED.add(Paths.get("fifo_mod", "fifo.v").toString()); // Assignment type mismatch.
    EXCLUDED.add(Paths.get("plasma", "reg_bank.vhd").toString()); // Array assignments.
    //EXCLUDED.add(Paths.get("adder", "adder4.v").toString()); // Multiple variable assignments.
    //EXCLUDED.add(Paths.get("blocks", "blocks.v").toString()); // Multiple variable assignments.
    EXCLUDED.add(Paths.get("empty", "empty.v").toString()); // Empty file. No model is extracted.
    //EXCLUDED.add(Paths.get("fifo", "fifo.v").toString()); // Multiple variable assignments.

    /* Multiple variable assignments. */
    //EXCLUDED.add(Paths.get("fifo", "fifo_testbench.v").toString());
    //EXCLUDED.add(Paths.get("ram", "ram.v").toString());
    //EXCLUDED.add(Paths.get("ram", "ram_testbench.v").toString());

    /* Type mismatch. (32-bit const value) */
    EXCLUDED.add(Paths.get("mips16", "hazard_detection_unit.v").toString());

    /* Type mismatch. (see "decoding_op_src1 != 0" for example) */
    EXCLUDED.add(Paths.get("mips16", "IF_stage.v").toString());

    /* Contains only define directives. No model is extracted. */
    EXCLUDED.add(Paths.get("mips16", "mips_16_defs.v").toString());

    /* Case value type mismatch. */
    //EXCLUDED.add(Paths.get("mips16", "instruction_mem.v").toString());
    //EXCLUDED.add(Paths.get("mips16", "mips_16_core_top.v").toString());

    //EXCLUDED.add(Paths.get("for_loop", "for_loop.v").toString()); // Initial value type mismatch.
    //EXCLUDED.add(Paths.get("mips16", "data_mem.v").toString()); // Initial value type mismatch.

    /* Expressions not allowed in array subscripts on left hand side of assignments. */
    EXCLUDED.add(Paths.get("fifo0", "mem_2p.v").toString());
    EXCLUDED.add(Paths.get("fifo0", "myfifo.v").toString());

    /* Too big model. */
    EXCLUDED.add(Paths.get("fifo0", "mul_fifo.v").toString());
    EXCLUDED.add(Paths.get("mips16", "data_mem.v").toString());
    EXCLUDED.add(Paths.get("mips16", "MEM_stage.v").toString());



    METHOD.put("b01", ModelCheckingMethod.CTL_BDD);
    METHOD.put("b02", ModelCheckingMethod.CTL_BDD);
    //METHOD.put("b03", ModelCheckingMethod.CTL_BDD);
    METHOD.put("b04", ModelCheckingMethod.LTL_BMC);
    //METHOD.put("b06", ModelCheckingMethod.CTL_BDD);
    METHOD.put("b07", ModelCheckingMethod.LTL_BMC);
    METHOD.put("b08", ModelCheckingMethod.LTL_BMC);
    //METHOD.put("b09", ModelCheckingMethod.CTL_BDD);
    //METHOD.put("b10", ModelCheckingMethod.CTL_BDD);
    METHOD.put("b11", ModelCheckingMethod.LTL_MSAT);

   /* EXPECTED_COUNTEREXAMPLE_COUNT.put("b01", 29);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b02", 48);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b03", 41);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b04", 46);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b06", 75);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b07", 52);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b08", 63);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b09", 32);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b10", 95);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b11", 78);*/

  }

  public SmvUtilTest() {
  }

  @Test
  public void test() {
    runTest();
  }

  private void runTest() {
    final Map<String, Map<String, List<String>>> files = getDesigns();

    for (Entry<String, Map<String, List<String>>> project : files.entrySet()) {
      for (Entry<String, List<String>> file : project.getValue().entrySet()) {

        final String topName = file.getKey();
        final String folderName =
            ReadWriteUtils.getParentDirName(file.getValue().iterator().next());

        final Map<String, List<String>> onlyFile = new LinkedHashMap<>();
        onlyFile.put(topName, file.getValue());

        if (!isExcluded(folderName)) {

          SmvModelCheckerLauncher.setBmcBound(
              BOUND_VALUES.getOrDefault(topName, SmvModelCheckerLauncher.DEFAULT_BOUND));

          HdlTestUtils.runHdl(onlyFile, getMainTarget(), getParameters(topName));

          checkResults(topName);
        }
      }
    }
  }

  /**
   * The main target of the launch.
   *
   * @return The main target.
   */
  protected abstract String getMainTarget();

  /**
   * Additional parameters of the launch.
   *
   * @param designName - the specified design name.
   *
   * @return The array of parameters.
   */
  protected abstract String[] getParameters(final String designName);

  /**
   * HDL designs used by the test.
   *
   * @return The map of designs.
   */
  protected abstract Map<String, Map<String, List<String>>> getDesigns();

  /**
   * Indicates whether the specified design path is not supported by the checker.
   *
   * @param path - the specified design path (in the form {@code "folder/file"}).
   *
   * @return {@code True} if the design is NOT supported, {@code false} otherwise.
   */
  private boolean isExcluded(final String path) {
    return EXCLUDED.contains(path);
  }

  /**
   * Returns the preferred method of the specified design model checking.
   *
   * @param designName - the specified design name (filename without extension).
   *
   * @return {@code bdd} for symbolic model checking or {@code bmc} for bounded model checking.
   *     Bounded model checking is the default value.
   */
  private String getCheckMethod(final String designName) {
    return METHOD.containsKey(designName)
        ? METHOD.get(designName).getStringValue()
        : ModelCheckingMethod.LTL_BMC.getStringValue();
  }

  /**
   * Targets used by any SMV-based test.
   *
   * @param designName - the specified design name.
   *
   * @return The list of targets.
   */
  protected List<String> commonParameters(final String designName) {
    final List<String> targets = new ArrayList<>();
    final String smvFileName = designName + ".smv";
    targets.add(HlddPropertySmvPrinter.SMV_FILE_NAME.getName());
    targets.add(smvFileName);
    targets.add(HlddPropertySmvPrinter.CHECK_METHOD.getName());
    targets.add(getCheckMethod(designName));
    if (designName.equals("b07") || designName.equals("b11") || "b08".equals(designName)) {
      targets.add(GaddEfsmTransformer.STATE_VARIABLES.getName());
      targets.add("STATO");
    }
    return targets;
  }

  /*
   * Checks nuXmv run results.
   */
  private void checkResults(final String designName) {
    if (EXPECTED_COUNTEREXAMPLE_COUNT.containsKey(designName)) {
      final Engine launcher =
          Retrascope.getEngineRegistry().getEngine(SmvModelCheckerLauncher.ENGINE_ID);
      final int actual = ((SmvModelCheckerLauncher) launcher).getCounterexampleCount();
      final int expected = EXPECTED_COUNTEREXAMPLE_COUNT.get(designName);
      Assert.assertEquals(expected, actual);
    }
  }
}
