/*
 * Copyright 2016-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.generator.test.random;

import org.junit.Test;
import ru.ispras.retrascope.HdlTestFiles;
import ru.ispras.retrascope.HdlTestUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.result.test.TestModel;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Test for Random test generator based on Control Flow Graph Model.
 * The test runs test generator on every separate HDL file and the engine produces
 * {@link TestModel} representation of test.
 *
 * @see CfgRandomTestGenerator
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRandomTestGeneratorTestCase {

  static final int TEST_LENGTH = 10;
  private static final String ITC99 = "itc99";
  private static final String MINIMIPS = "minimips";
  private static final String PLASMA = "plasma";
  private static final String CLK = "CLK";
  private static final String RESET = "RESET";
  private static final String RESET_IN = "RESET_IN";
  private static final String CLOCK = "CLOCK";
  private static final String CLK_LOW = "clk";
  private static final String RST_LOW = "rst";
  private static final Map<String, List<ModuleDescriptor>> MAP = new LinkedHashMap<>();

  static class ModuleDescriptor {

    private final String filePath;
    private final String clkName;
    private final String rstName;
    private final String inputValues;

    ModuleDescriptor(final Path filePath, final String clkName, final String rstName) {

      this.filePath = filePath.toString();
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = "";
    }

    ModuleDescriptor(
        final Path filePath,
        final String clkName,
        final String rstName,
        final String inputValues) {

      this.filePath = filePath.toString();
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = inputValues;
    }

    String getFilePath() {

      return this.filePath;
    }

    String getClkName() {
      return this.clkName;
    }

    String getRstName() {
      return this.rstName;
    }

    String getInputValues() {
      return this.inputValues;
    }
  }

  private static void put(final String topName, final ModuleDescriptor descriptor) {

    if (MAP.containsKey(topName)) {

      MAP.get(topName).add(descriptor);
    } else {

      final List<ModuleDescriptor> list = new LinkedList<>();
      list.add(descriptor);
      MAP.put(topName, list);
    }
  }

  static {

    /* === VHDL test projects === */

    /* sample */
    put("bvsample", new ModuleDescriptor(Paths.get("bvsample", "bvsample.vhd"), CLK, ""));

    /* sample */
    put("sample", new ModuleDescriptor(Paths.get("sample", "sample.vhd"), CLK, "RST"));

    /* count */
    put("count", new ModuleDescriptor(Paths.get("count", "count.vhd"), CLK, RESET));

    /* example */
    put("stub", new ModuleDescriptor(Paths.get("example", "stub.vhd"), CLK, RESET));
    put("driver", new ModuleDescriptor(Paths.get("example", "driver.vhd"), CLK, RESET));
    put("example", new ModuleDescriptor(Paths.get("example", "example.vhd"), CLK, RESET));

    /* ITC'99  benchmark */
    put("b01", new ModuleDescriptor(Paths.get(ITC99, "b01", "b01.vhd"), CLOCK, RESET));
    put("b02", new ModuleDescriptor(Paths.get(ITC99, "b02", "b02.vhd"), CLOCK, RESET));
    put("b03", new ModuleDescriptor(Paths.get(ITC99, "b03", "b03.vhd"), CLOCK, RESET));
    put("b04",
        new ModuleDescriptor(
            Paths.get(ITC99, "b04", "b04.vhd"), CLOCK, RESET, "DATA_IN=[-128,127,10]"));
    put("b05", new ModuleDescriptor(Paths.get(ITC99, "b05", "b05.vhd"), CLOCK, RESET));
    put("b06", new ModuleDescriptor(Paths.get(ITC99, "b06", "b06.vhd"), CLOCK, RESET));
    put("b07", new ModuleDescriptor(Paths.get(ITC99, "b07", "b07.vhd"), CLOCK, RESET));
    put("b08", new ModuleDescriptor(Paths.get(ITC99, "b08", "b08.vhd"), CLOCK, RESET));
    put("b09", new ModuleDescriptor(Paths.get(ITC99, "b09", "b09.vhd"), CLOCK, RESET));
    put("b10", new ModuleDescriptor(Paths.get(ITC99, "b10", "b10.vhd"), CLOCK, RESET));
    put("b11",
        new ModuleDescriptor(Paths.get(ITC99, "b11", "b11.vhd"), CLOCK, RESET, "X_IN=[0,63,10]"));
    put("b12", new ModuleDescriptor(Paths.get(ITC99, "b12", "b12.vhd"), CLOCK, RESET));
    put("b13", new ModuleDescriptor(Paths.get(ITC99, "b13", "b13.vhd"), CLOCK, RESET));
    put("b14", new ModuleDescriptor(Paths.get(ITC99, "b14", "b14.vhd"), CLOCK, RESET));
    put("b14_1", new ModuleDescriptor(Paths.get(ITC99, "b14_1", "b14_1.vhd"), CLOCK, RESET));
    put("b15", new ModuleDescriptor(Paths.get(ITC99, "b15", "b15.vhd"), CLOCK, RESET));
    put("b15_1", new ModuleDescriptor(Paths.get(ITC99, "b15_1", "b15_1.vhd"), CLOCK, RESET));
    put("b17", new ModuleDescriptor(Paths.get(ITC99, "b17", "b17.vhd"), CLOCK, RESET));
    put("b17_1", new ModuleDescriptor(Paths.get(ITC99, "b17_1", "b17_1.vhd"), CLOCK, RESET));
    put("b18", new ModuleDescriptor(Paths.get(ITC99, "b18", "b18.vhd"), CLOCK, RESET));
    put("b18_1", new ModuleDescriptor(Paths.get(ITC99, "b18_1", "b18_1.vhd"), CLOCK, RESET));
    put("b19", new ModuleDescriptor(Paths.get(ITC99, "b19", "b19.vhd"), CLOCK, RESET));
    put("b19_1", new ModuleDescriptor(Paths.get(ITC99, "b19_1", "b19_1.vhd"), CLOCK, RESET));
    put("b20", new ModuleDescriptor(Paths.get(ITC99, "b20", "b20.vhd"), CLOCK, RESET));
    put("b20_1", new ModuleDescriptor(Paths.get(ITC99, "b20_1", "b20_1.vhd"), CLOCK, RESET));
    put("b21", new ModuleDescriptor(Paths.get(ITC99, "b21", "b21.vhd"), CLOCK, RESET));
    put("b21_1", new ModuleDescriptor(Paths.get(ITC99, "b21_1", "b21_1.vhd"), CLOCK, RESET));
    put("b22", new ModuleDescriptor(Paths.get(ITC99, "b22", "b22.vhd"), CLOCK, RESET));
    put("b22_1", new ModuleDescriptor(Paths.get(ITC99, "b22_1", "b22_1.vhd"), CLOCK, RESET));

    /* MiniMIPS */
    put("alu", new ModuleDescriptor(Paths.get(MINIMIPS, "alu.vhd"), CLOCK, RESET));
    put("banc", new ModuleDescriptor(Paths.get(MINIMIPS, "banc.vhd"), CLOCK, RESET));
    put("bus_ctrl", new ModuleDescriptor(Paths.get(MINIMIPS, "bus_ctrl.vhd"), CLOCK, RESET));
    put("minimips", new ModuleDescriptor(Paths.get(MINIMIPS, "minimips.vhd"), CLOCK, RESET));
    put("pps_di", new ModuleDescriptor(Paths.get(MINIMIPS, "pps_di.vhd"), CLOCK, RESET));
    put("pps_ei", new ModuleDescriptor(Paths.get(MINIMIPS, "pps_ei.vhd"), CLOCK, RESET));
    put("pps_ex", new ModuleDescriptor(Paths.get(MINIMIPS, "pps_ex.vhd"), CLOCK, RESET));
    put("pps_mem", new ModuleDescriptor(Paths.get(MINIMIPS, "pps_mem.vhd"), CLOCK, RESET));
    put("pps_pf", new ModuleDescriptor(Paths.get(MINIMIPS, "pps_pf.vhd"), CLOCK, RESET));
    put("ram", new ModuleDescriptor(Paths.get(MINIMIPS, "ram.vhd"), "", ""));
    put("renvoi", new ModuleDescriptor(Paths.get(MINIMIPS, "renvoi.vhd"), "", ""));
    put("rom", new ModuleDescriptor(Paths.get(MINIMIPS, "rom.vhd"), "", ""));
    put("syscop", new ModuleDescriptor(Paths.get(MINIMIPS, "syscop.vhd"), CLOCK, RESET));

    /* Plasma */
    put("alu", new ModuleDescriptor(Paths.get(PLASMA, "alu.vhd"), "", ""));
    put("bus_mux", new ModuleDescriptor(Paths.get(PLASMA, "bus_mux.vhd"), "", ""));
    put("control", new ModuleDescriptor(Paths.get(PLASMA, "control.vhd"), "", ""));
    put("mem_ctrl", new ModuleDescriptor(Paths.get(PLASMA, "mem_ctrl.vhd"), CLK, RESET_IN));
    put("mlite2sram", new ModuleDescriptor(Paths.get(PLASMA, "mlite2sram.vhd"), CLK, ""));
    put("mlite2uart", new ModuleDescriptor(Paths.get(PLASMA, "mlite2uart.vhd"), CLK, RESET));
    put("mlite_cpu", new ModuleDescriptor(Paths.get(PLASMA, "mlite_cpu.vhd"), CLK, RESET_IN));
    put("mult", new ModuleDescriptor(Paths.get(PLASMA, "mult.vhd"), CLK, ""));
    put("pc_next", new ModuleDescriptor(Paths.get(PLASMA, "pc_next.vhd"), CLK, RESET_IN));
    put("pipeline", new ModuleDescriptor(Paths.get(PLASMA, "pipeline.vhd"), CLK, RESET));
    put("plasma", new ModuleDescriptor(Paths.get(PLASMA, "plasma.vhd"), "BOARD_CLK", "BOARD_RES"));
    put("ram", new ModuleDescriptor(Paths.get(PLASMA, "ram.vhd"), CLK, ""));
    put("reg_bank", new ModuleDescriptor(Paths.get(PLASMA, "reg_bank.vhd"), CLK, RESET_IN));
    put("shifter", new ModuleDescriptor(Paths.get(PLASMA, "shifter.vhd"), "", ""));
    put("sram2mlite", new ModuleDescriptor(Paths.get(PLASMA, "sram2mlite.vhd"), CLK, ""));
    put("uart", new ModuleDescriptor(Paths.get(PLASMA, "uart.vhd"), CLK, RESET));

    /* === Verilog test projects === */

    /* adder */
    put("adder4", new ModuleDescriptor(Paths.get("adder", "adder4.v"), "", ""));
    put("adder4_testbench", new ModuleDescriptor(Paths.get("adder", "adder4_testbench.v"), "", ""));

    /* empty */
    put("", new ModuleDescriptor(Paths.get("empty", "empty.v"), "", ""));

    /* fifo */
    put("fifo", new ModuleDescriptor(Paths.get("fifo", "fifo.v"), CLK_LOW, RST_LOW));
    put("fifo_testbench",
        new ModuleDescriptor(Paths.get("fifo", "fifo_testbench.v"), CLK_LOW, RST_LOW));

    /* fifo_mod */
    put("fifo", new ModuleDescriptor(Paths.get("fifo_mod", "fifo.v"), CLK_LOW, RST_LOW));

    /* for_loop */
    put("for_loop", new ModuleDescriptor(Paths.get("for_loop", "for_loop.v"), CLK_LOW, ""));

    /* incompat_cond */
    put("incompat_cond",
        new ModuleDescriptor(Paths.get("incompat_cond", "incompat_cond.v"), CLK_LOW, ""));

    /* fifo0 */
    put("mem_2p", new ModuleDescriptor(Paths.get("fifo0", "mem_2p.v"), "rd_clk", ""));
    put("mul_fifo", new ModuleDescriptor(Paths.get("fifo0", "mul_fifo.v"), CLK_LOW, RST_LOW));
    put("myfifo", new ModuleDescriptor(Paths.get("fifo0", "myfifo.v"), CLK_LOW, RST_LOW));

    /* mux */
    put("mux4_1", new ModuleDescriptor(Paths.get("mux", "mux4_1.v"), "", ""));
    put("mux_1_testbench", new ModuleDescriptor(Paths.get("mux", "mux_1_testbench.v"), "", ""));

    /* ram */
    put("ram", new ModuleDescriptor(Paths.get("ram", "ram.v"), CLK_LOW, RST_LOW));
    put("ram_testbench",
        new ModuleDescriptor(Paths.get("ram", "ram_testbench.v"), CLK_LOW, RST_LOW));
  }

  @Test
  public void runTest() {
    final Map<String, Map<String, List<String>>> hdlProjects = HdlTestFiles.getProjects();

    for (final Map.Entry<String, Map<String, List<String>>> prjEntry : hdlProjects.entrySet()) {

      for (final Map.Entry<String, List<String>> entry : prjEntry.getValue().entrySet()) {

        final String topName = entry.getKey();
        final List<String> files = entry.getValue();
        final Map<String, List<String>> oneFile = new LinkedHashMap<>();
        oneFile.put(topName, files);

        final ModuleDescriptor descriptor = getDescriptor(topName, files);

        HdlTestUtils.runHdl(
            oneFile,
            TestModel.TYPE.getId(), Retrascope.ENGINE_PARAM.getName(),
            CfgRandomTestGenerator.ENGINE_ID,
            CfgRandomTestGenerator.TEST_LEN.getName(),
            String.valueOf(TEST_LENGTH),
            CfgRandomTestGenerator.CLK_NAME.getName(),
            descriptor.getClkName(),
            CfgRandomTestGenerator.RST_NAME.getName(),
            descriptor.getRstName(),
            CfgRandomTestGenerator.INPUT_VALUES.getName(),
            descriptor.getInputValues());
      }
    }
  }

  ModuleDescriptor getDescriptor(final String topName, final List<String> filePaths) {

    if (MAP.containsKey(topName)) {

      final List<ModuleDescriptor> descriptors = MAP.get(topName);

      for (final ModuleDescriptor descriptor : descriptors) {

        final String topPath = descriptor.getFilePath();
        if (filePaths.contains(topPath)) {
          return descriptor;
        }
      }

    }
    return new ModuleDescriptor(Paths.get(""), "", "");
  }
}
