/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import org.junit.Assert;
import org.junit.Test;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.SolverUtilTest;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.efsm.examples.B04;
import ru.ispras.retrascope.model.efsm.examples.SimpleEfsm;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * This class represents the EFSM production method test.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class ExampleEfsmProductTestCase extends SolverUtilTest {

  private static final String STATES = "States: ";
  private static final String TRANSITIONS = "Transitions: ";

  @Test
  public void testProduct() {
    final Efsm b04 = getEfsm(B04.EFSM_MODEL);
    final Efsm simpleEfsm = getEfsm(SimpleEfsm.EFSM_MODEL);

    printEfsmInfo(b04);
    printEfsmInfo(simpleEfsm);

    makeProduct("product1", b04, simpleEfsm);
    makeProduct("product2", simpleEfsm, b04);
    makeProduct("product3", simpleEfsm, simpleEfsm, simpleEfsm);
  }

  private Efsm getEfsm(final EfsmModel model) {
    final EfsmModule module = (EfsmModule) model.getModules().iterator().next();
    return module.getNonInitialEfsms().iterator().next();
  }

  private void makeProduct(final String name, final Efsm... efsms) {
    final EfsmModule productModule = new EfsmModule(name);
    for (final Efsm efsm : efsms) {
      declareVariables(efsm.getParentContainer(), productModule);
    }
    final Efsm product = EfsmUtils.product(productModule, efsms);

    Assert.assertNotNull(product);
    Assert.assertEquals(product.getStateNum(), getStateProduct(efsms));

    printEfsmInfo(product);
  }

  private void printEfsmInfo(final Efsm efsm) {
    Log.info(efsm.getName());
    Log.info(STATES + efsm.getStateNum());
    Log.info(efsm.getStates().toString());
    Log.info(TRANSITIONS + efsm.getTransitionNum());
    for (final EfsmTransition item : efsm.getTransitions()) {
      Log.info(item.toString());
    }
    Log.info("Initial state: " + efsm.getInitialState().toString());
    Log.info("");
  }

  private int getStateProduct(final Efsm... efsms) {
    int product = efsms[0].getStateNum();
    for (int i = 1; i < efsms.length; i++) {
      product = product * efsms[i].getStateNum();
    }
    return product;
  }

  private void declareVariables(final EfsmModule srcModule, final EfsmModule dstModule) {
    for (Map.Entry<NodeVariable, Declaration> entry : srcModule.getDeclarationMap().entrySet()) {
      if (!dstModule.declares(entry.getKey())) {
        dstModule.declareVariable(entry.getKey(), entry.getValue());
      }
    }
  }
}
