/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Guard;

/**
 * Class providing utility methods for manual construction of example EFSMs.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class ExampleEfsmUtils {

  /**
   * A guard whose condition is always true.
   */
  public static final Guard ALWAYS_TRUE_GUARD = new Guard(Nodes.TRUE);
  private static final String UNSUPPORTED_HDL_TYPE = "Unsupported HDL type: ";

  /**
   * Constructs an input variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareInput(
      final DecStatement variableContainer,
      final String name,
      final String hdlType) {

    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.IN, hdlType);
  }

  /**
   * Constructs an output variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareOutput(
      final DecStatement variableContainer,
      final String name,
      final String hdlType) {

    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.OUT, hdlType);
  }

  /**
   * Constructs a register variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareRegister(
      final DecStatement variableContainer,
      final String name,
      final String hdlType) {
    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.REG, hdlType);
  }

  /**
   * Constructs a variable with the specified parameters and declares it in the specified variable
   * container. The variable has no condition on value which is initially set to zero.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param dataType a data type of the variable
   * @param variableType a type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareVariable(
      final DecStatement variableContainer,
      final String name,
      final DataType dataType,
      final VariableType variableType) {

    final NodeVariable variable = new NodeVariable(name, dataType);
    variable.setUserData(new VariableDescriptor(variableType));

    variableContainer.declareVariable(variable, new NodeValue(dataType.valueOf("0", 10)), null);

    return variable;
  }

  /**
   * Constructs a variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param container the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param variableType a type of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  private static NodeVariable declareVariable(
      final DecStatement container,
      final String name,
      final VariableType variableType,
      final String hdlType) {

    final VariableDescriptor variableDescriptor = new VariableDescriptor(variableType);

    final String[] splittedHdlType = hdlType.replace('(', ' ').replace(')', ' ').split(" +");

    DataType dataType;
    String decInitValue;

    String minValue = null;
    String maxValue = null;

    switch (splittedHdlType[0].toLowerCase()) {

      case "bit": {
        dataType = DataType.bitVector(1);
        decInitValue = "0";

        break;
      }

      case "bit_vector": {

        if (splittedHdlType.length == 4) {

          final int firstIndex = Integer.parseInt(splittedHdlType[1]);
          final int secondIndex = Integer.parseInt(splittedHdlType[3]);

          switch (splittedHdlType[2]) {
            case "to": {
              dataType = DataType.bitVector(secondIndex - firstIndex);

              break;
            }

            case "downto": {
              dataType = DataType.bitVector(firstIndex - secondIndex);

              break;
            }

            default: {
              throw new IllegalArgumentException(UNSUPPORTED_HDL_TYPE + hdlType);
            }
          }
        } else {
          throw new IllegalArgumentException(UNSUPPORTED_HDL_TYPE + hdlType);
        }

        decInitValue = "0";

        break;
      }

      case "integer": {
        dataType = DataType.INTEGER;

        switch (splittedHdlType.length) {
          case 1: {
            minValue = "-2147483648";
            maxValue = "2147483647";

            decInitValue = minValue;

            break;
          }

          case 5: {
            switch (splittedHdlType[3].toLowerCase()) {
              case "to": {
                minValue = splittedHdlType[2];
                maxValue = splittedHdlType[4];

                break;
              }

              case "downto": {
                minValue = splittedHdlType[4];
                maxValue = splittedHdlType[2];

                break;
              }

              default: {
                throw new IllegalArgumentException(UNSUPPORTED_HDL_TYPE + hdlType);
              }
            }

            decInitValue = splittedHdlType[2];

            break;
          }

          default: {
            throw new IllegalArgumentException(UNSUPPORTED_HDL_TYPE + hdlType);
          }
        }

        break;
      }

      default: {
        throw new IllegalArgumentException(UNSUPPORTED_HDL_TYPE + splittedHdlType[0]);
      }
    }

    variableDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, hdlType);

    final NodeVariable variable = new NodeVariable(name, dataType);
    variable.setUserData(variableDescriptor);

    NodeValue high = null;
    NodeValue low = null;

    if (minValue != null) {
      low = new NodeValue(dataType.valueOf(minValue, 10));
    }

    if (maxValue != null) {
      high = new NodeValue(dataType.valueOf(maxValue, 10));
    }

    if (high == null || low == null) {
      container.declareVariable(variable,
          new NodeValue(dataType.valueOf(decInitValue, 10)));
    } else {
      container.declareVariable(variable,
          new NodeValue(dataType.valueOf(decInitValue, 10)), new Range(low, high));
    }

    return variable;
  }

  /**
   * Constructs a value of the same type as the specified variable has.
   *
   * @param variable the variable which value should be constructed for
   * @param decimalValue a string representation of the value
   *
   * @return the constructed value
   */
  static NodeValue constructValue(final NodeVariable variable, final String decimalValue) {

    return new NodeValue(variable.getDataType().valueOf(decimalValue, 10));
  }

  /**
   * Constructs an expression of the form "variable == value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Node constructEquality(final NodeVariable variable, final String decimalValue) {

    return Nodes.eq(variable, ExampleEfsmUtils.constructValue(variable, decimalValue));
  }

  /**
   * Constructs an expression of the form "variable != value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Node constructInequality(final NodeVariable variable, final String decimalValue) {

    return ExprUtils.getNegation(constructEquality(variable, decimalValue));
  }

  /**
   * Constructs an assignment of the form "variable := value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Assignment constructAssignment(final NodeVariable variable, final String decimalValue) {

    return new Assignment(variable, ExampleEfsmUtils.constructValue(variable, decimalValue));
  }

  private ExampleEfsmUtils() {
  }
}
