/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Arrays;
import java.util.Collections;

/**
 * The test class which contains manually created simple EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class SimpleEfsm {
  /**
   * An EFSM model which contains a simple EFSM module.
   */
  public static final EfsmModel EFSM_MODEL = new EfsmModel();

  static {
    final EfsmModule efsmModule = new EfsmModule(SimpleEfsm.class.getSimpleName());

    final NodeVariable clockVariable = ExampleEfsmUtils.declareVariable(efsmModule, "clock",
        DataType.bitVector(1), VariableType.IN);

    final NodeVariable i0Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i0",
        DataType.bitVector(4), VariableType.IN);
    final NodeVariable i1Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i1",
        DataType.bitVector(1), VariableType.IN);
    final NodeVariable i2Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i2",
        DataType.bitVector(4), VariableType.IN);

    final NodeVariable varX = ExampleEfsmUtils.declareVariable(efsmModule, "x",
        DataType.bitVector(4), VariableType.REG);
    final NodeVariable varY = ExampleEfsmUtils.declareVariable(efsmModule, "y",
        DataType.bitVector(4), VariableType.REG);
    final NodeVariable varZ = ExampleEfsmUtils.declareVariable(efsmModule, "z",
        DataType.bitVector(4), VariableType.REG);

    final NodeVariable stateVariable = ExampleEfsmUtils.declareVariable(efsmModule, "state",
        DataType.bitVector(4), VariableType.REG);

    final String state0Value = "0";
    final String state1Value = "1";
    final String state2Value = "2";
    final String state3Value = "3";
    final String state4Value = "4";

    final Efsm efsm = new Efsm(efsmModule, "EFSM1");

    final EfsmState state0 =
        efsm.addState(ExampleEfsmUtils.constructEquality(stateVariable, state0Value));
    final EfsmState state1 =
        efsm.addState(ExampleEfsmUtils.constructEquality(stateVariable, state1Value));
    final EfsmState state2 =
        efsm.addState(ExampleEfsmUtils.constructEquality(stateVariable, state2Value));
    final EfsmState state3 =
        efsm.addState(ExampleEfsmUtils.constructEquality(stateVariable, state3Value));
    final EfsmState state4 =
        efsm.addState(ExampleEfsmUtils.constructEquality(stateVariable, state4Value));

    efsm.setInitialState(state0);

    final Event clockEvent = new Event(clockVariable, EventType.POSITIVE_EDGE);

    {
      final Action action01 = new Action(Arrays.asList(new Assignment(varZ, i0Variable),
          ExampleEfsmUtils.constructAssignment(stateVariable, state1Value)));

      final GuardedAction guardedAction01 =
          new GuardedAction(ExampleEfsmUtils.ALWAYS_TRUE_GUARD, action01);
      guardedAction01.addClock(clockEvent);

      efsm.addTransition(state0, state1, guardedAction01);
    }

    {
      final Action action12 = new Action(Arrays.asList(new Assignment(varX, varZ),
          ExampleEfsmUtils.constructAssignment(stateVariable, state2Value)));

      final GuardedAction guardedAction12 =
          new GuardedAction(ExampleEfsmUtils.constructEquality(i1Variable, "1"), action12);
      guardedAction12.addClock(clockEvent);

      efsm.addTransition(state1, state2, guardedAction12);
    }

    {
      final Action action13 = new Action(Arrays.asList(new Assignment(varX, varZ),
          ExampleEfsmUtils.constructAssignment(stateVariable, state3Value)));

      final GuardedAction guardedAction13 =
          new GuardedAction(ExampleEfsmUtils.constructInequality(i1Variable, "1"), action13);
      guardedAction13.addClock(clockEvent);

      efsm.addTransition(state1, state3, guardedAction13);
    }

    {
      final Action action23 = new Action(Arrays.asList(new Assignment(varY, i2Variable),
          ExampleEfsmUtils.constructAssignment(stateVariable, state3Value)));

      final GuardedAction guardedAction23 =
          new GuardedAction(ExampleEfsmUtils.ALWAYS_TRUE_GUARD, action23);
      guardedAction23.addClock(clockEvent);

      efsm.addTransition(state2, state3, guardedAction23);
    }

    {
      final Node guardPredicate34 =
          Nodes.and(ExampleEfsmUtils.constructEquality(varX, "4"),
          ExampleEfsmUtils.constructEquality(varY, "2"));
      final Action action34 =
          new Action(ExampleEfsmUtils.constructAssignment(stateVariable, state4Value));

      final GuardedAction guardedAction34 = new GuardedAction(guardPredicate34, action34);
      guardedAction34.addClock(clockEvent);

      efsm.addTransition(state3, state4, guardedAction34);
    }

    final GuardedAction rstGuardedAction =
        new GuardedAction(
            ExampleEfsmUtils.ALWAYS_TRUE_GUARD,
            new Action(Collections.emptyList()));
    rstGuardedAction.addClock(clockEvent);
    efsm.addResetGuardedAction(rstGuardedAction);

    efsmModule.addNonInitialEfsm(efsm);
    SimpleEfsm.EFSM_MODEL.addModule(efsmModule);
  }

  private SimpleEfsm() {
    /* Empty. */
  }
}
