/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Level;
import org.junit.Assert;
import org.junit.Test;
import org.zamia.BuildPath;
import org.zamia.ERManager;
import org.zamia.SourceFile;
import org.zamia.ZamiaException;
import org.zamia.ZamiaLogger;
import org.zamia.ZamiaProject;
import org.zamia.ZamiaProjectBuilder;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGModule;
import org.zamia.zdb.ZDBException;
import ru.ispras.retrascope.HdlTestFiles;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * This test runs ZamiaCAD tool on every HDL file from project test suite.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class IgElaborationTestCase {

  private String zamiaDataBasePath;

  @Test
  public void runTest() {
    elaborateAll();
  }

  private void elaborateAll() {

    final Collection<Map<String, List<String>>> vhdlProjects =
        HdlTestFiles.VHDL_MAP.values();

    for (final Map<String, List<String>> vhdlProject : vhdlProjects) {

      for (final Map.Entry<String, List<String>> entry : vhdlProject.entrySet()) {

        final ZamiaProject zamiaProject = getProject(entry.getKey(), entry.getValue());
        final List<IGModule> igModules = getIgModules(zamiaProject);

        if (igModules.isEmpty()) {
          Log.warn("No modules found");
        }
        zamiaProject.shutdown();
        zamiaProject.getZDB().clear();
        try {
          FileUtils.forceDeleteOnExit(new File(this.zamiaDataBasePath + File.separator + "db"));
        } catch (final IOException e) {
          e.printStackTrace();
          Assert.fail(e.getMessage());
        }
      }
    }
  }

  private ZamiaProject getProject(final String topName, final List<String> files) {

    ZamiaLogger.setup(Level.OFF);
    ZamiaProject zamiaProject = null;

    /*
    * Create BuildPath.txt
    */

    /* Get full path to VHDL source file directory. */
    final String sourceDirPath = getSourceDirPath(files.get(0));

    final File buildFile = new File(sourceDirPath + File.separator + "BuildPath.txt");

    try {
      if (!buildFile.exists() && !buildFile.createNewFile()) {
        throw new RetrascopeException("Can't create " + buildFile.toPath().toString());
      }

      final BufferedWriter writer = ReadWriteUtils.newBufferedWriter(buildFile);

      writer.flush();
      writer.write(String.format("local WORK \"%s\"\n", sourceDirPath));
      writer.write(String.format("toplevel %s\n", topName.toUpperCase(Locale.US)));
      writer.close();

      /* Create Zamia IG. */
      this.zamiaDataBasePath = sourceDirPath;
      zamiaProject = new ZamiaProject("", sourceDirPath, new SourceFile(buildFile), sourceDirPath);
    } catch (IOException | ZamiaException | ZDBException e) {
      Log.error(e.getMessage());
    }

    return zamiaProject;
  }

  private static String getSourceDirPath(final String relativeFilePath) {

    String sourceDirPath;

    final int separatorIndex = relativeFilePath.lastIndexOf(File.separatorChar);

    if (separatorIndex >= 0) {

      sourceDirPath = relativeFilePath.substring(0, separatorIndex);
    } else {

      /* The source file is in current dir. */
      sourceDirPath = System.getProperty("user.dir");
    }
    return sourceDirPath;
  }

  private List<IGModule> getIgModules(final ZamiaProject project) {

    try {
      project.clean();

      final ZamiaProjectBuilder builder = project.getBuilder();
      ZamiaLogger.getInstance().getLogger().setLevel(Level.OFF);

      builder.build(true, true, null);
    } catch (IOException | ZamiaException e) {
      throw new RetrascopeException(e);
    }

    final ERManager erm = project.getERM();
    final int errNum = erm.getNumErrors();

    if (errNum > 0) {
      for (int i = 0; i < errNum; i++) {
        try {
          throw erm.getError(i);
        } catch (ZamiaException e) {
          Log.error(e.getMessage());
        }
      }
    }

    final BuildPath buildPath = project.getBuildPath();
    final int topLevelNum = buildPath.getNumToplevels();
    final List<IGModule> result = new LinkedList<>();

    if (topLevelNum > 0) {

      final IGManager igm = project.getIGM();
      for (int i = 0; i < topLevelNum; i++) {
        result.add(igm.findModule(buildPath.getToplevel(i)));
      }
    }
    return result;
  }
}
