/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.SolverUtilTest;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * The test for {@link ModelTransformUtils} class.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ModelTransformUtilsTestCase extends SolverUtilTest {

  private static final NodeValue ZERO = NodeValue.newInteger(0);
  private static final NodeValue ONE = NodeValue.newInteger(1);
  private static final NodeValue TWO = NodeValue.newInteger(2);
  private static final NodeValue THREE = NodeValue.newInteger(3);

  @Test
  public void runTest() {

    runSubstituteBackwardTest();
  }

  /* The method for {@code ModelTransformUtils.transformBackward} method testing. */
  private void runSubstituteBackwardTest() {

    final NodeVariable nodeVariableX = NodeVariable.newInteger("x");
    final NodeVariable nodeVariableY = NodeVariable.newInteger("y");
    final NodeVariable nodeVariableZ = NodeVariable.newBitVector("z", 3);
    final NodeVariable nodeVariableV = NodeVariable.newMap("v", DataType.INTEGER, DataType.INTEGER);
    final List<CfgBlockStatement> blocks = new LinkedList<>();

    runSbdSimpleTest(nodeVariableX, nodeVariableY, blocks);
    runSbNoSsaTest(nodeVariableX, nodeVariableY, blocks);
    runSbBitVectorTest(nodeVariableZ, blocks);
    runSbArrayTest(nodeVariableX, nodeVariableV, blocks);
  }

  private void runSbdSimpleTest(
      final NodeVariable nodeVariableX,
      final NodeVariable nodeVariableY,
      final List<CfgBlockStatement> blocks) {

    /*
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      x := 1;
     *        |
     *      y := 1;
     *        |
     * switch(x + y <= 2)
     *
     * The result should be: true
     *
     */
    blocks.clear();
    blocks.add(new CfgBlockStatement(nodeVariableX, ONE));
    blocks.add(new CfgBlockStatement(nodeVariableY, ONE));

    final NodeOperation switchExpr = Nodes.lesseq(Nodes.add(nodeVariableX, nodeVariableY), TWO);

    final NodeValue correctResult = (NodeValue) Nodes.TRUE;

    final Node substituteResult = ModelTransformUtils.substituteBackward(blocks, switchExpr);

    if (!substituteResult.equals(correctResult)) {
      Assert.fail();
    }
  }

  private void runSbNoSsaTest(
      final NodeVariable nodeVariableX,
      final NodeVariable nodeVariableY,
      final List<CfgBlockStatement> blocks) {

    /*
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      y := 1;
     *        |
     *      x := y;
     *        |
     *      y := 2;
     *        |
     * switch(x + y <= 3)
     *
     * The result should be: 1 + 2 <= 3
     *
     */

    blocks.clear();
    blocks.add(new CfgBlockStatement(nodeVariableY, ONE));
    blocks.add(new CfgBlockStatement(nodeVariableX, nodeVariableY));
    blocks.add(new CfgBlockStatement(nodeVariableY, TWO));

    final NodeOperation switchExpr = Nodes.lesseq(Nodes.add(nodeVariableX, nodeVariableY), THREE);

    final NodeValue correctResult = (NodeValue) Nodes.TRUE;

    final Node substituteResult = ModelTransformUtils.substituteBackward(blocks, switchExpr);

    if (!substituteResult.equals(correctResult)) {
      Assert.fail();
    }
  }

  private void runSbBitVectorTest(
      final NodeVariable nodeVariableZ,
      final List<CfgBlockStatement> blocks) {

    /*
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      z[0] := '1';
     *        |
     *      z[1] := '1';
     *        |
     *      z[2] := '0';
     *        |
     * switch((z - '011') == '000')
     *
     * The result should be: true
     *
     */

    blocks.clear();
    blocks.add(newBlock(nodeVariableZ, ZERO, NodeValue.newBitVector(1, 1)));
    blocks.add(newBlock(nodeVariableZ, ONE, NodeValue.newBitVector(1, 1)));
    blocks.add(newBlock(nodeVariableZ, TWO, NodeValue.newBitVector(0, 1)));

    final NodeValue constVector = NodeValue.newBitVector("011", 2, 3);
    final Node switchExpr =
        Nodes.eq(Nodes.bvsub(nodeVariableZ, constVector), NodeValue.newBitVector(0, 3));

    final Node transformedExpr = ModelTransformUtils.substituteBackward(blocks, switchExpr);

    final Node resultNode =
        ExprUtils.getConjunction(
            transformedExpr,
            Nodes.eq(nodeVariableZ, constVector));

    /* Check whether the transformation result is satisfiable. */
    if (!SolverUtils.isSat(resultNode)) {
      Assert.fail();
    }
  }

  private static CfgBlockStatement newBlock(
      final NodeVariable var,
      final NodeValue index,
      final  NodeValue value) {

    return new CfgBlockStatement(new RangedVariable(var, new Range(index, index)), value);
  }

  private void runSbArrayTest(
      final NodeVariable nodeVariableX,
      final NodeVariable nodeVariableV,
      final List<CfgBlockStatement> blocks) {

    /*
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      x := 0;
     *        |
     *      v[x] := 2;
     *        |
     *      v[1] := 3;
     *        |
     * switch(x[0] == 2)
     *
     * The result should be: true
     *
     */

    blocks.clear();
    blocks.add(
        new CfgBlockStatement(nodeVariableX, ZERO));
    blocks.add(
        new CfgBlockStatement(
            new RangedVariable(nodeVariableV, Collections.singletonList(nodeVariableX)), TWO));
    blocks.add(
        new CfgBlockStatement(
            new RangedVariable(nodeVariableV, Collections.singletonList(ONE)), THREE));

    final Node switchExpr = Nodes.eq(Nodes.select(nodeVariableV, ZERO), TWO);

    final Node transformedExpr = ModelTransformUtils.substituteBackward(blocks, switchExpr);

    if (!SolverUtils.isSat(transformedExpr)) {
      Assert.fail();
    }
  }
}
