/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.SolverUtilTest;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

/**
 * Test for {@link NodeUtils} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class NodeUtilTestCase extends SolverUtilTest {

  private static final Node INT_ONE = NodeValue.newInteger(1);
  private static final Node INT_SEVEN = NodeValue.newInteger(7);
  private static final Node BV_SEVEN = NodeValue.newBitVector("111", 2, 3);
  private static final Node BV_ONE = NodeValue.newBitVector(BitVector.TRUE);

  private static final Node I_VAR = NodeVariable.newInteger("i");
  private static final Node J_VAR = NodeVariable.newInteger("j");
  private static final Node K_VAR = NodeVariable.newInteger("k");

  @Test
  public void runTest() {

    runCastMethodTest();
    runSplitByVarTest();
  }

  private void runSplitByVarTest() {

    final Node iNode = Nodes.eq(I_VAR, INT_ONE);
    final Node jNode = Nodes.greater(J_VAR, INT_ONE);
    final Node kNode = Nodes.less(K_VAR, INT_ONE);
    final Node ijNode = Nodes.eq(Nodes.add(I_VAR, J_VAR), INT_SEVEN);

    final Collection<Collection<Node>> splitByVarNodes =
        NodeUtils.splitByVar(iNode, jNode, kNode, ijNode);

    final Collection<Collection<Node>> result = new HashSet<>();
    final Collection<Node> ijNodes = new HashSet<>();
    Collections.addAll(ijNodes, iNode, jNode, ijNode);
    result.add(ijNodes);
    final Collection<Node> kNodes = new HashSet<>();
    kNodes.add(kNode);
    result.add(kNodes);

    Assert.assertEquals(splitByVarNodes, result);
  }

  private void runCastMethodTest() {

    /* integer -> boolean */
    Assert.assertTrue(
        isSatEquation(Nodes.TRUE, NodeUtils.castToDataType(INT_SEVEN, DataType.BOOLEAN)));

    /* bool -> integer */
    Assert.assertTrue(
        isSatEquation(INT_ONE, NodeUtils.castToDataType(Nodes.TRUE, DataType.INTEGER)));

    /* bit vector -> boolean */
    Assert.assertTrue(
        isSatEquation(Nodes.TRUE, NodeUtils.castToDataType(BV_SEVEN, DataType.BOOLEAN)));

    /* boolean -> bit vector */
    Assert.assertTrue(
        isSatEquation(BV_ONE, NodeUtils.castToDataType(Nodes.TRUE, DataType.bitVector(1))));

    /* integer -> bit vector */
    Assert.assertTrue(
        isSatEquation(BV_SEVEN, NodeUtils.castToDataType(INT_SEVEN, DataType.bitVector(3))));

    /* bit vector -> integer */
    Assert.assertTrue(
        isSatEquation(INT_SEVEN, NodeUtils.castToDataType(BV_SEVEN, DataType.INTEGER)));
  }

  private boolean isSatEquation(final Node first, final Node second) {
    return SolverUtils.isSat(Nodes.eq(first, second));
  }
}
