/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.SolverUtilTest;

/**
 * Test for {@link SolverUtils} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class SolverUtilTestCase extends SolverUtilTest {

  @Test
  public void runTest() {

    runIsSatTest();
  }

  private void runIsSatTest() {

    /* Test constraint is: true == true.*/
    final Node satNode = Nodes.eq(Nodes.TRUE, Nodes.TRUE);

    checkExpression(satNode);

    /* Test constraint is: true == false. */
    final Node unSatNode = Nodes.eq(Nodes.TRUE, Nodes.FALSE);

    checkExpression(unSatNode);

    /* Test constraint is: (var1 == 1) AND (var1 != 1). */
    final NodeVariable var1 = NodeVariable.newInteger("var1");
    final NodeValue one = NodeValue.newInteger(1);
    final Node unSatIntegerNode = Nodes.and(Nodes.eq(var1, one), Nodes.noteq(var1, one));
    checkExpression(unSatIntegerNode);

    /* Test constraint is: (var1 == 1) AND ((var2 == 0) OR (var2 != 0)). */
    final NodeVariable var2 = NodeVariable.newInteger("var2");
    final NodeValue zero = NodeValue.newInteger(0);
    final Node satIntegerNode =
        Nodes.and(Nodes.eq(var1, one), Nodes.or(Nodes.eq(var2, zero), Nodes.noteq(var2, zero)));
    checkExpression(satIntegerNode);

    /* Test constraint is: (0 >= 0) AND (0 <= 2) AND (0 >= -128) AND (0 <= 127)
     * AND (0 >= -128) AND (0 <= 127) AND (0 == 0) */
    final NodeValue two = NodeValue.newInteger(2);
    final NodeValue max = NodeValue.newInteger(127);
    final NodeValue min = NodeValue.newInteger(-128);
    final Node constantNode =
        ExprUtils.getConjunction(
            Nodes.and(Nodes.greatereq(zero, zero), Nodes.lesseq(zero, two)),
            Nodes.and(Nodes.greatereq(zero, min), Nodes.lesseq(zero, max)),
            Nodes.and(Nodes.greatereq(zero, min), Nodes.lesseq(zero, max)),
            Nodes.eq(zero, zero));
    checkExpression(constantNode);

    /* Test constraint is: (var2 <= 2) AND (var1 > 0) AND (var2 > 0) */
    final Node positiveNode =
        ExprUtils.getConjunction(
            Nodes.lesseq(var2, two),
            Nodes.greater(var1, zero),
            Nodes.greater(var2, zero));
    checkExpression(positiveNode);

    /* Test constraint is: (bool1 AND bool2) AND (bool3 -> bool2) AND (NOT(bool1 AND bool2)) */
    final NodeVariable bool1 = NodeVariable.newBoolean("bool1");
    final NodeVariable bool2 = NodeVariable.newBoolean("bool2");
    final NodeVariable bool3 = NodeVariable.newBoolean("bool3");
    final Node forthOrderNode =
        ExprUtils.getConjunction(
            Nodes.and(bool1, bool2),
            Nodes.impl(bool3, bool2),
            Nodes.not(Nodes.and(bool1, bool2)));
    checkExpression(forthOrderNode);

    /* Test constraint is: (NOT(bool1 AND bool2)) AND (bool3 -> bool2) AND (bool1 AND bool2) */
    final Node backOrderNode =
        ExprUtils.getConjunction(
            Nodes.not(Nodes.and(bool1, bool2)),
            Nodes.impl(bool3, bool2),
            Nodes.and(bool1, bool2));
    checkExpression(backOrderNode);
  }

  private void checkExpression(final Node satNode) {

    if (differentSolverResults(satNode)) {
      Assert.fail();
    }
  }

  private boolean differentSolverResults(final Node satNode) {

    return SolverUtils.isSat(satNode) != ExprUtils.isSat(satNode);
  }
}
