/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.basis;

import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * The after-engine component.
 *
 * <p>The {@link ru.ispras.retrascope.basis.Engine} engine component
 * that should be executed after engine's main method.
 * The backend elaborates {@link Entity} entity that engine returns.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class Backend {

  private final String name;

  private boolean active;

  private File debugFile;

  /**
   * Constructs backend with the specified name.
   * @param name The backend name.
   */
  protected Backend(final String name) {

    this.name = name;
    this.active = true;
  }

  /**
   * Returns the backend name.
   * @return The string identifier for the backend.
   */
  public String getName() {

    return this.name;
  }

  /**
   * Checks whether the backend is active.
   * @return {@code true} when the backend is active, {@code false} otherwise.
   */
  public boolean isActive() {

    return this.active;
  }

  /**
   * Makes the backend active or not depending on the flag value.
   * @param active The activity flag.
   */
  public void setActive(final boolean active) {
    this.active = active;
  }

  /**
   * Returns the backend's debug file.
   *
   * @return The backend's debug file.
   */
  public File getDebugFile() {
    return this.debugFile;
  }

  /**
   * Initializes the backend's debug file.
   *
   * @param file The backend's debug file.
   */
  public void setDebugFile(final File file) {
    this.debugFile = file;
  }

  /**
   * Writes the line to the backend's debug file.
   * @param line The line to be written.
   */
  public void debugWrite(final String line) {

    if (getDebugFile() == null) {
      return;
    }

    try {

      final FileOutputStream stream = new FileOutputStream(getDebugFile(), true);
      final BufferedWriter writer = ReadWriteUtils.newBufferedWriter(stream);
      writer.write(line);
      writer.newLine();
      writer.close();

    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Elaborates the specified entity.
   *
   * @param entity Entity to be elaborated.
   */
  public abstract void start(Entity entity);
}
