/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

/**
 * This is a base class of an engine (tool) that transforms one {@link Entity} entity to another.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class Engine {

  /**
   * Parameter that disables all the engine backends.
   */
  public static final Parameter NO_BACKEND =
      new Parameter("no-backends", false, "Disables HDL parser backends");

  public static final Parameter DISABLE_BACKENDS =
      new Parameter(
          "disable-backends",
          true,
          "Disables HDL parser backends with the specified names");

  /* Debug file default name. */
  private final String debugFileName;

  /// The engine identifier.
  private final String engineId;

  /// The input types of the engine.
  private final Set<EntityType> inputTypes = new HashSet<>();
  /// The output type of the engine.
  private EntityType outputType;

  // The parameters of the engine.
  private final Parameters parameters = new Parameters();

  // The engine event handler.
  private EngineEventHandler eventHandler;

  // The engine back-ends
  private final List<Backend> backEnds = new LinkedList<>();

  /* Debug mode flag. */
  private boolean debugMode;

  /* File for debug output. */
  private File debugFile;

  /**
   * Constructs an engine.
   *
   * @param engineId the engine identifier.
   */
  protected Engine(final String engineId) {
    this.engineId = engineId;
    this.debugFileName = engineId + ".log";
    addParameter(NO_BACKEND);
    addParameter(DISABLE_BACKENDS);
  }

  /**
   * Constructs an engine.
   *
   * @param engineId the engine identifier.
   * @param inputType the input type of the engine.
   * @param outputType the output type of the engine.
   */
  protected Engine(final String engineId, final EntityType inputType, final EntityType outputType) {
    this(engineId);

    this.inputTypes.add(inputType);
    this.outputType = outputType;
  }

  private EngineEventHandler getEventHandler() {
    return this.eventHandler;
  }

  private void setEventHandler(final EngineEventHandler eventHandler) {
    this.eventHandler = eventHandler;
  }

  /**
   * Returns the list of back-ends.
   * @return The list of back-ends are registered for this object.
   */
  protected List<Backend> getBackends() {
    return Collections.unmodifiableList(this.backEnds);
  }

  /**
   * Returns the unique engine identifier.
   *
   * @return the unique engine identifier.
   */
  public final String getId() {
    return engineId;
  }

  /**
   * Returns the input types of the engine (i.e., the types of input entities).
   *
   * @return the input types of the engine.
   */
  public final Set<EntityType> getInputTypes() {
    return inputTypes;
  }

  /**
   * Adds the given input type of the engine.
   *
   * @param inputType the input type to be added.
   */
  public final void addInputType(final EntityType inputType) {
    this.inputTypes.add(inputType);
  }

  /**
   * Checks whether the engine uses an entity of the given type.
   *
   * @param inputType the input type to be checked.
   * @return {@code true} if it uses; {@code false} otherwise.
   */
  final boolean uses(final EntityType inputType) {
    return this.inputTypes.contains(inputType);
  }

  /**
   * Returns the output type of the engine (i.e., the type of output entities).
   *
   * @return the output type of the engine.
   */
  public final EntityType getOutputType() {
    return outputType;
  }

  /**
   * Sets the given output type of the engine.
   *
   * @param outputType the output type to be set.
   */
  public final void setOutputType(final EntityType outputType) {
    this.outputType = outputType;
  }

  /**
   * Checks whether the engine defines an entity of the given type.
   *
   * @param outputType the output type to be checked.
   * @return {@code true} if it defines; {@code false} otherwise.
   */
  public final boolean defines(final EntityType outputType) {
    return this.outputType.equals(outputType);
  }

  /**
   * Returns the parameters of the engine.
   *
   * @return the engine parameters.
   */
  public final Parameters getParameters() {
    return parameters;
  }

  /**
   * Adds the given parameter to the engine.
   *
   * @param parameter the parameter to be added.
   */
  protected final void addParameter(final Parameter parameter) {
    parameters.addParameter(parameter);
  }

  /**
   * Parses the command line.
   *
   * @param args the command line.
   * @return the arguments
   * @throws ParseException the parse exception.
   */
  protected final Arguments parseCommandLine(final String... args) throws ParseException {
    return parameters.parseCommandLine(args);
  }

  /**
   * Returns the logger of the engine.
   *
   * @return the logger of the engine.
   */
  protected Logger getLogger() {
    return Log.getLogger(this.getClass());
  }

  /**
   * Checks whether debug mode for the engine is enabled.
   *
   * @return {@code} true when debug mode is enabled, {@code false} otherwise.
   */
  protected boolean isDebugMode() {
    return this.debugMode;
  }

  /**
   * Enables\disables engine's debug mode.
   *
   * @param debugMode {@code true} to enable debug mode, {@code false} otherwise.
   */
  protected void setDebugMode(final boolean debugMode) {
    this.debugMode = debugMode;
  }

  /**
   * Returns the engine's debug file.
   *
   * @return The engine's debug file.
   */
  private File getDebugFile() {
    return this.debugFile;
  }

  /**
   * Initializes the engine's debug file.
   *
   * @param file The engine's debug file.
   */
  private void setDebugFile(final File file) {
    this.debugFile = file;

    if (hasBackEnds()) {
      for (final Backend backend : getBackends()) {
        backend.setDebugFile(file);
      }
    }
  }

  /**
   * Writes the line to the engine's debug file.
   * @param line The line to be written.
   */
  protected void debugWrite(final String line) {

    final File file = getDebugFile();
    if (file == null) {
      return;
    }

    try {

      final FileOutputStream stream = new FileOutputStream(getDebugFile(), true);
      final BufferedWriter writer = ReadWriteUtils.newBufferedWriter(stream);
      writer.write(line);
      writer.newLine();
      writer.close();

    } catch (final IOException e) {
      throw new RetrascopeException(e);
    }
  }

  /**
   * Writes the line to the engine's debug file.
   * <p>Prints the line of the specified format that represents the described data objects.</p>
   * @param format The line format.
   * @param objects The objects to be printed.
   */
  protected void debugWrite(final String format, final Object ... objects) {
    debugWrite(String.format(format, objects));
  }

  /**
   * Processes entities of the input types and produces an entity of the output type.
   *
   * @param inputs the entities to be processed.
   * @return the result of the processing.
   */
  public abstract Entity start(final Map<EntityType, Entity> inputs);

  /**
   * Processes entities of the input types and produces an entity of the output type.
   * While calling this method the engine back-ends (if exist) will be also executed.
   *
   * @param inputs the entities to be processed.
   * @param eventHandler the engine event handler.
   * @return the result of the processing.
   */
  public final Entity start(
      final Map<EntityType, Entity> inputs, final EngineEventHandler eventHandler) {

    setEventHandler(eventHandler);

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    setDebugOptions(arguments);
    setBackends(arguments);

    // Run the main method & back-ends if they are exist.
    final Entity result = start(inputs);
    if (hasBackEnds()) {
      runBackEnds(result);
    }

    return result;
  }

  private void setBackends(final Arguments args) {

    if (args == null) {
      return;
    }
    try {

      final Arguments engineArgs = parseCommandLine(args.getCommandLine());

      final String disabledBackends = engineArgs.getValue(DISABLE_BACKENDS);

      if (disabledBackends != null) {

        final String[] disabled = disabledBackends.split(File.pathSeparator);

        if (Arrays.asList(disabled).contains("help")) {

          final StringBuilder builder = new StringBuilder();

          for (final Backend backend : getBackends()) {
            builder.append(String.format("\n\t%s", backend.getName()));
          }

          Log.info("The list of available HDL parser backends: %s", builder.toString());
        }

        deactivateBackends(disabled);
      } else if (engineArgs.contains(NO_BACKEND)) {
        disableAllBackends();
      } else {
        enableAllBackends();
      }

    } catch (final ParseException e) {
      e.printStackTrace();
    }
  }

  private void setDebugOptions(final Arguments args) {

    if (args != null && args.contains(Retrascope.TOOL_DEBUG_PARAM)) {

      setDebugMode(true);
      final String dbgFileName = args.getValue(Retrascope.TOOL_DEBUG_PARAM, debugFileName);

      final File dbgFile = new File(dbgFileName);
      setDebugFile(dbgFile);
      ReadWriteUtils.clear(dbgFile);
    } else {
      setDebugMode(false);
      setDebugFile(null);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Event Handlers
  //------------------------------------------------------------------------------------------------

  /**
   * Sets the engine progress.
   *
   * @param progress the engine progress (from 0 to 100).
   */
  final void setProgress(final int progress) {
    if (getEventHandler() != null) {
      eventHandler.onSetProgress(this, progress);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Back-ends Methods
  //------------------------------------------------------------------------------------------------

  /**
   * Returns {@code true} if engine has back-ends, {@code false} otherwise.
   * @return rather engine has back-ends or not
   */
  private boolean hasBackEnds() {
    return !getBackends().isEmpty();
  }

  /**
   * Adds back-end.
   *
   * @param backend parser back-end
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  protected void addBackend(final Backend backend) {
    InvariantChecks.checkNotNull(backend);
    this.backEnds.add(backend);
  }

  /**
   * Enables all the engine back ends.
   */
  protected void enableAllBackends() {

    for (final Backend backend : getBackends()) {
      backend.setActive(true);
    }
  }

  /**
   * Disables all the engine back ends.
   */
  protected void disableAllBackends() {

    for (final Backend backend : getBackends()) {
      backend.setActive(false);
    }
  }

  /**
   * Deactivates back-ends with the specified names.
   * @param names The names for registered back-ends.
   */
  protected void deactivateBackends(final String ... names) {
    for (final Backend backend : getBackends()) {

      if (Arrays.asList(names).contains(backend.getName())) {
        backend.setActive(false);
      }
    }
  }

  /**
   * Runs all the back-ends that were added to this engine.
   *
   * @param entity Entity to be processed by back-ends
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private void runBackEnds(final Entity entity) {
    InvariantChecks.checkNotNull(entity);

    for (final Backend backend : getBackends()) {
      if (backend.isActive()) {
        backend.start(entity);
      }
    }
  }

  //------------------------------------------------------------------------------------------------
  // Auxiliary Methods
  //------------------------------------------------------------------------------------------------

  protected static <T extends Entity> T getOptionalInputEntity(
      final Map<EntityType, Entity> inputs,
      final Class<T> entityClass,
      final EntityType entityType) {

    final Entity entity = inputs.get(entityType);

    if (entity == null) {
      return null;
    } else {
      InvariantChecks.checkTrue(entityClass.isInstance(entity));

      return entityClass.cast(entity);
    }
  }

  protected static <T extends Entity> T getInputEntity(
      final Map<EntityType, Entity> inputs,
      final Class<T> entityClass,
      final EntityType entityType) {

    final Entity entity = inputs.get(entityType);

    InvariantChecks.checkNotNull(entity, "Null entity for type: " + entityType.toString());
    InvariantChecks.checkTrue(entityClass.isInstance(entity));

    return entityClass.cast(entity);
  }

  @Override
  public String toString() {
    return getId();
  }

  @Override
  public boolean equals(final Object obj) {
    if (!(obj instanceof Engine)) {
      return false;
    }

    final Engine rhs = (Engine) obj;

    return getId().equals(rhs.getId());
  }

  @Override
  public int hashCode() {
    return getId().hashCode();
  }
}
