/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.util.Log;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Meta-information storage.
 *
 * <p>This object stores some meta-information. Meta-information is presented in pairs
 * like ({@link MetaInfoType} key, {@link MetaInfoValue} value).</p>
 *
 * @see MetaInfoType
 * @see MetaInfoValue
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class MetaInfo {

  private final Map<String, Object> metaInfoMap;

  /**
   * Default constructor.
   */
  protected MetaInfo() {
    this.metaInfoMap = new LinkedHashMap<>();
  }

  /**
   * Constructs an object with the specified data.
   * @param metaMap The meta-information data map.
   */
  protected MetaInfo(final Map<String, Object> metaMap) {
    this.metaInfoMap = new LinkedHashMap<>();
    this.metaInfoMap.putAll(metaMap);
  }

  /**
   * Checks whether object contains meta-information for the specified user-defined type.
   *
   * @param key The meta-information key.
   *
   * @return {@code true} when the object contains meta-information for the specified key,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean hasMetaInfo(final String key) {
    InvariantChecks.checkNotNull(key);

    return this.metaInfoMap.containsKey(key);
  }

  /**
   * Checks whether object contains meta-information for the specified enum key.
   *
   * @param key The meta-information key.
   *
   * @return {@code true} when the object contains meta-information for the specified key,
   *         {@code false} otherwise.
   */
  public boolean hasMetaInfo(final Enum<?> key) {
    return hasMetaInfo(key.name());
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * <p>It is supposed that this method is used for the predefined types and values.</p>
   *
   * @param type Meta-information type.
   * @param value Meta-information value.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final Enum<?> type, final Enum<?> value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value, "Null value for type: " + type);

    this.metaInfoMap.put(type.name(), value.name());
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * <p>It is supposed that this method is used for the predefined types.</p>
   *
   * @param type Meta-information type.
   * @param value Meta-information value.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final Enum<?> type, final Object value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value, "Null value for type: " + type);

    this.metaInfoMap.put(type.name(), value);
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * @param type Meta-information type.
   * @param value Meta-information value.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final String type, final Object value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value, "Null value for type: " + type);

    this.metaInfoMap.put(type, value);
  }

  /**
   * Add all the meta-information that is stored in the specified mapping.
   *
   * @param map The meta-information that is to be added into this object's storage.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addMetaInfo(final Map<String, Object> map) {
    InvariantChecks.checkNotNull(map);

    this.metaInfoMap.putAll(map);
  }

  /**
   * Returns meta-information of the specified type.
   *
   * @param key Meta-information key.
   *
   * @return Information if an object contains it, {@code null} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Object getMetaInfo(final String key) {
    InvariantChecks.checkNotNull(key);

    return this.metaInfoMap.get(key);
  }

  /**
   * Returns meta-information for the specified key.
   *
   * @param key The meta-information key.
   *
   * @return The meta-information string for the specified key.
   */
  public Object getMetaInfo(final Enum<?> key) {
    return getMetaInfo(key.name());
  }

  /**
   * Returns the string representation of the data is stored with the specified key.
   *
   * @param key The data key.
   *
   * @return The string representation of the data is stored with the specified key.
   */
  public String getStringMetaInfo(final String key) {

    if (!hasMetaInfo(key)) {
      Log.error("No meta-information for key: %s", key);
    }
    return getMetaInfo(key).toString();
  }

  /**
   * Returns the string representation of the data is stored with the specified key.
   *
   * @param key The data key.
   *
   * @return The string representation of the data is stored with the specified key.
   */
  public String getStringMetaInfo(final Enum<?> key) {

    if (!hasMetaInfo(key)) {
      Log.error("No meta-information for key: %s", key.name());
    }

    return getMetaInfo(key).toString();
  }

  /**
   * Returns all the meta-information that is stored in this object.
   *
   * @return The meta-information that is stored in this object.
   */
  public Map<String, Object> getAllMetaInfo() {
    return Collections.unmodifiableMap(this.metaInfoMap);
  }
}
