/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import org.apache.commons.cli.Option;

/**
 * This class represents a specification of a command line parameter.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Parameter {

  // The option in the Apache Commons CLI format.
  private final Option option;
  // The predefined parameter values (if applicable).
  private String[] values;

  /**
   * Constructs a parameter.
   *
   * @param longName the multi-character name of the parameter.
   * @param hasArgument the flag signaling whether an argument is required after the parameter.
   * @param description the self-documenting description of the parameter.
   */
  public Parameter(final String longName, final boolean hasArgument,
      final String description) {
    option = new Option(null, longName, hasArgument, description);
  }

  /**
   * Constructs a parameter.
   *
   * @param longName the multi-character name of the parameter.
   * @param values the predefined values of the parameter.
   * @param description the self-documenting description of the parameter.
   */
  public Parameter(final String longName, final String[] values,
      final String description) {
    this(longName, true, description);
    this.values = values;
  }

  /**
   * Returns the long name of the parameter.
   *
   * @return the long name.
   */
  public String getLongName() {
    return getOption().getLongOpt();
  }

  /**
   * Returns terminal name for parameter.
   *
   * @return The name to be printed in terminal.
   */
  public String getName() {

    return String.format("--%s", getLongName());
  }

  /**
   * Checks whether an argument is required after the parameter.
   *
   * @return {@code true} if an argument is required; {@code false} otherwise.
   */
  public boolean hasArgument() {
    return getOption().hasArg();
  }

  /**
   * Checks whether the parameter has predefined values only.
   *
   * @return {@code true} if predefined values are used; {@code false} otherwise.
   */
  public boolean usesPredefinedValues() {
    return getValues() != null && getValues().length != 0;
  }

  /**
   * Returns possible values of the parameter.
   *
   * @return the array of possible values.
   */
  private String[] getValues() {
    return values;
  }

  /**
   * Returns the description of the parameter.
   *
   * @return the description.
   */
  public String getDescription() {
    return getOption().getDescription();
  }

  /**
   * Returns the parameter in the Apache Commons CLI format.
   *
   * @return the option.
   */
  public Option getOption() {
    return option;
  }
}
