/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.model.basis.memory.HashMapMemory;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.TestCoverage;
import ru.ispras.retrascope.result.test.TestVector;

/**
 * AN interface for components that simulate models.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public interface Simulated {

  /**
   * Simulates processing of the specified input vector and returns produced output vector.
   *
   * @param inputVector the vector to be processed
   *
   * @return the output vector
   *
   * @throws IllegalArgumentException if {@code inputVector} is {@code null}
   */
  TestVector simulateVector(final TestVector inputVector);

  /**
   * Simulates processing of the specified input vector and returns produced output vector.
   *
   * @param inputSequence the vector to be processed
   *
   * @return the output vector
   *
   * @throws IllegalArgumentException if {@code inputSequence} is {@code null}
   */
  Sequence simulateSequence(final Sequence inputSequence);

  /**
   * Returns a test coverage achieved by this simulator since the last reset (if any).
   *
   * @return the test coverage
   */
  TestCoverage getCoverage();

  /**
   * Resets a test coverage achieved by this simulator.
   */
  void resetCoverage();

  /**
   * Returns a state of an underlying model which this simulator is currently in.
   *
   * @return the current state of this simulator
   */
  State getCurrentState();

  /**
   * Returns input variables and their values.
   * @return The mapping from input variables to their values at the present moment.
   */
  HashMapMemory getInputsMemory();

  /**
   * Returns internal variables and their values.
   * @return The mapping from internal variables to their values at the present moment.
   */
  HashMapMemory getRegMemory();

  /**
   * Returns output variables and their values.
   * @return The mapping from output variables to their values at the present moment.
   */
  HashMapMemory getOutputsMemory();
}
