/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.memory.HashMapMemory;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The class representing an abstract simulator.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public abstract class Simulator<T extends DecStatement> implements Simulated {

  protected final T simulatedModel;

  private HashMapMemory inputsMemory;
  protected HashMapMemory regMemory;
  protected HashMapMemory outputsMemory;

  /**
   * Constructs a simulator for the specified model.
   *
   * @param simulatedModel the simulated model
   */
  protected Simulator(final T simulatedModel) {
    this.simulatedModel = simulatedModel;

    this.inputsMemory = new HashMapMemory(simulatedModel.getInputNames());
    this.regMemory = new HashMapMemory(simulatedModel.getRegisterNames());
    this.outputsMemory = new HashMapMemory(simulatedModel.getOutputNames());
  }

  /**
   * Returns a model which is simulated by this simulator.
   *
   * @return the simulated model
   */
  public abstract T getSimulatedModel();

  public HashMapMemory getInputsMemory() {
    return this.inputsMemory;
  }

  public HashMapMemory getRegMemory() {
    return this.regMemory;
  }

  public HashMapMemory getOutputsMemory() {
    return this.outputsMemory;
  }

  /**
   * Resets this simulator, that is, returns it at the same state as it had immediately after
   * construction. Default implementation only sets variables of this simulator to their initial
   * values.
   */
  protected void reset() {
    this.resetInputs();

    for (final String register : this.simulatedModel.getRegisterNames()) {

      final NodeValue initValue = getInitValue(register);

      if (initValue != null) {
        this.regMemory.setValue(register, initValue);
      }
    }

    for (final String output : this.simulatedModel.getOutputNames()) {

      final NodeValue initValue = getInitValue(output);

      if (initValue != null) {
        this.outputsMemory.setValue(output, initValue);
      }
    }
  }

  private NodeValue getInitValue(final String varName) {
    final Node initValue = this.simulatedModel.getInitialValue(varName);

    if (initValue == null) {
      return null;
    }

    if (initValue.getKind() != Node.Kind.VALUE) {
      Log.warn("Wrong init value for '%s': %s", varName, initValue);
      return null;
    }
    return (NodeValue) initValue;
  }

  /**
   * Sets all inputs of this simulator to their initial values.
   */
  private void resetInputs() {
    for (final String input : this.simulatedModel.getInputNames()) {

      final NodeValue initValue = getInitValue(input);

      if (initValue != null) {
        this.inputsMemory.setValue(input, initValue);
      }
    }
  }

  /**
   * Substitute variables in the expression by ones that are in mapping.
   * @param expression The expression that should be transformed.
   * @param bufferedVariables The "variable's name - node" mapping.
   * @return The expression that is built from the specified one by substitution of variables
   *         on their names to related by mapping nodes.
   */
  public Node substituteVariables(
      final Node expression,
      final Map<String, NodeValue> bufferedVariables) {

    final Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    final List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

    for (final NodeVariable usedVariable : usedVariables) {
      final String variableName = usedVariable.getName();
      NodeValue variableValue = null;

      if (bufferedVariables.containsKey(variableName)) {
        variableValue = bufferedVariables.get(variableName);
      } else {
        if (this.hasVariable(variableName)) {
          variableValue = this.getVariableValue(variableName);
        }
      }

      if (variableValue != null) {
        bindingList.add(NodeBinding.bindVariable(usedVariable, variableValue));
      }
    }

    return Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));
  }

  /**
   * Reads the specified values and puts them to internal memory.
   * @param vector The set of values to be read.
   */
  public void readVector(final TestVector vector) {
    InvariantChecks.checkNotNull(vector);

    for (final String inputName : this.simulatedModel.getInputNames()) {
      this.inputsMemory.setValue(inputName, (NodeValue) vector.getValue(inputName));
    }

  }

  /**
   * Returns output vector that is produced by this object.
   *
   * @return output vector
   */
  public TestVector getOutputVector() {
    final Transaction transaction = new Transaction();

    for (final String variableName: this.simulatedModel.getOutputNames()) {
      transaction.addVariable(variableName, this.outputsMemory.getValue(variableName));
    }

    return new TestVector(transaction);
  }

  /**
   * Checks rather this operates with input variable that has the specified name.
   *
   * @param variableName Input variable's name.
   *
   * @return {@code true} when it operates with input variable
   *         that has the specified name, {@code false} otherwise.
   */
  private boolean hasInput(final String variableName) {
    return this.inputsMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this operates with inner variable that has the specified name.
   *
   * @param variableName Inner variable's name.
   *
   * @return {@code true} when it operates with inner variable
   *         that has the specified name, {@code false} otherwise.
   */
  private boolean hasInnerVariable(final String variableName) {
    return this.regMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this operates with output variable that has the specified name.
   *
   * @param variableName Output variable's name.
   *
   * @return {@code true} when it operates with output variable
   *         that has the specified name, {@code false} otherwise.
   */
  private boolean hasOutput(final String variableName) {
    return this.outputsMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this object operates with variable of the specified name.
   *
   * @param name Variable name.
   * @return {@code true} when this object operates with variable that has the specified name,
   *         {@code false} otherwise.
   */
  private boolean hasVariable(final String name) {
    return hasInput(name) || hasInnerVariable(name) || hasOutput(name);
  }

  /**
   * Returns value of variable with the specified name.
   *
   * @param variableName variable name
   *
   * @return variable value, if this operates with it,
   *         {@code null} otherwise
   */
  public NodeValue getVariableValue(final String variableName) {
    if (this.hasInput(variableName)) {
      return this.inputsMemory.getValue(variableName);
    } else {
      if (this.hasInnerVariable(variableName)) {
        return this.regMemory.getValue(variableName);
      } else {
        if (this.hasOutput(variableName)) {
          return this.outputsMemory.getValue(variableName);
        } else {
          return null;
        }
      }
    }
  }

  /**
   * Returns a snapshot representing a current state of the simulation.
   *
   * @return the snapshot
   */
  public SimulationSnapshot makeSimulationSnapshot() {
    return new SimulationSnapshot(this);
  }

  /**
   * Restores a simulation state from the specified snapshot.
   *
   * @param snapshot the snapshot
   */
  protected void loadSnapshot(final SimulationSnapshot snapshot) {
    this.inputsMemory = new HashMapMemory(snapshot.getInMemory());
    this.regMemory = new HashMapMemory(snapshot.getRegMemory());
    this.outputsMemory = new HashMapMemory(snapshot.getOutMemory());
  }
}
