/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.engine.basis.GraphMlUtils;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.DfsWalker;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgAssertStatement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.File;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * Control flow graph visitor for GraphML printer.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgGraphMlVisitor extends CfgEmptyVisitor {

  /* The print-to-GraphMl-file writer. */
  private PrintWriter printWriter;

  /* The expression printer. */
  private ExprTreePrinter exprPrinter;

  /* The identifier of current process sink node. */
  private String processSinkNodeId;

  /* The GraphML edges style */
  private String edgeStyle;

  private ModelWalker walker;

  /**
   * Default constructor.
   */
  CfgGraphMlVisitor() {
    super();
    setEdgeStyle(GraphMlUtils.STYLE_LINE);
  }

  /**
   * Sets GraphML edges style.
   *
   * @param edgeStyle Style of GraphML edges.
   */
  void setEdgeStyle(final String edgeStyle) {
    this.edgeStyle = edgeStyle;
  }

  private String getEdgeStyle() {
    return this.edgeStyle;
  }

  /**
   * Sets print writer.
   * @param printWriter Print writer.
   */
  void setPrintWriter(final PrintWriter printWriter) {
    InvariantChecks.checkNotNull(printWriter);
    this.printWriter = printWriter;
  }

  /**
   * Returns the current file writer.
   * @return The current file writer.
   */
  private PrintWriter getPrintWriter() {
    return this.printWriter;
  }

  /**
   * Returns the current expression printer.
   * @return The current expression printer.
   */
  private ExprTreePrinter getExprPrinter() {
    return this.exprPrinter;
  }

  /**
   * Returns the group node identifier for the specified statement.
   * @param statement The statement which group node is returned.
   * @return The group node identifier for the specified statement.
   */
  private String getGroupNodeId(final Statement statement) {
    return "Group" + statement.getId();
  }

  /**
   * Sets expression printer.
   * @param printer Expression printer.
   */
  void setExprPrinter(final ExprTreePrinter printer) {
    InvariantChecks.checkNotNull(printer);

    this.exprPrinter = printer;
  }

  void setWalker(final ModelWalker walker) {
    InvariantChecks.checkNotNull(walker);

    this.walker = walker;
  }

  private ModelWalker getWalker() {
    return this.walker;
  }

  /**
   * Initializes object with the specified file name.
   *
   * @param fileName GraphML file name.
   */
  public void initialize(final String fileName) {

    /* Create a GraphML file. */
    final File graphMlFile = FileCreator.newFile(fileName);
    setPrintWriter(ReadWriteUtils.newPrintWriter(graphMlFile));

    /* The expression printer initialization. */
    setExprPrinter(Retrascope.getConfiguration().getStylePrinter());
  }

  @Override
  public Status getStatus() {
    return Status.OK;
  }

  @Override
  public void onRootBegin() {
    GraphMlUtils.onGraphMlBegin(getPrintWriter());
  }

  @Override
  public void onRootEnd() {
    GraphMlUtils.onGraphMlEnd(getPrintWriter());
    GraphMlUtils.closeGraphMlFile(getPrintWriter());
  }

  @Override
  public void onModelBegin(final Model model) {
    GraphMlUtils.onGraphBegin(getPrintWriter());
  }

  @Override
  public void onModelEnd(final Model model) {
    GraphMlUtils.onGraphEnd(getPrintWriter());
  }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   *
   * @param node CFG model node.
   */
  private void printNodeBegin(final Statement node) {
    printNodeBegin(node, GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_PURPLE);
  }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   *
   * @param node CFG model node.
   * @param form Node form.
   * @param color Node color.
   */
  private void printNodeBegin(
      final Statement node,
      final String form,
      final String color) {

    /* Add the node to the GraphML file. */
    GraphMlUtils.printNode(getPrintWriter(), node.getId(), node.getDescription(getExprPrinter()),
        form, color);

    /* Add the edges to the GraphML file. */
    final Collection<Statement> successors = StatementUtils.getSuccessors(node);

    if (successors.isEmpty() && this.processSinkNodeId != null) {

      /* Add the edge to the process SINK node */
      GraphMlUtils.printEdge(getPrintWriter(), node.getId(), this.processSinkNodeId);

    } else if (!successors.isEmpty()) {

      for (final Statement successor : successors) {

        /* Case nodes will not be added to GraphML. */
        if (successor.isType(CfgStatementType.CASE)) {

          printGrandEdge(node, successor);

        } else {

          GraphMlUtils.printEdge(getPrintWriter(), node.getId(), getEdgeStyle(), successor.getId());
        }
      }
    }
  }

  private void printGrandEdge(final Statement node, final Statement successor) {

    final Collection<Statement> grandChildren = StatementUtils.getSuccessors(successor);

    if (grandChildren.isEmpty()) {

      /* Add the edge to the process SINK node */
      GraphMlUtils.printEdge(
          getPrintWriter(),
          node.getId(),
          getEdgeStyle(),
          successor.getDescription(getExprPrinter()),
          this.processSinkNodeId);

    } else {

      final Statement caseChild = StatementUtils.getSuccessors(successor).iterator().next();

      GraphMlUtils.printEdge(
          getPrintWriter(),
          node.getId(),
          getEdgeStyle(),
          successor.getDescription(getExprPrinter()),
          caseChild.getId());
    }
  }

  @Override
  public void onModuleBegin(final Module module) {
    printNodeBegin(module, GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    GraphMlUtils.printGroup(
        getPrintWriter(),
        process.getId(),
        process.getDescription(getExprPrinter()));
    GraphMlUtils.onGraphBegin(getPrintWriter(), getGroupNodeId(process));

    /* Print auxiliary SINK node. */
    this.processSinkNodeId = "SINK" + process.getId();
    GraphMlUtils.printNode(
        getPrintWriter(),
        this.processSinkNodeId,
        "SINK",
        GraphMlUtils.FORM_RECTANGLE,
        GraphMlUtils.COLOR_GRAY);
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    GraphMlUtils.onGraphEnd(getPrintWriter());
    GraphMlUtils.closeGroup(getPrintWriter());

    this.processSinkNodeId = null;
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {
    printNodeBegin(block);
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    printNodeBegin(switchNode, GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_RED);
  }

  @Override
  public void onWaitBegin(final CfgWaitStatement wait) {
    printNodeBegin(wait, GraphMlUtils.FORM_OCTAGON, GraphMlUtils.COLOR_ORANGE);
  }

  @Override
  public void onAssertBegin(final CfgAssertStatement assertStatement) {
    printNodeBegin(assertStatement, GraphMlUtils.FORM_TRAPEZOID, GraphMlUtils.COLOR_GRAY);
  }

  @Override
  public void onLoopBegin(final CfgLoopStatement loopStatement) {

    GraphMlUtils.printGroup(
        getPrintWriter(),
        loopStatement.getId(),
        loopStatement.getDescription(getExprPrinter()));
    GraphMlUtils.onGraphBegin(getPrintWriter(), getGroupNodeId(loopStatement));

    /* Print loop internals. */

    class CfgLoopVisitor extends CfgGraphMlVisitor {

      private Status status;
      private final Collection<String> visitedIds = new LinkedHashSet<>();

      private Collection<String> getVisitedStmtIds() {
        return this.visitedIds;
      }

      @Override
      public void onStatementBegin(final Statement statement) {

        if (!loopStatement.contains(statement)) {
          this.status = Status.SKIP;
        }
        visitedIds.add(statement.getId());
      }

      @Override
      public Status getStatus() {
        return this.status;
      }
    }

    final CfgLoopVisitor loopVisitor = new CfgLoopVisitor();
    loopVisitor.setPrintWriter(getPrintWriter());
    loopVisitor.setExprPrinter(getExprPrinter());
    loopVisitor.setEdgeStyle(GraphMlUtils.STYLE_DASH);
    final ModelWalker loopWalker = new DfsWalker(loopVisitor, WalkerMode.DFS_NO_RPT);

    for (int i = 0; i < loopStatement.getStatementNum(); i++) {

      loopWalker.visitNode(loopStatement.getStatement(i));
    }

    getWalker().addVisited(loopVisitor.getVisitedStmtIds());

    GraphMlUtils.onGraphEnd(getPrintWriter());
    GraphMlUtils.closeGroup(getPrintWriter());
  }
}
