/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml.depend;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.basis.ModelWalkerEngine;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * GraphML printer for dependencies in the Control Flow Graph model.
 * <p>The printer shows control dependencies between processes of
 * Control Flow Graph model.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgDependencyGraphMlPrinter extends ModelWalkerEngine {

  /**
   * Engine identifier.
   */
  public static final String ENGINE_ID = "cfg-dep-graphml-printer";

  /**
   * The engine output entity type.
   */
  public static final EntityType OUT_TYPE = EntityType.get("cfg-dep-graphml");

  /**
   * The command line parameters.
   */
  public static final Parameter OUT_FILE_NAME =
      new Parameter("cfg-dep-graphml", true, "Output GraphML file name");

  /* The default file name. */
  private static final String DEFAULT_FILE_NAME = "cfg-dependency-model.graphml";

  /* The printer result. */
  private FileNames fileNames;

  /**
   * Constructs object with the specified identifier, input/output types,
   * walkers.
   */
  public CfgDependencyGraphMlPrinter() {
    super(
        ENGINE_ID,
        CfgModel.TYPE,
        OUT_TYPE,
        new CfgDfsWalker(new CfgDependencyGraphVisitor(), WalkerMode.DFS_NO_RPT));
  }

  private FileNames getFileNames() {
    return this.fileNames;
  }

  private void setFileNames(final FileNames fileNames) {
    this.fileNames = fileNames;
  }

  private void initGraphMlVisitor(final String fileName) {
    ((CfgDependencyGraphVisitor)getWalkers()[0].getVisitor()).setFileName(fileName);
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {

      /* Get arguments from command line. */
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      /* Create GraphML file. */
      setFileNames(new FileNames("graphml"));
      fileNames.add(engineArgs.getValue(OUT_FILE_NAME, DEFAULT_FILE_NAME));

      initGraphMlVisitor(getFileNames().getList().get(0));

    } catch (ParseException e) {
      Log.error(e.toString());
    }
  }

  @Override
  public Entity getOutput() {
    return getFileNames();
  }
}
