/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml.depend;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.basis.GraphMlUtils;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.File;
import java.io.PrintWriter;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * GraphML printer of CFG processes dependencies.
 * <p>An experimental engine that prints a dependency scheme
 * of CFG model processes dependencies. It is said that a process {@code p1} depends from
 * process {@code p2} when {@code p2} defines variable(s) that is (are) used in the control part
 * of {@code p1}.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgDependencyGraphVisitor extends CfgEmptyVisitor {

  private PrintWriter writer;
  private final Map<String, Set<NodeVariable>> processDefines;
  private final Map<String, Set<NodeVariable>> processCtrlUses;
  private final Map<String, Set<NodeVariable>> processDataUses;

  /**
   * Constructs dependency printer.
   */
  public CfgDependencyGraphVisitor() {
    super();
    this.processDefines = new LinkedHashMap<>();
    this.processCtrlUses = new LinkedHashMap<>();
    this.processDataUses = new LinkedHashMap<>();
  }

  private PrintWriter getPrintWriter() {
    return this.writer;
  }

  private Map<String, Set<NodeVariable>> getDefines() {
    return this.processDefines;
  }

  private Map<String, Set<NodeVariable>> getControlUses() {
    return this.processCtrlUses;
  }

  private Map<String, Set<NodeVariable>> getDataUses() {
    return this.processDataUses;
  }

  private void addDefines(final String processId, final Set<NodeVariable> variables) {
    this.processDefines.put(processId, variables);
  }

  private void addControlUses(final String processId, final Set<NodeVariable> variables) {
    this.processCtrlUses.put(processId, variables);
  }

  private void addDataUses(final String processId, final Set<NodeVariable> variables) {
    this.processDataUses.put(processId, variables);
  }

  /**
   * Sets the specified string as name of the GraphML file to be generated.
   *
   * @param fileName GraphML file name.
   */
  public void setFileName(final String fileName) {
    final File file = FileCreator.newFile(fileName);
    this.writer = ReadWriteUtils.newPrintWriter(file);
  }

  @Override
  public void onRootBegin() {
    GraphMlUtils.onGraphMlBegin(getPrintWriter());
  }

  @Override
  public void onRootEnd() {

    printDependencyEdges();

    GraphMlUtils.onGraphMlEnd(getPrintWriter());
    GraphMlUtils.closeGraphMlFile(getPrintWriter());
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    if (getDefines().containsKey(process.getId())) {
      Log.error(LogMessage.makeMessage(LogMessage.ALREADY_CONTAIN, process.getId()));
    }

    addDefines(process.getId(), getProcessDefines(process));

    if (getControlUses().containsKey(process.getId())) {
      Log.error(LogMessage.makeMessage(LogMessage.ALREADY_CONTAIN, process.getId()));
    }

    if (getDataUses().containsKey(process.getId())) {
      Log.error(LogMessage.makeMessage(LogMessage.ALREADY_CONTAIN, process.getId()));
    }

    addControlUses(process.getId(), getProcessControlUses(process));
    addDataUses(process.getId(), getProcessDataUses(process));

    final boolean useInputs = useInputs(process);
    final boolean defOutputs = defOutputs(process);

    final boolean useInDefOut = useInputs && defOutputs;

    GraphMlUtils.printNode(
        getPrintWriter(),
        process.getId(),
        getInOutDescription(process),
        GraphMlUtils.FORM_ELLIPSE,
        useInDefOut
            ? GraphMlUtils.COLOR_BLUE
            : useInputs
            ? GraphMlUtils.COLOR_GREEN
            : defOutputs
            ? GraphMlUtils.COLOR_RED : GraphMlUtils.COLOR_YELLOW);
  }

  private boolean useInputs(final ModuleProcess process) {
    final Set<NodeVariable> controlUses = getControlUses().get(process.getId());
    final Set<NodeVariable> dataUses = getDataUses().get(process.getId());
    final Set<NodeVariable> uses = CollectionUtils.uniteSets(controlUses, dataUses);

    for (final NodeVariable controlUse : uses) {

      final VariableDescriptor descriptor = getVariableDescriptor(controlUse.getUserData());
      final VariableType type = descriptor.getVariableType();

      if (VariableType.isInputSignal(type)) {
        return true;
      }
    }
    return false;
  }

  private VariableDescriptor getVariableDescriptor(final Object userData) {

    if (!(userData instanceof VariableDescriptor)) {

      throw new RetrascopeException(
          String.format(
              "Wrong class for user data '%s': '%s' is expected.",
              userData.getClass(),
              VariableDescriptor.class));
    }

    return (VariableDescriptor) userData;
  }

  private boolean defOutputs(final ModuleProcess process) {

    final Set<NodeVariable> defines = getDefines().get(process.getId());

    for (final NodeVariable define : defines) {

      final VariableDescriptor descriptor = getVariableDescriptor(define.getUserData());
      final VariableType type = descriptor.getVariableType();

      if (VariableType.isOutputSignal(type)) {
        return true;
      }
    }
    return false;
  }

  private void printDependencyEdges() {

    for (final Map.Entry<String, Set<NodeVariable>> defEntry : getDefines().entrySet()) {

      final String srcId = defEntry.getKey();

      for (final Map.Entry<String, Set<NodeVariable>> ctrlUse : getControlUses().entrySet()) {

        final String dstId = ctrlUse.getKey();
        final Set<NodeVariable> intersect =
            CollectionUtils.intersectSets(defEntry.getValue(), ctrlUse.getValue());

        if (!intersect.isEmpty()) {

          GraphMlUtils.printEdge(
              getPrintWriter(),
              srcId,
              GraphMlUtils.STYLE_LINE,
              getDependDescription(intersect),
              GraphMlUtils.ALIGN_CENTER,
              GraphMlUtils.FONT_ITALIC,
              dstId);
        }
      }

      for (final Map.Entry<String, Set<NodeVariable>> dataUse : getDataUses().entrySet()) {

        final String dstId = dataUse.getKey();
        final Set<NodeVariable> intersect =
            CollectionUtils.intersectSets(defEntry.getValue(), dataUse.getValue());

        if (!intersect.isEmpty()) {

          GraphMlUtils.printEdge(
              getPrintWriter(),
              srcId,
              GraphMlUtils.STYLE_DASH,
              getDependDescription(intersect),
              GraphMlUtils.ALIGN_CENTER,
              GraphMlUtils.FONT_ITALIC,
              dstId);
        }
      }
    }
  }

  private String getDependDescription(final Set<NodeVariable> variables) {
    final Set<NodeVariable> inputs = new LinkedHashSet<>();
    final Set<NodeVariable> outputs = new LinkedHashSet<>();
    final Set<NodeVariable> registers = new LinkedHashSet<>();

    for (final NodeVariable variable : variables) {

      final VariableDescriptor descriptor = getVariableDescriptor(variable.getUserData());
      final VariableType type = descriptor.getVariableType();

      switch (type) {

        case IN:
          inputs.add(variable);
          break;

        case OUT:
          outputs.add(variable);
          break;

        case INOUT:
          inputs.add(variable);
          outputs.add(variable);
          break;

        case REG:
          registers.add(variable);
          break;

        default:
          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, type);
      }
    }

    return String.format(
        "%s%s%s",
        printSet("IN", inputs),
        printSet("OUT", outputs),
        printSet("REG", registers));
  }

  private String printSet(
      final String prefix,
      final Set<?> variables) {

    return variables.isEmpty() ? "" : String.format("%s %s;", prefix, variables.toString());
  }

  private String getInOutDescription(final ModuleProcess process) {

    final Set<NodeVariable> readInputs = new LinkedHashSet<>();
    final Set<NodeVariable> readOutputs = new LinkedHashSet<>();
    final Set<NodeVariable> writeInputs = new LinkedHashSet<>();
    final Set<NodeVariable> writeOutputs = new LinkedHashSet<>();

    for (int i = 0; i < process.getStatementNum(); i++) {
      elaborate(readInputs, readOutputs, writeInputs, writeOutputs, process.getStatement(i));
    }

    if (!writeInputs.isEmpty()) {
      throw new RetrascopeException("Wrong writing to input signals: " + writeOutputs);
    }

    return printAllSets(readInputs, readOutputs, writeInputs, writeOutputs);
  }

  private String printAllSets(
      final Set<?> readInputs,
      final Set<?> readOutputs,
      final Set<?> writeInputs,
      final Set<?> writeOutputs) {

    return String.format(
        "%s %s",
        printSetPair("Read", readInputs, readOutputs),
        printSetPair("Write", writeInputs, writeOutputs));
  }

  private String printSetPair(
      final String prefix,
      final Set<?> first,
      final Set<?> second) {

    return first.isEmpty() && second.isEmpty()
        ? "" : String.format("%s{%s;%s}", prefix, printSet("In", first), printSet("Out", second));
  }

  private void elaborate(
      final Set<NodeVariable> readInputs,
      final Set<NodeVariable> readOutputs,
      final Set<NodeVariable> writeInputs,
      final Set<NodeVariable> writeOutputs,
      final Statement node) {

    for (final NodeVariable use : node.getUses()) {
      selectSet(readInputs, readOutputs, use);
    }

    for (final NodeVariable define : node.getDefines()) {
      selectSet(writeInputs, writeOutputs, define);
    }
  }

  private void selectSet(
      final Set<NodeVariable> inputs,
      final Set<NodeVariable> outputs,
      final NodeVariable variable) {

    final VariableDescriptor descriptor = getVariableDescriptor(variable.getUserData());
    final VariableType type = descriptor.getVariableType();

    if (VariableType.isInputSignal(type)) {
      inputs.add(variable);
    }
    if (VariableType.isOutputSignal(type)) {
      outputs.add(variable);
    }
  }

  private Set<NodeVariable> getProcessControlUses(final ModuleProcess process) {
    final Set<NodeVariable> controlUses = new LinkedHashSet<>();

    addProcessControlUses(process, controlUses);

    return controlUses;
  }

  private void addProcessControlUses(final Statement composite, final Set<NodeVariable> ctrlUses) {

    for (int i = 0; i < composite.getStatementNum(); i++) {

      final Statement statement = composite.getStatement(i);

      if (statement.getStatementNum() == 0) {

        if (!statement.isType(CfgStatementType.BASIC_BLOCK)) {
          ctrlUses.addAll(statement.getUses());
        }
      } else {
        addProcessControlUses(statement, ctrlUses);
      }
    }
  }

  private Set<NodeVariable> getProcessDataUses(final ModuleProcess process) {
    final Set<NodeVariable> dataUses = new LinkedHashSet<>();

    addProcessDataUses(process, dataUses);
    return dataUses;
  }

  private void addProcessDataUses(final Statement composite, final Set<NodeVariable> dataUses) {

    for (int i = 0; i < composite.getStatementNum(); i++) {

      final Statement statement = composite.getStatement(i);

      if (statement.isType(CfgStatementType.BASIC_BLOCK)) {

        dataUses.addAll(statement.getUses());

      } else if (statement.getStatementNum() > 0) {

        addProcessDataUses(statement, dataUses);
      }
    }
  }

  private Set<NodeVariable> getProcessDefines(final ModuleProcess process) {
    final Set<NodeVariable> defines = new LinkedHashSet<>();

    addProcessDefines(process, defines);

    return defines;
  }

  private void addProcessDefines(final Statement composite, final Set<NodeVariable> defines) {

    for (int i = 0; i < composite.getStatementNum(); i++) {

      final Statement statement = composite.getStatement(i);

      if (statement.isType(CfgStatementType.BASIC_BLOCK)) {

        defines.addAll(statement.getDefines());

      } else if (statement.getStatementNum() > 0) {

        addProcessDefines(statement, defines);
      }
    }
  }
}
