/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.gadd.GaddModel;

import java.util.ArrayDeque;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Visitor component of transformer to {@link GaddModel} model.
 *
 * @see CfgGaddTransformer
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgGaddTransformerVisitor extends CfgEmptyVisitor {

  /* The current new module. */
  private final Deque<CfgModule> modules;

  /* The mapping of GADD modules to related CFG processes. */
  private final Map<CfgModule, List<CfgProcess>> moduleMap;

  /* The mapping of CFG modules into GADD modules. */
  private final Map<CfgModule, CfgModule> oldNewModuleMap;

  private Status status;

  /**
   * Default constructor.
   */
  CfgGaddTransformerVisitor() {
    super();
    this.modules = new ArrayDeque<>();
    this.moduleMap = new LinkedHashMap<>();
    this.oldNewModuleMap = new LinkedHashMap<>();
    this.status = Status.OK;
  }

  private Deque<CfgModule> getModules() {
    return this.modules;
  }

  /**
   * Returns the mapping of GADD modules to related CFG processes.
   *
   * @return The mapping of GADD modules to related CFG processes.
   */
  Map<CfgModule, List<CfgProcess>> getModuleMap() {
    return Collections.unmodifiableMap(this.moduleMap);
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  @Override
  public void onModuleBegin(final Module module) {

    final CfgModule newModule = new CfgModule(module);

    final Statement moduleParent = module.getParent();

    if (moduleParent != null && moduleParent instanceof CfgModule
        && this.oldNewModuleMap.containsKey(moduleParent)) {

      newModule.setParent(this.oldNewModuleMap.get(moduleParent));
    }
    this.moduleMap.put(newModule, new LinkedList<>());
    this.oldNewModuleMap.put((CfgModule) module, newModule);
    this.modules.push(newModule);
  }

  @Override
  public void onModuleEnd(final Module module) {
    this.modules.pop();
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    final CfgModule currentModule = getModules().peek();
    final List<CfgProcess> processes = this.moduleMap.get(currentModule);
    processes.add((CfgProcess)process);
    this.status = Status.SKIP;
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    if (getStatus().equals(Status.SKIP)) {
      this.status = Status.OK;
    }
  }
}
