/*
 * Copyright 2017-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgConditionStatement;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Deque;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * The component that performs Control Flow Graph (CFG) model statements placement
 * upon different phases for the specified process.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgProcessPhasePlacer {

  /**
   * The phase dependencies graph. The graph vertices are process\switch statements
   * and "merge" points.
   * The graph edges are annotated with phase values and some other related data (blocks between
   * statements, case statement data, if exists, etc.).
   */
  private Graph<Statement, PhaseEdge> phaseDepGraph;

  /* Phase value. */
  private Integer phase;

  /* The initial value for phase variable. */
  private static final int INIT_PHASE = 0;

  /**
   * The phase placement heuristic.
   */
  class PhasePlacementVisitor extends CfgEmptyVisitor {

    /* Blocks that are executed at current phase. */
    private List<CfgBlockStatement> phaseBlocks;

    /* The current process. */
    private ModuleProcess process;

    private Set<Integer> phases;

    /* Visited case statements. */
    private Set<CfgCaseStatement> visitedCases;

    /* Deque for statements that are current 'src' nodes in dependency graph. */
    private Deque<Statement> srcDeque;

    private Deque<Node> switchConds;

    /* Data that is related to the case statement that is being visited at the moment. */
    private CfgCaseData currentCaseData;

    private Status status;

    PhasePlacementVisitor() {
      setStatus(Status.OK);
    }

    Integer getPhase() {
      return this.phases.size() - 1;
    }

    void createNewPhase() {
      this.phases.add(this.phases.size());
    }

    private List<CfgBlockStatement> getPhaseBlocks() {
      return this.phaseBlocks;
    }

    private void addToPhasePath(final CfgBlockStatement block) {
      this.phaseBlocks.add(block);
    }

    private Deque<Statement> getSrcDeque() {
      return this.srcDeque;
    }

    private Statement getCurrentSrc() {
      return this.srcDeque.peek();
    }

    private Node getCurrentCond() {
      return this.switchConds.isEmpty() ? null : this.switchConds.peek();
    }

    private CfgCaseData getCurrentCaseData() {

      return this.currentCaseData;
    }

    @Override
    public Status getStatus() {
      return this.status;
    }

    void setStatus(final Status status) {
      this.status = status;
    }

    @Override
    public void onProcessBegin(final ModuleProcess process) {

      this.process = process;

      this.phaseBlocks = new LinkedList<>();
      this.visitedCases = new LinkedHashSet<>();
      this.srcDeque = new LinkedList<>();
      this.switchConds = new LinkedList<>();

      phaseDepGraph = new DirectedSparseMultigraph<>();
      phaseDepGraph.addVertex(process);

      /* Entering the process, phase changes from 0 to 1. */
      this.phases = new LinkedHashSet<>();
      this.phases.add(INIT_PHASE);
      createNewPhase();

      getSrcDeque().push(process);
    }

    @Override
    public void onProcessEnd(final ModuleProcess process) {
      getSrcDeque().pop();
    }

    @Override
    public void onBlockBegin(final CfgBlockStatement block) {

      final Statement predecessor = StatementUtils.getPredecessor(block);
      final Statement blockGrandParent = block.getParent().getParent();

      final boolean mergePointBlock = (predecessor instanceof CfgSwitchStatement)
          && !predecessor.equals(blockGrandParent);

      if (!mergePointBlock) {
        addToPhasePath(block);
        return;
      }

      /* So, this statement is a merge point, treat it as phase dependency graph vertex. */
      if (!phaseDepGraph.containsVertex(block)) {
        phaseDepGraph.addVertex(block);
      }

      final Statement currentSrc = getCurrentSrc();
      final boolean blockAfterSwitch = currentSrc instanceof CfgSwitchStatement;
      final Node cond = blockAfterSwitch ? getCurrentCond() : null;

      final List<CfgBlockStatement> blocks = new LinkedList<>();
      if (blockAfterSwitch) {
        blocks.addAll(getPhaseBlocks());
      } else {
        createNewPhase();
        blocks.add(block);
      }
      final PhaseEdge edge = new PhaseEdge(getPhase(), blocks, cond, getCurrentCaseData());

      if (!edge.containSameFieldEdge(phaseDepGraph, currentSrc, block)) {

        addEdge(phaseDepGraph, edge, currentSrc, block);
        getPhaseBlocks().clear();
        this.currentCaseData = null;
      }
      getSrcDeque().push(block);
    }

    @Override
    public void onBlockEnd(final CfgBlockStatement block) {

      if (getPhaseBlocks().isEmpty()) {
        return;
      }

      final int last = getPhaseBlocks().size() - 1;
      final CfgBlockStatement lastBlock = getPhaseBlocks().get(last);

      if (lastBlock.equals(block)) {
        getPhaseBlocks().remove(last);
      }
    }

    @Override
    public void onCaseBegin(final CfgCaseStatement caseNode) {

      if (this.visitedCases.contains(caseNode)) {

        setStatus(Status.SKIP);

      } else if (!StatementUtils.isLast(caseNode)) {

        createNewPhase();
        this.visitedCases.add(caseNode);
        this.currentCaseData = caseNode.getCfgCaseData();
      }
    }

    @Override
    public void onSwitchBegin(final CfgSwitchStatement switchNode) {

      final boolean prevVisited = phaseDepGraph.containsVertex(switchNode);

      if (!prevVisited) {

        phaseDepGraph.addVertex(switchNode);
      }

      final Statement src = getCurrentSrc();
      final boolean blockSrc = src instanceof CfgBlockStatement;

      final Node edgeCond = blockSrc ? null : getCurrentCond();
      final CfgCaseData data = blockSrc ? null : getCurrentCaseData();

      if (src instanceof CfgBlockStatement) {
        createNewPhase();
        getPhaseBlocks().add((CfgBlockStatement) src);
      }

      final PhaseEdge edge = new PhaseEdge(getPhase(), getPhaseBlocks(), edgeCond, data);

      if (!edge.containSameFieldEdge(phaseDepGraph, src, switchNode)) {

        addEdge(phaseDepGraph, edge, src, switchNode);
        getPhaseBlocks().clear();
        this.currentCaseData = null;
      }

      if (!getCurrentSrc().equals(switchNode)) {
        getSrcDeque().push(switchNode);
        this.switchConds.push(switchNode.getCondition());
      }

      if (prevVisited) {
        setStatus(Status.SKIP);
      }
    }

    @Override
    public void onSwitchEnd(final CfgSwitchStatement switchNode) {

      if (getStatus().equals(Status.SKIP)) {
        setStatus(Status.OK);
      }

      if (getCurrentSrc().equals(switchNode)) {
        getSrcDeque().pop();
        this.switchConds.pop();
      }
    }

    @Override
    public void onStatementEnd(final Statement statement) {

      if (!StatementUtils.isLast(statement)) {
        if (!this.srcDeque.isEmpty() && getCurrentSrc().equals(statement)) {
          getSrcDeque().pop();
        }
        return;
      }

      final Statement src = (getCurrentSrc() == null && statement instanceof CfgProcess
              || phaseDepGraph.containsVertex(statement))
          ? statement
          : getCurrentSrc();

      final CfgConditionStatement parentCondStmt =
          CfgConditionStatement.getSuperStatement(statement, getCurrentCond());

      final Node edgeCond = parentCondStmt == null ? null : getCurrentCond();
      final List<CfgBlockStatement> edgeBlocks = new LinkedList<>(getPhaseBlocks());

      if (statement instanceof CfgBlockStatement) {
        edgeBlocks.add((CfgBlockStatement) statement);
      }

      final CfgCaseData data = statement instanceof CfgCaseStatement
          ? ((CfgCaseStatement) statement).getCfgCaseData()
          : getCurrentCaseData();

      createNewPhase();
      final PhaseEdge edge = new PhaseEdge(getPhase(), edgeBlocks, edgeCond, data);

      if (!edge.containSameFieldEdge(phaseDepGraph, src, this.process)) {

        addEdge(phaseDepGraph, edge, src, this.process);
        getPhaseBlocks().clear();
      }

      if (getCurrentSrc() != null && getCurrentSrc().equals(statement)) {
        getSrcDeque().pop();
      }
    }

    private void checkWhenBlock(final Statement stmt) {

      if (stmt instanceof CfgBlockStatement && phaseDepGraph.getOutEdges(stmt).size() > 0) {
        throw RetrascopeException.exception(LogMessage.UNEXPECTED,
            "More than one out edge for block vertex");
      }
    }

    private void addEdge(
        final Graph<Statement, PhaseEdge> graph,
        final PhaseEdge edge,
        final Statement src,
        final Statement dst) {

      checkWhenBlock(src);
      graph.addEdge(edge, src, dst, EdgeType.DIRECTED);
    }
  }

  /**
   * Constructs the phase placement component for the specified statement.
   * <p>The constructor call includes the phase placement procedure itself.</p>
   * @param statement The statement to that the phase placement is done.
   */
  CfgProcessPhasePlacer(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    this.phase = INIT_PHASE;

    final PhasePlacementVisitor visitor = new PhasePlacementVisitor();
    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS);
    walker.start(statement);
    this.phase = visitor.getPhase();
  }

  /**
   * Returns the phase dependency tree.
   * @return The phase dependency tree.
   */
  Graph<Statement, PhaseEdge> getPhaseGraph() {
    return this.phaseDepGraph;
  }

  Integer getMaxPhase() {
    return this.phase;
  }
}
