/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.VariableVersionFactory;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The {@link ru.ispras.retrascope.model.gadd.GaddModel} model visitor
 * that transforms the target model into an SSA form and transforms
 * non-concurrent basic block nodes into concurrent non-ranged ones
 * and compresses basic block sequences.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgSsaVersionVisitor extends CfgEmptyVisitor {

  /* The current module. */
  private final Module module;

  /* The mapping from variables to their versions. */
  private final Map<NodeVariable, LinkedList<NodeVariable>> versionMap;

  /* Variables that are defined in the current model block path. */
  private final Set<NodeVariable> pathDefines;

  /* The list of blocking assignments for current basic block statement. */
  private final List<Assignment> blocking = new LinkedList<>();

  /* The list of concurrent assignments for current basic block statement. */
  private final List<Assignment> concurrent = new LinkedList<>();

  /**
   * Default constructor.
   */
  CfgSsaVersionVisitor(final Module module) {
    super();
    this.module = module;
    this.versionMap = new LinkedHashMap<>();
    initVersionMap(this.versionMap, this.module.getVariables());
    this.pathDefines = new LinkedHashSet<>();
  }

  private void initVersionMap(
      final Map<NodeVariable, LinkedList<NodeVariable>> map,
      final Collection<NodeVariable> variables) {

    for (final NodeVariable variable : variables) {

      final LinkedList<NodeVariable> versions = new LinkedList<>();
      versions.add(variable);
      map.put(variable, versions);
    }
  }

  @Override
  public void onBlockEnd(final CfgBlockStatement block) {

    /* If predecessor is basic block, skip it - this will be elaborated later. */
    final Statement preStmt = StatementUtils.getPredecessor(block);
    if (preStmt.isType(CfgStatementType.BASIC_BLOCK)) {
      return;
    }

    final Set<NodeVariable> defines = new LinkedHashSet<>();
    this.pathDefines.clear();

    /*
     * Classify all the statements of this execution path (this block and it's successors).
     */

    classify(block);

    /* Elaborate next statement if it is basic block. */
    final Collection<Statement> blockSuccessors = StatementUtils.getSuccessors(block);

    for (final Statement current : blockSuccessors) {

      if (current.isType(CfgStatementType.BASIC_BLOCK)) {

        classify((CfgBlockStatement) current);
        block.getParent().remove(current);
      } else {
        Log.error(LogMessage.WRONG_TYPE + " for GADD path statement: %s\n", current.getType());
      }
    }

    /* Find variables that are defined once - for them versions should not be included. */
    final List<NodeVariable> oneDefVars = getOnceDefined();

    /* Map for current path versions. */
    final Map<NodeVariable, LinkedList<NodeVariable>> blockVersionMap = new LinkedHashMap<>();
    initVersionMap(blockVersionMap, this.versionMap.keySet());

    /* Substitute variable entries with their right versions. */
    final List<Assignment> result = new LinkedList<>();

    for (final Assignment blockAssignment : this.blocking) {

      final RangedVariable target = blockAssignment.getTarget();
      final NodeVariable targetVar = target.getVariable();

      defines.add(targetVar);
      this.pathDefines.add(targetVar);

      Node value = blockAssignment.getValue();

      if (target.ranged() || target.indexed()) {
        value = RangedVariableUtils.transformNodeWithRanged(target, value);
      }

      value = applyVersions(value, lastVersionMap(blockVersionMap));

      final NodeVariable version =
          oneDefVars.contains(targetVar)
              ? targetVar
              : createVersion(targetVar, blockVersionMap, true);

      result.add(new Assignment(version, value));
    }

    final Map<NodeVariable, NodeVariable> blockMap = lastVersionMap(blockVersionMap);

    for (final Assignment concAssignment : this.concurrent) {

      Node concValue = applyVersions(concAssignment.getValue(), blockMap);
      final RangedVariable concTarget = concAssignment.getTarget();
      final NodeVariable concTargetVar = concTarget.getVariable();
      defines.add(concTargetVar);
      this.pathDefines.add(concTargetVar);

      if (concTarget.ranged() || concTarget.indexed()) {

        final Node rangeToRhs = RangedVariableUtils.transformNodeWithRanged(concTarget, concValue);

        /* Substitute only non-versioned entries with last version. */
        final Map<NodeVariable, NodeVariable> nonVersionTargetMap = new LinkedHashMap<>();
        nonVersionTargetMap.put(concTargetVar, blockVersionMap.get(concTargetVar).getLast());

        concValue = applyVersions(rangeToRhs, nonVersionTargetMap);
      }

      final NodeVariable targetVersion =
          oneDefVars.contains(concTargetVar)
              ? concTargetVar
              : createVersion(concTargetVar, blockVersionMap, false);
      result.add(new Assignment(targetVersion, concValue));
    }
    block.removeAssignments();
    block.addAssignments(result);

    /* Assign already defined variables to their real prototypes. */
    for (final NodeVariable defVar : defines) {

      if (oneDefVars.contains(defVar)) {
        continue;
      }
      final LinkedList<NodeVariable> versions = this.versionMap.get(defVar);
      InvariantChecks.checkNotNull(versions, "No versions for: " + defVar.getName());

      final NodeVariable lastVersion = blockVersionMap.get(defVar).getLast();

      block.addAssignment(new Assignment(defVar, lastVersion));
    }
    this.blocking.clear();
    this.concurrent.clear();
    this.pathDefines.clear();
  }

  private List<NodeVariable> getOnceDefined() {

    final List<NodeVariable> onceDefined = new LinkedList<>();
    final List<NodeVariable> moreThanOne = new LinkedList<>();

    onceDefined(this.blocking, onceDefined, moreThanOne);
    onceDefined(this.concurrent, onceDefined, moreThanOne);
    moreThanOne.clear();

    return onceDefined;
  }

  private void onceDefined(
      final List<Assignment> assignments,
      final List<NodeVariable> onceDefined,
      final List<NodeVariable> moreThanOne) {

    for (final Assignment assignment : assignments) {

      final NodeVariable define = assignment.getVariable();

      if (!onceDefined.contains(define) && !moreThanOne.contains(define)) {
        onceDefined.add(define);
      } else if (onceDefined.contains(define)) {
        onceDefined.remove(define);
        moreThanOne.add(define);
      }
    }
  }

  private NodeVariable createVersion(
      final NodeVariable variable,
      final Map<NodeVariable, LinkedList<NodeVariable>> map,
      final boolean blocking) {

    final LinkedList<NodeVariable> targetVersions = this.versionMap.get(variable);

    /*if (!this.pathDefines.contains(variable)) {
      this.pathDefines.add(variable);
      return variable;
    }*/

    final LinkedList<NodeVariable> pathVersions = map.get(variable);

    final NodeVariable version = pathVersions.size() == targetVersions.size()
        ? VariableVersionFactory.getVersion(variable, targetVersions)
        : targetVersions.get(pathVersions.size());

    if (pathVersions.size() == targetVersions.size()) {
      this.versionMap.get(variable).add(version);
      final Declaration declaration = this.module.getDeclaration(variable);
      this.module.declareVariable(version, new Declaration(declaration));
    }
    //if (blocking) {
    map.get(variable).add(version);
    //}

    return version;
  }

  private Node applyVersions(final Node node, final Map<NodeVariable, NodeVariable> versions) {

    final List<NodeBinding.BoundVariable> bindings = bindVersions(versions.keySet(), versions);

    return Transformer.substituteAllBindings(new NodeBinding(node, bindings));
  }

  private Map<NodeVariable, NodeVariable> lastVersionMap(
      final Map<NodeVariable, LinkedList<NodeVariable>> versionMap) {

    final Map<NodeVariable, NodeVariable> map = new LinkedHashMap<>();

    for (final Map.Entry<NodeVariable, LinkedList<NodeVariable>> entry : versionMap.entrySet()) {

      map.put(entry.getKey(), entry.getValue().getLast());
    }

    return map;
  }

  private List<NodeBinding.BoundVariable> bindVersions(
      final Collection<NodeVariable> variables,
      final Map<NodeVariable, NodeVariable> versions) {

    final List<NodeBinding.BoundVariable> bindings = new LinkedList<>();

    for (final NodeVariable variable : variables) {

      if (!this.pathDefines.contains(variable)) {
        continue;
      }
      final NodeVariable version = versions.get(variable);
      InvariantChecks.checkNotNull(version, "No versions found for: " + variable.getName());

      bindings.add(NodeBinding.bindVariable(variable, version));
    }

    return bindings;
  }

  private void classify(final CfgBlockStatement block) {

    final List<Assignment> assignments = block.getAssignments();

    if (block.isConcurrent()) {
      this.concurrent.addAll(assignments);
    } else {
      this.blocking.addAll(assignments);
    }
  }
}
