/*
 * Copyright 2017-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.gadd;

import edu.uci.ics.jung.graph.Graph;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.gadd.Phase;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Edge for phase placement graph.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class PhaseEdge {

  private final List<Statement> pathBlocks;
  private final Node switchCondition;
  private final CfgCaseData caseData;
  private final Phase pathPhase;
  private final Integer edgeHash;

  PhaseEdge(
      final Integer phaseValue,
      final List<CfgBlockStatement> blocks,
      final Node condition,
      final CfgCaseData data) {

    this.pathPhase = new Phase(phaseValue);
    this.pathBlocks = new LinkedList<>();
    this.pathBlocks.addAll(blocks);
    this.switchCondition = condition;
    this.caseData = data == null ? null : new CfgCaseData(data);
    this.edgeHash = System.identityHashCode(this);
  }

  Phase getPhase() {
    return this.pathPhase;
  }

  List<Statement> getPathBlocks() {
    return Collections.unmodifiableList(this.pathBlocks);
  }

  /**
   * Returns the data descriptor that is related to this edge.
   * @return The data descriptor that is related to this edge.
   */
  CfgCaseData getCaseData() {
    return this.caseData;
  }

  Node getCond() {
    return this.switchCondition;
  }

  private boolean hasData() {
    return this.caseData != null;
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }

    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final PhaseEdge phaseEdge = (PhaseEdge) obj;

    return getPathBlocks().equals(phaseEdge.getPathBlocks())
        && switchCondition.equals(phaseEdge.switchCondition)
        && edgeHash.equals(phaseEdge.edgeHash)
        && pathPhase.equals(phaseEdge.pathPhase)
        && (hasData() ? getCaseData().equals(phaseEdge.getCaseData()) : !phaseEdge.hasData());
  }

  @Override
  public int hashCode() {

    int result = getPathBlocks().hashCode();
    result = 31 * result + edgeHash;
    result = 31 * result + (getCaseData() == null ? 0 : getCaseData().hashCode());
    result = 31 * result + pathPhase.hashCode();

    return result;
  }

  public boolean containSameFieldEdge(
      final Graph<Statement, PhaseEdge> graph,
      final Statement src,
      final Statement dst) {

    final Collection<PhaseEdge> outEdges = graph.getOutEdges(src);

    if (outEdges == null || outEdges.isEmpty()) {
      return false;
    }

    final Collection<PhaseEdge> inEdges = graph.getInEdges(dst);

    if (inEdges == null || inEdges.isEmpty()) {
      return false;
    }

    final Set<PhaseEdge> hammock =
        CollectionUtils.intersectSets(new LinkedHashSet<>(outEdges), new LinkedHashSet<>(inEdges));

    for (final PhaseEdge edge : hammock) {

      final CfgCaseData edgeData = edge.getCaseData();
      final List<Statement> edgeBlocks = edge.getPathBlocks();
      final Node edgeCond = edge.getCond();

      final boolean equalCaseData =
          getCaseData() == null ? edgeData == null : getCaseData().equals(edgeData);
      final boolean equalConditions =
          getCond() == null ? edgeCond == null : getCond().equals(edgeCond);
      final boolean equalPaths = edgeBlocks.equals(getPathBlocks());

      if (equalCaseData && equalConditions && equalPaths) {
        return true;
      }
    }

    return false;
  }
}
