/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.SolverResultBuilder;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearUncoverChooser;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.engine.efsm.simulator.ModuleSimulator;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;

/**
 * Directed test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomGenerator
 * @see EfsmTestGenerator
 */
final class DirectedGenerator extends AtomicTestGenerator {

  private final String logEntryHeader;

  private final Set<List<EfsmTransition>> traversable;
  private final Map<EfsmTransition, Set<EfsmTransition>> testedCore;

  private final EfsmSimulator simulator;
  private final int loopLimit;

  private final EfsmTransitionCoverage totalCoverage;

  private NearUncoverChooser directedChooser;
  private final EfsmAnalyser analyser;

  private final Map<EfsmState, Map<EfsmState, List<EfsmTransition>>> preferredBridges;
  private final Set<List<List<EfsmTransition>>> unTraverseFragm;
  private final Set<List<EfsmTransition>> unTraversePlain;

  /**
   * Constructs a new EFSM atomic test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   * @param traversable a set of paths which are guaranteed to be traversable
   * @param loopLimit max number of single EFSM loop traversing in one path
   */
  DirectedGenerator(
      final Efsm efsm,
      final Set<List<EfsmTransition>> traversable,
      final int loopLimit,
      final ModuleSimulator simulator) {
    super();

    this.logEntryHeader = String.format("EFSM.TestGenerator.DirectedGenerator.%s:", efsm.getName());
    this.loopLimit = loopLimit;
    this.simulator = simulator.getSubSimulator(efsm);
    this.totalCoverage = new EfsmTransitionCoverage(efsm);
    this.analyser = new EfsmAnalyser(efsm);

    for (final List<EfsmTransition> traversablePath : traversable) {

      for (final EfsmTransition covered: traversablePath) {

        this.totalCoverage.markTransitionAsCovered(covered);
      }
    }

    this.traversable = new HashSet<>(traversable);
    this.testedCore = new HashMap<>();
    this.directedChooser = null;
    this.preferredBridges = new HashMap<>();
    this.unTraverseFragm = new HashSet<>();
    this.unTraversePlain = new HashSet<>();
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getEfsm();
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation,
   *         {@code null} is returned
   */
  @Override
  public Iterator<TestVector> getNextSequenceIterator() {

    if (!this.totalCoverage.isFull()) {

      if (this.directedChooser == null) {

        this.directedChooser = new NearUncoverChooser(getEfsm(), this.totalCoverage);
      }

      for (EfsmTransition transition = this.directedChooser.chooseNext(); transition != null;
          transition = this.directedChooser.chooseNext()) {

        final List<TestVector> vectors = this.generateSequence(transition);

        if (!vectors.isEmpty()) {

          this.directedChooser = new NearUncoverChooser(getEfsm(), this.totalCoverage);
          return vectors.iterator();
        }
      }
    }

    return null;
  }

  private List<TestVector> generateSequence(final EfsmTransition transition) {

    Log.debug(this.logEntryHeader
        + " generating a sequence to cover the transition: " + transition);

    List<TestVector> vectors = tryDirectPath(transition);

    if (!vectors.isEmpty()) {
      return vectors;
    }

    final Map<EfsmTransition, Set<List<EfsmTransition>>> defPaths = new HashMap<>();
    final Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefPaths = new HashMap<>();

    for (final List<EfsmTransition> defPath : this.analyser.getDefPaths(transition)) {

      final EfsmTransition coreTransition = defPath.get(0);
      Set<List<EfsmTransition>> relatedPaths = defPaths.get(coreTransition);

      if (relatedPaths == null) {
        relatedPaths = new HashSet<>();
        defPaths.put(coreTransition, relatedPaths);
      }

      relatedPaths.add(defPath);
    }

    for (final Set<List<EfsmTransition>> defLoops : this.analyser.getLoopDefPaths(transition)) {

      final EfsmTransition coreTransition = defLoops.iterator().next().get(0);
      Set<Set<List<EfsmTransition>>> relatedLoops = loopDefPaths.get(coreTransition);

      if (relatedLoops == null) {
        relatedLoops = new HashSet<>();
        loopDefPaths.put(coreTransition, relatedLoops);
      }

      relatedLoops.add(defLoops);
    }

    final Set<EfsmTransition> coreTransitions = new HashSet<>(defPaths.keySet());
    coreTransitions.addAll(loopDefPaths.keySet());

    if (coreTransitions.isEmpty()) {
      Log.debug(this.logEntryHeader + " the transition is independent of inner variables");
    }

    for (final EfsmTransition coreTransition : coreTransitions) {
      if (this.testedCore.containsKey(transition)
          && this.testedCore.get(transition).contains(coreTransition)) {
        continue;
      }

      if (this.totalCoverage.isCovered(coreTransition)) {

        Log.debug(this.logEntryHeader
            + " trying full definition paths from the transition: " + coreTransition);

        final List<EfsmTransition> shortestHeadPath = findShortestTraversablePath(coreTransition);

        if (shortestHeadPath.isEmpty()) {

          Log.debug(this.logEntryHeader + " the shortest next head path is empty");

        } else {

          Log.debug(this.logEntryHeader + " the shortest head path: ");
          logPlainPath(shortestHeadPath, LogLevel.DEBUG);
        }

        if (loopDefPaths.containsKey(coreTransition)) {

          for (final Set<List<EfsmTransition>> loopsGroup : loopDefPaths.get(coreTransition)) {

            vectors = traverseLoop(loopsGroup, shortestHeadPath, transition);

            if (!vectors.isEmpty()) {
              break;
            }
          }
        }

        if (vectors.isEmpty() && defPaths.containsKey(coreTransition)) {

          for (final List<EfsmTransition> definitionPath : defPaths.get(coreTransition)) {

            final List<List<EfsmTransition>> fragmentedDefPath = fragmentPath(definitionPath);
            Log.debug(this.logEntryHeader + " trying the definition path: ");
            logFragmentedPath(fragmentedDefPath, LogLevel.DEBUG);

            final List<List<EfsmTransition>> corePath = new ArrayList<>();
            corePath.add(shortestHeadPath);
            corePath.addAll(fragmentedDefPath);
            corePath.add(Collections.singletonList(transition));
            vectors = traverseFragmentedPath(corePath);

            if (!vectors.isEmpty()) {
              break;
            }
          }
        }

        if (!vectors.isEmpty()) {
          return vectors;
        }

        Set<EfsmTransition> tested = this.testedCore.get(transition);

        if (tested == null) {
          tested = new HashSet<>();
          this.testedCore.put(transition, tested);
        }

        tested.add(coreTransition);
      }
    }

    return new LinkedList<>();
  }

  private List<TestVector> tryDirectPath(final EfsmTransition transition) {

    Log.debug(
        this.logEntryHeader + " trying the shortest direct path to the target transition");

    List<EfsmTransition> directPath = null;

    for (final EfsmTransition pred : getEfsm().getIncoming(transition.getSrc())) {

      if (this.totalCoverage.isCovered(pred)) {

        final List<EfsmTransition> candidate = new ArrayList<>(findShortestTraversablePath(pred));

        if ((directPath == null) || (candidate.size() + 1 < directPath.size())) {
          candidate.add(pred);
          directPath = candidate;
        }
      }
    }

    assert directPath != null;
    directPath.add(transition);

    return traversePlainPath(directPath);
  }

  private List<List<EfsmTransition>> fragmentPath(final List<EfsmTransition> plainPath) {

    if (Log.getLogger().isLoggable(LogLevel.DEBUG)) {

      Log.debug(this.logEntryHeader + " fragmenting the path: ");
      logPlainPath(plainPath, LogLevel.DEBUG);
    }

    final List<List<EfsmTransition>> fragmentedPath = new ArrayList<>();
    List<EfsmTransition> currentFragment = new ArrayList<>();
    fragmentedPath.add(currentFragment);
    EfsmState currentState = null;

    for (final EfsmTransition transition : plainPath) {

      if (!currentFragment.isEmpty()) {

        final EfsmState nextState = transition.getSrc();

        assert currentState != null;

        if (!currentState.equals(nextState)) {

          currentFragment = new ArrayList<>();
          fragmentedPath.add(currentFragment);
        }
      }

      currentFragment.add(transition);
      currentState = transition.getDst();
    }

    if (Log.getLogger().isLoggable(LogLevel.DEBUG)) {

      Log.debug(this.logEntryHeader + " fragmented path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return fragmentedPath;
  }

  private List<List<EfsmTransition>> smoothPath(final List<List<EfsmTransition>> fragmentedPath) {

    if (Log.getLogger().isLoggable(LogLevel.DEBUG)) {

      Log.debug(this.logEntryHeader + " smoothing the fragmented path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    final List<List<EfsmTransition>> smoothPath = new ArrayList<>();
    EfsmState currentState = null;

    for (final List<EfsmTransition> fragment : fragmentedPath) {
      if (!fragment.isEmpty()) {
        if (currentState != null) {
          final EfsmState nextState = fragment.get(0).getSrc();

          if (!currentState.equals(nextState)) {
            List<EfsmTransition> bridge;

            if (this.preferredBridges.containsKey(currentState)
                && this.preferredBridges.get(currentState).containsKey(nextState)) {
              bridge = this.preferredBridges.get(currentState).get(nextState);
            } else {
              bridge = getEfsm().getShortestPath(
                  currentState,
                  nextState,
                  TransitionWeightBasis.GUARD_REGISTER_AMOUNT);
            }

            smoothPath.add(bridge);
          }
        }

        smoothPath.add(fragment);
        currentState = fragment.get(fragment.size() - 1).getDst();
      }
    }

    if (Log.getLogger().isLoggable(LogLevel.DEBUG)) {

      Log.debug(this.logEntryHeader + " smooth path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return smoothPath;
  }

  private List<EfsmTransition> plainPath(final List<List<EfsmTransition>> fragmentedPath) {

    final List<EfsmTransition> plainPath = new ArrayList<>();

    for (final List<EfsmTransition> fragment: fragmentedPath) {
      plainPath.addAll(fragment);
    }

    return plainPath;
  }

  private List<TestVector> traverseLoop(
      final Set<List<EfsmTransition>> plainLoopCores,
      final List<EfsmTransition> headPath,
      final EfsmTransition target) {

    Log.debug(this.logEntryHeader + " fragmenting the loop cores");

    final Set<List<List<EfsmTransition>>> loopCores = new HashSet<>(plainLoopCores.size());

    for (final List<EfsmTransition> plainLoopCore : plainLoopCores) {

      final LinkedList<EfsmTransition> normLoopCore = new LinkedList<>(plainLoopCore);
      normLoopCore.addLast(normLoopCore.removeFirst());
      loopCores.add(fragmentPath(normLoopCore));
    }

    Log.debug(this.logEntryHeader + " traversing the loop");
    final List<EfsmTransition> currentPath = new ArrayList<>(headPath);
    currentPath.add(plainLoopCores.iterator().next().get(0));

    return traverseMixedLoop(loopCores, fragmentPath(currentPath), target, 0);
  }

  private List<TestVector> traverseMixedLoop(
      final Set<List<List<EfsmTransition>>> loopCores,
      final List<List<EfsmTransition>> currentPath,
      final EfsmTransition targetTransition,
      final int limit) {

    if (limit == this.loopLimit) {

      Log.debug(this.logEntryHeader + " loop limit is reached: " + this.loopLimit);

      return new LinkedList<>();
    }

    Log.debug(this.logEntryHeader + " current path at iteration " + limit + ":");
    logFragmentedPath(currentPath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> loopCore : loopCores) {

      Log.debug(this.logEntryHeader + " proceeding with the loop:");
      logFragmentedPath(loopCore, LogLevel.DEBUG);

      final List<List<EfsmTransition>> newCurrentPath = new ArrayList<>(currentPath);
      newCurrentPath.addAll(loopCore);

      if (checkFragmentedPathTraversability(newCurrentPath)) {

        Log.debug(this.logEntryHeader
            + " the new loop is traversable. Trying to reach the target transition");

        final List<List<EfsmTransition>> completePath = new ArrayList<>(newCurrentPath);
        completePath.add(Collections.singletonList(targetTransition));

        List<TestVector> vectors = traverseFragmentedPath(completePath);

        if (vectors.isEmpty()) {
          Log.debug(this.logEntryHeader + " the target transition is still unreachable");

          vectors = traverseMixedLoop(loopCores, newCurrentPath, targetTransition, limit + 1);

          if (!vectors.isEmpty()) {
            return vectors;
          }
        } else {

          return vectors;
        }
      } else {

        Log.debug(this.logEntryHeader + " the new loop is untraversable. Rolling back");
      }
    }

    return new LinkedList<>();
  }

  private boolean checkFragmentedPathTraversability(final List<List<EfsmTransition>> corePath) {

    Log.debug(this.logEntryHeader + " checking the fragmented path for traversability: ");
    logFragmentedPath(corePath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> untraversablePath: this.unTraverseFragm) {

      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return false;
      }
    }

    final List<List<EfsmTransition>> smoothPath = this.smoothPath(corePath);
    boolean traversability = checkPlainPathTraversability(plainPath(smoothPath));

    if (!traversability) {

      final List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      final List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      Log.debug(this.logEntryHeader + " looking for untraversable fragment");

      for (final List<EfsmTransition> fragment : smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        final List<EfsmTransition> plainPath = plainPath(newSmoothPath);
        traversability = checkPlainPathTraversability(plainPath);

        if (!traversability) {

          Log.debug(this.logEntryHeader + " the untraversable fragment: ");
          logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {

            Log.debug(
                this.logEntryHeader + " the untraversable fragment is in the core path");

            this.unTraverseFragm.add(currentCorePath);

            return false;
          } else {
            final EfsmState srcState = fragment.get(0).getSrc();
            final EfsmState dstState = fragment.get(fragment.size() - 1).getDst();
            final Set<EfsmTransition> forbidden = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            final List<List<EfsmTransition>> pathCandidates =
                findAllPaths(srcState, dstState, forbidden);
            boolean success = false;

            for (final List<EfsmTransition> pathCandidate : pathCandidates) {

              if (!pathCandidate.equals(fragment)) {

                Log.debug(
                    this.logEntryHeader + " replacing the untraversable fragment with: ");
                logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (checkPlainPathTraversability(plainPath(newSmoothPath))) {
                  success = true;

                  Log.debug(this.logEntryHeader + " replacement is successful");

                  Map<EfsmState, List<EfsmTransition>> bridges =
                      this.preferredBridges.get(srcState);

                  if (bridges == null) {
                    bridges = new HashMap<>();
                    this.preferredBridges.put(srcState, bridges);
                  }

                  bridges.put(dstState, pathCandidate);

                  break;
                }
              }
            }

            if (!success) {

              Log.debug(this.logEntryHeader + " replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.unTraverseFragm.add(currentCorePath);

              return false;
            }
          }
        }
      }
    }

    return traversability;
  }

  private List<TestVector> traverseFragmentedPath(final List<List<EfsmTransition>> corePath) {

    Log.debug(this.logEntryHeader + " traversing the fragmented path: ");
    this.logFragmentedPath(corePath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> untraversablePath: this.unTraverseFragm) {

      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return new LinkedList<>();
      }
    }

    final List<List<EfsmTransition>> smoothPath = smoothPath(corePath);
    List<TestVector> vectors = traversePlainPath(plainPath(smoothPath));

    if (vectors.isEmpty()) {

      final List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      final List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      Log.debug(this.logEntryHeader + " looking for untraversable fragment");

      for (final List<EfsmTransition> fragment : smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        final List<EfsmTransition> plainPath = plainPath(newSmoothPath);

        boolean success;

        if (currentCorePath.size() == corePath.size()) {
          vectors = traversePlainPath(plainPath);
          success = vectors != null;
        } else {
          success = checkPlainPathTraversability(plainPath);
        }

        if (!success) {

          Log.debug(this.logEntryHeader + " the untraversable fragment: ");
          logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {

            Log.debug(
                this.logEntryHeader + " the untraversable fragment is in the core path. "
                    + "Finishing traverse of this core path");

            this.unTraverseFragm.add(currentCorePath);

            return new LinkedList<>();
          } else {
            final EfsmState srcState = fragment.get(0).getSrc();
            final EfsmState dstState = fragment.get(fragment.size() - 1).getDst();
            final Set<EfsmTransition> forbidden = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            final List<List<EfsmTransition>> pathCandidates =
                findAllPaths(srcState, dstState, forbidden);
            success = false;

            for (final List<EfsmTransition> pathCandidate : pathCandidates) {

              if (!pathCandidate.equals(fragment)) {

                Log.debug(
                    this.logEntryHeader + " replacing the untraversable fragment with: ");
                logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (checkPlainPathTraversability(plainPath(newSmoothPath))) {
                  success = true;

                  Log.debug(this.logEntryHeader + " replacement is successful");

                  break;
                }
              }
            }

            if (!success) {

              Log.debug(this.logEntryHeader + " replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.unTraverseFragm.add(currentCorePath);

              return new LinkedList<>();
            }
          }
        }
      }
    }

    return vectors;
  }

  private List<TestVector> traversePlainPath(final List<EfsmTransition> path) {

    final SolverResult solverResult = resolvePlainPath(path);

    if (solverResult.getStatus().equals(Status.SAT)) {

      Log.debug(this.logEntryHeader + " the path is traversable. Generating input vectors");

      final Map<Integer, Map<String, Data>> inputs =
          Substitutor.deannotateInputs(solverResult.getVariables());

      final List<TestVector> vectors = new ArrayList<>(path.size());

      for (final ListIterator<EfsmTransition> i = path.listIterator(); i.hasNext();) {
        final int index = i.nextIndex();
        final EfsmTransition transition = i.next();

        final Map<String, Data> vectorInputs = inputs.get(index);
        final Transaction transaction = new Transaction();

        if (vectorInputs != null) {
          for (final Entry<String, Data> input: vectorInputs.entrySet()) {
            transaction.addVariable(input.getKey(), new NodeValue(input.getValue()));
          }
        }

        final TestVector vector = new TestVector();
        vector.addTransaction(transaction);
        vector.addEvents(transition.getClocks());
        vectors.add(vector);

        Log.debug(this.logEntryHeader + " generated vector: " + vector);

        this.totalCoverage.markTransitionAsCovered(transition);
      }

      this.traversable.add(path);

      return vectors;
    } else {

      Log.debug(this.logEntryHeader + " the path is untraversable");

      this.unTraversePlain.add(path);

      return new LinkedList<>();
    }
  }

  private boolean checkPlainPathTraversability(final List<EfsmTransition> path) {
    final boolean traversability = resolvePlainPath(path).getStatus().equals(Status.SAT);

    if (!traversability) {
      this.unTraversePlain.add(path);
    }

    return traversability;
  }

  private SolverResult resolvePlainPath(final List<EfsmTransition> path) {

    for (final List<EfsmTransition> untraversablePath : this.unTraversePlain) {
      if ((path.size() >= untraversablePath.size())
          && path.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return new SolverResultBuilder(Status.UNSAT).build();
      }
    }

    Log.debug(this.logEntryHeader + " checking the plain path for traversability: ");
    logPlainPath(path, LogLevel.DEBUG);

    if (!this.simulator.getCurrentState().equals(path.get(0).getSrc())) {
      throw new IllegalArgumentException("The path is started not in the current state of the"
          + " simulator");
    }

    final Efsm efsm = this.simulator.getEfsm();
    final Substitutor substitutor = new Substitutor(efsm);
    Node pathConstraint = substitutor.backtrack(path);

    Log.debug(this.logEntryHeader + " the traverse constraint: " + pathConstraint);

    pathConstraint = this.simulator.fillExpression(pathConstraint, Collections.emptyMap());

    Log.debug(this.logEntryHeader + " the constraint to resolve: " + pathConstraint);

    return SolverUtils.solve(pathConstraint);
  }

  private List<List<EfsmTransition>> findAllPaths(
      final EfsmState src,
      final EfsmState dst,
      final Set<EfsmTransition> forbidden) {

    List<List<EfsmTransition>> foundPaths;
    final Efsm efsm = this.simulator.getEfsm();

    if (efsm.isReachable(src, dst)) {
      foundPaths = new ArrayList<>();

      for (final EfsmTransition firstTransition : efsm.getOutgoing(src)) {
        foundPaths.addAll(this.buildPaths(
            Collections.singletonList(firstTransition), dst,
            forbidden));
      }
    } else {
      foundPaths = Collections.emptyList();
    }

    Collections.sort(foundPaths, new Comparator<List<EfsmTransition>>() {

      @Override
      public int compare(final List<EfsmTransition> list1, final List<EfsmTransition> list2) {
        return list1.size() - list2.size();
      }
    });

    return foundPaths;
  }

  private Set<List<EfsmTransition>> buildPaths(
      final List<EfsmTransition> path,
      final EfsmState dst,
      final Set<EfsmTransition> forbidden) {

    final Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    final EfsmState currentState = path.get(path.size() - 1).getDst();

    if (currentState.equals(dst)) {
      foundPaths.add(path);
    } else {

      final Efsm efsm = this.simulator.getEfsm();

      for (final EfsmTransition nextTransition : efsm.getOutgoing(currentState)) {

        if (!path.contains(nextTransition) && !forbidden.contains(nextTransition)) {

          final List<EfsmTransition> newCurrentPath = new ArrayList<>(path);
          newCurrentPath.add(nextTransition);
          foundPaths.addAll(buildPaths(newCurrentPath, dst, forbidden));
        }
      }
    }

    return foundPaths;
  }

  private void logFragmentedPath(
      final List<List<EfsmTransition>> fragmentedPath,
      final Level logLevel) {

    for (final List<EfsmTransition> fragment : fragmentedPath) {

      Log.getLogger().log(logLevel, this.logEntryHeader + " path fragment: ");
      logPlainPath(fragment, logLevel);
    }
  }

  private void logPlainPath(
      final List<EfsmTransition> plainPath,
      final Level logLevel) {

    for (final EfsmTransition transition : plainPath) {
      Log.getLogger().log(logLevel,
          String.format("%s %s", this.logEntryHeader, transition.toString()));
    }
  }

  private List<EfsmTransition> findShortestTraversablePath(final EfsmTransition transition) {
    List<EfsmTransition> shortestPath = null;

    for (final List<EfsmTransition> traversablePath : this.traversable) {
      final int index = traversablePath.indexOf(transition);

      if (index != -1 && ((shortestPath == null) || (index + 1 < shortestPath.size()))) {
        shortestPath = traversablePath.subList(0, index);
      }
    }

    return shortestPath;
  }

  @Override
  public void reset() {
    // TODO Auto-generated method stub
  }
}
