/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;

/**
 * The class which performs the learning phase of the FATE algorithm.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 */
final class EfsmAnalyser {

  private final String logEntryHeader;

  private final Efsm efsm;

  private final Graph<EfsmTransition, VariableDependency> ctrlDepGraph;
  private final Graph<EfsmTransition, VariableDependency> dataDepGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;
  private final Map<EfsmTransition, Set<List<EfsmTransition>>> defPaths;
  private final Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefPaths;

  EfsmAnalyser(final Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE.Analyser." + efsm.getName();

    this.efsm = efsm;

    this.ctrlDepGraph = detectCtroDeps(efsm);
    this.dataDepGraph = detectDataDeps(efsm);
    this.counters = new HashMap<>();
    this.defPaths = new HashMap<>();
    this.loopDefPaths = new HashMap<>();
    this.detectCounters(efsm);
  }

  private Graph<EfsmTransition, VariableDependency> detectCtroDeps(final Efsm efsm) {

    Log.debug(this.logEntryHeader + ": detecting control dependencies");

    final Graph<EfsmTransition, VariableDependency> ctrlDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition trans : efsm.getTransitions()) {

      if (!ctrlDepGraph.containsVertex(trans)) {
        ctrlDepGraph.addVertex(trans);
      }

      for (final NodeVariable usedVar : trans.getGuardUses()) {

        final String usedVarName = usedVar.getName();
        final Object varData = usedVar.getUserData();

        if (varData instanceof VariableDescriptor) {

          if (((VariableDescriptor)varData).getVariableType() == VariableType.REG) {

            for (final EfsmTransition define : EfsmAnalyser.findVarDef(efsm, trans, usedVarName)) {

              if (!ctrlDepGraph.containsVertex(define)) {
                ctrlDepGraph.addVertex(define);
              }

              final VariableDependency edge = new VariableDependency(define, trans, usedVar);
              ctrlDepGraph.addEdge(edge, define, trans, EdgeType.DIRECTED);
            }
          }
        } else {
          throw RetrascopeException.exception(
              LogMessage.UNSUPPORTED_TYPE,
              varData.getClass());
        }
      }
    }

    Log.debug(this.logEntryHeader + ": control dependencies have been detected successfully");

    return ctrlDepGraph;
  }

  private Graph<EfsmTransition, VariableDependency> detectDataDeps(final Efsm efsm) {

    Log.debug(this.logEntryHeader + ": detecting data dependencies");

    final Graph<EfsmTransition, VariableDependency> dataDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition trans : efsm.getTransitions()) {

      if (!dataDepGraph.containsVertex(trans)) {
        dataDepGraph.addVertex(trans);
      }

      for (final NodeVariable usedVar : trans.getActionUses()) {

        final String usedName = usedVar.getName();
        final Object varData = usedVar.getUserData();

        if (varData instanceof VariableDescriptor) {

          final VariableDescriptor descriptor = (VariableDescriptor) varData;
          final VariableType varType = descriptor.getVariableType();

          if (varType.equals(VariableType.REG)) {

            for (final EfsmTransition defTrans : EfsmAnalyser.findVarDef(efsm, trans, usedName)) {

              if (!dataDepGraph.containsVertex(defTrans)) {
                dataDepGraph.addVertex(defTrans);
              }

              final VariableDependency edge = new VariableDependency(defTrans, trans, usedVar);
              dataDepGraph.addEdge(edge, defTrans, trans, EdgeType.DIRECTED);
            }
          }
        } else {
          throw RetrascopeException.exception(
              LogMessage.UNSUPPORTED_TYPE,
              varData.getClass());
        }
      }
    }

    Log.debug(this.logEntryHeader + ": data dependencies have been detected successfully");

    return dataDepGraph;
  }

  private static List<EfsmTransition> findVarDef(
      final Efsm efsm,
      final EfsmTransition trans,
      final String useName) {

    final Set<EfsmTransition> visited = new HashSet<>();
    final Queue<EfsmTransition> queue = new LinkedList<>();
    queue.add(trans);
    final List<EfsmTransition> defining = new ArrayList<>();

    while (!queue.isEmpty()) {

      final EfsmTransition current = queue.poll();

      for (final EfsmTransition prev : efsm.getIncoming(current.getSrc())) {

        if (!visited.contains(prev)) {

          boolean useDefMatching = false;

          for (final NodeVariable defVar : prev.getDefines()) {

            if (defVar.getName().equals(useName)) {
              useDefMatching = true;

              break;
            }
          }

          if (useDefMatching) {
            defining.add(prev);
          } else {
            queue.add(prev);
            visited.add(prev);
          }
        }
      }
    }

    return defining;
  }

  private void detectCounters(final Efsm efsm) {

    Log.debug(this.logEntryHeader + ": detecting counters");

    final Map<NodeVariable, Set<EfsmTransition>> processed = new HashMap<>();

    for (final VariableDependency dep : this.ctrlDepGraph.getEdges()) {

      final EfsmTransition defTrans = this.ctrlDepGraph.getSource(dep);

      if (processed.containsKey(dep.getVariable())) {

        final Set<EfsmTransition> processedTrans = processed.get(dep.getVariable());

        if (processedTrans.contains(defTrans)) {
          continue;
        } else {
          processedTrans.add(defTrans);
        }
      } else {
        final Set<EfsmTransition> transitions = new HashSet<>();
        transitions.add(defTrans);
        processed.put(dep.getVariable(), transitions);
      }

      final EfsmState srcState = defTrans.getSrc();
      final EfsmState dstState =
          this.ctrlDepGraph.getDest(dep).getDst();
      final Set<NodeVariable> defVars =
          EfsmUtils.getResultingDependencies(defTrans.getAction())
              .get(dep.getVariable());

      for (final Iterator<NodeVariable> iterator = defVars.iterator(); iterator.hasNext();) {
        if (efsm.getParentContainer().declaresInput(iterator.next().getName())) {
          iterator.remove();
        }
      }

      if (!defVars.isEmpty() && efsm.isReachable(dstState, srcState)) {

        final Set<NodeVariable> variables = new HashSet<>(1);
        variables.add(dep.getVariable());

        for (final VariableDependency dataDep : this.dataDepGraph.getOutEdges(defTrans)) {

          if (dataDep.getVariable().equals(dep.getVariable())) {

            final EfsmTransition transition = this.dataDepGraph.getDest(dataDep);

            final List<EfsmTransition> path = Collections.singletonList(defTrans);
            findCounterLoop(dataDep.getVariable(), transition, defVars, variables, path);
          }
        }
      }
    }

    Log.debug(this.logEntryHeader + ": counters have been detected successfully");
  }

  //target transition is the first in the returned list
  private void findCounterLoop(
      final NodeVariable probCounter,
      final EfsmTransition current,
      final Set<NodeVariable> targets,
      final Set<NodeVariable> currentVars,
      final List<EfsmTransition> path) {

    if (path.contains(current)) {

      if (current.equals(path.get(0))) {

        final Set<NodeVariable> intersection = new HashSet<>(targets);
        intersection.retainAll(currentVars);

        if (!intersection.isEmpty()) {

          Set<List<EfsmTransition>> counterData = this.counters.get(probCounter);

          if (counterData == null) {

            counterData = new HashSet<>(1);
            this.counters.put(probCounter, counterData);
          }

          counterData.add(path);

          Log.debug(
              "%s : the variable %s defined by the transition %s  is a counter in the loop %s",
              this.logEntryHeader,
              probCounter.getName(),
              path.get(0),
              path);
        }
      }
    } else {

      final Set<NodeVariable> newVars = new HashSet<>();

      for (final Entry<NodeVariable, Set<NodeVariable>> dep :
          EfsmUtils.getResultingDependencies(current.getAction()).entrySet()) {

        final NodeVariable variable = dep.getKey();
        final Set<NodeVariable> intersection = new HashSet<>(currentVars);
        intersection.retainAll(dep.getValue());

        if (!intersection.isEmpty()) {
          newVars.add(variable);
        }
      }

      if (!newVars.isEmpty()) {

        final List<EfsmTransition> newCurrentPath = new ArrayList<>(path);
        newCurrentPath.add(current);

        for (final VariableDependency dataDep : this.dataDepGraph.getOutEdges(current)) {

          if (newVars.contains(dataDep.getVariable())) {

            final EfsmTransition transition = this.dataDepGraph.getDest(dataDep);

            this.findCounterLoop(probCounter, transition, targets, newVars, newCurrentPath);
          }
        }
      }
    }
  }

  public Set<List<EfsmTransition>> getDefPaths(final EfsmTransition transition) {

    Set<List<EfsmTransition>> cachedPaths = this.defPaths.get(transition);

    if (cachedPaths == null) {

      cachedPaths = new HashSet<>();
      this.defPaths.put(transition, cachedPaths);

      final Map<EfsmTransition, Set<NodeVariable>> deps = new HashMap<>();

      for (final VariableDependency dep : this.ctrlDepGraph.getInEdges(transition)) {

        final NodeVariable depVar = dep.getVariable();
        final EfsmTransition defTrans = this.ctrlDepGraph.getSource(dep);
        final Map<NodeVariable, Set<NodeVariable>> resDeps =
            EfsmUtils.getResultingDependencies(defTrans.getAction());
        final Set<NodeVariable> variables = resDeps.get(depVar);

        Set<NodeVariable> stored = deps.get(defTrans);

        if (stored == null) {
          stored = new HashSet<>();
          deps.put(defTrans, stored);
        }

        stored.addAll(variables);
      }

      for (final Entry<EfsmTransition, Set<NodeVariable>> dep : deps.entrySet()) {

        final List<EfsmTransition> path = Collections.singletonList(dep.getKey());
        cachedPaths.addAll(this.findDefinitionPaths(path, dep.getValue()));
      }
    }

    return cachedPaths;
  }

  public Set<Set<List<EfsmTransition>>> getLoopDefPaths(final EfsmTransition transition) {

    Set<Set<List<EfsmTransition>>> cached = this.loopDefPaths.get(transition);

    if (cached == null) {

      cached = new HashSet<>();
      this.loopDefPaths.put(transition, cached);

      for (final VariableDependency dep : this.ctrlDepGraph.getInEdges(transition)) {

        final NodeVariable depVar = dep.getVariable();
        final EfsmTransition defTrans = this.ctrlDepGraph.getSource(dep);
        final Set<List<EfsmTransition>> paths = new HashSet<>();

        if (this.counters.containsKey(depVar)) {

          for (final List<EfsmTransition> path : this.counters.get(depVar)) {
            if (path.get(0).equals(defTrans)) {
              paths.add(path);
            }
          }

          if (!paths.isEmpty()) {
            cached.add(paths);
          }
        }
      }
    }

    return cached;
  }

  private Set<List<EfsmTransition>> findDefinitionPaths(
      final List<EfsmTransition> path,
      final Set<NodeVariable> vars) {

    final Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    final Set<NodeVariable> innerDeps = new HashSet<>();

    for (final NodeVariable variable : vars) {

      if (this.efsm.getParentContainer().declaresRegister(variable.getName())) {

        innerDeps.add(variable);
      }
    }

    if (innerDeps.isEmpty()) {
      foundPaths.add(path);
    } else {
      if (innerDeps.size() < vars.size()) {
        foundPaths.add(path);
      }

      final EfsmTransition current = path.get(0);
      final Map<EfsmTransition, Map<NodeVariable, Set<NodeVariable>>> deps = new HashMap<>();

      for (final VariableDependency dep : this.dataDepGraph.getInEdges(current)) {

        final NodeVariable depVar = dep.getVariable();

        if (innerDeps.contains(depVar)) {

          final EfsmTransition defTrans = this.dataDepGraph.getSource(dep);
          final Map<NodeVariable, Set<NodeVariable>> resDeps =
              EfsmUtils.getResultingDependencies(defTrans.getAction());
          final Set<NodeVariable> variables = resDeps.get(depVar);

          Map<NodeVariable, Set<NodeVariable>> stored = deps.get(defTrans);

          if (stored == null) {
            stored = new HashMap<>();
            deps.put(defTrans, stored);
          }

          stored.put(depVar, variables);
        }
      }

      for (final Entry<EfsmTransition, Map<NodeVariable,Set<NodeVariable>>> dep : deps.entrySet()) {

        if (path.contains(dep.getKey())) {

          final int limit = dep.getKey().getAction().getAssignments().size();
          int occurrences = 0;

          for (final EfsmTransition transition : path) {

            if (transition.equals(dep.getKey())) {
              occurrences++;
            }
          }

          if (occurrences >= limit) {
            continue;
          }
        }

        final List<EfsmTransition> newPath = new ArrayList<>(path);
        final Set<NodeVariable> variables = new HashSet<>(innerDeps);
        variables.removeAll(dep.getValue().keySet());

        for (final Set<NodeVariable> newVariables : dep.getValue().values()) {
          variables.addAll(newVariables);
        }

        newPath.add(0, dep.getKey());
        foundPaths.addAll(this.findDefinitionPaths(newPath, variables));
      }
    }

    return foundPaths;
  }
}
