/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.util.SolverUtils;

/**
 * The utility class for EFSM test generators.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class TestGeneratorUtils {

  private TestGeneratorUtils() {
    /* Empty. */
  }

  /**
   * Adds a transaction with randomised inputs to the specified vector. Only the specified module's
   * inputs which are currently aren't in the vector are included to the transaction.
   *
   * @param efsmModule the EFSM module whose inputs are tested
   * @param inputVector a vector to be supplemented with random transaction
   *
   * @throws IllegalArgumentException if at least one argument is {@code null}
   */
  static void randomiseUnusedInputs(final EfsmModule efsmModule, final TestVector inputVector) {
    InvariantChecks.checkNotNull(efsmModule);
    InvariantChecks.checkNotNull(inputVector);

    final Transaction rndFilledTrans = new Transaction();

    for (final String inputName : efsmModule.getInputNames()) {
      if (!inputVector.containsVariable(inputName)) {
        NodeValue value;
        final Node constraint = efsmModule.getInvariant(inputName);

        if (constraint == null) {
          value = new NodeValue(Random.newValue(efsmModule.getDataType(inputName)));
        } else {
          if ((efsmModule.getDataType(inputName).getTypeId() == DataTypeId.LOGIC_INTEGER)
              && (constraint.getKind() == Node.Kind.OPERATION)
              && (((NodeOperation) constraint).getOperandCount() == 2)
              && (((NodeOperation) constraint).getOperand(0).getKind() == Node.Kind.OPERATION)
              && (((NodeOperation) constraint).getOperand(1).getKind() == Node.Kind.OPERATION)) {
            final NodeOperation firstCond =
                (NodeOperation) ((NodeOperation) constraint).getOperand(0);
            final NodeOperation secondCond =
                (NodeOperation) ((NodeOperation) constraint).getOperand(1);

            if (ExprUtils.isOperation(firstCond,
                StandardOperation.GREATEREQ, StandardOperation.LESSEQ)) {
              final int min = ((NodeValue)firstCond.getOperand(1)).getInteger().intValue();
              final int max = ((NodeValue)secondCond.getOperand(1)).getInteger().intValue();

              final NodeValue module = new NodeValue(Data.newInteger(max - min));
              final NodeValue minValueNode = new NodeValue(Data.newInteger(min));
              value = new NodeValue(Random.newValue(DataType.INTEGER));
              final NodeVariable targetVariable = NodeVariable.newInteger(inputName);

              final NodeOperation modValueNode = Nodes.mod(value, module);
              final NodeOperation valueComputation = Nodes.add(modValueNode, minValueNode);
              final Node assignment = Nodes.eq(targetVariable, valueComputation);
              final SolverResult result = SolverUtils.solve(assignment);

              if (result.getStatus() == Status.SAT) {
                value = new NodeValue(result.getVariables().iterator().next().getData());
              } else {
                throw new RetrascopeException("Failed to resolve invariant constraint on "
                    + "input " + inputName + " due to the errors: " + result.getErrors());
              }
            } else {
              throw new IllegalArgumentException("Such invariant is currently unsupported: "
                  + constraint);
            }
          } else {
            final SolverResult result = SolverUtils.solve(constraint);

            if (result.getStatus() == Status.SAT) {
              value = new NodeValue(result.getVariables().iterator().next().getData());
            } else {
              throw new RetrascopeException("Failed to resolve invariant constraint on "
                  + "input " + inputName + " due to the errors: " + result.getErrors());
            }
          }
        }

        rndFilledTrans.addVariable(inputName, value);
      }
    }

    if (!rndFilledTrans.isEmpty()) {
      inputVector.addTransaction(rndFilledTrans);
    }
  }
}
