/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction;

import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * The class representing a "nearest uncovered transition" priority-based heuristic for choosing
 * a transitions to traverse.
 * <br>
 * Each instance of RandomEfsmTransitionChooser is bounded to the specified {@link Efsm} and
 * {@link EfsmTransitionCoverage}.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class NearUncoverChooser extends EfsmTransitionChooser {

  private final Set<EfsmTransition> possible;

  /**
   * Creates a new NearUncoverChooser for the specified EFSM and its coverage.
   * <br>
   * Note that list of choosable transitions is constructed only during a
   * RandomEfsmTransitionChooser instance initialisation, so any future changes in the specified
   * EFSM and coverage instances will not affect a result of choosing by that
   * NearUncoverChooser instance.
   * <br>
   * The list of choosable transitions is constructed of all uncovered transitions outgoing from end
   * states of covered transitions. An order of these transitions is unspecified.
   *
   * @param efsm the EFSM
   * @param coverage the test coverage of the EFSM
   */
  public NearUncoverChooser(
      final Efsm efsm,
      final EfsmTransitionCoverage coverage) {

    this.possible = new HashSet<>();

    for (final EfsmTransition uncovered : coverage.getUncovered()) {

      if (uncovered.getSrc().equals(efsm.getInitialState())) {
        this.possible.add(uncovered);
      }

      for (final EfsmTransition previous : efsm.getIncoming(uncovered.getSrc())) {

        if (coverage.getCovered().contains(previous)) {
          this.possible.add(uncovered);
        }
      }
    }
  }

  /**
   * Chooses and returns a transition for traversing.
   * <br>
   * The chosen transition is a first element in the choosable transitions list conhstructed during
   * this NearUncoverChooser initialisation. Once chosen the transition is
   * removed from the list.
   *
   * @return the transition proposed for traversing. {@code null} is returned if all choosable
   *         transitions were chosen or there are no choosable transitions at all
   */
  @Override
  public EfsmTransition chooseNext() {

    if (this.possible.isEmpty()) {

      return null;

    } else {

      final Iterator<EfsmTransition> iterator = this.possible.iterator();
      final EfsmTransition nextTransition = iterator.next();
      iterator.remove();

      return nextTransition;
    }
  }

  @Override
  public boolean hasNext() {
    return !this.possible.isEmpty();
  }
}
