/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.engine.basis.SimulatorEngine;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.result.test.SequenceModel;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * The simulator for {@link EfsmModel}.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmModel
 */
public class EfsmModelSimulator extends SimulatorEngine {

  /**
   * The engine ID of all {@code EfsmModelSimulator} instances.
   */
  private static final String ENGINE_ID = "efsm-simulator";

  private static final String LOG_ENTRY_HEADER = "EFSM.Simulator:";

  /**
   * Creates a simulator.
   */
  public EfsmModelSimulator() {
    super(EfsmModelSimulator.ENGINE_ID, EfsmModel.TYPE);
  }

  /**
   * Simulates the specified input sequences on the specified EFSM modules and returns produced
   * output sequences. Amounts of the sequences and modules must be equal.
   *
   * @param inputs a map containing the input sequences and the EFSM modules. It must contain
   *        mappings from {@link SequenceModel} entity type to a {@link SequenceModel} object
   *        and from {@link EfsmModel} entity type to an {@link EfsmModel} object
   *
   *
   * @return the output sequences
   *
   * @throws IllegalArgumentException if {@code inputs} is {@code null}
   * @throws IllegalArgumentException if {@code inputs} contains either no mapping from
   *         {@link SequenceModel} entity type to a {@link SequenceModel} object or from
   *         {@link EfsmModel} entity type to an {@link EfsmModel} object
   * @throws IllegalArgumentException if amounts of sequences in {@code SequenceModel} and top-level
   *         modules in {@code EfsmModel} are different
   */
  @Override
  public SequenceModel start(final Map<EntityType, Entity> inputs) {
    InvariantChecks.checkNotNull(inputs);

    final EfsmModel efsmModel = Engine.getInputEntity(inputs, EfsmModel.class, EfsmModel.TYPE);
    final SequenceModel sequenceModel =
        Engine.getInputEntity(inputs, SequenceModel.class, SequenceModel.TYPE);

    InvariantChecks.checkTrue(efsmModel.getModules().size() == sequenceModel.getSequences().size());

    final SequenceModel output = new SequenceModel();

    for (final Module module : efsmModel.getModules()) {
      final String prefix = String.format("%s module %s:", LOG_ENTRY_HEADER, module.getName());

      Log.info("%s starting simulation", prefix);
      Log.info("%s simulation has been finished successfully", prefix);
    }

    return output;
  }
}
