/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.basis.Simulated;
import ru.ispras.retrascope.engine.basis.SimulationSnapshot;
import ru.ispras.retrascope.engine.efsm.generator.test.Substitutor;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.memory.HashMapMemory;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.TestVector;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

/**
 * The EFSM-based representation of a hardware model which encapsulates a runtime state along with a
 * static structure of an original device.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmSimulator implements Simulated {

  private final String logEntryHeader;

  private final ModuleSimulator parent;

  private final Efsm efsm;

  private final String name;

  private EfsmState currentState;

  private EventList currentEvents;

  private final EfsmTransitionCoverage coverage;

  private EfsmTransition lastTraversed;

  /**
   * Constructs a simulator above the specified EFSM.
   *
   * @param parent a simulator of a module which the specified EFSM belongs to
   * @param efsm an EFSM which the simulator is based on
   *
   * @throws IllegalArgumentException if at least one argument is {@code null}
   *
   * @see Efsm#getInitialState()
   * @see Efsm#getResetGuardedActions()
   */
  EfsmSimulator(final ModuleSimulator parent, final Efsm efsm) {
    InvariantChecks.checkNotNull(parent);
    this.parent = parent;
    this.efsm = efsm;

    this.logEntryHeader = String.format("EFSM.Simulator.%s:", efsm.getName());

    this.name = efsm.getName();

    this.coverage = new EfsmTransitionCoverage(efsm);
  }

  /**
   * Returns a name of this simulator.
   *
   * @return the name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns the simulated EFSM.
   * @return The simulated model.
   */
  public Efsm getEfsm() {
    return efsm;
  }

  public String getLogEntryHeader() {
    return logEntryHeader;
  }

  /**
   * Returns a state of an underlying EFSM which this simulator is currently in.
   *
   * @return the current state of this simulator. Returns the same state as
   *         {@link Efsm#getInitialState()} do if called immediately
   *         after construction/reset of this simulator
   *
   * @see EfsmSimulator#reset()
   */
  @Override
  public EfsmState getCurrentState() {
    return this.currentState;
  }

  public void setCurrentState(final EfsmState state) {

    this.currentState = state;
  }

  @Override
  public HashMapMemory getInputsMemory() {

    return this.parent.getInputsMemory();
  }

  @Override
  public HashMapMemory getRegMemory() {

    return this.parent.getRegMemory();
  }

  @Override
  public HashMapMemory getOutputsMemory() {

    return this.parent.getOutputsMemory();
  }

  /**
   * Returns a transition which was traversed by this simulator on last cycle.
   *
   * @return the last traversed transition
   */
  public EfsmTransition getLastTraversed() {
    return this.lastTraversed;
  }

  public void setLastTraversed(final EfsmTransition transition) {
    this.lastTraversed = transition;
  }

  /**
   * Returns a test coverage reached during simulation since last reset.
   *
   * @return the test coverage
   */
  @Override
  public EfsmTransitionCoverage getCoverage() {
    return new EfsmTransitionCoverage(this.coverage);
  }

  /**
   * Resets this simulator, that is, returns it to an uninitialised state.
   */
  public void reset() {

    Log.debug("%s resetting", this.logEntryHeader);

    this.coverage.reset();

    setCurrentState(getEfsm().getInitialState());

    if (getCurrentState().equals(Efsm.UNINIT_STATE)) {

      final Collection<EfsmTransition> currentOut = getEfsm().getOutgoing(getCurrentState());
      final EfsmTransition initTransition = currentOut.iterator().next();

      if (initTransition == null) {
        throw new RetrascopeException("No init transition has been found.");
      }
      executeAction(initTransition.getAction());
      setCurrentState(initTransition.getDst());
      setLastTraversed(initTransition);
    }

    setLastTraversed(null);
  }

  private void traverse(final EfsmTransition transition) {

    Log.debug("%s traversing the transition %s", getLogEntryHeader(), transition);

    executeAction(transition.getAction());
    setCurrentState(transition.getDst());
    this.coverage.markTransitionAsCovered(transition);
    setLastTraversed(transition);

    Log.debug("%s traversing has been finished successfully", getLogEntryHeader());
  }

  private List<EfsmTransition> getTraversableTransitions() {

    Log.debug("%s performing transitions traversability test", getLogEntryHeader());

    final Collection<EfsmTransition> currentOut = getEfsm().getOutgoing(getCurrentState());
    final List<EfsmTransition> possible = new LinkedList<>(currentOut);

    for (final Iterator<EfsmTransition> possibleIter = possible.iterator();
        possibleIter.hasNext();) {

      final EfsmTransition testedTransition = possibleIter.next();

      Log.debug("%s testing the transition %s", getLogEntryHeader(), testedTransition);

      if (this.currentEvents == null && testedTransition.clockFree()
          || this.currentEvents != null && testedTransition.isSensible(this.currentEvents)) {

        final Node testedConstraint = guardToConstraint(testedTransition.getGuardNode());

        if (SolverUtils.isSat(testedConstraint)) {

          Log.debug("%s the transition is traversable", getLogEntryHeader());

        } else {

          Log.debug("%s the transition isn't traversable now", getLogEntryHeader());

          possibleIter.remove();
        }

      } else {

        Log.debug("%s the transition isn't sensible to the occurred events", this.logEntryHeader);

        possibleIter.remove();
      }
    }

    return possible;
  }

  private EfsmTransition chooseTransitionToTraverse(final List<EfsmTransition> transitions) {

    EfsmTransition traversed;

    if (transitions.size() == 1) {

      traversed = transitions.get(0);

    } else {

      Log.warn("%s there are multiple traversable transitions now. Choosing one of them at random",
          getLogEntryHeader());

      for (final EfsmTransition transition: transitions) {

        Log.warn("%s possible transition: %s", getLogEntryHeader(), transition);
      }

      final Random randomGenerator = new Random(transitions.size());

      traversed =
          transitions.get(randomGenerator.nextInt(transitions.size()));
    }

    return traversed;
  }

  private Node guardToConstraint(final Node guardNode) {

    return this.parent.substituteVariables(guardNode, Collections.emptyMap());
  }

  private void executeAction(final Action action) {
    InvariantChecks.checkNotNull(action);

    Log.debug("%s executing the action: %s", getLogEntryHeader(), action);

    final List<Assignment> assignments = action.getAssignments();

    Map<String, NodeValue> bufferedVariables = new HashMap<>(assignments.size());

    for (final Assignment assignment : assignments) {

      bufferedVariables = this.executeAssignment(assignment, bufferedVariables);
    }

    for (final Entry<String, NodeValue> bufferedVariable : bufferedVariables.entrySet()) {

      final String variableName = bufferedVariable.getKey();
      final NodeValue variableValue = bufferedVariable.getValue();

      this.parent.modifyVariable(this, variableName, variableValue);
    }

    Log.debug("%s the action has been executed successfully: %s", getLogEntryHeader(), action);
  }

  private Map<String, NodeValue> executeAssignment(
      final Assignment assignment,
      final Map<String, NodeValue> bufferedVars) {

    Log.debug("%s executing the assignment: %s", getLogEntryHeader(), assignment);

    final RangedVariable rangedVariable = assignment.getTarget();
    Node value = assignment.getValue();

    final NodeVariable nodeVar = rangedVariable.getVariable();
    final DataType dataType = nodeVar.getDataType();

    value = RangedVariableUtils.transformNodeWithRanged(rangedVariable, value);

    value = this.parent.substituteVariables(value, bufferedVars);
    final NodeValue assignedValue = this.computeExpression(value, dataType);
    bufferedVars.put(rangedVariable.getVariableName(), assignedValue);

    Log.debug("%s the assigned value: %s", this.logEntryHeader, assignedValue);

    return bufferedVars;
  }

  private NodeValue computeExpression(Node expression, final DataType resultDataType) {

    if (expression instanceof NodeValue) {

      final NodeValue value = (NodeValue) expression;

      if (!value.getDataType().equals(resultDataType)) {
        throw new IllegalArgumentException("Incompatible data types: " + value.getDataType()
            + " vs " + resultDataType);
      }

      return value;
    } else {
      if (!ExprUtils.getVariables(expression).isEmpty()) {
        throw new IllegalArgumentException("The expression to be computed (" + expression
            + ") contains unevaluated variables: " + ExprUtils.getVariables(expression));
      }

      Log.debug("%s computing the expression %s", getLogEntryHeader(), expression);

      expression = Nodes.eq(new NodeVariable("assignmentResult", resultDataType), expression);

      final SolverResult result = SolverUtils.solve(expression);

      if (result.getStatus().equals(Status.SAT)) {

        final NodeValue resultValue =
            new NodeValue(result.getVariables().iterator().next().getData());

        Log.debug("%s the computation result: %s", getLogEntryHeader(), resultValue);

        return resultValue;
      } else {
        throw new RetrascopeException(
            "Failed to resolve the assignment constraint due to errors: " + result.getErrors());
      }
    }
  }

  private void logInnerDump() {

    Log.debug("%s current state: %s", getLogEntryHeader(), getCurrentState());
    Log.debug("%s inner variables' dump: %s", getLogEntryHeader(), getRegMemory());
    Log.debug("%s output dump: %s", getLogEntryHeader(), getOutputsMemory());
    Log.debug("%s current coverage: %s", getLogEntryHeader(), getCoverage());
  }

  @Override
  public TestVector simulateVector(final TestVector inputVector) {

    Log.debug("%s processing vector: %s", getLogEntryHeader(), inputVector);

    this.parent.readVector(inputVector);

    if (inputVector.hasEvents()) {
      this.currentEvents = inputVector.getEventList();
    }

    final List<EfsmTransition> traversable = getTraversableTransitions();

    if (traversable.isEmpty()) {

      Log.debug("%s there are no transitions to traverse. Finishing event handling",
          getLogEntryHeader());

      setLastTraversed(null);

    } else {
      this.traverse(this.chooseTransitionToTraverse(traversable));
    }

    final List<EfsmTransition> enabled = getEnabledTransitions();

    while (!enabled.isEmpty()) {

      this.traverse(this.chooseTransitionToTraverse(enabled));
      enabled.clear();
      enabled.addAll(getEnabledTransitions());
    }

    this.logInnerDump();

    Log.debug("%s model has finished to process the events", this.logEntryHeader);

    return this.parent.getOutputVector();
  }

  private List<EfsmTransition> getEnabledTransitions() {

    final Integer currentPhase = this.currentState.getPhase();
    if (currentPhase == 0) {
      this.currentEvents = null;
    }

    final List<EfsmTransition> enabled = new LinkedList<>();

    final Collection<EfsmTransition> possible = getEfsm().getOutgoing(this.currentState);

    for (final EfsmTransition transition : possible) {

      Log.debug("%s testing the transition %s", getLogEntryHeader(), transition);

      boolean eventTriggered = false;

      if (!transition.clockFree()) {

        if (this.currentEvents == null) {
          break;
        }

        if (transition.isSensible(this.currentEvents)) {
          eventTriggered = true;
        }
      }

      if (transition.clockFree() || (eventTriggered && this.currentEvents != null)) {

        final Node testedConstraint = guardToConstraint(transition.getGuardNode());

        if (SolverUtils.isSat(testedConstraint)) {

          Log.debug("%s the transition is traversable", getLogEntryHeader());
          enabled.add(transition);

        } else {

          Log.debug("%s the transition isn't traversable now", getLogEntryHeader());
        }
      }
    }

    return enabled;
  }

  @Override
  public Sequence simulateSequence(final Sequence inputSequence) {

    throw new UnsupportedOperationException("simulateSequence(Sequence) isn't implemented yet");
  }

  /**
   * Generates an input vector to enable the specified guarded action considering the specified
   * fixed events and inputs.
   *
   * @param guardedAction the guarded action to be enabled
   * @param fixedEvents the fixed events
   * @param fixedInputs the fixed inputs
   *
   * @return the generated vector; if one failed to enable the guarded action, {@code null} is
   *     returned
   */
  public TestVector enableGuardedAction(
      final GuardedAction guardedAction,
      final EventList fixedEvents,
      final Map<String, NodeValue> fixedInputs) {

    if (!fixedEvents.isEmpty()) {

      final EventList gaEventList = new EventList();
      gaEventList.addEvents(guardedAction.getClocks());

      if (!gaEventList.compatibleWith(fixedEvents)) {
        return null;
      }
    }

    final Node guardExpression = this.fillExpression(guardedAction.getGuardNode(), fixedInputs);

    final SolverResult solverResult = SolverUtils.solve(guardExpression);

    if (solverResult.getStatus().equals(Status.SAT)) {

      final TestVector inputVector = new TestVector();
      final Transaction transaction = new Transaction();

      for (final Variable variable : solverResult.getVariables()) {

        final String variableName = variable.getName();

        if (this.efsm.getParentContainer().declaresInput(variableName)) {

          transaction.addVariable(variableName, new NodeValue(variable.getData()));
        }
      }

      inputVector.addEvents(guardedAction.getClocks());

      if (!transaction.isEmpty()) {
        inputVector.addTransaction(transaction);
      }

      inputVector.addEvents(fixedEvents);

      if (!fixedInputs.isEmpty()) {
        inputVector.addTransaction(new Transaction(fixedInputs));
      }

      return inputVector;
    } else {
      return null;
    }
  }

  /**
   * Fills the specified expression with the specified inputs and current values of inner variables.
   *
   * @param expression the expression to be filled
   * @param fixedInputs the fixed inputs
   *
   * @return an expression where variables replaced by their values
   */
  public Node fillExpression(Node expression, final Map<String, NodeValue> fixedInputs) {

    final Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    final List<BoundVariable> bindingList = new LinkedList<>();
    final Set<Node> inputConstraints = new HashSet<>();

    for (final NodeVariable usedVar: usedVariables) {

      final String varName = Substitutor.deannotateName(usedVar.getName());

      if (this.efsm.getParentContainer().declaresInput(varName)) {

        final Node varConstraint = this.efsm.getParentContainer().getInvariant(varName);

        if (varConstraint != null) {

          inputConstraints.add(Transformer.substitute(varConstraint, varName, usedVar));
        }

        if (fixedInputs.containsKey(varName)) {

          bindingList.add(NodeBinding.bindVariable(usedVar, fixedInputs.get(varName)));
        }
      } else if (this.efsm.getParentContainer().declaresRegister(varName)) {

        bindingList.add(NodeBinding.bindVariable(usedVar, this.parent.getVariableValue(varName)));
      }
    }

    expression = Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

    if (!inputConstraints.isEmpty()) {
      expression = Nodes.and(expression, Nodes.and(inputConstraints));
    }

    return expression;
  }

  @Override
  public void resetCoverage() {
    this.coverage.reset();
  }

  /**
   * A process simulation snapshot.
   */
  class EfsmSimulationSnapshot extends SimulationSnapshot {

    private final EfsmState currentState;
    private final EfsmTransition lastTraversed;
    private final EfsmTransitionCoverage coverage;

    EfsmSimulationSnapshot(final EfsmSimulator simulator) {
      super(simulator);

      this.currentState = EfsmSimulator.this.currentState;
      this.lastTraversed = EfsmSimulator.this.lastTraversed;
      this.coverage = new EfsmTransitionCoverage(EfsmSimulator.this.coverage);
    }
  }

  public EfsmSimulationSnapshot makeSimulationSnapshot() {
    return new EfsmSimulationSnapshot(this);
  }

  /**
   * Loads the specified simulation snapshot discarding current simulation state and history.
   *
   * @param snapshot the snapshot
   */
  void loadSnapshot(final EfsmSimulationSnapshot snapshot) {

    Log.debug("%s loading a simulation snapshot", this.logEntryHeader);

    this.currentState = snapshot.currentState;
    this.lastTraversed = snapshot.lastTraversed;
    this.coverage.reset();
    this.coverage.include(snapshot.coverage);

    this.parent.loadSnapshot(snapshot);
  }
}
