/*
 * Copyright 2016-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.efsm;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.Backend;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmDfsWalker;
import ru.ispras.retrascope.model.efsm.EfsmEmptyVisitor;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Backend to EFSM transformer that applies heuristics for initial state detection.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class EfsmInitStateBackend extends Backend {

  private final InitStateDetectorVisitor visitor = new InitStateDetectorVisitor();

  private static final String CANT_FIND_RST = "Can't find resetting transition";

  /**
   * Default constructor.
   */
  EfsmInitStateBackend() {
    super("efsm_init_state");
  }

  static class InitStateDetectorVisitor extends EfsmEmptyVisitor {

    private Map<String, ModuleProcess> efsmCfgProcessMap;

    private ModuleProcess process;
    private NodeVariable phaseVariable;

    /**
     * Sets mapping of EFSM names to related CFG processes.
     *
     * @param map EFSM names mapping to control flow graph processes.
     */
    void setEfsmCfgProcessMap(final Map<String, ModuleProcess> map) {
      this.efsmCfgProcessMap = map;
    }

    private NodeVariable getPhaseVariable() {
      return this.phaseVariable;
    }

    @Override
    public void onProcessBegin(final ModuleProcess efsm) {

      /* NOTE: supposing that EFSM name is equal to CFG process name. */
      this.process = this.efsmCfgProcessMap.get(efsm.getName());

      if (efsm.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
        final String phaseVarName = efsm.getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);
        this.phaseVariable = ((DecStatement)efsm.getParent()).getVariable(phaseVarName);
      }

      getInitialState((Efsm) efsm);
    }

    private void getInitialState(final Efsm efsm) {

      final Map<NodeVariable, Node> initValueMap = getInitialOrZeroValues(this.process);

      final List<NodeBinding.BoundVariable> initValuePairs = createInitState(initValueMap);
      final Collection<EfsmState> efsmStates = efsm.getStates();

      final Integer phaseValue = getPhaseValue(initValueMap);

      final EfsmState initState = getEfsmInitState(initValuePairs, phaseValue, efsmStates);

      if (initState == null) {

        efsm.addState(Efsm.UNINIT_STATE);
        Log.warn("Can't find initial state, an auxiliary is selected: %s", Efsm.UNINIT_STATE);
        efsm.setInitialState(Efsm.UNINIT_STATE);

      } else {

        Log.info("The initial state for %s: %s", efsm.getName(), initState);
        efsm.setInitialState(initState);
      }

      final EfsmRstVisitor rstVisitor = new EfsmRstVisitor(efsm);
      final EfsmDfsWalker walker = new EfsmDfsWalker(rstVisitor);
      walker.start(efsm);

      final Deque<EfsmTransition> rstTransitions = rstVisitor.getRstTransitions();

      if (rstTransitions.isEmpty()) {

        /* Create a transition from uninitialized state to some compliant state.
         * The transition should use only input signals and define all the internal variables. */

        final EfsmTransition inputOnlyUseTrans = getDefAllFromInit(efsm);

        if (inputOnlyUseTrans == null) {

          efsm.remove(Efsm.UNINIT_STATE);
          final EfsmState someState = efsmStates.iterator().next();
          efsm.setInitialState(someState);
          efsm.setSingleTestable(false);

          Log.warn("%s for %s", CANT_FIND_RST, efsm.getName());
          Log.warn("The extracted EFSM is not single testable: " + efsm.getName());
          Log.warn("The initial state is selected arbitrarily: " + someState);

        } else {

          final GuardedAction guardedAction = new GuardedAction(inputOnlyUseTrans);

          efsm.addTransition(Efsm.UNINIT_STATE, inputOnlyUseTrans.getDst(), guardedAction);
        }

      } else {

        for (final EfsmTransition rstTransition : rstTransitions) {

          efsm.addResetGuardedAction(rstTransition);
        }
      }
      Log.info("======================================");
    }

    private Integer getPhaseValue(final Map<NodeVariable, Node> valueMap) {

      final Node value = valueMap.get(getPhaseVariable());
      final Node toInt = TypeConversion.coerce(value, DataType.INTEGER);
      final NodeValue toIntValue = (NodeValue) toInt;
      final BigInteger bigInt = toIntValue.getInteger();

      return bigInt.intValue();
    }

    private EfsmTransition getDefAllFromInit(final Efsm efsm) {

      for (final EfsmTransition transition : efsm.getTransitions()) {

        final Set<NodeVariable> uses = transition.getUses();

        boolean allUsesAreInputs = true;

        for (final NodeVariable use : uses) {

          final Object data = use.getUserData();

          if (data instanceof VariableDescriptor) {

            /* VHDL case. */
            final VariableType type = ((VariableDescriptor) data).getVariableType();

            if (type != null && !VariableType.isInputSignal(type)) {

              allUsesAreInputs = false;
              break;
            }
          } else if (data instanceof VerilogReference) {

            /* Verilog case.*/
            final VerilogDeclaration.Type type =
                ((VerilogReference) data).getDeclaration().getType();

            if (VerilogDeclaration.Type.INPUT != type
                && VerilogDeclaration.Type.INOUT != type) {

              allUsesAreInputs = false;
              break;
            }
          }
        }
        if (allUsesAreInputs) {

          final Set<NodeVariable> transitionDefines = transition.getDefines();
          final boolean definesAll = transitionDefines.equals(efsm.getDefines());

          return definesAll ? transition : null;
        }
      }
      return null;
    }

    private EfsmState getEfsmInitState(
        final List<NodeBinding.BoundVariable> boundVariables,
        final Integer phaseValue,
        final Collection<EfsmState> efsmStates) {

      EfsmState initialState = null;

      for (final EfsmState state : efsmStates) {

        final Integer phase = state.getPhase();
        if (!phase.equals(phaseValue)) {
          continue;
        }

        final NodeBinding bindNode = new NodeBinding(state.getNode(), boundVariables);
        final Node conjunction = Transformer.substituteAllBindings(bindNode);

        if (SolverUtils.isSat(conjunction)) {

          initialState = state;
          break;
        }
      }
      return initialState;
    }

    private Map<NodeVariable, Node> getInitialOrZeroValues(final ModuleProcess process) {

      final Map<NodeVariable, Node> initValueMap = new LinkedHashMap<>();

      Statement container = process.getParent();

      do {

        final Map<NodeVariable, Declaration> varMap =
            ((DecStatement) container).getDeclarationMap();

        for (final Map.Entry<NodeVariable, Declaration> entry : varMap.entrySet()) {

          final NodeVariable var = entry.getKey();
          final Declaration declaration = entry.getValue();

          if (!initValueMap.containsKey(var)) {

            final Node value =
                declaration.containsInitialValue() ? declaration.getInitialValue() : makeZero(var);

            initValueMap.put(var, value);
          }
        }

        container = container.getParent();
      } while (container != null);

      return initValueMap;
    }

    private static Node makeZero(final NodeVariable var) {

      final DataType dataType = var.getDataType();
      final DataTypeId typeId = dataType.getTypeId();

      switch (typeId) {

        case BIT_VECTOR:
          return NodeValue.newBitVector(0, dataType.getSize());

        case LOGIC_BOOLEAN:
          return NodeValue.newBoolean(false);

        case LOGIC_INTEGER:
          return NodeValue.newInteger(0);

        case LOGIC_REAL:
          return NodeValue.newReal(0);

        case MAP:
          return NodeValue.newMap(createZeroMap(var));

        default:
          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE,
              String.format("datatype=%s, var=%s", dataType, var));
      }
    }

    private static DataMap createZeroMap(final NodeVariable var) {

      final DataType type = var.getDataType();
      final DataType keyType = (DataType) type.getAttribute(DataTypeId.Attribute.KEY);
      final DataType valueType = (DataType) type.getAttribute(DataTypeId.Attribute.VALUE);

      final DataMap map = new DataMap(keyType, valueType);
      for (int i = 0; i < type.getSize(); i++) {
        map.put(makeValue(i, keyType), makeValue(0, valueType));
      }
      return map;
    }

    private static Data makeValue(final int num, final DataType type) {

      final DataTypeId typeId = type.getTypeId();

      switch (typeId) {

        case BIT_VECTOR:
          return Data.newBitVector(num, type.getSize());

        case LOGIC_INTEGER:
          return Data.newInteger(num);

        default:
          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, typeId);
      }
    }

    private List<NodeBinding.BoundVariable> createInitState(final Map<NodeVariable, Node> map) {

      final List<NodeBinding.BoundVariable> boundVars = new LinkedList<>();

      for (final Map.Entry<NodeVariable, Node> entry : map.entrySet()) {
        boundVars.add(NodeBinding.bindVariable(entry.getKey(), entry.getValue()));
      }

      return boundVars;
    }
  }

  /**
   * Sets mapping of EFSM names to related CFG processes.
   *
   * @param map EFSM names mapping to control flow graph processes.
   */
  void setEfsmCfgProcessMap(final Map<String, ModuleProcess> map) {
    this.visitor.setEfsmCfgProcessMap(map);
  }

  @Override
  public void start(final Entity entity) {

    Log.info("======================================");
    final EfsmModel efsmModel = (EfsmModel) entity;

    final ModelWalker modelWalker = new EfsmDfsWalker(this.visitor);
    modelWalker.start(efsmModel);
  }
}
