/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.efsm;

import edu.uci.ics.jung.graph.Graph;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.basis.ModelWalkerEngine;
import ru.ispras.retrascope.engine.cfg.transformer.gadd.PhaseSeqEdge;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.model.gadd.Phase;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeOrthogonalizer;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.io.File;
import java.math.BigInteger;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The transformer of {@link GaddModel} model into {@link EfsmModel} model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class GaddEfsmTransformer extends ModelWalkerEngine {

  private static final String DEBUG_PRINT_STAT_FMT =
      "engine=%s modules=%s states=%d transitions=%d efsms=%d";

  /* The model that is to be transformed into EFSM. */
  private GaddModel gaddModel;

  /* Map from GADD model processes to their phase sequence graphs. */
  private Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processToGraph;

  /* The state-like variables extractor. */
  private GaddStateVarDetector stateDetector;

  /* The state-like variables statistics keeping visitor. */
  private GaddStatisticsKeeper statKeeper;

  /* The states expressions' extractor. */
  private GaddPathExtractor pathExtractor;

  /* Auxiliary mapping between EFSM names and processes that are extracted from. */
  private final Map<String, ModuleProcess> efsmProcessMap = new LinkedHashMap<>();

  /* Back-ends. */
  private final EfsmInitStateBackend initStateBackend;

  /* Maximum number of EFSM states. */
  private int stateMaxNum;

  /* Model component that is under construction. */
  private Efsm efsm;

  /* States to be checked as possible sources for transition under construction. */
  private LinkedList<EfsmState> canBeSrcStates;

  private Set<EfsmState> visited;

  private final Map<Integer, LinkedHashSet<GaddPath>> srcPathMap;

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "gadd-efsm-transformer";

  private static final String STATE_VAR_INFO = "Names of state-like variables";

  /**
   * State-like variables names command line parameter.
   */
  public static final Parameter STATE_VARIABLES = new Parameter("state-vars", true, STATE_VAR_INFO);

  private static final String STATE_MAX_INFO = "Maximum number of EFSM states.";

  /**
   * State maximum number parameter.
   */
  public static final Parameter STATE_NUM =
      new Parameter("state-max-num", true, STATE_MAX_INFO);

  private static final int STATE_MAX_NUM = -1;

  static class StateVariableContainer {

    final Map<NodeVariable, Integer> stateNumMap;
    final int stateMaxNum;
    final boolean complete;

    StateVariableContainer(
        final Map<NodeVariable, Integer> stateNumMap,
        final int stateMaxNum,
        final boolean complete) {
      this.stateNumMap = stateNumMap;
      this.stateMaxNum = stateMaxNum;
      this.complete = complete;
    }

    Integer getNum(final NodeVariable variable) {
      return this.stateNumMap.get(variable);
    }

    int getStateMaxNum() {
      return this.stateMaxNum;
    }

    boolean hasVariable(final NodeVariable variable) {
      return this.stateNumMap.containsKey(variable);
    }

    boolean isComplete() {
      return this.complete;
    }

    @Override public String toString() {
      return "StateVariableContainer{" + "stateNumMap=" + stateNumMap + ", stateMaxNum="
          + stateMaxNum + ", complete=" + complete + '}';
    }
  }

  /**
   * Default constructor.
   */
  public GaddEfsmTransformer() {
    super(
        GaddEfsmTransformer.ENGINE_ID,
        GaddModel.TYPE,
        EfsmModel.TYPE,
        new CfgDfsWalker(new GaddStateVarDetector(), WalkerMode.DFS),
        new CfgDfsWalker(new GaddStatisticsKeeper(), WalkerMode.DFS),
        new CfgDfsWalker(new GaddPathExtractor(), WalkerMode.DFS));

    /* Engine command line parameters. */
    addParameter(Retrascope.TOOL_DEBUG_PARAM);
    addParameter(STATE_VARIABLES);
    addParameter(STATE_NUM);

    this.srcPathMap = new LinkedHashMap<>();

    /* Back-end registration. */
    this.initStateBackend = new EfsmInitStateBackend();
    getInitStateBackend().setEfsmCfgProcessMap(this.efsmProcessMap);
    addBackend(getInitStateBackend());
  }

  private EfsmInitStateBackend getInitStateBackend() {
    return this.initStateBackend;
  }

  private CfgModel getCfgModel() {
    return this.gaddModel.getModel();
  }

  private void setGaddModel(final GaddModel model) {
    InvariantChecks.checkNotNull(model);

    this.gaddModel = model;
  }

  private GaddStateVarDetector getStateDetector() {
    return this.stateDetector;
  }

  private void setStateDetector(final GaddStateVarDetector stateDetector) {
    InvariantChecks.checkNotNull(stateDetector);

    this.stateDetector = stateDetector;
  }

  private GaddStatisticsKeeper getStatKeeper() {
    return this.statKeeper;
  }

  private void setStatKeeper(final GaddStatisticsKeeper statKeeper) {
    InvariantChecks.checkNotNull(statKeeper);

    this.statKeeper = statKeeper;
  }

  private GaddPathExtractor getPathExtractor() {
    return this.pathExtractor;
  }

  private void setPathExtractor(final GaddPathExtractor pathExtractor) {
    InvariantChecks.checkNotNull(pathExtractor);

    this.pathExtractor = pathExtractor;
  }

  private int getStateMaxNum() {
    return this.stateMaxNum;
  }

  private void setStateMaxNum(final int num) {
    this.stateMaxNum = num;
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    /* Extract control flow graph from wrapper. */

    final GaddModel gaddModel = (GaddModel) inputs.get(GaddModel.TYPE);
    inputs.remove(GaddModel.TYPE);
    setGaddModel(gaddModel);

    this.processToGraph = gaddModel.getProcessPhaseMap();

    /* treat internal model as a GADD. */
    inputs.put(GaddModel.TYPE, getCfgModel());

    initWalkers();

    /* Elaborate command line parameters. */
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      if (engineArgs.contains(STATE_VARIABLES)) {

        final String[] stateNames = engineArgs.getValue(STATE_VARIABLES).split(File.pathSeparator);
        getStateDetector().setStateNames(stateNames);
      }

      setStateMaxNum(
          Integer.parseInt(
              engineArgs.getValue(STATE_NUM, Integer.toString(STATE_MAX_NUM))));

    } catch (final ParseException e) {
      throw new RetrascopeException(e);
    }
  }

  private void initWalkers() {

    setStateDetector(new GaddStateVarDetector());
    setStatKeeper(new GaddStatisticsKeeper());
    setPathExtractor(new GaddPathExtractor());

    setWalkers(new CfgDfsWalker(getStateDetector(), WalkerMode.DFS),
        new CfgDfsWalker(getStatKeeper(), WalkerMode.DFS),
        new CfgDfsWalker(getPathExtractor(), WalkerMode.DFS_NO_RPT));

    /* Create mappings for internal components. */
    createClockMap();
    createStateMap();

    getStateDetector().setPhaseSeqMap(this.processToGraph);
  }

  private void createStateMap() {

    final Map<ModuleProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getStateDetector().setStateMap(map);
    getStatKeeper().setStateMap(map);
    getPathExtractor().setStateMap(map);
  }

  private void createClockMap() {

    final Map<ModuleProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getPathExtractor().setClockMap(map);
  }

  @Override
  public Entity getOutput() {

    final EfsmModel model = new EfsmModel();

    /* Create GADD model visitor that transforms modules. */
    final GaddModuleVisitor moduleVisitor = new GaddModuleVisitor(model);

    final ModelWalker walker = new CfgDfsWalker(moduleVisitor, WalkerMode.DFS_NO_RPT);
    walker.start(getCfgModel());

    final Map<CfgModule, EfsmModule> moduleEfsmMap = moduleVisitor.getModuleMap();

    final Map<ModuleProcess, Map<Integer, LinkedHashSet<GaddPath>>> pathMap =
        getPathExtractor().getProcessDataMap();

    final Map<ModuleProcess, Set<NodeVariable>> stateVariableMap =
        getStateDetector().getStateLikeVarMap();

    Log.info("======================================");
    int efsmCount = 0;
    int efsmStateNum = 0;
    int efsmTransNum = 0;

    for (final Map.Entry<ModuleProcess, Map<Integer, LinkedHashSet<GaddPath>>> entry
        : pathMap.entrySet()) {

      final ModuleProcess keyProcess = entry.getKey();
      final Statement keyParent = keyProcess.getParent();

      /* Check that Process parent is Module object. */
      assert keyParent instanceof CfgModule
          : LogMessage.makeMessage(LogMessage.WRONG_TYPE, keyParent.getClass().getName());

      final EfsmModule parentContainer = moduleEfsmMap.get(keyParent);

      Log.info("Transforming the process '%s'.", keyProcess.getName());

      this.srcPathMap.putAll(entry.getValue());

      this.efsm = new Efsm(parentContainer, keyProcess.getId());
      this.efsm.addMetaInfo(
          MetaInfoType.PHASE_VAR_NAME, keyProcess.getMetaInfo(MetaInfoType.PHASE_VAR_NAME.name()));

      this.efsmProcessMap.put(this.efsm.getName(), keyProcess);

      if (this.gaddModel.hasClock(((CfgModule) keyParent).getName())) {

        final NodeVariable moduleClock = this.gaddModel.getClock(((CfgModule) keyParent).getName());
        if (keyProcess.hasEvent(moduleClock)) {
          this.efsm.addClock(moduleClock);
        }
      }

      final Set<NodeVariable> stateVariables = stateVariableMap.get(keyProcess);

      createEfsm(stateVariables);

      if (keyProcess.isInitial()) {
        parentContainer.addInitialEfsm(this.efsm);
      } else {
        parentContainer.addNonInitialEfsm(this.efsm);
      }

      removeIsolatedStates();

      Log.info("Model states number: %d.", this.efsm.getStateNum());

      Log.info("Model transitions number: %d.", this.efsm.getTransitionNum());
      Log.info("======================================");

      efsmCount++;
      efsmStateNum += this.efsm.getStateNum();
      efsmTransNum += this.efsm.getTransitionNum();
      this.srcPathMap.clear();
    }
    Log.info("The number of extracted models: %d.", efsmCount);
    Log.info("The total number of states: %d.", efsmStateNum);
    Log.info("The total number of transitions: %d.", efsmTransNum);

    /* If debug log mode is enabled, collect & write statistical info to log file. */
    if (isDebugMode()) {

      final EfsmDebugInfo debugInfo = new EfsmDebugInfo(0, 0, 0);

      final List<String> topNames = new LinkedList<>();

      for (final Module efsmModule : model.getModules()) {

        getModuleInfo(efsmModule, debugInfo);
        topNames.add(efsmModule.getName());
      }
      debugWrite(
          String.format(
              DEBUG_PRINT_STAT_FMT,
              getId(),
              StringUtils.join(topNames, File.pathSeparator),
              debugInfo.getStateNum(),
              debugInfo.getTransitionNum(),
              debugInfo.getModelNum()));
    }

    return model;
  }

  private void removeIsolatedStates() {

    final Set<EfsmState> isolated = new LinkedHashSet<>();

    for (final EfsmState state : this.efsm.getStates()) {

      final Collection<EfsmTransition> incoming = this.efsm.getIncoming(state);
      final Collection<EfsmTransition> outgoing = this.efsm.getOutgoing(state);

      if (incoming.isEmpty() && outgoing.isEmpty()) {
        isolated.add(state);
      }
    }
    for (final EfsmState isolatedState : isolated) {
      this.efsm.remove(isolatedState);
    }
  }

  private void getModuleInfo(final Module efsmModule, final EfsmDebugInfo info) {

    final Collection<Module> nestedModules = efsmModule.getModules();
    for (final Module subModule : nestedModules) {
      getModuleInfo(subModule, info);
    }

    final Set<ModuleProcess> efsms = new LinkedHashSet<>(efsmModule.getProcesses());

    for (final ModuleProcess efsm : efsms) {

      info.setModelNum(info.getModelNum() + 1);
      info.setStateNum(info.getStateNum() + ((Efsm)efsm).getStateNum());
      info.setTransitionNum(info.getTransitionNum() + ((Efsm)efsm).getTransitionNum());
    }
  }

  private void createEfsm(final Set<NodeVariable> stateVariables) {

    debugWrite("EFSM building...");
    final List<Node> states = createStates(stateVariables);

    if (states.isEmpty()) {
      states.add(Nodes.TRUE);
    }

    printStateVarInfo(states);
    createTransitions();
    debugWrite("done.");
  }

  private void printStateVarInfo(final List<Node> states) {

    final Collection<String> names = new LinkedHashSet<>();

    for (final Node state : states) {

      this.efsm.addState(state);
      names.addAll(NodeUtils.getVariableNames(ExprUtils.getVariables(state)));
    }

    Log.info("State-like variables: %s.", names.isEmpty() ? "<none>" : names);
  }

  private void createTransitions() {

    /* Copy all the model states into temporary collection. */
    this.canBeSrcStates = new LinkedList<>(this.efsm.getStates());

    /* Sort states for deterministic behaviour. */
    Collections.sort(this.canBeSrcStates, Comparator.comparing(EfsmState::toString));

    /* Collection for states that are processed already. */
    this.visited = new LinkedHashSet<>();

    while (!this.canBeSrcStates.isEmpty()) {

      /* Take next state from collection. */
      final EfsmState src = this.canBeSrcStates.poll();

      this.visited.add(src);

      final Integer srcPhase = src.getPhase();

      if (this.srcPathMap.containsKey(srcPhase)) {

        final LinkedHashSet<GaddPath> phasePaths = this.srcPathMap.get(srcPhase);

        createTransitionsFrom(src, phasePaths);

      } else {

        Log.warn("No execution GADD paths for phase='%s'", srcPhase);
      }
    }
    this.visited.clear();
  }

  private void createTransitionsFrom(final EfsmState srcState, final Set<GaddPath> paths) {

    for (final GaddPath path : paths) {

      /* When the stabilization splits the state, skip the rest. */
      if (!this.efsm.containsState(srcState)) {
        break;
      }

      final Node preCond = Nodes.and(path.getState());
      final Node pathGuardNode = Nodes.and(path.getGuardNode(), preCond);
      final Node clockNode = path.getClockNode();

      final Node srcNode = srcState.getNode();
      final Node guardSatCond = getNodesWithInvariants(pathGuardNode, clockNode, srcNode);
      final boolean guardSat = SolverUtils.isSat(guardSatCond);

      if (!guardSat) {
        continue;
      }

      addTransitionsFromSrc(srcState, path);
    }
  }

  private Node getNodesWithInvariants(final Node pathGuard, final Node clock, final Node srcState) {

    return clock == null
        ? ExprUtils.getConjunction(pathGuard, srcState, getNodeVarInvariants(pathGuard, srcState))
        : ExprUtils.getConjunction(
            clock,
            pathGuard,
            srcState,
            getNodeVarInvariants(clock, pathGuard, srcState));
  }

  private void addTransitionsFromSrc(final EfsmState src, final GaddPath path) {

    final GuardedAction guardedAction = path.getGuardedAction();

    final Action action = guardedAction.getAction();
    final Set<EfsmState> dstStates = getDstStates(src, path);

    if (dstStates.size() > 1) {

      for (final EfsmState dst : dstStates) {

        final CfgBlockStatement block = action.toBlock();
        final Node wpCond = CfgUtils.getWeakestPreCondition(dst.getNode(), block);
        final Node guard =
            Nodes.and(guardedAction.getGuardNode(), TransformerUtils.standardize(wpCond));

        final GuardedAction transitionData =
            new GuardedAction(guard, action, guardedAction.getClocks());
        this.efsm.addTransition(src, dst, transitionData);
      }

    } else if (dstStates.size() == 1) {
      final EfsmState dst = dstStates.iterator().next();
      this.efsm.addTransition(src, dst, guardedAction);
    }
  }

  private List<Node> createStates(final Set<NodeVariable> stateVariables) {

    final StateVariableContainer container = createStateContainer(stateVariables);

    if (container.isComplete()
        && getStateMaxNum() > 0
        && container.getStateMaxNum() <= getStateMaxNum()) {

      final List<List<Node>> states = new LinkedList<>();

      for (final NodeVariable stateVariable : stateVariables) {

        if (container.hasVariable(stateVariable)) {
          final Integer varValueNum = container.getNum(stateVariable);

          states.add(createVarEquations(stateVariable, varValueNum));
        }
      }

      return conjunctions(cartesianProduct(states));

    } else {

      final Collection<GaddPath> gaddPaths = new LinkedHashSet<>();
      final Collection<LinkedHashSet<GaddPath>> paths = this.srcPathMap.values();

      for (final Set<GaddPath> pathSet : paths) {
        gaddPaths.addAll(pathSet);
      }

      final Map<NodeVariable, Collection<Collection<Node>>> stateVarNodes = splitByVars(gaddPaths);
      final List<List<Node>> ortNodes = new LinkedList<>();

      /* Sort state predicate collections for determinism. */
      final LinkedList<Collection<Collection<Node>>> preStates =
          new LinkedList<>(stateVarNodes.values());
      preStates.sort(Comparator.comparing(l -> l.iterator().next().iterator().next().toString()));

      for (final Collection<Collection<Node>> nodes : stateVarNodes.values()) {
        ortNodes.add(NodeOrthogonalizer.orthogonalizeNodes(nodes));
      }

      return conjunctions(cartesianProduct(ortNodes));
    }
  }

  private List<Node> createVarEquations(final NodeVariable variable, final Integer valueNum) {

    final List<Node> equations = new LinkedList<>();

    for (int i = 0; i < valueNum; i++) {

      final Node value = TypeConversion.coerce(NodeValue.newInteger(i), variable.getDataType());
      equations.add(Nodes.eq(variable, value));
    }

    return equations;
  }

  private StateVariableContainer createStateContainer(final Set<NodeVariable> stateVariables) {

    final Map<NodeVariable, Integer> valueNumMap = new LinkedHashMap<>();
    int stateMaxNum = 1;
    boolean complete = true;

    for (final NodeVariable variable : stateVariables) {

      final Statement topParent = StatementUtils.getTopParent(this.efsm);
      final DecStatement decStatement = StatementUtils.getVariableContainer(topParent, variable);
      final Declaration declaration = decStatement.getDeclaration(variable);

      if (declaration.containsInvariant()) {

        final Range invariant = declaration.getInvariant();
        final Node maxNode = invariant.getHigh();
        final Node minNode = invariant.getLow();

        if (ExprUtils.isKind(Node.Kind.VALUE, maxNode, minNode)) {

          final NodeValue maxValue = (NodeValue) maxNode;
          final NodeValue minValue = (NodeValue) minNode;

          final BigInteger maxInt = maxValue.getInteger();
          final BigInteger minInt = minValue.getInteger();

          final int valueNum = getValueNum(maxInt, minInt);
          valueNumMap.put(variable, valueNum);

          stateMaxNum = stateMaxNum * valueNum;
        }
      } else if (constrainedDataType(variable)) {

        final DataType dataType = variable.getDataType();
        final BigInteger maxInt = getMaxInt(dataType);
        final BigInteger minInt = getMinInt(dataType);

        final int valueNum = getValueNum(maxInt, minInt);
        valueNumMap.put(variable, valueNum);

        stateMaxNum = stateMaxNum * valueNum;
      } else {
        complete = false;
        break;
      }
    }

    return new StateVariableContainer(valueNumMap, stateMaxNum, complete);
  }

  private int getValueNum(final BigInteger max, final BigInteger min) {

    return (max.subtract(min)).intValue() + 1;
  }

  private BigInteger getMinInt(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return BigInteger.ZERO;

      case BIT_VECTOR:
        final BitVector minVector = BitVector.newEmpty(dataType.getSize());
        minVector.reset();
        return new BigInteger(minVector.toByteArray());

      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private BigInteger getMaxInt(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return BigInteger.ONE;

      case BIT_VECTOR:
        final BitVector minVector = BitVector.newEmpty(dataType.getSize());
        minVector.setAll();
        return new BigInteger(minVector.toByteArray());

      default:
        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private boolean constrainedDataType(final NodeVariable variable) {

    return variable.isType(DataType.BOOLEAN) || variable.isType(DataTypeId.BIT_VECTOR);
  }

  private List<Node> conjunctions(final List<List<Node>> lists) {

    final List<Node> conjunctions = new LinkedList<>();

    for (final List<Node> list : lists) {

      if (!list.isEmpty()) {

        final Node conjunction = Nodes.and(list);

        if (SolverUtils.areCompatible(conjunction, getNodeVarInvariants(conjunction))) {
          conjunctions.add(conjunction);
        }
      }
    }

    return conjunctions;
  }

  private List<List<Node>> cartesianProduct(final List<List<Node>> lists) {
    final List<List<Node>> resultLists = new LinkedList<>();

    if (lists.isEmpty()) {

      resultLists.add(new LinkedList<>());
      return resultLists;

    } else {

      final List<Node> firstList = lists.get(0);
      final List<List<Node>> remainingLists = cartesianProduct(lists.subList(1, lists.size()));

      for (final Node firstListElem : firstList) {

        for (final List<Node> remainingList : remainingLists) {

          final List<Node> resultList = new LinkedList<>();
          resultList.add(firstListElem);
          resultList.addAll(remainingList);
          resultLists.add(resultList);
        }
      }
    }
    return resultLists;
  }

  private Map<NodeVariable, Collection<Collection<Node>>> splitByVars(
      final Collection<GaddPath> paths) {

    final Map<NodeVariable, Collection<Collection<Node>>> varNodeMap = new LinkedHashMap<>();

    for (final GaddPath path : paths) {

      final Set<Node> state = path.getState();

      /* Mapping from variables to nodes that are dependent from them. */
      final Map<NodeVariable, Collection<Node>> varDepNodeMap = new LinkedHashMap<>();

      for (final Node node : state) {

        final Collection<NodeVariable> variables = ExprUtils.getVariables(node);

        if (variables.isEmpty()) {
          continue;
        }

        /* It is supposed that every node of the state depends on exactly one variable. */
        InvariantChecks.checkTrue(variables.size() == 1);
        final NodeVariable variable = variables.iterator().next();

        if (varDepNodeMap.containsKey(variable)) {

          varDepNodeMap.get(variable).add(node);

        } else {

          final Collection<Node> set = new LinkedHashSet<>();
          set.add(node);
          varDepNodeMap.put(variable, set);
        }
      }

      for (final Map.Entry<NodeVariable, Collection<Node>> varEntry : varDepNodeMap.entrySet()) {

        final NodeVariable stateVar = varEntry.getKey();
        final Collection<Node> stateDepNodes = varEntry.getValue();

        if (varNodeMap.containsKey(stateVar)) {

          varNodeMap.get(stateVar).add(stateDepNodes);

        } else {

          final Collection<Collection<Node>> pathCollections = new LinkedHashSet<>();
          pathCollections.add(stateDepNodes);
          varNodeMap.put(stateVar, pathCollections);
        }
      }
    }

    return varNodeMap;
  }

  private Node getNodeVarInvariants(final Node ... nodes) {

    final ModuleProcess process = this.efsmProcessMap.get(this.efsm.getName());
    final Map<NodeVariable, Node> invariants = new LinkedHashMap<>();

    for (final Node node : nodes) {

      final Collection<NodeVariable> variables = ExprUtils.getVariables(node);

      for (final NodeVariable variable : variables) {

        if (!invariants.containsKey(variable)) {

          final Node invariant =
              process == null ? null : GaddEfsmTransformerUtils.getInvariant(process, variable);

          if (invariant != null) {
            invariants.put(variable, invariant);
          }
        }
      }
    }
    return invariants.isEmpty() ? Nodes.TRUE : Nodes.and(invariants.values());
  }

  private Set<EfsmState> getDstStates(final EfsmState src, final GaddPath path) {

    final Integer dst = path.getDstPhase();
    final Node guard = path.getGuardNode();
    final Action action = path.getGuardedAction().getAction();
    final Node srcNode = src.getNode();
    final Set<EfsmState> destinations = new LinkedHashSet<>();

    final Set<NodeVariable> stateVars = new LinkedHashSet<>(ExprUtils.getVariables(srcNode));

    if (CollectionUtils.areIntersectedSets(action.getDefines(), stateVars)) {

      final Node preCond = Nodes.and(srcNode, guard, getNodeVarInvariants(srcNode, guard));

      final CfgBlockStatement block = action.toBlock();

      for (final EfsmState dstState : this.efsm.getStates()) {

        if (dstState.getPhase().equals(dst)) {

          final Node dstNode = dstState.getNode();
          final Node postCond = Nodes.and(dstNode, getNodeVarInvariants(dstNode));
          final CfgBlockStatement actionBlock = action.toBlock();
          final Node spCond = CfgUtils.getStrongestPostCondition(preCond, postCond, actionBlock);

          if (SolverUtils.isSat(spCond)) {
            destinations.add(dstState);
          }
        }
      }

      if (destinations.isEmpty()) {

        final Node postCond = CfgUtils.getPostCondition(block);
        final Set<NodeVariable> currentEfsmVars = this.efsm.getStateVariables();
        final Pair<Node, Node> stateOtherPair = NodeUtils.splitByVar(postCond, currentEfsmVars);
        final Node stateDep = stateOtherPair.getLeft();
        final Node stateFreeOrAll = stateOtherPair.getRight();
        final Node stateVarPart = stateDep == null ? stateFreeOrAll : stateDep;
        final EfsmState nextPhaseState = createNewPhaseState(dst, stateVarPart);
        destinations.add(nextPhaseState);
      }
    } else {

      final EfsmState nextPhaseState = createNewPhaseState(dst, srcNode);
      destinations.add(nextPhaseState);
    }

    return destinations;
  }

  private EfsmState createNewPhaseState(final Integer dst, final Node srcNode) {
    final EfsmState nextPhaseState = new EfsmState(srcNode, dst);

    if (!this.efsm.containsState(nextPhaseState)) {

      this.efsm.addState(nextPhaseState);

      if (!this.canBeSrcStates.contains(nextPhaseState)
          && !this.visited.contains(nextPhaseState)) {

        this.canBeSrcStates.add(nextPhaseState);
      }
    }
    return nextPhaseState;
  }
}
