/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.LinkedHashSet;
import java.util.Set;

/**
 * GADD diagram path that includes a conjunction of state-like variable depending expressions
 * and related guarded action.
 *
 * @see GuardedAction
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class GaddPath {

  /* The conjunction of state sub-expressions. */
  private final Set<Node> state;

  /* The outgoing-from-state guarded actions. */
  private final GuardedAction guardedAction;

  private final Set<Integer> srcPhases;

  private final Integer dstPhase;

  /**
   * Constructs object from the specified state expression
   * and the specified guarded action.
   *
   * @param state state expression
   * @param guardedAction guarded action
   * @param srcPhases Source phase values.
   * @param dstPhase Destination phase value.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  GaddPath(
      final Set<Node> state,
      final GuardedAction guardedAction,
      final Set<Integer> srcPhases,
      final Integer dstPhase) {
    InvariantChecks.checkNotNull(state);
    InvariantChecks.checkNotNull(guardedAction);

    this.state = state;
    this.guardedAction = guardedAction;
    this.srcPhases = new LinkedHashSet<>(srcPhases);
    this.dstPhase = dstPhase;
  }

  /**
   * Returns state expression.
   *
   * @return state expression
   */
  public Set<Node> getState() {
    return this.state;
  }

  /**
   * Returns guarded action.
   *
   * @return guarded action
   */
  GuardedAction getGuardedAction() {
    return this.guardedAction;
  }

  /**
   * Returns guard expression node.
   *
   * @return Guard expression node.
   */
  public Node getGuardNode() {
    return getGuardedAction().getGuardNode();
  }

  /**
   * Returns action as a block statement.
   *
   * @return Action as a block statement.
   */
  public CfgBlockStatement getBlock() {
    return getAction().toBlock();
  }

  /**
   * Returns action of the internal guarded action.
   *
   * @return action.
   */
  private Action getAction() {
    return getGuardedAction().getAction();
  }

  /**
   * Returns the set of source values of phase variable.
   *
   * @return The set of source values of phase variable.
   */
  Set<Integer> getSrcPhases() {
    return this.srcPhases;
  }

  /**
   * Returns the destination value of phase variable.
   *
   * @return The destination value of phase variable.
   */
  Integer getDstPhase() {
    return this.dstPhase;
  }

  /**
   * Returns the clock-related node for the internal guarded action.
   *
   * @return The clock-related node for the internal guarded action.
   */
  Node getClockNode() {

    return guardedAction.getClockNode();
  }

  @Override
  public String toString() {
    return String.format(
        "state=%s src=%s GA=%s dst=%s",
        getState(),
        getSrcPhases(),
        getGuardedAction(),
        getDstPhase());
  }
}
