/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;

import java.util.ArrayList;
import java.util.Map;
import java.util.Objects;

/**
 * This class contains some methods that are used in gadd-to-hldd transformation process.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class GaddHlddUtils {

  private GaddHlddUtils() {
  }

  /**
   * Gets the phase variable of the specified process.
   *
   * @param module  - the specified parent module.
   * @param process - the specified process.
   * @return the corresponding phase variable of the process, if contained, {@code null} otherwise.
   */
  public static NodeVariable getPhaseVariable(final Module module, final ModuleProcess process) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(module);
    if (process.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
      final String phaseVarName = process.getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);
      return module.getVariable(phaseVarName);
    }
    return null;
  }

  /**
   * Constructs the blocking variable HLDD using the specifed parameters.
   *
   * @param module           - the specified parent module.
   * @param blockingVariable - the specifed blocking variable.
   * @param phaseVar         - the specified phase variable.
   * @param blockingVars     - the specified process blocking variables mapping.
   * @return the constructed {@link Hldd}.
   */
  public static Hldd constructBlockingVarHldd(final Module module,
      final NodeVariable blockingVariable, final NodeVariable phaseVar,
      final Map<ModuleProcess, NodeVariable> blockingVars) {
    final Hldd hldd = new Hldd(blockingVariable);

    final HlddConditionNode rootNode = new HlddConditionNode();
    rootNode.setLevel(0);
    hldd.addNode(rootNode);
    hldd.setRootNode(rootNode);

    final HlddEdge trueValEdge = new HlddEdge(rootNode);
    rootNode.addChild(trueValEdge);
    trueValEdge.addExpression(constructTrueEdgeCondition(module, phaseVar, blockingVars));
    trueValEdge.setLevel(1);
    hldd.addNode(trueValEdge);

    final HlddTerminalNode trueNode =
        new HlddTerminalNode(trueValEdge, blockingVariable, NodeValue.newBoolean(true));
    trueValEdge.addChild(trueNode);
    trueNode.setLevel(2);
    hldd.addNode(trueNode);

    final HlddEdge falseValEdge = new HlddEdge(rootNode);
    rootNode.addChild(falseValEdge);
    falseValEdge.addExpression(constructFalseEdgeCondition(module, blockingVars));
    falseValEdge.setLevel(1);
    hldd.addNode(falseValEdge);

    final HlddTerminalNode falseNode =
        new HlddTerminalNode(falseValEdge, blockingVariable, NodeValue.newBoolean(false));
    falseValEdge.addChild(falseNode);
    falseNode.setLevel(2);
    hldd.addNode(falseNode);

    return hldd;
  }

  /*
   * Constructs the blocking condition for the specified phase variable.
   */
  private static Node constructTrueEdgeCondition(final Module module,
      final NodeVariable phaseVariable, final Map<ModuleProcess, NodeVariable> blockingVars) {
    final ArrayList<NodeVariable> otherPhaseVars = new ArrayList<>();
    for (ModuleProcess process : blockingVars.keySet()) {
      final NodeVariable otherVar = getPhaseVariable(module, process);
      if (!Objects.equals(otherVar, phaseVariable)) {
        otherPhaseVars.add(getPhaseVariable(module, process));
      }
    }

    final int size = phaseVariable.getDataType().getSize();
    final Node blockCond = Nodes.eq(phaseVariable, new NodeValue(Data.newBitVector(0, size)));
    final ArrayList<Node> otherCond = new ArrayList<>();
    for (NodeVariable var : otherPhaseVars) {
      final int otherSize = var.getDataType().getSize();
      otherCond.add(Nodes.noteq(var, NodeValue.newBitVector(0, otherSize)));
    }
    final Node otherCondConj = ExprUtils.getDisjunction(otherCond.toArray(new Node[0]));
    return ExprUtils.getConjunction(blockCond, otherCondConj);
  }

  /*
   * Constructs the unblocking condition for the specified phase variable.
   */
  private static Node constructFalseEdgeCondition(final Module module,
      final Map<ModuleProcess, NodeVariable> blockingVars) {
    final ArrayList<Node> conditions = new ArrayList<>();
    for (ModuleProcess process : blockingVars.keySet()) {
      final NodeVariable phaseVar = getPhaseVariable(module, process);
      if (phaseVar != null) {
        final int size = phaseVar.getDataType().getSize();
        conditions.add(Nodes.eq(phaseVar, new NodeValue(Data.newBitVector(0, size))));
      }
    }
    return ExprUtils.getConjunction(conditions.toArray(new Node[0]));
  }

  /**
   * Adds the blocking condition to the specified phase variable HLDD.
   *
   * @param hldd     - the specified phase variable HLDD.
   * @param blockVar - the specified blocking variable.
   */
  public static void addBlockingNodes(final Hldd hldd, final NodeVariable blockVar) {
    final HlddConditionNode rootNode = new HlddConditionNode(null, blockVar);
    final HlddEdge falseEdge = new HlddEdge(rootNode);
    falseEdge.addExpression(NodeValue.newBoolean(false));
    rootNode.addChild(falseEdge);

    falseEdge.addChild(hldd.getRootNode());
    hldd.getRootNode().setParent(falseEdge);


    final HlddEdge trueEdge = new HlddEdge(rootNode);
    trueEdge.setDefault();
    trueEdge.addExpression(NodeValue.newBoolean(true));
    rootNode.addChild(trueEdge);

    final NodeVariable currentPhaseVar = hldd.getVariable();
    final HlddTerminalNode assignment =
        new HlddTerminalNode(trueEdge, currentPhaseVar, currentPhaseVar);
    trueEdge.addChild(assignment);

    falseEdge.setLevel(hldd.getRootNode().getLevel() - 1);
    trueEdge.setLevel(falseEdge.getLevel());
    rootNode.setLevel(falseEdge.getLevel() - 1);
    assignment.setLevel(falseEdge.getLevel() + 1);

    hldd.addNode(rootNode);
    hldd.addNode(falseEdge);
    hldd.addNode(trueEdge);
    hldd.addNode(assignment);
    hldd.setRootNode(rootNode);
  }

  /**
   * Constructs a one-node HLDD from the specified switch node and the corresponding guard variable.
   *
   * @param switchNode    - the specified node.
   * @param guardVariable - the specified guard variable.
   * @return the guard variable {@link Hldd}.
   */
  public static Hldd createGuardHldd(final Node switchNode, final NodeVariable guardVariable) {
    final Hldd result = new Hldd(guardVariable);
    final HlddTerminalNode singleNode = new HlddTerminalNode(guardVariable, switchNode);
    singleNode.setLevel(0);
    result.addNode(singleNode);
    result.setRootNode(singleNode);
    result.setImmediateAssignment(true);
    return result;
  }

  /**
   * Removes all terminal nodes from the specified HLDD which are not associated with
   * the specified variable.
   *
   * @param newHldd  - the specified HLDD.
   * @param variable - the specified variable
   * @return the specified variable targeted {@link Hldd}.
   */
  public static Hldd removeWrongVariableNodes(final Hldd newHldd, final NodeVariable variable) {
    newHldd.setVariable(variable);
    for (HlddNode node : newHldd.getNodes()) {
      if (node.isTerminal()) {
        if (!((HlddTerminalNode) node).getTargetVariable().equals(variable)) {
          newHldd.removeNode(node);
        } else {
          final HlddEdge parent = (HlddEdge) node.getParent();
          if (parent != null) {
            parent.addChild(node);
          }
        }
      }
    }
    // If a HLDD consists of only one assignment node.
    if (newHldd.getNodes().size() == 1) {
      newHldd.setRootNode(newHldd.getNodes().iterator().next());
    }
    return newHldd;
  }

  /**
   * Returns the specified array size (for one-dimensional arrays).
   *
   * @param variable - the specified variable.
   * @return array size value, if possible, {@code 0} otherwise.
   */
  public static int getArraySize(final NodeVariable variable) {
    final VariableDescriptor descriptor = getDescriptor(variable);
    if (descriptor != null && descriptor.getIndexRanges() != null) {
      return descriptor.getIndexRanges().get(0).toIntegerRange().size().intValue();
    }
    return 0;
  }

  /**
   * Returns the version number of the specified variable.
   *
   * @param variable - the specified variable.
   * @return the version number value, if possible, {@code -1} otherwise.
   */
  public static int getVersionNumber(final NodeVariable variable) {
    final VariableDescriptor descriptor = getDescriptor(variable);
    if (descriptor != null && descriptor.getOriginalVariable() != null) {
      return descriptor.getVersionNumber();
    }
    return -1;
  }

  /**
   * Returns the original variable of the specified version variable.
   *
   * @param variable - the specified version variable.
   * @return the original {@link NodeVariable}, if possible, {@code null} otherwise.
   */
  public static NodeVariable getOriginalVariable(final NodeVariable variable) {
    final VariableDescriptor descriptor = getDescriptor(variable);
    if (descriptor != null) {
      return descriptor.getOriginalVariable();
    }
    return null;
  }

  /*
   * Returns the variable descriptor.
   */
  private static VariableDescriptor getDescriptor(final NodeVariable variable) {
    if (variable.getUserData() instanceof VariableDescriptor) {
      return (VariableDescriptor) variable.getUserData();
    } else {
      return null;
    }
  }

  /**
   * Indicates that the specified node is the STORE operation.
   *
   * @param node - the specified node.
   * @return {@code true} if the specified node represents the STORE operation, {@code false}
   *     otherwise.
   */
  public static boolean isStore(final Node node) {
    return node instanceof NodeOperation
        && ((NodeOperation) node).getOperationId() == StandardOperation.STORE;
  }

  /**
   * Indicates that the specified node is the SELECT operation.
   *
   * @param node - the specified node.
   * @return {@code true} if the specified node represents the SELECT operation, {@code false}
   *     otherwise.
   */
  public static boolean isSelect(final Node node) {
    return node instanceof NodeOperation
        && ((NodeOperation) node).getOperationId() == StandardOperation.SELECT;
  }

  /**
   * Compares the specified variable names.
   *
   * @param var1 - the specified variable.
   * @param var2 - the specified variable.
   * @return the value 0 if the argument string is equal to this string; a value less than 0
   *     if this string is lexicographically less than the string argument; and a value greater
   *     than 0 if this string is lexicographically greater than the string argument.
   */
  public static int compareVariablesByName(final NodeVariable var1, final NodeVariable var2) {
    return var1.getName().compareTo(var2.getName());
  }
}
