/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.version.VersionRegistry;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.version.VersionUtils;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.version.Versions;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddBuilder;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Stack;

/**
 * This class represents a visitor which is used to construct a HLDD model from a GADD model.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class GaddHlddVisitor extends CfgEmptyVisitor implements HlddBuilder {

  // Blocking variable name prefix.
  private static final String PHASE_BLOCK_NAME = "phase_block_";

  // Resulting model.
  private final HlddModel hlddModel;
  // Switch - corresponding variable mapping.
  private Map<Node, NodeVariable> guardVariables;
  // Counter of guard conditions.
  private int guardCounter;
  // Contains resulting process HLDDs.
  private HlddProcess hlddProcess;
  // HlddModule stack.
  private final LinkedList<Module> moduleStack;
  // Blocking flag.
  private boolean needBlock;
  // Map of process blocking variables.
  private Map<ModuleProcess, NodeVariable> blockingVars;
  // Phase variable of the current process.
  private NodeVariable currentPhaseVar;
  // Variable versions registry.
  private VersionRegistry versionRegistry;
  // Collection of used arrays.
  private Collection<NodeVariable> usedVersions;
  // Stack of path conditions.
  private Stack<Node> conditionStack;
  // variable-value-condition mapping
  private Map<NodeVariable, Map<Node, List<Node>>> assignedValues;

  /**
   * Constructs a visitor.
   */
  public GaddHlddVisitor() {
    super();
    hlddModel = new HlddModel();
    guardVariables = new HashMap<>();
    guardCounter = 0;
    moduleStack = new LinkedList<>();
    needBlock = false;
    blockingVars = new HashMap<>();
    versionRegistry = new VersionRegistry();
    usedVersions = new HashSet<>();
  }

  @Override
  public HlddModel getModel() {
    return hlddModel;
  }

  @Override
  public void onModuleBegin(final Module module) {
    final HlddModule hlddModule = new HlddModule(module);
    moduleStack.push(hlddModule);
    declareBlockingVariables(module, hlddModule);
    searchVersions(module);
  }

  @Override
  public void onModuleEnd(final Module module) {
    final Module hlddModule = moduleStack.pop();
    if (moduleStack.isEmpty()) {
      hlddModel.addModule(hlddModule);
    } else {
      moduleStack.peek().add(hlddModule);
    }
    removeUnusedVersions(hlddModule);
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    hlddProcess = new HlddProcess(process);
    moduleStack.peek().add(hlddProcess);
    conditionStack = new Stack<>();
    assignedValues = new HashMap<>();
    guardVariables = new HashMap<>();
    createBlockingHldds(process);
    versionRegistry.initAllValues();
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {
    constructHldds(process);
    constructGuardHldds();
    currentPhaseVar = null;
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {

    final CfgSwitchStatement parent = (CfgSwitchStatement) caseNode.getParent();
    final Node condition = parent.getCondition();
    final NodeVariable guard = guardVariables.get(condition);
    if (caseNode.isDefault()) {
      conditionStack.push(Nodes.TRUE);
    } else {
      final Set<NodeValue> values = caseNode.getValues();
      if (values.size() > 0) {
        final List<Node> conditions = new LinkedList<>();
        for (NodeValue value : values) {
          conditions.add(Nodes.eq(guard, value));
        }
        conditionStack.push(conditions.size() > 1 ? Nodes.or(conditions) : conditions.get(0));
      } else {
        throw new RetrascopeException(
            String.format("Non-default case of %s has no values!", condition));
      }
    }
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    conditionStack.pop();
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    createGuardVariable(switchNode);
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {
    final Map<NodeVariable, Node> mapping =
        VersionUtils.getAssignmentMap(block.getAssignments(), versionRegistry, true);

    for (Entry<NodeVariable, Node> entry : mapping.entrySet()) {
      final NodeVariable variable = entry.getKey();
      final Node value = entry.getValue();
      final Map<Node, List<Node>> values =
          assignedValues.containsKey(variable) ? assignedValues.get(variable) : new HashMap<>();
      final List<Node> paths = values.containsKey(value) ? values.get(value) : new LinkedList<>();
      if (conditionStack.size() > 1) {
        paths.add(Nodes.and(conditionStack));
      } else if (conditionStack.size() == 1) {
        paths.add(conditionStack.peek());
      } else if (conditionStack.empty()) {
        paths.add(Nodes.TRUE);
      }
      values.put(value, paths);
      assignedValues.put(variable, values);
    }
    collectUsedArrays(mapping.keySet());

    // Reset values in the end of the block.
    versionRegistry.initAllValues();
  }

  /*
   * Creates HLDDs for process variables.
   */
  private void constructHldds(final ModuleProcess process) {

    for (Entry<NodeVariable, Map<Node, List<Node>>> entry : assignedValues.entrySet()) {
      final NodeVariable variable = entry.getKey();
      final Map<Node, List<Node>> values = entry.getValue();
      final Hldd hldd = new Hldd(variable);
      // if variable is a version
      if (GaddHlddUtils.getOriginalVariable(variable) != null) {
        hldd.setImmediateAssignment(true);
      }
      final HlddConditionNode root = new HlddConditionNode();
      root.setLevel(0);
      hldd.addNode(root);
      hldd.setRootNode(root);

      for (Entry<Node, List<Node>> valuePath : values.entrySet()) {
        final Node conditions = Nodes.or(valuePath.getValue());
        final HlddEdge edge = new HlddEdge(root);
        edge.addExpression(conditions);
        edge.setLevel(1);
        root.addChild(edge);
        final HlddTerminalNode assignment =
            new HlddTerminalNode(edge, variable, valuePath.getKey());
        assignment.setLevel(2);
        edge.addChild(assignment);
        hldd.addNode(edge);
        hldd.addNode(assignment);
      }

      if (needBlock && variable.equals(currentPhaseVar)) {
        GaddHlddUtils.addBlockingNodes(hldd, blockingVars.get(process));
      }
      hlddProcess.add(hldd);
    }
  }

  /*
   * Creates HLDDs for guard variables.
   */
  private void constructGuardHldds() {
    final List<Entry<Node, NodeVariable>> entries = new ArrayList<>(guardVariables.entrySet());
    entries.sort((o1, o2) -> GaddHlddUtils.compareVariablesByName(o1.getValue(), o2.getValue()));
    for (Entry<Node, NodeVariable> entry : entries) {
      final Node switchNode = entry.getKey();
      final NodeVariable var = entry.getValue();
      this.hlddProcess.add(GaddHlddUtils.createGuardHldd(switchNode, var));
    }
  }

  /*
   * Declares process blocking variables if the amount of processes is > 1. Blocking variables
   * are used for synchronization.
   */
  private void declareBlockingVariables(final Module module, final HlddModule hlddModule) {
    if (module.getProcesses().size() > 1) {
      needBlock = true;
      int num = 0;
      for (ModuleProcess process : module.getProcesses()) {
        final NodeVariable blockingVariable =
            new NodeVariable(PHASE_BLOCK_NAME + num, DataType.BOOLEAN);
        blockingVars.put(process, blockingVariable);
        num++;
        hlddModule.declareVariable(blockingVariable, NodeValue.newBoolean(false));
      }
    }
  }

  /*
   * Creates HLDDs for blocking variables.
   */
  private void createBlockingHldds(final ModuleProcess process) {
    if (needBlock) {
      currentPhaseVar = GaddHlddUtils.getPhaseVariable(moduleStack.peek(), process);
      final NodeVariable blockingVar = blockingVars.get(process);
      final Hldd blockingVarHldd = GaddHlddUtils
          .constructBlockingVarHldd(moduleStack.peek(), blockingVar, currentPhaseVar, blockingVars);
      hlddProcess.add(blockingVarHldd);
    }
  }

  /*
   * Creates a guard variable from a switch statement.
   */
  private void createGuardVariable(final CfgSwitchStatement switchNode) {
    if (!guardVariables.keySet().contains(switchNode.getCondition())) {
      final NodeVariable variable =
          new NodeVariable("guard" + guardCounter, switchNode.getCondition().getDataType());
      guardVariables.put(switchNode.getCondition(), variable);
      guardCounter++;
    }
  }

  /*
   * Collects array variables marked by SELECT operation.
   */
  private void collectUsedArrays(final Collection<NodeVariable> targets) {
    for (NodeVariable variable : targets) {
      if (variable.getUserData() instanceof NodeOperation) {
        final NodeOperation node = (NodeOperation) variable.getUserData();
        final NodeVariable array = (NodeVariable) node.getOperand(0);
        usedVersions.add(array);
      }
    }
  }

  /*
   * Removes declarations of variable versions that are never used.
   */
  private void removeUnusedVersions(final DecStatement stmt) {
    final Collection<NodeVariable> unusedVersions =
        new ArrayList<>(versionRegistry.getAllVersions());
    unusedVersions.removeAll(usedVersions);
    for (NodeVariable variable : unusedVersions) {
      stmt.removeDeclaration(variable);
    }
  }

  /*
   * Searches for variable versions and adds them to the registry.
   */
  private void searchVersions(final DecStatement stmt) {
    final Map<NodeVariable, List<NodeVariable>> versionMap = VersionUtils.getVersionMap(stmt);
    addToRegistry(versionMap);
  }

  /*
   * Adds variable versions to the versions registry.
   */
  private void addToRegistry(final Map<NodeVariable, List<NodeVariable>> versionMap) {
    for (Entry<NodeVariable, List<NodeVariable>> entry : versionMap.entrySet()) {
      if (!entry.getValue().isEmpty()) {
        final Versions versions = new Versions(entry.getKey(), entry.getValue());
        versionRegistry.addVersions(versions);
      }
    }
  }



}
