/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.version;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * This class contains methods that deal with variable versions and are used in gadd-to-hldd
 * transformation process.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class VersionUtils {

  private VersionUtils() {
  }

  /**
   * Searches for variable versions in the specified {@link DecStatement}.
   *
   * @param stmt - the specified statement.
   * @return the map containing original variables and their versions.
   */
  public static Map<NodeVariable, List<NodeVariable>> getVersionMap(final DecStatement stmt) {
    final Map<NodeVariable, List<NodeVariable>> versionMap = new HashMap<>();
    for (NodeVariable variable : stmt.getVariables()) {
      if (variable.isType(DataTypeId.MAP)) {
        final NodeVariable original = GaddHlddUtils.getOriginalVariable(variable);
        if (original != null) {
          final List<NodeVariable> versions =
              versionMap.containsKey(original) ? versionMap.get(original) : new ArrayList<>();
          versions.add(variable);
          versionMap.put(original, versions);
        } else if (!versionMap.containsKey(variable)) {
          final List<NodeVariable> versions = new ArrayList<>();
          versionMap.put(variable, versions);
        }
      }
    }
    return versionMap;
  }

  /**
   * Creates a variable that represents an array item using the specified array variable and
   * the index value. The constructed variable contains (SELECT array index) operation in its
   * user data.
   *
   * @param array - the specified array variable.
   * @param index - the specified index.
   * @return the {@link NodeVariable} that represents an array item.
   */
  public static NodeVariable getArrayElement(final NodeVariable array, final Node index) {
    final Node intIndex = tryIntConversion(index);
    final String elementName =
        String.format("%s_%s", array.getName(), intIndex.toString()); // array_index
    final DataType elementType = (DataType) array.getDataType().getParameters()[1];
    final NodeVariable element = new NodeVariable(elementName, elementType);
    final NodeOperation elementSelect =
        new NodeOperation(StandardOperation.SELECT, array, intIndex); // array[index]
    element.setUserData(elementSelect);
    return element;
  }

  /**
   * Creates the following node using the specified parameters:
   * {@code (index == intIndex) ? storedValue : array[intIndex]}
   *
   * @param array       - the specified array variable.
   * @param index       - the specified index node.
   * @param intIndex    - the specified index int value.
   * @param storedValue - the specified alternative value node.
   * @return If {@code index} is a value, then the resulting expression can be reduced to:
   *     if {@code index} equals {@code intIndex} then {@code storedValue} is returned,
   *     {@code array[intIndex]} otherwise.
   *     If {@code index} is an expression, then the whole if-then-else node is returned.
   */
  public static Node getValueExpr(final Node array, final Node index, final int intIndex,
      final Node storedValue) {
    final Node select = Nodes.select(array, NodeValue.newInteger(intIndex));
    final DataType indexType = index.getDataType();
    final Node indexValue;
    if (indexType.isTypeId(DataTypeId.BIT_VECTOR)) {
      indexValue = NodeValue.newBitVector(intIndex, indexType.getSize());
    } else if (indexType.isTypeId(DataTypeId.LOGIC_INTEGER)) {
      indexValue = NodeValue.newInteger(intIndex);
    } else {
      throw new RetrascopeException(
          String.format("%s is not an appropriate index type.", indexType.toString()));
    }
    final Node condition = Nodes.eq(index, indexValue);
    final Node reducedNode = Reducer.reduce(condition);
    if (reducedNode.equals(Nodes.TRUE)) {
      return storedValue;
    } else if (reducedNode.equals(Nodes.FALSE)) {
      return select;
    } else {
      return Nodes.ite(condition, storedValue, select);
    }
  }

  /**
   * Tries to transform the specified array index node to INTEGER data type.
   *
   * @param index - the specified index node.
   * @return the transformed node, if the transformation is possible, input node otherwise.
   */
  public static Node tryIntConversion(final Node index) {
    return TransformerUtils.toInt(index) != null ? TransformerUtils.toInt(index) : index;
  }

  /**
   * Tries to get the integer value of the specified node.
   *
   * @param index - the specified node.
   * @return integer value of the node, if possible, {@code -1} otherwise.
   */
  public static int tryGetIntIndex(final Node index) {
    if (index instanceof NodeValue) {
      return ((NodeValue) tryIntConversion(index)).getInteger().intValue();
    }
    return -1;
  }

  /**
   * Creates a variable-assigned value map from the specified collection of assignments.
   *
   * @param assignments     - the specified collection of assignments.
   * @param versionRegistry - the specified {@link VersionRegistry}.
   * @param handleMaps      - enables map versions handling.
   * @return the variable-assigned value map.
   */
  public static Map<NodeVariable, Node> getAssignmentMap(final Collection<Assignment> assignments,
      final VersionRegistry versionRegistry, final boolean handleMaps) {
    InvariantChecks.checkNotNull(assignments);
    final Map<NodeVariable, Node> mapping = new LinkedHashMap<>();
    for (Assignment assignment : assignments) {
      final Node value = assignment.getValue();
      final NodeVariable variable = assignment.getVariable();
      if (variable.isType(DataTypeId.MAP) && handleMaps) {
        // Get array assignments.
        handleMap(variable, value, mapping, versionRegistry);
      } else {
        // Get simple assignment.
        mapping.put(variable, value);
      }
    }
    return mapping;
  }

  /*
   * Handles map variable assignment.
   */
  private static void handleMap(final NodeVariable variable, final Node value,
      final Map<NodeVariable, Node> mapping, final VersionRegistry versionRegistry) {

    // Assigned value is a STORE operation.
    if (GaddHlddUtils.isStore(value)) {
      parseStore((NodeOperation) value, variable, mapping, versionRegistry);
    } else if (value instanceof NodeVariable && value.isType(DataTypeId.MAP)) {
      // Array variable is assigned at once.
      collectArrayAssignments(variable, (NodeVariable) value, mapping, versionRegistry);
    }
  }

  /*
   * Gets array assignments from the STORE operation.
   */
  private static void parseStore(final NodeOperation operation, final NodeVariable array,
      final Map<NodeVariable, Node> mapping, final VersionRegistry versionRegistry) {
    final Node storeDest = operation.getOperand(0);
    final Node index = operation.getOperand(1);
    final Node storedValue = operation.getOperand(2);
    final NodeVariable element = getArrayElement(array, index);
    final Node intIndex = tryIntConversion(index);

    // Specified array variable is a version
    if (versionRegistry.contains(array)) {
      // Index is not const
      // TODO: add situation: array == storeDest && index != const
      if ((storeDest instanceof NodeVariable && !array.equals(storeDest))
          || !(intIndex instanceof NodeValue)) {
        for (int i = 0; i < versionRegistry.getValues(array).size(); i++) {
          versionRegistry.updateValue(array, i, getValueExpr(storeDest, intIndex, i, storedValue));
        }
      } else {
        // Const index
        final int idx = ((NodeValue) intIndex).getInteger().intValue();
        versionRegistry.updateValue(array, idx, storedValue);
        mapping.put(element, storedValue);
      }
      // Specified array variable is original
    } else {
      // Const index
      if (intIndex instanceof NodeValue) {
        mapping.put(element, storedValue);
      } else {
        // Index is not const
        for (int i = 0; i < GaddHlddUtils.getArraySize(array); i++) {
          final NodeVariable item = getArrayElement(array, NodeValue.newInteger(i));
          final Node value = getValueExpr(array, intIndex, i, storedValue);
          mapping.put(item, value);
        }
      }
    }

  }

  /*
   * Searches for and assigns the values of the specified right-hand side variable to the specified
   * original variable
   */
  private static void collectArrayAssignments(final NodeVariable lhs, final NodeVariable rhs,
      final Map<NodeVariable, Node> mapping, final VersionRegistry versionRegistry) {
    if (lhs.getUserData() instanceof VariableDescriptor && rhs
        .getUserData() instanceof VariableDescriptor) {
      final VariableDescriptor lhsDescriptor = (VariableDescriptor) lhs.getUserData();
      final VariableDescriptor rhsDescriptor = (VariableDescriptor) rhs.getUserData();
      if (!lhsDescriptor.getIndexRanges().equals(rhsDescriptor.getIndexRanges())) {
        throw new RetrascopeException("Array types are not equal!");
      }
    }
    mapping.putAll(versionRegistry.collectAssignments(lhs, rhs));
  }

}
