/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.gadd.transformer.hldd.version;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents a container of versions of the original variable.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class Versions {

  private final NodeVariable originalVariable;
  private final List<Version> versions = new ArrayList<>();
  private final int size;

  private final Map<NodeVariable, Map<Integer, Node>> initialValues;

  /**
   * Constructs a new container using the specified original variable and version variables.
   *
   * @param variable - the specified original variable.
   * @param vars     - the specified version variables.
   */
  public Versions(final NodeVariable variable, Collection<NodeVariable> vars) {
    InvariantChecks.checkNotNull(variable);
    originalVariable = variable;
    size = GaddHlddUtils.getArraySize(variable);
    versions.add(new Version(originalVariable, originalVariable, 0, size));
    initialValues = new HashMap<>();
    addVersions(vars);
  }

  /**
   * Adds a version variable to the container.
   *
   * @param versionVar - the specified version var.
   */
  public void addVersion(final NodeVariable versionVar) {
    if (versionVar.getUserData() instanceof VariableDescriptor) {
      final VariableDescriptor descriptor = (VariableDescriptor) versionVar.getUserData();
      if (descriptor.getOriginalVariable().equals(originalVariable)) {
        final Version version =
            new Version(versionVar, descriptor.getOriginalVariable(), descriptor.getVersionNumber(),
                size);
        versions.add(version);
      }
    }
  }

  /**
   * Adds a collection of version variables to the container.
   *
   * @param vars - the specified collection of version variables.
   */
  public void addVersions(final Collection<NodeVariable> vars) {
    for (NodeVariable variable : vars) {
      addVersion(variable);
    }
    versions.sort(Comparator.comparingInt(Version::getVersionNumber));
  }

  /**
   * Returns the original variable this container associated to.
   *
   * @return the original {@link NodeVariable}.
   */
  public NodeVariable getOriginalVariable() {
    return originalVariable;
  }

  /**
   * Returns the current values of the specifed variable.
   *
   * @param variable - the specified variable.
   * @return {@code null} if the specified variable does not belong to this container, a list of
   *     current values otherwise.
   */
  public List<Node> getValues(final NodeVariable variable) {
    final Version version = getVersion(variable);
    return version == null ? null : Arrays.asList(version.getValues());
  }

  /**
   * Updates the value of the specified variable.
   *
   * @param variable - the specified variable.
   * @param index    - the specified index of the value.
   * @param value    - the specified value.
   */
  public void updateValue(final NodeVariable variable, final int index, final Node value) {
    final Version version = getVersion(variable);
    if (version != null && index < size) {
      version.getValues()[index] = value;
    }
  }

  /**
   * Fills the values of the variables.
   * This container is assumed to contain arrays, so each value is initialized by the corresponding
   * SELECT operation (e.g. {@code i}th value of variable {@code x} is assigned by {@code x[i]}).
   */
  public void initValues() {
    for (Version version : versions) {
      for (int i = 0; i < size; i++) {
        final Node initialValue;
        final NodeVariable variable = version.getVariable();
        if (initialValues.containsKey(variable) && initialValues.get(variable).containsKey(i)) {
          initialValue = initialValues.get(variable).get(i);
        } else {
          final Node select = Nodes.select(variable, NodeValue.newInteger(i));
          initialValue = select;
          final Map<Integer, Node> values =
              initialValues.containsKey(variable) ? initialValues.get(variable) : new HashMap<>();
          values.put(i, select);
          if (!initialValues.containsKey(variable)) {
            initialValues.put(variable, values);
          }
        }
        version.getValues()[i] = initialValue;
      }
    }
  }

  /**
   * Returns the size of the array of values.
   *
   * @return value array size.
   */
  public int getSize() {
    return size;
  }

  /*
   * Searches for the corresponding Version of a variable.
   */
  private Version getVersion(final NodeVariable variable) {
    if (variable.equals(originalVariable)) {
      return versions.get(0);
    } else if (originalVariable.equals(GaddHlddUtils.getOriginalVariable(variable))) {
      return versions.get(GaddHlddUtils.getVersionNumber(variable));
    } else {
      return null;
    }
  }

  /**
   * Returns the versions of the original variable.
   *
   * @return the collection of versions except the original variable.
   */
  public Collection<NodeVariable> getVersions() {
    final Collection<NodeVariable> result = new ArrayList<>();
    for (int i = 1; i < versions.size(); i++) {
      result.add(versions.get(i).getVariable());
    }
    return result;
  }

}
