/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.engine.gadd.transformer.hldd.GaddHlddUtils;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.MapUtils;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvPrinterUtils;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvReservedWords;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;
import ru.ispras.retrascope.util.ReadWriteUtils;
import ru.ispras.retrascope.util.TemporalOperation;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Basic HLDD visitor that provides SMV representation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class HlddSmvVisitor extends HlddEmptyVisitor {

  // Prints expressions into file.
  private final PrintWriter printWriter;

  // SMV expression printer.
  private final SmvExprPrinter exprPrinter;

  // Indentation.
  private int indentationLevel;

  // Current HLDD variable.
  private NodeVariable hlddVariable;

  // Enables printing.
  private boolean printingEnabled;

  private Collection<NodeVariable> currentInputs;

  protected Map<NodeVariable, Declaration> currentPhaseVars;

  protected Module currentModule;

  private static final int BASIC_INDENT = 3;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddSmvVisitor(final String fileName) {
    super();
    final File smvFile = new File(fileName);
    exprPrinter = new SmvExprPrinter();
    printingEnabled = true;
    printWriter = ReadWriteUtils.newPrintWriter(smvFile);
    currentInputs = new ArrayList<>();
    indentationLevel = BASIC_INDENT;
  }

  /**
   * Returns the current SMV expression printer.
   *
   * @return The current SMV expression printer.
   */
  protected SmvExprPrinter getExprPrinter() {
    return this.exprPrinter;
  }

  /**
   * Appends the specified string to the output file.
   *
   * @param string The string to be added to the output file.
   */
  protected void append(final String string) {
    if (printingEnabled) {
      printWriter.print(string);
    }
  }

  /**
   * Appends the specified line to the output file.
   *
   * @param line The line to be added to the output file.
   */
  protected void appendLine(final String line) {
    if (printingEnabled) {
      printWriter.println(line);
    }
  }

  /**
   * Appends an empty line to the output file.
   */
  protected void appendLine() {
    if (printingEnabled) {
      printWriter.println();
    }
  }

  /**
   * Returns the expression description string.
   *
   * @param expression The expression to be represented in string format.
   * @return The SMV string representation for the specified expression.
   */
  protected String getExprDescription(final Node expression) {
    return SmvExprPrinter.replaceLogicValues(
        exprPrinter.toString(SmvPrinterUtils.makeSubstitutions(expression, true)));
  }

  protected String getExprDescription(final Node expression, final boolean standardize) {
    return SmvExprPrinter.replaceLogicValues(
        exprPrinter.toString(SmvPrinterUtils.makeSubstitutions(expression, standardize)));
  }

  @Override
  public void onModelBegin(final Model model) {

    appendLine();
    // Print "main".
    appendLine(SmvPrinterUtils.getModuleDeclaration(SmvReservedWords.MAIN_MODULE.toString()));

    final Map<NodeVariable, Declaration> inputVariables = new HashMap<>();

    // Declare nested modules.
    for (final Module hlddModule : model.getModules()) {
      final String moduleName = SmvPrinterUtils.getName(hlddModule.getName());
      final String instanceName = "m_" + moduleName;
      for (final NodeVariable input : hlddModule.getInputs()) {
        inputVariables.put(input, hlddModule.getDeclaration(input));
      }
      appendLine(
          SmvPrinterUtils.getModuleInstantiation(instanceName, moduleName, hlddModule.getInputs()));
    }

    // Declare input variables.
    appendLine(SmvPrinterUtils.getVarDeclareSection(inputVariables, exprPrinter));

  }

  @Override
  public void onModelEnd(final Model model) {
    appendLine();
    printWriter.close();
  }

  @Override
  public void onModuleBegin(final Module hlddModule) {
    currentInputs = hlddModule.getInputs();
    currentModule = hlddModule;
    final String moduleName = SmvPrinterUtils.getName(hlddModule.getName());
    appendLine();
    // Print module.
    appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, currentInputs));
    final List<NodeVariable> regs = SmvPrinterUtils.getRegs(hlddModule);
    for (NodeVariable output : hlddModule.getOutputs()) {
      final VariableType type = ((VariableDescriptor) output.getUserData()).getVariableType();
      if (type == VariableType.OUT) {
        regs.add(output);
      }
    }
    if (!regs.isEmpty()) {
      final Map<NodeVariable, Declaration> regDeclarations = new HashMap<>();
      for (NodeVariable reg : regs) {
        regDeclarations.put(reg, hlddModule.getDeclaration(reg));
      }
      appendLine(SmvPrinterUtils.getVarDeclareSection(regDeclarations, exprPrinter));
      currentInputs.addAll(regs);
    }

    currentPhaseVars = getPhaseVariables(hlddModule.getProcesses());

    // Declare nested processes.
    //appendLine(SmvPrinterUtils.getProcessInstantiation(hlddModule.getProcesses(), currentInputs));
  }

  private Map<NodeVariable, Declaration> getPhaseVariables(
      final Collection<ModuleProcess> processes) {
    final Map<NodeVariable, Declaration> result = new HashMap<>();
    for (ModuleProcess process : processes) {
      if (process.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
        final String phaseVarName = process.getStringMetaInfo(MetaInfoType.PHASE_VAR_NAME);
        final NodeVariable phaseVar = currentModule.getVariable(phaseVarName);
        if (phaseVar != null) {
          final Declaration declaration = currentModule.getDeclaration(phaseVar);
          result.put(phaseVar, declaration);
        }
      }
    }
    return result;
  }

  @Override
  public void onModuleEnd(final Module module) {
    for (Node node : module.getProperties()) {
      appendLine(getExprDescription(node));
    }
    appendLine();
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    //final String moduleName = SmvPrinterUtils.getName(process.getName());
    appendLine();
    // Print process.
    //appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, currentInputs));
    //appendLine();

    final Node invariant = ((HlddProcess) process).getInvariant();
    if (invariant != null) {
      appendLine(SmvReservedWords.INVAR.toString());
      appendLine(getExprDescription(invariant));
      appendLine();
    }

    appendLine(SmvReservedWords.ASSIGNMENT.toString());
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    appendLine(MapUtils.printMaps(exprPrinter));
    MapUtils.resetMaps();
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    final NodeVariable variable = hldd.getVariable();
    final String name = (variable.getUserData() instanceof NodeOperation) ? getExprDescription(
        (NodeOperation) variable.getUserData()) : SmvPrinterUtils.getName(variable);
    if (hldd.isImmediateAssignment()) {
      append(String.format("  %s := ", name));
    } else {
      append(String.format("  %s(%s) := ", SmvReservedWords.NEXT.toString(), name));
    }
    hlddVariable = hldd.getVariable();
  }

  @Override
  public void onHlddEnd(final Hldd hldd) {
    appendLine();
  }

  @Override
  public void onConditionBegin(final HlddConditionNode node) {
    indentationLevel++;
    appendLine(String.format("%n%s%s", SmvPrinterUtils.getWhitespace(indentationLevel),
        SmvReservedWords.CASE_BEGIN.toString()));
  }

  @Override
  public void onConditionEnd(final HlddConditionNode node) {
    if (!node.hasDefaultEdge()) {
      appendLine(String.format("%s  %s : %s;", SmvPrinterUtils.getWhitespace(indentationLevel),
          SmvReservedWords.TRUE.toString(), getAssignedValue()));
    }
    appendLine(String.format("%s%s;", SmvPrinterUtils.getWhitespace(indentationLevel),
        SmvReservedWords.CASE_END.toString()));
    indentationLevel--;
  }

  private void printProperties(final Collection<Node> properties) {
    for (Node property : properties) {
      if (TemporalOperation.isTemporal(property)) {
        switch (TemporalOperation.getFamily((NodeOperation) property)) {
          case CTL:
            appendLine(SmvReservedWords.CTLSPEC.toString());
            break;
          case LTL:
            appendLine(SmvReservedWords.LTLSPEC.toString());
            break;
          case PSL:
            appendLine(SmvReservedWords.PSLSPEC.toString());
            break;
          default:
            throw new IllegalArgumentException(String
                .format("Properties of type %s are not supported by nuXmv",
                    TemporalOperation.getFamily((NodeOperation) property)));
        }
      } else {
        appendLine(SmvReservedWords.INVARSPEC.toString());
      }
    }
  }

  private String getAssignedValue() {
    final String assignedValue;

    if (userDataIsOp(hlddVariable)) {
      assignedValue = getExprDescription((NodeOperation) hlddVariable.getUserData());
    } else if (userDataIsVar(hlddVariable)) {
      final NodeVariable var = (NodeVariable) hlddVariable.getUserData();
      if (userDataIsOp(var)) {
        assignedValue = getExprDescription((NodeOperation) var.getUserData());
      } else {
        assignedValue = getExprDescription((NodeVariable) hlddVariable.getUserData());
      }
    } else if (GaddHlddUtils.getOriginalVariable(hlddVariable) != null) {
      assignedValue = getExprDescription(GaddHlddUtils.getOriginalVariable(hlddVariable));
    } else {
      assignedValue = SmvPrinterUtils.getName(hlddVariable);
    }
    return assignedValue;
  }

  private boolean userDataIsOp(final Node node) {
    return node.getUserData() instanceof NodeOperation;
  }

  private boolean userDataIsVar(final Node node) {
    return node.getUserData() instanceof NodeVariable;
  }

  @Override
  public void onEdgeBegin(final HlddEdge node) {
    indentationLevel++;
    final Node finalExpression =
        SmvPrinterUtils.makeSubstitutions(node.getConditionEquation(), true);
    append(String.format("%s%s : ", SmvPrinterUtils.getWhitespace(indentationLevel),
        getExprDescription(finalExpression)));
  }

  @Override
  public void onEdgeEnd(final HlddEdge node) {
    indentationLevel--;
  }

  @Override
  public void onAssignment(final HlddTerminalNode node) {
    final Node expression = node.getAssignedValue().deepCopy();
    if (expression.getKind() == Node.Kind.VALUE) {
      appendLine(getExprDescription(coerceTypes(expression)) + ";");
    } else {
      if (expression.getKind() == Node.Kind.OPERATION
          && ((NodeOperation) expression).getOperationId() == SmvReservedWords.Operation.NEXT) {
        final Node arg = ((NodeOperation) expression).getOperand(0);
        appendLine(
            String.format("%s(%s);", SmvReservedWords.NEXT.toString(), getExprDescription(arg)));
      } else {
        final Node finalExpression = SmvPrinterUtils.makeSubstitutions(expression, true);
        appendExpression(getExpressionType(finalExpression).getTypeId(),
            getExprDescription(finalExpression));
      }
    }
  }

  /*
   * Coerces expression type to the target variable type.
   */
  private Node coerceTypes(final Node expression) {
    final DataType targetDataType = hlddVariable.getDataType();
    if (!expression.getDataType().equals(targetDataType)
        && expression.getDataTypeId() != DataTypeId.MAP) {
      final Node coercedExpr = TypeConversion.coerce(expression, targetDataType);
      if (coercedExpr != null) {
        return coercedExpr;
      }
    }
    return expression;
  }

  /*
   * Returns the specified expressions data type if possible.
   */
  private DataType getExpressionType(final Node expression) {
    DataType exprType = DataType.UNKNOWN;
    try {
      exprType = expression.getDataType();
    } catch (IllegalStateException | NullPointerException e) {
      exprType = DataType.UNKNOWN;
    } finally {
      if (exprType == null) {
        exprType = DataType.UNKNOWN;
      }
    }
    return exprType;
  }

  /*
   * Prints the specified expression description and adds the type conversion expression if needed.
   */
  private void appendExpression(final DataTypeId exprType, final String exprDescription) {
    if (conversionNeeded(exprType)) {
      appendLine(String.format("%s(%s);",
          SmvPrinterUtils.getTypeConversionPrefix(exprType, hlddVariable.getDataTypeId()),
          exprDescription));
    } else {
      appendLine(exprDescription + ";");
    }
  }

  /*
   * Shows whether the specified type needs conversion to target variable type.
   */
  private boolean conversionNeeded(final DataTypeId exprType) {
    return exprType != hlddVariable.getDataTypeId() && exprType != DataTypeId.UNKNOWN;
  }

}
