/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.spec;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.HlddClockedVisitorBase;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedReader;
import java.io.IOException;

/**
 * This class represents the visitor that prints the SMV model together with
 * the given specifications.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddSpecSmvVisitor extends HlddClockedVisitorBase {

  // Additional specification file.
  private String specFile = null;
  private GaddModel gadd = null;
  private Module currentModule = null;

  /**
   * Adds the specification file to the visitor.
   *
   * @param specFile - the specified file name.
   */
  public void setSpecificationFile(final String specFile) {
    this.specFile = specFile;
  }

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddSpecSmvVisitor(final String fileName) {
    super(fileName);
  }

  public void setGaddModel(final GaddModel model) {
    gadd = model;
  }

  @Override
  protected NodeVariable getClock() {
    if (gadd != null && currentModule != null && gadd.hasClock(currentModule.getName())) {
      return gadd.getClock(currentModule.getName());
    }
    return null;
  }

  @Override
  public void onModelBegin(final Model model) {
    super.onModelBegin(model);

    // Add specifications from file.
    if (null != specFile) {
      appendLine(readSpecifications(specFile));
    }
  }

  @Override
  public void onModuleBegin(final Module module) {
    currentModule = module;
    super.onModuleBegin(module);
  }

  /*
   * Returns the contents of the specification file as a string.
   */
  private String readSpecifications(final String fileName) {
    final StringBuilder builder = new StringBuilder();
    try {
      final BufferedReader reader = ReadWriteUtils.newBufferedReader(fileName);
      String line;
      while ((line = reader.readLine()) != null) {
        builder.append(line);
        builder.append("\n");
      }
      reader.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return builder.toString();
  }

}
