/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.usedef;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.engine.hldd.printer.smv.basis.HlddClockedVisitorBase;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvReservedWords;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.gadd.GaddModel;
import ru.ispras.retrascope.result.usedef.UseDefContainer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * This visitor prints the HLDD model and the corresponding use-def properties.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddUseDefSmvVisitor extends HlddClockedVisitorBase {

  private GaddModel gaddModel;
  private Map<NodeVariable, UseDefContainer> useDefMap;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddUseDefSmvVisitor(String fileName) {
    super(fileName);
    this.useDefMap = new LinkedHashMap<>();
  }

  /**
   * Adds a GADD model to the visitor.
   *
   * @param model - the specified GaddModel.
   */
  public void setGaddModel(final GaddModel model) {
    gaddModel = model;
  }

  /**
   * Adds a variable-use/def map to the visitor.
   *
   * @param map - the specified map.
   */
  public void setUseDefMap(final Map<NodeVariable, UseDefContainer> map) {
    if (map != null) {
      useDefMap = map;
    }
  }

  @Override
  public void onModuleEnd(final Module module) {
    super.onModuleEnd(module);

    printModuleProperties();
  }

  private void printModuleProperties() {
    for (Map.Entry<NodeVariable, UseDefContainer> entry : useDefMap.entrySet()) {
      if (!entry.getKey().equals(getClock())) {
        final UseDefContainer useDef = entry.getValue();

        if (useDef.getDefines().size() > 1) {
          appendLine(String.format("%n-- %s ww", entry.getKey()));
          printWwConditions(useDef.getDefines().values());
        }

        if (useDef.getDefines().size() > 0 && useDef.getUses().size() > 0 && !HlddUseDefUtils
            .useDefSameProcess(useDef)) {
          appendLine(String.format("%n-- %s rw", entry.getKey()));
          printRwCondition(useDef);
          final Node defCondition = Nodes.or(useDef.getDefines().values());
          final Node useCondition = Nodes.or(useDef.getUses().values());
          appendLine(String.format("%n-- %s undef", entry.getKey()));
          defExistsBeforeUse(defCondition, useCondition);
          appendLine(String.format("%n-- %s wrw", entry.getKey()));
          useExistsBetweenDefs(defCondition, useCondition, defCondition);
        }
      }
    }
  }

  private void printWwConditions(final Collection<Node> defs) {
    final List<Node> defList = new ArrayList<>(defs);
    final List<Node> conditions = new ArrayList<>();
    for (int i = 0; i < defList.size(); i++) {
      for (int j = i + 1; j < defList.size(); j++) {
        conditions.add(Nodes.and(defList.get(i), defList.get(j)));
      }
    }
    final Node wwCondition = conditions.size() == 1 ? conditions.get(0) : Nodes.or(conditions);
    appendLine(SmvReservedWords.LTLSPEC.toString());
    appendLine(String.format("! F (%s)", getExprDescription(wwCondition)));
  }

  private void printRwCondition(final UseDefContainer useDef) {
    final List<Map.Entry<String, Node>> defs = new ArrayList<>(useDef.getDefines().entrySet());
    final List<Map.Entry<String, Node>> uses = new ArrayList<>(useDef.getUses().entrySet());

    for (int i = 0; i < defs.size(); i++) {
      for (int j = 0; j < uses.size(); j++) {
        final String defProcess = getProcessName(defs, i);
        final Node defCondition = getCondition(defs, i);

        final String useProcess = getProcessName(uses, j);
        final Node useCondition = getCondition(uses, j);

        if (defProcess.compareTo(useProcess) != 0) {
          final Node rwCondition = Nodes.and(defCondition, useCondition);
          appendLine(SmvReservedWords.LTLSPEC.toString());
          appendLine(String.format("! F (%s)", getExprDescription(rwCondition)));
        }
      }
    }

  }

  private String getProcessName(final List<Map.Entry<String, Node>> list, final int index) {
    return list.get(index).getKey();
  }

  private Node getCondition(final List<Map.Entry<String, Node>> list, final int index) {
    return list.get(index).getValue();
  }

  private void defExistsBeforeUse(final Node def, final Node use) {
    appendLine(SmvReservedWords.LTLSPEC.toString());
    appendLine(String.format("G (!%s U %s)", getExprDescription(use), getExprDescription(def)));
    // G (!use U def)
  }

  private void useExistsBetweenDefs(final Node firstDef, final Node lastDef, final Node use) {
    appendLine(SmvReservedWords.LTLSPEC.toString());
    appendLine(String
        .format("F (%s & ((F %s) U %s))", getExprDescription(firstDef), getExprDescription(use),
            getExprDescription(lastDef)));
    // F(def_1 & ((F use) U def_2)
  }

  @Override
  protected NodeVariable getClock() {
    if (gaddModel != null && currentModule != null && gaddModel.hasClock(currentModule.getName())) {
      return gaddModel.getClock(currentModule.getName());
    }
    return null;
  }
}
