/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.List;

/**
 * Utility methods for ranged variable class.
 *
 * @see RangedVariable
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class RangedVariableUtils {

  private RangedVariableUtils() {
    /* Empty. */
  }

  /**
   * Transforms the specified node with assistance of the specified ranged variable.
   * <p>For example, we have a pair '(a[x], b)' where 'a[x]' is ranged variable and 'b'
   * is node. Such pair may come from {@link Assignment} node constructor parameters
   * that should be transformed to (NodeVariable, Node) pair. This method transforms 'b'
   * expression in such way that the result can be assigned to RangedVariable's NodeVariable field
   * without any Range.</p>
   * @param rangedVariable Ranged variable that is used in transformation.
   * @param value Node to be transformed.
   * @return Result of the specified node's transformation with using of the specified
   *         ranged variable.
   */
  public static Node transformNodeWithRanged(
      final RangedVariable rangedVariable,
      final Node value) {

    if (!rangedVariable.ranged() && !rangedVariable.indexed()) {

      return value;
    }

    if (!rangedVariable.ranged()) {

      return indicesToStores(rangedVariable, value);
    }

    final NodeVariable variable = rangedVariable.getVariable();

    Node result = value;

    final Range range = rangedVariable.getRange();

    /* Calculate the type for range fields. */
    final DataType rangeType = getFieldType(range);

    final NodeValue rangeZero = createValue(rangeType, 0);
    final NodeValue rangeOne = createValue(rangeType, 1);

    /* Calculate the type of ARRAY element for indexed target (or variable type for non-indexed). */
    final DataType nonRangeVarType =
        getDataType(new RangedVariable(variable, rangedVariable.getIndices()));

    if (nonRangeVarType.isTypeId(DataTypeId.BIT_VECTOR)) {

      if (ExprUtils.isKind(Node.Kind.VALUE, range.getHigh(), range.getLow())) {

        result = TransformerUtils.standardize(getNodeValueRanged(rangedVariable, value));

      } else {

        final NodeValue nonRangeVarSize = NodeValue.newInteger(nonRangeVarType.getSize());

        final Node rangeTypeSize = TypeConversion.coerce(nonRangeVarSize, rangeType);
        final StandardOperation subOp =
            rangeType == DataType.INTEGER ? StandardOperation.SUB : StandardOperation.BVSUB;
        final Node maxNum = new NodeOperation(subOp, rangeTypeSize, rangeOne);

        /* low == 0 && high == size - 1 */
        final Node nonRanged = getNonRangedCondition(rangeZero, maxNum, range);

        /* low > 0 && high == size - 1 */
        final Node highBorder = getHighBorderCondition(rangeZero, maxNum, range);
        final Node highBorderNode = getHighBorderExpr(value, rangedVariable, rangeZero, rangeOne);

        /* low == 0 && high < size - 1 */
        final Node lowBorder = getLowBorderCondition(rangeZero, maxNum, range);
        final Node lowBorderNode = getLowBorderExpr(value, rangedVariable, rangeOne, maxNum);

        final Node generalNode =
            getGeneralCaseExpr(value, rangedVariable, rangeZero, rangeOne, maxNum);

        result = Nodes.ite(
            nonRanged,
            value,
            Nodes.ite(
                highBorder, highBorderNode,
                Nodes.ite(lowBorder, lowBorderNode, generalNode)));
      }
    }

    if (rangedVariable.indexed()) {
      result = indicesToStores(rangedVariable, result);
    }
    return result;
  }

  private static Node indicesToStores(final RangedVariable target, final Node value) {

    final List<Node> indices = target.getIndices();
    final NodeVariable targetVar = target.getVariable();

    if (indices == null || indices.isEmpty()) {
      return value;
    }

    Node result = value;

    for (int i = indices.size() - 1; i >= 0; i--) {

      result = Nodes.store(targetVar, indices.get(i), result);
    }

    return result;
  }

  private static Node indicesToSelects(final RangedVariable rangedVariable) {

    final NodeVariable variable = rangedVariable.getVariable();

    if (!rangedVariable.indexed()) {
      return variable;
    }

    final List<Node> indices = rangedVariable.getIndices();

    Node result = variable;

    for (final Node index : indices) {

      result = Nodes.select(result, index);
    }

    return result;
  }

  private static NodeOperation getGeneralCaseExpr(
      final Node node,
      final RangedVariable rangedVariable,
      final NodeValue zero,
      final NodeValue one,
      final Node maxNum) {

    final Range range = rangedVariable.getRange();
    final DataType dataType = getFieldType(range);

    final Node maxNumExpr = maxNum instanceof NodeValue ? TransformerUtils.toInt(maxNum) : maxNum;

    final Node highPlusOneExpr = new NodeOperation(
        dataType == DataType.INTEGER ? StandardOperation.ADD : StandardOperation.BVADD,
        rangedVariable.getRange().getHigh(),
        one);
    final Node highPlusOneInt = TransformerUtils.toInt(highPlusOneExpr);
    final Node highPlusOne = highPlusOneInt == null ? highPlusOneExpr : highPlusOneInt;

    final Node lowMinusOneExpr = new NodeOperation(
        dataType == DataType.INTEGER ? StandardOperation.SUB : StandardOperation.BVSUB,
        rangedVariable.getRange().getLow(),
        one);
    final Node lowMinusOneInt = TransformerUtils.toInt(lowMinusOneExpr);
    final Node lowMinusOne = lowMinusOneInt == null ? lowMinusOneExpr : lowMinusOneInt;

    return Nodes.bvconcat(
        Nodes.bvextract(
            maxNumExpr,
            highPlusOne,
            indicesToSelects(rangedVariable)),
        node,
        Nodes.bvextract(
            lowMinusOne,
            TransformerUtils.toInt(zero),
            indicesToSelects(rangedVariable)));
  }

  private static Node getLowBorderCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Nodes.and(
        Nodes.eq(range.getLow(), zero),
        new NodeOperation(
            range.getHigh().isType(DataType.INTEGER)
                ? StandardOperation.LESS : StandardOperation.BVULT,
            range.getHigh(),
            maxNum));
  }

  private static Node getHighBorderCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Nodes.and(
        new NodeOperation(
            getFieldType(range) == DataType.INTEGER
                ? StandardOperation.GREATER : StandardOperation.BVUGT,
            range.getLow(),
            zero),
        Nodes.eq(range.getHigh(), maxNum));
  }

  private static Node getNonRangedCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Nodes.and(Nodes.eq(range.getLow(), zero), Nodes.eq(range.getHigh(), maxNum));
  }

  private static NodeOperation getLowBorderExpr(
      final Node node,
      final RangedVariable rangedVar,
      final NodeValue one,
      final Node maxNum) {

    final Range range = rangedVar.getRange();
    final Node high = range.getHigh();
    final Node highPlusOneExpr =
        new NodeOperation(
            getFieldType(range) == DataType.INTEGER
                ? StandardOperation.ADD
                : StandardOperation.BVADD,
            high,
            one);
    final Node highPlusOne = TransformerUtils.toInt(highPlusOneExpr);
    final Node maxInt = TransformerUtils.toInt(maxNum);

    return Nodes.bvconcat(
        Nodes.bvextract(
            maxInt,
            highPlusOne == null ? highPlusOneExpr : highPlusOne,
            indicesToSelects(rangedVar)),
        node);
  }

  private static NodeOperation getHighBorderExpr(
      final Node value,
      final RangedVariable rangedVar,
      final NodeValue zero,
      final NodeValue one) {

    final Range range = rangedVar.getRange();
    final Node low = range.getLow();
    final Node lowMinusOneExpr =
        new NodeOperation(
            getFieldType(range) == DataType.INTEGER
                ? StandardOperation.SUB : StandardOperation.BVSUB,
            low,
            one);

    final Node lowMinusOne = TransformerUtils.toInt(lowMinusOneExpr);
    final Node zeroInt = TransformerUtils.toInt(zero);

    return Nodes.bvconcat(
        value,
        Nodes.bvextract(
            lowMinusOne == null ? lowMinusOneExpr : lowMinusOne,
            zeroInt,
            indicesToSelects(rangedVar)));
  }

  private static Node getNodeValueRanged(final RangedVariable rangedVariable, final Node value) {

    Range range = rangedVariable.getRange();
    final NodeVariable variable = rangedVariable.getVariable();

    /* Calculate 'one' and 'zero' values depending on range data type. */
    final DataType rangeFieldType = getFieldType(range);

    final NodeValue rangeZero = createValue(rangeFieldType, 0);
    final NodeValue rangeOne = createValue(rangeFieldType, 1);

    final int arrayElemSize =
        getDataType(new RangedVariable(variable, rangedVariable.getIndices())).getSize();
    final NodeValue size = createValue(rangeFieldType, arrayElemSize);

    final Node maxRangeFieldValue = Reducer.reduce(
        rangeFieldType == DataType.INTEGER
            ? Nodes.sub(size, rangeOne)
            : Nodes.bvsub(size, rangeOne));

    /* Check that range fields lie between 0 and size - 1. */
    InvariantChecks.checkTrue(
        checkRangeFieldInvariant(range, maxRangeFieldValue),
        String.format("for '%s' range is broken: 0 <= low='%s' <= high='%s' < '%s'",
            variable, range.getLow(), range.getHigh(), maxRangeFieldValue));

    /* low == 0 && high == size - 1 */
    if (SolverUtils.isSat(getNonRangedCondition(rangeZero, maxRangeFieldValue, range))) {

      /* Actually, range coincides with complete value, so ignore the range. */
      return value;
    }

    /* low > 0 && high == size - 1 */
    if (SolverUtils.isSat(getHighBorderCondition(rangeZero, maxRangeFieldValue, range))) {

      return getHighBorderExpr(value, rangedVariable, rangeZero, rangeOne);
    }

    /* low == 0 && high < size - 1 */
    if (SolverUtils.isSat(getLowBorderCondition(rangeZero, maxRangeFieldValue, range))) {

      return getLowBorderExpr(value, rangedVariable, rangeOne, maxRangeFieldValue);
    }

    return getGeneralCaseExpr(value, rangedVariable, rangeZero, rangeOne, maxRangeFieldValue);
  }

  private static boolean checkRangeFieldInvariant(final Range range, final Node maxValue) {

    final DataType rangeFieldType = getFieldType(range);

    final Enum<?> lessEq =
        rangeFieldType == DataType.INTEGER ? StandardOperation.LESSEQ : StandardOperation.BVULE;

    final Enum<?> greaterEq =
        rangeFieldType == DataType.INTEGER ? StandardOperation.GREATEREQ : StandardOperation.BVUGE;

    final Node high = range.getHigh();
    final Node low = range.getLow();

    return SolverUtils.isSat(Nodes.and(
        new NodeOperation(greaterEq, low, createValue(rangeFieldType, 0)),
        new NodeOperation(lessEq, high, maxValue),
        new NodeOperation(greaterEq, high, low)));
  }

  private static NodeValue createValue(
      final DataType dataType,
      final int value) {

    return dataType == DataType.INTEGER
        ? NodeValue.newInteger(value)
        : NodeValue.newBitVector(value, dataType.getSize());
  }

  /**
   * Returns the data type of this object.
   * @param rangedVariable Variable that type is calculated.
   *
   * @return The data type of this object or DataType.UNKNOWN if it fails.
   */
  public static DataType getDataType(final RangedVariable rangedVariable) {

    DataType result = rangedVariable.getVariable().getDataType();

    if (rangedVariable.indexed()) {

      while (result.isTypeId(DataTypeId.MAP)) {

        result = (DataType) result.getParameters()[1];
      }
    }

    if (rangedVariable.ranged()) {

      final Node high = rangedVariable.getRange().getHigh();
      final DataType highType = high.getDataType();
      final Node low = rangedVariable.getRange().getLow();
      final DataType lowType = low.getDataType();

      if (ExprUtils.isKind(Node.Kind.VALUE, high, low) && highType.equals(lowType)) {

        final boolean areInt = ExprUtils.isType(DataType.INTEGER, high, low);

        final NodeOperation size =
            new NodeOperation(areInt ? StandardOperation.ADD : StandardOperation.BVADD,
                new NodeOperation(
                    areInt
                        ? StandardOperation.SUB
                        : StandardOperation.BVSUB, high, low),
                areInt
                    ? NodeValue.newInteger(1)
                    : NodeValue.newBitVector(1, highType.getSize()));

        final NodeValue intSize = (NodeValue) TransformerUtils.toInt(size);
        result = DataType.bitVector((intSize).getInteger().intValue());
      } else {
        result = DataType.UNKNOWN;
      }
    }

    return result;
  }

  private static DataType getFieldType(final Range range) {

    final Node high = range.getHigh();
    final DataType highType = high.getDataType();

    final Node low = range.getLow();
    final DataType lowType = low.getDataType();

    return highType == null || highType.equals(DataType.UNKNOWN) ? lowType : highType;
  }

  /**
   * Returns the invariant in range form.
   * @param left The first border of the range invariant.
   * @param right The second border of the ranged invariant.
   * @param ascend Flag that shows whether second border is greater than the first one.
   * @return The range, which left border is the smaller one, and the right border
   *         is the greater one from the specified nodes.
   */
  public static Range getRangeInvariant(final Node left, final Node right, final boolean ascend) {

    final Node greater = ascend ? right : left;
    final Node less = ascend ? left : right;

    return new Range(less, greater);
  }
}
