/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.EventList;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Control flow graph case statement's data descriptor.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgCaseData {

  private final EventList list = new EventList();
  private final Set<NodeValue> values = new LinkedHashSet<>();
  private boolean valueFree;

  /**
   * Constructs data descriptor with the specified parameters.
   * @param events The list of events.
   * @param values The collection of same-type values.
   * @param valueFree The "free of values" indicator.
   */
  public CfgCaseData(
      final EventList events,
      final Collection<NodeValue> values,
      final boolean valueFree) {

    haveSameDataType(values);

    if (events != null) {
      this.list.addEvents(events);
    }
    this.values.addAll(values);
    this.valueFree = valueFree;
  }

  /**
   * Constructs data descriptor with the specified parameters.
   * @param events The list of events.
   * @param value The value.
   */
  public CfgCaseData(final EventList events, final NodeValue value) {

    if (events != null) {
      this.list.addEvents(events);
    }
    this.values.add(value);
    this.valueFree = false;
  }

  /**
   * Constructs data descriptor with the specified parameters.
   * @param events The list of events.
   * @param values The collection of same-type values.
   */
  public CfgCaseData(final EventList events, final Collection<NodeValue> values) {
    this(events, values, false);
  }

  /**
   * Constructs data descriptor with the specified "free-of-values" value.
   * @param valueFree The "free-of-values" flag.
   */
  public CfgCaseData(final boolean valueFree) {
    this.valueFree = valueFree;
  }

  /**
   * Constructs data descriptor with the specified value.
   * @param value The value.
   */
  public CfgCaseData(final NodeValue value) {
    this(null, value);
  }

  /**
   * Copy constructor.
   * @param data The data object which fields are to be copied.
   */
  public CfgCaseData(final CfgCaseData data) {

    if (data.hasEvents()) {

      this.list.addEvents(data.getEventList());
    }

    if (data.isValueFree()) {

      this.valueFree = data.isValueFree();

    } else {
      this.values.addAll(data.getValues());
    }
  }

  private static void haveSameDataType(final Collection<NodeValue> values) {

    DataType dataType = null;

    for (final NodeValue value : values) {

      if (dataType == null) {

        dataType = value.getDataType();

      } else if (!dataType.equals(value.getDataType())) {

        throw new RetrascopeException(
            String.format(
                "Argument should contain same data type but it contains: %s and %s.",
                dataType,
                value.getDataType()));
      }
    }
  }

  /**
   * Returns the collection of descriptor-related values.
   * @return The collection of descriptor-related values.
   */
  public Set<NodeValue> getValues() {
    return Collections.unmodifiableSet(this.values);
  }

  /**
   * Checks whether this descriptor does not contain any values.
   * @return {@code true} when this object does not contain values, {@code false} otherwise.
   */
  public boolean isValueFree() {
    return this.valueFree;
  }

  /**
   * Sets the specified value to "free of values" flag.
   * @param isDefault The value to be set to "free of values" flag.
   * @return {@code true} when this value has been successfully set, {@code false} otherwise.
   */
  public boolean setValueFree(final boolean isDefault) {

    if (!this.values.isEmpty() && isDefault) {
      return false;
    }
    this.valueFree = isDefault;
    return true;
  }

  /**
   * Adds the specified values to this descriptor.
   * @param value The value to be added to this descriptor.
   */
  public void add(final NodeValue value) {
    this.values.add(value);
  }

  /**
   * Removes the specified value from the descriptor.
   * @param value The value to be removed from the descriptor.
   */
  public void remove(final NodeValue value) {
    this.values.remove(value);
  }

  /**
   * Removes all the related values for this descriptor.
   */
  public void removeValues() {
    this.values.clear();
  }

  /**
   * Returns the list of events that is related to this descriptor.
   * @return The list of events that is related to this descriptor.
   */
  public EventList getEventList() {
    return this.list;
  }

  /**
   * Checks whether this descriptor contains events.
   * @return {@code true} when the descriptor contains at least one event, {@code false} otherwise.
   */
  public boolean hasEvents() {
    return this.list != null && !this.list.isEmpty();
  }

  /**
   * Returns the string description for this object in the format of the specified printer.
   * @param printer The printer that formats the description.
   * @return The string description for this object in the format of the specified printer.
   */
  public String getDescription(final ExprTreePrinter printer) {

    final Iterator<NodeValue> iterator = this.values.iterator();
    final String prefix =
        String.format(
            "%s%s",
            getEventList().isEmpty() ? "" : String.format("[%s] ", getEventList().toString()),
            iterator.hasNext() ? "" : "default ");
    final StringBuilder builder = new StringBuilder(prefix);

    while (iterator.hasNext()) {

      builder.append(printer.toString(iterator.next()));

      if (iterator.hasNext()) {
        builder.append(',');
      }
    }
    return builder.toString();
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final CfgCaseData that = (CfgCaseData) obj;

    return valueFree == that.valueFree
        && (list == null
            ? that.list == null && getValues().equals(that.getValues()) : list.equals(that.list));
  }

  @Override
  public int hashCode() {
    int result = list == null ? 0 : list.hashCode();
    result = 31 * result + getValues().hashCode();
    result = 31 * result + (valueFree ? 1 : 0);
    return result;
  }
}
