/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * Sequential statement that is enhanced with condition.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgConditionStatement extends CfgStatement {

  private Node condition;

  /**
   * Constructs statement with {@code null} condition.
   */
  CfgConditionStatement() {
    super();
    this.condition = null;
  }

  /**
   * Constructs statement with the specified condition.
   * @param condition The statement's condition.
   */
  CfgConditionStatement(final Node condition) {
    super();
    this.condition = condition;
  }

  /**
   * Constructs the statement that is a copy of the specified one.
   * @param statement The statement to be copied from.
   */
  CfgConditionStatement(final CfgConditionStatement statement) {
    super(statement);
    this.condition = statement.getCondition();
  }

  /**
   * Checks whether statement contains non-{@code null} condition.
   * @return {@code false} when condition is {@code null}, {@code true} otherwise.
   */
  public boolean hasCondition() {
    return this.condition != null;
  }

  /**
   * Returns {@link Node} condition.
   *
   * @return condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Sets condition to this statement.
   *
   * @param condition condition node
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  void setCondition(final Node condition) {
    InvariantChecks.checkNotNull(condition);

    this.condition = condition;
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    InvariantChecks.checkNotNull(printer);

    return printer.toString(getCondition());
  }

  @Override
  public Set<NodeVariable> getUses() {

    final Set<NodeVariable> uses = new LinkedHashSet<>();

    /* Get variables that are used in expression. */
    final Node expr = getCondition();
    final Collection<NodeVariable> usedExprVariables = ExprUtils.getVariables(expr);
    uses.addAll(usedExprVariables);

    uses.addAll(super.getUses());
    return uses;
  }

  @Override
  public void apply(final Map<String, Node> bindings) {

    if (hasCondition()) {
      for (final Map.Entry<String, Node> entry : bindings.entrySet()) {
        setCondition(Transformer.substitute(getCondition(), entry.getKey(), entry.getValue()));
      }
    }
    super.apply(bindings);
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final CfgConditionStatement that = (CfgConditionStatement) object;

    return (hasCondition() ? getCondition().equals(that.getCondition()) : !hasCondition());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (hasCondition() ? getCondition().hashCode() : 0);
    return result;
  }

  /**
   * Returns super-statement (sub-...parent) for the specified one with the specified fields.
   * @param statement The statement which parent is to be found.
   * @param condition The condition field for the statement to be found.
   * @return The conditional super-statement for the specified one that has the specified field
   *         if it exists, {@code null} otherwise.
   */
  public static CfgConditionStatement getSuperStatement(
      final Statement statement,
      final Node condition) {

    if (statement instanceof CfgConditionStatement) {

      final CfgConditionStatement condStatement = (CfgConditionStatement) statement;

      if (condition == null && !condStatement.hasCondition()
          || condStatement.getCondition().equals(condition)) {
        return (CfgConditionStatement) statement;
      }
    }

    Statement parent = statement.getParent();

    while (parent != null) {

      if (parent instanceof CfgConditionStatement) {

        final Node parentCond = ((CfgConditionStatement) parent).getCondition();
        if (parentCond == null && condition == null || parentCond.equals(condition)) {
          return (CfgConditionStatement) parent;
        }
      }

      parent = parent.getParent();
    }
    return null;
  }
}
