/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.ModelTransformUtils;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * Utility methods for control flow graph model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgUtils {

  private CfgUtils() {
    /* Empty. */
  }

  /**
   * Returns sub-statement with the specified identifier.
   *
   * @param statement The root statement.
   * @param stmtId The target statement identifier.
   *
   * @return Sub-statement that has the specified identifier.
   *
   * @throws NoSuchElementException when there is no such sub-statement with the specified
   *         identifier.
   */
  public static Statement getStatement(final Statement statement, final String stmtId) {

    class CfgGetByIdVisitor extends CfgEmptyVisitor {

      private final String stmtId;
      private Statement result;

      private CfgGetByIdVisitor(final String stmtId) {
        super();
        this.stmtId = stmtId;
        this.result = null;
      }

      private Statement getStatementWithId() {
        return this.result;
      }

      @Override
      public void onStatementBegin(final Statement statement) {
        if (statement.getId().equals(this.stmtId)) {
          this.result = statement;
        }
      }
    }

    final CfgGetByIdVisitor visitor = new CfgGetByIdVisitor(stmtId);
    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS_NO_RPT);
    walker.visitNode(statement);

    final Statement statementWithId = visitor.getStatementWithId();

    if (statementWithId == null) {
      throw new NoSuchElementException(
          String.format("%s node with '%s' id.", LogMessage.CANT_FIND, stmtId));
    }
    return statementWithId;
  }

  /**
   * Substitutes usages of variable with the specified name
   * by the specified expression in all the elements of the specified collection.
   * <p>
   * Substitutes usages of variable with the specified {@link java.lang.String} name
   * by the specified {@link ru.ispras.fortress.expression.Node} expression
   * in all the elements of the specified collection. The resulting collection
   * consists of node copies of the specified collection. In which node the every
   * usage of the variable with the specified name will be substituted by the
   * specified expression.
   * </p>
   *
   * @param list collection which nodes will be copied and transformed
   * @param name variable name
   * @param node expression
   *
   * @return collection of modified nodes
   */
  public static List<Statement> substitute(
      final List<Statement> list,
      final String name,
      final Node node) {

    final int listSize = list.size();
    final List<Statement> result = new LinkedList<>();
    final Map<Statement, Statement> map = new LinkedHashMap<>(listSize);

    /* Substitute variable with the specified name by the specified node. */
    for (final Statement elem : list) {

      final Statement copy = substituteStatement(elem, name, node);

      map.put(elem, copy);
      result.add(copy);
    }

    /* Establish parent-child links between nodes if they have them. */
    for (int i = 0; i < listSize; i++) {

      final Statement one = list.get(i);

      for (final Statement another : list) {

        if (one.contains(another)) {
          map.get(one).add(map.get(another));
        }
      }
    }

    return result;
  }

  private static Statement substituteStatement(
      final Statement elem,
      final String name,
      final Node node) {

    final Enum<?> elemType = elem.getType();

    Statement substitutedStmt;

    if (elemType.equals(CfgStatementType.SWITCH)) {

      final Node switchCondition = ((CfgSwitchStatement) elem).getCondition();
      final Node substitutedCond = Transformer.substitute(switchCondition, name, node);

      substitutedStmt = new CfgSwitchStatement(substitutedCond);

    } else if (elemType.equals(CfgStatementType.BASIC_BLOCK)) {

      final CfgBlockStatement blockElem = (CfgBlockStatement) elem;
      final List<Assignment> elemAssignments = blockElem.getAssignments();
      final List<Assignment> substAssignments = new LinkedList<>();

      for (final Assignment assign : elemAssignments) {

        final Node copyValue = Transformer.substitute(assign.getValue(), name, node);

        final RangedVariable target = assign.getTarget();

        if (target.ranged()) {

          final Range range = target.getRange();
          final Node high = range.getHigh();
          final Node low = range.getLow();

          final Node newHigh = TransformerUtils.substitute(high, name, node);
          final Node newLow = TransformerUtils.substitute(low, name, node);
          final Range newRange = new Range(newLow, newHigh);
          target.setRange(newRange);
        }

        if (target.indexed()) {

          final List<DataType> indexTypes = getIndexTypes(target.getVariable());
          final List<Node> newIndices = new LinkedList<>();

          final List<Node> indices = target.getIndices();
          for (final Node index : target.getIndices()) {
            newIndices.add(
                TypeConversion.coerce(
                    TransformerUtils.substitute(index, name, node),
                    indexTypes.get(indices.indexOf(index))));
          }
          target.setIndices(newIndices);
        }

        substAssignments.add(new Assignment(target, copyValue));
      }

      substitutedStmt = new CfgBlockStatement(substAssignments, blockElem.isConcurrent());

    } else if (elemType.equals(CfgStatementType.ASSERT)) {

      final CfgAssertStatement assertElem = (CfgAssertStatement) elem;
      final Node newCond = Transformer.substitute(assertElem.getCondition(), name, node);

      substitutedStmt =
          new CfgAssertStatement(newCond, assertElem.getSeverity(), assertElem.getReport());

    } else if (elemType.equals(CfgStatementType.WAIT)) {

      final CfgWaitStatement waitElem = (CfgWaitStatement) elem;
      substitutedStmt = new CfgWaitStatement(
          waitElem.getEventList(), Transformer.substitute(waitElem.getCondition(), name, node));

    } else if (elemType.equals(CfgStatementType.CASE)) {

      substitutedStmt =
          new CfgCaseStatement(
              substituteData(((CfgCaseStatement)elem).getCfgCaseData(), name, node));
    } else {
      throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, elem.getType());
    }

    for (int i = 0; i < elem.getStatementNum(); i++) {

      final Statement stmt = elem.getStatement(i);

      if (stmt == null) {
        throw new RetrascopeException("Null statement is found at: " + elem.toString());
      }

      substitutedStmt.add(substituteStatement(stmt, name, node));
    }

    return substitutedStmt;
  }

  private static List<DataType> getIndexTypes(NodeVariable variable) {

    final List<DataType> types = new LinkedList<>();

    DataTypeId typeId = variable.getDataTypeId();

    while (typeId.equals(DataTypeId.MAP)) {

      DataType type = variable.getDataType();
      types.add((DataType) type.getAttribute(DataTypeId.Attribute.KEY));

      DataType valueType = (DataType) type.getAttribute(DataTypeId.Attribute.VALUE);
      typeId = valueType.getTypeId();
    }

    return types;
  }

  private static CfgCaseData substituteData(
      final CfgCaseData data,
      final String name,
      final Node node) {

    final boolean valueFree = data.isValueFree();
    final Set<NodeValue> values = data.getValues();

    final EventList substituted = new EventList();

    for (final Event event : data.getEventList()) {

      final RangedVariable variable = substituteVariable(event.getRangedVariable(), name, node);
      substituted.add(new Event(variable, event.getType()));
    }

    return new CfgCaseData(substituted, values, valueFree);
  }

  private static RangedVariable substituteVariable(
      final RangedVariable rangedVariable,
      final String name,
      final Node node) {

    final NodeVariable variable = rangedVariable.getVariable();
    final String varName = variable.getName();

    if (varName.equals(name) && node.getKind() != Node.Kind.VARIABLE) {
      Log.warn("Loop unrolling: suspicious substitution: '%s' -> '%s'; ignored", name, node);
    }

    final RangedVariable substituted =
        new RangedVariable(
            node.getKind() == Node.Kind.VARIABLE
                ? (NodeVariable) Transformer.substitute(variable, name, node)
                : variable);

    if (rangedVariable.ranged()) {

      final Range range = rangedVariable.getRange();
      substituted.setRange(
          Transformer.substitute(range.getLow(), name, node),
          Transformer.substitute(range.getHigh(), name, node));
    }

    if (rangedVariable.indexed()) {

      final List<Node> indices = new LinkedList<>();
      for (final Node index : rangedVariable.getIndices()) {
        indices.add(Transformer.substitute(index, name, node));
      }
      substituted.setIndices(indices);
    }

    return substituted;
  }

  /**
   * Constructs an expression from the expression of the specified switch
   * {@link CfgSwitchStatement} node and it's possible values
   * in the specified case {@link CfgCaseStatement} node.
   *
   * @param caseNode case node
   * @param switchNode switch node
   *
   * @return expression
   */
  public static Node getExpression(
      final CfgCaseStatement caseNode,
      final CfgSwitchStatement switchNode) {

    Node expr;

    if (caseNode.isDefault()) {

      /* The 'default' case processing. */
      final Collection<Statement> children = StatementUtils.getSuccessors(switchNode);
      final List<NodeOperation> list = new LinkedList<>();

      for (final Statement child : children) {

        if (!child.equals(caseNode)) {
          list.add(constructExpr((CfgCaseStatement)child, switchNode));
        }
      }
      expr = ExprUtils.getComplement(list.toArray(new Node[0]));
    } else {

      /* The 'normal' case processing. */
      expr = constructExpr(caseNode, switchNode);
    }
    return TransformerUtils.standardize(expr);
  }

  private static NodeOperation constructExpr(
      final CfgCaseStatement child,
      final CfgSwitchStatement parent) {

    NodeOperation expr = null;

    for (final NodeValue value: child.getValues()) {

      if (expr == null) {

        expr = Nodes.eq(parent.getCondition(), value);

      } else {

        expr = Nodes.or(expr, Nodes.eq(parent.getCondition(), value));
      }
    }
    return expr;
  }

  /**
   * Creates weakest pre-condition for the specified post-condition and the specified
   * sequence of basic blocks.
   *
   * @param postCondition Expression that is formed after all the specified blocks.
   * @param blocks Sequence of basic blocks.
   * @return Weakest pre-condition.
   */
  private static Node getWeakestPreCondition(
      final Node postCondition,
      final List<CfgBlockStatement> blocks) {

    return ModelTransformUtils.substituteBackward(blocks, postCondition);
  }

  /**
   * Creates the weakest pre-condition for the specified post-condition and the specified
   * basic block.
   *
   * @param postCond The expression that is formed after the specified block.
   * @param block The basic block.
   * @return Weakest pre-condition.
   */
  public static Node getWeakestPreCondition(final Node postCond, final CfgBlockStatement block) {

    return getWeakestPreCondition(postCond, Collections.singletonList(block));
  }

  /**
   * Returns the post-condition for the specified sequence of block statements.
   * @param blocks The sequence of block statements.
   * @return The post-condition for the specified sequence of block statements.
   */
  public static Node getPostCondition(final CfgBlockStatement ... blocks) {

    final CfgContext cfgContext = new CfgContext(null, blocks).invoke(false);
    return cfgContext.getPerformedContext();
  }

  /**
   * Creates strongest post-condition for the specified conditions and the specified list
   * of block nodes.
   * <p>IMPORTANT NOTE: all the block assignments should be range-free.</p>
   *
   * @param pre Expression that is formed before all the specified blocks (pre-condition).
   * @param post Expression that is formed after all the specified blocks (post-condition).
   * @param blocks The sequence of basic blocks.
   *
   * @return Strongest post-condition.
   */
  public static Node getStrongestPostCondition(
      final Node pre,
      final Node post,
      final CfgBlockStatement ... blocks) {

    final CfgContext cfgContext = new CfgContext(pre, blocks).invoke(true);

    final Map<NodeVariable, Deque<NodeVariable>> versionMap = cfgContext.getVersionMap();
    final Map<NodeVariable, Node> context = cfgContext.getContext();
    final Node performedContext = cfgContext.getPerformedContext();
    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final Map.Entry<NodeVariable, Deque<NodeVariable>> entry : versionMap.entrySet()) {

      final NodeVariable key = entry.getKey();
      final NodeVariable value = entry.getValue().peek();

      if (!key.equals(value)) {
        boundVariables.add(NodeBinding.bindVariable(entry.getKey(), value));
      }
    }
    return context.isEmpty()
        ? post == null ? pre : Nodes.and(pre, NodeUtils.substituteAllBindings(post, boundVariables))
        : (post == null
            ? Nodes.and(pre, performedContext)
            : ExprUtils.getConjunction(
                  pre,
                  performedContext,
                  NodeUtils.substituteAllBindings(post, boundVariables)));
  }

  /**
   * Returns all the variables and their declarations that are defined at the specified statement
   * and their parents (if they're exist).
   * @param statement The statement for which all the variables that can be used in it are returned.
   * @return The collection of variables and their declarations that can be used or defined
   *         at the specified statement.
   */
  static Map<NodeVariable, Declaration> getDeclared(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    final Map<NodeVariable, Declaration> declared = new LinkedHashMap<>();

    Statement nearDecStatement = statement;

    while (!(nearDecStatement instanceof DecStatement)) {
      nearDecStatement = nearDecStatement.getParent();
    }

    addDeclared(declared, (DecStatement) nearDecStatement);

    return declared;
  }

  private static void addDeclared(
        final Map<NodeVariable, Declaration> declared,
        final DecStatement statement) {

    declared.putAll(statement.getDeclarationMap());

    final Statement parent = statement.getParent();

    if (parent instanceof DecStatement) {
      addDeclared(declared, (DecStatement) parent);
    }
  }

  /**
   * Applies the specified bindings to the specified statement.
   *
   * @param bindings The "variable's name - node" mapping.
   * @param stmt The statement that internals are to be substituted.
   */
  public static void apply(final Map<String, Node> bindings, final Assignment stmt) {

    final RangedVariable rangedVariable = stmt.getTarget();
    final String name = rangedVariable.getVariableName();

    RangedVariable newTarget = rangedVariable;

    /* If defined variable is used-in-bindings. */
    if (bindings.containsKey(name)) {

      final Node binding = bindings.get(name);

      if (bindings.get(name).getKind() == Node.Kind.VARIABLE) {

        final Range range = rangedVariable.getRange();
        final List<Node> indices = rangedVariable.getIndices();
        newTarget = new RangedVariable((NodeVariable) binding, range, indices);

      } else {
        throw RetrascopeException.exception(LogMessage.WRONG_TYPE, binding.getKind());
      }
    }

    final Set<String> usedVarNames = stmt.getUsedVariableNames();

    for (final String usedVarName : usedVarNames) {

      /* If used-in-expression variable is used-in-bindings. */
      if (bindings.containsKey(usedVarName)) {

        final Node bindVar = bindings.get(usedVarName);
        stmt.setValue(Transformer.substitute(stmt.getValue(), usedVarName, bindVar));

        if (stmt.ranged()) {
          final Node newLow =
              Transformer.substitute(stmt.getRange().getLow(), usedVarName, bindVar);
          final Node newHigh =
              Transformer.substitute(stmt.getRange().getHigh(), usedVarName, bindVar);
          newTarget.setRange(newLow, newHigh);
        }

        if (stmt.indexed()) {

          final List<Node> newIndices = new LinkedList<>();

          for (final Node index : stmt.getIndices()) {
            newIndices.add(Transformer.substitute(index, usedVarName, bindVar));
          }

          newTarget.setIndices(newIndices);
        }
      }
    }
    stmt.setVariable(newTarget);
  }
}
