/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.ModuleProcess;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Extended finite state machine.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class Efsm extends ModuleProcess {

  private static final String SRC_NOT_FOUND =
      "The specified source state of a new transition is not found in this EFSM: ";

  private static final String DST_NOT_FOUND =
      "The specified destination state of a new transition is not found in this EFSM: ";

  private static final String NO_VAR =
      "There is no declaration of variable neither in this EFSM nor in its ancestors: ";

  /**
   * Transition weight types.
   */
  public enum TransitionWeightBasis {

    /**
     * All transitions are considered as having one and the same weight.
     */
    CONSTANT,

    /**
     * A transition's weight is equal to amount of registers used in the transition's guard.
     */
    GUARD_REGISTER_AMOUNT
  }

  /**
   * An object representing a state based on uninitialised values.
   */
  public static final EfsmState UNINIT_STATE = new EfsmState(Nodes.FALSE);

  /**
   * A string prefix for "state not found" error message.
   */
  public static final String STATE_NOT_FOUND = "The specified state is not found in this EFSM: ";

  private final Set<EfsmState> states;
  private final Set<EfsmTransition> transitions;

  private EfsmState initialState;
  private final List<GuardedAction> resets;
  private boolean testAsSingle;

  /**
   * Constructs an empty EFSM with the specified name and parent module.
   *
   * @param parent the module which contains the constructed EFSM
   * @param name the EFSM name
   *
   * @throws IllegalArgumentException if either {@code parent}
   *         or {@code name} is {@code null}
   */
  public Efsm(
      final EfsmModule parent,
      final String name) {
    super(name);

    InvariantChecks.checkNotNull(parent);
    InvariantChecks.checkNotNull(name);

    setParent(parent);

    this.states = new LinkedHashSet<>();
    this.transitions = new LinkedHashSet<>();
    this.initialState = null;
    this.resets = new LinkedList<>();

    setSingleTestable(true);
  }

  /**
   * Returns a deep copy of the object.
   * @return A deep copy of the object.
   */
  public Efsm deepCopy() {
    return new Efsm(this);
  }

  private Efsm(final Efsm efsm) {
    super(efsm);

    /* These fields are simply copied. */
    setParent(efsm.getParentContainer());
    this.resets = new LinkedList<>(efsm.getResetGuardedActions());

    /* These fields are deeply copied. */

    final Map<EfsmState, EfsmState> oldNewMap = new LinkedHashMap<>();

    this.transitions = new LinkedHashSet<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {

      final EfsmState src = transition.getSrc();
      final EfsmState srcDeepCopy = src.deepCopy();
      final boolean containSrc = oldNewMap.containsKey(src);
      final EfsmState srcCopy = containSrc ? oldNewMap.get(src) : srcDeepCopy;

      if (!containSrc) {
        oldNewMap.put(src, srcDeepCopy);
      }

      final EfsmState dst = transition.getDst();
      final EfsmState dstDeepCopy = dst.deepCopy();
      final boolean containDst = oldNewMap.containsKey(dst);
      final EfsmState dstCopy = containDst ? oldNewMap.get(dst) : dstDeepCopy;

      if (!containDst) {
        oldNewMap.put(dst, dstDeepCopy);
      }

      this.transitions.add(new EfsmTransition(srcCopy, dstCopy, transition));
    }

    this.states = new LinkedHashSet<>();

    for (final EfsmState state : efsm.getStates()) {

      final EfsmState stateDeepCopy = state.deepCopy();
      final boolean containState = oldNewMap.containsKey(state);
      this.states.add(containState ? oldNewMap.get(state) : stateDeepCopy);

      if (!containState) {
        oldNewMap.put(state, stateDeepCopy);
      }
    }

    final EfsmState initState = efsm.getInitialState();
    final EfsmState initDeepCopy = initState == null ? null : initState.deepCopy();
    final boolean containInitCopy = efsm.hasInitState() && oldNewMap.containsKey(initState);
    this.initialState = containInitCopy ? oldNewMap.get(initState) : initDeepCopy;
    setSingleTestable(efsm.singleTestable());
  }

  /**
   * Checks whether this object can be tested as single one, or not.
   * <p>If the object is single testable, it means, that it can be run in an isolated
   * environment. Either it does not have any dependencies from other EFSM objects, or they're
   * negligible.</p>
   * @return {@code true} if this object can be tested as a single one,
   *         {@code false} otherwise.
   */
  public boolean singleTestable() {
    return this.testAsSingle;
  }

  /**
   * Make the object to be non or single testable depending on the value of the specified parameter.
   * @param singleTestable The single testability flag.
   */
  public void setSingleTestable(final boolean singleTestable) {
    this.testAsSingle = singleTestable;
  }

  /**
   * Returns parent container of this EFSM.
   *
   * @return the parent EFSM container
   */
  public EfsmModule getParentContainer() {
    return (EfsmModule) getParent();
  }

  /**
   * Creates a new state based on the specified logical expression and adds it into this EFSM. All
   * variables which are in the expression must be declared either in this EFSM or in its parent
   * before the creation.
   *
   * @param expression the expression representing the state
   *
   * @return the created state
   *
   * @throws IllegalArgumentException if at least one variable which is in the expression is not
   *         declared neither in this EFSM nor in its parent
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EfsmState addState(final Node expression) {
    final EfsmState state = new EfsmState(expression);
    this.addState(state);

    return state;
  }

  /**
   * Adds the specified state into this EFSM.
   * <p>Do nothing when the specified state is already added to this object.</p>
   * @param state the state to be added
   *
   * @throws NullPointerException if {@code state} is {@code null}
   * @throws IllegalArgumentException if at least one variable which is in the expression is not
   *         declared neither in this EFSM nor in its parent
   * @throws IllegalArgumentException when argument is {@code null}
   */
  public void addState(final EfsmState state) {
    InvariantChecks.checkNotNull(state);

    this.processAddedExpression(state.getNode());

    this.states.add(state);
  }

  /**
   * Returns the set of state variables that are used in this object.
   * @return The set of state variables that are used in this object.
   */
  public Set<NodeVariable> getStateVariables() {

    final Set<NodeVariable> variables = new LinkedHashSet<>();

    for (final EfsmState state : this.states) {

      final Collection<NodeVariable> stateVariables = ExprUtils.getVariables(state.getNode());
      variables.addAll(stateVariables);
    }

    return variables;
  }

  /**
   * Creates a new transition with the specified guarded action between the specified states and
   * adds it into this EFSM. All variables which are in the guarded action must be declared either
   * in this EFSM or in its parent before the creation.
   *
   * @param src a source state
   * @param dst a destination state
   * @param guardedAction the guarded action
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM or at least one variable which is in {@code guardedAction} is not
   *         declared neither in this EFSM nor in its parent
   */
  public void addTransition(
      final EfsmState src,
      final EfsmState dst,
      final GuardedAction guardedAction) {

    if (!this.states.contains(src)) {
      throw new IllegalArgumentException(SRC_NOT_FOUND + src);
    }

    if (!this.states.contains(dst)) {
      throw new IllegalArgumentException(DST_NOT_FOUND + dst);
    }

    final EfsmTransition transition = new EfsmTransition(src, dst, guardedAction);
    this.transitions.add(transition);
    this.processAddedExpression(guardedAction.getGuardNode());

    for (final Assignment assignment: guardedAction.getAction().getAssignments()) {

      this.processAddedExpression(assignment.getValue());

      if (assignment.getTarget() != null) {
        this.processAddedExpression(assignment.getVariable());
      }
    }
  }

  /**
   * Checks if the specified state is in this EFSM.
   *
   * @param state the state whose presence in the EFSM to be tested
   *
   * @return {@code true} if this EFSM contains the specified state. Otherwise returns {@code false}
   */
  public boolean containsState(final EfsmState state) {
    return this.states.contains(state);
  }

  /**
   * Marks the specified state of this EFSM as an initial state.
   *
   * @param state the new initial state of this EFSM
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public void setInitialState(final EfsmState state) {

    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(STATE_NOT_FOUND + state);
    }

    this.initialState = state;
  }

  /**
   * Returns an initial state of this EFSM.
   *
   * @return the initial state of this EFSM
   */
  public EfsmState getInitialState() {
    return this.initialState;
  }

  /**
   * Adds the specified guarded action to this EFSM. This guarded action represents a precondition
   * of the initial state and an action which is performed during initialisation of this EFSM.
   *
   * @param rstGuardedAction the reset guarded action
   *
   * @throws IllegalArgumentException if the {@code resetGuardedAction} is {@code null}
   */
  public void addResetGuardedAction(final GuardedAction rstGuardedAction) {
    InvariantChecks.checkNotNull(rstGuardedAction);

    /* If variable is used at reset guarded action guard, it cannot be in model clocks. */
    final Collection<NodeVariable> resetVariables =
        ExprUtils.getVariables(rstGuardedAction.getGuardNode());

    final Collection<NodeVariable> filtered = new LinkedList<>();

    for (final NodeVariable reset : resetVariables) {

      if (getClocks().contains(reset)) {
        filtered.add(reset);
      }
    }
    clearEvents(filtered);

    this.resets.add(rstGuardedAction);
  }

  /**
   * Returns reset guarded actions of this EFSM, that is, such guarded actions which represent an
   * initialisation of this EFSM.
   *
   * @return the reset guarded actions
   */
  public List<GuardedAction> getResetGuardedActions() {
    return this.resets;
  }

  /**
   * Add the specified object as clock-like variable for this model.
   * <p>When the specified variable is used in a guard of some resetting transition,
   * this method does nothing, because a variable cannot be placed both at clocks and resets.</p>
   * @param clockVariable clock-like variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClock(final NodeVariable clockVariable) {
    InvariantChecks.checkNotNull(clockVariable);

    if (!usedAsReset(clockVariable)) {
      addEvent(new Event(clockVariable, EventType.ANY_EDGE));
    }
  }

  private boolean usedAsReset(final NodeVariable variable) {

    for (final GuardedAction reset : getResetGuardedActions()) {
      if (ExprUtils.getVariables(reset.getGuardNode()).contains(variable)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns clock-like variables related to this model.
   * @return clock-like variables
   */
  public Collection<NodeVariable> getClocks() {
    final Collection<NodeVariable> clocks = new LinkedHashSet<>();
    final EventList eventList = getEventList();

    final Collection<NodeVariable> resets = getResets();

    for (final Event event : eventList) {
      final NodeVariable variable = event.getVariable();
      if (!resets.contains(variable)) {
        clocks.add(variable);
      }
    }
    return clocks;
  }

  private Collection<NodeVariable> getResets() {
    final Collection<NodeVariable> resets = new LinkedHashSet<>();
    final List<GuardedAction> rstGuardedActions = getResetGuardedActions();

    for (final GuardedAction rstGuardedAction : rstGuardedActions) {
      resets.addAll(ExprUtils.getVariables(rstGuardedAction.getClockGuardNode()));
    }

    return resets;
  }

  /**
   * Returns all states of this EFSM.
   *
   * @return a collection containing the states of this EFSM
   */
  public Collection<EfsmState> getStates() {
    return this.states;
  }

  /**
   * Returns the number of states for this EFSM.
   * @return The number of states for this EFSM.
   */
  public int getStateNum() {
    return this.states.size();
  }

  /**
   * Returns the number of incoming transitions for the specified state.
   * @param state The model state.
   * @return The number of incoming transitions for the specified state.
   */
  public int getInCount(final EfsmState state) {

    int inCount = 0;

    for (final EfsmTransition transition : getTransitions()) {

      if (transition.getDst().equals(state)) {
        inCount++;
      }
    }

    return inCount;
  }

  /**
   * Returns the number of outgoing transitions for the specified state.
   * @param state The model state.
   * @return The number of outgoing transitions for the specified state.
   */
  public int getOutCount(final EfsmState state) {

    int outCount = 0;

    for (final EfsmTransition transition : getTransitions()) {

      if (transition.getSrc().equals(state)) {
        outCount++;
      }
    }

    return outCount;
  }

  /**
   * Removes the specified state from this EFSM object.
   *
   * @param state State to ber removed.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void remove(final EfsmState state) {
    InvariantChecks.checkNotNull(state);

    this.states.remove(state);
  }

  /**
   * Removes the specified transition from this object.
   *
   * @param transition Transition to be removed from this object.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void remove(final EfsmTransition transition) {
    InvariantChecks.checkNotNull(transition);

    this.transitions.remove(transition);
  }

  /**
   * Returns all transitions of this EFSM.
   *
   * @return a collection containing the transitions of this EFSM
   */
  public Collection<EfsmTransition> getTransitions() {
    return this.transitions;
  }

  /**
   * Returns the number of transitions for this EFSM.
   *
   * @return The number of transitions for this EFSM.
   */
  public int getTransitionNum() {
    return this.transitions.size();
  }

  private void processAddedExpression(final Node expression) {

    if (expression != null) {

      for (final NodeVariable variable: ExprUtils.getVariables(expression)) {

        final String variableName = variable.getName();

        if (!EfsmUtils.hasDeclaringAncestor(this, variable)) {
          throw new IllegalArgumentException(NO_VAR + variableName);
        }
      }
    }
  }

  /**
   * Checks rather this object has non-{@code null} initial state.
   * @return {@code true} if this object contains initial state, {@code false} otherwise.
   */
  public boolean hasInitState() {
    return this.initialState != null;
  }

  /**
   * Checks if there is a path from the specified source state to the specified destination state
   * at the model.
   *
   * @param sourceState The source state.
   * @param destinationState The destination state.
   *
   * @return {@code true} if there is a path from the source state to the destination state.
   *         Otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM
   */
  public boolean isReachable(
      final EfsmState sourceState,
      final EfsmState destinationState) {

    if (!containsState(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!containsState(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return true;
    }

    final List<EfsmState> visitedStates = new LinkedList<>();
    final LinkedList<EfsmState> queue = new LinkedList<>();
    visitedStates.add(sourceState);
    queue.add(sourceState);

    while (!queue.isEmpty()) {

      for (final EfsmState nextState : getDestStates(queue.poll())) {

        if (nextState.equals(destinationState)) {
          return true;
        }

        if (!visitedStates.contains(nextState)) {
          queue.add(nextState);
          visitedStates.add(nextState);
        }
      }
    }

    return false;
  }

  /**
   * Returns transitions of the EFSM that end at the specified state.
   *
   * @param state The model state.
   *
   * @return Collection containing incoming transitions for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Collection<EfsmTransition> getIncoming(final EfsmState state) {

    if (!containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final List<EfsmTransition> incoming = new LinkedList<>();

    for (final EfsmTransition transition : getTransitions()) {

      if (transition.getDst().equals(state)) {
        incoming.add(transition);
      }
    }

    return incoming;
  }

  /**
   * Returns transitions of the EFSM that start at the specified state.
   *
   * @param state The model state.
   *
   * @return A list containing outgoing transitions for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Collection<EfsmTransition> getOutgoing(final EfsmState state) {

    if (!containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final List<EfsmTransition> outgoing = new LinkedList<>();

    for (final EfsmTransition transition: getTransitions()) {

      if (transition.getSrc().equals(state)) {
        outgoing.add(transition);
      }
    }

    return outgoing;
  }

  /**
   * Returns states which are end points of outgoing transitions of the specified state.
   *
   * @param state The model state.
   *
   * @return A list containing the destination states for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  private List<EfsmState> getDestStates(final EfsmState state) {

    if (!containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final Collection<EfsmTransition> outgoing = getOutgoing(state);

    if (outgoing.isEmpty()) {
      return Collections.emptyList();
    }

    final List<EfsmState> destinationStates = new LinkedList<>();

    for (final EfsmTransition transition : outgoing) {
      destinationStates.add(transition.getDst());
    }

    return destinationStates;
  }

  /**
   * Looks for the shortest path from the specified source state
   * to the specified destination state with the specified principle of transition
   * weight defining. This method is based on the Dijkstra's algorithm.
   * <br>
   * Currently CONST and GUARD_REGISTER_AMOUNT TransitionWeightBasis types are supported.
   *
   * @param src the source state
   * @param dst the destination state
   * @param weightBasis the principle used for transition weight computing
   *
   * @return a list containing the transitions of found path.
   *         It's empty if the source state matches
   *         to the destination state. If there is no path
   *         from the source state to the destination
   *         state {@code null} is returned
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM or if unsupported
   *         {@link ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis} is passed.
   */
  public List<EfsmTransition> getShortestPath(
      final EfsmState src,
      final EfsmState dst,
      final Efsm.TransitionWeightBasis weightBasis) {

    if (!containsState(src)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!containsState(dst)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (src.equals(dst)) {
      return Collections.emptyList();
    }

    if (!isReachable(src, dst)) {
      return null;
    }

    final int stateNum = getStateNum();
    final LinkedHashMap<EfsmState, Integer> distances = new LinkedHashMap<>(stateNum);
    final LinkedHashMap<EfsmState, EfsmTransition> previousNodes = new LinkedHashMap<>(stateNum);

    for (final EfsmState state : getStates()) {
      distances.put(state, Integer.MAX_VALUE);
      previousNodes.put(state, null);
    }

    distances.put(src, 0);
    final LinkedHashMap<EfsmTransition, Integer> transitionWeights = new LinkedHashMap<>();

    for (final EfsmTransition transition : getTransitions()) {

      switch (weightBasis) {

        case CONSTANT:
          transitionWeights.put(transition, 1);
          break;

        case GUARD_REGISTER_AMOUNT:
          int weight = 0;

          for (final NodeVariable variable : ExprUtils.getVariables(transition.getGuardNode())) {
            final String variableName = variable.getName();
            if (getParentContainer().declaresRegister(variableName)) {
              weight++;
            }
          }

          transitionWeights.put(transition, weight);
          break;

        default:
          throw new IllegalArgumentException("Unsupported weight defining principle: "
              + weightBasis);
      }
    }

    final Comparator<EfsmTransition> transComparator =
        Comparator.comparingInt(transitionWeights::get);

    final Comparator<EfsmState> stateComparator = Comparator.comparingInt(distances::get);
    final List<EfsmState> visitedStates = new LinkedList<>();
    visitedStates.add(src);
    final LinkedList<EfsmState> queue = new LinkedList<>();
    queue.add(src);

    while (!queue.isEmpty()) {

      final EfsmState currentState = queue.poll();
      final int currentDistance = distances.get(currentState);
      final LinkedList<EfsmTransition> nextTrans = new LinkedList<>(getOutgoing(currentState));

      nextTrans.sort(transComparator);

      for (final EfsmTransition nextTransition : nextTrans) {

        final EfsmState nextState = nextTransition.getDst();
        final int newDistance = currentDistance + transitionWeights.get(nextTransition);

        if (newDistance < distances.get(nextState)) {

          previousNodes.put(nextState, nextTransition);
          distances.put(nextState, newDistance);

          if (nextState.equals(dst)) {
            break;
          }

          if (!visitedStates.contains(nextState)) {
            visitedStates.add(nextState);
            queue.add(nextState);
          }
        }
      }

      queue.sort(stateComparator);
    }

    final LinkedList<EfsmTransition> result = new LinkedList<>();
    EfsmState currentState = dst;

    while (previousNodes.get(currentState) != null) {

      result.addFirst(previousNodes.get(currentState));
      currentState = previousNodes.get(currentState).getSrc();
    }

    return result;
  }

  @Override
  public Set<NodeVariable> getUses() {

    final Set<NodeVariable> uses = new LinkedHashSet<>(getStateVariables());

    for (final EfsmTransition transition : getTransitions()) {
      uses.addAll(transition.getUses());
    }

    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {

    final Set<NodeVariable> defines = new LinkedHashSet<>();

    for (final EfsmTransition transition : getTransitions()) {
      defines.addAll(transition.getDefines());
    }
    return defines;
  }

  @Override
  public String toString() {
    return ((EfsmModule)getParent()).getName() + "." + getName();
  }
}
