/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents a High-Level Decision Diagram (HLDD) of a variable assignment function.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class Hldd {

  // Diagram nodes.
  private final Collection<HlddNode> nodes;
  // Terminal nodes.
  private final Collection<HlddNode> terminalNodes;
  // Assigned variable.
  private NodeVariable variable;
  // Root node.
  private HlddNode rootNode;
  // Immediate assignment flag.
  private boolean isImmAssignment;

  /**
   * Constructs a new HLDD using the specified variable.
   *
   * @param variable - the specified variable.
   */
  public Hldd(final NodeVariable variable) {
    setVariable(variable);
    nodes = new ArrayList<>();
    terminalNodes = new ArrayList<>();
  }

  /**
   * Constructs a new HLDD using the specified values.
   *
   * @param variable - the specified variable.
   * @param nodes - the specified collection of nodes.
   */
  public Hldd(final NodeVariable variable, final Collection<HlddNode> nodes) {
    setVariable(variable);
    this.nodes = new ArrayList<>();
    terminalNodes = new ArrayList<>();
    addNodes(nodes);
  }

  /**
   * Constructs a new HLDD.
   */
  public Hldd() {
    nodes = new ArrayList<>();
    terminalNodes = new ArrayList<>();
  }

  /**
   * Sets the root node.
   *
   * @param root - the specified node.
   */
  public void setRootNode(final HlddNode root) {
    rootNode = root;
  }

  /**
   * Returns the root node.
   *
   * @return The root node.
   */
  public HlddNode getRootNode() {
    return rootNode;
  }

  /**
   * Adds the specified node to the diagram.
   *
   * @param node - the specified node.
   * @return This diagram.
   */
  public Hldd addNode(final HlddNode node) {
    InvariantChecks.checkNotNull(node);

    nodes.add(node);
    if (node.isTerminal()) {
      terminalNodes.add(node);
    }
    return this;
  }

  /**
   * Adds the collection of nodes to the diagram.
   *
   * @param nodes - the specified collection of nodes.
   */
  private void addNodes(final Collection<HlddNode> nodes) {
    for (HlddNode node : nodes) {
      addNode(node);
    }
  }

  /**
   * Returns the diagram's nodes.
   *
   * @return The collection of nodes.
   */
  public Collection<HlddNode> getNodes() {
    return new ArrayList<>(nodes);
  }

  // Searches for the nodes following the start node (all or terminal only).
  private Collection<HlddNode> getNodes(final HlddNonTerminalNode start, final boolean terminal) {
    InvariantChecks.checkNotNull(start);
    final ArrayList<HlddNode> followingNodes = new ArrayList<>();
    for (HlddNode child : start.getChildren()) {
      if (!terminal || child.isTerminal()) {
        followingNodes.add(child);
      }
      if (child.isNonTerminal()) {
        followingNodes.addAll(getNodes((HlddNonTerminalNode) child, terminal));
      }
    }
    return followingNodes;
  }

  /**
   * Returns the terminal nodes.
   *
   * @return The collection of terminal nodes.
   */
  public Collection<HlddNode> getTerminalNodes() {
    return Collections.unmodifiableCollection(terminalNodes);
  }

  /**
   * Returns the following nodes of the specified node.
   *
   * @param start - the specified starting node.
   * @return A collection of the following nodes.
   */
  public Collection<HlddNode> getSuccessors(final HlddNode start) {
    if (start.isNonTerminal()) {
      return getNodes((HlddNonTerminalNode) start, false);
    }
    return new ArrayList<>();
  }

  /**
   * Returns the following terminal nodes of the specified node.
   *
   * @param start The specified starting node.
   * @return A collection of the following terminal nodes.
   */
  public Collection<HlddNode> getTerminalSuccessors(final HlddNode start) {
    if (start.isNonTerminal()) {
      return getNodes((HlddNonTerminalNode) start, true);
    }
    return new ArrayList<>();
  }

  /**
   * Removes the specified nodes from the HLDD.
   *
   * @param nodes - the specified collection of nodes.
   */
  public void removeNodes(final Collection<HlddNode> nodes) {
    InvariantChecks.checkNotNull(nodes);
    for (HlddNode node : nodes) {
      removeNode(node);
    }
  }

  /**
   * Removes the specified node from the HLDD.
   *
   * @param node - the specified node.
   */
  public void removeNode(final HlddNode node) {
    this.nodes.remove(node);
    if (node.isTerminal()) {
      terminalNodes.remove(node);
    }
    if (node.getParent() != null) {
      final HlddNonTerminalNode parent = (HlddNonTerminalNode) node.getParent();
      parent.removeChild(node);
    }
  }

  /**
   * Returns a deep copy of the diagram using the specified associated variable.
   *
   * @return A deep copy.
   */
  public Hldd deepCopy() {
    final Hldd result = new Hldd(variable);
    final Map<HlddNode, HlddNode> nodeCopies = copyNodes();
    connectCopies(nodeCopies);
    result.setRootNode(nodeCopies.get(this.getRootNode()));
    result.addNodes(nodeCopies.values());
    return result;
  }

  private void connectCopies(final Map<HlddNode, HlddNode> nodeCopies) {
    for (Entry<HlddNode, HlddNode> entry : nodeCopies.entrySet()) {
      final HlddNode node = entry.getKey();
      final HlddNode nodeCopy = entry.getValue();
      nodeCopy.setParent(nodeCopies.get(node.getParent()));
      if (node.isNonTerminal()) {
        setChildren((HlddNonTerminalNode) node, (HlddNonTerminalNode) nodeCopy, nodeCopies);
      }
    }
  }

  private Map<HlddNode, HlddNode> copyNodes() {
    final Map<HlddNode, HlddNode> nodeCopies = new HashMap<>();
    for (HlddNode node : nodes) {
      nodeCopies.put(node, node.deepCopy());
    }
    return nodeCopies;
  }

  private void setChildren(final HlddNonTerminalNode node, final HlddNonTerminalNode copy,
      final Map<HlddNode, HlddNode> nodeCopies) {
    for (HlddNode child : node.getChildren()) {
      copy.addChild(nodeCopies.get(child));
    }
  }

  /**
   * Returns the associated variable.
   *
   * @return The associated variable.
   */
  public NodeVariable getVariable() {
    return variable;
  }

  /**
   * Sets the associated variable.
   *
   * @param variable - the specified variable.
   */
  public void setVariable(final NodeVariable variable) {
    this.variable = variable;
  }

  /**
   * Shows whether the HLDD implements an immediate assignment.
   *
   * @return {@code true} if the HLDD implements condition, {@code false} otherwise.
   */
  public boolean isImmediateAssignment() {
    return isImmAssignment;
  }

  /**
   * Sets an immediate assignment property.
   *
   * @param isGuard - the specified value.
   */
  public void setImmediateAssignment(final boolean isGuard) {
    this.isImmAssignment = isGuard;
  }

}
