/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This class represents a process of a HLDD model. Contains a collection of HLDDs and the
 * corresponding variables.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddProcess extends ModuleProcess {

  // HLDDs of the process.
  private final Collection<Hldd> hldds;
  // Process-level invariant.
  private Node invariant;

  /**
   * Constructs a new process with the specified name.
   *
   * @param name - the specified name.
   */
  public HlddProcess(final String name) {
    super(name);
    hldds = new ArrayList<>();
  }

  /**
   * Constructs a new process using the specified process.
   *
   * @param process - the specified process.
   */
  public HlddProcess(final ModuleProcess process) {
    super(process.getId(), process.getEventList(), process.isInitial());
    hldds = new ArrayList<>();
    final Enum<?> infoType = MetaInfoType.PHASE_VAR_NAME;
    if (process.hasMetaInfo(infoType)) {
      this.addMetaInfo(infoType, process.getMetaInfo(infoType));
    }
  }

  /*
   * Copy constructor.
   */
  private HlddProcess(final HlddProcess process) {
    super(process.getName(), process.getEventList(), process.isInitial());
    this.hldds = new ArrayList<>();
    add(process.getHldds());
    setInvariant(process.getInvariant());
  }

  /**
   * Adds a collection of HLDDs to this process.
   *
   * @param hldds - the specified collection.
   */
  private void add(Collection<Hldd> hldds) {
    for (Hldd hldd : hldds) {
      add(hldd);
    }
  }

  /**
   * Adds a HLDD to this process.
   *
   * @param hldd - the specified HLDD.
   */
  public void add(final Hldd hldd) {
    InvariantChecks.checkNotNull(hldd);
    hldds.add(hldd);
    if (hldd.getVariable() != null
        && !checkDeclaration(hldd.getVariable(), (DecStatement) getParent())) {
      ((DecStatement)getParent()).declareVariable(hldd.getVariable());
    }
  }

  private boolean checkDeclaration(final NodeVariable variable, final DecStatement statement) {
    return statement.declares(variable)
        || statement.getParent() != null
            && checkDeclaration(variable, (DecStatement) statement.getParent());
  }

  /**
   * Removes the specified HLDDs from this process.
   *
   * @param hldds - the specified collection of HLDDs.
   */
  public void removeHldds(final Collection<Hldd> hldds) {
    this.hldds.removeAll(hldds);
  }

  /**
   * Returns the variables that are assigned in this process.
   *
   * @return The collection of assigned variables.
   */
  public List<NodeVariable> getAssignedVariables() {
    final List<NodeVariable> result = new ArrayList<>();
    for (Hldd hldd : hldds) {
      result.add(hldd.getVariable());
    }
    return result;
  }

  /**
   * Returns the contents of the container.
   *
   * @return The collection of objects.
   */
  public Collection<Hldd> getHldds() {
    final Collection<Hldd> result = new ArrayList<>(hldds);
    ((ArrayList<Hldd>) result).sort((o1, o2) -> {
      final String name1 = o1.getVariable().getName();
      final String name2 = o2.getVariable().getName();
      return name1.compareTo(name2);
    });
    return result;
  }

  /**
   * Returns the HLDD for the specified variable.
   *
   * @param variable - the specififed target variable.
   * @return The specific HLDD, if exists, {@code null} otherwise.
   */
  public Hldd getHldd(final NodeVariable variable) {
    for (Hldd hldd : hldds) {
      if (hldd.getVariable().equals(variable)) {
        return hldd;
      }
    }
    return null;
  }

  @Override
  public Statement deepCopy() {
    return new HlddProcess(this);
  }

  /**
   * Returns the process-level invariant.
   *
   * @return The invariant node.
   */
  public Node getInvariant() {
    return invariant;
  }

  /**
   * Sets the specified expression as a process-level invariant.
   *
   * @param invariant - the specified expression node.
   */
  public void setInvariant(final Node invariant) {
    this.invariant = invariant;
  }
}
