/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.Node;

/**
 * This class represents the terminal node of a HLDD.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddTerminalNode extends HlddNode {

  // Target variable.
  private Node targetVariable;
  // Assigned expression.
  private Node assignedValue;

  /**
   * Constructs a new terminal node.
   *
   * @param parent - the specified parent.
   */
  public HlddTerminalNode(final HlddNode parent) {
    super(parent, NodeType.ASSIGNMENT);
  }

  /**
   * Constructs a new terminal node.
   *
   * @param parent   - the specified parent.
   * @param variable - the specified target variable.
   * @param value    - the specified assigned expression.
   */
  public HlddTerminalNode(final HlddNode parent, final Node variable, final Node value) {
    super(parent, NodeType.ASSIGNMENT);
    setTargetVariable(variable);
    setAssignedValue(value);
  }

  /**
   * Constructs a new terminal node.
   *
   * @param variable - the specified target variable.
   * @param value    - the specified assigned expression.
   */
  public HlddTerminalNode(final Node variable, final Node value) {
    super(null, NodeType.ASSIGNMENT);
    setTargetVariable(variable);
    setAssignedValue(value);
  }

  /**
   * Constructs a new terminal node.
   */
  public HlddTerminalNode() {
    super(NodeType.ASSIGNMENT);
  }

  private HlddTerminalNode(final HlddTerminalNode node) {
    super(node);
    setTargetVariable(node.getTargetVariable());
    setAssignedValue(node.getAssignedValue());
  }

  /**
   * Returns the target variable.
   *
   * @return The target variable.
   */
  public Node getTargetVariable() {
    return targetVariable;
  }

  /**
   * Sets the target variable.
   *
   * @param targetVariable - the specified target variable.
   */
  public void setTargetVariable(final Node targetVariable) {
    this.targetVariable = targetVariable;
  }

  /**
   * Returns the assigned expression.
   *
   * @return The assigned expression.
   */
  public Node getAssignedValue() {
    return assignedValue;
  }

  /**
   * Sets the assigned expression.
   *
   * @param assignedValue - - the specified assigned expression.
   */
  public void setAssignedValue(final Node assignedValue) {
    this.assignedValue = assignedValue;
  }

  @Override
  public HlddNode deepCopy() {
    return new HlddTerminalNode(this);
  }

  @Override
  public boolean isNonTerminal() {
    return false;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((assignedValue == null) ? 0 : assignedValue.hashCode());
    result = prime * result + ((targetVariable == null) ? 0 : targetVariable.hashCode());
    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (!super.equals(obj)) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    final HlddTerminalNode other = (HlddTerminalNode) obj;
    if (assignedValue == null) {
      if (other.assignedValue != null) {
        return false;
      }
    } else if (!assignedValue.equals(other.assignedValue)) {
      return false;
    }
    if (targetVariable == null) {
      return other.targetVariable == null;
    } else {
      return targetVariable.equals(other.targetVariable);
    }
  }

  @Override
  public String toString() {
    return "HlddTerminalNode [targetVariable=" + targetVariable + ", assignedValue=" + assignedValue
        + "]";
  }

}
