/*
 * Copyright 2014-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.cfg.CfgVisitorBackend;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.parser.basis.backend.CfgAssertBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgBlockSeqBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgEmptyCaseBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgEventListToNodeBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgInitialProcessBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgLoopBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgOneBlockProcessBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgRangedSwitchBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgStatisticBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgSwitchSequenceBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgVarRangeBlockBackend;
import ru.ispras.retrascope.parser.basis.backend.CfgWaitBackend;
import ru.ispras.retrascope.util.Log;

import java.io.File;
import java.util.Map;

/**
 * HDL parser basic class.
 *
 * <p>Returns {@link ru.ispras.retrascope.model.cfg.CfgModel}
 * control flow graph model representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class HdlParser extends Engine {

  /**
   * The HDL top-level module names.
   */
  public static final Parameter MODULE_NAME =
      new Parameter("module-name", true, "Top-level module names");

  /**
   * Parameter that assigns top level module name by the name of the first input HDL file.
   */
  public static final Parameter FIRST_FILE_MODULE =
      new Parameter("first-file-module", false, "Get top level module name from first file");

  /**
   * Constructs parser with the specified identifier.
   *
   * @param parserId Parser identifier.
   */
  protected HdlParser(final String parserId) {
    super(parserId);

    addParameter(MODULE_NAME);
    addParameter(NO_BACKEND);
    addParameter(DISABLE_BACKENDS);
    addParameter(FIRST_FILE_MODULE);

    /* Parser back-ends registration. */
    addNoRptBackend(CfgLoopBackend.NAME, new CfgLoopBackend());
    addNoRptBackend(CfgEmptyCaseBackend.NAME, new CfgEmptyCaseBackend());
    addNoRptBackend(CfgEventListToNodeBackend.NAME, new CfgEventListToNodeBackend());
    addNoRptBackend(CfgAssertBackend.NAME, new CfgAssertBackend());
    addNoRptBackend(CfgWaitBackend.NAME, new CfgWaitBackend());
    addNoRptBackend(CfgRangedSwitchBackend.NAME, new CfgRangedSwitchBackend());
    addNoRptBackend(CfgSwitchSequenceBackend.NAME, new CfgSwitchSequenceBackend());
    addNoRptBackend(CfgVarRangeBlockBackend.NAME, new CfgVarRangeBlockBackend());
    addNoRptBackend(CfgBlockSeqBackend.NAME, new CfgBlockSeqBackend());
    addNoRptBackend(CfgOneBlockProcessBackend.NAME, new CfgOneBlockProcessBackend());
    addNoRptBackend(CfgInitialProcessBackend.NAME, new CfgInitialProcessBackend());
    addBackend(new CfgStatisticBackend());
  }

  private void addNoRptBackend(final String name, final CfgVisitor visitor) {

    addVisitorBackend(name, visitor, WalkerMode.DFS_NO_RPT);
  }

  private void addVisitorBackend(
      final String name,
      final CfgVisitor visitor,
      final WalkerMode mode) {

    addBackend(new CfgVisitorBackend(name, visitor, mode));
  }

  /**
   * Parses input entities.
   *
   * @param inputs Input entities collection.
   *
   * @return Output entity.
   */
  protected abstract Entity parse(final Map<EntityType, Entity> inputs);

  /**
   * Returns the type of HDL source inputs.
   * @return The type of HDL source inputs.
   */
  protected abstract EntityType getSrcType();

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    return parse(inputs);
  }

  protected String[] getModuleNames(final Arguments arguments, final FileNames fileNames) {
    if (arguments.contains(MODULE_NAME)) {

      final String names = arguments.getValue(MODULE_NAME);
      return names.split(File.pathSeparator);

    } else if (arguments.contains(FIRST_FILE_MODULE)) {

      final String onlyFile = fileNames.getList().get(0);
      final int left = onlyFile.lastIndexOf(File.separator) + 1;
      final int right = onlyFile.lastIndexOf(getSrcType().getId()) - 1;

      return new String[] {onlyFile.substring(left, right)};
    } else {
      Log.error(
          "No HDL module names are specified: use either '%s' or '%s' option.",
          MODULE_NAME.getName(),
          FIRST_FILE_MODULE.getName());
      return new String[]{};
    }
  }
}
