/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.LinkedList;
import java.util.List;

/**
 * Backend for the HDL parser that transforms processes' event lists {@link EventList}
 * into top level {@link CfgSwitchStatement} switch statements.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgEventListToNodeBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "events_to_node";

  @Override
  public void onModuleEnd(final Module module) {

    for (final ModuleProcess process : module.getProcesses()) {

      final EventList eventList = process.getEventList();

      final Node eventNode = toNode(eventList);

      if (eventNode == null) {
        continue;
      }

      final List<Statement> statements = new LinkedList<>();
      for (int i = 0; i < process.getStatementNum(); i++) {
        statements.add(process.getStatement(i));
      }

      process.removeAll();
      final CfgSwitchStatement eventSwitch = new CfgSwitchStatement(eventNode);
      final CfgCaseStatement trueCase = new CfgCaseStatement((NodeValue) Nodes.TRUE);
      final CfgCaseStatement falseCase = new CfgCaseStatement((NodeValue) Nodes.FALSE);

      trueCase.addAll(statements);
      eventSwitch.add(trueCase);
      eventSwitch.add(falseCase);
      process.add(eventSwitch);
    }
  }

  private Node toNode(final EventList eventList) {

    Node result = null;

    for (final Event event : eventList) {

      final RangedVariable variable = event.getRangedVariable();
      final Node varNode = variable.toNode();

      switch (event.getType()) {

        case ANY_EDGE:
          break;

        case NEGATIVE_EDGE:
          final Node falseNode = Nodes.eq(varNode, NodeUtils.getValue(0, varNode.getDataType()));
          result = (result == null) ? falseNode : Nodes.or(result, falseNode);
          break;

        case POSITIVE_EDGE:
          final Node trueNode = Nodes.eq(varNode, NodeUtils.getValue(1, varNode.getDataType()));
          result = (result == null) ? trueNode : Nodes.or(result, trueNode);
          break;

        default:
          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, event.getType());
      }
    }

    return result;
  }
}
