/*
 * Copyright 2017-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.castle.ast.IntegerRange;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Configuration;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;

/**
 * Parser backend that transforms sequences of switch statements with ranged variables in
 * conditions.
 *
 * <p>The backend optimizes sequences of control flow graph
 * {@link CfgSwitchStatement} switch statements. Statements to be
 * optimized should have conditions of the {@code x[i:j]} form, where {@code x} is a common
 * variable for sequence statements; {@code i} and {@code j} indices are constants; ranges
 * of subsequent statements are non-intersecting but can be joined.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgRangedSwitchBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "ranged_switch";

  /* The next-after-current statement that can be merged. */
  private CfgSwitchStatement next;

  /* The condition for current statement. */
  private NodeOperation cond;

  /* The condition for next-after-current statement. */
  private NodeOperation nextCond;

  /* The range for current statement. */
  private IntegerRange range;

  /* The range for next-after-current statement. */
  private IntegerRange nextRange;

  @Override
  public void onSwitchBegin(final CfgSwitchStatement stmt) {

    while (canCombineWithNext(stmt)) {

      combineWithNext(stmt);
    }
  }

  private boolean canCombineWithNext(final CfgSwitchStatement switchStmt) {

    final Statement parent = switchStmt.getParent();
    final int switchIndex = parent.indexOf(switchStmt);

    if (switchIndex == parent.getStatementNum() - 1) {

      /* Switch statement is the last, no statements to be combined with. */
      return false;
    }

    final Statement next = parent.getStatement(switchIndex + 1);

    if (!next.isType(CfgStatementType.SWITCH)) {

      /* Next statement is not a switch, no statements to be combined with. */
      return false;
    }

    this.next = (CfgSwitchStatement) next;

    final Node currentCond = switchStmt.getCondition();
    final Node nextCond = this.next.getCondition();

    if (!ExprUtils.isKind(Node.Kind.OPERATION, currentCond, nextCond)) {
      return false;
    }
    this.cond = (NodeOperation) currentCond;
    this.nextCond = (NodeOperation) nextCond;

    return sameRanged() && noDefaultCases(switchStmt, this.next);
  }

  private boolean noDefaultCases(final CfgSwitchStatement ... stmts) {

    for (final CfgSwitchStatement switchStmt : stmts) {

      for (int i = 0; i < switchStmt.getStatementNum(); i++) {

        final CfgCaseStatement caseStmt = (CfgCaseStatement) switchStmt.getStatement(i);

        if (caseStmt.isDefault()) {
          return false;
        }
      }
    }

    return true;
  }

  private boolean sameRanged() {

    if (!bvExtract(this.cond, this.nextCond)) {
      return false;
    }

    this.range = getRange(this.cond);
    this.nextRange = getRange(this.nextCond);

    return this.range.border(this.nextRange);
  }

  private IntegerRange getRange(final NodeOperation operation) {

    final NodeValue high = (NodeValue) operation.getOperand(0);
    final NodeValue low = (NodeValue) operation.getOperand(1);

    final NodeValue minIntNode = (NodeValue) TypeConversion.coerce(low, DataType.INTEGER);
    final NodeValue maxIntNode = (NodeValue) TypeConversion.coerce(high, DataType.INTEGER);

    return new IntegerRange(minIntNode.getInteger(), maxIntNode.getInteger());
  }

  private static boolean bvExtract(final Node ... nodes) {

    for (final Node node : nodes) {

      if (!ExprUtils.isOperation(node, StandardOperation.BVEXTRACT)) {
        return false;
      }
    }
    return true;
  }

  private void combineWithNext(final CfgSwitchStatement current) {

    final Statement parent = current.getParent();

    final int compareRanges = this.range.compareTo(this.nextRange);

    final Collection<Statement> cases;

    if (compareRanges > 0) {

      cases = uniteCases(this.next, current, true);

    } else if (compareRanges < 0) {

      cases = uniteCases(current, this.next, false);

    } else {

      final Configuration config = Retrascope.getConfiguration();
      final ExprTreePrinter printer = config.getStylePrinter();

      throw RetrascopeException.exception(
          LogMessage.WRONG_ARG,
          printer.toString(this.cond));
    }

    parent.remove(this.next);

    current.setCondition(
        compareRanges > 0 ? glue(this.nextCond, this.cond) : glue(this.cond, this.nextCond));

    current.removeAll();
    current.addAll(cases);
  }

  private Node glue(final NodeOperation young, final NodeOperation old) {

    final Node expr = young.getOperand(2);
    final Node min = TransformerUtils.toInt(young.getOperand(1));
    final Node max = TransformerUtils.toInt(old.getOperand(0));

    return Nodes.bvextract(max, min, expr);
  }

  private Collection<Statement> uniteCases(
      final CfgSwitchStatement top,
      final CfgSwitchStatement bottom,
      final boolean reverted) {

    final Collection<Statement> union = new LinkedList<>();

    for (int i = 0; i < top.getStatementNum(); i++) {

      final CfgCaseStatement topCase = (CfgCaseStatement) top.getStatement(i);

      for (int j = 0; j < bottom.getStatementNum(); j++) {

        final CfgCaseStatement bottomCase = (CfgCaseStatement) bottom.getStatement(j);

        final EventList unionList = new EventList();
        unionList.addEvents(topCase.getEventList());
        unionList.addEvents(bottomCase.getEventList());

        final Collection<NodeValue> composedValues =
            composeValues(topCase.getValues(), bottomCase.getValues(), reverted);

        final CfgCaseStatement unionCase =
            unionList.isEmpty()
                ? new CfgCaseStatement(composedValues)
                : new CfgCaseStatement(unionList, composedValues);

        final Collection<Statement> topStatements = copyTopStatements(topCase);
        final Collection<Statement> bottomStatements = copyTopStatements(bottomCase);

        if (reverted) {
          unionCase.addAll(bottomStatements);
          unionCase.addAll(topStatements);
        } else {
          unionCase.addAll(topStatements);
          unionCase.addAll(bottomStatements);
        }

        union.add(unionCase);
      }
    }

    return union;
  }

  private Collection<Statement> copyTopStatements(final Statement stmt) {

    final Collection<Statement> topStatements = StatementUtils.getTopStatements(stmt);

    final Collection<Statement> copies = new LinkedHashSet<>();

    for (final Statement topStmt : topStatements) {
      copies.add(topStmt.deepCopy());
    }

    return copies;
  }

  private Collection<NodeValue> composeValues(
      final Collection<NodeValue> oldParts,
      final Collection<NodeValue> youngParts,
      final boolean reverted) {

    final Collection<NodeValue> composed = new LinkedList<>();

    for (final Node oldPart : oldParts) {

      for (final Node youngPart : youngParts) {

        final Node concat =
            reverted ? Nodes.bvconcat(youngPart, oldPart) : Nodes.bvconcat(oldPart, youngPart);

        composed.add((NodeValue) Reducer.reduce(concat));
      }
    }

    return composed;
  }
}
