/*
 * Copyright 2017-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.castle.ast.IntegerRange;
import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.TransformerUtils;

import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Parser backend that transforms assignments with variable ranges into control flow
 * sub-graphs that contain assignments with constant ranges only.
 * <p>This backend appears because of SMT solvers limitations.The current implementation
 * supports only assignments like "x[f(i), g(i)] = y". It means, that range borders (or index)
 * depend only from one variable. Another limitation comes from the fact that the "i"
 * variable should be of integer type.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgVarRangeBlockBackend extends CfgEmptyVisitor {

  /**
   * The backend name.
   */
  public static final String NAME = "var_range_block";

  /* The process under observation. */
  private Module module;

  @Override
  public void onModuleBegin(final Module module) {
    this.module = module;
  }

  @Override
  public void onModuleEnd(final Module module) {
    this.module = null;
  }

  @Override
  public void onStatementEnd(final Statement parent) {

    final int stmtNum = parent.getStatementNum();
    final List<Statement> newStatements = new LinkedList<>();

    boolean shouldBeModified = false;

    for (int i = 0; i < stmtNum; i++) {

      final Statement child = parent.getStatement(i);

      if (!child.isType(CfgStatementType.BASIC_BLOCK)) {
        newStatements.add(child);
        continue;
      }

      final CfgBlockStatement block = (CfgBlockStatement) child;
      final boolean concurrent = block.isConcurrent();

      final List<Assignment> assignments = block.getAssignments();

      for (final Assignment assignment : assignments) {

        final RangedVariable target = assignment.getTarget();

        final Set<NodeVariable> params = getParamVariables(target);

        /* NOTE: if range has 2 or more variables inside it,
         * the backend doesn't try to modify the assignment.
         * So is for non-parametric assignments. */
        if (params.size() != 1) {
          newStatements.add(new CfgBlockStatement(assignment, concurrent));
          continue;
        }

        final NodeVariable param = params.iterator().next();

        final Collection<NodeValue> invValues = getInvariantValues(param);

        final Collection<NodeValue> descriptorValues = getIndices(target, param);

        final int invariantValNum = invValues.size();
        final int descriptorValNum = descriptorValues.size();

        final Collection<NodeValue> values =
            invariantValNum == 0
                ? descriptorValues
                : (descriptorValNum == 0
                      ? invValues
                      : (descriptorValNum <= invariantValNum ? descriptorValues : invValues));

        if (values.isEmpty()) {

          /* No info is available to modify an assignment, so skip it. */
          newStatements.add(new CfgBlockStatement(assignment, concurrent));

        } else {

          shouldBeModified = true;

          final CfgSwitchStatement indexSwitch = new CfgSwitchStatement(param);
          indexSwitch.addMetaInfo(MetaInfoType.TOOL_ADDED, "");

          for (final NodeValue indexValue : values) {

            final CfgCaseStatement indexValCase = new CfgCaseStatement(indexValue);
            indexValCase.addMetaInfo(MetaInfoType.TOOL_ADDED, "");

            indexSwitch.add(indexValCase);

            final Node assignedValue = assignment.getValue();

            final RangedVariable ranged = substituteWithValue(target, param, indexValue);
            final Assignment constIndexAssign = new Assignment(ranged, assignedValue);
            final CfgBlockStatement caseBlock = new CfgBlockStatement(constIndexAssign, concurrent);
            indexValCase.add(caseBlock);
          }
          newStatements.add(indexSwitch);
        }
      }
    }

    if (shouldBeModified) {
      parent.removeAll();
      parent.addAll(newStatements);
    } else {
      newStatements.clear();
    }
  }

  private RangedVariable substituteWithValue(
      final RangedVariable rangedVar,
      final NodeVariable param,
      final NodeValue value) {

    final RangedVariable result = new RangedVariable(rangedVar.getVariable());

    if (rangedVar.ranged()) {

      final Range range = rangedVar.getRange();
      final Node high = range.getHigh();
      final Node resultHigh = TransformerUtils.substitute(high, param.getName(), value);
      final Node low = range.getLow();
      final Node resultLow = TransformerUtils.substitute(low, param.getName(), value);

      result.setRange(resultLow, resultHigh);
    }

    if (rangedVar.indexed()) {

      final List<Node> indices = rangedVar.getIndices();
      final List<Node> resultIndices = new LinkedList<>();

      for (final Node index : indices) {

        resultIndices.add(TransformerUtils.substitute(index, param.getName(), value));
      }

      result.setIndices(resultIndices);
    }

    return result;
  }

  private Set<NodeVariable> getParamVariables(final RangedVariable variable) {
    final Set<NodeVariable> params = new LinkedHashSet<>();

    if (variable.ranged()) {

      final Range range = variable.getRange();
      params.addAll(range.getVariables());
    }

    if (variable.indexed()) {

      for (final Node index : variable.getIndices()) {
        params.addAll(ExprUtils.getVariables(index));
      }
    }
    return params;
  }

  private Collection<NodeValue> getIndices(final RangedVariable ranged, final NodeVariable param) {

    final Collection<NodeValue> indices = new LinkedHashSet<>();

    final NodeVariable var = ranged.getVariable();
    final VariableDescriptor desc = (VariableDescriptor) var.getUserData();

    /*final Node length = desc.getLength();

    if (length != null) {
      final NodeValue lValue = (NodeValue) TypeConversion.coerce(length, DataType.INTEGER);
      final int lInt = lValue.getInteger().intValue();

      final DataType type = var.getDataType();
      final DataTypeId typeId = type.getTypeId();

      for (int i = 0; i < lInt; i++) {

        switch (typeId) {
          case LOGIC_INTEGER:
            indices.add(NodeValue.newInteger(i));
            break;
          case BIT_VECTOR:
            indices.add(NodeValue.newBitVector(i, type.getSize()));
            break;
          default:
            throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, type);
        }
      }
    }*/

    final List<Range> indexRanges = desc.getIndexRanges();
    final List<Node> rangedIndices = ranged.getIndices();

    if (indexRanges != null && rangedIndices != null
        && ExprUtils.getVariables(rangedIndices).contains(param)) {

      int index = -1;

      for (final Node rangedIndex : rangedIndices) {

        if (ExprUtils.getVariables(rangedIndex).contains(param)) {
          index = rangedIndices.indexOf(rangedIndex);
          break;
        }
      }

      final Range paramUsedRange = indexRanges.get(index);

      return getRangeValues(param, paramUsedRange);
    }

    return indices;
  }

  private Collection<NodeValue> getInvariantValues(final NodeVariable variable) {

    final String varName = variable.getName();
    final DecStatement container = StatementUtils.getVariableContainer(this.module, varName);

    if (container == null) {

      Log.warn("No declaration has been found for variable: '%s'", varName);
      return new LinkedHashSet<>();
    }

    final Declaration declaration = container.getDeclaration(variable);

    if (declaration == null) {

      Log.warn("Null declaration has been found for variable: '%s'", varName);
      return new LinkedHashSet<>();
    }

    final Range invariant = declaration.getInvariant();

    /* NOTE: the invariant must be of integer type to be elaborated. */
    return getRangeValues(variable, invariant);
  }

  private Collection<NodeValue> getRangeValues(NodeVariable variable, Range invariant) {

    final Collection<NodeValue> values = new LinkedHashSet<>();

    final DataType type = variable.getDataType();
    final int size = type.getSize();

    if (invariant == null || !invariant.isConstantRange() || !invariant.isIntegerRange()) {

      /* BIT_VECTOR case is special: all the possible values will be calculated. *//*
      if (ExprUtils.isType(DataTypeId.BIT_VECTOR, variable)) {

        for (int i = 0; i < Math.pow(2, size); i++) {
          values.add(NodeValue.newBitVector(i, size));
        }
      }*/
      if (invariant != null) {
        Log.warn(
            "Unable to get possible values for: var='%s', invariant='%s'", variable, invariant);
      }
      return values;
    }

    final IntegerRange intRange = invariant.toIntegerRange();
    final BigInteger rangeSize = intRange.size();

    for (int i = 0; i < rangeSize.intValue(); i++) {

      final BigInteger value = intRange.getMin().add(BigInteger.valueOf(i));

      final DataTypeId typeId = type.getTypeId();

      switch (typeId) {
        case LOGIC_INTEGER:
          values.add(NodeValue.newInteger(value));
          break;
        case BIT_VECTOR:
          values.add(NodeValue.newBitVector(value, size));
          break;
        default:
          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, type);
      }
    }

    return values;
  }
}
