/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.verilog.parser.VerilogDesignBackend;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.elaborator.VerilogDesign;
import ru.ispras.verilog.parser.elaborator.VerilogProcess;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.walker.VerilogTreeWalker;

import java.util.Map;

/**
 * This class creates {@link CfgModel} from Verilog AST provided by Verilog Translator.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VerilogCfgBuilder extends VerilogDesignBackend {

  private static final String BACKEND_NAME = "verilog-cfg-backend";

  private CfgModel model;

  VerilogCfgBuilder() {
    super(BACKEND_NAME);
    this.model = new CfgModel();
  }

  /**
   * Returns the constructed CFG model.
   *
   * @return The constructed CFG model.
   */
  public CfgModel getModel() {
    return this.model;
  }

  @Override
  public void start(VerilogDesign verilogDesign) {

    this.model = new CfgModel();

    final Map<String, VerilogDescriptor> variables = verilogDesign.getVariables();
    final String designName = verilogDesign.getName();
    final String moduleName = designName == null ? "" : designName;

    final CfgModule module = new CfgModule(moduleName);

    declareVariables(variables, module);

    for (final VerilogProcess process : verilogDesign.getProcesses()) {
      final VerilogStatement statement = process.getStatement();

      final VerilogProcess.Type type = process.getType();
      final VerilogCfgProcessBuilder processBuilder =
          new VerilogCfgProcessBuilder(module, type, variables);
      final VerilogTreeWalker walker = new VerilogTreeWalker(statement, processBuilder);

      walker.start();
      module.add(processBuilder.getProcess());
    }
    this.model.addModule(module);
  }

  private void declareVariables(final Map<String, VerilogDescriptor> variables,
      final CfgModule module) {

    for (final VerilogDescriptor verilogDesc : variables.values()) {

      final NodeVariable variable = verilogDesc.getVariable();
      final VerilogDeclaration declaration = verilogDesc.getDeclaration();

      final NodeVariable varCopy = (NodeVariable) variable.deepCopy();
      final VariableDescriptor varDescriptor = VerilogUtils.getDescriptor(verilogDesc);
      varCopy.setUserData(varDescriptor);

      final VerilogMinTypMax decExpression = declaration.getExpression();

      if (decExpression == null) {
        module.declareVariable(varCopy);
      } else {
        if (!decExpression.isUsual()) {
          throw RetrascopeException.exception(LogMessage.CANT_PARSE, decExpression.toString());
        }

        final VerilogExpression verilogInitValue = decExpression.getMinExpression();
        final DataType varDataType = variable.getDataType();
        Node initNode = verilogInitValue.getNode();
        final DataType initNodeType = initNode.getDataType();
        if (!(varDataType.equals(initNodeType)
            || variable.isType(DataType.UNKNOWN)
            || initNode.isType(DataType.UNKNOWN))) {
          initNode = TypeConversion.coerce(initNode, varDataType);
        }
        module.declareVariable(varCopy, initNode);
      }
    }
  }
}
