/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License") {} you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.MetaInfoValue;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.elaborator.VerilogProcess;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogAttribute;
import ru.ispras.verilog.parser.model.VerilogBlockGenerate;
import ru.ispras.verilog.parser.model.VerilogBlockStatement;
import ru.ispras.verilog.parser.model.VerilogCaseGenerate;
import ru.ispras.verilog.parser.model.VerilogCaseGenerateItem;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.VerilogCode;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDefineParameter;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogDisableStatement;
import ru.ispras.verilog.parser.model.VerilogGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerateBranch;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogIfStatementBranch;
import ru.ispras.verilog.parser.model.VerilogInstantiation;
import ru.ispras.verilog.parser.model.VerilogLoopGenerate;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNullStatement;
import ru.ispras.verilog.parser.model.VerilogPathDeclaration;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogPortConnection;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.VerilogPulseStyle;
import ru.ispras.verilog.parser.model.VerilogShowCancelled;
import ru.ispras.verilog.parser.model.VerilogSpecify;
import ru.ispras.verilog.parser.model.VerilogTable;
import ru.ispras.verilog.parser.model.VerilogTableEntry;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogTriggerStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;
import ru.ispras.verilog.parser.model.basis.VerilogEdge;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.model.util.ModelUtils;
import ru.ispras.verilog.parser.walker.VerilogNodeVisitor;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * This class creates {@link CfgProcess} from Verilog AST provided by Verilog Translator.
 *
 * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VerilogCfgProcessBuilder extends VerilogNodeVisitor {

  /*
   * Current CFG process to add statements.
   */
  private CfgProcess process;

  /*
   * CFG process for initializations.
   */
  private CfgProcess initialProcess;

  /**
   * Collection of temporary identifiers that should be removed at the end of 'case' elaboration.
   */
  private Map<VerilogCaseStatement, Set<String>> caseIds;

  /*
   * List of currently base block identifiers.
   */
  private Stack<String> currentbases;

  /*
   * List of data type in condition expressions.
   */
  private Stack<DataType> conditions;

  /*
   * Storage of CFG modules.
   */
  private List<CfgModule> modules;

  /*
   * Shows whether it is always process or initial.
   */
  private VerilogProcess.Type activityType;

  /*
   * Mapping from variables names to their descriptors.
   */
  private final Map<String, VerilogDescriptor> variables;

  VerilogCfgProcessBuilder(
      final CfgModule module,
      final VerilogProcess.Type processType,
      final Map<String, VerilogDescriptor> variables) {

    this.variables = variables;
    this.modules = new LinkedList<>();
    this.modules.add(module);

    this.activityType = processType;
    if (activityType == VerilogProcess.Type.INITIAL) { // or INITIAL or ALWAYS
      if (initialProcess == null) {
        initialProcess = new CfgProcess(true);
      }
    } else {
      process = new CfgProcess();
    }

    if (caseIds == null) {
      caseIds = new LinkedHashMap<>();
    }
    if (conditions == null) {
      conditions = new Stack<>();
    }
    if (currentbases == null) {
      currentbases = new Stack<>();
    }
  }

  /*
   * {@link VariableTransformer} changes variables' user data format.
   *
   * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
   */
  private final class VariableTransformer extends ExprTreeVisitorDefault {

    private final Map<String, VerilogDescriptor> variables;

    VariableTransformer(Map<String, VerilogDescriptor> variables) {

      this.variables = variables;
    }

    Map<String, VerilogDescriptor> getVariables() {

      return this.variables;
    }

    @Override
    public void onVariable(final NodeVariable variable) {
      InvariantChecks.checkNotNull(variable);

      final String name = variable.getName();
      final VerilogDescriptor descriptor = getVariables().get(name);
      InvariantChecks.checkNotNull(descriptor, "No descriptor for variable: " + name);

      final VariableDescriptor variableDescriptor = VerilogUtils.getDescriptor(descriptor);

      variable.setUserData(variableDescriptor);
    }
  }

  /**
   * Returns the constructed CFG process.
   *
   * @return The constructed CFG process.
   */
  public CfgProcess getProcess() {
    return this.activityType == VerilogProcess.Type.INITIAL ? this.initialProcess : this.process;
  }

  /**
   * Returns the CFG model reference kept in the storage according to a number of current CFG model.
   *
   * @return Current CFG model.
   */
  private CfgModule getCurrentModule() {
    return modules.get(modules.size() - 1);
  }

  /**
   * Changes userData type kept in the node from Reference to VariableType.
   *
   * @param node Node to be updated.
   * @return Updated node.
   */
  private Node transformNode(final Node node) {

    final Node castConstNode = transformDataOf(node);

    return TransformerUtils.standardize(castConstNode);
  }

  private Node transformDataOf(final Node node) {
    final VariableTransformer transformer = new VariableTransformer(this.variables);
    final ExprTreeWalker walker = new ExprTreeWalker(transformer);
    walker.visitNode(node);
    Node castConstNode = node;

    try {

      castConstNode = TypeConversion.castConstants(node);

    } catch (final Exception e) {

      Log.warn(e.getMessage());

    } finally {

      if (castConstNode == null) {
        castConstNode = node;
      }
    }
    castConstNode.setUserData(node.getUserData());
    return castConstNode;
  }

  private List<Assignment> parseAssignments(final Node value, final Node target) {

    RangedVariable singleVar;

    if (target instanceof NodeVariable) {

      final NodeVariable variable = (NodeVariable) transformDataOf(target);
      singleVar = new RangedVariable(variable);
      return Collections.singletonList(getSingleAssignment(value, singleVar));

    } else if (target instanceof NodeOperation) {

      // Transform target because of BVEXTRACT(... BVEXTRACT (i j x))
      return parseOperationAssignments(value, (NodeOperation) target);
    } else {
      throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, target.getClass());
    }
  }

  private List<Assignment> parseOperationAssignments(final Node value, final NodeOperation target) {

    Node node = transformDataOf(target);

    if (ExprUtils.isOperation(node, StandardOperation.BVCONCAT)) {

      return parseConcatAssignments((NodeOperation) node, value);
    }

    return Collections.singletonList(parseSingleAssignment(value, node));
  }

  private Assignment parseSingleAssignment(final Node value, Node node) {

    final List<Node> indices = new LinkedList<>();
    Range range = null;

    while (ExprUtils.isKind(Node.Kind.OPERATION, node)) {

      final NodeOperation nodeOp = (NodeOperation) node;

      if (ExprUtils.isOperation(node, StandardOperation.SELECT)) {

        indices.add(nodeOp.getOperand(1));

        node = nodeOp.getOperand(0);
      }

      if (ExprUtils.isOperation(node, StandardOperation.BVEXTRACT)) {

        Node high = nodeOp.getOperand(0);
        Node low = nodeOp.getOperand(1);

        final DataType highType = high.getDataType();
        final DataType lowType = low.getDataType();

        if (!highType.equals(lowType) && NodeUtils.oneConstAnotherNot(high, low)) {

          /* When both const and var-containing params appear, cast first to last. */
          high = high.getKind() == Node.Kind.VALUE ? TypeConversion.coerce(high, lowType) : high;
          low = low.getKind() == Node.Kind.VALUE ? TypeConversion.coerce(low, highType) : low;
        }

        range = new Range(low, high);

        node = nodeOp.getOperand(2);
      }
    }

    InvariantChecks.checkTrue(node instanceof NodeVariable, "Unexpected target: " + node);

    final RangedVariable targetVar =
        indices.isEmpty()
            ? range == null
                ? new RangedVariable((NodeVariable) node)
                : new RangedVariable((NodeVariable) node, range)
            : range == null
                ? new RangedVariable((NodeVariable) node, indices)
                : new RangedVariable((NodeVariable) node, range, indices);

    return getSingleAssignment(value, targetVar);
  }

  private List<Assignment> parseConcatAssignments(final NodeOperation concatOp, final Node value) {

    final List<Assignment> list = new LinkedList<>();

    /*NOTE: it is supposed that BVCONCAT operands are atomic (except for BVEXTRACT, SELECT). */

    int low = 0;

    for (int i = concatOp.getOperandCount() - 1; i >= 0; i--) {

      final Node operand = concatOp.getOperand(i);
      final int high = operand.getDataType().getSize() + low - 1;
      final Node assigned = Nodes.bvextract(high, low, value);

      if (operand instanceof NodeVariable) {

        list.add(new Assignment(new RangedVariable((NodeVariable) operand), assigned));

      } else if (operand instanceof NodeOperation) {

        list.add(parseSingleAssignment(assigned, operand));
      } else {
        throw RetrascopeException.exception(LogMessage.WRONG_TYPE, operand.getClass());
      }

      low += high + 1;
    }

    return list;
  }

  private Assignment getSingleAssignment(final Node expression, final RangedVariable singleVar) {

    final Node node = getRhsValue(expression, RangedVariableUtils.getDataType(singleVar));

    return new Assignment(singleVar, node);
  }

  private Node getRhsValue(final Node node, final DataType varType) {
    Node expr = TypeConversion.castConstants(node);

    /* Check that assigned expression has the same type as the defined variable. */
    final DataType localType = expr.getDataType();

    Node coerced = null;
    if (!(varType.equals(localType)
        || varType.equals(DataType.UNKNOWN)
        || localType.equals(DataType.UNKNOWN))) {
      coerced = TypeConversion.coerce(expr, varType);
    }
    return coerced == null ? expr : coerced;
  }

  /**
   * Processes assignments of different types.
   *
   * @param assignment The assignment expression itself.
   * @param type Blocking, non-blocking, assign (deassign, force, release will not be supported).
   * @param insideProcess Add the assignment to the current process or create a new one.
   */
  private void parseAssignment(
      final VerilogAssignment assignment,
      final VerilogAssignStatement.Type type,
      final boolean insideProcess) {

    final EventList eventList = new EventList();
    if (!insideProcess) {
      final Set<NodeVariable> useVars = ModelUtils.getUsedVariables(assignment);

      for (final NodeVariable node : useVars) {

        final RangedVariable rangedVar = new RangedVariable((NodeVariable)transformNode(node));
        final Event retrascopeEvent = new Event(rangedVar);

        eventList.add(retrascopeEvent);
      }
    }

    final VerilogMinTypMax rhsExpr = assignment.getRhsExpression();
    final Node rhsMinNode = rhsExpr.getMinExpression().getNode();
    final Node expression = rhsExpr.isUsual() ? transformNode(rhsMinNode) : null;
    // NOTICE: for unusual expressions which are not supported
    // System.out.println(assignment.getCondition().getMinExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getTypExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getMaxExpression());

    if (expression == null) {
      throw new IllegalStateException();
    }

    final CfgBlockStatement block =
        new CfgBlockStatement(type == VerilogAssignStatement.Type.NON_BLOCKING);

    final List<Assignment> assignments =
        parseAssignments(expression, transformDataOf(assignment.getLhsNode()));
    block.addAssignments(assignments);

    addBlock(block, eventList, insideProcess);
  }

  private Collection<Assignment> collectAssignments(final VerilogAssignment assignment) {
    InvariantChecks.checkTrue(assignment.getRhsExpression().isUsual());

    final Node rhs = transformNode(assignment.getRhsExpression().getMinExpression().getNode());

    final Collection<Assignment> newAssignments = new LinkedList<>();

    newAssignments.addAll(parseAssignments(rhs, transformDataOf(assignment.getLhsNode())));

    return newAssignments;
  }

  /**
   * Simplified adding of process nodes into Cfg.
   *
   * <p>Required are only two parameters: the CFGNode itself and flag forcing
   * to add the Node into initial or normal process.
   * The other parameters are set to the following default values:
   * empty eventList and adding into existing process (if it exists).</p>
   *
   * @param block Node to be added.
   */
  private void addBlock(final Statement block) {
    final EventList eventList = new EventList();
    addBlock(block, eventList, true);
  }

  /**
   * Add process node (some block, e.g. assignment) to CFG.
   *
   * @param block Node to be added.
   * @param eventList Should contain all the events touching the process.
   *                  Used in case of continuous assignments.
   * @param insideProcess If {@code true}, the block is added to existing process,
   *                      If {@code false}, new process is created for it.
   */
  private void addBlock(final Statement block, final EventList eventList,
      final boolean insideProcess) {
    if (insideProcess) {
      attachBlock(block);
    } else {
      final CfgProcess process = new CfgProcess(eventList);
      process.add(block);
      getCurrentModule().add(process);
    }
  }

  /**
   * Attach block into existing process.
   * @param block Node to be added.
   */
  private void attachBlock(final Statement block) {
    final CfgProcess localProcess =
        activityType == VerilogProcess.Type.INITIAL ? initialProcess : process;

    if (currentbases.isEmpty()) {
      localProcess.add(block);
    } else {
      final Statement statement = CfgUtils.getStatement(localProcess, currentbases.peek());
      statement.add(block);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Activity
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onActivityBegin(final VerilogActivity node) {
    activityType = node.isInitial() ? VerilogProcess.Type.INITIAL : VerilogProcess.Type.ALWAYS;
    if (activityType == VerilogProcess.Type.INITIAL) { // or INITIAL or ALWAYS
      if (initialProcess == null) {
        initialProcess = new CfgProcess(true);
      }
    } else {
      process = new CfgProcess();
    }

    if (caseIds == null) {
      caseIds = new LinkedHashMap<>();
    }
    if (conditions == null) {
      conditions = new Stack<>();
    }
    if (currentbases == null) {
      currentbases = new Stack<>();
    }

    return Result.OK;
  }

  @Override
  public Result onActivityEnd(final VerilogActivity node) {
    if (activityType == VerilogProcess.Type.INITIAL) {
      getCurrentModule().add(initialProcess);
      activityType = VerilogProcess.Type.INITIAL;
      initialProcess = null;
    } else {
      getCurrentModule().add(process);
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assign
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignBegin(final VerilogAssign node) {
    // node.getDelay() is not used
    parseAssignment(node.getAssignment(), VerilogAssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignEnd(final VerilogAssign node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // AssignStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignStatementBegin(final VerilogAssignStatement node) {
    switch (node.getType()) {
      case BLOCKING:
      case NON_BLOCKING:
      case ASSIGN: {
        parseAssignment(node.getAssignment(), node.getType(), true);
        break;
      }
      default: {
        Log.warn(node.getType() + " is unsupported.");
        break;
      }
    }

    return Result.OK;
  }

  @Override
  public Result onAssignStatementEnd(final VerilogAssignStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assignment
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignmentBegin(final VerilogAssignment node) {
    parseAssignment(node, VerilogAssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignmentEnd(final VerilogAssignment node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Attribute
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAttributeBegin(final VerilogAttribute node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onAttributeEnd(final VerilogAttribute node) {
    // Nothing
    return Result.OK;
  }

  // -------------------------------------------------------------------------
  // BlockGenerate
  // -------------------------------------------------------------------------

  @Override
  public Result onBlockGenerateBegin(final VerilogBlockGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onBlockGenerateEnd(final VerilogBlockGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgBlockStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onBlockStatementBegin(final VerilogBlockStatement node) {

    if (node.getType() == VerilogBlockStatement.Type.FORK) {
      Log.warn(LogMessage.UNSUPPORTED_TYPE + node.getType().name());
    }

    return Result.OK;
  }

  @Override
  public Result onBlockStatementEnd(final VerilogBlockStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateBegin(final VerilogCaseGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateEnd(final VerilogCaseGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerateItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateItemBegin(final VerilogCaseGenerateItem node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateItemEnd(final VerilogCaseGenerateItem node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgCaseStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementBegin(final VerilogCaseStatement node) {

    caseIds.put(node, new LinkedHashSet<>());
    return Result.OK;
  }

  @Override
  public Result onCaseStatementEnd(final VerilogCaseStatement node) {

    String base = currentbases.peek();
    final Set<String> nodeCaseIds = caseIds.get(node);
    while (!nodeCaseIds.isEmpty() && nodeCaseIds.contains(base)) {
      nodeCaseIds.remove(base);
      currentbases.pop();
      base = currentbases.isEmpty() ? null : currentbases.peek();
    }
    caseIds.remove(node);
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseStatementItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementItemBegin(final VerilogCaseStatementItem node) {

    final Node exprNode = node.getExpression().getNode();
    final CfgSwitchStatement switchBlock = new CfgSwitchStatement(transformNode(exprNode));
    final String switchId = switchBlock.getId();
    attachBlock(switchBlock);
    currentbases.push(switchId);
    caseIds.get(node.getParentNode()).add(switchId);

    final CfgCaseStatement trueCase = new CfgCaseStatement((NodeValue) Nodes.TRUE);
    final String trueCaseId = trueCase.getId();
    attachBlock(trueCase);
    currentbases.push(trueCaseId);
    caseIds.get(node.getParentNode()).add(trueCaseId);

    return Result.OK;
  }

  @Override
  public Result onCaseStatementItemEnd(final VerilogCaseStatementItem node) {
    final String removed = currentbases.pop();
    caseIds.get(node.getParentNode()).remove(removed);

    final CfgCaseStatement caseBlock = new CfgCaseStatement((NodeValue) Nodes.FALSE);
    final String caseId = caseBlock.getId();

    attachBlock(caseBlock);
    currentbases.push(caseId);
    caseIds.get(node.getParentNode()).add(caseId);

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Code
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCodeBegin(final VerilogCode node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onCodeEnd(final VerilogCode node) {
    // Creates new empty CFG model and adds it into storage of CFG models.
    if (modules == null || modules.size() == 0) {
      modules = new LinkedList<>();
      final String fileName = node.getFileName();
      final CfgModule module = new CfgModule(fileName == null ? "" : fileName);
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);
      modules.add(module);
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Declaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDeclarationBegin(final VerilogDeclaration node) {

    if (node.getExpression() != null) {
      if (initialProcess == null) {
        initialProcess = new CfgProcess(true);
      }

      final VerilogAssignment assignment = new VerilogAssignment(null);
      final VerilogReference reference = new VerilogReference();
      final VerilogPath path = new VerilogPath();
      path.addItem(new VerilogPathItem(node.getName()));
      reference.setDeclaration(node);
      reference.setPath(path);
      assignment.addReference(reference);
      assignment.setRhsExpression(node.getExpression());
      activityType = VerilogProcess.Type.INITIAL;
      parseAssignment(assignment, VerilogAssignStatement.Type.ASSIGN, true);
    }
    return Result.OK;
  }

  @Override
  public Result onDeclarationEnd(final VerilogDeclaration node) {
    activityType = VerilogProcess.Type.ALWAYS;
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DefineParameter
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDefineParameterBegin(final VerilogDefineParameter node) {

    return Result.OK;
  }

  @Override
  public Result onDefineParameterEnd(final VerilogDefineParameter node) {

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DelayedStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDelayedStatementBegin(final VerilogDelayedStatement node) {
    final VerilogEventControl eventControl = node.getEventControl();

    if (eventControl.isEvents()) { //|| eventControl.isRepeat() eventControl.isDelays() not parsed
      for (final VerilogEvent event : eventControl.getEvents()) {
        EventType eventType = EventType.ANY_EDGE;
        if (event.getEdge().equals(VerilogEdge.POSITIVE)) {
          eventType = EventType.POSITIVE_EDGE;
        } else if (event.getEdge() == VerilogEdge.NEGATIVE) {
          eventType = EventType.NEGATIVE_EDGE;
        } else {
          assert event.getEdge().equals(VerilogEdge.UNDEFINED);
        }

        final EventList eventList = new EventList();
        final VerilogExpression eventExpr = event.getExpression();
        for (final NodeVariable nodeVariable : ExprUtils.getVariables(eventExpr.getNode())) {
          final RangedVariable rangedVariable =
              new RangedVariable((NodeVariable)transformNode(nodeVariable));

          final Event retrascopeEvent = new Event(rangedVariable, eventType);

          if (!getCurrentModule().declares(nodeVariable)) {
            Log.warn(
                "Variable %s is not a variable in module %s!",
                nodeVariable.getName(),
                getCurrentModule().getName());
          }

          final CfgProcess localProcess =
              activityType == VerilogProcess.Type.INITIAL ? initialProcess : process;
          if (activityType == VerilogProcess.Type.INITIAL) {
            eventList.add(retrascopeEvent);
          } else {
            localProcess.addEvent(retrascopeEvent);
          }
        }

        if (activityType == VerilogProcess.Type.INITIAL) {
          final CfgWaitStatement waitBlock =
              new CfgWaitStatement(eventList
              /*, transformNode(event.getExpression().getNode())*/);
          addBlock(waitBlock);
        }
      }
    }

    return Result.OK;
  }

  @Override
  public Result onDelayedStatementEnd(final VerilogDelayedStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DisableStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDisableStatementBegin(final VerilogDisableStatement node) {
    return Result.OK;
  }

  @Override
  public Result onDisableStatementEnd(final VerilogDisableStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Generate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onGenerateBegin(final VerilogGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onGenerateEnd(final VerilogGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBegin(final VerilogIfGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateEnd(final VerilogIfGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerateBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBranchBegin(final VerilogIfGenerateBranch node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateBranchEnd(final VerilogIfGenerateBranch node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBegin(final VerilogIfStatement node) {

    final VerilogExpression expression = node.getExpression();
    final Node exprNode = expression.getNode();

    if (exprNode.isType(DataType.UNKNOWN)) {
      Log.error("[onIfStatementBegin] DataType of %s is UNKNOWN.", node.getExpression().getNode());
    }

    final CfgSwitchStatement switchBlock = new CfgSwitchStatement(transformDataOf(exprNode));
    attachBlock(switchBlock);
    currentbases.push(switchBlock.getId());
    conditions.push(exprNode.getDataType());
    return Result.OK;
  }

  @Override
  public Result onIfStatementEnd(final VerilogIfStatement node) {
    currentbases.pop();
    conditions.pop();

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatementBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBranchBegin(final VerilogIfStatementBranch node) {

    CfgCaseStatement caseBlock;

    if (node.isThen()) {
      caseBlock = new CfgCaseStatement(getCaseValue(1, true));
    } else if (node.isElse()) {
      caseBlock = new CfgCaseStatement(getCaseValue(0, false));
    } else {
      throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, node);
    }

    attachBlock(caseBlock);
    currentbases.push(caseBlock.getId());
    return Result.OK;
  }

  private NodeValue getCaseValue(
      final int bvValue,
      final boolean boolValue) {
    return conditions.peek().getTypeId() == DataTypeId.BIT_VECTOR
        ? NodeValue.newBitVector(bvValue, conditions.peek().getSize())
        : NodeValue.newBoolean(boolValue);
  }

  @Override
  public Result onIfStatementBranchEnd(final VerilogIfStatementBranch node) {
    currentbases.pop();
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Instantiation
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onInstantiationBegin(final VerilogInstantiation node) {

    return Result.OK;
  }

  @Override
  public Result onInstantiationEnd(final VerilogInstantiation node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // LoopGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopGenerateBegin(final VerilogLoopGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onLoopGenerateEnd(final VerilogLoopGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgLoopStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopStatementBegin(final VerilogLoopStatement node) {
    CfgLoopStatement loopStatement = null;

    final VerilogExpression verExpr = node.getExpression();

    switch (node.getType()) {
      case FOREVER: {
        loopStatement = new CfgLoopStatement(CfgLoopStatement.Type.FOREVER, null, null, null);
        break;
      }
      case REPEAT: {

        final Node exprNode = verExpr.getNode();
        final DataType dataType = exprNode.getDataType();

        NodeValue initValue;
        NodeValue incValue;
        Enum<?> incOpId;
        Enum<?> cmpOpId;

        final NodeVariable nodeVariable = getCurrentModule().declareNewVariable(dataType);

        if (dataType.getTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
          initValue = NodeValue.newInteger(0);
          incValue = NodeValue.newInteger(1);
          incOpId = StandardOperation.ADD;
          cmpOpId = StandardOperation.LESS;
        } else if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          initValue = NodeValue.newBitVector(0, dataType.getSize());
          incValue = NodeValue.newBitVector(1, dataType.getSize());
          incOpId = StandardOperation.BVADD;
          cmpOpId = StandardOperation.BVULT;
        } else {
          throw new IllegalArgumentException(
              LogMessage.makeMessage(LogMessage.CANT_PARSE, verExpr));
        }

        final NodeOperation incExpr = new NodeOperation(incOpId, nodeVariable, incValue);
        final RangedVariable incVar = new RangedVariable(nodeVariable);
        final Assignment incAssignment = new Assignment(incVar, incExpr);

        final Assignment initAssignment = new Assignment(nodeVariable, initValue);
        final NodeOperation checkExpr = new NodeOperation(cmpOpId, nodeVariable, exprNode);

        loopStatement =
            new CfgLoopStatement(
                CfgLoopStatement.Type.REPEAT,
                checkExpr,
                initAssignment,
                incAssignment);
        break;
      }
      case WHILE: {
        final Node exprNode = verExpr.getNode();

        if (exprNode.isType(DataType.UNKNOWN)) {
          Log.error("[onLoopStatementBegin] DataType of %s is UNKNOWN.", verExpr.getNode());
        }

        loopStatement = new CfgLoopStatement(CfgLoopStatement.Type.WHILE, exprNode, null, null);
        break;
      }
      case FOR: {

        final Node exprNode = verExpr.getNode();

        if (exprNode.isType(DataType.UNKNOWN)) {
          Log.error("[onLoopStatementItemBegin2] DataType of %s is UNKNOWN.", exprNode);
        }

        final Collection<Assignment> initAssignment = collectAssignments(node.getInitialization());
        InvariantChecks.checkTrue(initAssignment.size() == 1);

        final Collection<Assignment> iterAssignment = collectAssignments(node.getIteration());
        InvariantChecks.checkTrue(iterAssignment.size() == 1);

        loopStatement =
            new CfgLoopStatement(
                CfgLoopStatement.Type.FOR,
                exprNode,
                initAssignment.iterator().next(),
                iterAssignment.iterator().next());
        break;
      }
      default:
        Log.error(LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));
    }
    attachBlock(loopStatement);
    currentbases.push(loopStatement.getId());

    return Result.OK;
  }

  @Override
  public Result onLoopStatementEnd(final VerilogLoopStatement node) {
    currentbases.pop();
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Module
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onModuleBegin(final VerilogModule node) {
    if (node.isMacromodule()) {
      Log.error(LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));

    } else if (node.isGatePrimitive() || node.isUserPrimitive()) {
      Log.error(LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));
    } else {
      final CfgModule module =
          new CfgModule(node.getName());
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);

      Log.debug("Start observing module %s.", module.getName());

      if (modules == null) {
        modules = new LinkedList<>();
      }
      if (caseIds == null) {
        caseIds = new LinkedHashMap<>();
      }
      if (conditions == null) {
        conditions = new Stack<>();
      }
      if (currentbases == null) {
        currentbases = new Stack<>();
      }

      modules.add(module);
    }

    return Result.OK;
  }

  @Override
  public Result onModuleEnd(final VerilogModule node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // NullStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onNullStatementBegin(final VerilogNullStatement node) {
    return Result.OK;
  }

  @Override
  public Result onNullStatementEnd(final VerilogNullStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PathDeclaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPathDeclarationBegin(final VerilogPathDeclaration node) {
    return Result.OK;
  }

  @Override
  public Result onPathDeclarationEnd(final VerilogPathDeclaration node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Port
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortBegin(final VerilogPort node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onPortEnd(final VerilogPort node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PortConnection
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortConnectionBegin(final VerilogPortConnection node) {

    return Result.OK;
  }

  @Override
  public Result onPortConnectionEnd(final VerilogPortConnection node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Procedure
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onProcedureBegin(final VerilogProcedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) { // TODO: implement functions and tasks.
      text("function");
    } else {
      text("task");
    }

    if (node.isAutomatic()) {
      text(" ");
      text("automatic");
    }

    if (type.isSpecified()) {
      text(" ");
      text(type);
    }

    text(" ");
    text(node.getName());
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onProcedureEnd(final VerilogProcedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) {
      text("endfunction");
    } else {
      text("endtask");
    }

    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PulseStyle
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPulseStyleBegin(final VerilogPulseStyle node) {
    return Result.OK;
  }

  @Override
  public Result onPulseStyleEnd(final VerilogPulseStyle node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // ShowCancelled
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onShowCancelledBegin(final VerilogShowCancelled node) {
    return Result.OK;
  }

  @Override
  public Result onShowCancelledEnd(final VerilogShowCancelled node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Specify
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onSpecifyBegin(final VerilogSpecify node) {
    return Result.OK;
  }

  @Override
  public Result onSpecifyEnd(final VerilogSpecify node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Table
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableBegin(final VerilogTable node) {
    /*text("table"); TODO: implement tables.
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEnd(final VerilogTable node) {
    /*text("endtable");
    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TableEntry
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableEntryBegin(final VerilogTableEntry node) {
    /*text(" ", node.getInputs1());
    text(" ");
    text(" ", node.getEdge()); TODO: implement tables.
    text(" ");
    text(" ", node.getInputs2());
    text(" : ");
    text(" ", node.getOutputs());
    text(";");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEntryEnd(final VerilogTableEntry node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TaskStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTaskStatementBegin(final VerilogTaskStatement node) {
    /*text(node.getPath());
    text("(");

    boolean comma = false;
    for (final Expression expression : node.getArguments()) {
      if (comma) {
        text(", ");
      }
      comma = true;

      text(expression); TODO: implement task statements
    }

    text(");");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTaskStatementEnd(final VerilogTaskStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TriggerStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTriggerStatementBegin(final VerilogTriggerStatement node) {
    return Result.OK;
  }

  @Override
  public Result onTriggerStatementEnd(final VerilogTriggerStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgWaitStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onWaitStatementBegin(final VerilogWaitStatement node) {

    final NodeVariable nodeVariable = (NodeVariable)transformNode(node.getExpression().getNode());
    NodeValue nodeValue;
    if (nodeVariable.isType(DataTypeId.LOGIC_INTEGER)) {
      nodeValue = NodeValue.newInteger(0);
    } else {
      if (nodeVariable.isType(DataTypeId.BIT_VECTOR)) {
        nodeValue = NodeValue.newBitVector(0, nodeVariable.getDataType().getSize());
      } else {
        throw new IllegalArgumentException(
            LogMessage.makeMessage(LogMessage.CANT_PARSE, node.getExpression()));
      }
    }

    final NodeOperation nodeExpr = Nodes.noteq(nodeVariable, nodeValue);
    final CfgWaitStatement waitBlock = new CfgWaitStatement(nodeExpr);

    addBlock(waitBlock);

    return Result.OK;
  }

  @Override
  public Result onWaitStatementEnd(final VerilogWaitStatement node) {
    return Result.OK;
  }
}
