/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import ru.ispras.castle.ast.Range;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.elaborator.VerilogIndexRange;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogRange;

import java.util.LinkedList;
import java.util.List;

/**
 * Utility methods for Verilog AST elements processing.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VerilogUtils {

  private VerilogUtils() {
    /* Empty. */
  }

  /**
   * Constructs variable description container from the specified description object.
   * @param descriptor The Verilog AST description object.
   * @return The variable description container.
   */
  static VariableDescriptor getDescriptor(final VerilogDescriptor descriptor) {

    final VerilogDeclaration declaration = descriptor.getDeclaration();
    final VerilogDeclaration.Type verilogType = declaration.getType();

    VariableType type = VariableType.REG;

    switch (verilogType) {

      case INPUT:
        type = VariableType.IN;
        break;
      case OUTPUT:
        type = VariableType.OUT;
        break;
      case INOUT:
        type = VariableType.INOUT;
        break;
      default:
        break;
    }

    final VariableDescriptor varDescriptor = new VariableDescriptor(type);
    final VerilogElementType elemType = declaration.getElementType();
    final VerilogRange range = elemType.getRange();
    final String typeString = range == null ? "" : getTypeString(descriptor);

    varDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, typeString);

    final List<Range> indexRanges = getRanges(descriptor.getIndexRanges());
    varDescriptor.setIndexRanges(indexRanges);
    return varDescriptor;
  }

  private static List<Range> getRanges(final List<VerilogIndexRange> indexRanges) {
    final List<Range> list = new LinkedList<>();

    for (final VerilogIndexRange verilogRange : indexRanges) {

      list.add(new Range(verilogRange.min, verilogRange.max));
    }

    return list;
  }

  private static String getTypeString(final VerilogDescriptor descriptor) {

    final DataType dataType = descriptor.getDataType();

    if (dataType.isTypeId(DataTypeId.BIT_VECTOR)) {

      final int size = dataType.getSize();
      final int offset = descriptor.getOffset();

      return String.format("[%s:%s]", size + offset - 1, offset);
    }
    return dataType.toString();
  }
}
