/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.lang3.StringUtils;
import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGConcurrentStatement;
import org.zamia.instgraph.IGContainer;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGDefaultSequentialStatementVisitor;
import org.zamia.instgraph.IGInstantiation;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationArrayAggregate;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGRecordField;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssert;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialExit;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialNext;
import org.zamia.instgraph.IGSequentialProcedureCall;
import org.zamia.instgraph.IGSequentialReport;
import org.zamia.instgraph.IGSequentialReturn;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGStructure;
import org.zamia.instgraph.IGSubProgram;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.util.PathName;
import ru.ispras.castle.ast.Range;
import ru.ispras.castle.ast.RangedVariable;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgAssertStatement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.parser.vhdl.mapper.IgSubprogramMapper;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * IG structure objects parser.
 *
 * <p>Component of VHDL parser that transforms IG (Instantiation Graph)
 * structure objects representing processes, modules, instances to
 * control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgStructureVisitor extends IGDefaultSequentialStatementVisitor {

  /* The IG representation manager. */
  private final IGManager manager;

  /* The IG interpreter runtime environment. */
  private final IGInterpreterRuntimeEnv runtimeEnv;

  /* The parent top-level module. */
  private final CfgModule topLevel;

  /* The list of current structure statements. */
  private final List<Statement> statements;

  /* The current process. */
  private CfgProcess process;

  /* The builder for VHDL variables. */
  private VariableBuilder builder;

  /* Mapping from process id to ids of related instances. */
  private Map<Long, Deque<Long>> processInstMap;

  /* The identifier of the instance which process is elaborated. */
  private Long instanceId;

  /* Mapping from instance id to related object-to-object bindings. */
  private Map<Long, Map<IGObject, IGObject>> instIgObjMap;

  /* Mapping from instance id to related object-to-operation bindings. */
  private Map<Long, Map<IGObject, IGOperation>> instIgOperationMap;

  /* Instance hierarchical name. */
  private LinkedList<String> namePath;

  /* Mapping from IG object to their hierarchical name prefixes. */
  private Map<IGObject, String> namePrefixMap;

  /* The mapping from sub-program calls to the new variables that keep their return values. */
  private final Map<String, NodeVariable> progReturnMap;

  /* The stack for sub-program tags that are processed right now. */
  private final Deque<String> progTags;

  /* Formal-actual param mapping for the processed sub-program. */
  private Map<String, IGOperation> progParams;

  /* Constants that keep severity string values. */
  private static final String SEVERITY_NOTE = "NOTE";
  private static final String SEVERITY_WARN = "WARNING";
  private static final String SEVERITY_ERROR = "ERROR";
  private static final String SEVERITY_FAIL = "FAILURE";

  /**
   * Creates object with the specified IG manager, IG runtime environment and the top-level module.
   *
   * @param manager IG representation manager.
   * @param runtimeEnv IG runtime environment.
   * @param module Top-level module.
   *
   * @throws IllegalArgumentException when {@code module} argument is {@code null}.
   */
  IgStructureVisitor(
      final IGManager manager,
      final IGInterpreterRuntimeEnv runtimeEnv,
      final CfgModule module) {

    super(new IgOperationVisitor());
    InvariantChecks.checkNotNull(module);

    getOperationCfgVisitor().setParentVisitor(this);
    setBuilder(new VariableBuilder(this));

    this.manager = manager;
    this.runtimeEnv = runtimeEnv;
    this.topLevel = module;
    this.statements = new LinkedList<>();
    this.progReturnMap = new LinkedHashMap<>();
    this.progTags = new LinkedList<>();
    this.progParams = new LinkedHashMap<>();
    this.processInstMap = new LinkedHashMap<>();
    this.instIgObjMap = new LinkedHashMap<>();
    this.instIgOperationMap = new LinkedHashMap<>();
    this.namePath = new LinkedList<>();
    this.namePrefixMap = new LinkedHashMap<>();
  }

  /**
   * Creates the visitor object as a copy of the specified one.
   * @param visitor The object to be copied.
   */
  private IgStructureVisitor(final IgStructureVisitor visitor) {
    this(visitor.getManager(), visitor.getRuntimeEnv(), visitor.getTopLevelModule());

    this.process = visitor.getProcess();
    this.progParams = new LinkedHashMap<>(visitor.progParams);
    this.processInstMap = new LinkedHashMap<>(visitor.processInstMap);
    this.instIgObjMap = new LinkedHashMap<>(visitor.instIgObjMap);
    this.instIgOperationMap = new LinkedHashMap<>(visitor.instIgOperationMap);
    this.instanceId = visitor.instanceId;
    this.namePath = new LinkedList<>(visitor.namePath);
    this.namePrefixMap = new LinkedHashMap<>(visitor.namePrefixMap);
  }

  /**
   * Returns sub-visitor for IGOperation objects.
   *
   * @return IGOperation objects sub-visitor.
   */
  IgOperationVisitor getOperationCfgVisitor() {

    return (IgOperationVisitor) this.fOperationVisitor;
  }

  /**
   * Returns top-level module.
   *
   * @return Top-level module that is processed at the current moment.
   */
  CfgModule getTopLevelModule() {
    return this.topLevel;
  }

  /**
   * Returns control flow graph model statements collection.
   *
   * @return Collection of statements that are constructed at the current moment.
   */
  public List<Statement> getStatements() {
    return this.statements;
  }

  /**
   * Returns IG interpreter runtime environment.
   *
   * @return Runtime environment.
   */
  IGInterpreterRuntimeEnv getRuntimeEnv() {
    return this.runtimeEnv;
  }

  /**
   * Returns IG manager.
   *
   * @return IG manager.
   */
  public IGManager getManager() {
    return this.manager;
  }

  /**
   * Returns the current process that is under construction.
   * @return The current process that is under construction.
   */
  public CfgProcess getProcess() {
    return this.process;
  }

  /**
   * Sets the specified process to be constructed by this visitor.
   *
   * @param process Process to be constructed.
   */
  private void setProcess(final CfgProcess process) {
    this.process = process;
  }

  /**
   * Returns inner variable and input\output signal builder.
   *
   * @return Variable builder.
   */
  public VariableBuilder getBuilder() {
    return this.builder;
  }

  private void setBuilder(final VariableBuilder builder) {
    this.builder = builder;
  }

  /**
   * Registers the specified mapping for arguments of the processed sub-program.
   * @param mapping The formal-to-actual mapping of the processed sub-program arguments.
   */
  private void registerMapping(final IGMapping mapping) {

    this.progParams.put(mapping.getFormal().toHRString(), mapping.getActual());
  }

  /**
   * Clears all the mapping for the processed sub-program.
   */
  void clearMappings() {
    this.progParams.clear();
  }

  /**
   * Returns the hierarchical name prefix for variable name.
   * @return The hierarchical name prefix for variable name.
   */
  String getPathPrefix() {

    return StringUtils.join(this.namePath, "_");
  }

  private void putToNamePath(final String name) {

    this.namePath.push(name);
  }

  private void popPrefix() {

    this.namePath.poll();
  }

  /**
   * Stores the specified hierarchical name prefix for the specified IG object.
   * @param igObject The IG object.
   * @param prefix The prefix for the IG object name.
   */
  void storeNamePrefix(final IGObject igObject, final String prefix) {

    this.namePrefixMap.put(igObject, prefix);
  }

  /**
   * Checks whether name prefix is registered for the specified IG object.
   * @param igObject The IG object that is checked whether name prefix is stored for it.
   * @return {@code true} when the name prefis is already stored for the specified IG object,
   *         {@code false} otherwise.
   */
  boolean hasNamePrefix(final IGObject igObject) {

    return this.namePrefixMap.containsKey(igObject);
  }

  /**
   * Returns name prefix for the specified IG object.
   * @param igObject The IG object that name prefix is returned.
   * @return The name prefix for the specified IG object or {@code null} if there is no such.
   */
  String getNamePrefix(final IGObject igObject) {

    return this.namePrefixMap.get(igObject);
  }

  /**
   * Checks whether this visitor processes a sub-program with the specified formal argument.
   * @param formalArg The name of the formal argument for the processed sub-program.
   * @return {@code true} when this visitor processes a sub-program with the specified formal
   *         argument, {@code false} otherwise.
   */
  boolean hasFormalArg(final String formalArg) {
    return this.progParams.containsKey(formalArg);
  }

  /**
   * Returns the actual argument for the specified formal one.
   * @param formalArg The formal argument name.
   * @return The actual argument for the specified formal one.
   */
  IGOperation getActualArg(final String formalArg) {
    return this.progParams.get(formalArg);
  }

  /**
   * Checks whether does the sub-program with the specified tag is already elaborated.
   * @param subProgramTag The sub-program tag.
   * @return {@code true} when sub-program with the specified tag has already been elaborated,
   *         {@code false} otherwise.
   */
  boolean hasProgramElaborated(final String subProgramTag) {
    return this.progReturnMap.containsKey(subProgramTag);
  }

  /**
   * Returns the resulting value variable that is related to the sub-program with the specified tag.
   * @param subProgramTag The sub-program tag.
   * @return The resulting value variable that is related to the sub-program with the specified tag.
   */
  NodeVariable getReturnVar(final String subProgramTag) {
    return this.progReturnMap.get(subProgramTag);
  }

  private void registerTag(final String subProgramTag, final NodeVariable tagVariable) {
    this.progReturnMap.put(subProgramTag, tagVariable);
  }

  private void addSubProgramTag(final String subProgramTag) {
    this.progTags.push(subProgramTag);
  }

  private void removeSubProgramTag() {
    this.progTags.pop();
  }

  private String getCurrentSubProgramTag() {
    return this.progTags.peek();
  }

  /**
   * Checks whether the specified object is mapped to something through this component execution.
   * @param object The object that is mapped to something.
   * @return {@code true} when the specified object is mapped to something through this component
   *         execution, {@code false} otherwise.
   */
  boolean hasMapping(final IGObject object) {

    if (this.instanceId == null) {
      return false;
    }

    return this.instIgObjMap.get(this.instanceId).containsKey(object);
  }

  /**
   * Checks whether the specified IG object is mapped by another one,
   * @param object The object that is checked.
   * @return {@code true} if the specified IG object is mapped to another one,
   *         {@code false} otherwise.
   */
  boolean hasMapped(final IGObject object) {

    if (this.instanceId == null) {
      return false;
    }

    return this.instIgObjMap.get(this.instanceId).containsValue(object);
  }

  /**
   * Returns the object that is mapped to the specified one through this component execution.
   * @param object The object that is mapped to the specified one through this component execution.
   * @return the object that is mapped to the specified one through this component execution.
   */
  IGObject getMapping(final IGObject object) {

    return this.instIgObjMap.get(this.instanceId).get(object);
  }

  boolean hasIgMapping(final IGObject object) {

    if (this.instanceId == null) {
      return false;
    }

    return this.instIgOperationMap.get(this.instanceId).containsKey(object);
  }

  /**
   * Returns the IG parent object that contains the specified one.
   * @param object The object that parent one is to be returned.
   * @return The IG parent object that contains the specified one if it is already elaborated,
   *         {@code null} otherwise.
   */
  IGOperation getIgMapping(final IGObject object) {

    return this.instIgOperationMap.get(this.instanceId).get(object);
  }

  private void visitProcess(final IGProcess igProcess) {

    final long processId = igProcess.getDBID();
    if (this.processInstMap.containsKey(processId)) {

      final Deque<Long> instIds = this.processInstMap.get(processId);

      if (instIds.isEmpty()) {
        this.processInstMap.remove(processId);
        this.instanceId = null;
      } else {
        this.instanceId = instIds.pop();
      }
    } else {
      this.instanceId = null;
    }

    final EventList sensList = getBuilder().getProcessEventList(igProcess);
    final CfgProcess newProcess = new CfgProcess(igProcess.getLabel(), sensList, false);
    setProcess(newProcess);

    /* Elaborate internal statements. */
    final IGSequenceOfStatements sos = igProcess.getSequenceOfStatements();
    visit(sos);

    final List<Statement> processNodes = new LinkedList<>(this.statements);
    newProcess.addAll(processNodes);

    /* Declare process variables at the parent module. */
    final IGContainer container = igProcess.getContainer();
    final Map<NodeVariable, Declaration> declarations = getDeclarations(container);

    if (!declarations.isEmpty()) {

      Map<String, Node> bindings = new LinkedHashMap<>();

      for (final Map.Entry<NodeVariable, Declaration> entry : declarations.entrySet()) {

        final NodeVariable variable = entry.getKey();
        final String name = variable.getName();
        final DataType type = variable.getDataType();
        final Declaration declaration = entry.getValue();

        final NodeVariable global =
            new NodeVariable(getProcessVarName(name, igProcess, getTopLevelModule()), type);
        global.setUserData(variable.getUserData());

        getTopLevelModule().declareVariable(global, declaration);
        bindings.put(name, global);
      }

      newProcess.apply(bindings);
    }

    /* Link process node with parent & child statements. */
    getTopLevelModule().add(newProcess);

    this.statements.clear();
    setProcess(null);
  }

  private String getProcessVarName(
      final String varName,
      final IGProcess process,
      final CfgModule module) {

    final String processPrefix =
        process.getLabel() == null
            ? String.format("process_%s", module.getStatementNum())
            : process.getLabel();

    return String.format("%s.%s", processPrefix, varName);
  }

  /**
   * Returns variables and their declarations from the specified container.
   * @param container The container that keeps variable declarations.
   * @return The mapping from variables to their declarations.
   */
  Map<NodeVariable, Declaration> getDeclarations(final IGContainer container) {

    final Map<NodeVariable, Declaration> map = new LinkedHashMap<>();

    for (int i = 0; i < container.getNumLocalItems(); i++) {

      final IGContainerItem item = container.getLocalItem(i);

      /* Elaborate variables (not-a-constants). */
      if (item instanceof IGObject && !IgUtils.isConstant((IGObject) item)) {

        if (((IGObject) item).getType().isRecord()) {
          final SourceLocation src = item.computeSourceLocation();
          final IGType type = ((IGObject) item).getType();
          try {
            for (int fieldIndex = 0; fieldIndex < type.getNumRecordFields(src); fieldIndex++) {
              final String name = item.getId();
              final  IGRecordField recordField = type.getRecordField(fieldIndex, src);
              final NodeVariable variable =
                  getBuilder().getRecordField(name, recordField, VariableType.varType);

              if (variable == null) {

                /* Non-synthesizable. */
                continue;
              }

              final Range invariant = getBuilder().getInvariant(variable, recordField.getType());
              map.put(variable, new Declaration(null, invariant));
            }
          } catch (ZamiaException e) {
            e.printStackTrace();
          }
        } else {

          final NodeVariable variable = getBuilder().getObjectVariable((IGObject) item);
          addDeclaration(map, (IGObject) item, variable);
        }
      }
    }
    return map;
  }

  private void addDeclaration(Map<NodeVariable, Declaration> map, IGObject item,
      NodeVariable variable) {
    if (variable == null) {

      /* Non-synthesizable. */
      return;
    }

    final Node initValue = getInitialValue(item);
    final Range invariant = getBuilder().getInvariant(variable, item.getType());

    map.put(variable, new Declaration(initValue, invariant));
  }

  private Node getInitialValue(final IGObject igObject) {
    Node initValue;

    final IGOperation igValue = igObject.getInitialValue();

    /* Try to calculate LEFT expression since it is initial for VHDL variable. */
    if (igValue == null) {
      try {

        final IGType type = igObject.getType();
        final IGTypeStatic stType = type.computeStaticType(getRuntimeEnv(), null, null);
        final SourceLocation srcLoc = igObject.computeSourceLocation();
        final IGStaticValue left =
            stType.getStaticRange() == null ? null : stType.getStaticLeft(srcLoc);
        final IGStaticValue leftVal = left == null ? IGStaticValue.generateZ(stType, srcLoc) : left;

        initValue = leftVal == null ? null : getOperationNode(leftVal);

      } catch (final ZamiaException e) {
        throw new RetrascopeException(e);
      }
    } else {
      initValue = getOperationNode(igValue);
    }

    if (initValue == null) {
      throw new RetrascopeException("Can't calculate initial value of: " + igObject.toString());
    }
    return initValue;
  }

  @Override
  public void visit(final IGSequentialStatement statement) {

    if (!(statement instanceof IGSequenceOfStatements) && statement != null) {

      final IGOperationInvokeSubprogram[] subPrograms = getSubPrograms(statement);

      for (final IGOperationInvokeSubprogram program : subPrograms) {

        for (int i = 0; i < program.getNumMappings(); i++) {
          registerMapping(program.getMapping(i));
        }

        final String tagName = IgUtils.getTag(program);

        addSubProgramTag(tagName);

        final IGSubProgram programCode = program.getSub();
        final DataType returnType = getBuilder().getDataType(programCode.getReturnType());
        final NodeVariable returnVar = new NodeVariable(tagName, returnType);
        final IgSubprogramBuilder builder = new IgSubprogramBuilder(new IgStructureVisitor(this));
        final Statement[] statements = builder.build(program, tagName);

        registerTag(tagName, returnVar);
        removeSubProgramTag();
        addToThis(statements);
        clearMappings();
      }
    }
    super.visit(statement);
  }

  @Override
  public void visit(final IGStructure igStructure, final PathName pathName) {

    final IGContainer igContainer = igStructure.getContainer();

    final String prefix = pathName.toString();
    putToNamePath(prefix);
    getMappings(igStructure);

    for (int i = 0; i < igStructure.getNumStatements(); i++) {

      final IGConcurrentStatement stmt = igStructure.getStatement(i);

      if (stmt instanceof IGProcess) {


        visitProcess((IGProcess) stmt);

      } else if (stmt instanceof IGStructure) {

        visit((IGStructure) stmt, pathName);
      }
    }

    final Map<NodeVariable, Declaration> declarations = getDeclarations(igContainer);
    getTopLevelModule().declareVariables(declarations);
    popPrefix();
  }

  private void getMappings(final IGStructure igStructure) {

    /* Extract mappings from instances */
    for (int i = 0; i < igStructure.getNumStatements(); i++) {

      final IGConcurrentStatement stmt = igStructure.getStatement(i);

      if (stmt instanceof IGInstantiation) {

        getMappings((IGInstantiation) stmt);
      }
    }
  }

  private void getMappings(final IGInstantiation igInstance) {

    final Map<IGObject, IGObject> igMap = new LinkedHashMap<>();
    final Map<IGObject, IGOperation> operationMap = new LinkedHashMap<>();

    for (int mapIdx = 0; mapIdx < igInstance.getNumMappings(); mapIdx++) {

      final IGMapping igMapping = igInstance.getMapping(mapIdx);

      final IGOperation formal = igMapping.getFormal();
      final IGOperation actual = igMapping.getActual();

      if (!(formal instanceof IGOperationObject)) {
        Log.warn(LogMessage.UNSUPPORTED_TYPE + formal.getClass().toString());
        continue;
      }

      if (!(actual instanceof IGOperationObject)) {
        Log.warn(LogMessage.UNSUPPORTED_TYPE + formal.getClass().toString());
        continue;
      }

      final IGOperationObject formalObj = (IGOperationObject) formal;
      final IGObject formalIgObj = formalObj.getObject();
      final IGOperationObject actualObj = (IGOperationObject) actual;
      final IGObject actualIgObj = actualObj.getObject();

      igMap.put(formalIgObj, getActual(actualIgObj));
      operationMap.put(formalIgObj, formalObj);
      operationMap.put(actualIgObj, actualObj);
    }

    final Long instId = igInstance.getDBID();
    this.instIgObjMap.put(instId, igMap);
    this.instIgOperationMap.put(instId, operationMap);

    IGModule instModule = (IGModule) igInstance.findChild(null);
    final IGStructure instStructure = instModule.getStructure();

    for (int j = 0; j < instStructure.getNumStatements(); j++) {

      final IGConcurrentStatement subStmt = instStructure.getStatement(j);

      if (subStmt instanceof IGProcess) {

        final long subProcessId = subStmt.getDBID();

        if (this.processInstMap.containsKey(subProcessId)) {

          this.processInstMap.get(subProcessId).add(instId);
        } else {

          final Deque<Long> instIds = new LinkedList<>();
          instIds.add(instId);

          this.processInstMap.put(subProcessId, instIds);
        }

      } else if (subStmt instanceof IGStructure) {

        getMappings((IGStructure) subStmt);

      } else if (subStmt instanceof IGInstantiation) {

        getMappings((IGInstantiation) subStmt);

      } else {

        throw RetrascopeException.exception(
            LogMessage.UNSUPPORTED_TYPE, subStmt.getClass().toString());
      }
    }
  }

  private IGObject getActual(final IGObject actual) {

    IGObject result = actual;

    if (this.instanceId == null) {
      return result;
    }

    while (this.instIgObjMap.get(this.instanceId).containsKey(result)) {

      result = this.instIgObjMap.get(this.instanceId).get(result);
    }

    return result;
  }

  private IGOperationInvokeSubprogram[] getSubPrograms(final IGSequentialStatement statement) {

    final List<IGOperationInvokeSubprogram> subPrograms = new LinkedList<>();

    final int childNum = statement.getNumChildren();

    for (int i = 0; i < childNum; i++) {
      final IGItem child = statement.getChild(i);
      if (child != null) {
        getSubPrograms(subPrograms, child);
      }
    }

    return subPrograms.toArray(new IGOperationInvokeSubprogram[0]);
  }

  private void getSubPrograms(
      final List<IGOperationInvokeSubprogram> subPrograms,
      final IGItem item) {

    if (item instanceof IGOperationInvokeSubprogram) {

      final String name = IgUtils.getName((IGOperationInvokeSubprogram) item);

      if (!IgSubprogramMapper.hasOperation(name) && !IgSubprogramMapper.hasMacro(name)) {
        subPrograms.add((IGOperationInvokeSubprogram) item);
      }
    } else {
      final int itemChildNum = item.getNumChildren();

      for (int i = 0; i < itemChildNum; i++) {
        final IGItem itemChild = item.getChild(i);

        if (itemChild != null) {
          getSubPrograms(subPrograms, itemChild);
        }
      }
    }
  }

  @Override
  public void visitWait(final IGSequentialWait igWait) {

    final IGOperation clause = igWait.getConditionClause();
    final Node condition = (clause == null) ? null : getOperationNode(clause);

    final EventList list = getBuilder().getEventList(igWait);
    final CfgWaitStatement waitNode =
        (condition == null) ? new CfgWaitStatement(list) : new CfgWaitStatement(list, condition);

    addToThis(waitNode);
  }

  @Override
  public void visitReturn(final IGSequentialReturn igReturn) {

    final IGItem igReturnValue = igReturn.getChild(0);
    final Node returnNode = getOperationNode((IGOperation) igReturnValue);

    final NodeVariable returnVar = getReturnVar(getCurrentSubProgramTag());
    final CfgBlockStatement returnBlock = new CfgBlockStatement(returnVar, returnNode);
    addToThis(returnBlock);
  }

  @Override
  public void visitReport(final IGSequentialReport igReport) {
    throw RetrascopeException.exception(LogMessage.CANT_PARSE, igReport.toHRString());
  }

  @Override
  public void visitCall(final IGSequentialProcedureCall igCall) {
    throw RetrascopeException.exception(LogMessage.CANT_PARSE, igCall.toHRString());
  }

  @Override
  public void visitNext(final IGSequentialNext igNext) {
    throw RetrascopeException.exception(LogMessage.CANT_PARSE, igNext.toHRString());
  }

  @Override
  public void visitLoop(final IGSequentialLoop igLoop) {

    /* IG loop node children. */
    final IGItem child0 = igLoop.getChild(0);

    Node left;
    Node right;
    boolean isAscending;

    if (child0 instanceof IGOperationAttribute) {

      final IGOperationAttribute.AttrOp attrOp = ((IGOperationAttribute) child0).getAttrOp();

      if (!attrOp.equals(IGOperationAttribute.AttrOp.RANGE)) {

        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, attrOp);
      }

      final IGType type = ((IGOperationAttribute) child0).getType();

      if (!(type instanceof IGTypeStatic)) {

        throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, type.getClass());
      }

      final IGTypeStatic staticType = (IGTypeStatic) type;
      final SourceLocation srcLoc = child0.computeSourceLocation();

      if (staticType.getStaticRange() == null) {

        IGType igElemType = staticType.getElementType();

        if (!(igElemType instanceof IGTypeStatic)) {

          throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, igElemType.getClass());
        }

        final IGTypeStatic elemType = (IGTypeStatic) igElemType;

        if (elemType.isUniversal()) {

          final IGItem item = ((IGOperationAttribute) child0).getItem();

          if (!(item instanceof IGOperationObject)) {

            throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, item.getClass());
          }

          final IGOperationObject igOpObject = (IGOperationObject) item;
          final IGObject igObject = igOpObject.getObject();

          final IGObject.IGObjectCat category = igObject.getCat();

          if (!category.equals(IGObject.IGObjectCat.CONSTANT)) {

            throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, category);
          }

          final IGOperation initValue = igObject.getInitialValue();

          if (!(initValue instanceof IGOperationArrayAggregate)) {

            throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, initValue);
          }

          final IGOperationArrayAggregate initArray = (IGOperationArrayAggregate) initValue;

          final int itemNum = initArray.getNumOperands();

          left = NodeValue.newInteger(0);
          right = NodeValue.newInteger(itemNum - 1);
          isAscending = true;

        } else {

          try {

            left = getOperationCfgVisitor().getNode(elemType.getStaticLeft(srcLoc));
            right = getOperationCfgVisitor().getNode(elemType.getStaticRight(srcLoc));

            isAscending = true;

          } catch (final ZamiaException e) {
            throw new RetrascopeException(e);
          }
        }

      } else {

        try {

          left = getOperationCfgVisitor().getNode(staticType.getStaticLeft(srcLoc));
          right = getOperationCfgVisitor().getNode(staticType.getStaticRight(srcLoc));

          isAscending = staticType.isAscending();

        } catch (final ZamiaException e) {
          throw new RetrascopeException(e);
        }
      }
    } else if (child0 instanceof IGRange) {

      final IGRange loopRange = (IGRange)child0;
      left = getOperationNode(loopRange.getLeft());
      right = getOperationNode(loopRange.getRight());
      final Node ascending = getOperationNode(loopRange.getAscending());
      isAscending = ascending.equals(Nodes.TRUE);

    } else {
      throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, child0.getClass());
    }

    final IGItem loopCounter = igLoop.getChild(1);
    final IGSequenceOfStatements loopBody = (IGSequenceOfStatements) igLoop.getChild(4);

    /* The range iteration variable. */
    final NodeVariable rangeVariable = getBuilder().getObjectVariable((IGObject) loopCounter);

    /* The loop condition. */

    final Node range = createRange(rangeVariable, left, right, isAscending);

    /* Create the "i := 0" initialization block. */
    final Assignment init = new Assignment(rangeVariable, left);

    /* Create the incrementing "i++" basic block. */
    checkRangeParam(rangeVariable);

    final StandardOperation addOp = StandardOperation.ADD;
    final NodeValue one = NodeValue.newInteger(1);
    final NodeOperation incrementNode = new NodeOperation(addOp, rangeVariable, one);
    final Assignment increment = new Assignment(rangeVariable, incrementNode);

    final CfgLoopStatement loopStatement =
        new CfgLoopStatement(CfgLoopStatement.Type.FOR, range, init, increment);
    loopStatement.addAll(getSubGraph(loopBody));
    addToThis(loopStatement);
  }

  private void checkRangeParam(final NodeVariable rangeVariable) {

    if (!rangeVariable.isType(DataTypeId.LOGIC_INTEGER)) {

      final DataType dataType = rangeVariable.getDataType();

      throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, dataType);
    }
  }

  @Override
  public void visitIf(final IGSequentialIf igIf) {

    final IGOperation ifCondition = igIf.getCond();
    final IGSequenceOfStatements elseSos = igIf.getElseSOS();
    final IGSequenceOfStatements thenSos = igIf.getThenSOS();

    final Node condition = getOperationNode(ifCondition);

    /* Transform operands & detect event-dependent part of "x'event && x='0'" condition .*/
    final EventList events = getEvents(ifCondition);

    final CfgSwitchStatement switchStatement = new CfgSwitchStatement(condition);

    final CfgCaseStatement trueCase = new CfgCaseStatement(events, (NodeValue) Nodes.TRUE);
    switchStatement.add(trueCase);
    if (thenSos != null) {
      trueCase.addAll(getSubGraph(thenSos));
    }

    final CfgCaseStatement falseCase = new CfgCaseStatement((NodeValue) Nodes.FALSE);
    switchStatement.add(falseCase);
    if (elseSos != null) {
      falseCase.addAll(getSubGraph(elseSos));
    }

    addToThis(switchStatement);
  }

  private EventList getEvents(final IGOperation ifCondition) {

    final EventList list = new EventList();

    for (int i = 0; i < ifCondition.getNumOperands(); i++) {

      final IGOperation operand = ifCondition.getOperand(i);

      if (isEvent(operand)) {

        final Event event = (Event)getOperationCfgVisitor().getData(operand);
        list.add(event);
      }
    }

    return list;
  }

  private static boolean isEvent(final IGOperation operand) {
    return operand instanceof IGOperationAttribute
        && ((IGOperationAttribute)operand).getAttrOp().equals(IGOperationAttribute.AttrOp.EVENT);
  }

  @Override
  public void visitExit(final IGSequentialExit igExit) {
    throw RetrascopeException.exception(LogMessage.CANT_PARSE, igExit.toHRString());
  }

  @Override
  public void visitAssert(final IGSequentialAssert igAssert) {

    final IGOperation assertOp = igAssert.getOp();
    final Node condition = getOperationNode(assertOp);
    final CfgAssertStatement.Severity severity = getSeverity(igAssert);
    final String report = igAssert.getReport().toHRString();

    final CfgAssertStatement assertStatement = new CfgAssertStatement(condition, severity, report);

    addToThis(assertStatement);
  }

  private CfgAssertStatement.Severity getSeverity(final IGSequentialAssert igAssert) {

    final String sevString = igAssert.getSeverity().toHRString();

    if (SEVERITY_NOTE.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.NOTE;

    } else if (SEVERITY_WARN.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.WARNING;

    } else if (SEVERITY_ERROR.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.ERROR;

    } else if (SEVERITY_FAIL.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.FAILURE;

    } else {
      Log.warn("%s%s.", LogMessage.CANT_PARSE, sevString);
    }
    return null;
  }

  @Override
  public void visitAssignment(final IGSequentialAssignment igAssignment) {

    final IGOperation target = igAssignment.getTarget();
    final IGOperation value = igAssignment.getValue();

    if (target.getType().isRecord() && value.getType().isRecord()) {

      addRecordAssignments(target,value);

    } else {
      addAssignment(target, value);
    }
  }

  private void addAssignment(final IGOperation target, final IGOperation value) {
    addToThis(getBlock(target, getOperationNode(value)));
  }

  private Node getOperationNode(final IGOperation node) {
    return getOperationCfgVisitor().getNode(node);
  }

  private void addRecordAssignments(final IGOperation target, final IGOperation value) {
    final SourceLocation srcTar = target.computeSourceLocation();
    final SourceLocation srcVal = value.computeSourceLocation();
    final IGType tarType = target.getType();
    final IGType valType = value.getType();
    if (!tarType.equals(valType)) {
      Log.error("Different record types");

      return;
    }

    final VariableType targetVarType = VariableBuilderUtils.getVariableType(target);
    final VariableType valueVarType = VariableBuilderUtils.getVariableType(value);

    try {

      for (int i = 0; i < tarType.getNumRecordFields(srcTar); i++) {
        final String valId = value.getId();
        final String tarId = target.getId();
        final String nameVal =
            valId == null ? ((IGOperationObject) value).getObject().getId() : value.getId();
        final String nameTar =
            tarId == null ? ((IGOperationObject) target).getObject().getId() : target.getId();

        final IGRecordField recordFieldTar = tarType.getRecordField(i, srcTar);
        final IGRecordField recordFieldVal = valType.getRecordField(i, srcVal);
        final NodeVariable tarVar =
            getBuilder().getRecordField(nameTar, recordFieldTar, targetVarType);
        final NodeVariable valVar =
            getBuilder().getRecordField(nameVal, recordFieldVal, valueVarType);

        addToThis(new CfgBlockStatement(tarVar, valVar));
      }
    } catch (final ZamiaException zamiaExc) {
      throw new RetrascopeException(zamiaExc);
    }


  }

  private CfgBlockStatement getBlock(final IGOperation target, final Node valueExpr) {

    final RangedVariable rangedVariable = getBuilder().getRangedVariable(target);

    final Assignment assignment = new Assignment(rangedVariable, valueExpr);

    /* If the variable is of signal type, then basic block is concurrent. */
    final boolean isConcurrent = isSignal(rangedVariable);

    return new CfgBlockStatement(assignment, isConcurrent);
  }

  private boolean isSignal(final RangedVariable rangedVariable) {

    final NodeVariable variable = rangedVariable.getVariable();
    final Object userData = variable.getUserData();

    return VariableType.isSignal(((VariableDescriptor) userData).getVariableType());
  }

  private List<Statement> getSubGraph(final IGSequenceOfStatements stmt) {

    final IgStructureVisitor visitor = new IgStructureVisitor(this);
    visitor.setProcess(getProcess());
    visitor.visit(stmt);

    return visitor.getStatements();
  }

  private Node createRange(
      final NodeVariable var,
      final Node left,
      final Node right,
      final boolean isAscending) {

    /* Select operation identifiers depending on data types. */
    final boolean areBitVector = ExprUtils.isType(DataTypeId.BIT_VECTOR, left, right);

    final StandardOperation gtId =
        areBitVector ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;
    final StandardOperation lessId =
        areBitVector ? StandardOperation.BVULE : StandardOperation.LESSEQ;

    Node expr;
    if (isAscending) {

      /* The result is: left <= var <= right. */
      expr = Nodes.and(new NodeOperation(gtId, var, left), new NodeOperation(lessId, var, right));
    } else {

      /* The result is: left >= var >= right. */
      expr = Nodes.and(new NodeOperation(lessId, var, left), new NodeOperation(gtId, var, right));
    }

    return expr;
  }

  private void addToThis(final Statement statement) {
    this.statements.add(statement);
  }

  private void addToThis(final Statement ... statements) {
    for (final Statement statement : statements) {
      addToThis(statement);
    }
  }
}
