/*
 * Copyright 2017-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.ZamiaException;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGType;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.LogMessage;

/**
 * Utility methods for {@code VariableBuilder} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VariableBuilderUtils {

  private VariableBuilderUtils() {
    /* Empty. */
  }

  /**
   * Extracts variable type of the specified item.
   *
   * @param item Item to be processed.
   *
   * @return A {@link VariableType} variable type that is extracted from the specified item.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when argument is neither
   *         of {@code IGOperation} nor of {@code IGObject} type.
   */
  static VariableType getVariableType(final IGContainerItem item) {
    InvariantChecks.checkNotNull(item);

    VariableType type;
    if (item instanceof IGOperation) {

      type = getOperationVariableType((IGOperation) item);

    } else if (item instanceof IGObject) {

      type = getObjectVariableType((IGObject) item);

    } else {
      throw RetrascopeException.exception(LogMessage.CANT_PARSE, item.getClass());
    }

    return type;
  }

  private static VariableType getOperationVariableType(final IGOperation operation) {

    try {
      final IGObject.OIDir direction = operation.getDirection();
      return toVariableType(direction);
    } catch (ZamiaException e) {
      throw new RetrascopeException(e);
    }
  }

  private static VariableType getObjectVariableType(final IGObject obj) {

    final  IGObject.OIDir direction = obj.getDirection();
    return toVariableType(direction);
  }

  private static VariableType toVariableType(final IGObject.OIDir direction) {
    VariableType type;
    switch (direction) {
      case NONE:
        type = VariableType.REG;
        break;
      case IN:
      case LINKAGE:
        type = VariableType.IN;
        break;
      case INOUT:
        type = VariableType.INOUT;
        break;
      case OUT:
        type = VariableType.OUT;
        break;
      default:
        throw RetrascopeException.exception(LogMessage.CANT_PARSE, direction);
    }
    return type;
  }

  /**
   * Returns the data type that corresponds to the specified IG type.
   * @param type The IG type.
   * @return The data type that corresponds to the specified IG type.
   */
  static DataType getEnumDataType(final IGType type) {

    DataType dataType;

    if (type.isBit()) {

      dataType = DataType.bitVector(1);

    } else if (type.isBool()) {

      dataType = DataType.BOOLEAN;

    } else if (type.isLogic()) {

      dataType = DataType.bitVector(1);

    } else if (type.isDiscrete()) {

      dataType = type.isCharEnum() ? DataType.bitVector(1) : DataType.INTEGER;

    } else {
      throw RetrascopeException.exception(LogMessage.CANT_PARSE, type.toHRString());
    }
    return dataType;
  }
}
