/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.cli.ParseException;
import org.apache.commons.io.FileUtils;
import org.apache.log4j.Level;
import org.zamia.BuildPath;
import org.zamia.SourceFile;
import org.zamia.ZamiaException;
import org.zamia.ZamiaLogger;
import org.zamia.ZamiaProject;
import org.zamia.ZamiaProjectBuilder;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.zdb.ZDBException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.MetaInfoValue;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.parser.basis.HdlParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.Map;

/**
 * Parser for VHDL hardware models.
 *
 * <p>Uses ZamiaCAD toolkit for making an intermediate representation (Instantiation Graph, IG)
 * before control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} building.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VhdlParser extends HdlParser {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "vhdl-parser";

  /**
   * The engine's input entity type.
   */
  private static final EntityType INPUT_TYPE = EntityType.get("vhd");

  /**
   * Default constructor.
   */
  public VhdlParser() {
    super(VhdlParser.ENGINE_ID);

    addInputType(VhdlParser.INPUT_TYPE);
    setOutputType(CfgModel.TYPE);

    addParameter(MODULE_NAME);
  }

  @Override
  public Entity parse(final Map<EntityType, Entity> inputs) {

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);
    final FileNames fileNames = (FileNames)inputs.get(VhdlParser.INPUT_TYPE);

    final CfgModel cfgModel = new CfgModel();

    try {

      /* Get top level modules names from command line. */
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      final String[] topLevels = getModuleNames(engineArgs, fileNames);

      for (int i = 0; i < topLevels.length; i++) {

        final String fileName = fileNames.getList().get(i);

        if (!new File(fileName).exists()) {

          throw RetrascopeException.exception(LogMessage.CANT_FIND, fileName);
        }

        /*
         * Create BuildPath.txt
         */

        /* Get full path to VHDL file directory. */
        final String path = getSourceDirPath(fileName);

        final File buildFile = FileCreator.newFile(path + File.separator + "BuildPath.txt");
        final BufferedWriter writer = ReadWriteUtils.newBufferedWriter(buildFile);

        writer.flush();
        writer.write(String.format("local WORK \"%s\"%n", path));
        writer.write(String.format("toplevel %s%n", topLevels[i].toUpperCase()));
        writer.close();

        /* Create Zamia IG. */

        final java.util.logging.Level logLevel = Log.getLogger().getLevel();

        ZamiaLogger.setup(logLevel.equals(LogLevel.DEBUG) ? Level.ALL : Level.OFF);

        final String dataBasePath = String.format("%s%sdb", path, File.separator);

        final ZamiaProject project = new ZamiaProject("", path, new SourceFile(buildFile), path);
        final ZamiaProjectBuilder builder = project.getBuilder();
        builder.build(true, true, null);

        /* Parse Zamia IG. */

        final IGManager igm = project.getIGM();
        final BuildPath buildPath = project.getBuildPath();

        /* Elaborate top level modules. */

        for (int j = 0; j < buildPath.getNumToplevels(); j++) {

          final IGModule topLevel = igm.findModule(buildPath.getToplevel(j));

          if (topLevel == null) {
            Log.error("%s%s module", LogMessage.CANT_FIND, topLevels[i]);
            System.exit(-1);
          }

          final CfgModule module = new CfgModule(topLevel.getDUUID().getId());
          module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
          module.addMetaInfo(MetaInfoType.FILE_NAME, fileName);

          final IGInterpreterRuntimeEnv runEnv = new IGInterpreterRuntimeEnv(null, project);
          final IgStructureVisitor structureVisitor = new IgStructureVisitor(igm, runEnv, module);

          topLevel.accept(structureVisitor, Integer.MAX_VALUE);
          cfgModel.addModule(structureVisitor.getTopLevelModule());
        }


        project.shutdown();
        project.clean();
        project.getZDB().clear();
        FileUtils.forceDeleteOnExit(new File(dataBasePath));
      }
    } catch (final ParseException | ZamiaException | IOException | ZDBException e) {
      throw new RetrascopeException(e);
    }
    return cfgModel;
  }

  @Override
  protected EntityType getSrcType() {
    return INPUT_TYPE;
  }

  private String getSourceDirPath(final String fileName) {

    final int sepIdx = fileName.lastIndexOf(File.separatorChar);

    return sepIdx < 0 ? System.getProperty("user.dir") : fileName.substring(0, sepIdx);
  }
}
