/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.property;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Named module-level property container.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class ModuleProperty extends DeclarationContainer {

  private final String name;

  private final ModuleProperty parent;

  private NodeVariable clock;

  private List<Node> properties;

  private final Map<String, ModuleProperty> nested;

  private final Map<String, ProcessProperty> processPropertyMap;

  /* Reset-like variable. */
  private NodeVariable resetVariable;

  /**
   * Constructs the object with the specified parameters.
   *
   * @param decStatement The statement which variable declarations are to be copied.
   * @param name Name of this object.
   * @param parent Parent container for this object.
   *
   * @throws IllegalArgumentException when name argument is {@code null}.
   */
  public ModuleProperty(
      final DecStatement decStatement,
      final String name,
      final ModuleProperty parent) {
    super();
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.parent = parent;
    this.properties = new LinkedList<>();
    this.nested = new LinkedHashMap<>();
    this.processPropertyMap = new LinkedHashMap<>();

    if (decStatement != null) {
      declareVariables(decStatement.getDeclarationMap());
    }
  }

  /**
   * Constructs the object with the specified parameters.
   *
   * @param name The object name.
   * @param parent The parent node for this object.
   */
  ModuleProperty(final String name, final ModuleProperty parent) {
    this(null, name, parent);
  }

  /**
   * Constructs the object with the specified name.
   * @param name The object name.
   */
  ModuleProperty(final String name) {
    this(null, name, null);
  }

  /**
   * Returns name of this object.
   *
   * @return Object name.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns parent container.
   *
   * @return Parent container for this object.
   */
  public ModuleProperty getParent() {
    return this.parent;
  }

  /**
   * Adds the specified expression that uses variables
   * that are already declared at the parent container.
   *
   * @param node Property to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addProperty(final Node node) {
    InvariantChecks.checkNotNull(node);

    this.properties.add(node);
  }

  /**
   * Returns a collection of properties.
   *
   * @return Properties of exactly this object.
   */
  public List<Node> getProperties() {
    return Collections.unmodifiableList(this.properties);
  }

  /**
   * Returns clock-like variable of the top-level container.
   *
   * @return Clock-like variable for this object.
   */
  public NodeVariable getClock() {
    return this.clock;
  }

  /**
   * Sets the specified variable as a clock-like for the top-level container.
   *
   * @param variable Clock-like variable to be set for this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when this is not top-level container.
   * @throws RetrascopeException when this container already has it's clock.
   */
  public void setClock(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    if (this.parent != null) {
      throw new RetrascopeException("Can't set clock to not top container.");
    }
    if (this.clock != null) {
      throw RetrascopeException.exception(LogMessage.CANT_REASSIGN, variable);
    }
    this.clock = variable;
  }

  /**
   * Adds the specified module-level property container object.
   *
   * @param container Module-level property container to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addModuleProperty(final ModuleProperty container) {
    InvariantChecks.checkNotNull(container);

    this.nested.put(container.getName(), container);
  }

  /**
   * Returns mapping of names to nested module-level containers.
   *
   * @return Mapping of names to property containers that this object contains.
   */
  Map<String, ModuleProperty> getModuleMap() {
    return Collections.unmodifiableMap(this.nested);
  }

  /**
   * Adds the specified process-level property container object.
   *
   * @param group Process-level property container to be added.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addProcessProperty(final ProcessProperty group) {
    InvariantChecks.checkNotNull(group);

    this.processPropertyMap.put(group.getName(), group);
  }

  /**
   * Returns mapping of names to nested process-level property containers.
   *
   * @return Mapping from names to related process-level property containers
   *         that a stored into this object.
   */
  public Map<String, ProcessProperty> getProcessMap() {
    return Collections.unmodifiableMap(this.processPropertyMap);
  }

  /**
   * Checks whether this object declares reset-like variable.
   *
   * @return {@code true} if this object declares reset-like variable, {@code false} otherwise.
   */
  private boolean hasResetVariable() {
    return resetVariable != null;
  }

  /**
   * Returns reset-like variable for this object.
   *
   * @return reset-like variable.
   */
  private NodeVariable getResetVariable() {
    return this.resetVariable;
  }

  /**
   * Sets the specified variable as reset-like one.
   * <p>Note that the specified variable should be declared at this object
   * before set as reset-like one.</p>
   * @param variable reset-like variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeException when the specified variable has not been already
   *         declared at this object.
   */
  public void setResetVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.resetVariable = variable;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    if (!super.equals(obj)) {
      return false;
    }

    final ModuleProperty that = (ModuleProperty) obj;

    if (getName() == null ? that.getName() != null : !getName().equals(that.getName())) {
      return false;
    }
    if (getParent() == null ? that.getParent() != null : !getParent().equals(that.getParent())) {
      return false;
    }
    if (getClock() == null ? that.getClock() != null : !getClock().equals(that.getClock())) {
      return false;
    }
    if (getProperties().isEmpty()
        ? !this.getProperties().isEmpty()
        : !getProperties().equals(this.getProperties())) {
      return false;
    }
    if (nested == null ? that.nested != null : !nested.equals(that.nested)) {
      return false;
    }
    if (processPropertyMap == null
        ? that.processPropertyMap != null :
        !processPropertyMap.equals(that.processPropertyMap)) {
      return false;
    }
    return getResetVariable() == null
        ? that.getResetVariable() == null
        : getResetVariable().equals(that.getResetVariable());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getName() == null ? 0 : getName().hashCode());
    result = 31 * result + (getParent() == null ? 0 : getParent().hashCode());
    result = 31 * result + (getClock() == null ? 0 : getClock().hashCode());
    result = 31 * result + (getProperties().isEmpty() ? 0 : getProperties().hashCode());
    result = 31 * result + (nested == null ? 0 : nested.hashCode());
    result = 31 * result + (processPropertyMap == null ? 0 : processPropertyMap.hashCode());
    result = 31 * result + (getResetVariable() == null ? 0 : getResetVariable().hashCode());
    return result;
  }
}
