/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.property;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * Process-level named group of properties.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class ProcessProperty {

  private final String name;

  private final ModuleProperty parentContainer;

  private final List<Node> properties;

  private final Node initValueState;

  private final List<GuardedAction> rstGuardedActions;

  /**
   * Constructs object with specified parameters.
   *
   * @param name Name of this object.
   * @param parent Parent container for this object.
   * @param initValueState The init value state node.
   * @param rstGuardedActions The sequence of resetting guarded actions.
   *
   * @throws IllegalArgumentException when name argument is {@code null}.
   */
  public ProcessProperty(
      final String name,
      final ModuleProperty parent,
      final Node initValueState,
      final List<GuardedAction> rstGuardedActions) {
    super();
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.parentContainer = parent;
    this.initValueState = initValueState;
    this.rstGuardedActions = rstGuardedActions;
    this.properties = new LinkedList<>();
  }

  /**
   * Constructs object with specified parameters.
   *
   * @param name Name of this object.
   * @param parent Parent container for this object.
   * @param node The init value state node.
   *
   * @throws IllegalArgumentException when name argument is {@code null}.
   */
  public ProcessProperty(
      final String name,
      final ModuleProperty parent,
      final Node node) {
    this(name, parent, node, null);
  }

  /**
   * Constructs object with the specified parameters.
   * @param name Object name.
   * @param parent Parent container for this object.
   */
  public ProcessProperty(final String name, final ModuleProperty parent) {
    this(name, parent, null, null);
  }

  /**
   * Returns object's name.
   *
   * @return Name of this object.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns parent container.
   *
   * @return Parent container for this object.
   */
  public ModuleProperty getParent() {
    return this.parentContainer;
  }

  /**
   * Returns node of the EFSM state that is compliant with initial values of state-like variables.
   * @return The EFSM state node that is compliant with initial values of state-like variables.
   */
  public Node getInitValueState() {
    return this.initValueState;
  }

  /**
   * Returns resetting guarded action.
   * @return The resetting guarded action.
   */
  public List<GuardedAction> getRstGuardedAction() {
    return this.rstGuardedActions;
  }

  /**
   * Adds the specified expression that uses variables
   * that are already declared at the parent container.
   *
   * @param node Assertion to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addProperty(final Node node) {
    InvariantChecks.checkNotNull(node);

    this.properties.add(node);
  }

  /**
   * Returns a collection of properties.
   *
   * @return Assertions of this object.
   */
  public List<Node> getProperties() {
    return Collections.unmodifiableList(this.properties);
  }

  /**
   * Returns the number of related properties.
   * @return The number of related properties.
   */
  public int getPropertyNum() {
    return getProperties().size();
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final ProcessProperty that = (ProcessProperty) object;

    return getName() == null
        ? that.getName() == null
            && (parentContainer == null
                ? that.parentContainer == null && getProperties().equals(that.getProperties())
                : parentContainer.equals(that.parentContainer))
        : getName().equals(that.getName());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getName() == null ? 0 : getName().hashCode());
    result = 31 * result + (parentContainer == null ? 0 : parentContainer.hashCode());
    result = 31 * result + getProperties().hashCode();
    result = 31 * result + (getInitValueState() == null ? 0 : getInitValueState().hashCode());
    result = 31 * result + (getRstGuardedAction() == null ? 0 : getRstGuardedAction().hashCode());
    return result;
  }
}
