/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.EventList;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of a test sequence. A sequence is consisted of a list of input vectors as
 * well as list of attendant events (for example clock or reset signal change).
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 * @see TestVector
 * @see EventList
 */
@XmlType(propOrder = {"vectors"}, name = "sequence")
@XmlRootElement
public final class Sequence implements Iterator<TestVector> {

  //@XmlElement(name = "vector")
  //private List<TestVector> vectors;

  private Integer index;

  private final BidiMap<TestVector, Integer> vectorMap;
  private final List<Integer> hashes;

  /**
   * Constructs an empty sequence.
   */
  public Sequence() {
    this.vectorMap = new DualHashBidiMap<>();
    this.hashes = new LinkedList<>();
    initIterator();
  }

  /**
   * Constructs the sequence that contains the specified vector only.
   * @param vector The vector that this sequence contains.
   */
  public Sequence(final TestVector vector) {
    this();
    addVector(vector);
  }

  /**
   * Adds the specified vector to the end of this sequence.
   *
   * @param vector TestVector to be added.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVector(final TestVector vector) {
    InvariantChecks.checkNotNull(vector);

    if (!this.vectorMap.containsKey(vector)) {
      this.vectorMap.put(vector, System.identityHashCode(vector));
    }
    this.hashes.add(this.vectorMap.get(vector));
  }

  /**
   * Adds the specified vectors to the end of this sequence. An original order is preserved.
   *
   * @param vectors a list of the vectors to be added
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVectors(final List<TestVector> vectors) {
    InvariantChecks.checkNotNull(vectors);

    for (final TestVector vector : vectors) {
      addVector(vector);
    }
  }

  /**
   * Returns an amount of vectors in this sequence.
   *
   * @return the amount of vectors in this sequence
   */
  public int getLength() {
    return this.hashes.size();
  }

  /**
   * Checks whether the object has no vectors..
   *
   * @return {@code true} if the object has no vectors, {@code false} otherwise.
   */
  public boolean isEmpty() {

    return getLength() == 0;
  }

  /**
   * Prepares internal iterator for vectors.
   */
  public void initIterator() {
    this.index = 0;
  }

  /**
   * Checks whether the object includes the specified one.
   * @param sequence The object for which the checking is applied.
   * @return {@code true} when the specified object is includes into the object,
   *         {@code false} otherwise.
   */
  public boolean includes(final Sequence sequence) {

    final int seqLength = sequence.getLength();
    if (seqLength > getLength()) {
      return false;
    }

    boolean includes = true;
    sequence.initIterator();
    initIterator();

    while (sequence.hasNext()) {

      final TestVector seqVector = sequence.next();
      final TestVector thisVector = next();

      if (!seqVector.equals(thisVector)) {
        includes = false;
        break;
      }
    }

    return includes;
  }

  @Override
  public boolean hasNext() {
    return this.hashes.size() > this.index;
  }

  @Override
  public TestVector next() {
    final TestVector next = this.vectorMap.getKey(this.hashes.get(this.index));
    this.index++;
    return next;
  }

  @Override
  public void remove() {
    /* Empty. */
  }

  @Override
  public String toString() {

    final StringBuilder stringBuilder = new StringBuilder("{");
    final int initLength = stringBuilder.length();

    for (final TestVector vector : this.vectorMap.keySet()) {
      if (stringBuilder.length() > initLength) {
        stringBuilder.append("; ");
      }

      stringBuilder.append("vector: ");
      stringBuilder.append(vector);
    }

    stringBuilder.append('}');

    return stringBuilder.toString();
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }

    final Sequence sequence = (Sequence) object;

    if (getLength() != sequence.getLength()) {
      return false;
    }

    final int thisIndex = this.index;
    final int sequenceIndex = sequence.index;
    initIterator();
    sequence.initIterator();

    while (hasNext() && sequence.hasNext()) {
      if (!next().equals(sequence.next())) {
        return false;
      }
    }
    this.index = thisIndex;
    sequence.index = sequenceIndex;

    return true;
  }

  @Override
  public int hashCode() {
    int result = 0;
    final int index = this.index;
    initIterator();
    while (hasNext()) {
      result = 31 * result + next().hashCode();
    }
    this.index = index;
    return result;
  }
}
