/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of a test. A test consisted of a list of input sequences.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Sequence
 */
@XmlType(name = "test")
@XmlRootElement
public class Test {

  @XmlElement(name = "testSequence")
  private final List<Sequence> sequences;

  /**
   * Constructs an empty test.
   */
  public Test() {
    this.sequences = new ArrayList<>();
  }

  /**
   * Constructs a test containing the specified list of sequences. An original order of the
   * sequences is preserved.
   *
   * @param sequences the list of sequences
   */
  public Test(final List<Sequence> sequences) {
    this.sequences = new ArrayList<>(sequences);
  }

  /**
   * Adds the specified sequence to this test. The sequence is appended to the end of this test.
   *
   * @param sequence the sequence to be added.
   */
  public void addSequence(final Sequence sequence) {
    this.sequences.add(sequence);
  }

  /**
   * Returns a list of sequences which this test consists of.
   *
   * @return the list of sequences
   */
  public List<Sequence> getSequences() {
    return new ArrayList<>(this.sequences);
  }

  /**
   * Returns the number of sequences which this test consists of.
   *
   * @return the number of sequences in this test.
   */
  public int getSequenceNum() {
    return this.sequences.size();
  }

  /**
   * Returns the number of vectors which this test consists of.
   *
   * @return the number of vectors in this test.
   */
  public int getVectorNum() {

    int vectorNum = 0;

    for (final Sequence sequence: this.sequences) {
      vectorNum = vectorNum + sequence.getLength();
    }

    return vectorNum;
  }

  /**
   * Checks whether the object contains empty sequences only.
   *
   * @return {@code true} if the object contains empty sequences only, {@code false} otherwise.
   */
  public boolean isEmpty() {

    boolean empty = true;

    for (final Sequence sequence : getSequences()) {

      if (!sequence.isEmpty()) {

        empty = false;
        break;
      }
    }

    return empty;
  }

  @Override
  public int hashCode() {

    return 31 * super.hashCode() + this.sequences.hashCode();
  }

  /**
   * Compares the specified object with this test for equality. The object is considered as equal to
   * this test if and only if it is also a test and it contains the same sequences in the same order
   * as this test does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code Test} class or its subclass;
   *    <li>it contains the same amount of sequences as this test does;
   *    <li>each its sequence is equal to the corresponding sequence in this test.
   * </ul>
   *
   * @param object the object to be compared for equality with this test
   *
   * @return {@code true} if the specified object is equal to this test. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }

    if ((object == null) || !(object instanceof Test)) {
      return false;
    }

    final Test thatTest = (Test) object;

    return this.sequences.equals(thatTest.sequences);
  }

  @Override
  public String toString() {

    final StringBuilder stringBuilder = new StringBuilder("{");

    final int initLength = stringBuilder.length();

    for (final Sequence sequence : this.sequences) {
      if (stringBuilder.length() > initLength) {
        stringBuilder.append("; ");
      }

      stringBuilder.append("sequence: ");
      stringBuilder.append(sequence);
    }

    stringBuilder.append('}');

    return stringBuilder.toString();
  }
}
