/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.TestCoverage;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * The representation of a transition-based coverage of an EFSM. In such metrics, an EFSM is
 * considered as fully covered if and only if all its transitions are covered.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmTransitionCoverage extends TestCoverage {

  private static final int PRECISION_BOUND = 100;
  private int defaultPrecision;

  private final Set<EfsmTransition> covered;
  private final Set<EfsmTransition> uncovered;

  private static final String NOT_FOUND =
      "The specified transition hasn't been found among both covered and uncovered transitions";

  /**
   * Constructs an empty coverage.
   */
  public EfsmTransitionCoverage() {
    super();

    this.uncovered = new HashSet<>();
    this.covered = new HashSet<>();

    this.computeDefaultPrecision();
  }

  /**
   * Constructs a new coverage representation for the specified EFSM.
   *
   * <p>Initially, all transitions of the EFSM are considered as uncovered.</p>
   *
   * @param efsm the EFSM for coverage monitoring
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EfsmTransitionCoverage(final Efsm efsm) {
    super();
    InvariantChecks.checkNotNull(efsm);

    this.uncovered = new HashSet<>(efsm.getTransitions());
    this.covered = new HashSet<>(getUncoveredNum());

    this.computeDefaultPrecision();
  }

  /**
   * Constructs a new coverage as a copy of the specified coverage.
   *
   * @param coverage the coverage to be copied
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EfsmTransitionCoverage(final EfsmTransitionCoverage coverage) {
    super();
    InvariantChecks.checkNotNull(coverage);

    this.uncovered = new HashSet<>(coverage.uncovered);
    this.covered = new HashSet<>(coverage.covered);

    this.defaultPrecision = coverage.defaultPrecision;
  }

  /**
   * Marks the specified transition as covered.
   *
   * @param coveredTransition the transition to be marked as covered
   *
   * @throws IllegalArgumentException if the specified transition is not found
   *         among the transitions extracted from the EFSM during initialisation
   */
  public void markTransitionAsCovered(final EfsmTransition coveredTransition) {

    if (this.uncovered.contains(coveredTransition)) {

      this.covered.add(coveredTransition);
      this.uncovered.remove(coveredTransition);

    } else if (!this.covered.contains(coveredTransition)) {

      throw new IllegalArgumentException(NOT_FOUND);
    }
  }

  /**
   * Checks if the specified transition is covered.
   *
   * @param transition the transition to be tested
   *
   * @return {@code true} if the transition is covered, otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified transition is not found
   *         among the transitions extracted from the EFSM during initialisation
   */
  public boolean isCovered(final EfsmTransition transition) {

    if (this.covered.contains(transition)) {
      return true;
    } else {
      if (this.uncovered.contains(transition)) {
        return false;
      } else {
        throw new IllegalArgumentException(NOT_FOUND);
      }
    }
  }

  /**
   * Returns a percentage representation of this test coverage computed with the default precision.
   *
   * <p>The returned value is computed as a ratio of amount of covered transitions to total amount
   * of transitions. It is guaranteed that the value is precise enough to reflect the smallest
   * possible change in a coverage information.</p>
   *
   * @return the percentage of this coverage
   */
  @Override
  public BigDecimal getPercentage() {
    return this.getPercentage(this.defaultPrecision);
  }

  /**
   * Returns a percentage representation of this test coverage computed with the specified
   * precision.
   *
   * <p>The returned value is computed as a ratio of amount of covered transitions to total
   * amount of transitions.</p>
   *
   * @param precision a number of decimal places in the result
   *
   * @return a percentage of this coverage
   */
  @Override
  public BigDecimal getPercentage(final int precision) {
    BigDecimal result;

    if (getUncovered().isEmpty() || getAllNum() == 0) {
      result = BigDecimal.valueOf(100);
    } else {

      result = new BigDecimal(getCoveredNum());

      final BigDecimal allTransNum = new BigDecimal(getCoveredNum() + getUncoveredNum());

      result = result.divide(allTransNum, precision + 3, RoundingMode.HALF_UP);
      result = result.multiply(BigDecimal.valueOf(100));
    }

    return result.setScale(precision, RoundingMode.HALF_UP);
  }

  /**
   * Tests if this coverage is full, that is, there are no uncovered transitions.
   *
   * @return {@code true} if this coverage is full. Otherwise returns {@code false}
   */
  @Override
  public boolean isFull() {
    return this.uncovered.isEmpty();
  }

  /**
   * Returns all covered transitions in this coverage.
   *
   * @return a set of covered transitions
   */
  public Set<EfsmTransition> getCovered() {
    return Collections.unmodifiableSet(this.covered);
  }

  private int getCoveredNum() {

    return getCovered().size();
  }

  /**
   * Returns all uncovered transitions in this coverage.
   *
   * @return a set of uncovered transitions
   */
  public Set<EfsmTransition> getUncovered() {
    return Collections.unmodifiableSet(this.uncovered);
  }

  private int getUncoveredNum() {

    return getUncovered().size();
  }

  private int getAllNum() {

    return getCoveredNum() + getUncoveredNum();
  }

  @Override
  public void reset() {
    this.uncovered.addAll(this.covered);
    this.covered.clear();
  }

  /**
   * Includes the specified coverage to this coverage.
   *
   * @param coverage the coverage to be included
   */
  public void include(final EfsmTransitionCoverage coverage) {

    final Set<EfsmTransition> newUncovered = new HashSet<>(coverage.uncovered);
    newUncovered.removeAll(this.covered);
    this.covered.addAll(coverage.covered);
    this.uncovered.removeAll(coverage.covered);
    this.uncovered.addAll(newUncovered);
  }

  private void computeDefaultPrecision() {

    final int transitionsAmount = getUncoveredNum();

    this.defaultPrecision =
        transitionsAmount > PRECISION_BOUND ? (int) Math.log10(transitionsAmount) - 1 : 0;
  }

  @Override
  public int hashCode() {

    int result = 1;
    result = 31 * result + this.covered.hashCode();
    result = 31 * result + this.uncovered.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this coverage for equality.
   *
   * <p>The object is considered as equal to this coverage if and only if
   * it is also an EFSM transition coverage and it contains
   * the same covered and covered transitions as this coverage does, or more formally:</p>
   * <ul>
   *   <li>it is also an instance of the {@code EfsmTransitionCoverage} class or its subclass;
   *   <li>it has the same amount of covered and uncovered transitions as this coverage does;
   *   <li>for each its covered and uncovered transition there is an equal one in this coverage.
   * </ul>
   *
   * @param object the object to be compared for equality with this coverage
   *
   * @return {@code true} if the specified object is equal to this coverage. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }

    if (!(object instanceof EfsmTransitionCoverage)) {
      return false;
    }

    final EfsmTransitionCoverage that = (EfsmTransitionCoverage) object;

    return this.covered.equals(that.covered)
        && this.uncovered.equals(that.uncovered);
  }

  @Override
  public String toString() {

    final DecimalFormat decFormat = new DecimalFormat("#.00");
    final String percentageStr = decFormat.format(getPercentage());

    return String.format(
        "%s%% (%d of %d transition%s)",
        percentageStr,
        getCoveredNum(),
        getAllNum(),
        getAllNum() == 1 ? "" : "s");
  }
}
