/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.retrascope.Retrascope;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

/**
 * This class implements logging facilities.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Log {
  // The maximum number of bytes to write to a log.
  private static final int LIMIT = 100 * 1024 * 1024;
  // The number of files to use.
  private static final int COUNT = 1;

  // The root logger.
  private static final Logger LOGGER = getLogger(Retrascope.class);

  static {
    addConsoleOutput();
    LOGGER.setUseParentHandlers(false);
  }

  private Log() {
    /* Empty. */
  }

  /**
   * Formatter of a log record.
   */
  private static class LogFormatter extends Formatter {

    @Override
    public String format(final LogRecord record) {

      final String format = "yyyy.MM.dd HH:mm:ss.SSS";
      final SimpleDateFormat dateFormat = new SimpleDateFormat(format, Locale.US);
      final String date    = dateFormat.format(new Date(record.getMillis()));
      final String level   = record.getLevel().getName();
      final String message = record.getMessage();

      if (record.getThrown() == null) {

        return String.format("%s. %s: %s%s", date, level, message, System.lineSeparator());
      } else {

        final StringWriter stackTraceWriter = new StringWriter();
        record.getThrown().printStackTrace(new PrintWriter(stackTraceWriter));

        return String.format("%s. %s: %s: %s%s", date, level, message,
            stackTraceWriter.toString(), System.lineSeparator());
      }
    }
  }

  /**
   * Returns the root logger.
   *
   * @return Root logger.
   */
  public static Logger getLogger() {
    return LOGGER;
  }

  /**
   * Returns the component logger.
   *
   * @param component Component class.
   * @return Component logger.
   */
  public static Logger getLogger(final Class<?> component) {
    return Logger.getLogger(component.getPackage().getName());
  }

  /**
   * Prints the specified objects in the specified format at the debug logging level.
   *
   * @see LogLevel
   *
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void debug(final String format, final Object ... args) {
    print(LogLevel.DEBUG, args.length == 0 ? format : String.format(format, args));
  }

  /**
   * Prints the specified objects in the specified format at the error logging level.
   *
   * @see LogLevel
   *
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void error(final String format, final Object ... args) {
    print(LogLevel.ERROR, args.length == 0 ? format : String.format(format, args));
  }

  /**
   * Prints the specified objects in the specified format at the info logging level.
   *
   * @see LogLevel
   *
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void info(final String format, final Object ... args) {
    print(LogLevel.INFO, args.length == 0 ? format : String.format(format, args));
  }

  /**
   * Prints the specified objects in the specified format at the warning logging level.
   *
   * @see LogLevel
   *
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void warn(final String format, final Object ... args) {
    print(LogLevel.WARNING, args.length == 0 ? format : String.format(format, args));
  }

  /**
   * Prints the message with the specified logging level.
   *
   * @param logLevel Logging level at which message is printed.
   * @param message Message to be printed.
   */
  private static void print(final Level logLevel, final String message) {

    if (LOGGER.isLoggable(logLevel)) {
      LOGGER.log(logLevel, message);
    }
  }

  /**
   * Sets the log level.
   *
   * @param level Log level to be set.
   */
  public static void setLevel(final Level level) {

    LOGGER.setLevel(level);

    for (final Handler handler : LOGGER.getHandlers()) {
      handler.setLevel(level);
    }
  }

  /**
   * Adds the console output to the log.
   */
  private static void addConsoleOutput() {
    addHandler(new ConsoleHandler());
  }

  /**
   * Adds the file output to the log.
   *
   * @param fileName Name of the file to log to.
   */
  public static void addFileOutput(final String fileName) {

    try {

      addHandler(new FileHandler(fileName, LIMIT, COUNT, true));

    } catch (final IOException | SecurityException e) {
      LOGGER.log(LogLevel.ERROR, "The exception has been encountered", e);
    }
  }

  /**
   * Adds the output handler to the log.
   *
   * @param handler Handler to be added.
   */
  private static void addHandler(final Handler handler) {

    handler.setFormatter(new LogFormatter());
    LOGGER.addHandler(handler);
  }

  /**
   * Closes all file handlers which are associated with the Retrascope root logger.
   */
  public static void closeFileHandlers() {

    for (final Handler handler: LOGGER.getHandlers()) {

      if (handler instanceof FileHandler) {
        handler.flush();
        handler.close();
      }
    }
  }
}
