/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeException;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for {@code Node} objects.
 *
 * @see Node
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class NodeUtils {

  private NodeUtils() {
    /* Empty. */
  }

  /**
   * Returns object of the specified type that contains the specified value.
   * @param value The value of the object.
   * @param type The type of the object.
   * @return The node of the specified type that contains the specified value.
   */
  public static Node getValue(final int value, final DataType type) {

    final DataTypeId typeId = type.getTypeId();

    switch (typeId) {

      case BIT_VECTOR:
        final int size = type.getSize();
        return NodeValue.newBitVector(value, size);

      case LOGIC_REAL:
        return NodeValue.newReal(value);

      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(value != 0);

      case LOGIC_INTEGER:
        return NodeValue.newInteger(value);

      default:
        throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE + typeId.toString());
    }
  }

  /**
   * Returns a collection of names for variables from the specified collection.
   *
   * @param nodes Collection of nodes.
   *
   * @return Collection of names for variable nodes from the specified collection.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Collection<String> getVariableNames(final Collection<? extends Node> nodes) {
    InvariantChecks.checkNotNull(nodes);

    final Collection<String> names = new LinkedList<>();

    for (final Node node : nodes) {

      if (ExprUtils.isKind(Node.Kind.VARIABLE, node)) {
        names.add(((NodeVariable)node).getName());
      }
    }

    return names;
  }

  /**
   * Casts the specified node to the specified data type.
   *
   * @param node Expression to be casted.
   * @param dataType Type to which the expression to be casted to.
   *
   * @return An expression which is equivalent to the specified node and has the specified type.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  static Node castToDataType(final Node node, final DataType dataType) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(dataType);

    if (node.getDataType() == dataType) {
      return node;
    }

    switch (dataType.getTypeId()) {

      case BIT_VECTOR:
        return castToBitVector(node, dataType.getSize());

      case LOGIC_BOOLEAN:
        return castToBool(node);

      case LOGIC_INTEGER:
        return castToInteger(node);

      default:
        unsupported(dataType.getTypeId());
    }
    return null;
  }

  private static Node castToInteger(final Node node) {

    final Node zeroValue = NodeValue.newInteger(0);
    final Node oneValue = NodeValue.newInteger(1);

    switch (node.getDataTypeId()) {

      case BIT_VECTOR:
        final int bvSize = node.getDataType().getSize();
        final BitVector oneBitVector = BitVector.valueOf(1, bvSize);
        Node intExpression = zeroValue;
        BitVector current = oneBitVector;

        for (int i = 1; i < Math.pow(2, bvSize); i++) {

          final Node newInt = NodeValue.newInteger(i);
          intExpression =
              Nodes.ite(Nodes.eq(node, NodeValue.newBitVector(current)), newInt, intExpression);
          current = BitVectorMath.add(current, oneBitVector);
        }
        return intExpression;

      case LOGIC_BOOLEAN:
        return Nodes.ite(Nodes.eq(node, Nodes.FALSE), zeroValue, oneValue);

      default:
        unsupported(node);
    }
    return null;
  }

  private static Node castToBool(final Node node) {

    switch (node.getDataTypeId()) {

      case BIT_VECTOR:

        final Node zeroBitVector = NodeValue.newBitVector(0, node.getDataType().getSize());

        return Nodes.ite(Nodes.eq(node, zeroBitVector), Nodes.FALSE, Nodes.TRUE);

      case LOGIC_INTEGER:
        final Node zeroInt = NodeValue.newInteger(0);
        return Nodes.ite(Nodes.eq(node, zeroInt), Nodes.FALSE, Nodes.TRUE);

      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static Node castToBitVector(
      final Node node,
      final int size) {

    final Node trueValue = NodeValue.newBitVector(BitVector.TRUE);
    final Node falseValue = NodeValue.newBitVector(BitVector.FALSE);

    switch (node.getDataTypeId()) {

      case LOGIC_BOOLEAN:
        return Nodes.ite(Nodes.eq(node, Nodes.FALSE), falseValue, trueValue);

      case LOGIC_INTEGER:
        Node bitVector = NodeValue.newBitVector(0, size);
        for (int i = 1; i < Math.pow(2, size); i++) {
          final Node intValue = NodeValue.newInteger(i);
          final Node bvValue = NodeValue.newBitVector(i, size);
          bitVector = Nodes.ite(Nodes.eq(node, intValue), bvValue, bitVector);
        }
        return bitVector;

      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static void unsupported(final Object obj) {
    throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, obj);
  }

  private static boolean dependExactlyOnVariables(
      final Collection<NodeVariable> exprVariables,
      final Collection<NodeVariable> variables) {

    if (exprVariables.isEmpty()) {
      return false;
    } else {
      exprVariables.removeAll(variables);
      return exprVariables.isEmpty();
    }
  }

  /**
   * Splits the specified node into two parts - one that depends exactly from the specified
   * variables and another that does not depend.
   * @param node Node to be split.
   * @param variables Variables dependence to which is a splitting criterion.
   * @return Two nodes that being united (with AND) form a node that is equal to the specified one.
   *         The first one depends exactly from the specified variables the second does not depend
   *         from all of them.
   */
  public static Pair<Node, Node> splitByVar(
      final Node node,
      final Collection<NodeVariable> variables) {

    class DepSplitVisitor extends ExprTreeVisitorDefault {
      private Collection<Node> dependOnVar;
      private Collection<Node> freeFromVar;
      private Collection<NodeVariable> variables;

      private DepSplitVisitor(final Collection<NodeVariable> variables,
          final Collection<Node> dependNodes, final Collection<Node> freeNodes) {
        super();
        InvariantChecks.checkNotNull(variables);

        this.variables = variables;
        this.dependOnVar = dependNodes;
        this.freeFromVar = freeNodes;
      }

      @Override
      public void onOperationBegin(final NodeOperation node) {
        if (!ExprUtils.isOperation(node, StandardOperation.AND)) {

          if (dependExactlyOnVariables(ExprUtils.getVariables(node), this.variables)) {
            this.dependOnVar.add(node);
          } else {
            this.freeFromVar.add(node);
          }

          setStatus(Status.SKIP);
        }
      }

      @Override
      public void onOperationEnd(final NodeOperation node) {
        if (getStatus() == Status.SKIP) {
          setStatus(Status.OK);
        }
      }
    }

    final List<Node> varDepends = new LinkedList<>();
    final List<Node> varFreeNodes = new LinkedList<>();

    final DepSplitVisitor visitor = new DepSplitVisitor(variables, varDepends, varFreeNodes);
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);
    walker.visit(node);

    return new ImmutablePair<>(
        varDepends.isEmpty() ? null : Nodes.and(varDepends),
        varFreeNodes.isEmpty() ? null : Nodes.and(varFreeNodes));
  }

  /**
   * Splits the specified array of nodes (conjunctions, probably) into the collection
   * of collections of nodes that depend on same variables.
   * @param nodes The nodes to be grouped by variable from that they are depend.
   * @return The collection of collections of nodes, such that every collection contains
   *         only the nodes that depend on common set of variables.
   */
  static Collection<Collection<Node>> splitByVar(final Node... nodes) {

    final Map<Set<NodeVariable>, Collection<Node>> varNodeMap = new HashMap<>();

    for (final Node node : nodes) {

      final Collection<Node> elementaryOps = splitConjunction(node);

      for (final Node element : elementaryOps) {

        final Set<NodeVariable> vars = new HashSet<>(ExprUtils.getVariables(element));

        final Collection<Set<NodeVariable>> intersected =
            getIntersected(varNodeMap.keySet(), vars);

        if (intersected.isEmpty()) {

          final Collection<Node> useNodes = new HashSet<>();
          useNodes.add(element);
          varNodeMap.put(vars, useNodes);

        } else if (intersected.size() == 1) {

          varNodeMap.get(intersected.iterator().next()).add(element);
        } else {

          final Set<NodeVariable> union = new HashSet<>();
          final Collection<Node> useNodes = new HashSet<>();

          for (final Set<NodeVariable> key : intersected) {
            union.addAll(key);
            useNodes.addAll(varNodeMap.get(key));
            varNodeMap.remove(key);
          }

          union.addAll(vars);
          useNodes.add(element);
          varNodeMap.put(union, useNodes);
        }
      }
    }

    return new HashSet<>(varNodeMap.values());
  }

  private static Collection<Set<NodeVariable>> getIntersected(
      final Set<Set<NodeVariable>> collections,
      final Set<NodeVariable> vars) {

    final Collection<Set<NodeVariable>> intersected = new HashSet<>();

    for (final Set<NodeVariable> collection : collections) {

      if (CollectionUtils.areIntersectedSets(collection, vars)) {
        intersected.add(collection);
      }
    }

    return intersected;
  }

  /**
   * Splits conjunction node into non-conjunction nodes.
   * @param node Node to be split.
   * @return Collection of nodes that being united by AND operation (a conjunction)
   *         forms an expression that is equivalent to the specified one.
   */
  private static Collection<Node> splitConjunction(final Node node) {

    class ConjunctionSplitVisitor extends ExprTreeVisitorDefault {

      private Collection<Node> nodes;

      private ConjunctionSplitVisitor() {
        this.nodes = new LinkedHashSet<>();
      }

      Collection<Node> getNodes() {
        return this.nodes;
      }

      @Override
      public void onOperationBegin(final NodeOperation operation) {
        elaborate(operation);
      }

      private void elaborate(final NodeOperation operation) {

        if (ExprUtils.isOperation(operation, StandardOperation.AND)) {

          for (int i = 0; i < operation.getOperandCount(); i++) {

            final Node operand = operation.getOperand(i);

            if (ExprUtils.isKind(Node.Kind.OPERATION, operand)) {

              elaborate((NodeOperation) operand);

            } else {
              this.nodes.add(operand);
            }
          }
        } else {
          this.nodes.add(operation);
          setStatus(Status.SKIP);
        }
      }

      @Override
      public void onOperationEnd(final NodeOperation node) {

        if (getStatus() == Status.SKIP) {
          setStatus(Status.OK);
        }
      }
    }

    InvariantChecks.checkNotNull(node);

    final ConjunctionSplitVisitor visitor = new ConjunctionSplitVisitor();
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);
    walker.visit(node);
    return visitor.getNodes();
  }

  /**
   * Substitutes variables that are used in the node by the specified bound variables.
   *
   * @param node The node to be performed.
   * @param boundVariables The substitutions.
   * @return The performed node, some of that variables are substituted by bound ones.
   */
  public static Node substituteAllBindings(
      final Node node,
      final List<NodeBinding.BoundVariable> boundVariables) {

    return Transformer.substituteAllBindings(new NodeBinding(node, boundVariables));
  }

  /**
   * Returns a "greater-or-equal" operation for the specified operands of the same type.
   * @param first The operand that is supposed to be greater or equal
   * @param second The operand that is supposed to be less
   * @return The "greater-or-equal" operation for the specified operands if they're of the same
   *         type, {@code false} otherwise.
   */
  public static Node greater(final Node first, final Node second) {

    if (ExprUtils.isType(DataType.INTEGER, first, second)) {

      return Nodes.greatereq(first, second);

    } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, first, second)) {

      return Nodes.bvuge(first, second);
    }

    return null;
  }

  /**
   * Checks that one node is a constant, but another is not for the two specified.
   * @param oneNode The first node.
   * @param anotherNode The second node.
   * @return {@code true} if in the specified pair of nodes the one is a constant and another is
   *         not, {@code false} otherwise.
   */
  public static boolean oneConstAnotherNot(final Node oneNode, final Node anotherNode) {
    return (oneNode.getKind() == Node.Kind.VALUE
        && (anotherNode.getKind() == Node.Kind.OPERATION
            || anotherNode.getKind() == Node.Kind.VARIABLE)
        || (anotherNode.getKind() == Node.Kind.VALUE
        && (oneNode.getKind() == Node.Kind.OPERATION
            || oneNode.getKind() == Node.Kind.VARIABLE)));
  }
}
