/*
 * Copyright 2015-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprTreeVisitor;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.SmtKeyword;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.exception.RetrascopeException;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Embedded (project-level) satisfiability (SAT) solver.
 *
 * <p>The internal solver calls external one for every constraint it is unable to solve.
 * It is able to solve constraints of the following kinds:</p>
 * <ul>
 * <li>{@code true}
 * <li>{@code false}
 * <li>single boolean variable or binding
 * <li>a conjunction that include "x AND !x" sub-expression
 * </ul>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class SatSolver {

  private static SatSolver instance;
  private static final Map<Node, SolverResult> SOLVED = new LinkedHashMap<>();

  private SatSolver() {
    /* Empty. */
  }

  /**
   * Returns an instance of the satisfiability solver.
   *
   * @return Solver instance.
   */
  public static synchronized SatSolver getInstance() {
    if (instance == null) {
      instance = new SatSolver();
    }
    return instance;
  }

  /**
   * Clear solver's internal data.
   */
  public void clear() {
    SOLVED.clear();
  }

  /**
   * Solves the specified node constraint.
   *
   * @param node Constraint to be solved.
   *
   * @return Result of solving the constraint.
   */
  SolverResult solve(final Node node) {

    if (SOLVED.containsKey(node)) {
      return SOLVED.get(node);
    } else {

      SolverResult result = null;
      try {
        if (canBeSolved(node)) {

          switch (node.getKind()) {

            case VALUE:
              final boolean isSatValue = isSatValue((NodeValue) node);

              final SolverResult.Status status =
                  isSatValue ? SolverResult.Status.SAT : SolverResult.Status.UNSAT;

              result = createResult(status, new LinkedList<>());
              break;

            case VARIABLE:
              final Variable variable =
                  new Variable(((NodeVariable) node).getName(), Data.newBoolean(true));
              result = createResult(SolverResult.Status.SAT, Collections.singletonList(variable));
              break;

            case OPERATION:
              /* Solve controversial operations only that are unsatisfiable. */
              result = createResult(SolverResult.Status.UNSAT, new LinkedList<>());
              break;

            default:
              throw RetrascopeException.exception(LogMessage.UNSUPPORTED_TYPE, node.getKind());
          }
        } else {

          /* Check node for SMT-LIB keywords, if appear, substitute them. */
          final List<NodeBinding.BoundVariable> keywordBindings = substituteKeywords(node);
          final Constraint constraint =
              ConstraintUtils.newConstraint(
                  keywordBindings.isEmpty()
                      ? node
                      : TransformerUtils.substituteBindings(node, keywordBindings));

          final SolverResult solved = ConstraintUtils.solve(constraint);
          result = keywordBindings.isEmpty() ? solved : applyBindings(solved, keywordBindings);
        }
      } catch (final Exception e) {

        Log.error("While solving the constraint:%n%s%n", node.toString());
        Log.error(e.getMessage());
        e.printStackTrace();
      }

      SOLVED.put(node, result);
      return result;
    }
  }

  private SolverResult applyBindings(
      final SolverResult result,
      final List<NodeBinding.BoundVariable> list) {

    final Map<Variable, Variable> map = createBindMap(list);
    final List<Variable> variables = new LinkedList<>();

    for (final Variable variable : result.getVariables()) {

      if (map.containsKey(variable)) {

        final Variable oldVar = map.get(variable);
        oldVar.setData(variable.getData());
        variables.add(oldVar);
      } else {
        variables.add(variable);
      }
    }

    return new SolverResult(result.getStatus(), result.getErrors(), variables);
  }

  private Map<Variable, Variable> createBindMap(final List<NodeBinding.BoundVariable> list) {

    final Map<Variable, Variable> map = new LinkedHashMap<>();

    for (final NodeBinding.BoundVariable boundVariable : list) {

      final Variable target = boundVariable.getVariable().getVariable();
      final Node value = boundVariable.getValue();

      if (value instanceof NodeVariable) {

        final Variable valueVar = ((NodeVariable) value).getVariable();
        map.put(valueVar, target);
      }
    }

    return map;
  }

  private List<NodeBinding.BoundVariable> substituteKeywords(final Node node) {

    final Collection<NodeBinding.BoundVariable> bindings = new LinkedHashSet<>();

    final ExprTreeVisitor keyWordFinder = new ExprTreeVisitor() {

      @Override
      public void onOperationBegin(final NodeOperation nodeOperation) {
        /* Empty. */
      }

      @Override
      public void onOperationEnd(final NodeOperation nodeOperation) {
        /* Empty. */
      }

      @Override
      public int[] getOperandOrder() {
        return null;
      }

      @Override
      public void onOperandBegin(
          final NodeOperation nodeOperation,
          final Node node,
          final int index) {
        /* Empty. */
      }

      @Override
      public void onOperandEnd(
          final NodeOperation nodeOperation,
          final Node node,
          final int index) {
        /* Empty. */
      }

      @Override
      public void onValue(final NodeValue nodeValue) {
        /* Empty. */
      }

      @Override
      public void onVariable(final NodeVariable nodeVariable) {

        final String varName = nodeVariable.getName();

        if (SmtKeyword.isKeyword(varName)) {

          final String name = String.format("_%s_", varName);
          final NodeVariable nonKeyword = new NodeVariable(name, nodeVariable.getDataType());
          nonKeyword.setData(nodeVariable.getData());

          bindings.add(NodeBinding.bindVariable(nodeVariable, nonKeyword));
        }
      }

      @Override
      public void onBindingBegin(final NodeBinding nodeBinding) {
        /* Empty. */
      }

      @Override
      public void onBindingListEnd(final NodeBinding nodeBinding) {
        /* Empty. */
      }

      @Override
      public void onBindingEnd(final NodeBinding nodeBinding) {
        /* Empty. */
      }

      @Override
      public void onBoundVariableBegin(
          final NodeBinding nodeBinding,
          final NodeVariable nodeVariable,
          final Node node) {
        /* Empty. */
      }

      @Override
      public void onBoundVariableEnd(
          final NodeBinding nodeBinding,
          final NodeVariable nodeVariable,
          final Node node) {
        /* Empty. */
      }

      @Override
      public Status getStatus() {
        return Status.OK;
      }

      @Override
      public void onBegin() {
        /* Empty. */
      }

      @Override
      public void onEnd() {
        /* Empty. */
      }
    };
    final ExprTreeWalker walker = new ExprTreeWalker(keyWordFinder);
    walker.visit(node);
    return new LinkedList<>(bindings);
  }

  /**
   * Checks whether the specified expression is satisfiable.
   *
   * @param expr Expression to be checked.
   * @return {@code true} if the expression is satisfiable, {@code false} otherwise.
   *
   */
  public boolean isSat(final Node expr) {

    final Node prepared = TypeConversion.castConstants(expr);

    if (SOLVED.containsKey(prepared)) {

      return SOLVED.get(prepared).getStatus() == SolverResult.Status.SAT;

    } else {

      final SolverResult result = solve(prepared);
      SOLVED.put(prepared, result);
      final List<String> errors = result.getErrors();

      if (!errors.isEmpty() && result.getStatus().equals(SolverResult.Status.ERROR)) {
        throw new IllegalArgumentException(
            "Constraint contains errors: " + prepared + "; errors=" + errors);
      }
      return result.getStatus() == SolverResult.Status.SAT;
    }
  }

  /**
   * Checks whether the specified logical expressions are compatible
   * {@code (getConjunction(nodes[0], ..., nodes[n-1]) is SAT)}.
   *
   * @param expressions Logical expressions to be checked.
   *
   * @return {@code true} if the specified logical expressions are compatible,
   *         {@code false} otherwise.
   */
  boolean areCompatible(final Node... expressions) {

    final Collection<Collection<Node>> splitByVarNodes = NodeUtils.splitByVar(expressions);

    if (splitByVarNodes.size() <= 1) {

      return isSat(ExprUtils.getConjunction(expressions));

    } else {

      for (final Collection<Node> sameVarNodes : splitByVarNodes) {

        final Node sameVarAndCond = Nodes.and(sameVarNodes);

        if (!isSat(sameVarAndCond)) {
          return false;
        }
      }
      return true;
    }
  }

  /**
   * Check whether the specified logical expressions are complete
   * {@code (!getComplement(nodes[0], ..., nodes[n-1]) is SAT)}.
   *
   * @param expressions Logical expressions to be checked.
   *
   * @return {@code true} if the specified logical expressions are complete,
   *         {@code false} otherwise.
   */
  boolean areComplete(final Node... expressions) {

    return !isSat(ExprUtils.getComplement(expressions));
  }

  private static boolean canBeSolved(final Node node) {

    if (ExprUtils.isValue(node) || ExprUtils.isVariable(node)) {
      return node.isType(DataType.BOOLEAN);
    }
    return ExprUtils.isOperation(node) && isControversial((NodeOperation) node);
  }

  private static boolean isSatValue(final NodeValue expr) {

    boolean isSatValue;
    if (expr.isType(DataType.BOOLEAN)) {

      if (expr.equals(Nodes.TRUE)) {

        isSatValue = true;

      } else if (expr.equals(Nodes.FALSE)) {

        isSatValue = false;

      } else {
        throw new RetrascopeException(
            new IllegalArgumentException("Incorrect boolean argument: " + expr.toString()));
      }
    } else {
      isSatValue = ExprUtils.isSat(expr);
    }
    return isSatValue;
  }

  private static boolean isControversial(final NodeOperation expr) {

    boolean controversial = false;

    if (ExprUtils.isOperation(expr, StandardOperation.AND)) {

      final Map<Node, Boolean> operandSignMap = new HashMap<>();

      for (int i = 0; i < expr.getOperandCount(); i++) {

        final Node operand = expr.getOperand(i);

        /* Check either object contains at least one controversial operand. */
        final boolean hasControversial =
            operand.getKind() == Node.Kind.OPERATION && isControversial((NodeOperation) operand);

        if (hasControversial) {

          controversial = true;
          break;
        }

        /* Mark as controversial if "x AND !x" is found in the expression. */
        if (containDifferentSignClauses(operandSignMap, operand)) {

          controversial = true;
          break;
        }
      }
    }
    return controversial;
  }

  private static boolean containDifferentSignClauses(
      final Map<Node, Boolean> map,
      final Node operand) {

    boolean result;

    final boolean isNotOperation = ExprUtils.isOperation(operand, StandardOperation.NOT);
    final Node clause = getClause(operand);

    result = map.containsKey(clause) && map.get(clause) != isNotOperation;

    if (!result) {
      map.put(clause, isNotOperation);
    }
    return result;
  }

  private static Node getClause(final Node operand) {

    final boolean isNotClause = ExprUtils.isOperation(operand, StandardOperation.NOT);

    return isNotClause ? ((NodeOperation)operand).getOperand(0) : operand;
  }

  private static SolverResult createResult(
      final SolverResult.Status status,
      final List<Variable> variables) {

    return new SolverResult(status, new LinkedList<>(), variables);
  }
}
